/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.vecmath.*;

import icc3dpackage.managers.*;


/**
 * The ApplicationLauncher provides the user with the possibility of modifying
 * Java VM options.
 */
public class ApplicationLauncher extends JFrame {

	/** The file used when writing and reading launcher options. */
	public static final String iniFile = "icc3d.ini";
	
	/** The absolute file name of the java executable. */
	protected static String javaExeCommand;
	
	/** The ICC3D class path. */
	protected static String javaClassPath;
	
	/** Deprecated. May be useful later. */
	private boolean skip;
	
	/** The initial memory heap size. */
	private int initialHeapSize;
	
	/** The maximum memory heap size. */
	private int maximumHeapSize;
	
	/** The initial memory heap size choices. */
	private JComboBox initCombo;
	
	/** The maximum memory heap size choices. */
	private JComboBox maxCombo;
	
	/** The Java executable. */
	private JButton javaButton;
	
	/** The Java class path. */
	private JTextArea javaClassPathArea;
	
	/** Disable the configuration dialog. */
	private JCheckBox neverAsk;
	
	/**
	 * The main function that is called when the ApplicationLauncher is started.
	 * @param args The command line arguments specified by the user.
	 */
	public static void main(String args[]) {
		javaExeCommand = System.getProperty("java.home") + File.separator + "bin"
				+ File.separator + "java";
		String fx = System.getProperty("java.home").substring(0,
				System.getProperty("java.home").length() - 3)
				+ "bin";
		File file = new File(fx);

		if (file.exists() && file.isDirectory()) {
			String bins[] = file.list();
			boolean found = false;

			for (int i = 0; i < bins.length; i++) {
				if (bins[i].indexOf("java") >= 0) {
					found = true;
				}
			}
			if (found) {
				javaExeCommand = fx + File.separator + "java";
			}
		}
		ApplicationLauncher launcher = new ApplicationLauncher();
	}
	
	/**
	 * Constructs am ApplicationLauncher.
	 */
	public ApplicationLauncher() {
		super("Configure application options");
		setIconImage(ICC3D.image);
		skip = false;
		setupGUI();
		getData();
	}
	
	/**
	 * Creates the dialog GUI.
	 */
	public void setupGUI() {
		initCombo = new JComboBox(new String[] {"32", "64", "128", "192", "256", "378",
			"512"});
		initCombo.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				skip = false;
			}
		});
		maxCombo = new JComboBox(new String[] {"32", "64", "128", "192", "256", "378",
			"512"});
		maxCombo.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				skip = false;
			}
		});
		JButton startButton = new JButton("Start ICC3D");

		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (initCombo.getSelectedIndex() > maxCombo.getSelectedIndex()) {
					SystemManager.displayError("Warning", "Initial heap size cannot exceed the maximum value");
				} else {
					if (neverAsk.isSelected()) {
						SystemManager.displayError("Information", "If you wish to see this dialog again, please delete\n"
								+ new File(iniFile).getAbsolutePath());
					}
					initialHeapSize = Integer.parseInt(initCombo.getSelectedItem().toString());
					maximumHeapSize = Integer.parseInt(maxCombo.getSelectedItem().toString());
					startICC3D(true);
				}
			}
		});
		javaButton = new JButton(javaExeCommand);
		javaButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser chooser = new JFileChooser(javaButton.getText().substring(0,
						javaButton.getText().lastIndexOf(File.separator)));
				int returnVal = chooser.showOpenDialog(ICC3D.getDialogParent());

				if (returnVal == JFileChooser.APPROVE_OPTION
						&& chooser.getSelectedFile() != null
						&& chooser.getSelectedFile().getName() != null
						&& !chooser.getSelectedFile().getName().equals("")) {
					javaExeCommand = chooser.getSelectedFile().getAbsolutePath();
					javaButton.setText(javaExeCommand);
				}
			}
		});
		javaClassPath = System.getProperty("java.class.path", ".");
		javaClassPathArea = new JTextArea(javaClassPath, 8, 20);
		javaClassPathArea.setLineWrap(true);
		neverAsk = new JCheckBox();
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		Container c = getContentPane();
		
		JPanel gplPanel = new JPanel(new GridLayout(5, 1));

		gplPanel.add(new JLabel("ICC3D version " + ICC3D.getVersion()
				+ ", Copyright (C) 2002-2003 Gjvik University College."));
		gplPanel.add(new JLabel("ICC3D comes with ABSOLUTELY NO WARRANTY. This is free"));
		gplPanel.add(new JLabel("software, and you are welcome to redistribute it under the terms"));
		gplPanel.add(new JLabel("specified in the GNU General Public License."));
		gplPanel.add(new JLabel(" "));
		
		JPanel p = new JPanel(new GridLayout(5, 2));

		p.add(new JLabel("Initial java heap size (MB):"));
		p.add(initCombo);
		p.add(new JLabel("Maximum java heap size (MB):"));
		p.add(maxCombo);
		p.add(new JLabel("Java executable:"));
		p.add(javaButton);
		p.add(new JLabel("Never ask me this again?"));
		p.add(neverAsk);
		c.setLayout(new BorderLayout());
		JPanel classPPanel = new JPanel(new GridLayout(1, 2));

		classPPanel.add(new JLabel("Java class path:"));
		classPPanel.add(new JScrollPane(javaClassPathArea));
		JPanel upperPanel = new JPanel(new BorderLayout());

		upperPanel.add(gplPanel, BorderLayout.NORTH);
		upperPanel.add(classPPanel, BorderLayout.CENTER);
		JPanel containerPanel = new JPanel(new BorderLayout());

		containerPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		containerPanel.add(upperPanel, BorderLayout.NORTH);
		containerPanel.add(p, BorderLayout.CENTER);
		containerPanel.add(startButton, BorderLayout.SOUTH);
		c.add(containerPanel, BorderLayout.CENTER);
		setSize(500, 400);
		//pack();
		setResizable(false);
		setLocation(Toolkit.getDefaultToolkit().getScreenSize().width / 2
				- (getWidth() / 2),
				Toolkit.getDefaultToolkit().getScreenSize().height / 2 - (getHeight() / 2));
	}
	
	/**
	 * Tries to read previous settings from the ini file.
	 */
	private void getData() {
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(iniFile));
			String initial = br.readLine();

			initialHeapSize = Integer.parseInt(initial);
			String maximum = br.readLine();

			maximumHeapSize = Integer.parseInt(maximum);
			javaExeCommand = br.readLine();
			javaButton.setText(javaExeCommand);
			javaClassPath = br.readLine();
			javaClassPathArea.setText(javaClassPath);
			neverAsk.setSelected(new Boolean(br.readLine()).booleanValue());
			skip = true;
			if (neverAsk.isSelected()) {
				startICC3D(false);
			}

			/* Timer timer = new Timer(5000, new ActionListener() {
			 public void actionPerformed(ActionEvent ae) {
			 if (skip) {
			 initialHeapSize = Integer.parseInt(initCombo.getSelectedItem().toString());
			 maximumHeapSize = Integer.parseInt(maxCombo.getSelectedItem().toString());
			 startICC3D();
			 }
			 }
			 });
			 timer.setRepeats(false);
			 timer.start();
			 */
		} catch (Exception e) {
			initialHeapSize = 128;
			maximumHeapSize = 256;
			javaExeCommand = "";
			javaClassPath = "";
		}
		if (javaExeCommand == null || javaExeCommand.equals("")) {
			javaExeCommand = System.getProperty("java.home") + File.separator + "bin"
					+ File.separator + "java";
			String fx = System.getProperty("java.home").substring(0,
					System.getProperty("java.home").length() - 3)
					+ "bin";
			File file = new File(fx);

			if (file.exists() && file.isDirectory()) {
				String bins[] = file.list();
				boolean found = false;

				for (int i = 0; i < bins.length; i++) {
					if (bins[i].indexOf("java") >= 0) {
						found = true;
					}
				}
				if (found) {
					javaExeCommand = fx + File.separator + "java";
				}
			}
			javaButton.setText(javaExeCommand);
			javaClassPath = System.getProperty("java.class.path", ".");
			javaClassPathArea.setText(javaClassPath);
		}
		boolean localSkip = skip;

		for (int i = 0; i < initCombo.getItemCount(); i++) {
			if (Integer.parseInt(initCombo.getItemAt(i).toString()) == initialHeapSize) {
				initCombo.setSelectedIndex(i);
			}
			if (Integer.parseInt(maxCombo.getItemAt(i).toString()) == maximumHeapSize) {
				maxCombo.setSelectedIndex(i);
			}
		}
		skip = localSkip;
		setVisible(true);
	}
	
	/**
	 * Initiates application startup.
	 * @param writeData Indicates whether the current options should be written to
	 * the ini file.
	 */
	private void startICC3D(boolean writeData) {
		if (writeData) {
			try {
				BufferedWriter bw = new BufferedWriter(new java.io.FileWriter(iniFile));

				bw.write(initCombo.getSelectedItem().toString());
				bw.newLine();
				bw.write(maxCombo.getSelectedItem().toString());
				bw.newLine();
				bw.write(javaButton.getText());
				bw.newLine();
				bw.write(javaClassPathArea.getText());
				bw.newLine();
				bw.write("" + neverAsk.isSelected());
				bw.newLine();
				bw.flush();
				bw.close();
			} catch (Exception e) {}
		}
		startApplication("\"" + javaExeCommand + "\"" + " -cp \""
				+ javaClassPathArea.getText() + "\" -Xms" + (initialHeapSize * 1000000)
				+ " -Xmx" + (maximumHeapSize * 1000000) + " icc3dpackage.ICC3D");
	}
	
	/**
	 * Attempts to start the main application.
	 * @param startCommand The shell command necessary to run the application.
	 */
	public void startApplication(String startCommand) {
		setVisible(false);
		System.out.println(startCommand);
		String os = System.getProperty("os.name");

		if (os.indexOf("Windows") >= 0) {
			// System.out.println("is windows");
			try {
				File f = new File("running.log");

				try {
					f.delete();
				} catch (Exception e) {}
				Runtime.getRuntime().exec(startCommand);
				for (int i = 0; i < 60; i++) {
					SystemManager.sleep(1000);
					if (f.exists()) {
						break;
					}
				}
				
				if (!f.exists()) {
					SystemManager.displayError(
						"Error",
						"The application failed to start.\nTry to manually execute \"java icc3dpackage.ICC3D\""
					);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			// System.out.println("isn't windows");
			try {
				File f = new File("running.log");

				try {
					f.delete();
				} catch (Exception e) {}
				BufferedWriter bw = new BufferedWriter(new java.io.FileWriter("shellscript"));

				bw.write("#!/bin/sh");
				bw.newLine();
				bw.write(startCommand);
				bw.newLine();
				bw.flush();
				bw.close();
				Runtime.getRuntime().exec(new String[] {"/bin/sh", "shellscript"});
				for (int i = 0; i < 60; i++) {
					SystemManager.sleep(1000);
					if (f.exists()) {
						break;
					}
				}
				if (!f.exists()) {
					SystemManager.displayError(
						"Error",
						"The application failed to start.\nTry to manually execute \"java icc3dpackage.ICC3D\""
					);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		System.exit(0);
	}
}
