/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage;


import java.awt.*;
import java.awt.event.*;
import javax.media.j3d.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import java.awt.color.*;

import icc3dpackage.image2d.Image2D;
import icc3dpackage.managers.*;
import icc3dpackage.modules.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.View3D;


/**
 * ICC3D
 * <p>ICC3D - Interactive Color Correction in 3 Dimensions.
 * <p>ICC3D is an application designed for visualization and manipulation of
 * color gamuts.
 * 
 */
public class ICC3D extends JFrame implements ActionListener {
	
	/** The ICC3D image icon. */
	public static final Image image = new ImageIcon("icons" + File.separator + "icc3d.png").getImage();
	
	/** A frame that is used as the parent of all dialog boxes. */
	public static final JFrame dialogParent = new JFrame();
	
	static {
		dialogParent.setIconImage(ICC3D.image);
	}
	
	/** The main application object. */
	protected static ICC3D icc3d;
	
	/** The session manager. */
	private SessionManager sessionManager;
	
	/** The splash screen. */
	private Splash splash;
	
	/** The control area where all parameter panels reside. */
	private ControlArea controlArea;
	
	/** A button that closes the application. */
	private JButton exit;
	
	/** A button that opens a new 3D view. */
	private JButton open3D_View;

	/** A button that opens a new image. */
	private JButton openImage;
	
	/** A button that opens an about box. */
	private JButton aboutButton;
	
	/**
	 * Default constructor.
	 */
	public ICC3D() {
		super("ICC3D version " + getVersion());
		setVisible(false);
		sessionManager = new SessionManager(this);
		System.out.println("ICC3D " + getVersion()
				+ " loading.");
		
		
		Container c = getContentPane();

		c.setLayout(new FlowLayout());
		c.add(createGui());
		controlArea = new ControlArea();
		pack();
		setResizable(false);
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				destroy();
			}
		});
		setIconImage(image);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		loadPlugIns();
	}
	
	/**
	 * Returns the main application object.
	 * @return The main application object.
	 */
	public static ICC3D getICC3D() {
		return icc3d;
	}
	
	/**
	 * Returns a frame that can be used as the parent of all dialog boxes.
	 * @return A frame that can be used as the parent of all dialog boxes.
	 */
	public static JFrame getDialogParent() {
		return dialogParent;
	}
	
	/**
	 * Initiates the dynamic loading of plugins used by the application .
	 * <p>Currently loads plugins for: ColorSpaces, Visualizations, Interactions, Icons ...
	 */
	private void loadPlugIns() {
		System.out.print(" #Loading plugins...");
		ColorSpaceManager.load(null);
		splash = new Splash(this);
		Thread thread = new Thread(splash);

		thread.start();
		ColorSpaceManager.load(splash.statusText);
		VisualizationManager.load(splash.statusText);
		InteractionManager.load(splash.statusText);
		ConstraintsManager.load(splash.statusText);
		MappingManager.load(splash.statusText);
		IconManager.load(splash.statusText);
		ModuleManager.load(splash.statusText);
		System.out.println(" [Done]");
	}
	
	/**
	 * Initiates termination of application.
	 */
	public void destroy() {
		
		boolean quit = true;
		Image2D imgs[] = sessionManager.getImages();
		View3D views[] = sessionManager.getView3Ds();

		if (imgs.length > 0 || views.length > 0) {
			int close = SystemManager.confirmYesNo(
				"Quit",
				"Are you sure you want to quit?");
			if (close == JOptionPane.NO_OPTION) {
				quit = false;
			}
		}
		
		if (quit) {
			System.exit(0);
		}
	}

	/**
	 * Prints information about supported fileformats in this version to standard output.
	 *	
	 * @see #getVersion
	 */
	public static void printHelp() {

		/* System.out.println("\n\nThis version("+getVersion()+") supports\n" +
		 "BMP, FlashPIX, GIF, JPEG, PNG, PNM, and TIFF\n" + 
		 "images as defined in the TIFF 6.0 specification.\n" +
		 "(TIFF G3 (1D and 2D), G4, PackBits, LZW, JPEG, and DEFLATE (Zip)\n" +
		 "compression types are understood.)\n");*/

		System.out.println("BMP file Handling:\n------------------------\n"
				+ "The BMP reader can read Version 2.x, 3.x and some 4.x BMP images.\n"
				+ "BMP images with 1, 4, 8, 24 bits can be read with this reader.\n"
				+ "Support for 16 and 32 bit images has also been implemented, although\n"
				+ "such images are not very common.\n"
				+ "Reading of compressed BMPs is supported. BI_RGB, BI_RLE8, BI_RLE4\n"
				+ "and BI_BITFIELDS compressions are handled.\n\n"
				+ "The BMP reader emits properties such as type of compression, bits\n"
				+ "per pixel etc. Use the getPropertyNames() method to get the names of\n"
				+ "all the properties emitted.\n\n" + "BMP Limitations:"
				+ "+ Only the default RGB color space is supported.\n"
				+ "+ Alpha channels are not supported.\n" + "BMP Writer:\n"
				+ "+ The BMP writer is capable of writing images in the Version 3\n"
				+ "format. Images which make use of a IndexColorModel with 2, 16,	\n"
				+ "or 256 palette entries will be written in palette form.\n"
				+ "Compression is not supported.\n");
		System.out.println("FlashPIX file handling:\n------------------------\n"
				+ "A limited FlashPIX reader is provided that is capable of extracting\n"
				+ "a single resolution from a FlashPIX image file. Only simple FlashPix\n"
				+ "files are decoded properly.\n"
				+ "The image view object is ignored, and image property information is\n"
				+ "not exported.\n" + "There is no FlashPIX writer.\n");

		System.out.println("GIF file handling:\n------------------------\n"
				+ "GIF files are read using the JDK's internal reader. There is no GIF\n"
				+ "writer. Due to a bug in the AWT GIF decoder, GIFs with a transparent\n"
				+ "background will be decoded as if their background were black.\n"
				+ "GIF files may not be read remotely, i.e., via RemoteImage, on hosts\n"
				+ "without window system access as the reader uses the AWT toolkit\n"
				+ "which requires window system access.\n");

		System.out.println("JPEG file handling:\n------------------------\n"
				+ "+ JPEG files are read and written using the classes found in the\n"
				+ "com.sun.image.codec.jpeg package of the JDK. A set of simple JAI\n"
				+ "wrapper classes around these classes is provided.\n"
				+ "+ The JPEG decoder cannot read abbreviated streams, either tables-only\n"
				+ "or image-only.\n");

		System.out.println("PNG file handling:\n------------------------\n"
				+ "All files in the PNGSuite test suite have been read and written\n"
				+ "successfully. See the documentation in\n"
				+ "com.sun.media.jai.codec.PNGDecodeParam and PNGEncodeParam for more\n"
				+ "information.\n");

		System.out.println("PNM file handling:\n------------------------\n"
				+ "PNM files may be read and written in both ASCII and raw formats. The\n"
				+ "encoder automatically selects between PBM (bitmap), PGM (grayscale)\n"
				+ "and PPM (RGB) files according to the number of bands and bit depth\n"
				+ "of the source image.\n"
				+ "Due to the limitations of the format, only images with 1 or 3 bands\n"
				+ "may be written.\n");
		System.out.println("TIFF file Handling:\n------------------------\n"
				+ "TIFF support has the following limitations:\n"
				+ "+ Bilevel-specific compressed formats may not be written. This\n"
				+ "includes CCITT Huffman (G3-1D), CCITT T.4 (G3-2D), and CCITT T.6\n"
				+ "(G4).\n"
				+ "+ LZW compressed format is not supported for the usual reason.\n"
				+ "+ Planar configuration (band sequential data) is not supported.\n");
	}

	/**
	 * Returns the control area.
	 *	
	 * @return Controlarea
	 * @see ControlArea
	 */
	public ControlArea getControlArea() {
		return controlArea;
	}
	
	/**
	 * Returns version information about this release. <xxx major.minor.bug-fix>
	 *	
	 * @return Version information
	 */
	public static String getVersion() {
		return "1.2.9";
	}

	/**
	 * Creates a new instance of ICC3D.
	 *	
	 * @param args array of string objects. If args[0] starts with "?", 
	 * printHelp() is invoked and application is then terminated.
	 * @see #printHelp
	 */
	public static void main(String args[]) {
		try {
			Class.forName("javax.media.j3d.Canvas3D");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to locate Java3D");
		}
		try {
			Class.forName("javax.media.jai.JAI");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to locate Java Advanced Imaging");
		}
		try {
			Class.forName("javax.media.datasink.DataSinkListener");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the Java Media Framework");
		}
		try {
			BufferedWriter bw;

			bw = new BufferedWriter(new java.io.FileWriter("running.log"));
			bw.write("running");
			bw.flush();
			bw.close();
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to write to the log files");
			System.exit(1);
		}
		try {
			PrintStream ps = new PrintStream(new FileOutputStream("log.txt"));

			System.setErr(ps);
			System.setOut(ps);
		} catch (Exception e) {}
		if (args.length == 1 && args[0].charAt(0) == '?') {
			printHelp();
			System.exit(0);
		}

		icc3d = new ICC3D();
	}	
	
	/**
	 * Creates a JPanel with given GUI for the main window, and assigns listeners 
	 * to objects.
	 *	
	 * @return Panel with given GUI. 
	 */
	public JPanel createGui() {
		JPanel main = new JPanel();

		main.setLayout(new BorderLayout());
		main.setBorder(BorderFactory.createLoweredBevelBorder());
		
		JPanel img = new JPanel();

		img.setBorder(BorderFactory.createLoweredBevelBorder());
		img.add(new JLabel(new ImageIcon("icc3dimg/icc3d-main.gif")));
		main.add(img, BorderLayout.WEST);
		
		JPanel tmpPanel = new JPanel();

		tmpPanel.setLayout(new BorderLayout());
						
		JPanel buttons = new JPanel();

		buttons.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				" Options: "));
		buttons.setLayout(new GridLayout(0, 1));
		openImage = new JButton("Open image", IconManager.getIcon("newsession"));
		openImage.setHorizontalAlignment(JButton.LEFT);
		openImage.addActionListener(this);
		buttons.add(openImage);
		open3D_View = new JButton("Open 3D view", IconManager.getIcon("newsession"));
		open3D_View.setHorizontalAlignment(JButton.LEFT);
		open3D_View.addActionListener(this);
		buttons.add(open3D_View);
		exit = new JButton("Exit", IconManager.getIcon("exit"));
		exit.setHorizontalAlignment(JButton.LEFT);
		exit.addActionListener(this);
		
		JPanel tmp = new JPanel();

		tmp.add(exit);
		tmpPanel.add(tmp, BorderLayout.SOUTH);
		aboutButton = new JButton("About", IconManager.getIcon("about"));
		aboutButton.addActionListener(this);
		aboutButton.setHorizontalAlignment(JButton.LEFT);
		buttons.add(aboutButton);
		tmpPanel.add(buttons, BorderLayout.NORTH);
		main.add(tmpPanel, BorderLayout.CENTER);
		return main;
	}
	
	/**
	 * Invokes necessary routines for a given event.
	 *	
	 * @param ae event
	 */
	public void actionPerformed(ActionEvent ae) {
		if (ae.getSource() == exit) {
			destroy();
		} else if (ae.getSource() == openImage) {
			sessionManager.newImage2D();
		} else if (ae.getSource() == aboutButton) {
			About a = new About(this);	
		} else if (ae.getSource() == open3D_View) {
			sessionManager.new3D_View();
		}
	}
	
	/**
	 * Returns the SessionManager object that keeps track of all open
	 * 2D and 3D windows.
	 * @return The session manager.
	 */
	public SessionManager getSessionManager() {
		return sessionManager;
	}
}
