/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.vecmath.*;
import javax.media.j3d.*;
import java.awt.color.*;
import java.awt.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.modules.ColorSpaceManager;
import icc3dpackage.utils.*;


/**
 * This class defines the CAT02 color space, including methods
 * for converting to and from CIEXYZ.
 *
 */
public class CAT02_ColorSpace3D extends ColorSpace3D {
	
	/** An sRGB color space. */
	private ColorSpace3D rgbSpace;
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/**
	 * Constructs a CAT02_ColorSpace3D.
	 */
	public CAT02_ColorSpace3D() {
		rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
		colorLabels = new char[] {'R', 'G', 'B'};
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}

	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int q, count = xyz.length / 3;
		float retColor[] = new float[xyz.length];
		float tmpColor;

		for (int i = 0; i < count; i++) {
			q = i * 3;
			retColor[q] = 0.7328f * xyz[q] + 0.4296f * xyz[q + 1] - 0.1624f * xyz[q + 2];
			retColor[q + 1] = -0.7036f * xyz[q] + 1.6975f * xyz[q + 1]
					+ 0.0061f * xyz[q + 2];
			retColor[q + 2] = 0.0030f * xyz[q] + 0.0136f * xyz[q + 1]
					+ 0.9834f * xyz[q + 2];
		}
		return retColor;		
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float xyz[], float retColor[], boolean dirtyColor[]) {
		int q, count = xyz.length / 3;
		float tmpColor;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * 3;
				retColor[q] = 0.7328f * xyz[q] + 0.4296f * xyz[q + 1] - 0.1624f * xyz[q + 2];
				retColor[q + 1] = -0.7036f * xyz[q] + 1.6975f * xyz[q + 1]
						+ 0.0061f * xyz[q + 2];
				retColor[q + 2] = 0.0030f * xyz[q] + 0.0136f * xyz[q + 1]
						+ 0.9834f * xyz[q + 2];
			}
		}
		return retColor;	
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int q, count = color.length / 3;
		float retColor[] = xyz;
		float a, b, c;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * 3;
				retColor[q] = 1.096124f * color[q] - 0.278869f * color[q + 1]
						+ 0.182745f * color[q + 2];
				retColor[q + 1] = 0.454369f * color[q] + 0.473533f * color[q + 1]
						+ 0.072098f * color[q + 2];
				retColor[q + 2] = -0.009628f * color[q] - 0.005698f * color[q + 1]
						+ 1.015326f * color[q + 2];
			}
		}
			
		return retColor;
	}

	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int q, count = color.length / 3;
		float retColor[] = new float[color.length];
		float a, b, c;

		for (int i = 0; i < count; i++) {
			q = i * 3;
			retColor[q] = 1.096124f * color[q] - 0.278869f * color[q + 1]
					+ 0.182745f * color[q + 2];
			retColor[q + 1] = 0.454369f * color[q] + 0.473533f * color[q + 1]
					+ 0.072098f * color[q + 2];
			retColor[q + 2] = -0.009628f * color[q] - 0.005698f * color[q + 1]
					+ 1.015326f * color[q + 2];
		}		
			
		return retColor;
	}

	/**
	 * Converts color to sRGB(byte) color.
	 *
	 * @param color CAT02 color.
	 * @return sRGB byte color.
	 */	
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}
	
	/**
	 * Creates a new BranchGroup which shows sRGB axis.
	 *
	 * @return sRGB axis
	 */		
	public BranchGroup getAxis() {
		return super.getStandardBranch();
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */		
	public String getName() {
		return "CAT02";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 1001;
	}

	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return The parameter panel
	 */	
	public Component getParameterPanel() {
		return getStandardParameterPanel();
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f, 1.0f, 1.0f};
	}
}
