/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.vecmath.*;
import javax.media.j3d.*;
import java.io.File;
import java.awt.color.*;
import java.awt.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;


/**
 * This class is a wrapper for the standard gray color space provided by Java.
 */
public class GrayColorSpace3D extends ColorSpace3D {
	
	/** The gray color space. */
	private ColorSpace colorSpace;
	
	/**
	 * Creates a GrayColorSpace3D object.
	 */
	public GrayColorSpace3D() {
		colorSpace = ColorSpace.getInstance(ColorSpace.CS_GRAY);
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return colorSpace.getNumComponents();
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return new char[] {'0', '0', '0'};
	}

	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int components = colorSpace.getNumComponents();
		int q, p, count = xyz.length / 3;
		float retColor[] = new float[count * components];
		float temp[] = new float[components];
		float tempRes[];

		for (int i = 0; i < count; i++) {
			q = i * components;
			p = i * 3;
			for (int j = 0; j < 3; j++) {
				temp[j] = xyz[p + j];
			}
			tempRes = colorSpace.fromCIEXYZ(temp);
			
			for (int j = 0; j < components; j++) {
				retColor[q + j] = tempRes[j];
			}
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int components = colorSpace.getNumComponents();
		int q, p, count = ciexyz.length / 3;
		float temp[] = new float[components];
		float tempRes[];

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * components;
				p = i * 3;
				for (int j = 0; j < 3; j++) {
					temp[j] = ciexyz[p + j];
				}
				tempRes = colorSpace.fromCIEXYZ(temp);
				
				for (int j = 0; j < components; j++) {
					retColor[q + j] = tempRes[j];
				}
			}
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int components = colorSpace.getNumComponents();
		int q, p, count = color.length / components;
		float retColor[] = xyz;
		float temp[] = new float[components];
		float tempRes[];

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * components;
				p = i * 3;
				for (int j = 0; j < components; j++) {
					temp[j] = color[q + j];
				}
				tempRes = colorSpace.toCIEXYZ(temp);
				
				for (int j = 0; j < 3; j++) {
					retColor[p + j] = tempRes[j];
				}
			}
		}
		return retColor;
	}

	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int components = colorSpace.getNumComponents();
		int q, p, count = color.length / components;
		float retColor[] = new float[count * 3];
		float temp[] = new float[components];
		float tempRes[];

		for (int i = 0; i < count; i++) {
			q = i * components;
			p = i * 3;
			for (int j = 0; j < components; j++) {
				temp[j] = color[q + j];
			}
			tempRes = colorSpace.toCIEXYZ(temp);
			
			for (int j = 0; j < 3; j++) {
				retColor[p + j] = tempRes[j];
			}
		}
		return retColor;
	}

	/**
	 * Converts color to sRGB(byte) color.
	 *
	 * @param color Grayscale color.
	 * @return sRGB byte color.
	 */	
	public byte[] toSRGB(float color[]) {
		ColorSpace3D rgbSpace = ColorSpaceManager.getColorSpace("sRGB");

		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}
	
	/**
	 * Creates a new BranchGroup which shows the sRGB space limits.
	 *
	 * @return sRGB axis
	 */
	public BranchGroup getAxis() {
		return getStandardAxis();
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */		
	public String getName() {
		return "Grayscale";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 200;
	}

	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */	
	public Component getParameterPanel() {
		return new JPanel();
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f};
	}
}
