/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.media.j3d.*;
import javax.vecmath.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.sun.j3d.utils.geometry.Text2D;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;


/**
 * This class is an implementation of the LCH color space.
 */
public class LCH_ColorSpace3D extends ColorSpace3D {
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** A CieXYZ white point value. */
	private float wpXn, wpYn, wpZn;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** An sRGB color space */
	private ColorSpace3D rgbSpace;
	
	/**
	 * Creates an LCH color space object.
	 */
	public LCH_ColorSpace3D() {
		colorLabels = new char[] {'L', 'C', 'H'};
		wpXn = 0.9642f;
		wpYn = 1.000f;
		wpZn = 0.8249f;
		rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int i, count = color.length / 3;
		float retColor[] = xyz;
		float varX, varY, varZ;
		float l, a, b, h, c;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				l = color[i * 3];
				c = color[i * 3 + 1];
				h = color[i * 3 + 2] * (2 * (float) Math.PI);
				a = c * (float) Math.cos(h) + 0.5f;
				b = c * (float) Math.sin(h) + 0.5f;
				varY = l * 0.86206896551724137931034482758621f
						+ 0.13793103448275862068965517241379f;
				varX = a * 0.512f - 0.256f + varY;
				varZ = varY - b * 1.28f + 0.64f;
				varY = LabColorSpace3D.thirdFuncReverse(varY);
				varX = LabColorSpace3D.thirdFuncReverse(varX);
				varZ = LabColorSpace3D.thirdFuncReverse(varZ);
				retColor[i * 3] = wpXn * varX;
				retColor[i * 3 + 1] = wpYn * varY;
				retColor[i * 3 + 2] = wpZn * varZ;
			}
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int i, count = color.length / 3;
		float retColor[] = new float[count * 3];
		float varX, varY, varZ;
		float l, a, b, h, c;
		
		for (i = 0; i < count; i++) {
			l = color[i * 3];
			c = color[i * 3 + 1];
			h = color[i * 3 + 2] * (2 * (float) Math.PI);
			a = c * (float) Math.cos(h) + 0.5f;
			b = c * (float) Math.sin(h) + 0.5f;
			varY = l * 0.86206896551724137931034482758621f
					+ 0.13793103448275862068965517241379f;
			varX = a * 0.512f - 0.256f + varY;
			varZ = varY - b * 1.28f + 0.64f;
			varY = LabColorSpace3D.thirdFuncReverse(varY);
			varX = LabColorSpace3D.thirdFuncReverse(varX);
			varZ = LabColorSpace3D.thirdFuncReverse(varZ);
			retColor[i * 3] = wpXn * varX;
			retColor[i * 3 + 1] = wpYn * varY;
			retColor[i * 3 + 2] = wpZn * varZ;
		}
		return retColor;
	}

	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int i, count = xyz.length / 3;
		float XXn, YYn, ZZn;
		float retColor[] = new float[count * 3];
		float l, a, b, h;
		
		for (i = 0; i < count; i++) {
			XXn = xyz[i * 3] / wpXn;
			YYn = xyz[i * 3 + 1] / wpYn;
			ZZn = xyz[i * 3 + 2] / wpZn;
			XXn = LabColorSpace3D.thirdFunc(XXn);
			YYn = LabColorSpace3D.thirdFunc(YYn);
			ZZn = LabColorSpace3D.thirdFunc(ZZn);
			l = 1.16f * YYn - 0.16f;
			a = 1.953125f * (XXn - YYn);
			b = 0.78125f * (YYn - ZZn);
			if (a == 0.0f) {
				a = 0.00000000000001f;
			}
			
			float res = (float) Math.atan(b / a);

			if (b < 0.0f && a > 0.0f) {
				res += (float) (2 * Math.PI);
			} else if (a < 0.0f) {
				res += (float) Math.PI;
			}
			if (res >= (float) (Math.PI * 2 - 0.000001f)) {
				res = (float) Math.PI * 2 - 0.000001f;
			}
			h = res / (float) (2 * Math.PI);
			retColor[i * 3] = l;
			retColor[i * 3 + 1] = (float) Math.sqrt(a * a + b * b);
			retColor[i * 3 + 2] = h;
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int i, count = ciexyz.length / 3;
		float XXn, YYn, ZZn;
		float l, a, b, h;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				XXn = ciexyz[i * 3] / wpXn;
				YYn = ciexyz[i * 3 + 1] / wpYn;
				ZZn = ciexyz[i * 3 + 2] / wpZn;
				XXn = LabColorSpace3D.thirdFunc(XXn);
				YYn = LabColorSpace3D.thirdFunc(YYn);
				ZZn = LabColorSpace3D.thirdFunc(ZZn);
				l = 1.16f * YYn - 0.16f;
				a = 1.953125f * (XXn - YYn);
				b = 0.78125f * (YYn - ZZn);
				if (a == 0.0f) {
					a = 0.00000000000001f;
				}
				
				float res = (float) Math.atan(b / a);

				if (b < 0.0f && a > 0.0f) {
					res += (float) (2 * Math.PI);
				} else if (a < 0.0f) {
					res += (float) Math.PI;
				}
				if (res >= (float) (Math.PI * 2 - 0.000001f)) {
					res = (float) Math.PI * 2 - 0.000001f;
				}
				h = res / (float) (2 * Math.PI);
				retColor[i * 3] = l;
				retColor[i * 3 + 1] = (float) Math.sqrt(a * a + b * b);
				retColor[i * 3 + 2] = h;
			}
		}
		return retColor;
	}
	
	/**
	 * Converts LCH color to sRGB color.
	 *
	 * @param color LCH color
	 * @return sRGB color
	 */	
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}

	/*
	 * Creates a new BranchGroup which shows sRGB axis.
	 *
	 * @return sRGB axis
	 */		
	public BranchGroup getAxis() {
		return super.getStandardBranch();
	}

	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return The color space name.
	 */	
	public String getName() {
		return "LCH";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An integer representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 12;
	}
	
	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */		
	public Component getParameterPanel() {
		return super.getStandardParameterPanel();
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f, 1.0f, 1.0f};
	}
}
