/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.media.j3d.*;
import javax.vecmath.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.sun.j3d.utils.geometry.Text2D;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;
import icc3dpackage.visualizations.appearance.*;


/**
 * This class is used in conjunction with operations involving Lab color space.
 * Converting colorvalues from other colorspaces to Lab color space and vice versa.
 */
public class LabColorSpace3D extends ColorSpace3D {
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** The CieXYZ white point values. */
	private float wpXn, wpYn, wpZn;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The size of the Lab color circle. */
	private JComboBox circleSize;
	
	/** Display the circle. */
	private JCheckBox showCircle;
	
	/** Displays the Lab axis. */
	private JCheckBox showLabAxis;
	
	/** The root branch group. */
	private BranchGroup labBranch;
	
	/** The circle branch group. */
	private BranchGroup circleBranch;
	
	/** The axis branch group. */
	private BranchGroup axisBranch;
	
	/** A TransformGroup used to translate the letters correctly. */
	private TransformGroup translateLetter[];
	
	/** A sRGB color space. */
	private ColorSpace3D rgbSpace;
	
	/**
	 * Default constructor.
	 */
	public LabColorSpace3D() {
		colorLabels = new char[] {'L', 'a', 'b'};
		createParameterPanel();
		wpXn = 0.9642f;
		wpYn = 1.000f;
		wpZn = 0.8249f;
		rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];
		float numColor = color.length / 3;

		for (int i = 0; i < numColor; i++) {
			retColor[i * 3 + 1] = color[i * 3] / 100.0f;
			retColor[i * 3 + 2] = 0.5f + color[i * 3 + 1] / 256.0f;
			retColor[i * 3] = 0.5f + color[i * 3 + 2] / 256.0f;
		}
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];
		float numColor = color.length / 3;

		for (int i = 0; i < numColor; i++) {
			retColor[i * 3] = color[i * 3 + 1] * 100.0f;
			retColor[i * 3 + 1] = (color[i * 3 + 2] - 0.5f) * 256.0f;
			retColor[i * 3 + 2] = (color[i * 3] - 0.5f) * 256.0f;
		}
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int i, count = color.length / 3;
		float retColor[] = xyz;
		float varX, varY, varZ;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				varY = color[i * 3 + 1] * 0.86206896551724137931034482758621f
						+ 0.13793103448275862068965517241379f;
				varX = color[i * 3 + 2] * 0.512f - 0.256f + varY;
				varZ = varY - color[i * 3] * 1.28f + 0.64f;
				varY = thirdFuncReverse(varY);
				varX = thirdFuncReverse(varX);
				varZ = thirdFuncReverse(varZ);
				retColor[i * 3] = wpXn * varX;
				retColor[i * 3 + 1] = wpYn * varY;
				retColor[i * 3 + 2] = wpZn * varZ;
			}
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int i, count = color.length / 3;
		float retColor[] = new float[count * 3];
		float varX, varY, varZ;

		for (i = 0; i < count; i++) {
			varY = color[i * 3 + 1] * 0.86206896551724137931034482758621f
					+ 0.13793103448275862068965517241379f;
			varX = color[i * 3 + 2] * 0.512f - 0.256f + varY;
			varZ = varY - color[i * 3] * 1.28f + 0.64f;
			varY = thirdFuncReverse(varY);
			varX = thirdFuncReverse(varX);
			varZ = thirdFuncReverse(varZ);
			retColor[i * 3] = wpXn * varX;
			retColor[i * 3 + 1] = wpYn * varY;
			retColor[i * 3 + 2] = wpZn * varZ;
		}
		return retColor;
	}
	
	/**
	 * Calculates f^(1/3)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */
	public static final float firstFunc(float f) {
		return ((float) Math.pow(f, 1.0 / 3.0));
	}
	
	/**
	 * Calculates f^(3)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */
	public static final float firstFuncReverse(float f) {
		return ((float) Math.pow(f, 3.0));
	}

	/**
	 * Calculates (7.787*f)+(16.0/116.0)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */	
	public static final float secondFunc(float f) {
		return (7.787f * f + 16.0f / 116.0f);
	}

	/**
	 * Calculates f-((16.0/116.0)/7.787)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */		
	public static final float secondFuncReverse(float f) {
		return ((f - 16.0f / 116.0f) / 7.787f);
	}
	
	/**
	 * Calculates either: firstFunc(f) if f>0.008856, otherwise secondFunc(f)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */	
	public static final float thirdFunc(float f) {
		if (f > 0.008856f) {
			return firstFunc(f);
		}
		return secondFunc(f);
	}

	/**
	 * Calculates either: firstFuncReverse(f) if firstFuncReverse(f)>0.008856, otherwise secondFuncReverse(f)
	 *
	 * @param f value to calculate from
	 * @return calculated value of f
	 */		
	public static final float thirdFuncReverse(float f) {
		if (firstFuncReverse(f) > 0.008856f) {
			return firstFuncReverse(f);
		}
		return secondFuncReverse(f);
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int i, count = xyz.length / 3;
		float XXn, YYn, ZZn;
		float retColor[] = new float[count * 3];
		
		for (i = 0; i < count; i++) {
			XXn = xyz[i * 3] / wpXn;
			YYn = xyz[i * 3 + 1] / wpYn;
			ZZn = xyz[i * 3 + 2] / wpZn;
			XXn = thirdFunc(XXn);
			YYn = thirdFunc(YYn);
			ZZn = thirdFunc(ZZn);
			retColor[i * 3 + 1] = 1.16f * YYn - 0.16f;
			retColor[i * 3 + 2] = 1.953125f * (XXn - YYn) + 0.5f;
			retColor[i * 3] = 0.78125f * (YYn - ZZn) + 0.5f;
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int i, count = ciexyz.length / 3;
		float XXn, YYn, ZZn;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				XXn = ciexyz[i * 3] / wpXn;
				YYn = ciexyz[i * 3 + 1] / wpYn;
				ZZn = ciexyz[i * 3 + 2] / wpZn;
				XXn = thirdFunc(XXn);
				YYn = thirdFunc(YYn);
				ZZn = thirdFunc(ZZn);
				retColor[i * 3 + 1] = 1.16f * YYn - 0.16f;
				retColor[i * 3 + 2] = 1.953125f * (XXn - YYn) + 0.5f;
				retColor[i * 3] = 0.78125f * (YYn - ZZn) + 0.5f;
			}
		}
		return retColor;
	}
	
	/**
	 * Converts Lab color to sRGB color.
	 *
	 * @param color CIELAB color
	 * @return sRGB color
	 */	
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}

	/**
	 * Creates a new BranchGroup which shows(if selected by user): Lab cicle and Lab axis
	 *
	 * @return Lab axis
	 */	
	public BranchGroup getAxis() {
		labBranch = new BranchGroup();
		labBranch.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		labBranch.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		circleBranch = null;
		axisBranch = null;
		labBranch.addChild(getStandardBranch());
		if (showCircle.isSelected()) {
			circleBranch = createLabCircle();
			circleBranch.setCapability(BranchGroup.ALLOW_DETACH);
			labBranch.addChild(circleBranch);
		}
		if (showLabAxis.isSelected()) {
			translateLetter = new TransformGroup[5];
			axisBranch = createColorAxis("b+", "L+", "a+", "b-", "L-", "a-",
					translateLetter);
			axisBranch.setCapability(BranchGroup.ALLOW_DETACH);
			labBranch.addChild(axisBranch);
		}
		return labBranch;
	} 
	
	/**
	 * Creates a new BranchGroup which shows Lab axis.
	 *
	 * @return Lab axis
	 */	
	public static final BranchGroup createColorAxis(String x, String y, String z, String xMinus, String yMinus, String zMinus, TransformGroup letterTranslate[]) {
		BranchGroup branchGroup = new BranchGroup();

		letterTranslate[0] = new TransformGroup();
		letterTranslate[0].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		letterTranslate[0].setCapability(TransformGroup.ALLOW_LOCAL_TO_VWORLD_READ);
		letterTranslate[1] = new TransformGroup();
		letterTranslate[1].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		letterTranslate[2] = new TransformGroup();
		letterTranslate[2].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		letterTranslate[3] = new TransformGroup();
		letterTranslate[3].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		letterTranslate[4] = new TransformGroup();
		letterTranslate[4].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		
		Transform3D l3d = new Transform3D();

		l3d.setTranslation(new Vector3f(0.5f, 1.1f, 0.5f));
		TransformGroup lGroup = new TransformGroup(l3d);
		Text2D lText2D = new Text2D(y, new Color3f(1.0f, 1.0f, 1.0f), "Helvetica", 15,
				Font.PLAIN);
		Transform3D lBill3d = new Transform3D();
		TransformGroup lBillGroup = new TransformGroup(lBill3d);

		lBillGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		Appearance app = lText2D.getAppearance();
		RenderingAttributes renAtt = new RenderingAttributes();

		renAtt.setAlphaTestFunction(RenderingAttributes.GREATER);
		renAtt.setAlphaTestValue(0.1f);
		app.setRenderingAttributes(renAtt);
		lText2D.setAppearance(app);
		letterTranslate[0].addChild(lText2D);
		lBillGroup.addChild(letterTranslate[0]);
		lGroup.addChild(lBillGroup);
		branchGroup.addChild(lGroup);
		l3d = new Transform3D();
		l3d.setTranslation(new Vector3f(0.5f, 0.0f, 1.1f));
		lGroup = new TransformGroup(l3d);
		lText2D = new Text2D(z, new Color3f(1.0f, 0.0f, 0.0f), "Helvetica", 15,
				Font.PLAIN);
		lBill3d = new Transform3D();
		lBillGroup = new TransformGroup(lBill3d);
		lBillGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		app = lText2D.getAppearance();
		renAtt = new RenderingAttributes();
		renAtt.setAlphaTestFunction(RenderingAttributes.GREATER);
		renAtt.setAlphaTestValue(0.1f);
		app.setRenderingAttributes(renAtt);
		lText2D.setAppearance(app);
		letterTranslate[1].addChild(lText2D);
		lBillGroup.addChild(letterTranslate[1]);
		lGroup.addChild(lBillGroup);
		branchGroup.addChild(lGroup);
		l3d = new Transform3D();
		l3d.setTranslation(new Vector3f(1.1f, 0.0f, 0.5f));
		lGroup = new TransformGroup(l3d);
		lText2D = new Text2D(x, new Color3f(1.0f, 1.0f, 0.0f), "Helvetica", 15,
				Font.PLAIN);
		lBill3d = new Transform3D();
		lBillGroup = new TransformGroup(lBill3d);
		lBillGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		app = lText2D.getAppearance();
		renAtt = new RenderingAttributes();
		renAtt.setAlphaTestFunction(RenderingAttributes.GREATER);
		renAtt.setAlphaTestValue(0.1f);
		app.setRenderingAttributes(renAtt);
		lText2D.setAppearance(app);
		letterTranslate[2].addChild(lText2D);
		lBillGroup.addChild(letterTranslate[2]);
		lGroup.addChild(lBillGroup);
		branchGroup.addChild(lGroup);
		l3d = new Transform3D();
		l3d.setTranslation(new Vector3f(0.5f, 0.0f, -0.1f));
		lGroup = new TransformGroup(l3d);
		lText2D = new Text2D(zMinus, new Color3f(0.0f, 1.0f, 0.0f), "Helvetica", 15,
				Font.PLAIN);
		lBill3d = new Transform3D();
		lBillGroup = new TransformGroup(lBill3d);
		lBillGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		app = lText2D.getAppearance();
		renAtt = new RenderingAttributes();
		renAtt.setAlphaTestFunction(RenderingAttributes.GREATER);
		renAtt.setAlphaTestValue(0.1f);
		app.setRenderingAttributes(renAtt);
		lText2D.setAppearance(app);
		letterTranslate[3].addChild(lText2D);
		lBillGroup.addChild(letterTranslate[3]);
		lGroup.addChild(lBillGroup);
		branchGroup.addChild(lGroup);
		l3d = new Transform3D();
		l3d.setTranslation(new Vector3f(-0.1f, 0.0f, 0.5f));
		lGroup = new TransformGroup(l3d);
		lText2D = new Text2D(xMinus, new Color3f(0.0f, 0.0f, 1.0f), "Helvetica", 15,
				Font.PLAIN);
		lBill3d = new Transform3D();
		lBillGroup = new TransformGroup(lBill3d);
		lBillGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		app = lText2D.getAppearance();
		renAtt = new RenderingAttributes();
		renAtt.setAlphaTestFunction(RenderingAttributes.GREATER);
		renAtt.setAlphaTestValue(0.1f);
		app.setRenderingAttributes(renAtt);
		lText2D.setAppearance(app);
		letterTranslate[4].addChild(lText2D);
		lBillGroup.addChild(letterTranslate[4]);
		lGroup.addChild(lBillGroup);
		branchGroup.addChild(lGroup);
		
		LineArray lineArray = new LineArray(6, GeometryArray.COORDINATES);
		Point3f points[] = new Point3f[6];
		Color3f colors[] = new Color3f[6];

		points[0] = new Point3f(0.5f, 0.0f, 0.5f);
		colors[0] = new Color3f(1.0f, 1.0f, 1.0f);
		points[1] = new Point3f(0.5f, 1.0f, 0.5f);
		colors[1] = new Color3f(1.0f, 1.0f, 1.0f);
		points[2] = new Point3f(0.0f, 0.0f, 0.5f);
		colors[2] = new Color3f(1.0f, 1.0f, 1.0f);
		points[3] = new Point3f(1.0f, 0.0f, 0.5f);
		colors[3] = new Color3f(1.0f, 1.0f, 1.0f);
		points[4] = new Point3f(0.5f, 0.0f, 0.0f);
		colors[4] = new Color3f(1.0f, 1.0f, 1.0f);
		points[5] = new Point3f(0.5f, 0.0f, 1.0f);
		colors[5] = new Color3f(1.0f, 1.0f, 1.0f);
		lineArray.setCoordinates(0, points);
		// lineArray.setColors(0, colors);
		
		
		Shape3D shape3D = new Shape3D(lineArray,
				StandardAppearance.getLineAppearance(new Color3f(0.0f, 0.0f, 0.0f)));

		branchGroup.addChild(shape3D);
		
		return branchGroup;
	}
	
	/**
	 * Creates a new BranchGroup which shows Lab circle. Radius may be either 100 or 128.
	 *
	 * @return Lab axis
	 */		
	private BranchGroup createLabCircle() {
		int circleDetail = 90;
		BranchGroup branchGroup = new BranchGroup();
		float adjustment = 1.0f;

		if (circleSize.getSelectedIndex() == 0) {
			adjustment = 100.0f / 128.0f;
		}
		LineStripArray lineArray = new LineStripArray(circleDetail + 1,
				GeometryArray.COORDINATES, new int[] {circleDetail + 1});
		Point3f points[] = new Point3f[circleDetail + 1];
		Color3f colors[] = new Color3f[circleDetail + 1];
		
		for (int i = 0; i < circleDetail + 1; i++) {
			points[i] = new Point3f(adjustment
					* (float) Math.sin((2 * Math.PI * i / (float) circleDetail)) / 2.0f
					+ 0.5f,
					0.0f,
							adjustment * ((float) Math.cos(2 * Math.PI * i / (float) circleDetail))
							/ 2.0f
					+ 0.5f);
			colors[i] = new Color3f(1.0f, 1.0f, 1.0f);
		}
		lineArray.setCoordinates(0, points);
		
		Shape3D shape3D = new Shape3D(lineArray,
				StandardAppearance.getLineAppearance(new Color3f(0.0f, 0.0f, 0.0f)));

		branchGroup.addChild(shape3D);		
		return branchGroup;
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */	
	public String getName() {
		return "Lab";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 1;
	}
	
	/**
	 * Creates a parameter panel for this ColorSpace3D.
	 */	
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		JPanel innerPanel = new JPanel(new BorderLayout());

		// innerPanel.add(new JLabel("Lab"), BorderLayout.NORTH);
		innerPanel.add(getStandardParameterPanel(), BorderLayout.SOUTH);
		
		ActionListener circleListener = new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (circleBranch != null) {
					circleBranch.detach();
					circleBranch = null;
				}
				if (showCircle.isSelected() && labBranch != null) {
					circleBranch = createLabCircle();
					circleBranch.setCapability(BranchGroup.ALLOW_DETACH);
					labBranch.addChild(circleBranch);
				}
			}
		};
		
		String circleSizes[] = new String[] {"100", "128"};

		circleSize = new JComboBox(circleSizes);
		circleSize.addActionListener(circleListener);
		showCircle = new JCheckBox("", true);
		showCircle.addActionListener(circleListener);
		showLabAxis = new JCheckBox("", true);
		showLabAxis.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (axisBranch != null) {
					axisBranch.detach();
					axisBranch = null;
				}
				if (showLabAxis.isSelected() && labBranch != null) {
					translateLetter = new TransformGroup[5];
					axisBranch = createColorAxis("b+", "L+", "a+", "b-", "L-", "a-",
							translateLetter);
					axisBranch.setCapability(BranchGroup.ALLOW_DETACH);
					labBranch.addChild(axisBranch);
					signalMovement();
				}
			}
		});
		JPanel labParamPanel = new JPanel(new GridLayout(0, 2));

		labParamPanel.add(new JLabel("Lab circle"));
		labParamPanel.add(showCircle);
		labParamPanel.add(new JLabel("Circle radius"));
		labParamPanel.add(circleSize);
		labParamPanel.add(new JLabel("Lab axis"));
		labParamPanel.add(showLabAxis);
		
		JPanel combinedPanel = new JPanel(new BorderLayout());

		combinedPanel.add(innerPanel, BorderLayout.NORTH);
		combinedPanel.add(labParamPanel, BorderLayout.SOUTH);
		parameterPanel.add(combinedPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel.
	 * @return parameter panel
	 */		
	public Component getParameterPanel() {
		return parameterPanel;
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		Transform3D transform = new Transform3D();

		transform.setScale(new Vector3d(1.0, 100.0 / 256.0, 1.0));
		transform.setTranslation(new Vector3d(0.0, (1.0 - 100.0 / 256.0) / 2.0, 0.0));
		return transform;
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {
		if (translateLetter != null && translateLetter[0] != null) {
			try {
				Transform3D t3d = new Transform3D();

				translateLetter[0].getLocalToVworld(t3d);
				t3d.invert();
				t3d.setTranslation(new Vector3d());
				for (int i = 0; i < 5; i++) {
					translateLetter[i].setTransform(t3d);
				}
			} catch (Exception e) {// ignore, no reason to report an error since this will happen
				// every now and then due to the Java3D scene graph being
				// undefined at that moment.
			}
		}
	}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {256.0f, 100.0f, 256.0f};
	}
}
