/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.media.j3d.*;
import javax.vecmath.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import com.sun.j3d.utils.geometry.Text2D;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;
import icc3dpackage.visualizations.appearance.*;


/**
 * This class is used in conjunction with operations involving Luv color space.
 * Converting colorvalues from other colorspaces to Luv color space and vice versa.
 */
public class LuvColorSpace3D extends ColorSpace3D {
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** The CieXYZ white point values. */
	private float wpXn, wpYn, wpZn;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The size of the Luv color circle. */
	private JComboBox circleSize;
	
	/** Display the circle. */
	private JCheckBox showCircle;
	
	/** Displays the Luv axis. */
	private JCheckBox showLuvAxis;
	
	/** The root branch group. */
	private BranchGroup luvBranch;
	
	/** The circle branch group. */
	private BranchGroup circleBranch;
	
	/** The axis branch group. */
	private BranchGroup axisBranch;
	
	/** Reference u value. */
	private float uRef;
	
	/** Reference v value. */
	private float vRef;
	
	/** A TransformGroup used to translate the letters correctly. */
	private TransformGroup translateLetter[];
	
	/** A sRGB color space. */
	private ColorSpace3D rgbSpace;
	
	/**
	 * Default constructor.
	 */
	public LuvColorSpace3D() {
		colorLabels = new char[] {'L', 'u', 'v'};
		createParameterPanel();
		wpXn = 0.9642f;
		wpYn = 1.000f;
		wpZn = 0.8249f;
		rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
		uRef = uFrom(wpXn, wpYn, wpZn);
		vRef = vFrom(wpXn, wpYn, wpZn);
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];
		float numColor = color.length / 3;

		for (int i = 0; i < numColor; i++) {
			retColor[i * 3 + 1] = color[i * 3] / 100.0f;
			retColor[i * 3 + 2] = 0.5f + color[i * 3 + 1] / 256.0f;
			retColor[i * 3] = 0.5f + color[i * 3 + 2] / 256.0f;
		}
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];
		float numColor = color.length / 3;

		for (int i = 0; i < numColor; i++) {
			retColor[i * 3] = color[i * 3 + 1] * 100.0f;
			retColor[i * 3 + 1] = (color[i * 3 + 2] - 0.5f) * 256.0f;
			retColor[i * 3 + 2] = (color[i * 3] - 0.5f) * 256.0f;
		}
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int i, count = color.length / 3;
		float retColor[] = xyz;
		float varX, varY, varZ;
		float l, u, v, x, y, z;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				l = color[i * 3 + 1];
				u = (color[i * 3 + 2] - 0.5f) * 256.0f / 100.0f;
				v = (color[i * 3] - 0.5f) * 256.0f / 100.0f;
				u /= (13 * l);
				v /= (13 * l);
				u += uRef;
				v += vRef;
				y = wpYn
						* LabColorSpace3D.thirdFuncReverse(color[i * 3 + 1]
										* 0.86206896551724137931034482758621f
								+ 0.13793103448275862068965517241379f);
				x = (9.0f / 4.0f) * (u * y / v);
				z = (-1.0f / 4.0f) * ((3 * u + 20 * v - 12) * y / v);
				if (Float.isNaN(z)) {
					z = 0.0f;
				}
				if (Float.isNaN(x)) {
					x = 0.0f;
				}
				retColor[i * 3] = x;
				retColor[i * 3 + 1] = y;
				retColor[i * 3 + 2] = z;
			}
		}
		return retColor;
	}
	
	public static final float uFrom(float x, float y, float z) {
		return 4 * x / (x + 15 * y + 3 * z);
	}
	
	public static final float vFrom(float x, float y, float z) {
		return 9 * y / (x + 15 * y + 3 * z);
	}
	
	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int i, count = color.length / 3;
		float retColor[] = new float[count * 3];
		float varX, varY, varZ;
		float l, u, v, x, y, z;
		
		for (i = 0; i < count; i++) {
			l = color[i * 3 + 1];
			u = (color[i * 3 + 2] - 0.5f) * 256.0f / 100.0f;
			v = (color[i * 3] - 0.5f) * 256.0f / 100.0f;
			u /= (13 * l);
			v /= (13 * l);
			u += uRef;
			v += vRef;
			y = wpYn
					* LabColorSpace3D.thirdFuncReverse(color[i * 3 + 1]
									* 0.86206896551724137931034482758621f
							+ 0.13793103448275862068965517241379f);
			x = (9.0f / 4.0f) * (u * y / v);
			z = (-1.0f / 4.0f) * ((3 * u + 20 * v - 12) * y / v);
			if (Float.isNaN(z)) {
				z = 0.0f;
			}
			if (Float.isNaN(x)) {
				x = 0.0f;
			}
			retColor[i * 3] = x;
			retColor[i * 3 + 1] = y;
			retColor[i * 3 + 2] = z;
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int i, count = xyz.length / 3;
		float XXn, YYn, ZZn;
		float retColor[] = new float[count * 3];
		float l, u, v;
		
		for (i = 0; i < count; i++) {
			l = 1.16f * LabColorSpace3D.thirdFunc(xyz[i * 3 + 1] / wpYn) - 0.16f;
			u = 13 * l * (uFrom(xyz[i * 3], xyz[i * 3 + 1], xyz[i * 3 + 2]) - uRef);
			v = 13 * l * (vFrom(xyz[i * 3], xyz[i * 3 + 1], xyz[i * 3 + 2]) - vRef);
			
			if (Float.isNaN(u)) {
				u = 0.0f;
			}
			if (Float.isNaN(v)) {
				v = 0.0f;
			}
			u = 0.5f + u * 100.0f / 256.0f;
			v = 0.5f + v * 100.0f / 256.0f;
			retColor[i * 3 + 1] = l;
			retColor[i * 3 + 2] = u;
			retColor[i * 3] = v;
			// System.out.println("L: "+l+", u:"+u+", v:"+v);
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int i, count = ciexyz.length / 3;
		float XXn, YYn, ZZn;
		float l, u, v;

		for (i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				l = 1.16f * LabColorSpace3D.thirdFunc(ciexyz[i * 3 + 1] / wpYn) - 0.16f;
				u = 13 * l
						* (uFrom(ciexyz[i * 3], ciexyz[i * 3 + 1], ciexyz[i * 3 + 2]) - uRef);
				v = 13 * l
						* (vFrom(ciexyz[i * 3], ciexyz[i * 3 + 1], ciexyz[i * 3 + 2]) - vRef);
				if (Float.isNaN(u)) {
					u = 0.0f;
				}
				if (Float.isNaN(v)) {
					v = 0.0f;
				}
				u = 0.5f + u * 100.0f / 256.0f;
				v = 0.5f + v * 100.0f / 256.0f;
				retColor[i * 3 + 1] = l;
				retColor[i * 3 + 2] = u;
				retColor[i * 3] = v;
				// System.out.println("L: "+l+", u:"+u+", v:"+v);
			}
		}
		return retColor;
	}
	
	/**
	 * Converts Luv color to sRGB color.
	 *
	 * @param color Luv color
	 * @return sRGB color
	 */	
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}

	/**
	 * Creates a new BranchGroup which shows(if selected by user): Luv cicle and Luv axis
	 *
	 * @return Luv axis
	 */	
	public BranchGroup getAxis() {
		luvBranch = new BranchGroup();
		luvBranch.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		luvBranch.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		circleBranch = null;
		axisBranch = null;
		luvBranch.addChild(getStandardBranch());
		if (showCircle.isSelected()) {
			circleBranch = createLuvCircle();
			circleBranch.setCapability(BranchGroup.ALLOW_DETACH);
			luvBranch.addChild(circleBranch);
		}
		if (showLuvAxis.isSelected()) {
			translateLetter = new TransformGroup[5];
			axisBranch = LabColorSpace3D.createColorAxis("v+", "L+", "u+", "v-", "L-",
					"u-", translateLetter);
			axisBranch.setCapability(BranchGroup.ALLOW_DETACH);
			luvBranch.addChild(axisBranch);
		}
		return luvBranch;
	}
	
	/**
	 * Creates a new BranchGroup which shows Luv circle. Radius may be either 100 or 128.
	 *
	 * @return Luv axis
	 */		
	private BranchGroup createLuvCircle() {
		int circleDetail = 90;
		BranchGroup branchGroup = new BranchGroup();
		float adjustment = 1.0f;

		if (circleSize.getSelectedIndex() == 0) {
			adjustment = 100.0f / 128.0f;
		}
		LineStripArray lineArray = new LineStripArray(circleDetail + 1,
				GeometryArray.COORDINATES, new int[] {circleDetail + 1});
		Point3f points[] = new Point3f[circleDetail + 1];
		Color3f colors[] = new Color3f[circleDetail + 1];
		
		for (int i = 0; i < circleDetail + 1; i++) {
			points[i] = new Point3f(adjustment
					* (float) Math.sin((2 * Math.PI * i / (float) circleDetail)) / 2.0f
					+ 0.5f,
					0.0f,
							adjustment * ((float) Math.cos(2 * Math.PI * i / (float) circleDetail))
							/ 2.0f
					+ 0.5f);
			colors[i] = new Color3f(1.0f, 1.0f, 1.0f);
		}
		lineArray.setCoordinates(0, points);
		// lineArray.setColors(0, colors);
		
		Shape3D shape3D = new Shape3D(lineArray,
				StandardAppearance.getLineAppearance(new Color3f(0.0f, 0.0f, 0.0f)));

		branchGroup.addChild(shape3D);		
		return branchGroup;
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */	
	public String getName() {
		return "Luv";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 13;
	}
	
	/**
	 * Creates a parameter panel for this ColorSpace3D.
	 */	
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		JPanel innerPanel = new JPanel(new BorderLayout());

		innerPanel.add(getStandardParameterPanel(), BorderLayout.SOUTH);
		
		ActionListener circleListener = new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (circleBranch != null) {
					circleBranch.detach();
					circleBranch = null;
				}
				if (showCircle.isSelected() && luvBranch != null) {
					circleBranch = createLuvCircle();
					circleBranch.setCapability(BranchGroup.ALLOW_DETACH);
					luvBranch.addChild(circleBranch);
				}
			}
		};
		
		String circleSizes[] = new String[] {"100", "128"};

		circleSize = new JComboBox(circleSizes);
		circleSize.addActionListener(circleListener);
		showCircle = new JCheckBox();
		showCircle.addActionListener(circleListener);
		showLuvAxis = new JCheckBox();
		showLuvAxis.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (axisBranch != null) {
					axisBranch.detach();
					axisBranch = null;
				}
				if (showLuvAxis.isSelected() && luvBranch != null) {
					translateLetter = new TransformGroup[5];
					axisBranch = LabColorSpace3D.createColorAxis("v+", "L+", "u+", "v-", "L-",
							"u-", translateLetter);
					axisBranch.setCapability(BranchGroup.ALLOW_DETACH);
					luvBranch.addChild(axisBranch);
					signalMovement();
				}
			}
		});
		JPanel luvParamPanel = new JPanel(new GridLayout(0, 2));

		luvParamPanel.add(new JLabel("Luv circle"));
		luvParamPanel.add(showCircle);
		luvParamPanel.add(new JLabel("Circle radius"));
		luvParamPanel.add(circleSize);
		luvParamPanel.add(new JLabel("Luv axis"));
		luvParamPanel.add(showLuvAxis);
		
		JPanel combinedPanel = new JPanel(new BorderLayout());

		combinedPanel.add(innerPanel, BorderLayout.NORTH);
		combinedPanel.add(luvParamPanel, BorderLayout.SOUTH);
		parameterPanel.add(combinedPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */		
	public Component getParameterPanel() {
		return parameterPanel;
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		Transform3D transform = new Transform3D();

		transform.setScale(new Vector3d(1.0, 100.0 / 256.0, 1.0));
		transform.setTranslation(new Vector3d(0.0, (1.0 - 100.0 / 256.0) / 2.0, 0.0));
		return transform;
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {
		if (translateLetter != null && translateLetter[0] != null) {
			try {
				Transform3D t3d = new Transform3D();

				translateLetter[0].getLocalToVworld(t3d);
				t3d.invert();
				t3d.setTranslation(new Vector3d());
				for (int i = 0; i < 5; i++) {
					translateLetter[i].setTransform(t3d);
				}
			} catch (Exception e) {// ignore, no reason to report an error since this will happen
				// every now and then due to the Java3D scene graph being
				// undefined at that moment.
			}
		}
	}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {256.0f, 100.0f, 256.0f};
	}
}
