/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import javax.vecmath.*;
import javax.media.j3d.*;
import java.awt.color.*;
import java.awt.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;


/**
 * This class defines an sRGB color space, including methods
 * for converting to and from CIEXYZ.
 *
 */
public class SRGB_ColorSpace3D extends ColorSpace3D {
	
	/** The CIEXYZ white point */
	protected static final float whitePoint[] = new float[] {0.95047f, 1.0f,
		1.08883f};
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/**
	 * Default constructor.
	 */
	public SRGB_ColorSpace3D() {
		colorLabels = new char[] {'R', 'G', 'B'};
		createParameterPanel();
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int q, count = xyz.length / 3;
		float retColor[] = new float[xyz.length];
		float tmpColor;

		for (int i = 0; i < count; i++) {
			q = i * 3;
			tmpColor = 3.1336026f * xyz[q] + -1.6168214f * xyz[q + 1]
					+ -0.49074236f * xyz[q + 2];
			if (tmpColor <= 0.0030402f) {
				tmpColor = tmpColor * 12.92f;
			} else {
				tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
			}
			retColor[q] = tmpColor;
			tmpColor = -0.9786503f * xyz[q] + 1.916061f * xyz[q + 1]
					+ 0.033512883f * xyz[q + 2];
			if (tmpColor <= 0.0030402f) {
				tmpColor = tmpColor * 12.92f;
			} else {
				tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
			}
			retColor[q + 1] = tmpColor;
			tmpColor = 0.07207655f * xyz[q] + -0.22906554f * xyz[q + 1]
					+ 1.4053595f * xyz[q + 2];
			if (tmpColor <= 0.0030402f) {
				tmpColor = tmpColor * 12.92f;
			} else {
				tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
			}
			retColor[q + 2] = tmpColor;
		}
		return retColor;		
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int q, count = ciexyz.length / 3;
		float tmpColor;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * 3;
				tmpColor = 3.1336026f * ciexyz[q] + -1.6168214f * ciexyz[q + 1]
						+ -0.49074236f * ciexyz[q + 2];
				if (tmpColor <= 0.0030402f) {
					tmpColor = tmpColor * 12.92f;
				} else {
					tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
				}
				retColor[q] = tmpColor;
				tmpColor = -0.9786503f * ciexyz[q] + 1.916061f * ciexyz[q + 1]
						+ 0.033512883f * ciexyz[q + 2];
				if (tmpColor <= 0.0030402f) {
					tmpColor = tmpColor * 12.92f;
				} else {
					tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
				}
				retColor[q + 1] = tmpColor;
				tmpColor = 0.07207655f * ciexyz[q] + -0.22906554f * ciexyz[q + 1]
						+ 1.4053595f * ciexyz[q + 2];
				if (tmpColor <= 0.0030402f) {
					tmpColor = tmpColor * 12.92f;
				} else {
					tmpColor = (1.055f * (float) Math.pow(tmpColor, (1 / 2.4))) - 0.055f;
				}
				retColor[q + 2] = tmpColor;
			}
		}
		return retColor;	
	}
	
	/**
	 * Calculates an intermediate value used in another calculation.
	 *
	 * @param rgb The input value.
	 * @return The resulting intermediate value.
	 */
	private float firstStep(float rgb) {
		float rgb1;

		if (rgb <= 0.03928f) {
			rgb1 = rgb / 12.92f;
		} else {
			rgb1 = (float) Math.pow(((0.055f + rgb) / 1.055f), 2.4);
		}
		return rgb1;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int q, count = color.length / 3;
		float retColor[] = xyz;
		float a, b, c;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				q = i * 3;
				
				if (color[q] <= 0.03928f) {
					a = color[q] / 12.92f;
				} else {
					a = (float) Math.pow(((0.055f + color[q]) / 1.055f), 2.4);
				}
					
				if (color[q + 1] <= 0.03928f) {
					b = color[q + 1] / 12.92f;
				} else {
					b = (float) Math.pow(((0.055f + color[q + 1]) / 1.055f), 2.4);
				}
				
				if (color[q + 2] <= 0.03928f) {
					c = color[q + 2] / 12.92f;
				} else {
					c = (float) Math.pow(((0.055f + color[q + 2]) / 1.055f), 2.4);
				}
				retColor[q] = 0.4361f * a + 0.3851f * b + 0.1431f * c;
				retColor[q + 1] = 0.2225f * a + 0.7169f * b + 0.0606f * c;
				retColor[q + 2] = 0.0139f * a + 0.0971f * b + 0.7141f * c;
			}
		}
			
		return retColor;
	}

	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int q, count = color.length / 3;
		float retColor[] = new float[color.length];
		float a, b, c;

		for (int i = 0; i < count; i++) {
			q = i * 3;
			
			if (color[q] <= 0.03928f) {
				a = color[q] / 12.92f;
			} else {
				a = (float) Math.pow(((0.055f + color[q]) / 1.055f), 2.4);
			}
				
			if (color[q + 1] <= 0.03928f) {
				b = color[q + 1] / 12.92f;
			} else {
				b = (float) Math.pow(((0.055f + color[q + 1]) / 1.055f), 2.4);
			}
			
			if (color[q + 2] <= 0.03928f) {
				c = color[q + 2] / 12.92f;
			} else {
				c = (float) Math.pow(((0.055f + color[q + 2]) / 1.055f), 2.4);
			}
			retColor[q] = 0.4361f * a + 0.3851f * b + 0.1431f * c;
			retColor[q + 1] = 0.2225f * a + 0.7169f * b + 0.0606f * c;
			retColor[q + 2] = 0.0139f * a + 0.0971f * b + 0.7141f * c;
		}		
			
		return retColor;
	}

	/**
	 * Converts sRGB(float) color to sRGB(byte) color.
	 *
	 * @param color sRGB floating point color.
	 * @return sRGB byte color.
	 */	
	public byte[] toSRGB(float color[]) {
		int count = color.length;
		int i;
		byte retColor[] = new byte[count];
		float tempColor;

		for (i = 0; i < count; i++) {
			tempColor = color[i];
			if (tempColor < 0.0f) {
				tempColor = 0.0f;
			}
			if (tempColor > 1.0f) {
				tempColor = 1.0f;
			}
			retColor[i] = (byte) (tempColor * 255.99f);
		}
		return retColor;
	}
	
	/**
	 * Creates a new BranchGroup which shows the sRGB color space cube.
	 *
	 * @return sRGB axis.
	 */		
	public BranchGroup getAxis() {
		return super.getStandardAxis();

		/*
		 float len=1.0f;
		 BranchGroup tmp = new BranchGroup();
		 
		 LineArray lineData = new LineArray(24, GeometryArray.COORDINATES|GeometryArray.COLOR_3);
		 
		 Point3f coords[] = new Point3f[24];
		 Color3f color[] = new Color3f[24];
		 
		 //----------------------------------------------------
		 // coordinates		
		 coords[0] = new Point3f(0.0f, 0.0f, 0.0f); 	// S->R
		 coords[1] = new Point3f(len, 0.0f, 0.0f); 	// S->R
		 coords[2] = new Point3f(0.0f, 0.0f, 0.0f); 	// S->G
		 coords[3] = new Point3f(0.0f, len, 0.0f); 	// S->G
		 coords[4] = new Point3f(0.0f, 0.0f, 0.0f); 	// S->B
		 coords[5] = new Point3f(0.0f, 0.0f, len); 	// S->B
		 
		 coords[6] = new Point3f(len, len, len); 		// H->C
		 coords[7] = new Point3f(0.0f, len, len); 		// H->C
		 coords[8] = new Point3f(len, len, len); 		// H->M
		 coords[9] = new Point3f(len, 0.0f, len); 		// H->M
		 coords[10] = new Point3f(len, len, len);		// H->Y
		 coords[11] = new Point3f(len, len, 0.0f); 	// H->Y
		 
		 coords[12] = new Point3f(0.0f, len, 0.0f);	// G->Y
		 coords[13] = new Point3f(len, len, 0.0f); 	// G->Y
		 coords[14] = new Point3f(0.0f, len, 0.0f);	// G->C
		 coords[15] = new Point3f(0.0f, len, len); 	// G->C
		 
		 coords[16] = new Point3f(len, 0.0f, 0.0f);	// R->Y
		 coords[17] = new Point3f(len, len, 0.0f); 	// R->Y
		 coords[18] = new Point3f(len, 0.0f, 0.0f);	// R->M
		 coords[19] = new Point3f(len, 0.0f, len); 	// R->M
		 
		 coords[20] = new Point3f(0.0f, 0.0f, len);	// B->C
		 coords[21] = new Point3f(0.0f, len, len); 	// B->C
		 coords[22] = new Point3f(0.0f, 0.0f, len);	// B->M
		 coords[23] = new Point3f(len, 0.0f, len); 	// B->M
		 
		 //----------------------------------------------------
		 // colors	
		 color[0] = new Color3f(0.0f, 0.0f, 0.0f); //
		 color[1] = new Color3f(1.0f, 0.0f, 0.0f); //
		 color[2] = new Color3f(0.0f, 0.0f, 0.0f); // 
		 color[3] = new Color3f(0.0f, 1.0f, 0.0f); //
		 color[4] = new Color3f(0.0f, 0.0f, 0.0f); //
		 color[5] = new Color3f(0.0f, 0.0f, 1.0f);	//

		 color[6] = new Color3f(1.0f, 1.0f, 1.0f);	//
		 color[7] = new Color3f(0.0f, 1.0f, 1.0f); //
		 color[8] = new Color3f(1.0f, 1.0f, 1.0f); // 
		 color[9] = new Color3f(1.0f, 0.0f, 1.0f);	//
		 color[10] = new Color3f(1.0f, 1.0f, 1.0f);//
		 color[11] = new Color3f(1.0f, 1.0f, 0.0f);//		
		 
		 color[12] = new Color3f(0.0f, 1.0f, 0.0f);	// G->Y
		 color[13] = new Color3f(1.0f, 1.0f, 0.0f); 	// G->Y
		 color[14] = new Color3f(0.0f, 1.0f, 0.0f);	// G->C
		 color[15] = new Color3f(0.0f, 1.0f, 1.0f); 	// G->C
		 
		 color[16] = new Color3f(1.0f, 0.0f, 0.0f);	// R->Y
		 color[17] = new Color3f(1.0f, 1.0f, 0.0f); 	// R->Y
		 color[18] = new Color3f(1.0f, 0.0f, 0.0f);	// R->M
		 color[19] = new Color3f(1.0f, 0.0f, 1.0f); 	// R->M
		 
		 color[20] = new Color3f(0.0f, 0.0f, 1.0f);	// B->C
		 color[21] = new Color3f(0.0f, 1.0f, 1.0f); 	// B->C
		 color[22] = new Color3f(0.0f, 0.0f, 1.0f);	// B->M
		 color[23] = new Color3f(1.0f, 0.0f, 1.0f); 	// B->M
		 
		 lineData.setCoordinates(0, coords);
		 lineData.setColors(0, color);
		 
		 
		 Shape3D tmpShape = new Shape3D(lineData, StandardAppearance.getLineAppearance(null));
		 tmp.addChild(tmpShape);
		 
		 return tmp;
		 */
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */		
	public String getName() {
		return "sRGB";		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 2;
	}
	
	/**
	 * Creates a parameter panel for this ColorSpace3D.
	 */		
	private void createParameterPanel() {
		parameterPanel = new JPanel();
		parameterPanel.add(new JLabel("sRGB"));
	}

	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */	
	public Component getParameterPanel() {
		return parameterPanel;
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f, 1.0f, 1.0f};
	}
	
	/**
	 * Returns the XYZ white point of the color space.
	 * @return A float array of length 3, or null if no white point is defined.
	 */
	public float[] getWhitePoint() {
		return whitePoint;
	}
}
