/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces;


import java.awt.*;
import javax.media.j3d.*;
import javax.swing.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;


/**
 * This class is used in conjunction with operations involving XYZ color space.
 * Converting colorvalues from other colorspaces to XYZ color space and vice versa.
 */
public class XYZ_ColorSpace3D extends ColorSpace3D {
	
	/** The CIEXYZ white point X value. */
	public static final float whiteX = 0.9642f;
	
	/** The CIEXYZ white point Y value. */
	public static final float whiteY = 1.0000f;
	
	/** The CIEXYZ white point Z value. */
	public static final float whiteZ = 0.8249f;
	
	/** A sRGB color space. */
	private ColorSpace3D rgbSpace;
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/**
	 * Default constructor.
	 */
	public XYZ_ColorSpace3D() {
		rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
		colorLabels = new char[] {'X', 'Y', 'Z'};
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		System.arraycopy(color, 0, xyz, 0, color.length);
		return xyz;
	}
	
	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		float retColor[] = new float[xyz.length];

		System.arraycopy(xyz, 0, retColor, 0, xyz.length);
		return retColor;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		System.arraycopy(ciexyz, 0, retColor, 0, ciexyz.length);
		return retColor;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}

	/**
	 * Converts CIEXYZ color to sRGB color.
	 *
	 * @param color CIEXYZ color
	 * @return sRGB color
	 */
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}
	
	/**
	 * Creates a new BranchGroup which shows XYZ axis.
	 *
	 * @return XYZ axis
	 */	
	public BranchGroup getAxis() {
		return getStandardBranch();
	}

	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */			
	public String getName() {
		return "XYZ";
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 3;
	}
	
	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */		
	public Component getParameterPanel() {
		return getStandardParameterPanel();
	}

	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */			
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return false;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f, 1.0f, 1.0f};
	}
}
