/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces.generic;


import java.awt.*;
import java.awt.event.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.SRGB_ColorSpace3D;
import icc3dpackage.managers.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;
import icc3dpackage.visualizations.appearance.*;


/**
 * This abstract class is used as a base for all other colorspaces.
 */
public abstract class ColorSpace3D extends Plugin {
	
	/** An sRGB color space */
	private static ColorSpace3D rgbSpace = new SRGB_ColorSpace3D();
	
	/** Detail level of lines (how smooth a line is drawn) */
	protected int detail;
	
	/** Detail level for wireframe view of the color space */
	protected int numSlices;
	
	/** Reference to the color space parameter panel */
	protected JPanel standardPanel;
	
	/** Display color cube */
	protected JCheckBox showBox;
	
	/** Used to determine which axis system is to be displayed */
	protected JComboBox useAxis;
	
	/** Reference to the colorcube surrounding the color space */
	protected BranchGroup colorSpaceCubeBranch;
	
	/** Reference to the wireframe view of the color space */
	protected BranchGroup sRGB_LimitBranch;
	
	/** Reference to the stadard view of the color space */
	protected BranchGroup standardBranch;
	
	/**
	 * Default constructor.
	 */
	public ColorSpace3D() {
		detail = 25;
		numSlices = 5;
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public abstract int getNumberOfComponents();
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public abstract void signalMovement();
	
	/**
	 * Converts colors in the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public abstract float[] toCIEXYZ(float color[]);
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public abstract float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]);

	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public abstract float[] toPresentation(float color[]);
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 *
	 * @return The scaling factors.
	 */
	public abstract float[] getScale();
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public abstract float[] fromPresentation(float color[]);

	/**
	 * Returns the color axis labels.
	 *
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public abstract char[] getColorLabels();
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public abstract float[] fromCIEXYZ(float xyz[]);
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public abstract float[] fromCIEXYZ(float xyz[], float retColor[], boolean dirtyColor[]);
	
	/**
	 * Converts colors in the given input color space to sRGB colors.
	 *
	 * @param color The input colors.
	 * @return The sRGB colors.
	 */
	public abstract byte[] toSRGB(float color[]);
	
	/**
	 * Creates a new BranchGroup which shows(if selected by user): Lab cicle and Lab axis
	 *
	 * @return Lab axis
	 */	
	public abstract BranchGroup getAxis();
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return The name.
	 */	
	public abstract String getName();
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 *
	 * @return True if the color space is based on an ICC profile.
	 */
	public abstract boolean isProfile();
	
	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */		
	public abstract Component getParameterPanel();
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */
	public abstract Transform3D getTransform();
	
	/**
	 * Returns the XYZ white point of the color space (if applicable).
	 *
	 * @return A float array of length 3, or null if no white point is defined.
	 */
	public float[] getWhitePoint() {
		return null;
	}
	
	/**
	 * Creates a standard parameter panel used by several derived classes.
	 * Every derived class will have these standard parameter options if chosen.
	 */
	private void createStandardParameterPanel() {
		standardPanel = new JPanel(new BorderLayout());
		
		JPanel standardPanelInner = new JPanel(new GridLayout(0, 2));

		showBox = new JCheckBox();
		showBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (colorSpaceCubeBranch != null) {
					colorSpaceCubeBranch.detach();
					colorSpaceCubeBranch = null;
				}
				if (showBox.isSelected() && standardBranch != null) {
					colorSpaceCubeBranch = getColorSpaceLimit();
					colorSpaceCubeBranch.setCapability(BranchGroup.ALLOW_DETACH);
					standardBranch.addChild(colorSpaceCubeBranch);
				}
			}
		});
		standardPanelInner.add(new JLabel("Color cube"));
		standardPanelInner.add(showBox);
		String choices[] = {"sRGB line limits", "None", "sRGB Wireframe"};

		useAxis = new JComboBox(choices);
		useAxis.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				if (sRGB_LimitBranch != null) {
					sRGB_LimitBranch.detach();
					sRGB_LimitBranch = null;
				}
				if (useAxis.getSelectedIndex() == 0 && standardBranch != null) {
					sRGB_LimitBranch = getStandardAxis();
					sRGB_LimitBranch.setCapability(BranchGroup.ALLOW_DETACH);
					standardBranch.addChild(sRGB_LimitBranch);
				}
				if (useAxis.getSelectedIndex() == 2 && standardBranch != null) {
					sRGB_LimitBranch = getWireFrameLimits();
					sRGB_LimitBranch.setCapability(BranchGroup.ALLOW_DETACH);
					standardBranch.addChild(sRGB_LimitBranch);
				}
			}
		});
		standardPanelInner.add(new JLabel("Display"));
		standardPanelInner.add(useAxis);
		standardPanel.add(standardPanelInner, BorderLayout.NORTH);
		standardPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Retrieves the standard parameter panel.
	 *
	 * @return standard parameter panel
	 */	
	public Component getStandardParameterPanel() {
		if (standardPanel == null) {
			createStandardParameterPanel();
		}
		return standardPanel;
	}
	
	/**
	 * Creates a new BranchGroup consisting of all requested visual aids from the user.
	 *
	 * @return The BranchGroup.
	 */	
	public BranchGroup getStandardBranch() {
		if (standardPanel == null) {
			createStandardParameterPanel();
		}
		standardBranch = new BranchGroup();
		standardBranch.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		standardBranch.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		colorSpaceCubeBranch = null;
		sRGB_LimitBranch = null;
		if (showBox.isSelected()) {
			colorSpaceCubeBranch = getColorSpaceLimit();
			colorSpaceCubeBranch.setCapability(BranchGroup.ALLOW_DETACH);
			standardBranch.addChild(colorSpaceCubeBranch);
		}
		if (useAxis.getSelectedIndex() == 0) {
			sRGB_LimitBranch = getStandardAxis();
			sRGB_LimitBranch.setCapability(BranchGroup.ALLOW_DETACH);
			standardBranch.addChild(sRGB_LimitBranch);
		}
		if (useAxis.getSelectedIndex() == 2) {
			sRGB_LimitBranch = getWireFrameLimits();
			sRGB_LimitBranch.setCapability(BranchGroup.ALLOW_DETACH);
			standardBranch.addChild(sRGB_LimitBranch);
		}
		return standardBranch;
	}
	
	/**
	 * Creates a new BranchGroup which shows a box surrounding this color space.
	 *
	 * @return BranchGroup representing color space limits
	 */	
	public BranchGroup getColorSpaceLimit() {
		BranchGroup branchGroup = new BranchGroup();
		
		LineArray lineArray = new LineArray(12 * 2, GeometryArray.COORDINATES);

		lineArray.setCoordinate(0, new float[] {0, 0, 0});
		lineArray.setCoordinate(1, new float[] {0, 0, 1});
		lineArray.setCoordinate(2, new float[] {0, 1, 0});
		lineArray.setCoordinate(3, new float[] {0, 1, 1});
		lineArray.setCoordinate(4, new float[] {1, 0, 0});
		lineArray.setCoordinate(5, new float[] {1, 0, 1});
		lineArray.setCoordinate(6, new float[] {1, 1, 0});
		lineArray.setCoordinate(7, new float[] {1, 1, 1});
		lineArray.setCoordinate(8, new float[] {0, 0, 0});
		lineArray.setCoordinate(9, new float[] {1, 0, 0});
		lineArray.setCoordinate(10, new float[] {0, 1, 0});
		lineArray.setCoordinate(11, new float[] {1, 1, 0});
		lineArray.setCoordinate(12, new float[] {0, 0, 1});
		lineArray.setCoordinate(13, new float[] {1, 0, 1});
		lineArray.setCoordinate(14, new float[] {0, 1, 1});
		lineArray.setCoordinate(15, new float[] {1, 1, 1});
		lineArray.setCoordinate(16, new float[] {0, 0, 0});
		lineArray.setCoordinate(17, new float[] {0, 1, 0});
		lineArray.setCoordinate(18, new float[] {0, 0, 1});
		lineArray.setCoordinate(19, new float[] {0, 1, 1});
		lineArray.setCoordinate(20, new float[] {1, 0, 0});
		lineArray.setCoordinate(21, new float[] {1, 1, 0});
		lineArray.setCoordinate(22, new float[] {1, 0, 1});
		lineArray.setCoordinate(23, new float[] {1, 1, 1});
				
		Shape3D shape3D = new Shape3D(lineArray,
				StandardAppearance.getLineAppearance());

		branchGroup.addChild(shape3D);
		return branchGroup;
	}
		
	/**
	 * Creates a new BranchGroup which shows the current color space as a wireframe.
	 *
	 * @return A wireframe representation of the color space.
	 */
	public BranchGroup getWireFrameLimits() {
		BranchGroup branchGroup = new BranchGroup();
		float startPoint[];
		float endPoint[];
		int vertexCount[] = new int[3 * (numSlices + 1) * (numSlices + 1)];

		for (int i = 0; i < 3 * (numSlices + 1) * (numSlices + 1); i++) {
			vertexCount[i] = detail + 1;
		}
	
		LineStripArray lineStripArray = new LineStripArray((detail + 1) * 3
				* (numSlices + 1) * (numSlices + 1),
				GeometryArray.COORDINATES | GeometryArray.COLOR_3, vertexCount);
		int startIndex = 0;
	
		float division = 1.0f / (float) numSlices;
		
		for (float k = 0; k <= 1.0001f; k += division) {
			for (float l = 0; l <= 1.0001f; l += division) {
				startPoint = new float[] {k, 0, l};
				endPoint = new float[] {k, 1, l};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
				startPoint = new float[] {k, l, 0};
				endPoint = new float[] {k, l, 1};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
				startPoint = new float[] {0, k, l};
				endPoint = new float[] {1, k, l};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
			}
		}

		Shape3D tmpShape = new Shape3D(lineStripArray,
				StandardAppearance.getLineAppearance());

		branchGroup.addChild(tmpShape);
		
		return branchGroup;
	}
	
	/**
	 * Creates a new BranchGroup which shows the boundary limits of this color space.
	 *
	 * @return The color space boundary.
	 */
	public BranchGroup getStandardAxis() {
		BranchGroup branchGroup = new BranchGroup();
		float startPoint[];
		float endPoint[];
		int vertexCount[] = new int[12];

		for (int i = 0; i < 12; i++) {
			vertexCount[i] = detail + 1;
		}
	
		LineStripArray lineStripArray = new LineStripArray((detail + 1) * 12,
				GeometryArray.COORDINATES | GeometryArray.COLOR_3, vertexCount);
		int startIndex = 0;
	
		for (int k = 0; k <= 1; k++) {
			for (int l = 0; l <= 1; l++) {
				startPoint = new float[] {k, 0, l};
				endPoint = new float[] {k, 1, l};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
				startPoint = new float[] {k, l, 0};
				endPoint = new float[] {k, l, 1};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
				startPoint = new float[] {0, k, l};
				endPoint = new float[] {1, k, l};
				makeLine(startPoint, endPoint, detail, lineStripArray, startIndex);
				startIndex += (detail + 1);
			}
		}

		Shape3D tmpShape = new Shape3D(lineStripArray,
				StandardAppearance.getLineAppearance());

		branchGroup.addChild(tmpShape);
		
		return branchGroup;
	}

	/**
	 * Creates a line.
	 *
	 * @param startPoint The start position of the line.
	 * @param endPoint The end position of the line.
	 * @param detail A value that determines the number of subdivisions used when
	 *   creating the line.
	 * @param lineStripArray The LineStripArray where the line coordinates and
	 *   colors are stored.
	 * @param startIndex The first index of the unused array positions.
	 */
	private void makeLine(float startPoint[], float endPoint[], int detail, LineStripArray lineStripArray, int startIndex) {
		float direction[] = new float[3];
		int i = 0;

		for (i = 0; i < 3; i++) {
			direction[i] = endPoint[i] - startPoint[i];
			direction[i] = direction[i] / (float) detail;
		}
		float newStart[] = new float[] {startPoint[0], startPoint[1], startPoint[2]};
		float inColorSpace[];

		for (i = 0; i <= detail; i++) {
			inColorSpace = ColorSpaceConverter.convert(newStart, rgbSpace, this);
			lineStripArray.setCoordinate(startIndex, inColorSpace);
			lineStripArray.setColor(startIndex, newStart);
			for (int j = 0; j < 3; j++) {
				newStart[j] += direction[j];
			}
			startIndex++;
		}
	}
}
