/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces.generic;


import java.awt.color.*;
import sun.awt.color.ICC_Transform;


/**
 * A color space that attempts to increase the speed slightly when converting to
 * and from ICC profile based color spaces.
 */
public class ExtendedICC_ColorSpace extends ICC_ColorSpace {
	
	/** A floating point representation of 2 - 1/32768 */
	private static final float ALMOST_TWO = 1.0f + (32767.0f / 32768.0f);
	
	/** The minimum values for each component */
	private float[] minVal2;
	
	/** The maximum values for each component */
	private float[] maxVal2;
	
	/** The difference between maximum and minimum values for each component */
	private float[] diffMinMax2;
	
	/** The inverted difference between maximum and minimum values for each component */
	private float[] invDiffMinMax2;
	
	/** Min and max values for components need to be computed once */
	private boolean needScaleInit2 = true;
	
	/** A transform to sRGB */
	private transient ICC_Transform this2srgb2;
	
	/** A transform from sRGB */
	private transient ICC_Transform srgb2this2;
	
	/** A transform to CIEXYZ */
	private transient ICC_Transform this2xyz2;
	
	/** A transform from CIEXYZ */
	private transient ICC_Transform xyz2this2;
	
	/**
	 * Constructs an extended ICC color space from an ICC prorfile.
	 * @param profile The ICC profile.
	 */
	public ExtendedICC_ColorSpace(ICC_Profile profile) {
		super(profile);
		setMinMax2();
	}
	
	/**
	 * Converts colors in the given color space to CIEXYZ.
	 *
	 * @param colorvalue The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float colorvalue[]) {
		if (this2xyz2 == null) {
			ICC_Transform[] transformList = new ICC_Transform[2];
			ICC_ColorSpace xyzCS = (ICC_ColorSpace) ColorSpace.getInstance(CS_CIEXYZ);

			try {
				transformList[0] = new ICC_Transform(this.getProfile(),
						ICC_Profile.icRelativeColorimetric, ICC_Transform.In);
			} catch (CMMException e) {
				transformList[0] = new ICC_Transform(this.getProfile(), ICC_Transform.Any,
						ICC_Transform.In);
			}
			transformList[1] = new ICC_Transform(xyzCS.getProfile(), ICC_Transform.Any,
					ICC_Transform.Out);
			this2xyz2 = new ICC_Transform(transformList);
			if (needScaleInit2) {
				setComponentScaling2();
			}
		}

		int nc = this.getNumComponents();
		int numColors = colorvalue.length / nc;
		short tmp[] = new short[nc];
		short tmp2[] = new short[3];
		short orig[] = tmp2;
		float result[] = new float[numColors * 3];
		float limit;

		for (int i = 0; i < numColors; i++) {
			for (int j = 0; j < nc; j++) {
				limit = (colorvalue[i * nc + j] - minVal2[j]) * invDiffMinMax2[j] + 0.5f;
				if (limit < 0) {
					limit = 0;
				} else if (limit > 65535) {
					limit = 65535;
				}
				tmp[j] = (short) limit;
			}
			tmp2 = this2xyz2.colorConvert(tmp, tmp2);
			for (int j = 0; j < 3; j++) {
				result[i * 3 + j] = (((float) (tmp2[j] & 0xffff)) / 65535.0f) * ALMOST_TWO;
			}
		}	
		return result;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param result The CIEXYZ color data.
	 * @param colorvalue The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float colorvalue[], float result[], boolean dirtyColor[]) {
		if (this2xyz2 == null) {
			ICC_Transform[] transformList = new ICC_Transform[2];
			ICC_ColorSpace xyzCS = (ICC_ColorSpace) ColorSpace.getInstance(CS_CIEXYZ);

			try {
				transformList[0] = new ICC_Transform(this.getProfile(),
						ICC_Profile.icRelativeColorimetric, ICC_Transform.In);
			} catch (CMMException e) {
				transformList[0] = new ICC_Transform(this.getProfile(), ICC_Transform.Any,
						ICC_Transform.In);
			}
			transformList[1] = new ICC_Transform(xyzCS.getProfile(), ICC_Transform.Any,
					ICC_Transform.Out);
			this2xyz2 = new ICC_Transform(transformList);
			if (needScaleInit2) {
				setComponentScaling2();
			}
		}

		int nc = this.getNumComponents();
		int numColors = colorvalue.length / nc;
		short tmp[] = new short[nc];
		short tmp2[] = new short[3];
		short orig[] = tmp2;
		float limit;
		
		for (int i = 0; i < numColors; i++) {
			if (dirtyColor[i]) {
				for (int j = 0; j < nc; j++) {
					limit = (colorvalue[i * nc + j] - minVal2[j]) * invDiffMinMax2[j] + 0.5f;
					if (limit < 0) {
						limit = 0;
					} else if (limit > 65535) {
						limit = 65535;
					}
					tmp[j] = (short) limit;
				}
				tmp2 = this2xyz2.colorConvert(tmp, tmp2);
				for (int j = 0; j < 3; j++) {
					result[i * 3 + j] = (((float) (tmp2[j] & 0xffff)) / 65535.0f) * ALMOST_TWO;
				}
			}
		}	
		return result;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param colorvalue The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float colorvalue[]) {
		if (xyz2this2 == null) {
			ICC_Transform[] transformList = new ICC_Transform[2];
			ICC_ColorSpace xyzCS = (ICC_ColorSpace) ColorSpace.getInstance(CS_CIEXYZ);

			transformList[0] = new ICC_Transform(xyzCS.getProfile(), ICC_Transform.Any,
					ICC_Transform.In);
			try {
				transformList[1] = new ICC_Transform(this.getProfile(),
						ICC_Profile.icRelativeColorimetric, ICC_Transform.Out);
			} catch (CMMException e) {
				transformList[1] = new ICC_Transform(this.getProfile(), ICC_Transform.Any,
						ICC_Transform.Out);
			}
			xyz2this2 = new ICC_Transform(transformList);
			if (needScaleInit2) {
				setComponentScaling2();
			}
		}
		
		int numColors = colorvalue.length / 3;
		int nc = this.getNumComponents();
		float result[] = new float[numColors * nc];
		short tmp[] = new short[3];
		short tmp2[] = new short[nc];
		float factor = 65535.0f / ALMOST_TWO;
		float limit;

		for (int i = 0; i < numColors; i++) {
			for (int j = 0; j < 3; j++) {
				limit = (colorvalue[i * 3 + j] * factor) + 0.5f;
				if (limit < 0) {
					limit = 0;
				} else if (limit > 65535) {
					limit = 65535;
				}
				tmp[j] = (short) limit;
			}
			tmp2 = xyz2this2.colorConvert(tmp, tmp2);
			for (int j = 0; j < nc; j++) {
				result[i * nc + j] = (((float) (tmp2[j] & 0xffff)) / 65535.0f)
								* diffMinMax2[j]
						+ minVal2[j];
			}
		}
		return result;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param colorvalue The CIEXYZ color data.
	 * @param result The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float colorvalue[], float result[], boolean dirtyColor[]) {
		if (xyz2this2 == null) {
			ICC_Transform[] transformList = new ICC_Transform[2];
			ICC_ColorSpace xyzCS = (ICC_ColorSpace) ColorSpace.getInstance(CS_CIEXYZ);

			transformList[0] = new ICC_Transform(xyzCS.getProfile(), ICC_Transform.Any,
					ICC_Transform.In);
			try {
				transformList[1] = new ICC_Transform(this.getProfile(),
						ICC_Profile.icRelativeColorimetric, ICC_Transform.Out);
			} catch (CMMException e) {
				transformList[1] = new ICC_Transform(this.getProfile(), ICC_Transform.Any,
						ICC_Transform.Out);
			}
			xyz2this2 = new ICC_Transform(transformList);
			if (needScaleInit2) {
				setComponentScaling2();
			}
		}
		
		int numColors = colorvalue.length / 3;
		int nc = this.getNumComponents();
		short tmp[] = new short[3];
		short tmp2[] = new short[nc];
		float factor = 65535.0f / ALMOST_TWO;
		float limit;

		for (int i = 0; i < numColors; i++) {
			if (dirtyColor[i]) {
				for (int j = 0; j < 3; j++) {
					limit = (colorvalue[i * 3 + j] * factor) + 0.5f;
					if (limit < 0) {
						limit = 0;
					} else if (limit > 65535) {
						limit = 65535;
					}
					tmp[j] = (short) limit;
				}
				tmp2 = xyz2this2.colorConvert(tmp, tmp2);
				for (int j = 0; j < nc; j++) {
					result[i * nc + j] = (((float) (tmp2[j] & 0xffff)) / 65535.0f)
									* diffMinMax2[j]
							+ minVal2[j];
				}
			}
		}
		return result;
	}
	
	/**
	 * Computes the difference between min and max component values.
	 */
	private void setComponentScaling2() {
		int nc = this.getNumComponents();

		diffMinMax2 = new float[nc];
		invDiffMinMax2 = new float[nc];
		for (int i = 0; i < nc; i++) {
			minVal2[i] = this.getMinValue(i);
			maxVal2[i] = this.getMaxValue(i);
			diffMinMax2[i] = maxVal2[i] - minVal2[i];
			invDiffMinMax2[i] = 65535.0f / diffMinMax2[i];
		}
		needScaleInit2 = false;
	}
	
	/**
	 * Sets the min and max component values.
	 */
	private void setMinMax2() {
		int nc = this.getNumComponents();
		int type = this.getType();

		minVal2 = new float[nc];
		maxVal2 = new float[nc];
		if (type == ColorSpace.TYPE_Lab) {
			minVal2[0] = 0.0f;    // L
			maxVal2[0] = 100.0f;
			minVal2[1] = -128.0f; // a
			maxVal2[1] = 127.0f;
			minVal2[2] = -128.0f; // b
			maxVal2[2] = 127.0f;
		} else if (type == ColorSpace.TYPE_XYZ) {
			minVal2[0] = minVal2[1] = minVal2[2] = 0.0f; // X, Y, Z
			maxVal2[0] = maxVal2[1] = maxVal2[2] = 1.0f + (32767.0f / 32768.0f);
		} else {
			for (int i = 0; i < nc; i++) {
				minVal2[i] = 0.0f;
				maxVal2[i] = 1.0f;
			}
		}
	}
}
