/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces.generic;


import java.awt.*;
import java.awt.color.*;
import java.io.File;
import java.util.Arrays;
import javax.media.j3d.*;
import javax.swing.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;


/**
 * This class represents a color space based on an ICC profile.
 * Converting colorvalues from other color spaces to this color space and vice versa.
 *
 */
public class ICC_ProfileColorSpace3D extends ColorSpace3D {
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** The ICC profile. */
	private ICC_Profile profile;
	
	/** The profile based color space. */
	private ExtendedICC_ColorSpace colorSpace;
	
	/** A sRGB color space. */
	private ColorSpace3D rgbSpace;
	
	/** The name/description of the color space. */
	private String name;
	
	/** The ICC profile file name. */
	private String fileName;
	
	/** The CIEXYZ white point defined in the profile. */
	private float[] whitePoint;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/**
	 * Constructs a color space from the given ICC profile.
	 */
	public ICC_ProfileColorSpace3D(String file) {
		try {
			profile = ICC_Profile.getInstance(file);
			byte b[] = profile.getData(ICC_Profile.icSigMediaWhitePointTag);

			if (b != null && b.length >= 4 + 4 + 3 * 4) {
				whitePoint = new float[3];
				for (int i = 0; i < 3; i++) {
					int o = (b[8 + i * 4]) & (0xff);

					o = o << 8;
					o += (b[9 + i * 4]) & (0xff);
					o = o << 8;
					o += (b[10 + i * 4]) & (0xff);
					o = o << 8;
					o += (b[11 + i * 4]) & (0xff);
					float f;

					f = o;
					f = f / 65536.0f;
					whitePoint[i] = f;
				}
			}
			colorSpace = new ExtendedICC_ColorSpace(profile);
			File f = new File(file);

			name = f.getName().substring(0, Math.min(f.getName().length(), 20));
			fileName = file;
			rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
		} catch (Exception e) {
			throw new RuntimeException("Unable to create color space from profile");
		}
		colorLabels = new char[] {'_', '_', '_'};
	}
	
	/**
	 * Constructs a color space from the given ICC profile.
	 */
	public ICC_ProfileColorSpace3D(ICC_Profile p) {
		try {
			profile = p;
			byte b[] = profile.getData(ICC_Profile.icSigMediaWhitePointTag);

			if (b != null && b.length >= 4 + 4 + 3 * 4) {
				whitePoint = new float[3];
				for (int i = 0; i < 3; i++) {
					int o = (b[8 + i * 4]) & (0xff);

					o = o << 8;
					o += (b[9 + i * 4]) & (0xff);
					o = o << 8;
					o += (b[10 + i * 4]) & (0xff);
					o = o << 8;
					o += (b[11 + i * 4]) & (0xff);
					float f;

					f = o;
					f = f / 65536.0f;
					whitePoint[i] = f;
				}
			}
			colorSpace = new ExtendedICC_ColorSpace(profile);
			name = "";
			rgbSpace = ColorSpaceManager.getColorSpace("sRGB");
		} catch (Exception e) {
			throw new RuntimeException("Unable to create color space from profile");
		}
		colorLabels = new char[] {'_', '_', '_'};
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return profile.getNumComponents();
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}

	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		return colorSpace.fromCIEXYZ(xyz);
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		return colorSpace.fromCIEXYZ(ciexyz, retColor, dirtyColor);
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		return colorSpace.toCIEXYZ(color, xyz, dirtyColor);
	}

	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		return colorSpace.toCIEXYZ(color);
	}

	/**
	 * Converts color to sRGB(byte) color.
	 *
	 * @param color ICC profile based color
	 * @return sRGB byte color
	 */	
	public byte[] toSRGB(float color[]) {
		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}
	
	/**
	 * Creates a new BranchGroup which shows sRGB axis.
	 *
	 * @return sRGB axis
	 */		
	public BranchGroup getAxis() {
		return getStandardBranch();
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */		
	public String getName() {
		return name;		
	}
	
	/**
	 * Returns the name of the file used to create this ICC_ProfileColorSpace3D
	 *
	 * @return The file name
	 */		
	public String getFileName() {
		return fileName;		
	}
	
	/**
	 * The priority of the color space.
	 * @return An int representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 1000;
	}

	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */	
	public Component getParameterPanel() {
		return getStandardParameterPanel();
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return true;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		float f[] = new float[this.getNumberOfComponents()];

		Arrays.fill(f, 1.0f);
		return f;
	}
	
	/**
	 * Returns the XYZ white point of the color space.
	 * @return A float array of length 3, or null if no white point is defined.
	 */
	public float[] getWhitePoint() {
		return whitePoint;
	}
}
