/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.colorspaces.generic;


import javax.vecmath.*;
import javax.media.j3d.*;
import java.io.*;
import java.awt.color.*;
import java.awt.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.modules.*;
import icc3dpackage.utils.*;


/**
 * This class represents a color space based on polynomial regression data.
 */
public class PolynomialColorSpace3D extends ColorSpace3D {
	
	/** The color axis labels. */
	private char[] colorLabels;
	
	/** The name/description of the color space. */
	private String name;
	
	/** The name of the file that contains the regression data. */
	private String fileName;
	
	/** The order of the polynomial. */
	private int order;
	
	/** Regression parameter. */
	private float xyzToRGBx[], xyzToRGBy[], xyzToRGBz[], rgbToXYZr[], rgbToXYZg[], rgbToXYZb[];
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/**
	 * Constructs a color space from the given data file.
	 */
	public PolynomialColorSpace3D(String file) {
		try {
			BufferedReader br = new BufferedReader(new java.io.FileReader(file));
			String s = br.readLine();

			order = Integer.parseInt(s);
			rgbToXYZr = new float[3 * order + 1];
			rgbToXYZg = new float[3 * order + 1];
			rgbToXYZb = new float[3 * order + 1];
			xyzToRGBx = new float[3 * order + 1];
			xyzToRGBy = new float[3 * order + 1];
			xyzToRGBz = new float[3 * order + 1];
			
			for (int j = 0; j < order * 3 + 1; j++) {
				rgbToXYZr[j] = Float.parseFloat(br.readLine());
			}
			
			for (int j = 0; j < order * 3 + 1; j++) {
				rgbToXYZg[j] = Float.parseFloat(br.readLine());
			}
			
			for (int j = 0; j < order * 3 + 1; j++) {
				rgbToXYZb[j] = Float.parseFloat(br.readLine());
			}
			
			for (int j = 0; j < order * 3 + 1; j++) {
				xyzToRGBx[j] = Float.parseFloat(br.readLine());
			}
			
			for (int j = 0; j < order * 3 + 1; j++) {
				xyzToRGBy[j] = Float.parseFloat(br.readLine());
			}
			
			for (int j = 0; j < order * 3 + 1; j++) {
				xyzToRGBz[j] = Float.parseFloat(br.readLine());
			}
			br.close();
			File f = new File(file);

			name = f.getName().substring(0, Math.min(f.getName().length(), 20));
			fileName = file;
		} catch (Exception e) {
			throw new RuntimeException("Unable to create color space from file.");
		}
		colorLabels = new char[] {'R', 'G', 'B'};
		createParameterPanel();
	}
	
	/**
	 * Returns the number of components (samples per pixel)
	 * in this color space;
	 */
	public int getNumberOfComponents() {
		return 3;
	}
	
	/**
	 * Converts a color from its presentation form.
	 *
	 * @param color Colors in a form suitable for presentation.
	 * @return Colors in the interal color format.
	 */
	public float[] fromPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Converts a color to its presentation form.
	 *
	 * @param color Colors in the interal color representation.
	 * @return Colors in a form suitable for presentation.
	 */
	public float[] toPresentation(float color[]) {
		float retColor[] = new float[color.length];

		System.arraycopy(color, 0, retColor, 0, color.length);
		return retColor;
	}
	
	/**
	 * Returns the color axis labels.
	 * @return The color axis labels. (For example L, a, b for the CIELAB color
	 * space);
	 */
	public char[] getColorLabels() {
		return colorLabels;
	}

	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param xyz The CIEXYZ color data.
	 * @return The output color
	 */
	public float[] fromCIEXYZ(float xyz[]) {
		int q, p, count = xyz.length / 3;
		float retColor[] = new float[count * 3];

		for (int i = 0; i < count; i++) {
			p = i * 3;
			float col[] = new float[] {xyz[p], xyz[p + 1], xyz[p + 2]};

			retColor[p] = get(col, xyzToRGBx);
			retColor[p + 1] = get(col, xyzToRGBy);
			retColor[p + 2] = get(col, xyzToRGBz);
		}
		return retColor;
	}
	
	private float get(float inVar[], float param[]) {
		float sum = param[0];

		for (int i = 1; i <= order; i++) {
			for (int j = 0; j < 3; j++) {
				sum += Math.pow(inVar[j], i) * param[1 + (i - 1) * 3 + j];
			}
		}
		return sum;
	}
	
	/**
	 * Converts colors from the CIEXYZ color space to the given color space.
	 * @param ciexyz The CIEXYZ color data.
	 * @param retColor The destination color array.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The converted colors.
	 */
	public float[] fromCIEXYZ(float ciexyz[], float retColor[], boolean dirtyColor[]) {
		int q, p, count = ciexyz.length / 3;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				p = i * 3;
				float col[] = new float[] {ciexyz[p], ciexyz[p + 1], ciexyz[p + 2]};

				retColor[p] = get(col, xyzToRGBx);
				retColor[p + 1] = get(col, xyzToRGBy);
				retColor[p + 2] = get(col, xyzToRGBz);
			}
		}
		return retColor;
	}
	
	/**
	 * Converts colors from the given color space to the CIEXYZ color space.
	 *
	 * @param xyz The CIEXYZ color data.
	 * @param color The colors that should be converted.
	 * @param dirtyColor Decides which colors that need to be recalculated.
	 * @return The CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[], float xyz[], boolean dirtyColor[]) {
		int q, p, count = color.length / 3;
		float retColor[] = xyz;

		for (int i = 0; i < count; i++) {
			if (dirtyColor[i]) {
				p = i * 3;
				float col[] = new float[] {color[p], color[p + 1], color[p + 2]};

				retColor[p] = get(col, rgbToXYZr);
				retColor[p + 1] = get(col, rgbToXYZg);
				retColor[p + 2] = get(col, rgbToXYZb);
			}
		}
		return retColor;
	}

	/**
	 * Converts colors from the given color space to CIEXYZ.
	 *
	 * @param color The colors that should be converted to CIEXYZ.
	 * @return The equivalent CIEXYZ colors.
	 */
	public float[] toCIEXYZ(float color[]) {
		int q, p, count = color.length / 3;
		float retColor[] = new float[count * 3];

		for (int i = 0; i < count; i++) {
			p = i * 3;
			float col[] = new float[] {color[p], color[p + 1], color[p + 2]};

			retColor[p] = get(col, rgbToXYZr);
			retColor[p + 1] = get(col, rgbToXYZg);
			retColor[p + 2] = get(col, rgbToXYZb);
		}
		return retColor;
	}

	/**
	 * Converts color to sRGB(byte) color.
	 *
	 * @param color Device specific color
	 * @return sRGB byte color
	 */	
	public byte[] toSRGB(float color[]) {
		ColorSpace3D rgbSpace = ColorSpaceManager.getColorSpace("sRGB");

		return rgbSpace.toSRGB(ColorSpaceConverter.convert(color, this, rgbSpace));
	}
	
	/**
	 * Creates a new BranchGroup which shows sRGB axis.
	 *
	 * @return sRGB axis
	 */		
	public BranchGroup getAxis() {
		return getStandardAxis();
	}
	
	/**
	 * Returns the name of this ColorSpace3D
	 *
	 * @return name
	 */		
	public String getName() {
		return name;		
	}
	
	/**
	 * Returns the name of the file used to create this color space.
	 *
	 * @return The file name
	 */		
	public String getFileName() {
		return fileName;		
	}
	
	/**
	 * The priority of the color space.
	 * @return An integer representing the priority of the color space (lower => higher priority).
	 */
	public int getPriority() {
		return 1000;
	}
	
	/**
	 * Creates a parameter panel for this ColorSpace3D.
	 */		
	private void createParameterPanel() {
		parameterPanel = new JPanel();
	}

	/**
	 * Returns the parameter panel of this ColorSpace3D
	 *
	 * @return parameter panel
	 */	
	public Component getParameterPanel() {
		return parameterPanel;
	}
	
	/**
	 * Returns a Transform3D which describes how objects should be
	 *	scaled in the current ColorSpace3D.
	 *
	 * @return transformation
	 */	
	public Transform3D getTransform() {
		return new Transform3D();
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {}
	
	/**
	 * Used to determine whether the color space is constructed from an ICC profile.
	 * @return True if the color space is based on an ICC profile.
	 */
	public boolean isProfile() {
		return true;
	}
	
	/**
	 * Returns the axis scaling factors used by this color space.
	 * @return The scaling factors.
	 */
	public float[] getScale() {
		return new float[] {1.0f, 1.0f, 1.0f};
	}
}
