/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.image2d.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.*;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.visualizations.ImprovedSegmentMaxima;


/**
 * An implementation of a gamut mapping algorithm that performs clipping
 * of all colors outside the destination gamut towards the point on
 * the achromatic axis that has the same lightness as the color that is being
 * mapped.
 */
public class ClipConstantHueAndLightness extends GamutMappingAlgorithm {
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		
		ColorSpace3D outColor = mo.getColorSpace();
		float scale[] = outColor.getScale();
		
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			clipColors(temporaryStorage, tetra, scale);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					clipColors(pos, tetra, scale);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				clipColors(pos, tetra, scale);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			float colors[] = mo.getColors();	
			Arrays.fill(dirty, true);
			clipColors(colors, tetra, scale);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */	
	public String getName() {
		return "Constant hue and lightness";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */
	public int getPriority() {
		return 12;
	}
	
	/**
	 * Clips all colors to the tetrahedron structure.
	 * @param color The color(s) that should be mapped.
	 * @param tetra The tetrahedrons.
	 * @param scale The color space scale.
	 */
	public void clipColors(float color[], Tetrahedron tetra[], float scale[]) {
		float xDiff, yDiff, zDiff;
		int num = color.length / 3;
		int q;
		Vector3f center = new Vector3f(0.5f, 0.5f, 0.5f);
		Vector3f colorExpanded = new Vector3f();
		Vector3f aboveColorExpanded = new Vector3f();
		ArrayList al = new ArrayList();
		Plane p = new Plane();
		Tetrahedron startSearch = tetra[0];
		
		int pnts;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			while (color[q] == 0.5f || color[q + 2] == 0.5f) {
				color[q] += (Math.random() - 0.5f) / 1000000.0f;
				color[q + 2] += (Math.random() - 0.5f) / 1000000.0f;
			}
			startSearch = startSearch.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			if (startSearch.plane[3].classifyPoint2(color[q], color[q + 1], color[q + 2])
					!= Plane.PLANE_OUTSIDE) {
				continue;
			}
			colorExpanded.x = 0.5f + (color[q] - 0.5f) * 10;
			colorExpanded.y = 0.5f + (color[q + 1] - 0.5f) * 10;
			colorExpanded.z = 0.5f + (color[q + 2] - 0.5f) * 10;
			aboveColorExpanded.x = colorExpanded.x;
			aboveColorExpanded.y = colorExpanded.y + 0.5f;
			aboveColorExpanded.z = colorExpanded.z;

			p.set(colorExpanded, center, aboveColorExpanded);

			float hue = ImprovedSegmentMaxima.hue(colorExpanded.x, colorExpanded.z);

			al.clear();
			if (!Tetrahedron.findBoundary(p, tetra, startSearch, al, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			
			float diffX = (colorExpanded.x - 0.5f) * scale[0];
			float diffZ = (colorExpanded.z - 0.5f) * scale[2];
			float px = (float) Math.sqrt(diffX * diffX + diffZ * diffZ);
			float py = color[q + 1] * scale[1];
			float rx = -px;
			float ry = 0;
			GamutBoundaryPoint gbp = Tetrahedron.findClosestIntersection(al, px, py, rx,
					ry);

			if (gbp == null) {
				color[q] = 0.5f;
				// color[q+1] = 0.5f;
				color[q + 2] = 0.5f;
				continue;
			}
			Vector3f v3f = gbp.getVector(hue, scale);

			color[q] = v3f.x;
			color[q + 1] = v3f.y;
			color[q + 2] = v3f.z;
		}		
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		return new JPanel();
	}
}
