/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.image2d.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.*;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.visualizations.ImprovedSegmentMaxima;


/**
 * An implementation of the cusp2cusp gamut mapping algorithm.
 */
public class Cusp2CuspAlgorithm extends GamutMappingAlgorithm {

	/** The lightness of the reproduction gamut black point */
	private float reproductionBlack;
	
	/** The lightness of the source gamut black point */
	private float originalBlack;
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */	
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		Tetrahedron sourceTetra[] = from;

		if (sourceTetra == null) {
			SystemManager.displayError(
				"Error",
				"Source gamut representation is not suitable for this algorithm");
			return;
		}
		
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		float scale[] = outColor.getScale();

		reproductionBlack = 1000000;
		for (int i = 0; i < tetra.length; i++) {
			Tetrahedron t = tetra[i];

			if (t.cornerpos0_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos0_1;
			}
			if (t.cornerpos1_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos1_1;
			}
			if (t.cornerpos2_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos2_1;
			}
			if (t.cornerpos3_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos3_1;
			}
		}
		reproductionBlack *= scale[1];
		originalBlack = 1000000;
		for (int i = 0; i < sourceTetra.length; i++) {
			Tetrahedron t = sourceTetra[i];

			if (t.cornerpos0_1 < originalBlack) {
				originalBlack = t.cornerpos0_1;
			}
			if (t.cornerpos1_1 < originalBlack) {
				originalBlack = t.cornerpos1_1;
			}
			if (t.cornerpos2_1 < originalBlack) {
				originalBlack = t.cornerpos2_1;
			}
			if (t.cornerpos3_1 < originalBlack) {
				originalBlack = t.cornerpos3_1;
			}
		}
		originalBlack *= scale[1];
		
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			map(temporaryStorage, scale, sourceTetra, tetra);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					map(pos, scale, sourceTetra, tetra);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				map(pos, scale, sourceTetra, tetra);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			float colors[] = mo.getColors();
			Arrays.fill(dirty, true);
			map(colors, scale, sourceTetra, tetra);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */
	public String getName() {
		return "Cusp2Cusp";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */	
	public int getPriority() {
		return 10;
	}
	
	/**
	 * Maps all colors to the tetrahedron structure.
	 * @param color The color(s) that should be mapped.
	 * @param scale The color space scale.
	 * @param originalTetra The source tetrahedrons.
	 * @param destinationTetra The reproduction tetrahedrons.
	 */
	private final void map(float color[], float scale[], Tetrahedron originalTetra[], Tetrahedron destinationTetra[]) {
		int num = color.length / 3;
		int q;
		Vector3f center = new Vector3f(0.5f, 0.5f, 0.5f);
		Vector3f colorExpanded = new Vector3f();
		Vector3f aboveColorExpanded = new Vector3f();
		ArrayList al1 = new ArrayList();
		ArrayList al2 = new ArrayList();
		Plane p = new Plane();
		Tetrahedron startSearch1 = originalTetra[0];
		Tetrahedron startSearch2 = destinationTetra[0];
		
		int pnts;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			while (color[q] == 0.5f || color[q + 2] == 0.5f) {
				color[q] += (Math.random() - 0.5f) / 1000000.0f;
				color[q + 2] += (Math.random() - 0.5f) / 1000000.0f;
			}
			startSearch1 = startSearch1.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			startSearch2 = startSearch2.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			float chroma = SGCK_Algorithm.chroma(color[q], color[q + 1], color[q + 2],
					scale);
			float lightness = color[q + 1] * scale[1];

			colorExpanded.x = 0.5f + (color[q] - 0.5f) * 10;
			colorExpanded.y = 0.5f + (color[q + 1] - 0.5f) * 10;
			colorExpanded.z = 0.5f + (color[q + 2] - 0.5f) * 10;
			aboveColorExpanded.x = colorExpanded.x;
			aboveColorExpanded.y = colorExpanded.y + 0.5f;
			aboveColorExpanded.z = colorExpanded.z;

			p.set(colorExpanded, center, aboveColorExpanded);

			float hue = ImprovedSegmentMaxima.hue(colorExpanded.x, colorExpanded.z);
			
			GamutBoundaryPoint.release(al1);
			al1.clear();
			if (!Tetrahedron.findBoundary(p, originalTetra, startSearch1, al1, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			
			if (i == 2) {
				ArrayList alll1 = new ArrayList();

				Tetrahedron.findBoundary(p, originalTetra, startSearch1, alll1, scale,
						color[q] > 0.5f, color[q + 2] > 0.5f, hue);
				System.out.println("Boundary 1:");
				for (int j = 0; j < al1.size(); j++) {
					GamutBoundaryPoint gbp = (GamutBoundaryPoint) al1.get(j);

					System.out.println("c " + gbp.chroma + ", l " + gbp.y);
				}
				System.out.println("Boundary 2:");
				for (int j = 0; j < alll1.size(); j++) {
					GamutBoundaryPoint gbp = (GamutBoundaryPoint) alll1.get(j);

					System.out.println("c " + gbp.chroma + ", l " + gbp.y);
				}
			}
			
			GamutBoundaryPoint.release(al2);
			al2.clear();
			if (!Tetrahedron.findBoundary(p, destinationTetra, startSearch2, al2, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			
			GamutBoundaryPoint originalCuspPoint = GamutBoundaryPoint.getCuspPoint(al1);
			GamutBoundaryPoint reproductionCuspPoint = GamutBoundaryPoint.getCuspPoint(al2);
			
			float ld = reproductionBlack
					+ (100 - reproductionBlack) * (lightness - originalBlack)
							/ (100 - originalBlack);
			float ly = reproductionBlack
					+ (100 - reproductionBlack) * (originalCuspPoint.y - originalBlack)
							/ (100 - originalBlack);
			float lc = ld
					+ (reproductionCuspPoint.y - ly) * (chroma / originalCuspPoint.chroma);
			float cc = chroma * (reproductionCuspPoint.chroma / originalCuspPoint.chroma);
			
			if (lc > 100 || lc < 0) {
				System.out.println("l " + lightness);
				System.out.println("c " + chroma);
				System.out.println("old black " + originalBlack + ", new black "
						+ reproductionBlack);
				System.out.println("oCusp " + originalCuspPoint.y + ", "
						+ originalCuspPoint.chroma);
				System.out.println("rCusp " + reproductionCuspPoint.y + ", "
						+ reproductionCuspPoint.chroma);
				System.out.println("ld " + ld);
				System.out.println("ly " + ly);
				System.out.println("lc " + lc);
				System.out.println("cc " + cc);
				if (originalCuspPoint.y < -10) {
					for (int j = 0; j < al1.size(); j++) {
						GamutBoundaryPoint gbp = (GamutBoundaryPoint) al1.get(j);

						System.out.println("gbp: " + gbp.chroma + "," + gbp.y);
					}
				}
			}
			
			GamutBoundaryPoint destination = GamutBoundaryPoint.getGBP();

			destination.y = lc;
			destination.chroma = cc;
			Vector3f v3f = destination.getVector(hue, scale);

			color[q] = v3f.x;
			color[q + 1] = v3f.y;
			color[q + 2] = v3f.z;
		}
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		return new JPanel();
	}
}
