/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.image2d.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.*;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.visualizations.ImprovedSegmentMaxima;


/**
 * An implementation of the gamma gamut mapping algorithm.
 */
public class GammaAlgorithm extends GamutMappingAlgorithm {
	private float reproductionBlack;
	private float lowestCommonL;
	private float highestCommonL;
	
	private JTextField factorField;
	private JSlider factorSlider;
	private JPanel parameterPanel;
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */	
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		Tetrahedron sourceTetra[] = from;

		if (sourceTetra == null) {
			SystemManager.displayError(
				"Error",
				"Source gamut representation is not suitable for this algorithm");
			return;
		}
		
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		float scale[] = outColor.getScale();

		reproductionBlack = 1000000;
		for (int i = 0; i < tetra.length; i++) {
			Tetrahedron t = tetra[i];

			if (t.cornerpos0_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos0_1;
			}
			if (t.cornerpos1_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos1_1;
			}
			if (t.cornerpos2_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos2_1;
			}
			if (t.cornerpos3_1 < reproductionBlack) {
				reproductionBlack = t.cornerpos3_1;
			}
		}
		reproductionBlack *= scale[1];
		
		RangeDescriptor rd1 = this.getRange(tetra, scale);
		RangeDescriptor rd2 = this.getRange(sourceTetra, scale);

		lowestCommonL = rd1.min;
		if (rd2.min > lowestCommonL) {
			lowestCommonL = rd2.min;
		}
		highestCommonL = (rd1.max < rd2.max) ? rd1.max : rd2.max;
		
		// System.out.println("low "+lowestCommonL+", high "+highestCommonL);
		
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			map(temporaryStorage, scale, sourceTetra, tetra);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					map(pos, scale, sourceTetra, tetra);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				map(pos, scale, sourceTetra, tetra);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			float colors[] = mo.getColors();
			Arrays.fill(dirty, true);
			map(colors, scale, sourceTetra, tetra);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */
	public String getName() {
		return "Gamma";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */	
	public int getPriority() {
		return 10;
	}
	
	/**
	 * Maps all colors to the tetrahedron structure.
	 * @param color The color(s) that should be mapped.
	 * @param scale The color space scale.
	 * @param originalTetra The source tetrahedrons.
	 * @param destinationTetra The reproduction tetrahedrons.
	 */
	private final void map(float color[], float scale[], Tetrahedron originalTetra[], Tetrahedron destinationTetra[]) {
		int num = color.length / 3;
		int q;
		Vector3f center = new Vector3f(0.5f, 0.5f, 0.5f);
		Vector3f colorExpanded = new Vector3f();
		Vector3f aboveColorExpanded = new Vector3f();
		ArrayList al1 = new ArrayList();
		ArrayList al2 = new ArrayList();
		Plane p = new Plane();
		Tetrahedron startSearch1 = originalTetra[0];
		Tetrahedron startSearch2 = destinationTetra[0];
		
		int pnts;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			while (color[q] == 0.5f || color[q + 2] == 0.5f) {
				color[q] += (Math.random() - 0.5f) / 1000000.0f;
				color[q + 2] += (Math.random() - 0.5f) / 1000000.0f;
			}
			startSearch1 = startSearch1.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			startSearch2 = startSearch2.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			float chroma = SGCK_Algorithm.chroma(color[q], color[q + 1], color[q + 2],
					scale);
			float lightness = color[q + 1] * scale[1];

			colorExpanded.x = 0.5f + (color[q] - 0.5f) * 10;
			colorExpanded.y = 0.5f + (color[q + 1] - 0.5f) * 10;
			colorExpanded.z = 0.5f + (color[q + 2] - 0.5f) * 10;
			aboveColorExpanded.x = colorExpanded.x;
			aboveColorExpanded.y = colorExpanded.y + 0.5f;
			aboveColorExpanded.z = colorExpanded.z;

			p.set(colorExpanded, center, aboveColorExpanded);

			float hue = ImprovedSegmentMaxima.hue(colorExpanded.x, colorExpanded.z);
			
			GamutBoundaryPoint.release(al1);
			al1.clear();
			if (!Tetrahedron.findBoundary(p, originalTetra, startSearch1, al1, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			GamutBoundaryPoint.release(al2);
			al2.clear();
			if (!Tetrahedron.findBoundary(p, destinationTetra, startSearch2, al2, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			
			GamutBoundaryPoint originalCuspPoint = GamutBoundaryPoint.getCuspPoint(al1);
			GamutBoundaryPoint reproductionCuspPoint = GamutBoundaryPoint.getCuspPoint(al2);
			
			// System.out.println("Original cusp: L "+originalCuspPoint.y+" c "+originalCuspPoint.chroma);
			// System.out.println("Reproduction cusp: L "+reproductionCuspPoint.y+" c "+reproductionCuspPoint.chroma);
			float reproductionBlackpointLightness = this.reproductionBlack;
			
			// System.out.println("Reproduction black point "+reproductionBlackpointLightness);
			
			float adjustableFactor = factorSlider.getValue() / 100.0f;
			
			// Use reproduction or original ??????????????????????????? ------- //
			float cuspLightness = originalCuspPoint.y;
			float cuspChroma = originalCuspPoint.chroma;
			
			float kcl = 32 * (float) Math.pow((chroma / cuspChroma), 0.25) - 25;
			float kll = cuspLightness
					- ((cuspLightness - lightness) / (cuspLightness))
							* (cuspLightness - reproductionBlackpointLightness);
			float convergenceLightness = kll + kcl;

			if (convergenceLightness > this.highestCommonL) {
				convergenceLightness = this.highestCommonL;
			} else if (convergenceLightness < this.lowestCommonL) {
				convergenceLightness = this.lowestCommonL;
			}
			
			// System.out.println("Convergence point lightness "+convergenceLightness);
			
			float diffX = (color[q] - 0.5f) * scale[0];
			float diffZ = (color[q + 2] - 0.5f) * scale[2];
			float px = (float) Math.sqrt(diffX * diffX + diffZ * diffZ);
			float py = color[q + 1] * scale[1];
			float rx = -px;
			float ry = convergenceLightness - py;
			
			GamutBoundaryPoint gbp1 = Tetrahedron.findExtremeIntersection(al1, px, py, rx,
					ry);

			if (gbp1 == null) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary intersection point");
				continue;
			}
			GamutBoundaryPoint gbp2 = Tetrahedron.findExtremeIntersection(al2, px, py, rx,
					ry);
			if (gbp2 == null) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary intersection point");
				continue;
			}
			
			float originalBoundaryChroma = gbp1.chroma;
			float reproductionBoundaryChroma = gbp2.chroma;
			
			float dr = chroma / originalBoundaryChroma;
			float compressionRatio = 1
					- (float) Math.pow(dr, adjustableFactor)
							* (1 - reproductionBoundaryChroma / originalBoundaryChroma);

			if (compressionRatio < 0) {
				compressionRatio = 0;
			}
			float reproductionChroma = chroma * compressionRatio;

			if (reproductionChroma > reproductionBoundaryChroma) {
				reproductionChroma = reproductionBoundaryChroma;
				compressionRatio = reproductionBoundaryChroma / chroma;
			}
			float reproductionLightness = convergenceLightness
					+ compressionRatio * (lightness - convergenceLightness);
			
			GamutBoundaryPoint destination = GamutBoundaryPoint.getGBP();

			/*
			 if (reproductionChroma>reproductionBoundaryChroma+5 || reproductionChroma<0) {
			 System.out.println("Convergence l: "+convergenceLightness);
			 System.out.println("c1 "+chroma+", lightness1 "+lightness);
			 System.out.println("Original cusp: L "+originalCuspPoint.y+" c "+originalCuspPoint.chroma);
			 System.out.println("Original GBP: L "+gbp1.y+" c "+gbp1.chroma);
			 System.out.println("Reproduction cusp: L "+reproductionCuspPoint.y+" c "+reproductionCuspPoint.chroma);
			 System.out.println("Reproduction GBP: L "+gbp2.y+" c "+gbp2.chroma);
			 System.out.println("Compression "+compressionRatio);
			 System.out.println("c2 "+reproductionChroma+", lightness2 "+reproductionLightness);
			 if (Math.random()<0.01) {
			 System.out.println("");
			 System.out.println("Num "+al1.size());
			 //System.out.println("Epsilon "+epsilon);
			 //System.out.println("Point: "+pointx+","+pointy);
			 //System.out.println("Ray: "+rayx+","+rayy);
			 for (int j=0; j<al1.size(); j++) {
			 GamutBoundaryPoint a = (GamutBoundaryPoint)al1.get(j);
			 System.out.println("gbp: "+a.chroma+","+a.y);
			 }
			 }	
			 }
			 */
			destination.y = reproductionLightness;
			destination.chroma = reproductionChroma;
			Vector3f v3f = destination.getVector(hue, scale);

			color[q] = v3f.x;
			color[q + 1] = v3f.y;
			color[q + 2] = v3f.z;
		}
	}
	
	/**
	 * Finds the range of tetrahedron points.
	 * @param tetra The tetrahedrons.
	 * @param scale The color space scale.
	 * @return The point range.
	 */
	public static final RangeDescriptor getRange(Tetrahedron tetra[], float scale[]) {
		Vector3f center = new Vector3f(0.5f, 0.5f, 0.5f);
		Vector3f colorExpanded = new Vector3f();
		Vector3f aboveColorExpanded = new Vector3f();
		ArrayList al1 = new ArrayList();
		Plane p = new Plane();
		Tetrahedron startSearch1 = tetra[0];

		colorExpanded.x = 0.5f + (float) Math.random();
		colorExpanded.y = 0.5f + (float) Math.random();
		colorExpanded.z = 0.5f + (float) Math.random();
		startSearch1 = startSearch1.findSegmentTetrahedron2(null, colorExpanded.x,
				colorExpanded.y, colorExpanded.z);
		aboveColorExpanded.x = colorExpanded.x;
		aboveColorExpanded.y = colorExpanded.y + 0.5f;
		aboveColorExpanded.z = colorExpanded.z;

		p.set(colorExpanded, center, aboveColorExpanded);

		float hue = ImprovedSegmentMaxima.hue(colorExpanded.x, colorExpanded.z);

		GamutBoundaryPoint.release(al1);
		
		if (!Tetrahedron.findBoundary(p, tetra, startSearch1, al1, scale,
				colorExpanded.x > 0.5f, colorExpanded.z > 0.5f, hue)) {
			return null;
		}
		RangeDescriptor range = new RangeDescriptor();
		float light[] = new float[2];
		int num = 0;

		for (int i = 0; i < al1.size(); i++) {
			if (num > 1) {
				break;
			}
			GamutBoundaryPoint gbp1 = (GamutBoundaryPoint) al1.get(i);
			GamutBoundaryPoint gbp2 = (GamutBoundaryPoint) al1.get((i + 1) % al1.size());

			if (gbp1.chroma == 0) {
				light[num++] = gbp1.y;
				// System.out.println("Added ligh t "+gbp1.y);
				continue;
			}
			if (((gbp1.chroma > 0) != (gbp2.chroma > 0)) && gbp2.chroma != 0) {
				float from1 = (gbp1.chroma > 0) ? gbp1.chroma : -gbp1.chroma;
				float from2 = (gbp2.chroma > 0) ? gbp2.chroma : -gbp2.chroma;
				float total = from1 + from2;
				float l = gbp2.y * (from1 / total) + gbp1.y * (from2 / total);

				// System.out.println("Added light "+l);
				light[num++] = l;
			}
		}
		if (light[0] > light[1]) {
			range.min = light[1];
			range.max = light[0];
		} else {
			range.min = light[0];
			range.max = light[1];
		}
		range.min += 1;
		range.max -= 1;
		return range;
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			factorSlider = new JSlider(JSlider.HORIZONTAL, 100, 180, 100);
			factorField = new JTextField();
			factorField.setEditable(false);
			factorSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent ce) {
					factorField.setText("" + (factorSlider.getValue() / 100.0f));
				}
			});
			factorField.setText("" + (factorSlider.getValue() / 100.0f));
			JPanel northPanel = new JPanel(new GridLayout(0, 1));
			JPanel splitPanel = new JPanel(new GridLayout(0, 2));

			splitPanel.add(new JLabel("Factor"));
			splitPanel.add(factorField);
			northPanel.add(splitPanel);
			northPanel.add(factorSlider);
			parameterPanel = new JPanel(new BorderLayout());
			parameterPanel.add(northPanel, BorderLayout.NORTH);
			parameterPanel.add(new JPanel(), BorderLayout.CENTER);
		}
		return parameterPanel;
	}
}


/**
 * A class used to describe the range of some values.
 */
class RangeDescriptor {
	
	/** The minimum value */
	public float min;
	
	/** The maximum value */
	public float max;
}
