/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.image2d.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.SystemManager;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;


/**
 * An implementation of linear lightness mapping.
 */
public class LinearLightnessAlgorithm extends GamutMappingAlgorithm {
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		
		if (!containsImage) {
			SystemManager.displayError("Error", "Algorithm requires a source image");
			return;
		}
		ColorSpace3D outColor = mo.getColorSpace();
		float scale[] = outColor.getScale();
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		
		float colors[] = mo.getColors();
		
		float max = findMax(colors);
		float min = findMin(colors);
		float newMax = findMax(tetra);
		float newMin = findMin(tetra);
		
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			map(temporaryStorage, max, min, newMax, newMin);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					map(pos, max, min, newMax, newMin);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				map(pos, max, min, newMax, newMin);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			Arrays.fill(dirty, true);
			map(colors, max, min, newMax, newMin);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}
	
	/**
	 * Finds the minimum lightness of the tetrahedron corner positions.
	 * @param tetra The tetrahedrons
	 * @return The minimum value.
	 */
	public static final float findMin(Tetrahedron tetra[]) {
		Tetrahedron current;
		float min = tetra[0].cornerpos0_1;

		for (int i = 0; i < tetra.length; i++) {
			current = tetra[i];
			if (current.cornerpos0_1 < min) {
				min = current.cornerpos0_1;
			}
			if (current.cornerpos1_1 < min) {
				min = current.cornerpos1_1;
			}
			if (current.cornerpos2_1 < min) {
				min = current.cornerpos2_1;
			}
			if (current.cornerpos3_1 < min) {
				min = current.cornerpos3_1;
			}
		}
		return min;
	}
	
	/**
	 * Finds the maximum lightness of the tetrahedron corner positions.
	 * @param tetra The tetrahedrons
	 * @return The maximum value.
	 */
	public static final float findMax(Tetrahedron tetra[]) {
		Tetrahedron current;
		float max = tetra[0].cornerpos0_1;

		for (int i = 0; i < tetra.length; i++) {
			current = tetra[i];
			if (current.cornerpos0_1 > max) {
				max = current.cornerpos0_1;
			}
			if (current.cornerpos1_1 > max) {
				max = current.cornerpos1_1;
			}
			if (current.cornerpos2_1 > max) {
				max = current.cornerpos2_1;
			}
			if (current.cornerpos3_1 > max) {
				max = current.cornerpos3_1;
			}
		}
		return max;
	}
	
	/**
	 * Finds the minimum lightness of the colors.
	 * @param color The color(s)
	 * @return The minimum value.
	 */
	public static final float findMin(float color[]) {
		int num = color.length / 3;
		float min = color[1];

		for (int i = 0; i < num; i++) {
			if (color[i * 3 + 1] < min) {
				min = color[i * 3 + 1];
			}
		}
		return min;
	}
	
	/**
	 * Finds the maximum lightness of the colors.
	 * @param color The color(s)
	 * @return The maximum value.
	 */
	public static final float findMax(float color[]) {
		int num = color.length / 3;
		float max = color[1];

		for (int i = 0; i < num; i++) {
			if (color[i * 3 + 1] > max) {
				max = color[i * 3 + 1];
			}
		}
		return max;
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */	
	public String getName() {
		return "Linear lightness mapping";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */
	public int getPriority() {
		return 1400;
	}
	
	/**
	 * Performs linear lightness scaling.
	 * @param color The color(s) that should be mapped.
	 * @param max The original maximum lightness
	 * @param min The original minimum lightness
	 * @param newMax The new maximum lightness
	 * @param newMin The new minimum lightness
	 */
	public static final void map(float color[], float max, float min, float newMax, float newMin) {
		int num = color.length / 3;
		float temp;

		for (int i = 0; i < num; i++) {
			temp = (newMax - newMin) * (color[i * 3 + 1] - min) / (max - min) + newMin;
			if (temp < newMin) {
				temp = newMin;
			} else if (temp > newMax) {
				temp = newMax;
			}
			color[i * 3 + 1] = temp;
		}
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		return new JPanel();
	}
}
