/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.image2d.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.*;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.visualizations.ImprovedSegmentMaxima;


/**
 * An implementation of the SGCK gamut mapping algorithm.
 */
public class SGCK_Algorithm extends GamutMappingAlgorithm {
	
	/** The "amount" of the reproduction gamut where colors remain unchanged */
	private float varVar = 0.9f;
	
	/** The parameter panel */
	private JPanel parameterPanel;
	
	/** The "amount" of the reproduction gamut where colors remain unchanged */
	private JSlider varSlider;
	
	/** The "amount" of the reproduction gamut where colors remain unchanged */
	private JTextField varField;
	
	/** Sigmoidal lightness optimized parameters */
	private static final float table[][] = { {53.7f, 43.0f}, {56.8f, 40.0f},
		{58.2f, 35.0f}, {60.6f, 34.5f}};
	
	/** Sigmoidal lightness lookup table */
	private float lutInit[];
	
	/** The maximum source lightness */
	private float maxOL;
	
	/** The minimum source lightness */
	private float minOL;
	
	/** The maximum reproduction lightness */
	private float maxRL;
	
	/** The minimum reproduction lightness */
	private float minRL;
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */	
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		Tetrahedron sourceTetra[] = from;

		if (sourceTetra == null) {
			SystemManager.displayError(
				"Error",
				"Source gamut representation is not suitable for this algorithm");
			return;
		}
		
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		float scale[] = outColor.getScale();

		initializeLUT(sourceTetra, tetra);
		
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			map(temporaryStorage, scale, sourceTetra, tetra);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					map(pos, scale, sourceTetra, tetra);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				map(pos, scale, sourceTetra, tetra);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			float colors[] = mo.getColors();
			Arrays.fill(dirty, true);
			map(colors, scale, sourceTetra, tetra);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */
	public String getName() {
		return "SGCK";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */
	public int getPriority() {
		return 8;
	}
	
	/**
	 * Maps the lightness of the gamut boundary points.
	 * @param al An ArrayList that contains gamut points.
	 */
	private final void mapLightness(ArrayList al) {
		GamutBoundaryPoint gbp = null;

		for (int i = 0; i < al.size(); i++) {
			gbp = (GamutBoundaryPoint) al.get(i);
			gbp.y = 100.0f
					* this.chromaDependentLightness(Math.abs(gbp.chroma), gbp.y / 100.0f,
							this.sigmoidalLightness(gbp.y / 100.0f, maxOL, minOL, lutInit,
							lutInit.length - 1));
		}
	}
	
	/**
	 * Maps all colors to the tetrahedron structure.
	 * @param color The color(s) that should be mapped.
	 * @param scale The color space scale.
	 * @param originalTetra The source tetrahedrons.
	 * @param destinationTetra The reproduction tetrahedrons.
	 */
	private final void map(float color[], float scale[], Tetrahedron originalTetra[], Tetrahedron destinationTetra[]) {
		int num = color.length / 3;
		int q;
		Vector3f center = new Vector3f(0.5f, 0.5f, 0.5f);
		Vector3f colorExpanded = new Vector3f();
		Vector3f aboveColorExpanded = new Vector3f();
		ArrayList al1 = new ArrayList();
		ArrayList al2 = new ArrayList();
		Plane p = new Plane();
		Tetrahedron startSearch1 = originalTetra[0];
		Tetrahedron startSearch2 = destinationTetra[0];
		
		int pnts;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			while (color[q] == 0.5f || color[q + 2] == 0.5f) {
				color[q] += (Math.random() - 0.5f) / 1000000.0f;
				color[q + 2] += (Math.random() - 0.5f) / 1000000.0f;
			}
			startSearch1 = startSearch1.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			startSearch2 = startSearch2.findSegmentTetrahedron2(null, color[q],
					color[q + 1], color[q + 2]);
			float chroma = chroma(color[q], color[q + 1], color[q + 2], scale);
			float sL = sigmoidalLightness(color[q + 1], maxOL, minOL, lutInit,
					lutInit.length - 1);
			float cL = chromaDependentLightness(chroma, color[q + 1], sL);

			color[q + 1] = cL;
			colorExpanded.x = 0.5f + (color[q] - 0.5f) * 10;
			colorExpanded.y = 0.5f + (color[q + 1] - 0.5f) * 10;
			colorExpanded.z = 0.5f + (color[q + 2] - 0.5f) * 10;
			aboveColorExpanded.x = colorExpanded.x;
			aboveColorExpanded.y = colorExpanded.y + 0.5f;
			aboveColorExpanded.z = colorExpanded.z;

			p.set(colorExpanded, center, aboveColorExpanded);

			float hue = ImprovedSegmentMaxima.hue(colorExpanded.x, colorExpanded.z);

			GamutBoundaryPoint.release(al2);
			al2.clear();
			if (!Tetrahedron.findBoundary(p, destinationTetra, startSearch2, al2, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			
			float cusp = GamutBoundaryPoint.getCusp(al2);
			
			float diffX = (color[q] - 0.5f) * scale[0];
			float diffZ = (color[q + 2] - 0.5f) * scale[2];
			float px = (float) Math.sqrt(diffX * diffX + diffZ * diffZ);
			float py = color[q + 1] * scale[1];
			float rx = -px;
			float ry = cusp - py;
			GamutBoundaryPoint gbp2 = Tetrahedron.findExtremeIntersection(al2, px, py, rx,
					ry);

			if (gbp2 == null) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary intersection point");
				continue;
			}
			
			float rayY = gbp2.y - cusp;
			float rayC = gbp2.chroma;
			float dist2 = (float) Math.sqrt(rayY * rayY + rayC * rayC);
			float colorDist = (float) Math.sqrt(rx * rx + ry * ry);
			
			if (colorDist < dist2 * varVar) {
				continue;
			}
			GamutBoundaryPoint.release(al1);
			al1.clear();
			if (!Tetrahedron.findBoundary(p, originalTetra, startSearch1, al1, scale,
					color[q] > 0.5f, color[q + 2] > 0.5f, hue)) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary");
				continue;
			}
			mapLightness(al1);
			
			GamutBoundaryPoint gbp1 = Tetrahedron.findExtremeIntersection(al1, px, py, rx,
					ry);
			
			if (gbp1 == null) {
				SystemManager.displayError(
					"Error",
					"Unable to find the gamut boundary intersection point");
				continue;
			}

			float dist1 = (float) Math.sqrt((gbp1.y - cusp) * (gbp1.y - cusp)
					+ (gbp1.chroma * gbp1.chroma));
			
			if (colorDist < dist2 && dist1 < dist2) {
				continue;
			} 
			
			float d = varVar * dist2
					+ (colorDist - varVar * dist2) * (1 - varVar) * dist2
							/ (dist1 - varVar * dist2);

			if (d > dist2) {
				d = dist2;
			} else if (d < varVar * dist2) {
				d = varVar * dist2;
			}
			gbp2.y = cusp + rayY * d / dist2;
			gbp2.chroma = rayC * d / dist2;
			Vector3f v3f = gbp2.getVector(hue, scale);

			color[q] = v3f.x;
			color[q + 1] = v3f.y;
			color[q + 2] = v3f.z;
		}
	}
	
	/**
	 * Performs chroma dependent sigmoidal lightness mapping.
	 * @param color The source colors.
	 * @param scale The color space scale.
	 */
	private final void changeLightness(float color[], float scale[]) {
		int num = color.length / 3;
		int p;

		for (int i = 0; i < num; i++) {
			p = i * 3;
			float chroma = this.chroma(color[p], color[p + 1], color[p + 2], scale);
			float sL = this.sigmoidalLightness(color[p + 1], maxOL, minOL, lutInit,
					lutInit.length - 1);
			float cL = this.chromaDependentLightness(chroma, color[p + 1], sL);

			color[p + 1] = cL;
		}
	} 

	/**
	 * Calculates the sigmoidal lightness mapping LUT.
	 * @param tetraOriginal The source tetrahedrons.
	 * @param tetraReproduction The reproduction tetrahedrons.
	 */
	private final void initializeLUT(Tetrahedron tetraOriginal[], Tetrahedron tetraReproduction[]) {
		float originalMinL = tetraOriginal[0].cornerpos0_1;
		float originalMaxL = tetraOriginal[0].cornerpos0_1;
		float reproductionMinL = tetraReproduction[0].cornerpos0_1;
		float reproductionMaxL = tetraReproduction[0].cornerpos0_1;
		
		for (int i = 0; i < tetraOriginal.length; i++) {
			if (tetraOriginal[i].cornerpos0_1 > originalMaxL) {
				originalMaxL = tetraOriginal[i].cornerpos0_1;
			} else if (tetraOriginal[i].cornerpos0_1 < originalMinL) {
				originalMinL = tetraOriginal[i].cornerpos0_1;
			}
			
			if (tetraOriginal[i].cornerpos1_1 > originalMaxL) {
				originalMaxL = tetraOriginal[i].cornerpos1_1;
			} else if (tetraOriginal[i].cornerpos1_1 < originalMinL) {
				originalMinL = tetraOriginal[i].cornerpos1_1;
			}
			
			if (tetraOriginal[i].cornerpos2_1 > originalMaxL) {
				originalMaxL = tetraOriginal[i].cornerpos2_1;
			} else if (tetraOriginal[i].cornerpos2_1 < originalMinL) {
				originalMinL = tetraOriginal[i].cornerpos2_1;
			}
			
			if (tetraOriginal[i].cornerpos3_1 > originalMaxL) {
				originalMaxL = tetraOriginal[i].cornerpos3_1;
			} else if (tetraOriginal[i].cornerpos3_1 < originalMinL) {
				originalMinL = tetraOriginal[i].cornerpos3_1;
			}
		}
		
		for (int i = 0; i < tetraReproduction.length; i++) {
			if (tetraReproduction[i].cornerpos0_1 > reproductionMaxL) {
				reproductionMaxL = tetraReproduction[i].cornerpos0_1;
			} else if (tetraReproduction[i].cornerpos0_1 < reproductionMinL) {
				reproductionMinL = tetraReproduction[i].cornerpos0_1;
			}
			
			if (tetraReproduction[i].cornerpos1_1 > reproductionMaxL) {
				reproductionMaxL = tetraReproduction[i].cornerpos1_1;
			} else if (tetraReproduction[i].cornerpos1_1 < reproductionMinL) {
				reproductionMinL = tetraReproduction[i].cornerpos1_1;
			}
			
			if (tetraReproduction[i].cornerpos2_1 > reproductionMaxL) {
				reproductionMaxL = tetraReproduction[i].cornerpos2_1;
			} else if (tetraReproduction[i].cornerpos2_1 < reproductionMinL) {
				reproductionMinL = tetraReproduction[i].cornerpos2_1;
			}
			
			if (tetraReproduction[i].cornerpos3_1 > reproductionMaxL) {
				reproductionMaxL = tetraReproduction[i].cornerpos3_1;
			} else if (tetraReproduction[i].cornerpos3_1 < reproductionMinL) {
				reproductionMinL = tetraReproduction[i].cornerpos3_1;
			}
		}
		originalMaxL *= 100.0f;
		originalMinL *= 100.0f;
		reproductionMaxL *= 100.0f;
		reproductionMinL *= 100.0f;
		
		int index = (int) Math.floor(reproductionMinL / 5) - 1;
		float w = (reproductionMinL - index * 5 - 5) / 5.0f;

		if (index > 2) {
			index = 2;
			w = 1;
		} else if (index < 0) {
			index = 0;
			w = 0;
		}
		float xo = (1.0f - w) * table[index][0] + w * table[index + 1][0];
		float sigma = (1.0f - w) * table[index][1] + w * table[index + 1][1];
		int numLUT = 256;
		float sLUT[] = new float[numLUT + 1];

		for (int i = 0; i <= numLUT; i++) {
			float si = (1.0f / (((float) Math.sqrt(2.0 * Math.PI)) * sigma))
					* (float) Math.exp(-((Math.pow(100.0 * i / (float) numLUT - xo, 2)))
							/ (2.0 * sigma * sigma));

			if (i == 0) {
				sLUT[0] = si;
			} else {
				sLUT[i] = sLUT[i - 1] + si;
			}
		}
		for (int i = 1; i <= numLUT; i++) {
			sLUT[i] = (sLUT[i] - sLUT[0]) / (sLUT[numLUT] - sLUT[0])
							* (reproductionMaxL - reproductionMinL)
					+ reproductionMinL;
		}
		sLUT[0] = reproductionMinL;
		this.maxOL = originalMaxL;
		this.minOL = originalMinL;
		this.maxRL = reproductionMaxL;
		this.minRL = reproductionMinL;
		this.lutInit = sLUT;
	}
	
	/**
	 * Calculates the chroma of a color.
	 * @param b The b coordinate of CIELAB.
	 * @param l The L coordinate of CIELAB.
	 * @param a The a coordinate of CIELAB.
	 * @param scale The color space scale.
	 * @return The chroma of the color.
	 */
	public static final float chroma(float b, float l, float a, float scale[]) {
		float bDiff = (b - 0.5f) * scale[0];
		float aDiff = (a - 0.5f) * scale[2];
		float chroma = (float) Math.sqrt(bDiff * bDiff + aDiff * aDiff);

		return chroma;
	}
	
	/**
	 * Performs an interpolated table lookup based sigmoidal lightness mapping.
	 * @param originalLightness The original lightness.
	 * @param maxOriginalLightness The maximum source lightness.
	 * @param minOriginalLighness The minimum source lightness.
	 * @param lut The lookup table.
	 * @param m The table size.
	 * @return The mapped lightness.
	 */
	private static final float sigmoidalLightness(float originalLightness, float maxOriginalLightness, float minOriginalLighness, float lut[], int m) {
		originalLightness *= 100.0f;
		originalLightness = 100.0f
				- (maxOriginalLightness - originalLightness) * 100.0f
						/ (maxOriginalLightness - minOriginalLighness);
		if (originalLightness < 100.0f) {
			float w = (float) Math.floor(originalLightness * m / 100.0f);

			if (w < 0.0f) {
				w = 0.0f;
			}
			originalLightness = (1.0f - originalLightness * m / 100.0f + w)
							* lut[(int) w]
					+ (originalLightness * m / 100.0f - w) * lut[(int) w + 1];
		} else {
			originalLightness = lut[m];
		}
		return (originalLightness /= 100.0f);
	}
	
	/**
	 * Performs chroma dependent lightness mapping.
	 * @param originalLightness The original lightness.
	 * @param originalChroma The original chroma.
	 * @param sigmoidalLighness The sigmoidal mapped lightness.
	 * @return The chroma dependent lightness.
	 */
	private static final float chromaDependentLightness(float originalChroma, float originalLightness, float sigmoidalLighness) {
		originalLightness *= 100.0f;
		sigmoidalLighness *= 100.0f;
		float weight = (float) Math.sqrt(Math.pow(originalChroma, 3.0)
				/ (Math.pow(originalChroma, 3.0) + 500000.0));

		return (((weight) * originalLightness + (1.0f - weight) * sigmoidalLighness)
				/ 100.0f);
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			parameterPanel = new JPanel(new BorderLayout());
			varField = new JTextField();
			varField.setEditable(false);
			varField.setText(Float.toString(varVar));
			JPanel pPanel = new JPanel(new GridLayout(0, 2));

			pPanel.add(new JLabel(" Factor "));
			pPanel.add(varField);
			varSlider = new JSlider(JSlider.HORIZONTAL, 1, 99, 90);
			varSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent ce) {
					varVar = (float) varSlider.getValue() / (float) 100;
					varField.setText(Float.toString(varVar));
				}
			});
			JPanel northPanel = new JPanel(new BorderLayout());

			northPanel.add(pPanel, BorderLayout.NORTH);
			northPanel.add(varSlider, BorderLayout.SOUTH);
			parameterPanel.add(northPanel, BorderLayout.NORTH);
			parameterPanel.add(new JPanel(), BorderLayout.CENTER);
			
		}
		return parameterPanel;
	}
}
