/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.gma;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.image2d.*;
import icc3dpackage.managers.SystemManager;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;


/**
 * An implementation of sigmoidal lightness mapping.
 */
public class SigmoidalLightnessAlgorithm extends GamutMappingAlgorithm {
	
	/** A slider that allows the user to specify the xo parameter. */
	private JSlider xoSlider;
	
	/** A slider that allows the user to specify the sigma parameter. */
	private JSlider sigmaSlider;
	
	/** A text field that displays the value of the xo parameter. */
	private JTextField xoField;
	
	/** A text field that displays the value of the sigma parameter. */
	private JTextField sigmaField;
	
	/** A check box that decides whether the algorithm automatically selects
	 * suitable default values for the xo and sigma parameters.
	 */
	private JCheckBox useDefault;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** A panel where the resulting sigmoidal function is displayed. */
	private JPanel sigPanel;
	
	/** A table that contains predefined sigma and xo values. */
	private static final float table[][] = { {53.7f, 43.0f}, {56.8f, 40.0f},
		{58.2f, 35.0f}, {60.6f, 34.5f}};
	
	/**
	 * Performs gamut mapping of the provided image gamut.
	 * @param mo The image gamut.
	 * @param tetra The destination tetrahedron structure.
	 * @param from The source tetrahedron structure.
	 */	
	public void map(MovingObject mo, Tetrahedron tetra[], Tetrahedron from[]) {
		boolean containsImage = (mo.getImage2D() != null);
		boolean containsObjects = (mo.getObjects() != null);
		if (!containsImage) {
			SystemManager.displayError("Error", "Algorithm requires a source image");
			return;
		}
		ColorSpace3D outColor = mo.getColorSpace();
		float scale[] = outColor.getScale();
		PickableObject objects[] = mo.getObjects();
		GeometryArray ga[] = mo.getGeometryArrays();
		
		float colors[] = mo.getColors();
		
		float max = LinearLightnessAlgorithm.findMax(colors);
		float min = LinearLightnessAlgorithm.findMin(colors);
		float newMax = LinearLightnessAlgorithm.findMax(tetra);
		float newMin = LinearLightnessAlgorithm.findMin(tetra);
		
		float lut[] = new float[257];

		initializeLut(lut, newMax, newMin);

		max *= 100.0f;
		min *= 100.0f;
		
		float pos[];

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
		           
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			map(temporaryStorage, max, min, lut);
			byte sRGB[] = outColor.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				pos = objects[i].getPositionReference();
				pos[0] = temporaryStorage[i * 3];
				pos[1] = temporaryStorage[i * 3 + 1];
				pos[2] = temporaryStorage[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			if (containsObjects) {
				for (int i = 0; i < objects.length; i++) {
					pos = objects[i].getPositionReference();
					map(pos, max, min, lut);
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				map(pos, max, min, lut);
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, outColor.toSRGB(pos));
			}
		}
		
		if (containsImage) {
			ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
			boolean dirty[] = image2d.getDirtyColors();
			Arrays.fill(dirty, true);
			map(colors, max, min, lut);
			outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
			image2d.imageUpdate();
		}
	}

	/**
	 * Returns the name of the gamut mapping algorithm.
	 * @return The name of the gamut mapping algorithm.
	 */	
	public String getName() {
		return "Sigmoidal lightness mapping";
	}
	
	/**
	 * The priority of the plugin.
	 * @return The priority.
	 */	
	public int getPriority() {
		return 1400;
	}
	
	/**
	 * Performs sigmoidal lightness mapping on the colors.
	 * @param color The color(s).
	 * @param max The maximum lightness.
	 * @param min The minimum lightness.
	 * @param lut The lookup table.
	 */
	public static final void map(float color[], float max, float min, float lut[]) {
		int num = color.length / 3;
		float temp;

		for (int i = 0; i < num; i++) {
			color[i * 3 + 1] = sigmoidalLightness(color[i * 3 + 1], max, min, lut);
		}
	}
	
	/**
	 * Initializes the lookup table.
	 * @param lut The lookup table.
	 * @param newMax The new maximum lightness.
	 * @param newMin The new minimum lightness.
	 */
	public final void initializeLut(float lut[], float newMax, float newMin) {
		newMax *= 100.0f;
		newMin *= 100.0f;
		
		int index = (int) Math.floor(newMin / 5) - 1;
		float w = (newMin - index * 5 - 5) / 5.0f;

		if (index > 2) {
			index = 2;
			w = 1;
		} else if (index < 0) {
			index = 0;
			w = 0;
		}
		float xo = (1.0f - w) * table[index][0] + w * table[index + 1][0];
		float sigma = (1.0f - w) * table[index][1] + w * table[index + 1][1];
		
		// System.out.println("xo: "+xo);
		// System.out.println("sigma: "+sigma);
		
		if (useDefault != null && !useDefault.isSelected()) {
			xo = xoSlider.getValue();
			sigma = sigmaSlider.getValue();
		}
		
		float numLUT = lut.length - 1;

		for (int i = 0; i <= numLUT; i++) {
			float si = (1.0f / (((float) Math.sqrt(2.0 * Math.PI)) * sigma))
					* (float) Math.exp(-((Math.pow(100.0 * i / numLUT - xo, 2)))
							/ (2.0 * sigma * sigma));

			if (i == 0) {
				lut[0] = si;
			} else {
				lut[i] = lut[i - 1] + si;
			}
		}
		for (int i = 1; i <= numLUT; i++) {
			lut[i] = (lut[i] - lut[0]) / (lut[(int) numLUT] - lut[0]) * (newMax - newMin)
					+ newMin;
		}
		lut[0] = newMin;	
	}
	
	/**
	 * Performs an interpolated table lookup based sigmoidal lightness mapping.
	 * @param originalLightness The original lightness.
	 * @param maxOriginalLightness The maximum source lightness.
	 * @param minOriginalLighness The minimum source lightness.
	 * @param lut The lookup table.
	 * @return The mapped lightness.
	 */
	private static final float sigmoidalLightness(float originalLightness, float maxOriginalLightness, float minOriginalLighness, float lut[]) {
		int m = lut.length - 1;

		originalLightness *= 100.0f;
		originalLightness = 100.0f
				- (maxOriginalLightness - originalLightness) * 100.0f
						/ (maxOriginalLightness - minOriginalLighness);
		if (originalLightness < 100.0f) {
			float w = (float) Math.floor(originalLightness * m / 100.0f);

			if (w < 0.0f) {
				w = 0.0f;
			}
			originalLightness = (1.0f - originalLightness * m / 100.0f + w)
							* lut[(int) w]
					+ (originalLightness * m / 100.0f - w) * lut[(int) w + 1];
		} else {
			originalLightness = lut[m];
		}
		return (originalLightness /= 100.0f);
	}
	
	/**
	 * Returns the parameter panel of the gamut mapping algorithm.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			parameterPanel = new JPanel(new BorderLayout());
			xoField = new JTextField();
			xoField.setEditable(false);
			xoSlider = new JSlider(JSlider.HORIZONTAL, 1, 99, 50);
			xoSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent ce) {
					xoField.setText("" + xoSlider.getValue());
					if (sigPanel != null) {
						sigPanel.repaint();
					}
				}
			});
			sigmaField = new JTextField();
			sigmaField.setEditable(false);
			sigmaSlider = new JSlider(JSlider.HORIZONTAL, 1, 99, 40);
			sigmaSlider.addChangeListener(new ChangeListener() {
				public void stateChanged(ChangeEvent ce) {
					sigmaField.setText("" + sigmaSlider.getValue());
					if (sigPanel != null) {
						sigPanel.repaint();
					}
				}
			});
			xoField.setText("" + xoSlider.getValue());
			sigmaField.setText("" + sigmaSlider.getValue());
			useDefault = new JCheckBox();
			useDefault.setSelected(true);
			useDefault.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					if (sigPanel != null) {
						sigPanel.repaint();
					}
				}
			});
			JPanel northPanel = new JPanel(new GridLayout(0, 1));
			JPanel splitPanel = new JPanel(new GridLayout(0, 2));

			splitPanel.add(new JLabel("xo"));
			splitPanel.add(xoField);
			northPanel.add(splitPanel);
			northPanel.add(xoSlider);
			splitPanel = new JPanel(new GridLayout(0, 2));
			splitPanel.add(new JLabel("sigma"));
			splitPanel.add(sigmaField);
			northPanel.add(splitPanel);
			northPanel.add(sigmaSlider);
			splitPanel = new JPanel(new GridLayout(0, 2));
			splitPanel.add(new JLabel("Use default"));
			splitPanel.add(useDefault);
			northPanel.add(splitPanel);
			sigPanel = new JPanel() {
				private static final float size = 200;
				public void paint(Graphics g) {
					super.paint(g);
					float lut[] = new float[257];

					initializeLut(lut, 1.0f, 0.0f);
					g.setColor(Color.blue);
					for (int i = 0; i < 100; i++) {
						int x1 = (int) (size * i / 100.0f);
						int y1 = (int) (size
								- size * sigmoidalLightness(i / 100.0f, 100.0f, 0.0f, lut));
						int x2 = (int) (size * (i + 1) / 100.0f);
						int y2 = (int) (size
								- size * sigmoidalLightness((i + 1) / 100.0f, 100.0f, 0.0f, lut));

						g.drawLine(x1, y1, x2, y2);
					}
				}

				public Dimension getPreferredSize() {
					return new Dimension((int) size + 1, (int) size + 1);
				}
			};
			JPanel innerPanel = new JPanel(new BorderLayout());

			innerPanel.add(northPanel, BorderLayout.NORTH);
			JPanel borderPanel = new JPanel(new BorderLayout());

			borderPanel.setBorder(BorderFactory.createBevelBorder(BevelBorder.RAISED));
			borderPanel.add(sigPanel, BorderLayout.CENTER);
			JPanel spaceCorrection = new JPanel(new BorderLayout());

			spaceCorrection.add(borderPanel, BorderLayout.WEST);
			spaceCorrection.add(new JPanel(), BorderLayout.CENTER);
			innerPanel.add(spaceCorrection, BorderLayout.SOUTH);
			parameterPanel.add(innerPanel, BorderLayout.NORTH);
			parameterPanel.add(new JPanel(), BorderLayout.CENTER);
		}
		return parameterPanel;
	}
}
