/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.iccreader;


import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * This class is capable of reading an ICC profile header, and
 * can be used to find certain information about such a file.
 */
public class IccHeader {
	
	/*
	public static final byte[] PROFILE_FILE_SIGNATURE = {0x61, 0x63, 0x73, 0x70};
	public static final byte[] INPUT_DEVICE_PROFILE = {0x73, 0x63, 0x6e, 0x72};
	public static final byte[] DISPLAY_DEVICE_PROFILE = {0x6d, 0x6e, 0x74, 0x72};
	public static final byte[] OUTPUT_DEVICE_PROFILE = {0x70, 0x72, 0x74, 0x72};
	public static final byte[] DEVICELINK_PROFILE = {0x6c, 0x69, 0x6e, 0x6b};
	public static final byte[] COLORSPACE_CONVERSION_PROFILE = {0x73, 0x70, 0x61,
		0x63};
	public static final byte[] ABSTRACT_PROFILE = {0x61, 0x62, 0x73, 0x74};
	public static final byte[] NAMED_COLOR_PROFILE = {0x6e, 0x6d, 0x63, 0x6c};
	*/
	
	/** The size of the profile. */
	private long profileSize;
	
	/** The CMM signature. */
	private long cmmTypeSignature;
	
	/** The profile version. */
	private byte profileVersion[] = new byte[4];
	
	/** The device class signature. */
	private String deviceClassSignature;
	
	/** The device dependent color space. */
	private String dataColorSpace;
	
	/** The device independent color space. */
	private String profileConnectionSpace;
	
	/** The profile creation year. */
	private int dateTimeYear;
	
	/** The profile creation month. */
	private int dateTimeMonth;
	
	/** The profile creation day. */
	private int dateTimeDay;
	
	/** The profile creation hour. */
	private int dateTimeHour;
	
	/** The profile creation minute. */
	private int dateTimeMinute;
	
	/** The profile creation second. */
	private int dateTimeSecond;
	
	/** The profile file signature. */
	private long profileFileSignature;
	
	/** The primary profile platform. */
	private String primaryPlatformSignature;
	
	/** Profile flags. */
	private byte flags[] = new byte[4];
	
	/** The profile device manufacturer. */
	private byte deviceManufacturer[] = new byte[4];
	
	/** The profile device model. */
	private byte deviceModel[] = new byte[4];
	
	/** The profile device attributes. */
	private byte deviceAttributes[] = new byte[8];
	
	/** The profile rendering intent. */
	private byte renderingIntent[] = new byte[4];
	
	/** The PCS illuminant. */
	private byte iluminantOfPCS[] = new byte[12];
	
	/** The profile creator signature. */
	private byte profileCreatorSignature[] = new byte[4];
	
	/** Reserved header bytes. */
	private byte reservedForExpansion[] = new byte[44];

	/**
	 * Constructs an IccHeader from a file.
	 */
	public IccHeader(RandomAccessFileWrapper rafw) {
		byte fourBytes[] = new byte[4];

		profileSize = rafw.getLong();
		if (profileSize != rafw.getLength()) {
			throw new RuntimeException("Profile size specified in header does not match file length");
		}
		cmmTypeSignature = rafw.getLong();
		rafw.getBytes(profileVersion);
		profileVersion[0] = (byte) (10 * (profileVersion[0] >> 4)
				+ ((profileVersion[0]) & (0x0f)));
		profileVersion[2] = (byte) ((profileVersion[1]) & (0x0f));
		profileVersion[1] = (byte) ((profileVersion[1]) >> 4);
		rafw.getBytes(fourBytes);
		deviceClassSignature = new String(fourBytes).trim();
		rafw.getBytes(fourBytes);
		dataColorSpace = new String(fourBytes).trim();
		rafw.getBytes(fourBytes);
		profileConnectionSpace = new String(fourBytes).trim();
		dateTimeYear = rafw.getInt();
		dateTimeMonth = rafw.getInt();
		dateTimeDay = rafw.getInt();
		dateTimeHour = rafw.getInt();
		dateTimeMinute = rafw.getInt();
		dateTimeSecond = rafw.getInt();
		profileFileSignature = rafw.getLong();
		rafw.getBytes(fourBytes);
		primaryPlatformSignature = new String(fourBytes).trim();
		rafw.getBytes(flags);
		rafw.getBytes(deviceManufacturer);
		rafw.getBytes(deviceModel);
		rafw.getBytes(deviceAttributes);
		rafw.getBytes(renderingIntent);
		rafw.getBytes(iluminantOfPCS);
		rafw.getBytes(profileCreatorSignature);
		rafw.getBytes(reservedForExpansion);
	}
	
	/**
	 * Returns a string with information about the profile.
	 * @return Information about the profile.
	 */
	public String getInfo() {
		String s;

		s = "Profile size: " + profileSize + "\nCMM Type Signature: "
				+ cmmTypeSignature + "\nProfile version: " + profileVersion[0] + "."
				+ profileVersion[1] + "." + profileVersion[2] + "\nDevice class: "
				+ deviceClassSignature + "\nData Color Space: " + dataColorSpace
				+ "\nProfile Connection Space: " + profileConnectionSpace
				+ "\nDate and time: " + dateTimeHour + ":" + dateTimeMinute + ":"
				+ dateTimeSecond + " " + dateTimeDay + "." + dateTimeMonth + "."
				+ dateTimeYear + "\nProfile file signature: " + profileFileSignature
				+ "\nPrimary platform: " + primaryPlatformSignature;
		return s;
	}
	
	/**
	 * Returns the profile size.
	 * @return The size of the profile.
	 */
	public long getProfileSize() {
		return profileSize;
	}
	
	/**
	 * Returns the CMM type signature.
	 * @return The CMM type signature.
	 */
	public long getCmmTypeSignature() {
		return cmmTypeSignature;
	}
	
	/**
	 * Returns the profile version.
	 * @return The profile version.
	 */
	public byte[] getProfileVersion() {
		return profileVersion;
	}
	
	/**
	 * Returns the device class signature.
	 * @return The device class signature.
	 */
	public String getDeviceClassSignature() {
		return deviceClassSignature;
	}
	
	/**
	 * Returns the data color space.
	 * @return The data color space.
	 */
	public String getDataColorSpace() {
		return dataColorSpace;
	}
	
	/**
	 * Returns the profile connection space.
	 * @return The profile connection space.
	 */
	public String getProfileConnectionSpace() {
		return profileConnectionSpace;
	}
	
	/**
	 * Returns the year of the profile.
	 * @return The year.
	 */
	public int getDateTimeYear() {
		return dateTimeYear;
	}
	
	/**
	 * Returns the month of the profile.
	 * @return The month.
	 */
	public int getDateTimeMonth() {
		return dateTimeMonth;
	}
	
	/**
	 * Returns the day of the profile.
	 * @return The day.
	 */
	public int getDateTimeDay() {
		return dateTimeDay;
	}
	
	/**
	 * Returns the hour of the profile.
	 * @return The hour.
	 */
	public int getDateTimeHour() {
		return dateTimeHour;
	}
	
	/**
	 * Returns the minute of the profile.
	 * @return The minute.
	 */
	public int getDateTimeMinute() {
		return dateTimeMinute;
	}
	
	/**
	 * Returns the second of the profile.
	 * @return The second.
	 */
	public int getDateTimeSecond() {
		return dateTimeSecond;
	}
	
	/**
	 * Returns the profile file signature.
	 * @return The profile file signature.
	 */
	public long getProfileFileSignature() {
		return profileFileSignature;
	}
	
	/**
	 * Returns the primary platform signature.
	 * @return The primary platform signature.
	 */
	public String getPrimaryPlatformSignature() {
		return primaryPlatformSignature;
	}
	
	/**
	 * Returns the flags.
	 * @return The flags.
	 */
	public byte[] getFlags() {
		return flags;
	}
	
	/**
	 * Returns the device manufacturer.
	 * @return The device manufacturer.
	 */
	public byte[] getDeviceManufacturer() {
		return deviceManufacturer;
	}
	
	/**
	 * Returns the device model.
	 * @return The device model.
	 */
	public byte[] getDeviceModel() {
		return deviceModel;
	}
	
	/**
	 * Returns the device attributes.
	 * @return The device attributes.
	 */
	public byte[] getDeviceAttributes() {
		return deviceAttributes;
	}
	
	/**
	 * Returns the rendering intent.
	 * @return The rendering intent.
	 */
	public byte[] getRenderingIntent() {
		return renderingIntent;
	}
	
	/**
	 * Returns the illuminant of the PCS.
	 * @return The illuminant of the PCS.
	 */
	public byte[] getIluminantOfPCS() {
		return iluminantOfPCS;
	}
	
	/**
	 * Returns the profile creator signature.
	 * @return The profile creator signature.
	 */
	public byte[] getProfileCreatorSignature() {
		return profileCreatorSignature;
	}
} 

