/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.iccreader;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.awt.color.*;
import java.text.*;

import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A class capable of reading and understanding certain parts of an ICC profile.
 */
public class IccReader {

	/** The LutReader used to read the gamut lut. */
	private LutReader lutReader;
	
	/** An IccHeader representing the information found in the file header. */
	private IccHeader iccHeader;
	
	/** The tag count. */
	private int tagCount;
	
	/** The profile white point. */
	private float whitePoint[];
	
	/** The tag offsets. */
	private long tagOffset[];
	
	/** The size of the elements. */
	private int tagElementSize[];
	
	/** The tag signatures. */
	private long tagSignature[];
	
	/** The actual content of the tags. */
	private byte tagContent[][];

	/**
	 * Constructs an IccReader from a given file.
	 * @param iccFile The name of the ICC file.
	 */	 
	public IccReader(String iccFile) {
		RandomAccessFile randomAccessFile;

		try {
			randomAccessFile = new RandomAccessFile(iccFile, "r");
		} catch (FileNotFoundException fnfe) {
			throw new RuntimeException("File not found!");
		}
		
		RandomAccessFileWrapper rafw = new RandomAccessFileWrapper(randomAccessFile,
				false);

		iccHeader = new IccHeader(rafw);

		tagCount = (int) rafw.getLong();
		tagOffset = new long[tagCount];
		tagElementSize = new int[tagCount];
		tagSignature = new long[tagCount];
		tagContent = new byte[tagCount][];
		
		for (int i = 0; i < tagCount; i++) {
			tagSignature[i] = rafw.getLong();
			tagOffset[i] = rafw.getLong();
			tagElementSize[i] = (int) rafw.getLong();
		}
		
		long gamutOffset = 0;
		long whitePointOffset = 0;
		
		for (int i = 0; i < tagCount; i++) {
			if (tagSignature[i] == 0x67616D74) {
				gamutOffset = tagOffset[i];
			}
			if (tagSignature[i] == 0x77747074) {
				whitePointOffset = tagOffset[i];
			}
			tagContent[i] = new byte[tagElementSize[i]];
			rafw.setOffset(tagOffset[i]);
			rafw.getBytes(tagContent[i]);
		}
		
		if (whitePointOffset != 0) {
			rafw.setOffset(whitePointOffset);
			long id = rafw.getLong();
			long reserved = rafw.getLong();
			float white[] = new float[3];
			byte b[] = new byte[4];

			for (int i = 0; i < 3; i++) {
				rafw.getBytes(b);
				int o = (b[0]) & (0xff);

				o = o << 8;
				o += (b[1]) & (0xff);
				o = o << 8;
				o += (b[2]) & (0xff);
				o = o << 8;
				o += (b[3]) & (0xff);
				float f;

				f = o;
				f = f / 65536.0f;
				white[i] = f;
				// System.out.println("White /**/: "+f);
			}
			whitePoint = white;
		}
		
		if (gamutOffset != 0) {
			lutReader = new LutReader(rafw, gamutOffset);
		} else {
			lutReader = null;
		}
	}

	/**
	 * Returns the output of the gamut tag given a certain input.
	 * @param input The input
	 * @return The output
	 */
	public float[] getOutput(float input[]) {
		return lutReader.getOutput(input);
	}
	
	/**
	 * Returns the string representing the given tag signature.
	 * @param sig The tag signature.
	 * @return The string describing the tag.
	 */
	public String tagType(long sig) {
		if (sig == 0x41324230) {
			return "AToB0Tag              ";
		}
		if (sig == 0x41324231) {
			return "AToB1Tag              ";
		}
		if (sig == 0x41324232) {
			return "AToB2Tag              ";
		}
		if (sig == 0x6258595a) {
			return "blueColorantTag       ";
		}
		if (sig == 0x62545243) {
			return "blueTRCTag            ";
		}
		if (sig == 0x42324130) {
			return "BToA0Tag              ";
		}
		if (sig == 0x42324131) {
			return "BToA1Tag              ";
		}
		if (sig == 0x42324132) {
			return "BToA2Tag              ";
		}
		if (sig == 0x63616C74) {
			return "calibrationDateTimeTag";
		}
		if (sig == 0x74617267) {
			return "charTargetTag         ";
		}
		if (sig == 0x63707274) {
			return "copyrightTag          ";
		}
		if (sig == 0x63726469) {
			return "crdInfoTag            ";
		}
		if (sig == 0x646D6E64) {
			return "deviceMfgDescTag      ";
		}
		if (sig == 0x646D6464) {
			return "deviceModelDescTag    ";
		}
		if (sig == 0x64657673) {
			return "deviceSettingsTag     ";
		}
		if (sig == 0x67616D74) {
			return "gamutTag              ";
		}
		if (sig == 0x6B545243) {
			return "grayTRCTag            ";
		}
		if (sig == 0x6758595a) {
			return "greenColorantTag      ";
		}
		if (sig == 0x67545243) {
			return "greenTRCTag           ";
		}
		if (sig == 0x6C756D69) {
			return "luminanceTag          ";
		}
		if (sig == 0x6D656173) {
			return "measurementTag        ";
		}
		if (sig == 0x626B7074) {
			return "mediaBlackPointTag    ";
		}
		if (sig == 0x77747074) {
			return "mediaWhitePointTag    ";
		}
		if (sig == 0x6E636F6C) {
			return "namedColorTag         ";
		}
		if (sig == 0x6E636C32) {
			return "namedColor2Tag        ";
		}
		if (sig == 0x72657370) {
			return "outputResponseTag     ";
		}
		if (sig == 0x70726530) {
			return "preview0Tag           ";
		}
		if (sig == 0x70726531) {
			return "preview1Tag           ";
		}
		if (sig == 0x70726532) {
			return "preview2Tag           ";
		}
		if (sig == 0x64657363) {
			return "profileDescriptionTag ";
		}
		if (sig == 0x70736571) {
			return "profileSequenceDescTag";
		}
		if (sig == 0x70736430) {
			return "ps2CRD0Tag            ";
		}
		if (sig == 0x70736431) {
			return "ps2CRD1Tag            ";
		}
		if (sig == 0x70736432) {
			return "ps2CRD2Tag            ";
		}
		if (sig == 0x70736433) {
			return "ps2CRD3Tag            ";
		}
		if (sig == 0x70733273) {
			return "ps2CSATag             ";
		}
		if (sig == 0x70733269) {
			return "ps2RenderingIntentTag ";
		}
		if (sig == 0x7258595A) {
			return "redColorantTag        ";
		}
		if (sig == 0x72545243) {
			return "redTRCTag             ";
		}
		if (sig == 0x73637264) {
			return "screeningDescTag      ";
		}
		if (sig == 0x7363726E) {
			return "screeningTag          ";
		}
		if (sig == 0x74656368) {
			return "technologyTag         ";
		}
		if (sig == 0x62666420) {
			return "ucrbgTag              ";
		}
		if (sig == 0x76756564) {
			return "viewingCondDescTag    ";
		}
		if (sig == 0x76696577) {
			return "viewingConditionsTag  ";
		}
		byte b[] = new byte[22];

		for (int i = 0; i < 22; i++) {
			b[i] = 0x20;
		}
		b[3] = (byte) (sig & 0xff);
		b[2] = (byte) ((sig >> 8) & (0xff));
		b[1] = (byte) ((sig >> 16) & (0xff));
		b[0] = (byte) ((sig >> 24) & (0xff));
		return new String(b);
	}
	
	/**
	 * Returns a string describing the ICC profile.
	 * @return A string describing the ICC profile.
	 */
	public String getInfo() {
		String headerInfo = iccHeader.getInfo();
		String tags = "\n\nTag count: " + tagCount;
		NumberFormat nf = NumberFormat.getInstance();

		nf.setMinimumIntegerDigits(0);
		for (int i = 0; i < tagCount; i++) {
			tags = tags + "\nTag signature: 0x" + Long.toHexString(tagSignature[i])
					+ " type: " + tagType(tagSignature[i]) + " offset: " + tagOffset[i];
		}
		String lutInfo = "";

		if (lutReader != null) {
			lutInfo = lutReader.getInfo();
		}
		return (headerInfo + tags + "\n\n" + lutInfo);
	}
	
	/**
	 * Returns the ICC header.
	 * @return The ICC header.
	 */
	public IccHeader getHeader() {
		return iccHeader;
	}
	
	/**
	 * Returns the tag count.
	 * @return The tag count.
	 */
	public int getTagCount() {
		return tagCount;
	}
	
	/**
	 * Returns the tag offsets.
	 * @return The tag offsets.
	 */
	public long[] getTagOffset() {
		return tagOffset;
	}
	
	/**
	 * Returns the tag element sizes.
	 * @return The tag element sizes.
	 */
	public int[] getTagElementSize() {
		return tagElementSize;
	}
	
	/**
	 * Returns the tag signatures.
	 * @return The tag signatures.
	 */
	public long[] getTagSignature() {
		return tagSignature;
	}
	
	/**
	 * Returns the tag content.
	 * @return The tag content.
	 */
	public byte[][] getTagContent() {
		return tagContent;
	}
	
	/**
	 * Returns a LutReader that contains the inside/outside gamut information.
	 * @return A LutReader.
	 */
	public LutReader getGamutReader() {
		return lutReader;
	}
	
	/**
	 * Returns the ICC profile white point.
	 * @return A float array of length 3 that contains the CIEXYZ white point.
	 */
	public float[] getWhitePoint() {
		return whitePoint;
	}
}
