/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.iccreader;


import javax.vecmath.*;
import javax.swing.JOptionPane;

import icc3dpackage.managers.SystemManager;
import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A class representing a lookup table in an ICC profile
 */	 
public class LutReader {

	/**
	 * A lut identifier.
	 */
	private long identifier;
	
	/**
	 * The number of input channels.
	 */
	private int inputChannels;
	
	/**
	 * The number of output channels.
	 */
	private int outputChannels;
	
	/**
	 * The number of clut grid points.
	 */
	private int clutGridPoints;
	
	/**
	 * The conversion matrix.
	 */
	private float matrix[][] = new float[3][3];
	
	/**
	 * The number of entries in the input table.
	 */
	private int inputTableEntries;
	
	/**
	 * The number of entries in the output table.
	 */
	private int outputTableEntries;
	
	/**
	 * An array representing the input table.
	 */
	private float inputTable[];
	
	/**
	 * The clut values.
	 */
	private float clutValues[];
	
	/**
	 * n array representing the input table.
	 */
	private float outputTable[];
	
	/**
	 * A string identifying the lut.
	 */
	private String type;
 
	/**
	 * Constructs a LutReader from a given file, at a given offset.
	 * @param rafw The file.
	 * @param offset The offset.
	 */
	public LutReader(RandomAccessFileWrapper rafw, long offset) {
		rafw.setOffset(offset);
		byte fourBytes[] = new byte[4];
		byte oneByte[] = new byte[1];
		byte twoBytes[] = new byte[2];

		identifier = rafw.getLong();
		if (identifier == 0x6d667431) {
			type = "Lut8";
		} else if (identifier == 0x6d667432) {
			type = "Lut16";
		} else {
			throw new RuntimeException("Unknown lut type: "
					+ Long.toHexString(identifier));
		}
		rafw.getBytes(fourBytes);
		if (fourBytes[0] != 0 || fourBytes[1] != 0 || fourBytes[2] != 0
				|| fourBytes[3] != 0) {
			SystemManager.displayError(
				"Warning",
				"Reserved bytes contain data.");
		}
		rafw.getBytes(oneByte);
		inputChannels = oneByte[0] & (0xff);
		rafw.getBytes(oneByte);
		outputChannels = oneByte[0] & (0xff);
		rafw.getBytes(oneByte);
		clutGridPoints = oneByte[0] & 0xff;
		rafw.getBytes(oneByte);
		if (oneByte[0] != 0) {
			SystemManager.displayError(
				"Warning",
				"Reserved bytes contain data.");
		}
		byte b[] = new byte[4];

		for (int i = 0; i < 3; i++) {
			for (int j = 0; j < 3; j++) {
				rafw.getBytes(b);
				int o = (b[0]) & (0xff);

				o = o << 8;
				o += (b[1]) & (0xff);
				o = o << 8;
				o += (b[2]) & (0xff);
				o = o << 8;
				o += (b[3]) & (0xff);
				float f;

				f = o;
				f = f / 65536.0f;
				matrix[i][j] = f;
				
			}
		}
		
		if (identifier == 0x6d667431) {
			// Lut8
			inputTableEntries = 256;
			inputTable = new float[inputTableEntries * inputChannels];
			for (int i = 0; i < inputTableEntries * inputChannels; i++) {
				rafw.getBytes(oneByte);
				inputTable[i] = ((oneByte[0]) & (0xff)) / 255.0f;
			}
			
			int p = (int) Math.pow(clutGridPoints, inputChannels) * outputChannels;

			clutValues = new float[p];
			for (int i = 0; i < p; i++) {
				rafw.getBytes(oneByte);
				clutValues[i] = ((oneByte[0]) & (0xff)) / 255.0f;
			}
			outputTableEntries = 256;
			outputTable = new float[outputTableEntries * outputChannels];
			for (int i = 0; i < outputTableEntries * outputChannels; i++) {
				rafw.getBytes(oneByte);
				outputTable[i] = ((oneByte[0]) & (0xff)) / 255.0f;
			}
			
		} else if (identifier == 0x6d667432) {
			// Lut16
			inputTableEntries = rafw.getInt();
			outputTableEntries = rafw.getInt();
			inputTable = new float[inputChannels * inputTableEntries];
			for (int i = 0; i < inputChannels * inputTableEntries; i++) {
				inputTable[i] = rafw.getInt() / 65535.0f;
			}
		
			int p = (int) Math.pow(clutGridPoints, inputChannels) * outputChannels;

			clutValues = new float[p];
			
			for (int i = 0; i < p; i++) {
				clutValues[i] = rafw.getInt() / 65535.0f;
			}
			outputTable = new float[outputChannels * outputTableEntries];
			for (int i = 0; i < outputChannels * outputTableEntries; i++) {
				long off = rafw.getOffset();
				int tempInt = rafw.getInt();

				outputTable[i] = tempInt / 65535.0f;
			}
			
		}
	}
	
	/**
	 * Returns the output for a given input. Implements certain CMM computations
	 * necessary for translating using the LUT.
	 * @param input The input.
	 * @return The output.
	 */
	public float[] getOutput(float input[]) {
		float result[] = null;
		float afterMatrix[] = new float[3];

		afterMatrix[0] = input[0] * matrix[0][0] + input[1] * matrix[0][1]
				+ input[2] * matrix[0][2];
		afterMatrix[1] = input[0] * matrix[1][0] + input[1] * matrix[1][1]
				+ input[2] * matrix[1][2];
		afterMatrix[2] = input[0] * matrix[2][0] + input[1] * matrix[2][1]
				+ input[2] * matrix[2][2];
		
		int intoInput[] = new int[3];
		int tmp;

		for (int i = 0; i < 3; i++) {
			tmp = (int) (afterMatrix[i] * (float) inputTableEntries);
			if (tmp > inputTableEntries - 1) {
				tmp = inputTableEntries - 1;
			}
			intoInput[i] = tmp + inputTableEntries * i;
		}
		float afterInput[] = new float[3];

		for (int i = 0; i < 3; i++) {
			afterInput[i] = (float) inputTable[intoInput[i]];
		}
		
		int intoClut[] = new int[3];

		for (int i = 0; i < 3; i++) {
			tmp = (int) (afterInput[i] * (clutGridPoints - 1));
			if (tmp > clutGridPoints - 1) {
				tmp = clutGridPoints - 1;
			}
			intoClut[i] = tmp;
		}
		
		int lookupClut = intoClut[0] * clutGridPoints * clutGridPoints
				+ intoClut[1] * clutGridPoints + intoClut[2];
		float afterClut = (float) clutValues[lookupClut];		
		int intoOutput = (int) (afterClut * (float) (outputTableEntries));

		if (intoOutput > outputTableEntries - 1) {
			intoOutput = outputTableEntries - 1;
		}
		result = new float[1];
		result[0] = (float) outputTable[intoOutput];
		return result;
	}
	
	/**
	 * Returns information about an ICC lookup table.
	 * @return  A string containing information about the lut.
	 */
	public String getInfo() {
		StringBuffer sb = new StringBuffer();

		sb.append("Gamut tag info:\n");
		sb.append("LUT type: ");
		sb.append(type);
		sb.append("\n");
		sb.append("Input channels: ");
		sb.append(inputChannels);
		sb.append("\n");
		sb.append("Input table entries: ");
		sb.append(inputTableEntries);
		sb.append("\n");
		sb.append("Output channels: ");
		sb.append(outputChannels);
		sb.append("\n");
		sb.append("Output table entries: ");
		sb.append(outputTableEntries);
		sb.append("\n");
		sb.append("CLUT grid points: ");
		sb.append(clutGridPoints);
		sb.append("\n\n");
		sb.append("Matrix (3x3):\n");
		Matrix3f m3f = new Matrix3f(matrix[0][0], matrix[0][1], matrix[0][2],
				matrix[1][0], matrix[1][1], matrix[1][2], matrix[2][0], matrix[2][1],
				matrix[2][2]);

		sb.append(m3f.toString());
		sb.append("\n");

		sb.append("\nInput Table: \n");
		for (int i = 0; i < inputTableEntries; i++) {
			sb.append(i);
			sb.append(": ");
			for (int lop = 0; lop < inputChannels; lop++) {
				sb.append(inputTable[lop * inputTableEntries + i]);
				sb.append(" ");
			}
			sb.append("\n");
		}
		sb.append("\nOutput Table: \n");
		for (int i = 0; i < outputTableEntries; i++) {
			sb.append(i);
			sb.append(": ");
			for (int lop = 0; lop < outputChannels; lop++) {
				sb.append(outputTable[lop * outputTableEntries + i]);
				sb.append(" ");
			}
			sb.append("\n");
		}

		sb.append("\nCLUT values: \n");
		int j[] = new int[inputChannels];
		int curChannel = inputChannels - 1;

		while (curChannel >= 0) {
			int h = 0;

			for (int u = 0; u < inputChannels; u++) {
				h *= clutGridPoints;
				h += j[u];
			}
			curChannel = inputChannels - 1;
			j[curChannel]++;
			while (curChannel >= 0 && j[curChannel] >= clutGridPoints) {
				j[curChannel] = 0;
				curChannel--;
				if (curChannel >= 0) {
					j[curChannel]++;
				}
			}
			if (curChannel >= 0) {
				for (int cp = 0; cp < inputChannels; cp++) {
					sb.append(j[cp]);
					sb.append(" ");
				}
				sb.append(": ");
				sb.append(clutValues[h]);
				sb.append("\n");
			}
		}
		return sb.toString();
	}
}
