/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.image2d;


import java.io.*;
import java.awt.*;
import java.awt.image.IndexColorModel;
import java.awt.image.RenderedImage;
import javax.media.jai.JAI;
import javax.media.jai.LookupTableJAI;
import javax.media.jai.PlanarImage;
import javax.swing.*;
import javax.swing.event.*;
import com.sun.media.jai.codec.*;

import icc3dpackage.*;
import icc3dpackage.managers.*;


/**
 * This class is used to write imagefiles(of a given format) to a specified path.
 *
 */
public abstract class FileWriter {

	/**
	 * Writes an imagefile to storage media. Image must consist of less than 4 bands.
	 * If the file exists, the user is presented with an owerwrite dialog.
	 *	
	 * For example: the following code will try to save a RenderedImage to present 
	 * working directory:
	 * FileWriter.writeImage(renderedImage, "foo.tiff");
	 *
	 * @param image renderedImage
	 * @param fileName absolute or relative path. For example: [path/]foo.tif
	 * @return true if successful, otherwise false
	 * @see java.awt.image.RenderedImage
	 */ 
	public static boolean writeImage(RenderedImage image, String fileName) {
		boolean result = true;

		if (PlanarImage.wrapRenderedImage(image).getNumBands() == 4) {
			SystemManager.displayError(
				"Error",
				"Image contains 4 bands which is not"
					+ " supported for output. File save operation was aborted.");
			result = false;
		} else {
			File file = new File(fileName);

			if (file.exists()
					&& !(SystemManager.confirmYesNo("Warning", "File already exists:\n" + fileName + "\n\nOverwrite?") == JOptionPane.YES_OPTION)) {
				result = false;
			}
			if (result == true) {
				file = null;
				try {
					ImageEncodeParam param = createImageEncodeParam(fileName);

					if (param != null) {
						int index = -1;
						String type = getFileType(fileName);
						String names[] = ImageCodec.getEncoderNames(image, param);
						FileOutputStream stream = new FileOutputStream(new File(fileName));

						if (index == -1) {
							param = null;
						}
						ImageEncoder enc = ImageCodec.createImageEncoder(type, stream, param);

						enc.encode(image);
						stream.close();
						System.out.println("File written successfully. (" + fileName + ")");	
					}
				} catch (Exception e) {
					result = false;
				}	
			}
		}
    
		return result;
	}

	/**
	 *	Constructs an ImageEncodeParam for fileName.
	 *	
	 * @param fileName absolute or relative path to file to create ImageEncodeParam for.
	 * @return Image encode parameter
	 *	@see com.sun.media.jai.codec.ImageEncodeParam
	 */
	private static ImageEncodeParam createImageEncodeParam(String fileName) {
		ImageEncodeParam imageParam = null;
		String type = getFileType(fileName);

		if (type != null) {
			if (type.equals("tiff")) {
				imageParam = new TIFFEncodeParam();
			} // ---------------------
			// JPEG
			else if (type.equals("jpeg")) {
				JPEGEncodeParam tmp = new JPEGEncodeParam();
				JPanel p = new JPanel();

				p.setLayout(new FlowLayout());
				final JSlider s = new JSlider(JSlider.HORIZONTAL, 0, 100, 50);

				p.add(s);
				final JTextField f = new JTextField(5);

				f.setText(Float.toString(s.getValue() / 100.0f));
				f.setEditable(false);
				p.add(f);
				s.addChangeListener(new ChangeListener() {
					public void stateChanged(ChangeEvent e) {
						f.setText(Float.toString(s.getValue() / 100.0f));
					}
				});
				if (JOptionPane.showConfirmDialog(
					ICC3D.getDialogParent(),
					p,
					"JPEG compression quality",
					JOptionPane.OK_CANCEL_OPTION,
					1,
					IconManager.getIcon("question")
					) == JOptionPane.OK_OPTION) {
					try {
						float q = Float.parseFloat(f.getText());

						tmp.setQuality(q);
					} catch (NumberFormatException nfe) {
						tmp.setQuality(0.75f);
					}
					imageParam = tmp;
				} else {
					imageParam = null;
				}
				
			} // ----------------------
			else if (type.equals("bmp")) {
				imageParam = new BMPEncodeParam();
			} else if (type.equals("gif")) {
				System.out.println("gif not supported for encoding");
			} else if (type.equals("png")) {
				imageParam = new PNGEncodeParam.RGB();
			} else if (type.equals("pnm")) {
				imageParam = new PNMEncodeParam();
			} else if (type.equals("fpx")) {
				System.out.println("fpx not supported for encoding");
				SystemManager.displayError(
					"Error",
					"FlashPiX is not supported for output.\nPlease choose another format.");
			} else {
				System.out.println("Can't determine fileformat: " + type
						+ " - File NOT saved...");
				SystemManager.displayError(
					"Error",
					"Output format not supported(\"" + type
						+ "\").\nChoose another format.\nFile NOT saved.");
			}
		}
		return imageParam;
	}
	
	/**
	 * Determines the type of file.
	 *	
	 * @param fileName absolute or relative path to file to determine filetype.
	 * @return Filetypes. For example: foo.tiff, foo.jpeg, foo.bmp ...
	 */
	private static String getFileType(String fileName) {
		String retValue;

		try {
			retValue = fileName.substring(fileName.lastIndexOf('.') + 1,
					fileName.length());
			retValue = retValue.toLowerCase();
			if (retValue.equals("tif")) {
				retValue = "tiff";
			} else if (retValue.equals("jpg")) {
				retValue = "jpeg";
			}
		} catch (IndexOutOfBoundsException ioob) {
			retValue = null;
		}
		return retValue;
	}
	
}

