/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.image2d;


import javax.swing.JPanel;
import javax.media.jai.*;
import java.awt.*;
import java.awt.image.*;
import java.util.Arrays;
import javax.swing.JOptionPane;
import javax.swing.BorderFactory;

import icc3dpackage.modules.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;


/**
 * A class used to display 2D images.
 */
public class ImageCanvas2D extends JPanel {
	
	/** The image */
	private Image image;
	
	/** The image source */
	private MemoryImageSource memoryImageSource;
	
	/** The image data */
	private int imageData[];
	
	/** The image width */
	public int width;
	
	/** The image height */
	public int height;
	
	/** The image background color */
	private Color backgroundColor;
	
	/** The image color space */
	private ColorSpace3D colorSpace3D;
	
	/** The image time stamp (last modified) */
	private long timeStamp;
	
	/** The image thumbnail */
	private ImageCanvas2D_Thumbnail thumbnail;
	
	/** The display color space */
	private ColorSpace3D viewColorSpace;
	
	/** The input color space */
	private ColorSpace3D inputColorSpace;
	
	/** The colors that need to be recalculated */
	private boolean dirtyColor[];
	
	/** The view image data */
	private float viewColor[];
	
	/** The CIEXYZ image data */
	private float xyzColor[];
	
	/** The original unmodified image colors */
	private float originalXYZ[];
	
	/** The color to pixel index */
	private int colorToPixelIndex[][];
	
	/** The pixel to color index */
	private int pixelToColorIndex[];
	
	/** The number of unique colors */
	private int uniqueColors;
	
	/**
	 * Constructs an image canvas.
	 * @param p The image data.
	 */
	public ImageCanvas2D(ImageOpenParameter p) {
		colorSpace3D = ColorSpaceManager.getColorSpace("sRGB");
		viewColorSpace = ColorSpaceManager.getColorSpace("sRGB");
		inputColorSpace = p.getInputColorSpace3D();
		setBorder(BorderFactory.createEmptyBorder());
		width = p.getWidth();
		height = p.getHeight();
		setSize(width, height);
		this.setPreferredSize(new Dimension(width, height));
		this.setMinimumSize(new Dimension(width, height));
		colorToPixelIndex = p.getColorToPixelIndex();
		pixelToColorIndex = p.getPixelToColorIndex();
		xyzColor = p.getXYZ();
		originalXYZ = new float[xyzColor.length];
		System.arraycopy(xyzColor, 0, originalXYZ, 0, xyzColor.length);
		viewColor = new float[xyzColor.length];
		uniqueColors = xyzColor.length / 3;
		dirtyColor = new boolean[uniqueColors];
		Arrays.fill(dirtyColor, true);
		imageData = new int[width * height];
		fromXYZ_ToView();
		fromViewToImage();
		memoryImageSource = new MemoryImageSource(width, height, imageData, 0, width);
		memoryImageSource.setAnimated(true);
		memoryImageSource.setFullBufferUpdates(true);
		image = getToolkit().createImage(memoryImageSource);
		backgroundColor = new Color(0.0f, 0.0f, 0.0f);
	}
	
	/**
	 * Converts colors from CIEXYZ to the display color space.
	 */
	private void fromXYZ_ToView() {
		viewColorSpace.fromCIEXYZ(xyzColor, viewColor, dirtyColor);
		for (int i = 0; i < viewColor.length; i++) {
			if (viewColor[i] > 1.0f) {
				viewColor[i] = 1.0f;
			} else if (viewColor[i] < 0.0f) {
				viewColor[i] = 0.0f;
			}
		}
		
	}
	
	/**
	 * Converts the display colors to a viewable image.
	 */
	private void fromViewToImage() {
		for (int i = 0; i < uniqueColors; i++) {
			if (dirtyColor[i]) {
				dirtyColor[i] = false;
				int r = ((int) (viewColor[i * 3] * 255.99f)) << 16;
				int g = ((int) (viewColor[i * 3 + 1] * 255.99f)) << 8;
				int b = (int) (viewColor[i * 3 + 2] * 255.99f);
				
				int color = (0xff000000) | r | g | b;

				for (int j = 0; j < colorToPixelIndex[i].length; j++) {
					imageData[colorToPixelIndex[i][j]] = color;
				}
			}
		}
	}
	
	/**
	 * Constructs an image canvas
	 * @param planarImage The image.
	 */
	public ImageCanvas2D(PlanarImage planarImage) {
		colorSpace3D = ColorSpaceManager.getColorSpace("sRGB");
		setBorder(BorderFactory.createEmptyBorder());
		PlanarImage pi = planarImage;// JAI.create("fileload", file);

		width = pi.getWidth();
		height = pi.getHeight();
		setSize(width, height);
		this.setPreferredSize(new Dimension(width, height));
		this.setMinimumSize(new Dimension(width, height));
		int temp[] = null;
		Raster raster = pi.getData();

		imageData = new int[width * height];
		for (int i = 0; i < width; i++) {
			for (int j = 0; j < height; j++) {
				temp = raster.getPixel(i, j, temp);
				imageData[j * width + i] = (0xff000000) + (temp[0] << 16) | (temp[1] << 8)
						| temp[2];
			}
		}
		memoryImageSource = new MemoryImageSource(width, height, imageData, 0, width);
		memoryImageSource.setAnimated(true);
		memoryImageSource.setFullBufferUpdates(true);
		image = getToolkit().createImage(memoryImageSource);
		backgroundColor = new Color(0.0f, 0.0f, 0.0f);
	}
	
	/**
	 * Returns a planar image representation of the image.
	 * @return A planar image representation of the image.
	 */
	public PlanarImage getPlanarImage() {
		SampleModel sampleModel = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE,
				width, height, 3, width * 3, new int[] {0, 1, 2});
		ColorModel colorModel = PlanarImage.createColorModel(sampleModel);
		
		byte tmp[] = new byte[imageData.length * 3];

		for (int i = 0; i < imageData.length; i++) {
			tmp[i * 3] = (byte) ((imageData[i] & 0x00ff0000) >> 16);
			tmp[i * 3 + 1] = (byte) ((imageData[i] & 0x0000ff00) >> 8);
			tmp[i * 3 + 2] = (byte) (imageData[i] & 0x000000ff);
		}
		DataBufferByte dataBuffer = new DataBufferByte(tmp, tmp.length);

		Raster raster = RasterFactory.createWritableRaster(sampleModel, dataBuffer,
				new Point(0, 0));
		TiledImage tiledImage = new TiledImage(0, 0, width, height, 0, 0, sampleModel,
				colorModel);

		tiledImage.setData(raster);
		return PlanarImage.wrapRenderedImage(tiledImage);	
	}
	
	/**
	 * Pants the image.
	 */
	public void paint(Graphics g) {
		super.paint(g);
		if (image != null) {
			g.drawImage(image, 0, 0, backgroundColor, null);
		}
	}
	
	/**
	 * Returns the image color space.
	 */
	public ColorSpace3D getColorSpace3D() {
		return colorSpace3D;
	}
	
	/**
	 * Updates the last modified time.
	 */
	private void updateTimeStamp() {
		timeStamp = System.currentTimeMillis();
	}
	
	/**
	 * Returns the last modified time stamp.
	 * @return The last modified time stamp.
	 */
	public long getTimeStamp() {
		return timeStamp;
	}
	
	/**
	 * Adds a constant to each of the red, green and blue values of the picture
	 * @param rr the red value you want to increase with
	 * @param gg the green value you want to increase with
	 * @param bb the blue value you want to increase with
	 */
	public void addConst(int rr, int gg, int bb) {
		int r, g, b;
		int len = imageData.length;

		for (int i = 0; i < len; i++) {
			r = ((imageData[i] & 0x00ff0000) >> 16);
			g = ((imageData[i] & 0x0000ff00) >> 8);
			b = (imageData[i] & 0x000000ff);
			
			if (r + rr > 255) {
				r = 255;
			} else if (r + rr < 0) {
				r = 0;
			} else {
				r += rr;
			}
			if (g + gg > 255) {
				g = 255;
			} else if (g + gg < 0) {
				g = 0;
			} else {
				g += gg;
			}
			if (b + bb > 255) {
				b = 255;
			} else if (b + bb < 0) {
				b = 0;
			} else {
				b += bb;
			}
			imageData[i] = (0xff000000) + (r << 16) | (g << 8) | b;
		}
		memoryImageSource.newPixels();
		repaint();
		if (thumbnail != null) {
			thumbnail.repaint();
		}
	}
	
	/**
	 * Returns a reference to the pixels in the picture 
	 * @return int[] reference to the pixels 
	 */
	public int[] getImageData() {
		return imageData;	
	}
	
	/**
	 * Returns a reference to the memoryImageSource picture
	 * @return MemoryImageSource reference to the picture
	 */
	public MemoryImageSource getMemoryImageSource() {
		return memoryImageSource;
	}
	
	/**
	 * Returns the width of the picture
	 * @return int picture width
	 */
	public int getImageWidth() {
		return width;
	}
	
	/**
	 * Returns the height of the picture
	 * @return int picture height
	 */
	public int getImageHeight() {
		return height;	
	}
	
	/**
	 * Updates the image area with the new image
	 * @param image PlanarImage object of the new image that shall be shown
	 */
	public void updateImage(PlanarImage image) {
		createWritableImage(image);
		updateTimeStamp();
		repaint();
	}
	
	/**
	 * Initializes a MemoryImageSource of an PlanarImage so you can have direct 
	 * access to all the pixels in the picture
	 * @param sourceImage PlanarImage image which is converted
	 */
	private void createWritableImage(PlanarImage sourceImage) {
		memoryImageSource = null;
		imageData = null;
		int temp[] = null;
		Raster raster = sourceImage.getData();

		imageData = new int[width * height];
		for (int i = 0; i < width; i++) {
			for (int j = 0; j < height; j++) {
				temp = raster.getPixel(i, j, temp);
				imageData[j * width + i] = (0xff000000) + (temp[0] << 16) | (temp[1] << 8)
						| temp[2];
			}
		}
		memoryImageSource = new MemoryImageSource(width, height, imageData, 0, width);
		memoryImageSource.setAnimated(true);
		memoryImageSource.setFullBufferUpdates(true);
		image = getToolkit().createImage(memoryImageSource);
		updateTimeStamp();
	}
	
	/**
	 * Perform an update of the image pixels.
	 */
	public void imageUpdate() {
		fromXYZ_ToView();
		fromViewToImage();
		memoryImageSource.newPixels();
		updateTimeStamp();
		repaint();
		if (thumbnail != null) {
			thumbnail.repaint();
		}
	}
	
	/**
	 * Creates a thumbnail of this image.
	 * @param thumbnailMaxWidth The maximum thumbnail width.
	 * @param thumbnailMaxHeight The maximum thumbnail height.
	 */
	public ImageCanvas2D_Thumbnail createThumbnail(int thumbnailMaxWidth, int thumbnailMaxHeight) {
		thumbnail = new ImageCanvas2D_Thumbnail(this, thumbnailMaxWidth,
				thumbnailMaxHeight);
		return thumbnail;
	}
	
	/**
	 * Returns the dirty flag array.
	 * @return The dirty flag array.
	 */
	public boolean[] getDirtyColors() {
		return dirtyColor;
	}
	
	/**
	 * Returns the CIEXYZ colors.
	 * @return the CIEXYZ colors.
	 */
	public float[] getXYZ_Colors() {
		return xyzColor;
	}
	
	/**
	 * Returns the input color space.
	 * @return The input color space.
	 */
	public ColorSpace3D getInputColorSpace() {
		return inputColorSpace;
	}
	
	/**
	 * Returns the color to pixel index.
	 * @return The color to pixel index.
	 */
	public int[][] getColorToPixelIndices() {
		return colorToPixelIndex;
	}
	
	/**
	 * Returns the pixel to color index.
	 * @return The pixel to color index.
	 */
	public int[] getPixelToColorIndex() {
		return pixelToColorIndex;
	}
	
	/**
	 * Returns the image.
	 * @return The image.
	 */
	public Image getImage() {
		return image;
	}
	
	/**
	 * Reverts to the original, unmodified image.
	 */
	public void revertToOriginal() {
		System.arraycopy(originalXYZ, 0, xyzColor, 0, originalXYZ.length);
		Arrays.fill(dirtyColor, true);
		this.imageUpdate();
	}
	
	/**
	 * Changes the display color space.
	 * @param cs3d The new display color space.
	 */
	public void setViewColorSpace(ColorSpace3D cs3d) {
		viewColorSpace = cs3d;
	}
}
