/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.image2d;


import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.jai.*;
import com.sun.media.jai.codec.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.*;
import icc3dpackage.tiffreader.TIFF_Reader;
import icc3dpackage.utils.*;


/**
 * A class that displays an image open dialog.
 */
public class ImageOpenDialog extends JDialog {
	
	/** The last image directory */
	private static String lastImageLocation = ".";
	
	/** The last ICC profile directory */
	private static String lastProfileLocation = ".";
	
	/** The image parameters */
	private ImageOpenParameter ioParam;
	
	/** The available color space choices */
	private ButtonGroup buttonGroup;
	
	/** Construct the color space from an independent ICC profile */
	private JRadioButton fromProfile;
	
	/** Constructs the color space from a profile embedded into a TIFF file */
	private JRadioButton fromTIFF;
	
	/** Uses one of the available application color spaces */
	private JRadioButton fromColorSpace;
	
	/** Loads an independent ICC profile */
	private JButton profileFile;
	
	/** Loads a profile from a TIFF file */
	private JComboBox tiffFile;
	
	/** The available color spaces */
	private JComboBox colorSpaces;
	
	/** Load the image */
	private JButton loadButton;
	
	/** The image type */
	private JButton typeButton;
	
	/** The image size */
	private JButton imgDimension;
	
	/** The number of unique colors */
	private JButton uniqueColors;
	
	/** The number of image components (samples per pixel) */
	private JButton componentButton;
	
	/** Quantize the image colors */
	private JCheckBox quantizeImage;
	
	/** Use relative colorimetry */
	private JCheckBox useRelative;
	
	/** Dismiss dialog and load image */
	private JButton okButton;
	
	/** Dismiss dialog and cancel loading of the image */
	private JButton cancelButton;
	
	/** The selected ICC profile */
	private String selectedProfile;
	
	/** The selected image */
	private String selectedImage;
	
	/** The device dependent image colors */
	private float tmpData[];
	
	/** The image width */
	private int width;
	
	/** The image height */
	private int height;
	
	/** The number of image components */
	private int components;
	
	/** The pixel to color index */
	private int pixelToColorIndex[];
	
	/** The color to pixel index */
	private int colorToPixelIndex[][];
	
	/** The ICC profile color space */
	private ICC_ProfileColorSpace3D iccColorSpace3D;
	
	/** The ICC profiles loaded from the TIFF file */
	private ICC_Profile profiles[];
	
	/**
	 * Constructs and displays an image open dialog with the specified parameters.
	 */
	public ImageOpenDialog(ImageOpenParameter ioParameter) {
		super(ICC3D.getDialogParent(), "Open image file", true);
		this.ioParam = ioParameter;
		okButton = new JButton("OK");
		okButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (selectedImage != null) {
					ColorSpace3D space = null;

					if (fromTIFF.isSelected()) {
						space = new ICC_ProfileColorSpace3D(profiles[tiffFile.getSelectedIndex()]);
					} else if (fromProfile.isSelected()) {
						space = iccColorSpace3D;
					} else {
						space = ColorSpaceManager.getColorSpace(colorSpaces.getSelectedItem().toString());
					}
					if (space.getNumberOfComponents() == components) {
						if (space.getName().equals("Lab")) {
							int count = tmpData.length / 3;
							float retColor[] = new float[tmpData.length];

							for (int i = 0; i < count; i++) {
								retColor[i * 3 + 1] = tmpData[i * 3];
								if (tmpData[i * 3 + 1] < 0.5f) {
									retColor[i * 3 + 2] = 0.5f + tmpData[i * 3 + 1];
								} else {
									retColor[i * 3 + 2] = tmpData[i * 3 + 1] - 0.5f;
								}
								if (tmpData[i * 3 + 2] < 0.5f) {
									retColor[i * 3] = 0.5f + tmpData[i * 3 + 2];
								} else {
									retColor[i * 3] = tmpData[i * 3 + 2] - 0.5f;
								}
							}
							tmpData = retColor;
						}
						float xyzData[] = space.toCIEXYZ(tmpData);
						float white[] = space.getWhitePoint();

						if (!useRelative.isSelected() && white != null) {
							Colorimetry.toAbsoluteXYZ_CAT02(xyzData, white[0], white[1], white[2]);
						}
						ioParam.set(space, width, height, xyzData, tmpData, pixelToColorIndex,
								colorToPixelIndex, quantizeImage.isSelected());
						dispose();
					} else {
						SystemManager.displayError(
							"Error",
							"The number of components in the color space does not match the image data");
					}
				} else {
					SystemManager.displayError(
						"Error",
						"No image selected");
				}
			}
		});
		cancelButton = new JButton("Cancel");
		cancelButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				ioParam.fail();
				dispose();
			}
		});
		buttonGroup = new ButtonGroup();
		profileFile = new JButton("None");
		profileFile.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser jfc = new JFileChooser(lastProfileLocation);
				CustomFileFilter cff = new CustomFileFilter("ICC profiles", true,
						new String[] {"ICC", "ICM"});

				jfc.setFileFilter(cff);
				int returnVal = jfc.showOpenDialog(ICC3D.getDialogParent());

				if (returnVal == JFileChooser.APPROVE_OPTION
						&& jfc.getSelectedFile() != null) {
					try {
						iccColorSpace3D = new ICC_ProfileColorSpace3D(jfc.getSelectedFile().getAbsolutePath());
						selectedProfile = jfc.getSelectedFile().getAbsolutePath();
						profileFile.setText(jfc.getSelectedFile().getName());
						fromProfile.setSelected(true);
					} catch (Exception e) {
						iccColorSpace3D = null;
						SystemManager.displayError(
							"Error",
							"Failed to load ICC profile");
						profileFile.setText("None");
						fromColorSpace.setSelected(true);
					}
					lastProfileLocation = jfc.getSelectedFile().getParent();
				} else if (profileFile.getText().equals("None") && fromProfile.isSelected()) {
					fromColorSpace.setSelected(true);
				}
			}
		});
		fromProfile = new JRadioButton("Use ICC profile");
		fromProfile.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (iccColorSpace3D == null) {
					ActionListener al[] = profileFile.getActionListeners();

					for (int i = 0; i < al.length; i++) {
						al[i].actionPerformed(new ActionEvent(this, ActionEvent.ACTION_PERFORMED,
								"yo"));
					}
				}
			}
		});
		buttonGroup.add(fromProfile);
		tiffFile = new JComboBox(new String[] {"None"});
		tiffFile.setEnabled(false);
		fromTIFF = new JRadioButton("Use ICC profile embedded in TIFF file");
		fromTIFF.setEnabled(false);
		buttonGroup.add(fromTIFF);
		colorSpaces = new JComboBox(ColorSpaceManager.getAllNames());
		for (int i = 0; i < colorSpaces.getItemCount(); i++) {
			if (colorSpaces.getItemAt(i).toString().equals("sRGB")) {
				colorSpaces.setSelectedIndex(i);
			}
		}
		fromColorSpace = new JRadioButton("Use color space");
		fromColorSpace.setSelected(true);
		buttonGroup.add(fromColorSpace);
		
		JPanel fromPanel = new JPanel(new GridLayout(3, 2));

		fromPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Color space"));
		fromPanel.add(fromColorSpace);
		fromPanel.add(colorSpaces);
		fromPanel.add(fromProfile);
		fromPanel.add(profileFile);
		fromPanel.add(fromTIFF);
		fromPanel.add(tiffFile);
		
		JLabel loadLabel = new JLabel("Load:");

		loadButton = new JButton("Click here to select a file");
		loadButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser jfc = new JFileChooser(lastImageLocation);
				CustomFileFilter cff, cff2 = new CustomFileFilter("All image formats", true,
						new String[] {"TIF", "TIFF", "PNG", "GIF", "JPG", "JPEG"});

				jfc.addChoosableFileFilter(cff2);
				cff = new CustomFileFilter("TIFF", true, new String[] {"TIF", "TIFF"});
				jfc.addChoosableFileFilter(cff);
				cff = new CustomFileFilter("JPEG", true, new String[] {"JPG", "JPEG"});
				jfc.addChoosableFileFilter(cff);
				cff = new CustomFileFilter("PNG", true, new String[] {"PNG"});
				jfc.addChoosableFileFilter(cff);
				cff = new CustomFileFilter("GIF", true, new String[] {"GIF"});
				jfc.addChoosableFileFilter(cff);
				jfc.setFileFilter(cff2);
				int returnVal = jfc.showOpenDialog(ICC3D.getDialogParent());

				if (returnVal == JFileChooser.APPROVE_OPTION
						&& jfc.getSelectedFile() != null) {
					ProgressBar bar = null;
					boolean hasAlpha = true;

					try {
						bar = new ProgressBar(7, "Processing image. Please stand by...",
								"Loading image");
						PlanarImage image = null;
						String fileName = jfc.getSelectedFile().getAbsolutePath();

						if (System.getProperty("JAI_IMAGE_READER_USE_CODECS") == null) {
							try {
								image = JAI.create("fileload", fileName).createInstance();
							} catch (Exception e) {
								try {
									image = RawLabReader.getPlanarImage(fileName, bar);
								} catch (Exception ee) {
									throw new RuntimeException("The file format was not recognized");
								}
							}
						} else {
							try {
								SeekableStream stream = new FileSeekableStream(fileName);
								String[] names = ImageCodec.getDecoderNames(stream);
								ImageDecoder dec = ImageCodec.createImageDecoder(names[0], stream, null);
								RenderedImage ri = dec.decodeAsRenderedImage();

								image = PlanarImage.wrapRenderedImage(ri);
								stream.close();
							} catch (Exception e) {
								try {
									image = RawLabReader.getPlanarImage(fileName, bar);
								} catch (Exception ee) {
									throw new RuntimeException("The file format was not recognized");
								}
							}
						}
						if (image != null && image.getColorModel() instanceof IndexColorModel) {
							IndexColorModel icm = (IndexColorModel) image.getColorModel();
							int mapSize = icm.getMapSize();
							int comSizes[] = icm.getComponentSize();
							byte[][] lutData = new byte[comSizes.length][mapSize];

							icm.getReds(lutData[0]);
							if (comSizes.length > 1) {
								icm.getGreens(lutData[1]);
							}
							if (comSizes.length > 2) {
								icm.getBlues(lutData[2]);
							}
							if (comSizes.length > 3) {
								icm.getAlphas(lutData[3]);
							}
							LookupTableJAI lut = new LookupTableJAI(lutData);

							image = JAI.create("lookup", image, lut);
						}
						PlanarImage pi = image;

						typeButton.setText(jfc.getSelectedFile().getName().substring(jfc.getSelectedFile().getName().lastIndexOf(".") + 1, jfc.getSelectedFile().getName().length()).toUpperCase());
						if (typeButton.getText().equals("TIF")) {
							typeButton.setText("TIFF");
						}
						if (typeButton.getText().equals("JPG")) {
							typeButton.setText("JPEG");
						}
						if (typeButton.getText().equals("TIFF")) {
							TIFF_Reader reader = new TIFF_Reader(jfc.getSelectedFile().getAbsolutePath());

							hasAlpha = reader.hasAlpha();
							profiles = reader.getICC_Profiles();
							if (profiles.length > 0) {
								fromTIFF.setEnabled(true);
								tiffFile.setEnabled(true);
								tiffFile.removeAllItems();
								for (int i = 0; i < profiles.length; i++) {
									tiffFile.addItem("Profile " + (i + 1));
								}
							} else {
								fromTIFF.setEnabled(false);
								tiffFile.setEnabled(false);
								fromColorSpace.setSelected(true);
								tiffFile.removeAllItems();
								tiffFile.addItem("None");
							}
						} else {
							fromTIFF.setEnabled(false);
							tiffFile.setEnabled(false);
							fromColorSpace.setSelected(true);
							tiffFile.removeAllItems();
							tiffFile.addItem("None");
						}
						width = pi.getWidth();
						height = pi.getHeight();
						imgDimension.setText("" + width + "x" + height);
						int temp[] = null;
						Raster raster = pi.getData();

						if (hasAlpha) {
							hasAlpha = pi.getColorModel().hasAlpha();
						}
						components = raster.getNumBands();
						if (components == 4) {
							int choice = JOptionPane.showOptionDialog(ICC3D.getDialogParent(),
									"Your image contains 4 components per pixel.\nShould this be interpreted as 4 components (CMYK) or 3+alpha (RGBA)?",
									"Image interpretation", JOptionPane.YES_NO_OPTION,
									JOptionPane.PLAIN_MESSAGE, IconManager.getIcon("question"),
									new String[] {"3 components + alpha channel", "4 components"}, null);

							hasAlpha = false;
							if (choice == 0) {
								hasAlpha = true;
							}
							bar.show();
							bar.doUpdate();
						}
						if (hasAlpha && components > 1) {
							components--;
							componentButton.setText("" + components + " (+alpha)");
						} else {
							componentButton.setText("" + components);
						}
						if (components == 3) {
							quantizeImage.setEnabled(true);
						} else {
							quantizeImage.setSelected(false);
							quantizeImage.setEnabled(false);
						}
						int sampleSize = raster.getDataBuffer().getDataTypeSize(raster.getDataBuffer().getDataType());

						if (sampleSize * components > 64) {
							throw new RuntimeException("Sample size * samples per pixel exceeds 64 bits");
						}
						long imageData[] = new long[width * height];
						int pix = 0;

						for (int i = 0; i < width; i++) {
							for (int j = 0; j < height; j++) {
								temp = raster.getPixel(i, j, temp);
								pix = j * width + i;
								imageData[pix] = temp[0];
								for (int l = 1; l < components; l++) {
									imageData[pix] = (imageData[pix] << sampleSize) | temp[l];
								}
							}
						}
						// ---------------------------------------------------------------------------------------------
						bar.inc("Counting unique colors.");
						long maxUnique = (long) Math.pow(2, sampleSize * components);
						float maxSampleValue = (float) Math.pow(2, sampleSize) - 1;
						int sampleMask = (int) Math.pow(2, sampleSize) - 1;
						byte pixelCount[];
						int index;
						int numUnique = 0;
						int nPixels = width * height;
						long uP[] = new long[nPixels];
						long data = 0;

						if (sampleSize * components <= 24) {
							BitSet bitSet = new BitSet((int) maxUnique);

							for (int i = 0; i < nPixels; i++) {
								index = (int) imageData[i];
								if (!bitSet.get(index)) {
									uP[numUnique] = imageData[i];
									numUnique++;
									bitSet.set(index);
								}
							}
						} else {
							long imgCopy[] = new long[imageData.length];

							System.arraycopy(imageData, 0, imgCopy, 0, imageData.length);
							Arrays.sort(imgCopy);
							long old = imgCopy[0] + 1;

							for (int i = 0; i < nPixels; i++) {
								data = imgCopy[i];
								if (data != old) {
									uP[numUnique++] = data;
									old = data;
								}
							}

							/*
							 for (int i=0; i<nPixels; i++) {
							 data = imageData[i];
							 boolean found = false;
							 for (int j=0; j<numUnique; j++) {
							 if (uP[j]==data) {
							 found = true;
							 break;
							 }
							 }
							 if (!found) {
							 uP[numUnique++] = data;
							 }
							 }
							 */
						}
						uniqueColors.setText(Integer.toString(numUnique));
						bar.inc("Sorting colors.");
						long newUP[] = new long[numUnique];

						System.arraycopy(uP, 0, newUP, 0, numUnique);
						uP = newUP;
						Arrays.sort(uP);
						bar.inc("Creating floating point representation of data.");
						tmpData = new float[numUnique * components];

						int q;

						for (int i = 0; i < numUnique; i++) {
							q = i * components;
							for (int j = 0; j < components; j++) {
								tmpData[q + j] = (float) (((uP[i] >> (sampleSize * (components - j - 1)))
												& sampleMask)
										/ maxSampleValue);
							}
						}
						bar.inc("Indexing pixels with corresponding color.");
						int numEqual[] = new int[numUnique];

						pixelToColorIndex = new int[nPixels];
						for (int i = 0; i < nPixels; i++) {
							index = Arrays.binarySearch(uP, imageData[i]);
							numEqual[index]++;
							pixelToColorIndex[i] = index;
						}
						bar.inc("Building necessary structures.");
						colorToPixelIndex = new int[numUnique][];
						int lastImageIndex[] = new int[numUnique];

						for (int i = 0; i < numUnique; i++) {
							colorToPixelIndex[i] = new int[numEqual[i]];
						}
						bar.inc("Indexing image pixels.");
						for (int i = 0; i < nPixels; i++) {
							index = pixelToColorIndex[i];
							colorToPixelIndex[index][lastImageIndex[index]++] = i;
						}
						bar.inc();
						bar.setVisible(false);
						bar.dispose();
						// ---------------------------------------------------------------------------------------------
						loadButton.setText(jfc.getSelectedFile().getName());
						selectedImage = jfc.getSelectedFile().getAbsolutePath();
					} catch (Exception e) {
						bar.setVisible(false);
						bar.dispose();
						loadButton.setText("Click here to select a file");
						typeButton.setText("None");
						uniqueColors.setText("0");
						imgDimension.setText("");
						componentButton.setText("");
						fromTIFF.setEnabled(false);
						tiffFile.setEnabled(false);
						fromColorSpace.setSelected(true);
						tiffFile.removeAllItems();
						tiffFile.addItem("None");
						selectedImage = null;
						tmpData = null;
						width = 0;
						height = 0;
						pixelToColorIndex = null;
						colorToPixelIndex = null;
						SystemManager.displayError(
							"Error",
							"Failed to load image file\n" + e.getMessage());
						e.printStackTrace();
					}
					lastImageLocation = jfc.getSelectedFile().getParent();
				}
			}
		});
		quantizeImage = new JCheckBox();
		useRelative = new JCheckBox();
		useRelative.setSelected(true);
		JLabel imgType = new JLabel("Image type:");

		typeButton = new JButton("None");
		uniqueColors = new JButton("0");
		imgDimension = new JButton("");
		typeButton.setEnabled(false);
		uniqueColors.setEnabled(false);
		imgDimension.setEnabled(false);
		componentButton = new JButton("");
		componentButton.setEnabled(false);
		JPanel imgPanel = new JPanel(new GridLayout(0, 2));

		imgPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Image"));
		imgPanel.add(loadLabel);
		imgPanel.add(loadButton);
		imgPanel.add(imgType);
		imgPanel.add(typeButton);
		imgPanel.add(new JLabel("Size"));
		imgPanel.add(imgDimension);
		imgPanel.add(new JLabel("Components"));
		imgPanel.add(componentButton);
		imgPanel.add(new JLabel("Unique colors"));
		imgPanel.add(uniqueColors);
		// imgPanel.add(new JLabel("Quantize colors"));
		// imgPanel.add(quantizeImage);
		imgPanel.add(new JLabel("Relative colorimetry"));
		imgPanel.add(useRelative);
		
		JPanel buttonPanel = new JPanel();
		JPanel buttonGrid = new JPanel(new GridLayout(1, 0));

		buttonGrid.add(okButton);
		buttonGrid.add(new JPanel() {
			public Dimension getPreferredSize() {
				return new Dimension(100, 0);
			}
		});
		buttonGrid.add(cancelButton);
		buttonPanel.add(buttonGrid);
		
		Container c = getContentPane();

		c.setLayout(new BorderLayout());
		c.add(fromPanel, BorderLayout.CENTER);
		c.add(buttonPanel, BorderLayout.SOUTH);
		c.add(imgPanel, BorderLayout.NORTH);
		setResizable(false);
		setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		pack();
		setLocation(Toolkit.getDefaultToolkit().getScreenSize().width / 2
				- (getWidth() / 2),
				Toolkit.getDefaultToolkit().getScreenSize().height / 2 - (getHeight() / 2));
		show();
	}
	
	/**
	 * Displays an image open dialog with default options.
	 */
	public static ImageOpenParameter open() {
		ImageOpenParameter p = new ImageOpenParameter();
		ImageOpenDialog ioDialog = new ImageOpenDialog(p);

		return p;
	}
}
