/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.image2d;


import icc3dpackage.colorspaces.generic.ColorSpace3D;


/**
 * This class is used to transfer data to and from the open image dialog.
 */
public class ImageOpenParameter {
	
	/** The image color space */
	private ColorSpace3D colorSpace;
	
	/** The image width */
	private int width;
	
	/** The image height */
	private int height;
	
	/** The CIEXYZ colors present in the image */
	private float xyzData[];
	
	/** The device dependent colors present in the image */
	private float inData[];
	
	/** An index used to find the color of an image pixel */
	private int pixelToColorIndex[];
	
	/** An index used to find pixels with a given color */
	private int colorToPixelIndex[][];
	
	/** Signals that the image was read successfully */
	private boolean imageReadSuccessfully;
	
	/** Signals that the image colors should be quantized */
	private boolean quantizeImage;
	
	/**
	 * Signals that the opening of the image failed.
	 */
	public void fail() {
		imageReadSuccessfully = false;
	}
	
	/**
	 * Sets the image data read from the file.
	 * @param colorSpace The image color space.
	 * @param width The image width.
	 * @param height The image height.
	 * @param xyzData The CIEXYZ colors.
	 * @param inData The device dependent colors.
	 * @param pixelToColorIndex The pixel to color index.
	 * @param colorToPixelIndex The color to pixel index.
	 * @param quantize Use quantization of the image colors.
	 */
	public void set(ColorSpace3D colorSpace, int width, int height, float xyzData[], float inData[], int pixelToColorIndex[], int colorToPixelIndex[][], boolean quantize) {
		imageReadSuccessfully = true;
		this.colorSpace = colorSpace;
		this.width = width;
		this.height = height;
		this.xyzData = xyzData;
		this.inData = inData;
		this.pixelToColorIndex = pixelToColorIndex;
		this.colorToPixelIndex = colorToPixelIndex;
		this.quantizeImage = quantize;
	}
	
	/**
	 * Returns true if the image was read successfully.
	 * @return True if the image was read successfully.
	 */
	public boolean imageReadSuccessfully() {
		return imageReadSuccessfully;
	}
	
	/**
	 * Returns the width of the image.
	 * @return The image width.
	 */
	public int getWidth() {
		return width;
	}
	
	/**
	 * Returns the image height.
	 * @return The image height.
	 */
	public int getHeight() {
		return height;
	}
	
	/**
	 * Returns the pixel to color index.
	 * @return The pixel to color index.
	 */
	public int[] getPixelToColorIndex() {
		return pixelToColorIndex;
	}
	
	/**
	 * Returns the color to pixel index.
	 * @return The color to pixel index.
	 */
	public int[][] getColorToPixelIndex() {
		return colorToPixelIndex;
	}
	
	/**
	 * Returns the CIEXYZ image colors.
	 * @return The CIEXYZ image colors.
	 */
	public float[] getXYZ() {
		return xyzData;
	}
	
	public ColorSpace3D getInputColorSpace3D() {
		return colorSpace;
	}
	
	/**
	 * Returns the device dependent image colors.
	 * @return The device dependent image colors.
	 */
	public float[] getInput() {
		return inData;
	}
	
	/**
	 * Returns true if the image should be quantized.
	 * @return True if the image should be quantized.
	 */
	public boolean getQuantize() {
		return quantizeImage;
	}
}
