/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.image2d;


import java.awt.*;
import java.awt.image.*;
import java.io.*;
import javax.swing.*;
import javax.media.jai.*;

import icc3dpackage.ProgressBar;
import icc3dpackage.managers.SystemManager;


/**
 * This class provides basic support for reading raw image data from a file without a
 * file header.
 */
public abstract class RawLabReader {
	
	/**
	 * This method attempts to read raw image data from the specified file.
	 * @param file The name of the input file.
	 * @param bar The ProgressBar used to display progress feedback.
	 * @return A PlanarImage.
	 */
	public static PlanarImage getPlanarImage(String file, ProgressBar bar) {
		
		JTextField widthField = new JTextField();
		JTextField heightField = new JTextField();
		
		JPanel optionPanel = new JPanel(new GridLayout(0, 2));

		optionPanel.add(new JLabel("Width"));
		optionPanel.add(widthField);
		optionPanel.add(new JLabel("Height"));
		optionPanel.add(heightField);
		
		int choice = 0;
		int width = 0;
		int height = 0;
		boolean done = false;

		do {
			choice = SystemManager.confirmYesNo(
				"Attempt to read raw CIELAB image data",
				"Read raw image data?");
			try {
				width = Integer.parseInt(widthField.getText());
				height = Integer.parseInt(heightField.getText());
			} catch (Exception e) {
				width = 0;
				height = 0;
			}
		} while ((width == 0 || height == 0) && choice == JOptionPane.OK_OPTION);
		
		if (choice != JOptionPane.YES_OPTION) {
			throw new RuntimeException("Unknown file format encountered");
		}
		bar.show();
		bar.setDescription("Attempting to read raw image data");
		FileInputStream fis = null;
		int bytes[] = new int[width * height * 3];

		try {
			fis = new FileInputStream(file);
			for (int i = 0; i < width * height * 3; i++) {
				int integer = fis.read();

				if (integer == -1) {
					throw new RuntimeException("Error reading raw data");
				}
				bytes[i] = (byte) integer;
			}
		} catch (Exception e) {
			try {
				fis.close();
			} catch (Exception ee) {}
			throw new RuntimeException("Error reading raw data");
		}
		
		SampleModel sampleModel = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE,
				width, height, 3, width * 3, new int[] {0, 1, 2});
		ColorModel colorModel = PlanarImage.createColorModel(sampleModel);
		
		byte tmp[] = new byte[width * height * 3];

		for (int i = 0; i < width * height; i++) {
			tmp[i * 3] = (byte) bytes[i * 3];
			tmp[i * 3 + 1] = (byte) (bytes[i * 3 + 1] - 128);
			tmp[i * 3 + 2] = (byte) (bytes[i * 3 + 2] - 128);
		}
		DataBufferByte dataBuffer = new DataBufferByte(tmp, tmp.length);

		Raster raster = RasterFactory.createWritableRaster(sampleModel, dataBuffer,
				new Point(0, 0));
		TiledImage tiledImage = new TiledImage(0, 0, width, height, 0, 0, sampleModel,
				colorModel);

		tiledImage.setData(raster);
		try {
			fis.close();
		} catch (Exception e) {}
		return PlanarImage.wrapRenderedImage(tiledImage);	
	}
}
