/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.awt.*;
import javax.swing.JPanel;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * This class represents a local transform. It gives you the opportunity to move
 * one object independent of the rest of the scene.
 */
public class BasicInteraction3D extends SimpleInteraction3D {

	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();

		boolean dirty[] = image2d.getDirtyColors();
		int colorIndex[] = objects[index].getColorIndices();
		float colors[] = mo.getColors();
		float movement[] = new float[3];

		for (int i = 0; i < 3; i++) {
			movement[i] = end[i] - start[i];
		}
		int col, inArray;

		for (int i = 0; i < colorIndex.length; i++) {
			col = colorIndex[i];
			inArray = col * 3;
			dirty[col] = true;
			colors[inArray] += movement[0];
			colors[inArray + 1] += movement[1];
			colors[inArray + 2] += movement[2];
		}
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
	}
	
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */		
	public boolean isSupported(PickableObject object) {
		return true;
	}
	
	/**
	 * Returns that the current movement is supported
	 * @return boolean true
	 */
	public Component getParameterPanel() {
		return new JPanel();
	}
	
	/**
	 * Returns the name of the class
	 * @return string name
	 */
	public String getName() {
		return "Standard";
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 1;
	}
	
	/**
	 * Returns what type of transform this is, in this case it is an local and not a global transform
	 * @return byte type of transform
	 */
	public byte getType() {
		return Interaction3D.LOCAL;
	}
}
