/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.awt.*;
import java.util.Arrays;
import javax.swing.JPanel;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * Global transform, scales all objects according to the movement on the L-axis (in the LAB color space).
 */ 
public class GlobalTransformScaleL extends Interaction3D {
	
	/** The movement starting point */
	private float startingPoint[];
	
	/** The movement ending point */
	private float endPoint[];
	
	/**
	 * Constructs a GlobalTransformScaleL.
	 */
	public GlobalTransformScaleL() {}
		
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public void update3D(int index, MovingObject movingObject, float start[], float end[]) {
		ColorSpace3D colorSpace3D = movingObject.getColorSpace();
		PickableObject objects[] = movingObject.getObjects();

		if (start[1] == 0.5f) { 
			return;
		}
			
		if ((end[1] >= 0.5f && start[1] < 0.5f) || (end[1] <= 0.5f && start[1] > 0.5f)) {
			return;
		}
		if (startingPoint == null) {
			startingPoint = start;
		}
		endPoint = end;
		
		float pos[];
		float yres = (end[1] - start[1]) / (start[1] - 0.5f);
		
		GeometryArray ga[] = movingObject.getGeometryArrays();

		if (ga == null || ga.length == 0) {
			if (temporaryStorage == null || temporaryStorage.length != objects.length * 3) {
				temporaryStorage = new float[objects.length * 3];
			}
	
			for (int i = 0; i < objects.length; i++) {
				// npos = new float[3];
				pos = objects[i].getPositionReference();
				if (pos[1] == 0.5f) { 
					break;
				}
				temporaryStorage[i * 3] = pos[0];
				temporaryStorage[i * 3 + 1] = pos[1] = ((pos[1] - 0.5f) * yres) + pos[1];
				temporaryStorage[i * 3 + 2] = pos[2];
			}
			byte sRGB[] = colorSpace3D.toSRGB(temporaryStorage);
			byte color[];

			for (int i = 0; i < objects.length; i++) {
				color = objects[i].getColorReference();
				color[0] = sRGB[i * 3];
				color[1] = sRGB[i * 3 + 1];
				color[2] = sRGB[i * 3 + 2];
				objects[i].updatePosition();
				objects[i].updateColor();
			}
		} else {
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				if (pos[1] == 0.5f) {
					break;
				}
				pos[1] += ((pos[1] - 0.5f) * yres);
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				for (int j = 0; j < vCount; j++) {
					pos[j * 3 + 1] += ((pos[j * 3 + 1] - 0.5f) * yres);
				}
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, colorSpace3D.toSRGB(pos));
			}
		}
	}

	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		if (startingPoint == null || endPoint == null) {
			return;
		}
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		
		float colors[] = mo.getColors();
		boolean dirty[] = image2d.getDirtyColors();

		Arrays.fill(dirty, true);
		
		float yres = (endPoint[1] - startingPoint[1]) / (startingPoint[1] - 0.5f);
		int count = colors.length / 3;
		
		for (int i = 0; i < count; i++) {
			colors[i * 3 + 1] = ((colors[i * 3 + 1] - 0.5f) * yres) + colors[i * 3 + 1];
		}
		startingPoint = null;
		endPoint = null;
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
	}
		
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */
	public Component getParameterPanel() {
		return new JPanel();
	}

	/**
	 * Returns that the current movement is supported
	 * @return boolean true
	 */
	public boolean isSupported(PickableObject object) {
		return true;
	}

	/**
	 * Returns the name of the class
	 * @return string name
	 */
	public String getName() {
		return "Scale L";		
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 3;
	}
	
	/**
	 * Returns what type of transform this is, in this case it is an global and not 
	 * local transform
	 * @return byte type of transform
	 */
	public byte getType() {
		return Interaction3D.GLOBAL;
	}
}
