/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.lang.Math;
import java.awt.*;
import java.awt.event.*;
import java.util.Arrays;
import javax.swing.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.image2d.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * Global transform, tilts all objects over one chosen object and by another chosen object.Gives you the opportunity to "raise" the entire
 * scene.
 */
public class GlobalTransformTilt extends Interaction3D {
	
	/**
	 * The different modes of operation available.
	 */
	private JCheckBox choosePoint;
 	
	/**
	 * The parameter panel.
	 */
	private JPanel parameterPanel; 
 	
	/**
	 * The selected tilt point.
	 */
	private float tiltPoint[];
	
	/**
	 * The old object index.
	 */
	private int oldIndex;
  
	/**
	 * A vector used when computing the movement plane.
	 */
	private Vector3f u;
  
	/**
	 * A point used to calculate several vectors.
	 */
	private float upos[];
  
	/**
	 * The index of the rotation point.
	 */
	private int rotationPointIndex = 0;
  
	/**
	 * The total tilt angle.
	 */
	private float totalAngle = 0.0f;
  
	/**
	 * A radio button offering a choice of operation mode.
	 */
	private JRadioButton one, two, three;
	
	/**
	 * Constructor that initializes the object
	 */

	public GlobalTransformTilt() {}
   
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public void update3D(int index, MovingObject movingObject, float start[], float end[]) {
		ColorSpace3D colorSpace3D = movingObject.getColorSpace();
		PickableObject objects[] = movingObject.getObjects();
		float scale[] = colorSpace3D.getScale();

		if (one.isSelected()) {
			if (oldIndex < objects.length) {
				objects[oldIndex].setColor(colorSpace3D.toSRGB(objects[oldIndex].getPosition()));
			}
			byte tmpArray[] = objects[index].getColor();

			tiltPoint = objects[index].getPosition();
			tiltPoint[0] *= scale[0];
			tiltPoint[1] *= scale[1];
			tiltPoint[2] *= scale[2];
			objects[index].setColor(new byte[] {(byte) 0xff, (byte) 0xff, 0});
			oldIndex = index;
			two.setSelected(true);
			return;
		}
		start = new float[] {start[0] * scale[0], start[1] * scale[1],
			start[2] * scale[2]};
		if (two.isSelected()) {
			if (tiltPoint[0] == start[0] && tiltPoint[1] == start[1]
					&& tiltPoint[2] == start[2]) {
				return;
			}
			rotationPointIndex = index;
			Vector3f vec2 = new Vector3f(start[0] - tiltPoint[0], start[1] - tiltPoint[1],
					start[2] - tiltPoint[2]);

			vec2.normalize();
		
			Vector3f vec1 = new Vector3f(0.0f, 1.0f, 0.0f);

			u = new Vector3f();
			u.cross(vec1, vec2);
			u.normalize();
			upos = new float[3];
			u.get(upos);
			three.setSelected(true);
			return;
		}
		
		if (tiltPoint == null) {
			return;
		}
		
		float lastPoint[] = objects[rotationPointIndex].getPosition();

		lastPoint[0] *= scale[0];
		lastPoint[1] *= scale[1];
		lastPoint[2] *= scale[2];
		Vector3f endPoint = new Vector3f(lastPoint[0] - tiltPoint[0],
				lastPoint[1] - tiltPoint[1], lastPoint[2] - tiltPoint[2]);
		float angle = 0;

		if (end[1] * scale[1] > start[1]) {
			angle = 0.01f;
		} else {
			angle = -0.01f;
		}
		totalAngle += angle;
	  
		float cosangle = (float) Math.cos(angle);
		float sinangle = (float) Math.sin(angle);
		float a, b, c, d;

		a = upos[0];
		b = upos[1];
		c = upos[2];
		d = (float) Math.sqrt(b * b + c * c);
    
		float coverd = c / d;
		float boverd = b / d;
    
		float cbSquared = c * c + b * b;
		float daSquared = d * d + a * a;
    
		float s2ai5 = c * d / cbSquared;
		float s2ai6 = d * b / cbSquared;
		float s2ai9 = 10 * s2ai6;
		float s2bi0 = d / daSquared;
		float s2bi2 = a / daSquared;
		float s2bi8 = -s2bi2;
		float pos1_1, pos1_2, pos2_0, pos2_1, pos2_2, pos3_0, pos3_2, pos4_0, pos4_1, pos5_2;
    
		float pos[];
		GeometryArray ga[] = movingObject.getGeometryArrays();

		if (ga == null || ga.length == 0) {
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPosition();
				pos[0] *= scale[0];
				pos[1] *= scale[1];
				pos[2] *= scale[2];
				pos1_1 = pos[1] - tiltPoint[1];
				pos1_2 = pos[2] - tiltPoint[2];
				pos2_0 = pos[0] - tiltPoint[0];
				pos2_1 = pos1_1 * coverd - pos1_2 * boverd;
				pos2_2 = pos1_1 * boverd + pos1_2 * coverd;
				pos3_0 = pos2_0 * d - pos2_2 * a;
				pos3_2 = pos2_0 * a + pos2_2 * d;
				pos4_0 = pos3_0 * cosangle - pos2_1 * sinangle;
				pos4_1 = pos3_0 * sinangle + pos2_1 * cosangle;
				pos5_2 = pos4_0 * s2bi8 + pos3_2 * s2bi0;
				pos[0] = ((pos4_0 * s2bi0 + pos3_2 * s2bi2) + tiltPoint[0]) / scale[0];
				pos[1] = ((pos4_1 * s2ai5 + pos5_2 * s2ai6) + tiltPoint[1]) / scale[1];
				pos[2] = ((pos4_1 * s2ai9 + pos5_2 * s2ai5) + tiltPoint[2]) / scale[2];
				objects[i].setPosition(pos);
				objects[i].setColor(colorSpace3D.toSRGB(pos));
			}
		} else {
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				pos1_1 = pos[1] * scale[1] - tiltPoint[1];
				pos1_2 = pos[2] * scale[2] - tiltPoint[2];
				pos2_0 = pos[0] * scale[0] - tiltPoint[0];
				pos2_1 = pos1_1 * coverd - pos1_2 * boverd;
				pos2_2 = pos1_1 * boverd + pos1_2 * coverd;
				pos3_0 = pos2_0 * d - pos2_2 * a;
				pos3_2 = pos2_0 * a + pos2_2 * d;
				pos4_0 = pos3_0 * cosangle - pos2_1 * sinangle;
				pos4_1 = pos3_0 * sinangle + pos2_1 * cosangle;
				pos5_2 = pos4_0 * s2bi8 + pos3_2 * s2bi0;
				pos[0] = ((pos4_0 * s2bi0 + pos3_2 * s2bi2) + tiltPoint[0]) / scale[0];
				pos[1] = ((pos4_1 * s2ai5 + pos5_2 * s2ai6) + tiltPoint[1]) / scale[1];
				pos[2] = ((pos4_1 * s2ai9 + pos5_2 * s2ai5) + tiltPoint[2]) / scale[2];
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				for (int j = 0; j < vCount; j++) {
					pos1_1 = pos[j * 3 + 1] * scale[1] - tiltPoint[1];
					pos1_2 = pos[j * 3 + 2] * scale[2] - tiltPoint[2];
					pos2_0 = pos[j * 3] * scale[0] - tiltPoint[0];
					pos2_1 = pos1_1 * coverd - pos1_2 * boverd;
					pos2_2 = pos1_1 * boverd + pos1_2 * coverd;
					pos3_0 = pos2_0 * d - pos2_2 * a;
					pos3_2 = pos2_0 * a + pos2_2 * d;
					pos4_0 = pos3_0 * cosangle - pos2_1 * sinangle;
					pos4_1 = pos3_0 * sinangle + pos2_1 * cosangle;
					pos5_2 = pos4_0 * s2bi8 + pos3_2 * s2bi0;
					pos[j * 3] = ((pos4_0 * s2bi0 + pos3_2 * s2bi2) + tiltPoint[0]) / scale[0];
					pos[j * 3 + 1] = ((pos4_1 * s2ai5 + pos5_2 * s2ai6) + tiltPoint[1])
							/ scale[1];
					pos[j * 3 + 2] = ((pos4_1 * s2ai9 + pos5_2 * s2ai5) + tiltPoint[2])
							/ scale[2];
				}
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, colorSpace3D.toSRGB(pos));
			}
		}
	}
   
	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		float scale[] = outColor.getScale();
		boolean dirty[] = image2d.getDirtyColors();

		Arrays.fill(dirty, true);
		if (totalAngle == 0.0f) {
			return;
		}
    
		float cosangle = (float) Math.cos(totalAngle);
		float sinangle = (float) Math.sin(totalAngle);
		float a, b, c, d;
    
		float colors[] = mo.getColors();
		int count = colors.length / 3;

		a = upos[0];
		b = upos[1];
		c = upos[2];
		d = (float) Math.sqrt(b * b + c * c);
    
		float coverd = c / d;
		float boverd = b / d;
    
		float cbSquared = c * c + b * b;
		float daSquared = d * d + a * a;
    
		float s2ai5 = c * d / cbSquared;
		float s2ai6 = d * b / cbSquared;
		float s2ai9 = 10 * s2ai6;
		float s2bi0 = d / daSquared;
		float s2bi2 = a / daSquared;
		float s2bi8 = -s2bi2;
		float pos1_1, pos1_2, pos2_0, pos2_1, pos2_2, pos3_0, pos3_2, pos4_0, pos4_1, pos5_2;
    
		for (int i = 0; i < count; i++) {
			pos1_1 = colors[i * 3 + 1] * scale[1] - tiltPoint[1];
			pos1_2 = colors[i * 3 + 2] * scale[2] - tiltPoint[2];
			pos2_0 = colors[i * 3] * scale[0] - tiltPoint[0];
			pos2_1 = pos1_1 * coverd - pos1_2 * boverd;
			pos2_2 = pos1_1 * boverd + pos1_2 * coverd;
			pos3_0 = pos2_0 * d - pos2_2 * a;
			pos3_2 = pos2_0 * a + pos2_2 * d;
			pos4_0 = pos3_0 * cosangle - pos2_1 * sinangle;
			pos4_1 = pos3_0 * sinangle + pos2_1 * cosangle;
			pos5_2 = pos4_0 * s2bi8 + pos3_2 * s2bi0;
			colors[i * 3] = ((pos4_0 * s2bi0 + pos3_2 * s2bi2) + tiltPoint[0]) / scale[0];
			colors[i * 3 + 1] = ((pos4_1 * s2ai5 + pos5_2 * s2ai6) + tiltPoint[1])
					/ scale[1];
			colors[i * 3 + 2] = ((pos4_1 * s2ai9 + pos5_2 * s2ai5) + tiltPoint[2])
					/ scale[2];
		}
		totalAngle = 0.0f;
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
	}
	
	/**
	 * Creates the parameter panel for this object
	 */

	private void createParameterPanel() {
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new GridLayout(0, 1));
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
				"Settings"));
		
		ButtonGroup tmpGroup = new ButtonGroup();

		one = new JRadioButton("Select tilt point", true);
		two = new JRadioButton("Select rotation point", false);
		three = new JRadioButton("Rotate", false);

		one.setVisible(true);
		two.setVisible(true);
		three.setVisible(true);
		tmpGroup.add(one);
		tmpGroup.add(two);
		tmpGroup.add(three);

		parameterPanel.add(one);
		parameterPanel.add(two);
		parameterPanel.add(three);
	}
	
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */	

	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Returns that the current movement is supported
	 * @return boolean true
	 */

	public boolean isSupported(PickableObject object) {
		return true;
	}

	/**
	 * Returns the name of the class
	 * @return string name
	 */

	public String getName() {
		return "Tilt";
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 6;
	}
	
	/**
	 * Returns what type of transform this is, in this case it is an global and not local transform
	 * @return byte type of transform
	 */

	public byte getType() {
		return Interaction3D.GLOBAL;
	}

}
