/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * Global transform. Allows you to stretch the points in a scene, moving
 * all objects withing a certain distance from the selected point in the same
 * direction.
 */
public class GlobalTransformTypeA extends Interaction3D {
	
	/** The radius of the effect */
	private JSlider effectRadius;
 	
	/** The power of the effect */
	private JSlider effectPower;
 	
	/** The radius of the effect */
	private float radius;
 	
	/** The scaled radius */
	private float scaledRadius;
 	
	/** The power of the effect */
	private float power;
 	
	/** The parameter panel */
	private JPanel parameterPanel; 
 	
	/** The startin points */
	private Vector accumulatedStarts;
 	
	/** The movements performed */
	private Vector accumulatedMovements;
 	
	/**
	 * Constructs a default interaction of type A
	 */
	public GlobalTransformTypeA() {}
   
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public void update3D(int index, MovingObject movingObject, float start[], float end[]) {
		ColorSpace3D colorSpace3D = movingObject.getColorSpace();
		PickableObject objects[] = movingObject.getObjects();
		float scale[] = colorSpace3D.getScale();

		scaledRadius = radius
				* (float) Math.sqrt(scale[0] * scale[0] + scale[1] * scale[1]
						+ scale[2] * scale[2]);
		start = new float[] {start[0] * scale[0], start[1] * scale[1],
			start[2] * scale[2]};
		end = new float[] {end[0] * scale[0], end[1] * scale[1], end[2] * scale[2]};
		float movement[] = new float[3];

		for (int i = 0; i < 3; i++) {
			movement[i] = end[i] - start[i];
		}
		float pos[];
		float scaledPos[] = new float[3];
		float diff[] = new float[3];
		float factor;

		accumulatedStarts.add(start);
		accumulatedMovements.add(movement);
		GeometryArray ga[] = movingObject.getGeometryArrays();

		if (ga == null || ga.length == 0) {
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPosition();
				pos[0] *= scale[0];
				pos[1] *= scale[1];
				pos[2] *= scale[2];
				
				float distance = (float) Math.sqrt((pos[0] - start[0]) * (pos[0] - start[0])
						+ (pos[1] - start[1]) * (pos[1] - start[1])
						+ (pos[2] - start[2]) * (pos[2] - start[2]));

				if (distance < scaledRadius) {
					factor = (float) Math.pow((scaledRadius - distance) / scaledRadius, power);
					pos[0] += movement[0] * factor;
					pos[1] += movement[1] * factor;
					pos[2] += movement[2] * factor;
					pos[0] /= scale[0];
					pos[1] /= scale[1];
					pos[2] /= scale[2];
					objects[i].setPosition(pos);
					objects[i].setColor(colorSpace3D.toSRGB(pos));
				}
			}
		} else {
			for (int i = 0; i < objects.length; i++) {
				pos = objects[i].getPositionReference();
				for (int j = 0; j < 3; j++) {
					scaledPos[j] = pos[j] * scale[j];
					diff[j] = scaledPos[j] - start[j];
				}
				
				float distance = (float) Math.sqrt(diff[0] * diff[0] + diff[1] * diff[1]
						+ diff[2] * diff[2]);

				if (distance < scaledRadius) {
					factor = (float) Math.pow((scaledRadius - distance) / scaledRadius, power);
					pos[0] = (scaledPos[0] + movement[0] * factor) / scale[0];
					pos[1] = (scaledPos[1] + movement[1] * factor) / scale[1];
					pos[2] = (scaledPos[2] + movement[2] * factor) / scale[2];
				}
			}
			for (int i = 0; i < ga.length; i++) {
				int vCount = ga[i].getVertexCount();

				if (temporaryStorage != null && temporaryStorage.length == vCount * 3) {
					pos = temporaryStorage;
				} else {
					temporaryStorage = pos = new float[vCount * 3];
				}
				ga[i].getCoordinates(0, pos);
				for (int j = 0; j < vCount; j++) {
					scaledPos[0] = pos[j * 3] * scale[0];
					scaledPos[1] = pos[j * 3 + 1] * scale[1];
					scaledPos[2] = pos[j * 3 + 2] * scale[2];
					diff[0] = scaledPos[0] - start[0];
					diff[1] = scaledPos[1] - start[1];
					diff[2] = scaledPos[2] - start[2];
					
					float distance = (float) Math.sqrt(diff[0] * diff[0] + diff[1] * diff[1]
							+ diff[2] * diff[2]);

					if (distance < scaledRadius) {
						factor = (float) Math.pow((scaledRadius - distance) / scaledRadius, power);
						pos[j * 3] = (scaledPos[0] + movement[0] * factor) / scale[0];
						pos[j * 3 + 1] = (scaledPos[1] + movement[1] * factor) / scale[1];
						pos[j * 3 + 2] = (scaledPos[2] + movement[2] * factor) / scale[2];
					}
				}
				ga[i].setCoordinates(0, pos);
				ga[i].setColors(0, colorSpace3D.toSRGB(pos));
			}
		}
	}
   
	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		
		int actions = accumulatedStarts.size();

		if (actions == 0) {
			return;
		}
			
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		float scale[] = outColor.getScale();
		boolean dirty[] = image2d.getDirtyColors();
    
		float colors[] = mo.getColors();
		int count = colors.length / 3;
		float pos[] = new float[3];
		float movement[];
		float factor;

		for (int j = 0; j < actions; j++) {
			start = (float[]) accumulatedStarts.elementAt(j);
			movement = (float[]) accumulatedMovements.elementAt(j);
			for (int i = 0; i < count; i++) {
				pos[0] = colors[i * 3] * scale[0];
				pos[1] = colors[i * 3 + 1] * scale[1];
				pos[2] = colors[i * 3 + 2] * scale[2];
    		
				float distance = (float) Math.sqrt((pos[0] - start[0]) * (pos[0] - start[0])
						+ (pos[1] - start[1]) * (pos[1] - start[1])
						+ (pos[2] - start[2]) * (pos[2] - start[2]));

				if (distance < scaledRadius) {
					dirty[i] = true;
					factor = (float) Math.pow((scaledRadius - distance) / scaledRadius, power);
					pos[0] += movement[0] * factor;
					pos[1] += movement[1] * factor;
					pos[2] += movement[2] * factor;
					colors[i * 3] = pos[0] / scale[0];
					colors[i * 3 + 1] = pos[1] / scale[1];
					colors[i * 3 + 2] = pos[2] / scale[2];
				}
			}
		}
		accumulatedStarts.clear();
		accumulatedMovements.clear();
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
		
	}
	
	/**
	 * Creates the parameter panel for this object
	 */
	private void createParameterPanel() {
		radius = 0.3f;
		power = 1.0f;
		effectRadius = new JSlider(JSlider.HORIZONTAL, 0, 1000, 300);
		effectRadius.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				radius = effectRadius.getValue() / 1000.0f;
			}
		});
		effectPower = new JSlider(JSlider.HORIZONTAL, 0, 500, 400);
		effectPower.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				power = 5.0f - effectPower.getValue() / 100.0f;
			}
		});
		accumulatedStarts = new Vector();
		accumulatedMovements = new Vector();
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new GridLayout(0, 1));
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
				"Settings"));
		parameterPanel.add(new JLabel("Effect radius"));
		parameterPanel.add(effectRadius);
		parameterPanel.add(new JLabel("Effect power"));
		parameterPanel.add(effectPower);
	}
	
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */	
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Returns that the current movement is supported
	 * @return boolean true
	 */
	public boolean isSupported(PickableObject object) {
		return true;
	}

	/**
	 * Returns the name of the class
	 * @return string name
	 */
	public String getName() {
		return "Stretch";
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 9;
	}
	
	/**
	 * Returns what type of transform this is, in this case it is an global and not local transform
	 * @return byte type of transform
	 */
	public byte getType() {
		return Interaction3D.GLOBAL;
	}

}
