/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.awt.*;
import javax.swing.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * Interpolates all the pixels within a segment. Their new values is calculated from how far they are from the center of the segment.
 */  
public class SegmentInteraction3D extends SimpleInteraction3D {

	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();
	
		PickableSegmentObject segObj[] = null;

		try {
			segObj = (PickableSegmentObject[]) objects;
		} catch (Exception ClassCastException) {
			return;
		}
		
		boolean dirty[] = image2d.getDirtyColors();
		int colorIndex[] = objects[index].getColorIndices();
		float center[] = segObj[index].getCenter();
		float colors[] = mo.getColors();
		float movement[] = new float[3];

		for (int i = 0; i < 3; i++) {
			movement[i] = end[i] - start[i];
		}
		
		float objectRadius = (float) Math.sqrt((start[0] - center[0])
						* (start[0] - center[0])
				+ (start[1] - center[1]) * (start[1] - center[1])
				+ (start[2] - center[2]) * (start[2] - center[2]));

		if (objectRadius == 0.0f) {
			return;
		}
		int col, inArray;
		float colorRadius;

		for (int i = 0; i < colorIndex.length; i++) {
			col = colorIndex[i];
			inArray = col * 3;
			dirty[col] = true;
			colorRadius = (float) Math.sqrt((colors[inArray] - center[0])
							* (colors[inArray] - center[0])
					+ (colors[inArray + 1] - center[1]) * (colors[inArray + 1] - center[1])
					+ (colors[inArray + 2] - center[2]) * (colors[inArray + 2] - center[2]));
			colors[inArray] += movement[0] * colorRadius / objectRadius;
			colors[inArray + 1] += movement[1] * colorRadius / objectRadius;
			colors[inArray + 2] += movement[2] * colorRadius / objectRadius;
		}
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
	}
	
	/**
	 * Returns wherether this interaction is supported within the current visualization
	 * @return boolean true if supported, false if the test fails
	 */

	public boolean isSupported(PickableObject object) {
		try {
			PickableSegmentObject segObj = (PickableSegmentObject) object;

			return true;
		} catch (Exception ClassCastException) {
			return false;
		}
	}
	
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */	

	public Component getParameterPanel() {
		return new JPanel();
	}
	
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public void update3D(int index, MovingObject movingObject, float start[], float end[]) {
		try {
			PickableSegmentObject segObj[] = (PickableSegmentObject[]) movingObject.getObjects();

			super.update3D(index, movingObject, start, end);
		} catch (Exception ClassCastException) {
			SystemManager.displayError(
				"Error",
				"Visualization does not support segment interaction.");
		}
	}
	
	/**
	 * Returns the name of the class
	 * @return string name
	 */
	public String getName() {
		return "Segment";
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 2;
	}
	
	/**
	 * Returns what type of interaction this is, in this case it is an local and not global interaction
	 * @return byte type of interaction
	 */
	public byte getType() {
		return Interaction3D.LOCAL;
	}

}
