/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions;


import java.awt.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;


/**
 * Interpolates all the pixels within a tetrahedron. Their new values are calculated 
 * from their part of the volume of the tetrahedron.
 */ 
public class TetrahedronInterpolationInteraction3D extends SimpleInteraction3D {
		
	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo) {
		
		ImageCanvas2D image2d = mo.getImage2D().getImageCanvas2D();
		ColorSpace3D outColor = mo.getColorSpace();
		PickableObject objects[] = mo.getObjects();
	
		PickableTetrahedronObject objs[] = null;

		try {
			objs = (PickableTetrahedronObject[]) objects;
		} catch (Exception e) {
			return;
		}
		
		boolean dirty[] = image2d.getDirtyColors();
		float colors[] = mo.getColors();
		Tetrahedron tetra[] = objs[index].getTetrahedra();

		if (tetra == null || tetra.length == 0) {
			SystemManager.displayError(
				"Error",
				"Object does not belong to any tetrahedron.\nAre you sure the visualization parameters are correct?");
		}
		int curColor, inArray;

		for (int i = 0; i < tetra.length; i++) {
			if (tetra[i] == null) {
				break;
			}
			float positions[] = tetra[i].getPositions();
			float weights[][] = tetra[i].getWeights();
			int colorIndices[] = tetra[i].getColorIndices();
			float corners[][] = tetra[i].getCorners();
			float startCorners[][] = tetra[i].getStartCorners();

			System.out.println("Colors " + colorIndices.length);
			for (int j = 0; j < colorIndices.length; j++) {
				curColor = colorIndices[j];
				inArray = curColor * 3;
				dirty[curColor] = true;
				colors[inArray] = positions[j * 3]
						+ weights[j][0] * (corners[0][0] - startCorners[0][0])
						+ weights[j][1] * (corners[1][0] - startCorners[1][0])
						+ weights[j][2] * (corners[2][0] - startCorners[2][0])
						+ weights[j][3] * (corners[3][0] - startCorners[3][0]);
				colors[inArray + 1] = positions[j * 3 + 1]
						+ weights[j][0] * (corners[0][1] - startCorners[0][1])
						+ weights[j][1] * (corners[1][1] - startCorners[1][1])
						+ weights[j][2] * (corners[2][1] - startCorners[2][1])
						+ weights[j][3] * (corners[3][1] - startCorners[3][1]);
				colors[inArray + 2] = positions[j * 3 + 2]
						+ weights[j][0] * (corners[0][2] - startCorners[0][2])
						+ weights[j][1] * (corners[1][2] - startCorners[1][2])
						+ weights[j][2] * (corners[2][2] - startCorners[2][2])
						+ weights[j][3] * (corners[3][2] - startCorners[3][2]);
					
			}
		}
		outColor.toCIEXYZ(colors, image2d.getXYZ_Colors(), dirty);
		image2d.imageUpdate();
	}
	
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public void update3D(int index, MovingObject movingObject, float start[], float end[]) {
		PickableTetrahedronObject objs[] = null;

		try {
			objs = (PickableTetrahedronObject[]) movingObject.getObjects();
			super.update3D(index, movingObject, start, end);
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Visualization does not support tetrahedron interpolation.");
			return;
		}
	}
	
	/**
	 * Returns wherether this interaction is supported within the current visualization
	 * @return boolean true if supported, false if the test fails
	 */
	public boolean isSupported(PickableObject object) {
		try {
			PickableTetrahedronObject objs = (PickableTetrahedronObject) object;
		} catch (Exception e) {
			return false;
		}
		return true;
	}
	
	/**
	 * Returns the parameter panel for this object.
	 * @return Component parameter panel
	 */	
	public Component getParameterPanel() {
		return new JPanel();
	}
	
	/**
	 * Returns the name of the class
	 * @return string name
	 */
	public String getName() {
		return "Tetrahedron interpolation";
	}
	
	/**
	 * The priority of the interaction.
	 * @return An int representing the priority of the interaction (lower => higher priority).
	 */
	public int getPriority() {
		return 3;
	}
	
	/**
	 * Returns what type of interaction this is, in this case it is an local 
	 * and not global interaction
	 * @return byte type of interaction
	 */
	public byte getType() {
		return Interaction3D.LOCAL;
	}

}
