/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.interactions.generic;


import java.awt.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;


/**
 * An abstract base class. Defines the functions
 * that allows a type of interaction to move the
 *	objects in the scene.
 */
public abstract class Interaction3D extends Plugin {
	
	/**
	 * Temporary storage available tp prevent constant reallocation of
	 * objects with a hort life span.
	 */
	protected static float temporaryStorage[];
	
	/**
	 * LOCAL identifies the Interaction3D as a local interaction, meaning
	 *	that only one object part is moved at a time.
	 */
	public static final byte LOCAL = 0;
	
	/**
	 * GLOBAL identifies the Interaction3D as a global interaction, meaning
	 *	that several object parts are moved simultaneously.
	 */
	public static final byte GLOBAL = 1;
	
	/**
	 * Updates the 3D view when you move one of the objects
	 * @param index The index of the object you have moved
	 * @param movingObject The object that ois being manipulated.
	 * @param start Start point of the object you move, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis		
	 * @param end End point of the object you move, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis
	 */
	public abstract void update3D(int index, MovingObject movingObject, float start[], float end[]);
	
	/**
	 * Updates the 2D view after you have released the mouse button, changes the pixels color values to their new values.
	 * @param index The index of the point you have moved.
	 * @param start Start point of the object you moved, start[0] is the location on the x-axis, start[1] y-axis and start[2] z-axis.
	 * @param end End point of the object you moved, end[0] is the location on the x-axis, end[1] y-axis and end[2] z-axis.
	 * @param oldColor Byte array representing the old color value for the object you moved.
	 * @param newColor Byte array representing the new color value for the object you moved.
	 * @param mo Reference to the object that is being manipulated.
	 */
	public abstract void update2D(int index, float start[], float end[], byte oldColor[], byte newColor[], MovingObject mo);
	
	/**
	 * Returns the parameter panel for this object
	 * @return Component the current parameter panel
	 */
	public abstract Component getParameterPanel();
	
	/**
	 * Returns whether the current interaction is supported by the scene graph built by the visualization.
	 * @return boolean true
	 */
	public abstract boolean isSupported(PickableObject object);
	
	/**
	 * Returns the description of the interaction.
	 * @return The description.
	 */
	public abstract String getName();
	
	/**
	 * The priority of the interaction.
	 * @return An integer representing the priority of the interaction (lower = higher priority).
	 */
	public abstract int getPriority();
	
	/**
	 * Returns the type of the interaction
	 * @return Interaction3D.GLOBAL or Interaction3D.LOCAL
	 */
	public abstract byte getType();
}
