/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.managers;


import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.beans.*;
import javax.media.j3d.*;


/**
 * This abstract class contains methods used for dynamically loading plugins
 *
 */
public abstract class Manager {
	
	/**
	 * Loads plugins dynamically from a directory.
	 * 
	 * @param output Status text is displayed in this JTextArea while loading the plugins.
	 * @param pluginClass The abstract base class extended by the plugins.
	 * @param pluginDirectory The (sub!)directory where the plugins are located.
	 * @param pluginType A string describing the plugin type
	 */
	public static void load(JTextArea output, Class pluginClass, String pluginDirectory, String pluginType, Manageable manageable) {
		if (pluginClass == null) {
			throw new RuntimeException("No plugin class specified. Unable to load plugins");
		}
		if (pluginDirectory == null) {
			throw new RuntimeException("No plugin directory specified. Unable to load plugins");
		}
		if (pluginType == null) {
			throw new RuntimeException("No plugin type specified. Unable to load plugins");
		}
		if (output == null) {
			output = new JTextArea();
		}
		Vector names = manageable.getNames();
		Vector plugins = manageable.getPlugins();
		int n = 0;

		try {
			File file = new File(pluginDirectory);
			String classes[] = file.list();

			append(output, "\nLoading " + pluginType + ": ");
			for (int i = 0; i < classes.length; i++) {
				if (classes[i].endsWith(".class") && classes[i].indexOf("$") == -1) {
					try {
						Plugin p = null;
						String className = pluginDirectory.replace(File.separatorChar, '.') + "."
								+ classes[i].substring(0, classes[i].length() - 6);
						Class pClass = Class.forName(className);

						p = (Plugin) pClass.newInstance();
						if (pluginClass.isInstance(p)) {
							plugins.add(p);
							names.add(p.getName());
							append(output, "+");
							n++;
						}
					} catch (Exception e) {
						append(output, "!");
					}
				}
			}
			append(output, "(" + Integer.toString(n) + ")");
		} catch (Exception e) {}
		sort(manageable);
	}
	
	public static void append(JTextArea output, String app) {
		output.append(app);
		output.setCaretPosition(output.getText().length());
		output.paintImmediately(output.getBounds(null));
	}
	
	/**
	 * Sorts the loaded plugins by their priority in ascending order.
	 */
	public static void sort(Manageable manageable) {
		Vector names = manageable.getNames();
		Vector plugins = manageable.getPlugins();

		if (names != null && !names.isEmpty()) {
			int len = names.size();
			Vector newNames = new Vector(len);
			Vector newPlugins = new Vector(len);
			
			while (!names.isEmpty()) {
				Plugin p = (Plugin) plugins.elementAt(0);
				String s = (String) names.elementAt(0);

				for (int i = 0; i < names.size(); i++) {
					Plugin pCurrent = (Plugin) plugins.elementAt(i);

					if (pCurrent.getPriority() < p.getPriority()) {
						p = pCurrent;
						s = (String) names.elementAt(i);
					}
				}
				names.remove(s);
				plugins.remove(p);
				newNames.add(s);
				newPlugins.add(p);
			}
			manageable.setNames(newNames);
			manageable.setPlugins(newPlugins);
		}
	}
		
	/**
	 * Retrieves the requested plugin object from this ColorSpaceManager.
	 *
	 * @param name The name of the plugin. Case-sensitive.
	 *	@return The requested plugin, or null if not found
	 */
	public static Plugin getPlugin(String name, Manageable manageable) {
		Plugin ret = null;
		Vector plugins = manageable.getPlugins();

		for (Enumeration e = plugins.elements(); e.hasMoreElements();) {
			Plugin p = (Plugin) e.nextElement();

			if (p.getName().equals(name)) {
				try {
					ret = (Plugin) p.getClass().newInstance();
				} catch (Exception ee) {
					SystemManager.displayError(
						"Error",
						"Unable to instantiate a plugin object of type " + p.getClass().getName());
					ret = null;
				}
			}
		}
		return ret;
	}
	
	/**
	 * Retrieves a list of names of the currently registered plugins.
	 *
	 * @return An array of names, or null if no plugins are registered.
	 */
	public static String[] getNames(Manageable manageable) {
		Vector names = manageable.getNames();
		String retNames[] = null;

		if (names != null && !names.isEmpty()) {
			retNames = new String[names.size()];
			for (int i = 0; i < names.size(); i++) {
				retNames[i] = (String) names.elementAt(i);
			}
		}
		return retNames;
	}
}
