/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.managers;


import java.io.File;
import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import icc3dpackage.view3d.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;


public class ModuleManager extends Manager {
	
	/** All available modules */
	private static Manageable availableModules;
	
	/** The modules loaded in this instance of the ModuleManager */
	private Vector modules;
	
	/** The View3D parent of this ModuleManager */
	private View3D view3D;
	
	/** The module manager parameter panel */
	private JPanel parameterPanel;
	
	/** The modules that can be added */
	private JMenuItem addMenuItem[];
	
	/** The menu which displays a list of available modules */
	private JPopupMenu addMenu;
	
	/**
	 * Creates a ModuleManager for the given View3D.
	 * @param view3D The View3D parent.
	 */
	public ModuleManager(View3D view3D) {
		this.view3D = view3D;
		modules = new Vector();
		String def[] = getDefaultModules();

		for (int i = 0; i < def.length; i++) {
			Module mod = getModule(def[i]);

			if (mod.initialize(view3D)) {
				modules.add(mod);
			}
		}
	}

	/**
	 * Load modules dynamically from the default directory: 
	 * icc3dpackage/modules.
	 * 
	 * @param output Status text is displayed in this JTextArea while loading the modules.
	 */
	public static void load(JTextArea output) {
		availableModules = new Manageable();
		Class c = null;

		try {
			c = Class.forName("icc3dpackage.modules.generic.Module");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the Module class");
			System.exit(1);
		}
		load(output, c, "icc3dpackage" + File.separator + "modules", "modules",
				availableModules);
	}
	
	/**
	 * Retrieves the requested module object.
	 *
	 * @param name of module
	 * @return requested module, or null if not found
	 */
	public static Module getModule(String name) {
		return (Module) getPlugin(name, availableModules);
	}
	
	/**
	 * Retrieves a list of names of the currently registered modules.
	 *
	 * @return An array of names.
	 */
	public static String[] getNames() {
		return getNames(availableModules);
	}
	
	/**
	 * Retrieves a list of modules that are present in every View3D by default.
	 * @return An array of names.
	 */
	public static String[] getDefaultModules() {
		Vector defMods = new Vector();

		for (int i = 0; i < availableModules.getNames().size(); i++) {
			Module m = (Module) availableModules.getPlugins().elementAt(i);

			if (m.existsAsDefault()) {
				defMods.add(availableModules.getNames().elementAt(i));
			}
		}
		String ret[] = new String[defMods.size()];

		for (int i = 0; i < defMods.size(); i++) {
			ret[i] = (String) defMods.elementAt(i);
		}
		return ret;
	}
	
	/**
	 * Retrieves a list of modules that the user can add to a View3D.
	 * @return An array of names.
	 */
	public static String[] getAddableModules() {
		Vector addMods = new Vector();

		for (int i = 0; i < availableModules.getNames().size(); i++) {
			Module m = (Module) availableModules.getPlugins().elementAt(i);

			if (m.userAdd()) {
				addMods.add(availableModules.getNames().elementAt(i));
			}
		}
		
		String ret[] = new String[addMods.size()];

		for (int i = 0; i < addMods.size(); i++) {
			ret[i] = (String) addMods.elementAt(i);
		}
		return ret;
	}
	
	/**
	 * Adds a Module to this manager and View3D.
	 * @param module The module.
	 */
	public void add(Module module) {
		if (module.initialize(view3D)) {
			modules.add(module);
			view3D.addModule(module);
		}
	}
	
	/**
	 * Removes a Module from this manager and View3D.
	 * @param module The module.
	 */
	public void remove(Module module) {
		modules.remove(module);
	}
	
	/**
	 * Removes the selected Module from this manager and View3D.
	 */
	public void removeModule() {
		int i = view3D.getTabs().getSelectedIndex();

		if (i < 0 || i >= modules.size()) {
			return;
		}
		Module mod = (Module) modules.get(i);

		if (mod == null) {
			return;
		}
		if (!mod.userRemove()) {
			SystemManager.displayError(
				"Error",
				"Selected module cannot be removed");
			return;
		}
		mod.deInitialize();
		modules.remove(i);
		view3D.getTabs().remove(i);
	}
	
	/**
	 * Retrieves a list of modules that have been added to this manager.
	 * @return An array of modules.
	 */
	public Module[] getModules() {
		Module mods[] = new Module[modules.size()];

		for (int i = 0; i < modules.size(); i++) {
			mods[i] = (Module) modules.elementAt(i);
		}
		return mods;
	}
	
	/**
	 * Returns whether the module has been added to this manager,
	 * @param mod The module.
	 * @return True if this manager contains the module, otherwise false.
	 */
	public boolean exists(Module mod) {
		boolean e = false;

		for (int i = 0; i < modules.size(); i++) {
			if (mod == (Module) modules.elementAt(i)) {
				e = true;
			}
		}
		return e;
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new GridLayout(1, 0));
		
		String s[] = getAddableModules();

		addMenuItem = new JMenuItem[s.length];
		for (int i = 0; i < s.length; i++) {
			addMenuItem[i] = new JMenuItem(s[i], getModule(s[i]).getIcon());
			addMenuItem[i].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					Module mod = getModule(((JMenuItem) ae.getSource()).getText());

					add(mod);
				}
			});
		}
		addMenu = new JPopupMenu("Add modules");
		for (int i = 0; i < addMenuItem.length; i++) {
			addMenu.add(addMenuItem[i]);
		}
		
		final JButton addButton = new JButton("Add module");

		addButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				addMenu.show(addButton, addButton.getX(), addButton.getY());
			}
		});
		JButton remButton = new JButton("Remove module");

		remButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				removeModule();
			}
		});
		parameterPanel.add(addButton);
		parameterPanel.add(remButton);
	}
	
	/**
	 * Deinitializes this ModuleManager and all associated modules.
	 */
	public void deInitialize() {
		Module mod;

		for (int i = 0; i < modules.size(); i++) {
			mod = (Module) modules.elementAt(i);
			mod.deInitialize();
		}
		this.addMenu = null;
		this.addMenuItem = null;
		this.modules = null;
		this.parameterPanel = null;
		this.view3D = null;
	}
}
