/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.managers;


import java.text.DecimalFormat;
import javax.swing.*;

import icc3dpackage.*;


/**
 * This class is used as a container for useful system functions.
 */
public abstract class SystemManager {
	
	/** The start time when a time recording is in progress */
	static private long time1 = 0;
	
	/** The stop time when a time recording is in progress */
	static private long time2 = 0;
	
	/**
	 * Performs garbage collection.
	 *
	 * @see #majorClean
	 */
	public static void simpleClean() {
		System.gc();
	}
	
	/**
	 * Performes garbage collection, and yields the current thread.
	 * @see #simpleClean
	 */
	public static void majorClean() {
		System.gc();
		Thread.currentThread().yield();
	}
	 
	/**
	 * Calculates total memory usage for this Java VM.
	 * (total memory available)-(free memory)
	 *
	 * @return The amount of memory used by this application
	 */
	public static long getUsedMemoryInBytes() {
		return Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();	
	}
	
	/**
	 * Calculates total memory usage for this Java VM.
	 * (total memory available)-(free memory)
	 *
	 * @return The amount of memory used by this application
	 */
	public static String getUsedMemory() {
		DecimalFormat df = new DecimalFormat("0.00");

		return df.format((getUsedMemoryInBytes() / (double) (1024 * 1024))) + " MB";
	}
	
	/**
	 * Retrieves the amount of free memory in this Java VM.
	 *
	 * @return The amount of memory that has been allocated by the VM, but not
	 * yet used by the application.
	 */
	public static long getFreeMemory() {
		return Runtime.getRuntime().freeMemory();
	}
	
	/**
	 * Retrieves the total amount of memory available to this Java VM.
	 *
	 * @return The amount of memory allocated by the Java VM.
	 */
	public static long getTotalMemory() {
		return Runtime.getRuntime().totalMemory();
	}
	
	/**
	 * Stores current system time as start. Used in coherence with stopRecordingTime().
	 *
	 * @see #stopRecordingTime
	 * @see #getTimeDifference
	 *	@see #resetTimeValues
	 */
	public static void startRecordingTime() {
		time1 = System.currentTimeMillis();	
	}
	
	/**
	 * Stores current system time as end. Used in coherence with startRecordingTime().
	 *
	 * @see #startRecordingTime
	 * @see #getTimeDifference
	 *	@see #resetTimeValues
	 */
	public static void stopRecordingTime() {
		time2 = System.currentTimeMillis();
	}
	
	/**
	 * Resets timevalues for recording. New value is 0.
	 *
	 * @see #stopRecordingTime
	 * @see #getTimeDifference
	 */	
	public static void resetTimeValues() {
		time1 = time2 = 0;	
	}
	
	/**
	 * Calculates the timedifference between time registered from startRecordingTime() 
	 * and stopRecordingTime().
	 * Timedifference is: stopRecordingTime()-startRecordingTime().
	 *
	 *	@return timedifference
	 * @see #stopRecordingTime
	 * @see #getTimeDifference
	 *	@see #resetTimeValues
	 */
	public static long getTimeDifference() {
		if ((time2 - time1) < 0) {
			resetTimeValues();
		}
		return time2 - time1;
	}
	
	/**
	 * Causes the currently executing thread to sleep (temporarily cease execution) for 
	 * the specified number of milliseconds.
	 *
	 * @param ms time to sleep in milliseconds
	 */
	public static void sleep(long ms) {
		try {
			Thread.currentThread().sleep(ms);
		} catch (InterruptedException ie) {}
	}
	
	/**
	 * Causes the currently executing thread object to temporarily pause and allow other 
	 * threads to execute.
	 */
	public static void yield() {
		Thread.currentThread().yield();
	}
	
	/**
	 * Display an error message.
	 * @param title The title of the message box.
	 * @param message The error message.
	 */
	public static void displayError(String title, String message) {
		JOptionPane.showMessageDialog(
			ICC3D.getDialogParent(),
			message,
			title,
			JOptionPane.ERROR_MESSAGE,
			IconManager.getIcon("exclamation"));
	}
	
	/**
	 * Displays a confirmation dialog box.
	 * @param title The title of the dialog box.
	 * @param question The question.
	 * @return One of the JOptionPane return values.
	 */
	public static int confirmYesNo(String title, String question) {
		return JOptionPane.showConfirmDialog(
			ICC3D.getDialogParent(),
			question,
			title,
			JOptionPane.YES_NO_OPTION,
			JOptionPane.INFORMATION_MESSAGE,
			IconManager.getIcon("question"));
	}
}
