/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import icc3dpackage.colorspaces.generic.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.beans.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.view3d.*;


/**
 * This class implements methods for changing the 3D view.
 */
public class CameraModule extends Module {
	
	/** The parameter panel that presents the options to the user. */
	private JPanel parameterPanel;
	
	/** Choose between perspective and ortographic view. */
	private JComboBox viewChooser;
	
	/** Choose the background color. */
	private JButton colorChooser;
	
	/** The View3D which can be modified by this module. */
	private View3D view3D;
	
	/**
	 * Constructs a CameraModule.
	 */
	public CameraModule() {}
	
	/**
	 * Creates the CameraModule parameter panel
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new BorderLayout());
		
		String choices[] = { "Perspective", "Ortographic" };

		viewChooser = new JComboBox(choices);
		viewChooser.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				String type = (String) viewChooser.getSelectedItem();

				view3D.changeView(type);
			}
		});
		colorChooser = new JButton("Background color");
		colorChooser.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Color3f color3f = new Color3f();

				view3D.background().getColor(color3f);
				float floatcol[] = new float[3];

				color3f.get(floatcol);
				Color color = new Color(floatcol[0], floatcol[1], floatcol[2]);
				Color retColor = JColorChooser.showDialog(
					ICC3D.getDialogParent(),
					"Choose a background color",
					color);

				if (retColor != null) {
					colorChooser.setBackground(retColor);
					view3D.background().setColor(new Color3f(retColor.getColorComponents(null)));
				}
			}
		});
		
		JButton viewReset_a = new JButton("Reset view to +a",
				IconManager.getIcon("+a"));

		viewReset_a.setHorizontalAlignment(JButton.LEFT);
		viewReset_a.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				view3D.animate(new Transform3D(), (String) viewChooser.getSelectedItem());
			}
		});
			
		JButton viewReset_b = new JButton("Reset view to +b",
				IconManager.getIcon("+b"));

		viewReset_b.setHorizontalAlignment(JButton.LEFT);
		viewReset_b.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Transform3D t3d = new Transform3D();

				t3d.rotY(-Math.PI / 2.0);
				view3D.animate(t3d, (String) viewChooser.getSelectedItem());
			}
		});
		
		JButton viewReset_l = new JButton("Reset view to +L",
				IconManager.getIcon("+l"));

		viewReset_l.setHorizontalAlignment(JButton.LEFT);
		viewReset_l.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Transform3D t3d = new Transform3D(), t3d2 = new Transform3D();

				t3d.rotX(Math.PI / 2.0);
				t3d2.rotY(-(3.0 / 2.0) * Math.PI);
				t3d.mul(t3d2);
				view3D.animate(t3d, (String) viewChooser.getSelectedItem());
			}
		});

		JButton viewReset_aa = new JButton("Reset view to -a",
				IconManager.getIcon("-a"));

		viewReset_aa.setHorizontalAlignment(JButton.LEFT);
		viewReset_aa.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Transform3D t3d = new Transform3D();

				t3d.rotY(Math.PI);
				view3D.animate(t3d, (String) viewChooser.getSelectedItem());
			}
		});
			
		JButton viewReset_bb = new JButton("Reset view to -b",
				IconManager.getIcon("-b"));

		viewReset_bb.setHorizontalAlignment(JButton.LEFT);
		viewReset_bb.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Transform3D t3d = new Transform3D();

				t3d.rotY(Math.PI / 2.0);
				view3D.animate(t3d, (String) viewChooser.getSelectedItem());
			}
		});
		
		JButton viewReset_ll = new JButton("Reset view to -L",
				IconManager.getIcon("-l"));

		viewReset_ll.setHorizontalAlignment(JButton.LEFT);
		viewReset_ll.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Transform3D t3d = new Transform3D(), t3d2 = new Transform3D();

				t3d.rotX(-Math.PI / 2.0);
				t3d2.rotY(Math.PI / 2.0);
				t3d.mul(t3d2);
				view3D.animate(t3d, (String) viewChooser.getSelectedItem());
			}
		});
		
		JPanel gridPanel = new JPanel(new GridLayout(0, 1));

		gridPanel.add(viewChooser);
		gridPanel.add(viewReset_ll);
		gridPanel.add(viewReset_l);
		gridPanel.add(viewReset_aa);
		gridPanel.add(viewReset_a);
		gridPanel.add(viewReset_bb);
		gridPanel.add(viewReset_b);
		gridPanel.add(colorChooser);
		
		JPanel flowPanel = new JPanel(new FlowLayout());

		flowPanel.add(gridPanel);
		parameterPanel.add(flowPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 1;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Camera";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("camera");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.colorChooser = null;
		this.parameterPanel = null;
		this.view3D = null;
		this.viewChooser = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		return null;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
