/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import java.io.*;
import javax.media.j3d.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.media.Format;
import javax.media.protocol.*;
import javax.media.jai.*;

import icc3dpackage.*;
import icc3dpackage.image2d.FileWriter;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.modules.capture.*;
import icc3dpackage.view3d.*;


/**
 * Captures images or movies from a 3D view.
 */
public class CaptureModule extends Module {
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** Starts a video capture. */
	private JButton capture;
	
	/** Captures a screen shot. */
	private JButton shoot;
	
	/** The movie maker that encodes the video stream. */
	private MovieMaker movieMaker;
	
	/** The frame that contains the video capture GUI. */
	private JFrame movieFrame;
	
	/** The ImageComponent2D used as an off screen buffer. */
	private ImageComponent2D offscreenBuffer;
	
	/** Frame rate selector. */
	private JComboBox frameRCBox;
	
	/** The script text area. */
	private JTextArea scriptArea;
	
	/** Video format selector. */
	private JComboBox videoFormat;
	
	/** Video codec selector. */
	private JComboBox videoCodec;
	
	/** Starts a video capture. */
	private JButton beginCapture;
	
	/** The Image3D from which the screen capture was initiated. */
	private View3D view3D;
	
	/**
	 * Constructs a CaptureModule.
	 */
	public CaptureModule() {}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new FlowLayout());
		
		JPanel buttonPanel = new JPanel(new GridLayout(0, 1));

		capture = new JButton("Video");
		capture.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (movieFrame == null) {
					Canvas3D canvas = view3D.getCanvas();

					movieMaker.prepare(((int) (canvas.getWidth() / 4)) * 4,
							((int) (canvas.getHeight() / 4)) * 4, 24.0f);
					createMovieFrame();
				}
				movieFrame.setVisible(true);
			}
		});
		shoot = new JButton("Screenshot");
		shoot.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				System.out.println(shoot.getWidth());
				JFileChooser saveChooser = new JFileChooser(".");
				String filename = null;

				if (saveChooser.showSaveDialog(ICC3D.getDialogParent()) != JFileChooser.CANCEL_OPTION) {
					filename = saveChooser.getSelectedFile().getAbsolutePath();
				}
				if (filename != null) {
					Canvas3D canvas = view3D.getCanvas();
					Canvas3D offscreenCanvas = view3D.getOffScreenCanvas();
					byte tmp[] = new byte[canvas.getWidth() * canvas.getHeight() * 3];
					DataBufferByte dataBuffer = new DataBufferByte(tmp, tmp.length);
					SampleModel sampleModel = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE,
							canvas.getWidth(), canvas.getHeight(), 3, canvas.getWidth() * 3,
							new int[] {0, 1, 2});
					ColorModel colorModel = PlanarImage.createColorModel(sampleModel);

					java.awt.image.Raster raster = RasterFactory.createWritableRaster(sampleModel,
							dataBuffer, new Point(0, 0));
					TiledImage tiledImage = new TiledImage(0, 0, canvas.getWidth(),
							canvas.getHeight(), 0, 0, sampleModel, colorModel);

					tiledImage.setData(raster);
					offscreenBuffer = new ImageComponent2D(ImageComponent2D.FORMAT_RGB8,
							tiledImage);
					offscreenBuffer.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
					offscreenBuffer.setCapability(ImageComponent2D.ALLOW_FORMAT_READ);
					offscreenBuffer.setCapability(ImageComponent2D.ALLOW_SIZE_READ);
					offscreenCanvas.setOffScreenBuffer(offscreenBuffer);
					offscreenCanvas.renderOffScreenBuffer();
					offscreenCanvas.waitForOffScreenRendering();
					RenderedImage ri = offscreenBuffer.getRenderedImage();

					tiledImage.setData(ri.getData());
					FileWriter.writeImage(tiledImage, filename);
				}
			}
		});
		
		JLabel captureLabel = new JLabel("Capture");

		captureLabel.setHorizontalAlignment(JLabel.CENTER);
		captureLabel.setBorder(BorderFactory.createEtchedBorder());
		captureLabel.setPreferredSize(new Dimension(150, 25));
		buttonPanel.add(captureLabel);
		buttonPanel.add(shoot);
		buttonPanel.add(capture);
		parameterPanel.add(buttonPanel);	
	}
	
	/**
	 * Creates the movie script frame, includes the file chooser, script area and launch button.
	 */
	private void createMovieFrame() {
		movieFrame = new JFrame();
		movieFrame.setIconImage(ICC3D.image);

		Container c = movieFrame.getContentPane();

		c.setLayout(new BorderLayout());
		
		String rates[] = {"3", "5", "7", "10", "13", "16", "20", "24", "30"};
		
		JLabel frameRLabel = new JLabel("Framerate :") {
			public Dimension getPreferredSize() {
				return(new Dimension(90, 25));
			}
		};

		frameRCBox = new JComboBox(rates);
		frameRCBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				Canvas3D canvas = view3D.getCanvas();
				Canvas3D offscreenCanvas = view3D.getOffScreenCanvas();

				movieMaker.prepare(((int) (canvas.getWidth() / 4)) * 4,
						((int) (canvas.getHeight() / 4)) * 4,
						Integer.parseInt((String) frameRCBox.getSelectedItem()));
			}
		});
		scriptArea = new JTextArea(10, 5);
		setDefaultText();
		JLabel codecLabel = new JLabel("Codec :") {
			public Dimension getPreferredSize() {
				return(new Dimension(90, 25));
			}
		};
		JLabel formatLabel = new JLabel("Video format :") {
			public Dimension getPreferredSize() {
				return(new Dimension(90, 25));
			}
		};

		videoFormat = new JComboBox(movieMaker.getContentDescriptors()) {
			public Dimension getPreferredSize() {
				return(new Dimension(300, 25));
			}
		};
		videoFormat.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				movieMaker.setContentDescriptor((ContentDescriptor) videoFormat.getSelectedItem());
				videoCodec.removeAllItems();
				Format formats[] = movieMaker.getFormats();

				for (int i = 0; i < formats.length; i++) {
					videoCodec.addItem(formats[i]);
				}
			}
		});
		videoCodec = new JComboBox(movieMaker.getFormats()) {
			public Dimension getPreferredSize() {
				return(new Dimension(300, 25));
			}
		};
		videoCodec.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				movieMaker.setFormat((Format) videoCodec.getSelectedItem());
			}
		});
	
		JButton parse = new JButton("Default");

		parse.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				setDefaultText();
			}
		});
		beginCapture = new JButton("Start");
		beginCapture.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser where = new JFileChooser(".");

				if (where.showSaveDialog(ICC3D.getDialogParent()) != JFileChooser.CANCEL_OPTION
						&& where.getSelectedFile() != null) {
					String filename = where.getSelectedFile().getName();

					if (filename != null) {
						Canvas3D canvas = view3D.getCanvas();
						Canvas3D offscreenCanvas = view3D.getOffScreenCanvas();

						movieMaker.setFileName(filename);
						movieMaker.prepare(((int) (canvas.getWidth() / 4)) * 4,
								((int) (canvas.getHeight() / 4)) * 4,
								Integer.parseInt((String) (frameRCBox.getSelectedItem())));
						movieMaker.setContentDescriptor((ContentDescriptor) videoFormat.getSelectedItem());
						movieMaker.setFormat((Format) videoCodec.getSelectedItem());
						if (movieMaker.doIt()) {
							canvas.stopRenderer();
							try {
								int w = 4 * (int) (canvas.getWidth() / 4);
								int h = 4 * (int) (canvas.getHeight() / 4);
								byte tmp[] = new byte[w * h * 3];
								DataBufferByte dataBuffer = new DataBufferByte(tmp, tmp.length);
								SampleModel sampleModel = new PixelInterleavedSampleModel(DataBuffer.TYPE_BYTE,
										w, h, 3, w * 3, new int[] {0, 1, 2});
								ColorModel colorModel = PlanarImage.createColorModel(sampleModel);
								java.awt.image.Raster raster = RasterFactory.createWritableRaster(sampleModel,
										dataBuffer, new Point(0, 0));
								TiledImage tiledImage = new TiledImage(0, 0, w, h, 0, 0, sampleModel,
										colorModel);

								tiledImage.setData(raster);
								offscreenBuffer = new ImageComponent2D(ImageComponent2D.FORMAT_RGB8,
										tiledImage);
								offscreenBuffer.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
								offscreenBuffer.setCapability(ImageComponent2D.ALLOW_FORMAT_READ);
								offscreenBuffer.setCapability(ImageComponent2D.ALLOW_SIZE_READ);
								offscreenCanvas.setOffScreenBuffer(offscreenBuffer);
								offscreenCanvas.renderOffScreenBuffer();
								offscreenCanvas.waitForOffScreenRendering();
								RenderedImage ri = offscreenBuffer.getRenderedImage();

								tiledImage.setData(ri.getData());
								movieMaker.addImage(tiledImage.getAsBufferedImage());
								Vector v = ScriptParser.parse(scriptArea.getText());

								for (Enumeration e = v.elements(); e.hasMoreElements();) {
									((MovieInterpolator) e.nextElement()).runMe(offscreenCanvas,
											offscreenBuffer, tiledImage, movieMaker, view3D.getObjectScale());
								}
								movieMaker.signalNoMoreData();
							} catch (Exception e) {
								System.out.println("Failed....");
							}
							canvas.startRenderer();
						} else {
							SystemManager.displayError("An error has occurred",
									"The application was unable to transcode RGB data\nto the specified format.");
							movieMaker = new MovieMaker();
							movieFrame.dispose();
							movieFrame = null;
						}
					}
				}
			}
		});
		
		JPanel labelPanel = new JPanel(new BorderLayout());

		labelPanel.add(formatLabel, BorderLayout.NORTH);
		labelPanel.add(codecLabel, BorderLayout.CENTER);
		labelPanel.add(frameRLabel, BorderLayout.SOUTH);
		
		JPanel comboPanel = new JPanel(new BorderLayout());

		comboPanel.add(videoFormat, BorderLayout.NORTH);
		comboPanel.add(videoCodec, BorderLayout.CENTER);
		comboPanel.add(frameRCBox, BorderLayout.SOUTH);
		
		JPanel gridPanel = new JPanel();

		gridPanel.add(labelPanel);
		gridPanel.add(comboPanel);
		c.add(gridPanel, BorderLayout.NORTH);
		c.add(scriptArea, BorderLayout.CENTER);
		c.add(parse, BorderLayout.WEST);
		c.add(beginCapture, BorderLayout.SOUTH);
		movieFrame.pack();
		movieFrame.setTitle("Capture video");
		movieFrame.setResizable(false);	
	}
	
	/**
	 * Replaces the current script with the default movie script
	 */
	private void setDefaultText() {
		scriptArea.setText("rotate 0 360 0\nframes 100\ngo");
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 5;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Capture";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("screencapture");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.beginCapture = null;
		this.capture = null;
		this.frameRCBox = null;
		this.movieFrame = null;
		this.movieMaker = null;
		this.offscreenBuffer = null;
		this.parameterPanel = null;
		this.scriptArea = null;
		this.shoot = null;
		this.videoCodec = null;
		this.videoFormat = null;
		this.view3D = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		this.view3D = view3D;
		movieMaker = new MovieMaker();
		return true;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		return null;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
