/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import icc3dpackage.colorspaces.generic.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.beans.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.view3d.*;


/**
 * This class is used to load and manage colorspaces (ColorSpace3D objects).
 *
 */
public class ColorSpaceManager extends Module {
	
	/**
	 * All the available plugins.
	 */
	private static Manageable manageable;
	
	/**
	 * The color space currently selected.
	 */
	private ColorSpace3D colorSpace3D;
	
	/**
	 * A vector containing the names of the available color spaces.
	 */
	private Vector names;
	
	/**
	 * The available color spaces.
	 */
	private Vector colorSpaces;
	
	/**
	 * The parameter panel that belongs to this color space manager.
	 */
	private JPanel parameterPanel;
	
	/**
	 * A split pane dividing the parameter panel in two.
	 */
	private JSplitPane jsp;
	
	/**
	 * The color spaces the user can select.
	 */
	private JComboBox currentColorSpace;
	
	/**
	 * Default constructor.
	 */
	public ColorSpaceManager() {}

	/**
	 * Load color space plugins dynamically from default directory: 
	 * icc3dpackage/colorspaces.
	 * 
	 * @param output status text is displayed using this reference.
	 */
	public static void load(JTextArea output) {
		manageable = new Manageable();
		Class c = null;

		try {
			c = Class.forName("icc3dpackage.colorspaces.generic.ColorSpace3D");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the ColorSpace3D class");
			System.exit(1);
		}
		Manager.load(output, c, "icc3dpackage" + File.separator + "colorspaces",
				"color spaces", manageable);
		
		if (output == null) {
			output = new JTextArea();
		}
		Manager.append(output, "\nLoading profiles: ");
			
		File profileDir = new File("profiles");
		String profiles[] = profileDir.list();
		int n = 0;
		
		if (profiles != null) {
			for (int i = 0; i < profiles.length; i++) {
				try {
					ColorSpace3D colorSpace = null;
	
					colorSpace = new ICC_ProfileColorSpace3D("profiles" + File.separator
							+ profiles[i]);
					manageable.getNames().add(colorSpace.getName());
					manageable.getPlugins().add(colorSpace);
					Manager.append(output, "+");
					n++;
				} catch (Exception e) {
					Manager.append(output, "!");
				}
			}
		}
		Manager.append(output, "(" + Integer.toString(n) + ")");
		Manager.append(output, "\nLoading regressions: ");
			
		File csDir = new File("prColorSpace");
		String csFiles[] = csDir.list();
		n = 0;
			if (csFiles != null) {
			for (int i = 0; i < csFiles.length; i++) {
				try {
					ColorSpace3D colorSpace = null;
	
					colorSpace = new PolynomialColorSpace3D("prColorSpace" + File.separator
							+ csFiles[i]);
					manageable.getNames().add(colorSpace.getName());
					manageable.getPlugins().add(colorSpace);
					Manager.append(output, "+");
					n++;
				} catch (Exception e) {
					Manager.append(output, "!");
				}
			}
		}
		Manager.append(output, "(" + Integer.toString(n) + ")");
	}
	
	/**
	 * Retrieves the requested ColorSpace3D object from this ColorSpaceManager.
	 *
	 * @param name of requeste ColorSpace3D. Case-sensitive.
	 *	@return requested ColorSpace3D object, or null if not found
	 */
	public static ColorSpace3D getColorSpace(String name) {
		ColorSpace3D colorSpace = null;
		int i = manageable.getNames().indexOf(name);

		if (i >= 0) {
			ColorSpace3D temp = (ColorSpace3D) manageable.getPlugins().elementAt(i);

			if (temp.isProfile()) {
				try {
					colorSpace = new ICC_ProfileColorSpace3D(((ICC_ProfileColorSpace3D) temp).getFileName());
				} catch (Exception e) {
					try {
						colorSpace = new PolynomialColorSpace3D(((PolynomialColorSpace3D) temp).getFileName());
					} catch (Exception e2) {
						return null;
					}
				}
			} else {
				return (ColorSpace3D) Manager.getPlugin(name, manageable);
			}
		}
		return colorSpace;
	}
	
	/**
	 * Retrieves a list of names of the currently registered 3 component ColorSpace3D plugins.
	 *
	 * @return list of names, or null if no 3 component ColorSpace3D objects are registered.
	 */
	public static String[] getNames() {
		Vector v = new Vector();
		Vector names = manageable.getNames();
		Vector plugins = manageable.getPlugins();

		for (int i = 0; i < names.size(); i++) {
			ColorSpace3D col = getColorSpace((String) names.elementAt(i));

			if (col.getNumberOfComponents() == 3) {
				v.add(names.elementAt(i));
			}
		}
		if (v.size() == 0) {
			return null;
		}
		String ret[] = new String[v.size()];

		for (int i = 0; i < v.size(); i++) {
			ret[i] = (String) v.elementAt(i);
		}
		return ret;
	}
	
	/**
	 * Retrieves a list of names of all the currently registered ColorSpace3D plugins.
	 *
	 * @return An array of names, or null if no ColorSpace3D objects are registered.
	 */
	public static String[] getAllNames() {
		return Manager.getNames(manageable);
	}
	
	/**
	 * Constructs the parameter panel for this ColorSpaceManager.
	 */
	private void createParameterPanel() {
		JPanel upperPanel;

		parameterPanel = new JPanel(new BorderLayout());
		currentColorSpace = new JComboBox(names);
		currentColorSpace.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				colorSpace3D = (ColorSpace3D) colorSpaces.elementAt(names.indexOf(currentColorSpace.getSelectedItem()));
				jsp.setBottomComponent(colorSpace3D.getParameterPanel());
			}
		});
		upperPanel = new JPanel(new BorderLayout());
		JPanel colSpace = new JPanel(new FlowLayout());
			
		JLabel csLabel = new JLabel("Color space");

		csLabel.setHorizontalAlignment(JLabel.CENTER);
		csLabel.setBorder(BorderFactory.createEtchedBorder());
		csLabel.setPreferredSize(new Dimension(150, 25));
		upperPanel.add(csLabel, BorderLayout.NORTH);
		upperPanel.add(currentColorSpace, BorderLayout.SOUTH);
		colSpace.add(upperPanel);
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT, colSpace,
				colorSpace3D.getParameterPanel());
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		parameterPanel.add(jsp, BorderLayout.CENTER);
		// parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Retrieves the currently selected ColorSpace3D object.
	 * @return ColorSpace3D 
	 */
	public ColorSpace3D getColorSpace3D() {
		return colorSpace3D;
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Notifies the color space that the user has moved the 3D objects.
	 */
	public void signalMovement() {
		for (int i = 0; i < colorSpaces.size(); i++) {
			ColorSpace3D cs3d = (ColorSpace3D) colorSpaces.elementAt(i);

			cs3d.signalMovement();
		}
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 2;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Color space";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("color");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.colorSpace3D = null;
		this.colorSpaces = null;
		this.currentColorSpace = null;
		this.jsp = null;
		this.names = null;
		this.parameterPanel = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		names = new Vector();
		colorSpaces = new Vector();
		String spaces[] = ColorSpaceManager.getNames();

		for (int i = 0; i < spaces.length; i++) {
			names.add(spaces[i]);
			colorSpaces.add(ColorSpaceManager.getColorSpace(spaces[i]));
		}
		colorSpace3D = (ColorSpace3D) colorSpaces.elementAt(0);
		view3D.setColorSpaceManager(this);
		return true;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		return new BranchGroup[] {null, colorSpace3D.getAxis()};
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
