/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.media.j3d.BranchGroup;

import icc3dpackage.*;
import icc3dpackage.constraints.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.view3d.*;


/**
 * This class is used to load and manage InteractionConstraint objects.
 *
 */
public class ConstraintsManager extends Module {
	
	/** All the managed plugins */
	protected static Manageable manageable;
	
	/** The paramater panel */
	private JPanel parameterPanel;
	
	/** The selectable constraints */
	private JRadioButton constraintButton[];
	
	/** The available constraints */
	private InteractionConstraint constraint[];
	
	/** The active constraint */
	private int activeConstraint;
	
	/**
	 * Creates a ConstraintsManager.
	 */
	public ConstraintsManager() {}
	
	/**
	 * Loads interactionconstraint plugins dynamically from default directory: 
	 * icc3dpackage/constraints.
	 * 
	 * @param output status text is displayed using this reference.
	 */
	static public void load(JTextArea output) {
		manageable = new Manageable();
		Class c = null;

		try {
			c = Class.forName("icc3dpackage.constraints.generic.InteractionConstraint");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the InteractionConstraint class");
			System.exit(1);
		}
		Manager.load(output, c, "icc3dpackage" + File.separator + "constraints",
				"constraints", manageable);
	}

	/**
	 * Retrieves a requested InteractionConstraint object from this ConstraintsManager.
	 *
	 * @param name of interactionconstraint
	 * @return requested InteractionConstraint, or null if not found
	 *	@see #getNames
	 */
	public static InteractionConstraint getConstraint(String name) {
		return (InteractionConstraint) Manager.getPlugin(name, manageable);
	}
	
	public static String[] getNames() {
		return Manager.getNames(manageable);
	}
	
	/**
	 * Creates the paramater panel.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		JPanel constraintPanel = new JPanel(new GridLayout(0, 1));
		
		String cons[] = ConstraintsManager.getNames();

		constraintButton = new JRadioButton[cons.length];
		constraint = new InteractionConstraint[cons.length];
		Icon icons[];
		ButtonGroup bg = new ButtonGroup();

		for (int i = 0; i < cons.length; i++) {
			InteractionConstraint ic = ConstraintsManager.getConstraint(cons[i]);

			constraint[i] = ic;
			icons = ic.getIcon();
			constraintButton[i] = new JRadioButton(ic.getName(), icons[0], false);
			constraintButton[i].setSelectedIcon(icons[1]);
			if (i == 0) {
				constraintButton[i].setSelected(true);
			}
			constraintButton[i].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent ae) {
					for (int j = 0; j < constraintButton.length; j++) {
						if (ae.getSource() == constraintButton[j]) {
							activeConstraint = j;
							break;
						}
					}
				}
			});
			bg.add(constraintButton[i]);
			constraintPanel.add(constraintButton[i]);
		}
		parameterPanel.add(constraintPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Returns the selected constraint.
	 * @return The active InteractionConstraint.
	 */
	public InteractionConstraint getConstraint() {
		return constraint[activeConstraint];
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 4;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Constraints";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("constraints");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.constraint = null;
		this.constraintButton = null;
		this.parameterPanel = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		view3D.setConstraintsManager(this);
		return true;
	}
	
	public BranchGroup[] getSceneBranch() {
		return null;
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
