/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.image.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.media.j3d.BranchGroup;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.image2d.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.generic.*;


/**
 * This class is used to manage a 2D image.
 * Keeps track of which visualization is currently active.
 * It is also used to create BranchGroups to represent it in 3D.
 */
public class ImageModule extends Module implements GamutData {
	
	/** The View3D parent. */
	private View3D view3D;
	
	/** The image */
	private Image2D image;
	
	/** The time when the image was modified */
	private long timeStamp;
	
	/** An array of all available visualization names */
	private String visualizationNames[];
		
	/** The current visualization */
	private Visualization3D vis3D = null;
	
	/** The available visualizations */
	private Vector visualizations;
	
	/** The choice of the available visualizations */
	private JComboBox visualizationChooser;
	
	/** A split pane dividing the image parameter panel in two parts */
	private JSplitPane jsp;
	
	/** The image parameter panel */
	private JPanel parameterPanel;
	
	/** Select the image source */
	private JButton imageSource;
	
	/**
	 * Creates an ImageModule.
	 */
	public ImageModule() {}

	/**
	 * Creates n number of BranchGroups of the current image to represent it as specified.
	 *
	 * @return newly created BranchGroups
	 * @see javax.media.j3d.BranchGroup
	 */
	public BranchGroup[] getSceneBranch() {
		if (image == null) {
			return null;
		}
		ImageCanvas2D image2D = image.getImageCanvas2D();
		ColorSpace3D colorSpaces[] = new ColorSpace3D[2];

		colorSpaces[1] = view3D.getColorSpaceManager().getColorSpace3D();
		colorSpaces[0] = image2D.getInputColorSpace();
		if (colorSpaces[0].getNumberOfComponents() != 3) {
			colorSpaces[0] = ColorSpaceManager.getColorSpace("XYZ");
		}
		float colors[][] = new float[2][];
		float xyz[] = image2D.getXYZ_Colors();

		colors[0] = colorSpaces[0].fromCIEXYZ(xyz);
		colors[1] = colorSpaces[1].fromCIEXYZ(xyz);
		byte col[] = colorSpaces[1].toSRGB(colors[1]);
		int numUnique = colors[1].length / 3;
		int numEqual[] = new int[numUnique];
		int colorToPixelIndex[][] = image2D.getColorToPixelIndices();

		for (int i = 0; i < numUnique; i++) {
			numEqual[i] = colorToPixelIndex[i].length;
		}
		return vis3D.createSceneObject(numUnique, colors, numEqual, col,
				colorToPixelIndex, colorSpaces, this, view3D);
	}
	
	/**
	 * Retrieves the currently selected Visualization3D object.
	 *	This Visualization3D object is used to construct subsequent BranchGroups until a 
	 * new visualization is selected.
	 *
	 * @return currently selected visualization
	 */
	public Visualization3D getVisualization3D() {
		return vis3D;
	}
	
	/**
	 * Constructs this ImageManager's parameter panel.
	 * Used to select which of the available visualizatins is to be used for representation.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		visualizationChooser = new JComboBox(visualizationNames);
		visualizationChooser.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				vis3D = (Visualization3D) visualizations.elementAt(visualizationChooser.getSelectedIndex());
				jsp.setBottomComponent(vis3D.getParameterPanel());
				Container c = jsp.getParent();

				if (c != null) {
					c.doLayout();
					Container cc = c.getParent();

					if (cc != null) {
						cc.doLayout();
					}
				}
				c.repaint();
			}
		});
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		
		JPanel upperPanel = new JPanel(new FlowLayout());

		imageSource = new JButton("None");
		imageSource.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Image2D img[] = view3D.getICC3D().getSessionManager().getImages();
				JPopupMenu popMenu = new JPopupMenu("Image source");
				JMenuItem items[] = new JMenuItem[img.length + 2];

				items[0] = new JMenuItem("None", IconManager.getIcon("none"));
				items[1] = new JMenuItem("New", IconManager.getIcon("new"));
				for (int i = 0; i < img.length; i++) {
					items[i + 2] = new JMenuItem(img[i].getTitle(), img[i].getIcon());
				}
				for (int i = 0; i < items.length; i++) {
					popMenu.add(items[i]);
					items[i].addActionListener(new ActionListener() {
						public void actionPerformed(ActionEvent ae) {
							JMenuItem item = (JMenuItem) ae.getSource();

							if (item.getText().equals("None")) {
								image = null;
								imageSource.setText("None");
							} else if (item.getText().equals("New")) {
								Image2D tmpImg = view3D.getICC3D().getSessionManager().newImage2D();

								if (tmpImg != null) {
									imageSource.setText(tmpImg.getTitle());
									image = tmpImg;
								} else {
									imageSource.setText("None");
									image = null;
								}
							} else {
								Image2D s = view3D.getICC3D().getSessionManager().getImage(item.getText());

								if (s == null) {
									imageSource.setText("None");
									image = null;
								} else {
									imageSource.setText(item.getText());
									image = s;
								}
							}
						}
					});
				}
				popMenu.show(imageSource, 0, 0);
			}
		});
		
		JLabel visLabel = new JLabel("  Image visualization  ");

		visLabel.setHorizontalAlignment(JLabel.CENTER);
		visLabel.setBorder(BorderFactory.createEtchedBorder());
		visLabel.setPreferredSize(new Dimension(150, 25));
		JLabel imgLabel = new JLabel("  Image source  ");

		imgLabel.setHorizontalAlignment(JLabel.CENTER);
		imgLabel.setBorder(BorderFactory.createEtchedBorder());
		imgLabel.setPreferredSize(new Dimension(150, 25));
		JPanel innerPanel = new JPanel(new GridLayout(0, 1));

		innerPanel.add(imgLabel);
		innerPanel.add(imageSource);
		innerPanel.add(visLabel);
		innerPanel.add(visualizationChooser);
		upperPanel.add(innerPanel);
		jsp.setTopComponent(upperPanel);
		jsp.setBottomComponent(vis3D.getParameterPanel());
		
		JPanel panel1 = new JPanel(new BorderLayout());

		panel1.add(jsp, BorderLayout.CENTER);
		parameterPanel.add(panel1, BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;	
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 100;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Image gamut";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("visualization");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.image = null;
		this.imageSource = null;
		this.jsp = null;
		this.parameterPanel = null;
		this.view3D = null;
		this.vis3D = null;
		this.visualizationChooser = null;
		this.visualizationNames = null;
		Visualization3D.deInitialize(visualizations);
		this.visualizations = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		visualizations = new Vector();
		String names[] = VisualizationManager.getNames();

		visualizationNames = new String[names.length];
		for (int i = 0; i < names.length; i++) {
			visualizationNames[i] = names[i];
			visualizations.add(VisualizationManager.getVisualization(names[i]));
		}
		vis3D = (Visualization3D) visualizations.elementAt(0);
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Returns the View3D parent.
	 * @return The View3D parent.
	 */
	public View3D getView3D() {
		return view3D;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return vis3D.checkParams();
	}
	
	/**
	 * Set the image source.
	 * @param img The image.
	 */
	public void setSource(Image2D img) {
		imageSource.setText(img.getTitle());
		image = img;
	}
	
	/**
	 * Returns the image source.
	 * @return The image source.
	 */
	public Image2D getImage2D() {
		return image;
	}
	
	/**
	 * Returns the image data for gamut comparisons.
	 * @return The image data for gamut comparisons.
	 */
	public float[] getData() {
		if (image == null) {
			return null;
		}
		ColorSpace3D labSpace = ColorSpaceManager.getColorSpace("Lab");

		return labSpace.fromCIEXYZ(image.getImageCanvas2D().getXYZ_Colors());
	}
	
	/**
	 * Returns the image filename.
	 * @return The image filename.
	 */
	public String getFileName() {
		if (image == null) {
			return "";
		}
		return image.getTitle();
	}
}
