/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.media.j3d.BranchGroup;

import icc3dpackage.*;
import icc3dpackage.interactions.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;

import icc3dpackage.view3d.*;


/**
 * This class is used to load and manage Interaction3D objects.
 *
 */
public class InteractionManager extends Module implements ActionListener {
	
	/** All local interactions currently available */
	private static Manageable localInteractions;
		
	/** All global interactions currently available */
	private static Manageable globalInteractions;
	
	/** The paramater panel */
	private JPanel parameterPanel;
	
	/** The split panel */
	private JSplitPane jsp;
	
	/** Continous 2D update */
	private JCheckBox interactivity;
	
	/** The selectable interactions */
	private JRadioButton interactionButton[];
	
	/** The available interactions */
	private Interaction3D interaction[];
	
	/** The active interaction */
	private int activeInteraction;
	
	/**
	 * Loads Interaction3D plugins dynamically from default directory: 
	 * icc3dpackage/interactions.
	 * 
	 * @param output statustext is displayed using this reference.
	 */	
	static public void load(JTextArea output) {
		Manageable temp = new Manageable();
		Class c = null;

		try {
			c = Class.forName("icc3dpackage.interactions.generic.Interaction3D");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the Interaction3D class");
			System.exit(1);
		}
		Manager.load(output, c, "icc3dpackage" + File.separator + "interactions",
				"interactions", temp);
		localInteractions = new Manageable();
		globalInteractions = new Manageable();
		
		for (int i = 0; i < temp.getNames().size(); i++) {
			Interaction3D i3d = (Interaction3D) temp.getPlugins().elementAt(i);

			if (i3d.getType() == Interaction3D.GLOBAL) {
				globalInteractions.getNames().add(temp.getNames().elementAt(i));
				globalInteractions.getPlugins().add(temp.getPlugins().elementAt(i));
			} else if (i3d.getType() == Interaction3D.LOCAL) {
				localInteractions.getNames().add(temp.getNames().elementAt(i));
				localInteractions.getPlugins().add(temp.getPlugins().elementAt(i));
			}
		}
	}

	/**
	 * Retrieves a list of names of the currently registered global Interaction3D plugins.
	 *
	 * @return a list of names, or null if no Interaction3D objects are registered.
	 * @see #getLocalInteraction
	 * @see #getGlobalInteraction
	 */
	static public String[] getGlobalNames() {
		return Manager.getNames(globalInteractions);
	}

	/**
	 * Retrieves a list of names of the currently registered local Interaction3D plugins.
	 *
	 * @return a list of names, or null if no Interaction3D objects are registered.
	 * @see #getLocalInteraction
	 * @see #getGlobalInteraction
	 */
	static public String[] getLocalNames() {
		return Manager.getNames(localInteractions);
	}
	
	/**
	 * Retrieves a requested local Interaction3D object from this InteractionManager.
	 *
	 * @param name of interaction
	 * @return requested Interaction3D, or null if not found
	 *	@see #getLocalNames
	 *	@see #getGlobalNames
	 */	
	public static Interaction3D getLocalInteraction(String name) {
		return (Interaction3D) Manager.getPlugin(name, localInteractions);
	}
	
	/**
	 * Retrieves a requested global Interaction3D object from this InteractionManager.
	 *
	 * @param name of interaction
	 * @return requested Interaction3D, or null if not found
	 *	@see #getLocalNames
	 *	@see #getGlobalNames
	 */	
	public static Interaction3D getGlobalInteraction(String name) {
		return (Interaction3D) Manager.getPlugin(name, globalInteractions);
	}
	
	/**
	 * Returns the selected interaction.
	 * @return The active Interaction3D.
	 */
	public Interaction3D getInteraction() {
		return interaction[activeInteraction];
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 3;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Interactions";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the module icon.
	 * @return The module icon.
	 */ 
	public Icon getIcon() {
		return IconManager.getIcon("interaction");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.interaction = null;
		this.interactionButton = null;
		this.interactivity = null;
		this.parameterPanel = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		view3D.setInteractionManager(this);
		return true;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		return null;
	}
	
	/**
	 * Creates the parameter panel.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		JPanel interactionPanel = new JPanel(new GridLayout(0, 1));

		interactivity = new JCheckBox("Continuous 2D update");
		interactionPanel.add(interactivity);
		
		String locs[] = InteractionManager.getLocalNames();
		String globs[] = InteractionManager.getGlobalNames();
		
		int num = locs.length + globs.length;

		interactionButton = new JRadioButton[num];
		interaction = new Interaction3D[num];
		ButtonGroup bg = new ButtonGroup();

		interactionPanel.add(new JLabel("Local interactions:"));
		for (int i = 0; i < locs.length; i++) {
			Interaction3D i3d = InteractionManager.getLocalInteraction(locs[i]);

			interaction[i] = i3d;
			interactionButton[i] = new JRadioButton(i3d.getName(), false);
			if (i == 0) {
				interactionButton[i].setSelected(true);
			}
			interactionButton[i].addActionListener(this);
			bg.add(interactionButton[i]);
			interactionPanel.add(interactionButton[i]);
		}
		interactionPanel.add(new JLabel("Global interactions:"));
		for (int i = 0; i < globs.length; i++) {
			Interaction3D i3d = InteractionManager.getGlobalInteraction(globs[i]);

			interaction[i + locs.length] = i3d;
			interactionButton[i + locs.length] = new JRadioButton(i3d.getName(), false);
			if (i + locs.length == 0) {
				interactionButton[i + locs.length].setSelected(true);
			}
			interactionButton[i + locs.length].addActionListener(this);
			bg.add(interactionButton[i + locs.length]);
			interactionPanel.add(interactionButton[i + locs.length]);
		}
		
		JPanel flowPanel = new JPanel(new FlowLayout());

		flowPanel.add(interactionPanel);
		jsp.setTopComponent(flowPanel);
		jsp.setBottomComponent(interaction[activeInteraction].getParameterPanel());
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		parameterPanel.add(jsp, BorderLayout.CENTER);
	}
	
	/**
	 * Handles changes in the interaction selection.
	 * @param ae The ActionEvent that triggered the change.
	 */
	public void actionPerformed(ActionEvent ae) {
		for (int j = 0; j < interactionButton.length; j++) {
			if (ae.getSource() == interactionButton[j]) {
				activeInteraction = j;
				jsp.setBottomComponent(interaction[activeInteraction].getParameterPanel());
				jsp.doLayout();
				break;
			}
		}
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
	
	/**
	 * Returns whether the 2D image should be continously updated.
	 * @return Whether the 2D image should be continously updated.
	 */
	public boolean getContinuousUpdate() {
		return interactivity.isSelected();
	}
}
