/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.media.j3d.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.*;
import icc3dpackage.parser.*;
import icc3dpackage.utils.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.*;
import icc3dpackage.visualizations.generic.*;


/**
 * This class is used to represent a gamut based on data from a file format
 * used by J. Morovic in some of his samples.
 */
public class JM_GamutModule extends Module implements GamutData {
	
	/** The filename. */
	private String fileName;
	
	/** The active visualization. */
	private Visualization3D visualization3D = null;
	
	/** The names of the available visualizations. */
	private Vector visualizationNames;
	
	/** The available visualizations. */
	private Vector visualizations;
	
	/** The gamut data as read from a file. */
	private float gamutData[];
	
	/** A Lab color space. */
	private ColorSpace3D labColorSpace3D;
	
	/** The parser that reads the input file. */
	private Parser3 parser;	
	
	/** The parameter panel. */
	private JPanel parameterPanel = null;
	
	/** The name of the file. */
	private JLabel name;
	
	/** Determines whether a file has been read. */
	private boolean gamutLoaded = false;
	
	/** A GUI button */
	private JButton loadGamut, saveMaxima;
	
	/** A choice of all the available visualizations. */
	private JComboBox gamutVisualization;
	
	/** Splits the parameter panel in two parts. */ 
	private JSplitPane jsp;
	
	/** The View3D parent. */
	private View3D view3D;
	
	/**
	 * Default constructor.
	 */
	public JM_GamutModule() {}
	
	/**
	 * Creates the parameter panel for this GamutView.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		gamutVisualization = new JComboBox(visualizationNames);
		gamutVisualization.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				String type = (String) gamutVisualization.getSelectedItem();

				visualization3D = (Visualization3D) visualizations.elementAt(visualizationNames.indexOf(type));
				jsp.setBottomComponent(visualization3D.getParameterPanel());
			}
		});
	  
		name = new JLabel("No gamut loaded");
		name.setHorizontalAlignment(JLabel.CENTER);
		name.setBorder(BorderFactory.createEtchedBorder());
		name.setPreferredSize(new Dimension(150, 25));
		
		JLabel fgVis = new JLabel("Gamut visualization");

		fgVis.setHorizontalAlignment(JLabel.CENTER);
		fgVis.setBorder(BorderFactory.createEtchedBorder());
		fgVis.setPreferredSize(new Dimension(150, 25));
		loadGamut = new JButton("Load gamut");
		loadGamut.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser chooser = new JFileChooser(".");

				if (chooser.showOpenDialog(ICC3D.getDialogParent()) != JFileChooser.CANCEL_OPTION) {
					saveMaxima.setEnabled(false);
					parser = new Parser3(chooser.getSelectedFile().getAbsolutePath());
					if (parser.parse() == true) {
						name.setText("File: " + chooser.getSelectedFile().getName());
						fileName = chooser.getSelectedFile().getName();
						float tmpData[] = parser.getData();

						if (tmpData != null && tmpData.length > 1) {
							gamutData = new float[tmpData.length];
							for (int i = 0; i < (tmpData.length / 3); i++) {
								gamutData[i * 3 + 1] = tmpData[i * 3] / 100.0f;
								gamutData[i * 3 + 2] = 0.5f + tmpData[i * 3 + 1] / 256.0f;
								gamutData[i * 3] = 0.5f + tmpData[i * 3 + 2] / 256.0f;
							}
							gamutLoaded = true;
						} else {
							name.setText("No gamut loaded");
							gamutLoaded = false;
							gamutData = null;
						}
					} else {
						name.setText("No gamut loaded");
						gamutLoaded = false;
						gamutData = null;
					}
				}
			}
		});
		saveMaxima = new JButton("Save SM");
		saveMaxima.setEnabled(false);
		saveMaxima.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				SystemManager.displayError("Error", "Not supported by the current version");
				//((SegmentMaximaVisualization3D) visualizations.elementAt(visualizationNames.indexOf("Segment Maxima"))).exportToFile();
			}
		});
		
		JPanel panel1 = new JPanel(new BorderLayout());

		panel1.add(loadGamut, BorderLayout.NORTH);
		
		JPanel panel2 = new JPanel(new BorderLayout());

		panel2.add(name, BorderLayout.NORTH);
		panel1.add(panel2, BorderLayout.SOUTH);
		
		JPanel panel3 = new JPanel(new BorderLayout());

		panel3.add(saveMaxima, BorderLayout.NORTH);
		panel2.add(panel3, BorderLayout.SOUTH);
		
		JPanel panel4 = new JPanel(new BorderLayout());

		panel4.add(fgVis, BorderLayout.NORTH);
		panel3.add(panel4, BorderLayout.SOUTH);
		
		JPanel panel5 = new JPanel(new BorderLayout());

		panel5.add(gamutVisualization, BorderLayout.NORTH);
		panel4.add(panel5, BorderLayout.SOUTH);
		
		JPanel upperPanel = new JPanel(new FlowLayout());

		upperPanel.add(panel1);

		jsp.setTopComponent(upperPanel);
		jsp.setBottomComponent(visualization3D.getParameterPanel());
		parameterPanel.add(jsp, BorderLayout.CENTER);
	}
	
	/**
	 * Creates n number of BranchGroups to represent the Gamut in 3D.
	 *
	 * @return An array of BranchGroups.
	 */
	public BranchGroup[] getSceneBranch() {
		if (gamutLoaded && visualization3D != null) {
			float vertex[][] = new float[2][];

			vertex[0] = gamutData;
			ColorSpace3D colorSpace3D = view3D.getColorSpaceManager().getColorSpace3D();

			vertex[1] = ColorSpaceConverter.convert(gamutData, labColorSpace3D,
					colorSpace3D);
			ColorSpace3D inpOup[] = new ColorSpace3D[2];

			inpOup[0] = labColorSpace3D;
			inpOup[1] = colorSpace3D;
			BranchGroup[] branchGroup = visualization3D.createGamutObject(gamutData.length
					/ 3,
					vertex,
					colorSpace3D.toSRGB(vertex[1]),
					inpOup,
					view3D);

			if (visualization3D.getName().equals("Segment Maxima")) {
				saveMaxima.setEnabled(true);
			}
			return branchGroup;	
		}
		return (new BranchGroup[] {new BranchGroup()});
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Returns the gamut point data.
	 * @return A float array containing Lab data.
	 */
	public float[] getData() {
		return this.gamutData;
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 100;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Lab data (abL text)";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("gamut_vis");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.gamutData = null;
		this.gamutVisualization = null;
		this.jsp = null;
		this.labColorSpace3D = null;
		this.loadGamut = null;
		this.name = null;
		this.parameterPanel = null;
		this.parser = null;
		this.saveMaxima = null;
		this.view3D = null;
		this.visualization3D = null;
		this.visualizationNames = null;
		Visualization3D.deInitialize(visualizations);
		this.visualizations = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		labColorSpace3D = ColorSpaceManager.getColorSpace("Lab");
		visualizationNames = new Vector();
		visualizations = new Vector();
		String names[] = VisualizationManager.getNames();

		for (int i = 0; i < names.length; i++) {
			visualizationNames.add(names[i]);
			visualizations.add(VisualizationManager.getVisualization(names[i]));
		}
		visualization3D = (Visualization3D) visualizations.elementAt(0);
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		if (visualization3D == null) {
			return true;
		}
		return visualization3D.checkParams();
	}
	
	/**
	 * Returns the filename.
	 * @return The filename.
	 */
	public String getFileName() {
		return fileName;
	}
}
