/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.media.j3d.BranchGroup;

import icc3dpackage.*;
import icc3dpackage.gma.generic.GamutMappingAlgorithm;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.*;


/**
 * This class is used to load and manage mapping algorithms.
 *
 */
public class MappingManager extends Module {
	
	/** All the managed plugins */
	private static Manageable manageable;
	
	/** The last shown algorithm parameter panel */
	private Component last;
	
	/** The parameter panel */
	private JPanel parameterPanel;
	
	/** The original gamut */
	private JComboBox sourceGamut;
	
	/** The reproduction gamut */
	private JComboBox destinationGamut;
	
	/** The source data */
	private JComboBox source;
	
	/** The algorithm */
	private JComboBox algorithmCombo;
	
	/** The available gamut mapping algorithms */
	private GamutMappingAlgorithm algorithm[];
	
	/** The View3D parent */
	private View3D view3D;
	
	/**
	 * Creates a MappingManager
	 */
	public MappingManager() {}
	
	/**
	 * Loads algorithm plugins dynamically from the default directory: 
	 * icc3dpackage/gma.
	 * 
	 * @param output status text is displayed using this reference.
	 */
	static public void load(JTextArea output) {
		manageable = new Manageable();
		Class c = null;

		try {
			c = Class.forName("icc3dpackage.gma.generic.GamutMappingAlgorithm");
		} catch (Exception e) {
			SystemManager.displayError(
				"Error",
				"Unable to find the GamutMappingAlgorithm class");
			System.exit(1);
		}
		Manager.load(output, c, "icc3dpackage" + File.separator + "gma",
				"gamut mapping algorithms", manageable);
	}

	/**
	 * Retrieves a requested GamutMappingAlgorithm.
	 *
	 * @param name The name of the GamutMappingAlgorithm
	 * @return The requested GamutMappingAlgorithm, or null if not found
	 */
	public static GamutMappingAlgorithm getAlgorithm(String name) {
		return (GamutMappingAlgorithm) Manager.getPlugin(name, manageable);
	}
	
	/**
	 * Returns the names of the available algorithms.
	 * @return The names of the available algorithms.
	 */
	public static String[] getNames() {
		return Manager.getNames(manageable);
	}
	
	/**
	 * Create the parameter panel.
	 */
	private void createParameterPanel() {
		sourceGamut = new JComboBox();
		source = new JComboBox();
		destinationGamut = new JComboBox();
		algorithmCombo = new JComboBox();
		parameterPanel = new JPanel(new BorderLayout());
		
		JPanel leftPanel = new JPanel(new GridLayout(0, 1));
		JPanel rightPanel = new JPanel(new GridLayout(0, 1));
		JPanel innerPanel = new JPanel(new BorderLayout());

		innerPanel.add(leftPanel, BorderLayout.WEST);
		innerPanel.add(rightPanel, BorderLayout.CENTER);
		leftPanel.add(new JLabel(" Map "));
		rightPanel.add(source);
		leftPanel.add(new JLabel(" from "));
		rightPanel.add(sourceGamut);
		leftPanel.add(new JLabel(" to "));
		rightPanel.add(destinationGamut);
		leftPanel.add(new JLabel(" using "));
		rightPanel.add(algorithmCombo);
		
		String cons[] = MappingManager.getNames();

		algorithm = new GamutMappingAlgorithm[cons.length];
		for (int i = 0; i < cons.length; i++) {
			GamutMappingAlgorithm ic = MappingManager.getAlgorithm(cons[i]);

			algorithm[i] = ic;
			algorithmCombo.addItem(cons[i]);
		}
		algorithmCombo.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				parameterPanel.remove(last);
				last = algorithm[algorithmCombo.getSelectedIndex()].getParameterPanel();
				parameterPanel.add(last, BorderLayout.CENTER);
				last.invalidate();
				last.doLayout();
				parameterPanel.invalidate();
				parameterPanel.doLayout();
				parameterPanel.validate();
				parameterPanel.repaint();
				last.repaint();
			}
		});
		
		JButton startButton = new JButton("Execute");
		JPanel startPanel = new JPanel();

		startPanel.add(startButton);
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (sourceGamut.getItemCount() == 0 || sourceGamut.getSelectedIndex() < 0) {
					SystemManager.displayError(
						"Error",
						"No source gamut selected");
					return;
				}
				if (destinationGamut.getItemCount() == 0
						|| destinationGamut.getSelectedIndex() < 0) {
					SystemManager.displayError(
						"Error",
						"No destination gamut selected");
					return;
				}
				ProgressBar pBar = new ProgressBar(1, "Performing gamut mapping", "Performing gamut mapping");
				MovingObject objectToMap = (MovingObject) view3D.getImageGamuts().get(source.getSelectedIndex());
				Tetrahedron mapFrom[] = null;

				if (sourceGamut.getSelectedIndex() < view3D.getImageGamuts().size()) {
					mapFrom = ((MovingObject) view3D.getImageGamuts().get(sourceGamut.getSelectedIndex())).getTetrahedrons();
				} else {
					mapFrom = (Tetrahedron[]) view3D.getDeviceGamuts().get(sourceGamut.getSelectedIndex()
							- view3D.getImageGamuts().size());
				}
				Tetrahedron mapTo[] = (Tetrahedron[]) view3D.getDeviceGamuts().get(destinationGamut.getSelectedIndex());
				pBar.setDescription("Please wait");
				algorithm[algorithmCombo.getSelectedIndex()].map(objectToMap, mapTo, mapFrom);
				pBar.dispose();
			}
		});
		parameterPanel.add(innerPanel, BorderLayout.NORTH);
		last = algorithm[algorithmCombo.getSelectedIndex()].getParameterPanel();
		parameterPanel.add(last, BorderLayout.CENTER);
		parameterPanel.add(startPanel, BorderLayout.SOUTH);
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 4;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Mapping";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return false;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("autoalgo");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.algorithm = null;
		this.parameterPanel = null;
		// TODO ....
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		this.view3D = view3D;
		view3D.setMappingManager(this);
		return true;
	}
	
	public BranchGroup[] getSceneBranch() {
		return null;
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
	
	/**
	 * Rebuilds the lists of available source and destination gamuts.
	 */
	public void notifyGamutsChanged() {
		ArrayList al = view3D.getImageGamuts();

		source.removeAllItems();
		sourceGamut.removeAllItems();
		if (al.size() > 0) {
			for (int i = 1; i <= al.size(); i++) {
				source.addItem("Image gamut " + i);
				sourceGamut.addItem("Image gamut " + i);
			}
		}
		al = view3D.getDeviceGamuts();
		destinationGamut.removeAllItems();
		if (al.size() > 0) {
			for (int i = 1; i <= al.size(); i++) {
				destinationGamut.addItem("Device gamut " + i);
				sourceGamut.addItem("Device gamut " + i);
			}
		}
	}
}
