/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.*;
import icc3dpackage.pickable.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.*;
import icc3dpackage.visualizations.appearance.*;


/**
 * A module that can visualize the effect of a gamut mapping algorithm.
 */
public class MappingTest extends Module {
	
	/** The mamimum number of subdivisions per axis. */
	private static final int MAX_DETAIL = 100;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The View3D parent. */
	private View3D view3D;
	
	/** The shape that is transformed by the algorithms. */
	private MovingObject mObject;
	
	/** The object that decides the appearance of the shape. */
	private TriangleAppearance triangleAppearance;
	
	/** The parent branch group. */
	private BranchGroup parent;
	
	/** The child branch group. */
	private BranchGroup child;
	
	/** Determines which planes should be visualized. */
	private JCheckBox addX, addY, addZ;
	
	/** Choose the number of subdivisions. */
	private JSlider xDetail, yDetail, zDetail;
	
	/** The number of subdivisions. */
	private JLabel xDetLabel, yDetLabel, zDetLabel;
	
	/** Choose the first row to be displayed. */
	private JSlider xStart, yStart, zStart;
	
	/** The first row to be displayed. */
	private JLabel xStartLabel, yStartLabel, zStartLabel;
	
	/** Choose the last row to be displayed. */
	private JSlider xStop, yStop, zStop;
	
	/** The last row to be displayed. */
	private JLabel xStopLabel, yStopLabel, zStopLabel;
	
	/** Create a new shape when the user performs an update. */
	private boolean newShapeOnUpdate;
	
	/** Disable update of shape. */
	private boolean disableUpdate;
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		this.view3D = view3D;
		newShapeOnUpdate = true;
		return true;
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		view3D = null;
		parameterPanel = null;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		disableUpdate = true;
		int detX = xStop.getMaximum();
		int detY = yStop.getMaximum();
		int detZ = zStop.getMaximum();
		int detailX  = xDetail.getValue();
		int detailY  = yDetail.getValue();
		int detailZ  = zDetail.getValue();
		
		if (!newShapeOnUpdate && (mObject != null && child != null && parent != null && detX == detailX && detY == detailY && detZ == detailZ)) {
			float oldPos[] = new float[detX*detY*detZ*3];
			mObject.getGeometryArrays()[0].getCoordinates(0, oldPos);
			child.detach();
			child = createChild(oldPos, detX, detY, detZ);
			
			BranchGroup[] bg = new BranchGroup[2];
			for (int i = 0; i < 2; i++) {
				bg[i] = new BranchGroup();
			}
			
			parent = new BranchGroup();
			parent.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
			parent.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
			parent.addChild(child);
			bg[1].addChild(parent);
			disableUpdate = false;
			return bg;
		}
		
		newShapeOnUpdate = false;
		
		xStart.setMaximum(detailX-1);
		yStart.setMaximum(detailY-1);
		zStart.setMaximum(detailZ-1);
		
		xStop.setMaximum(detailX);
		yStop.setMaximum(detailY);
		zStop.setMaximum(detailZ);
		
		float fDetailX = detailX - 1;
		float fDetailY = detailY - 1;
		float fDetailZ = detailZ - 1;
		
		float vertexPositions[] = new float[detailX * detailY * detailZ * 3];
		for (int i = 0; i < detailX; i++) {
			for (int j = 0; j < detailY; j++) {
				for (int k = 0; k < detailZ; k++) {
					vertexPositions[((i * detailY + j) * detailZ + k) * 3 + 0] = i / fDetailX;
					vertexPositions[((i * detailY + j) * detailZ + k) * 3 + 1] = j / fDetailY;
					vertexPositions[((i * detailY + j) * detailZ + k) * 3 + 2] = k / fDetailZ;
				}
			}
		}
		
		BranchGroup[] bg = new BranchGroup[2];
		for (int i = 0; i < 2; i++) {
			bg[i] = new BranchGroup();
		}
		parent = new BranchGroup();
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		
		child = createChild(vertexPositions, detailX, detailY, detailZ);
		parent.addChild(child);
		bg[1].addChild(parent);
		disableUpdate = false;
		return bg;
	}
	
	/**
	 * Creates the indices and the shape(s).
	 * @param vPos The vertex coordinates.
	 * @return The child branch group.
	 */
	private BranchGroup createChild(float vPos[], int detailX, int detailY, int detailZ) {
		boolean addX = this.addX.isSelected();
		boolean addY = this.addY.isSelected();
		boolean addZ = this.addZ.isSelected();
		if (!addX && !addY && !addZ) {
			BranchGroup bg = new BranchGroup();
			bg.setCapability(BranchGroup.ALLOW_DETACH);
			return bg;
		}
		
		int startX = xStart.getValue();
		int startY = yStart.getValue();
		int startZ = zStart.getValue();
		int stopX = xStop.getValue();
		int stopY = yStop.getValue();
		int stopZ = zStop.getValue();
		int nextIndex = 0;
		int indexCount = 0;
		if (addX) {
			indexCount += (stopX - startX) * (stopY - startY - 1) * (stopZ - startZ - 1) * 2 * 3;
		}
		if (addY) {
			indexCount += (stopY - startY) * (stopX - startX - 1) * (stopZ - startZ - 1) * 2 * 3;
		}
		if (addZ) {
			indexCount += (stopZ - startZ) * (stopX - startX - 1) * (stopY - startY - 1) * 2 * 3;
		}
		if (indexCount<3) {
			BranchGroup bg = new BranchGroup();
			bg.setCapability(BranchGroup.ALLOW_DETACH);
			return bg;
		}
		int indices[] = new int[indexCount];
		for (int i = 0; i < MAX_DETAIL; i++) {								// for each plane
			for (int j = 0; j < MAX_DETAIL; j++) {							// for each line
				for (int k = 0; k < MAX_DETAIL; k++) {						// for each rectangle
					
					if (addX) {
						if ((i >= startX) && (i < stopX)) {							// create plane where x is i
							if ((j >= startY) && (j < stopY - 1)) {
								if ((k >= startZ) && (k < stopZ - 1)) {
									indices[nextIndex++] = (i * detailY + j) * detailZ + k;
									indices[nextIndex++] = (i * detailY + j + 1) * detailZ + k;
									indices[nextIndex++] = (i * detailY + j + 1) * detailZ + k + 1;
									indices[nextIndex++] = (i * detailY + j) * detailZ + k;
									indices[nextIndex++] = (i * detailY + j + 1) * detailZ + k + 1;
									indices[nextIndex++] = (i * detailY + j) * detailZ + k + 1;
								}
							}
						}
					}
					
					if (addY) {
						if ((i >= startY) && (i < stopY)) {							// create plane where y is i
							if ((j >= startX) && (j < stopX - 1)) {
								if ((k >= startZ) && (k < stopZ - 1)) {
									indices[nextIndex++] = (j * detailY + i) * detailZ + k;
									indices[nextIndex++] = ((j + 1) * detailY + i) * detailZ + k;
									indices[nextIndex++] = ((j + 1) * detailY + i) * detailZ + k + 1;
									indices[nextIndex++] = (j * detailY + i) * detailZ + k;
									indices[nextIndex++] = ((j + 1) * detailY + i) * detailZ + k + 1;
									indices[nextIndex++] = (j * detailY + i) * detailZ + k + 1;
								}
							}
						}
					}
					
					if (addZ) {
						if ((i >= startZ) && (i < stopZ)) {							// create plane where z is i
							if ((j >= startX) && (j < stopX - 1)) {
								if ((k >= startY) && (k < stopY - 1)) {
									indices[nextIndex++] = (j * detailY + k) * detailZ + i;
									indices[nextIndex++] = ((j + 1) * detailY + k) * detailZ + i;
									indices[nextIndex++] = ((j + 1) * detailY + k + 1) * detailZ + i;
									indices[nextIndex++] = (j * detailY + k) * detailZ + i;
									indices[nextIndex++] = ((j + 1) * detailY + k + 1) * detailZ + i;
									indices[nextIndex++] = (j * detailY + k + 1) * detailZ + i;
								}
							}
						}
					}
					
				}
			}
		}
		
		int vertexCount = vPos.length / 3;
		int vertexFormat = GeometryArray.COORDINATES | GeometryArray.COLOR_3;
		IndexedTriangleArray ita = new IndexedTriangleArray(vertexCount, vertexFormat, indexCount);
		ita.setCoordinates(0, vPos);
		ColorSpace3D cs3d = view3D.getColorSpaceManager().getColorSpace3D();
		ita.setColors(0, cs3d.toSRGB(vPos));
		ita.setCoordinateIndices(0, indices);
		ita.setColorIndices(0, indices);
		
		GeometryArray ga[] = new GeometryArray[]{ita};
		
		
		if (mObject == null) {
			mObject = new MovingObject();
		}
		mObject.setColors(null);
		mObject.setColorSpace(cs3d);
		mObject.setImage2D(null);
		mObject.setObjects(null);
		mObject.setTetrahedrons(null);
		mObject.setStart(0);
		mObject.setGeometryArrays(ga);
		StandardVisualization3D.setCapability(ita);
		/*mo.setGeometry(pa);*/
		mObject.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
		/*branchGroup[0].addChild(mo);*/
		if (view3D.getImageGamuts().indexOf(mObject) < 0)
			view3D.getImageGamuts().add(mObject);
		
		Shape3D shape3D[] = triangleAppearance.getShapes(ita);
		child = new BranchGroup();
		child.setCapability(BranchGroup.ALLOW_DETACH);
		for (int i = 0; i < shape3D.length; i++) {
			child.addChild(shape3D[i]);
		}
		return child;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("gamuttest");
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel.
	 */
	private void createParameterPanel() {
		triangleAppearance = new TriangleAppearance();
		
		xDetail = new JSlider(JSlider.HORIZONTAL, 5, MAX_DETAIL, 20);
		xDetail.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (xDetail.getValue() < xStop.getValue()) {
					xStop.setValue(xDetail.getValue());
				}
				xDetLabel.setText(Integer.toString(xDetail.getValue()));
			}
		});
		yDetail = new JSlider(5, MAX_DETAIL, 20);
		yDetail.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (yDetail.getValue() < yStop.getValue()) {
					yStop.setValue(yDetail.getValue());
				}
				yDetLabel.setText(Integer.toString(yDetail.getValue()));
			}
		});
		zDetail = new JSlider(5, MAX_DETAIL, 20);
		zDetail.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (zDetail.getValue() < zStop.getValue()) {
					zStop.setValue(zDetail.getValue());
				}
				zDetLabel.setText(Integer.toString(zDetail.getValue()));
			}
		});
		
		xStart = new JSlider(0, MAX_DETAIL - 1, 0);
		xStart.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (xStart.getValue() >= xStop.getValue()) {
					xStop.setValue(xStart.getValue() + 1);
				}
				xStartLabel.setText(Integer.toString(xStart.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		yStart = new JSlider(0, MAX_DETAIL - 1, 0);
		yStart.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (yStart.getValue() >= yStop.getValue()) {
					yStop.setValue(yStart.getValue() + 1);
				}
				yStartLabel.setText(Integer.toString(yStart.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		zStart = new JSlider(0, MAX_DETAIL - 1, 0);
		zStart.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (zStart.getValue() >= zStop.getValue()) {
					zStop.setValue(zStart.getValue() + 1);
				}
				zStartLabel.setText(Integer.toString(zStart.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		
		xStop = new JSlider(1, MAX_DETAIL, 20);
		xStop.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (xStop.getValue() <= xStart.getValue()) {
					xStart.setValue(xStop.getValue() - 1);
				}
				if (xStop.getValue() > xDetail.getValue()) {
					xDetail.setValue(xStop.getValue());
				}
				xStopLabel.setText(Integer.toString(xStop.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		yStop = new JSlider(1, MAX_DETAIL, 20);
		yStop.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (yStop.getValue() <= yStart.getValue()) {
					yStart.setValue(yStop.getValue() - 1);
				}
				if (yStop.getValue() > yDetail.getValue()) {
					yDetail.setValue(yStop.getValue());
				}
				yStopLabel.setText(Integer.toString(yStop.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		zStop = new JSlider(1, MAX_DETAIL, 20);
		zStop.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (zStop.getValue() <= zStart.getValue()) {
					zStart.setValue(zStop.getValue() - 1);
				}
				if (zStop.getValue() > zDetail.getValue()) {
					zDetail.setValue(zStop.getValue());
				}
				zStopLabel.setText(Integer.toString(zStop.getValue()));
				if (!((JSlider) ce.getSource()).getValueIsAdjusting()) {
					updateDisplay();
				}
			}
		});
		
		xDetLabel = new JLabel(Integer.toString(xDetail.getValue()));
		yDetLabel = new JLabel(Integer.toString(yDetail.getValue()));
		zDetLabel = new JLabel(Integer.toString(zDetail.getValue()));
		
		xStartLabel = new JLabel(Integer.toString(xStart.getValue()));
		yStartLabel = new JLabel(Integer.toString(yStart.getValue()));
		zStartLabel = new JLabel(Integer.toString(zStart.getValue()));
		
		xStopLabel = new JLabel(Integer.toString(xStop.getValue()));
		yStopLabel = new JLabel(Integer.toString(yStop.getValue()));
		zStopLabel = new JLabel(Integer.toString(zStop.getValue()));
		
		JPanel lblPanel, addPanel;
		
		JPanel gridPanel = new JPanel(new GridLayout(0, 2));
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("X detail"), BorderLayout.WEST);
		lblPanel.add(xDetLabel, BorderLayout.EAST);
		addX = new JCheckBox("", true);
		addX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				updateDisplay();
			}
		});
		addPanel = new JPanel(new BorderLayout());
		addPanel.add(addX, BorderLayout.EAST);
		addPanel.add(new JPanel(), BorderLayout.CENTER);
		lblPanel.add(addPanel, BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(xDetail);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Y detail"), BorderLayout.WEST);
		lblPanel.add(yDetLabel, BorderLayout.EAST);
		addY = new JCheckBox("", true);
		addY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				updateDisplay();
			}
		});
		addPanel = new JPanel(new BorderLayout());
		addPanel.add(addY, BorderLayout.EAST);
		addPanel.add(new JPanel(), BorderLayout.CENTER);
		lblPanel.add(addPanel, BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(yDetail);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Z detail"), BorderLayout.WEST);
		lblPanel.add(zDetLabel, BorderLayout.EAST);
		addZ = new JCheckBox("", true);
		addZ.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				updateDisplay();
			}
		});
		addPanel = new JPanel(new BorderLayout());
		addPanel.add(addZ, BorderLayout.EAST);
		addPanel.add(new JPanel(), BorderLayout.CENTER);
		lblPanel.add(addPanel, BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(zDetail);
		
		
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("X start"), BorderLayout.WEST);
		lblPanel.add(xStartLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(xStart);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Y start"), BorderLayout.WEST);
		lblPanel.add(yStartLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(yStart);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Z start"), BorderLayout.WEST);
		lblPanel.add(zStartLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(zStart);
		
		
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("X stop"), BorderLayout.WEST);
		lblPanel.add(xStopLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(xStop);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Y stop"), BorderLayout.WEST);
		lblPanel.add(yStopLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(yStop);
		
		lblPanel = new JPanel(new BorderLayout());
		lblPanel.add(new JLabel("Z stop"), BorderLayout.WEST);
		lblPanel.add(zStopLabel, BorderLayout.EAST);
		lblPanel.add(new JPanel(), BorderLayout.CENTER);
		gridPanel.add(lblPanel);
		gridPanel.add(zStop);
		
		
		
		JButton reDisplay = new JButton("New shape");
		reDisplay.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				newShapeOnUpdate = true;
			}
		});
		gridPanel.add(new JPanel());
		gridPanel.add(reDisplay);
		
		
		JPanel boxPanel = new JPanel();
		boxPanel.setLayout(new BoxLayout(boxPanel, BoxLayout.Y_AXIS));
		boxPanel.add(gridPanel);
		boxPanel.add(triangleAppearance);
		
		parameterPanel = new JPanel(new BorderLayout());
		parameterPanel.add(boxPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Updates the display of the shape.
	 */
	private void updateDisplay() {
		if (disableUpdate)
			return;
		if (mObject != null && child != null && parent != null) {
			int detX = xStop.getMaximum();
			int detY = yStop.getMaximum();
			int detZ = zStop.getMaximum();
			float oldPos[] = new float[detX*detY*detZ*3];
			mObject.getGeometryArrays()[0].getCoordinates(0, oldPos);
			child.detach();
			child = createChild(oldPos, detX, detY, detZ);
			parent.addChild(child);
		}
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Mapping test";
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 5000;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}