/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.color.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.iccreader.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.*;
import icc3dpackage.utils.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.*;
import icc3dpackage.visualizations.generic.*;


/**
 * This class is used to represent a profile as a Gamut.
 */
public class ProfileGamutModule extends Module implements GamutData {
	
	/** The white point. */
	private float whitePoint[];
	
	/** The filename. */
	private String fileName;
	
	/** The active visualization. */
	private Visualization3D visualization3D = null;
	
	/** The names of the different visualizations. */
	private Vector visualizationNames;
	
	/** different visualizations. */
	private Vector visualizations;
	
	/** The gamut data extracted from the file. */
	private float gamutData[];
	
	/** A Lab color space. */
	private ColorSpace3D labColorSpace3D;
	
	/** The parameter panel. */
	private JPanel parameterPanel = null;
	
	/** The file name label. */
	private JLabel name;
	
	/** The View3D parent. */
	private View3D view3D;
	
	/** The complete filename. */
	private String completeName;
	
	/** The profile path. */
	private static String profilePath;
	
	/** Determines whether an ICC file has been read. */
	private boolean gamutLoaded = false;
	
	/** GUI button */
	private JButton loadGamut, saveMaxima;
	
	/** Use relative colorimetry. */
	private JCheckBox useRelative;
	
	/** The different visualization choices. */
	private JComboBox gamutVisualization;
	
	/** A JSplitPane dividing the parameter panel. */
	private JSplitPane jsp;
	
	/** The number of subdivisions used when determining the gamut data. */
	private int detail = 50;
	
	/** The IccReader used to read the file. */
	private IccReader iccReader;
	
	static {
		profilePath = ".";
	}
	
	/**
	 * Default constructor.
	 */
	public ProfileGamutModule() {}
	
	/**
	 * Creates the parameter panel of this GamutView.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		useRelative = new JCheckBox("Relative colorimetry");
		useRelative.setSelected(true);
		useRelative.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (whitePoint != null && gamutData != null) {
					ColorSpace3D xyzSpace = ColorSpaceManager.getColorSpace("XYZ");
					float xyz[] = labColorSpace3D.toCIEXYZ(gamutData);

					if (useRelative.isSelected()) {
						Colorimetry.toRelativeXYZ_CAT02(xyz, whitePoint[0], whitePoint[1],
								whitePoint[2]);
					} else {
						Colorimetry.toAbsoluteXYZ_CAT02(xyz, whitePoint[0], whitePoint[1],
								whitePoint[2]);
					}
					int indexHigh = 0;

					gamutData = labColorSpace3D.fromCIEXYZ(xyz);
					for (int i = 0; i < gamutData.length / 3; i++) {
						if (gamutData[i * 3 + 1] > gamutData[indexHigh * 3 + 1]) {
							indexHigh = i;
						}
					}
					float wh[] = new float[] {gamutData[indexHigh * 3],
						gamutData[indexHigh * 3 + 1], gamutData[indexHigh * 3 + 2]};

					wh = labColorSpace3D.toPresentation(wh);
					System.out.println("White " + wh[0] + ", " + wh[1] + ", " + wh[2]);
				}
			}
		});

		gamutVisualization = new JComboBox(visualizationNames);
		gamutVisualization.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				String type = (String) gamutVisualization.getSelectedItem();

				visualization3D = (Visualization3D) visualizations.elementAt(visualizationNames.indexOf(type));
				jsp.setBottomComponent(visualization3D.getParameterPanel());
			}
		});		
		name = new JLabel("No gamut loaded");
		name.setHorizontalAlignment(JLabel.CENTER);
		name.setBorder(BorderFactory.createEtchedBorder());
		name.setPreferredSize(new Dimension(150, 25));
		name.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent me) {
				if (!name.getText().equals("No gamut loaded")) {
					if (iccReader != null) {
						try {
							Profile$Information pi = new Profile$Information(iccReader.getInfo());

							pi = null;
						} catch (Exception e) {
							SystemManager.displayError(
								"Error",
								e.getMessage());
						}
					} else {
						try {
							ICC_Profile profile = ICC_Profile.getInstance(completeName);
							ICC_ColorSpace iccCS = new ICC_ColorSpace(profile);
							StringBuffer sb = new StringBuffer();

							sb.append("Nuber of color components: ");
							sb.append(profile.getNumComponents());
							sb.append("\n");
						} catch (Exception e) {
							SystemManager.displayError(
								"Error",
								e.getMessage());
						}
					}
				}
			}
		});
		
		JLabel fgVis = new JLabel("Gamut visualization");

		fgVis.setHorizontalAlignment(JLabel.CENTER);
		fgVis.setBorder(BorderFactory.createEtchedBorder());
		fgVis.setPreferredSize(new Dimension(150, 25));
		loadGamut = new JButton("gTag");
		loadGamut.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				loadProfile(false);
			}
		});
		
		JButton alt = new JButton("Load");

		alt.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				loadProfile(true);
			}
		});
		
		JPanel loadPanel = new JPanel(new GridLayout(0, 2));

		loadPanel.add(alt);
		loadPanel.add(loadGamut);
		saveMaxima = new JButton("Save SM");
		saveMaxima.setEnabled(false);
		saveMaxima.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				SystemManager.displayError("Error", "Not supported by the current version");
				//((SegmentMaximaVisualization3D) visualizations.elementAt(visualizationNames.indexOf("Segment Maxima"))).exportToFile();
			}
		});
		
		JPanel panel1 = new JPanel(new BorderLayout());

		panel1.add(loadPanel, BorderLayout.NORTH);
		
		JPanel panel2 = new JPanel(new BorderLayout());

		panel2.add(name, BorderLayout.NORTH);
		panel1.add(panel2, BorderLayout.SOUTH);
		
		JPanel panel3a = new JPanel(new BorderLayout());

		panel3a.add(useRelative, BorderLayout.NORTH);
		panel2.add(panel3a, BorderLayout.SOUTH);
		
		JPanel panel3 = new JPanel(new BorderLayout());

		panel3.add(saveMaxima, BorderLayout.NORTH);
		panel3a.add(panel3, BorderLayout.SOUTH);
		
		JPanel panel4 = new JPanel(new BorderLayout());

		panel4.add(fgVis, BorderLayout.NORTH);
		panel3.add(panel4, BorderLayout.SOUTH);
		
		JPanel panel5 = new JPanel(new BorderLayout());

		panel5.add(gamutVisualization, BorderLayout.NORTH);
		panel4.add(panel5, BorderLayout.SOUTH);
		
		JPanel upperPanel = new JPanel(new FlowLayout());

		upperPanel.add(panel1);

		jsp.setTopComponent(upperPanel);
		jsp.setBottomComponent(visualization3D.getParameterPanel());
		parameterPanel.add(jsp, BorderLayout.CENTER);
	}
	
	/**
	 * Creates n number of BranchGroups to represent the Gamut in 3D.
	 *
	 * @return array of BranchGroup
	 */
	public BranchGroup[] getSceneBranch() {
		if (gamutLoaded && visualization3D != null) {
			ColorSpace3D colorSpace3D = view3D.getColorSpaceManager().getColorSpace3D();
			float vertex[][] = new float[2][];

			vertex[0] = gamutData;
			vertex[1] = ColorSpaceConverter.convert(gamutData, labColorSpace3D,
					colorSpace3D);
			ColorSpace3D inpOup[] = new ColorSpace3D[2];

			inpOup[0] = labColorSpace3D;
			inpOup[1] = colorSpace3D;
			BranchGroup[] branchGroup = visualization3D.createGamutObject(gamutData.length
					/ 3,
					vertex,
					colorSpace3D.toSRGB(vertex[1]),
					inpOup,
					view3D);

			if (visualization3D.getName().equals("Segment Maxima")) {
				saveMaxima.setEnabled(true);
			}
			return branchGroup;	
		}
		return null;
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;	
	}
	
	/**
	 * Load a profile.
	 * @param useSecondMethod Find data in the gamut lut.
	 */
	private void loadProfile(boolean useSecondMethod) {
		JFileChooser chooser = new JFileChooser(profilePath);

		if (chooser.showOpenDialog(ICC3D.getDialogParent()) == JFileChooser.APPROVE_OPTION && chooser.getSelectedFile().exists()) {
			completeName = chooser.getSelectedFile().getAbsolutePath();
			profilePath = chooser.getSelectedFile().getPath();
			if (profilePath.equals(""))
				profilePath = ".";
			saveMaxima.setEnabled(false);
			if (!useSecondMethod) {
				try {
					iccReader = new IccReader(chooser.getSelectedFile().getAbsolutePath());
					if (iccReader.getGamutReader() == null) {
						iccReader = null;
						throw new RuntimeException("No gamut tag found");
					}
					name.setText("File: " + chooser.getSelectedFile().getName());
					fileName = chooser.getSelectedFile().getName();
					
					int count = 0;
					float out;

					for (int i = 0; i < detail; i++) {
						for (int j = 0; j < detail; j++) {
							for (int k = 0; k < detail; k++) {
								out = iccReader.getOutput(new float[] {i / (float) (detail - 1), j / (float) (detail - 1), k / (float) (detail - 1)})[0];
								if (out == 0.0f) {
									count++;
								}
							}
						}
					}
					gamutData = new float[3 * count];
					int index = 0;

					for (int i = 0; i < detail; i++) {
						for (int j = 0; j < detail; j++) {
							for (int k = 0; k < detail; k++) {
								out = iccReader.getOutput(new float[] {i / (float) (detail - 1), j / (float) (detail - 1), k / (float) (detail - 1)})[0];
								if (out == 0.0f) {
									gamutData[index * 3 + 1] = i / (float) (detail - 1);
									gamutData[index * 3 + 2] = j / (float) (detail - 1);
									gamutData[index * 3] = k / (float) (detail - 1);
									index++;
								}
							}
						}
					}
					gamutLoaded = true;
				} catch (Exception ee) {
					SystemManager.displayError(
						"Error",
						ee.getMessage());
					ee.printStackTrace();
					name.setText("No gamut loaded");
					gamutLoaded = false;
					gamutData = null;
					iccReader = null;
				}
			} else {
				try {
					iccReader = new IccReader(chooser.getSelectedFile().getAbsolutePath());
					whitePoint = iccReader.getWhitePoint();
					ICC_Profile profile = ICC_Profile.getInstance(chooser.getSelectedFile().getAbsolutePath());
					ICC_ColorSpace iccCS = new ICC_ColorSpace(profile);

					name.setText("File: " + chooser.getSelectedFile().getName());
					fileName = chooser.getSelectedFile().getName();
					int det = 10;

					gamutData = new float[(int) (Math.pow((det + 1), profile.getNumComponents()) * 3)];
					float tmpFloat[] = new float[profile.getNumComponents()];
					int inputChannels = profile.getNumComponents();
					int j[] = new int[inputChannels];
					int curChannel = inputChannels - 1;

					while (curChannel >= 0) {
						int h = 0;

						for (int u = 0; u < inputChannels; u++) {
							h *= (det + 1);
							h += j[u];
						}
						curChannel = inputChannels - 1;
						j[curChannel]++;
						while (curChannel >= 0 && j[curChannel] >= (det + 1)) {
							j[curChannel] = 0;
							curChannel--;
							if (curChannel >= 0) {
								j[curChannel]++;
							}
						}
						for (int cp = 0; cp < inputChannels; cp++) {
							tmpFloat[cp] = j[cp];
							tmpFloat[cp] /= (float) det;
						}
						float tmpFloat2[] = iccCS.toCIEXYZ(tmpFloat);

						gamutData[h * 3] = tmpFloat2[0];
						gamutData[h * 3 + 1] = tmpFloat2[1];
						gamutData[h * 3 + 2] = tmpFloat2[2];
					}
					gamutData = labColorSpace3D.fromCIEXYZ(gamutData);
					if (!useRelative.isSelected()) {
						float xyz[] = labColorSpace3D.toCIEXYZ(gamutData);

						Colorimetry.toAbsoluteXYZ_CAT02(xyz, whitePoint[0], whitePoint[1],
								whitePoint[2]);
						gamutData = labColorSpace3D.fromCIEXYZ(xyz);
					}
					gamutLoaded = true;					
				} catch (Exception ee) {
					SystemManager.displayError(
						"Error",
						ee.getMessage());
					ee.printStackTrace();
					name.setText("No gamut loaded");
					gamutLoaded = false;
					gamutData = null;
					iccReader = null;
				}
			}
		}
	}
	
	/**
	 * Returns the gamut point data.
	 * @return A float array containing Lab data.
	 */
	public float[] getData() {
		return this.gamutData;
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 100;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "ICC profile gamut";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("icc");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.completeName = null;
		this.gamutData = null;
		this.gamutVisualization = null;
		this.iccReader = null;
		this.jsp = null;
		this.labColorSpace3D = null;
		this.loadGamut = null;
		this.name = null;
		this.parameterPanel = null;
		this.profilePath = null;
		this.saveMaxima = null;
		this.view3D = null;
		this.visualization3D = null;
		this.visualizationNames = null;
		Visualization3D.deInitialize(visualizations);
		this.visualizations = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		labColorSpace3D = ColorSpaceManager.getColorSpace("Lab");
		visualizationNames = new Vector();
		visualizations = new Vector();
		String names[] = VisualizationManager.getNames();

		for (int i = 0; i < names.length; i++) {
			visualizationNames.add(names[i]);
			visualizations.add(VisualizationManager.getVisualization(names[i]));
		}
		visualization3D = (Visualization3D) visualizations.elementAt(0);
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		if (visualization3D == null) {
			return true;
		}
		return visualization3D.checkParams();
	}
	
	/**
	 * Returns the filename.
	 * @return The filename.
	 */
	public String getFileName() {
		return fileName;
	}
}


/**
 * A frame that displays information about a profile.
 */
class Profile$Information extends JFrame {
	
	/** The profile text */
	private JTextArea profileText;
	
	/**
	 * Create and show information about a profile.
	 * @param info Information about a profile.
	 */
	public Profile$Information(String info) {
		super("ICC Profile Information");
		Container c = getContentPane();

		c.setLayout(new BorderLayout());
		profileText = new JTextArea(25, 80);
		profileText.setFont(new Font("Monospaced", Font.PLAIN, 12));
		profileText.setText(info);
		profileText.setCaretPosition(0);
		c.add(new JScrollPane(profileText), BorderLayout.CENTER);
		pack();
		setVisible(true);
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
	}
}
