/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.BevelBorder;
import java.beans.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.modules.regression.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.generic.Visualization3D;


/**
 * This module adds the capability of performing polynomial regression.
 */
public class RegressionModule extends Module {
	
	/** The parameter panel that belongs to this color space manager */
	private JPanel parameterPanel;
	
	/** Select data */
	private JComboBox dataSelection;
	
	/** The regression frame */
	private RegressionFrame regressionFrame;
	
	/** The View3D parent */
	private View3D view3D;
	
	/** The active visualization */
	private Visualization3D visualization3D = null;
	
	/** The names of the available visualizations */
	private Vector visualizationNames;
	
	/** The available visualizations */
	private Vector visualizations;
	
	/** A CIEXYZ color space */
	private ColorSpace3D colorSpace3D_XYZ;
	
	/** Choice of data to be visualized */
	private JComboBox regressionVisualization;
	
	/** A split pane */
	private JSplitPane jsp;
	
	/**
	 * Default constructor.
	 */
	public RegressionModule() {}
	
	/**
	 * Constructs the parameter panel for this polynomial regression module.
	 */
	private void createParameterPanel() {
		
		JButton regressionButton = new JButton("Regression");

		regressionButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (regressionFrame == null) {
					regressionFrame = new RegressionFrame();
				}
				if (!regressionFrame.dataGood()) {
					regressionFrame = null;
				} else {
					if (!regressionFrame.dataGood()) {
						regressionFrame.dispose();
						regressionFrame = new RegressionFrame();
						if (!regressionFrame.dataGood()) {
							regressionFrame = null;
						}
					}
				}
				if (regressionFrame != null && regressionFrame.dataGood()) {
					regressionFrame.show();
				}
			}
		});
		dataSelection = new JComboBox(new String[] {"File data", "Regression data" /*,"debug"*/});
		regressionVisualization = new JComboBox(visualizationNames);
		regressionVisualization.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				String type = (String) regressionVisualization.getSelectedItem();

				visualization3D = (Visualization3D) visualizations.elementAt(visualizationNames.indexOf(type));
				jsp.setBottomComponent(visualization3D.getParameterPanel());
			}
		});
		
		JPanel gridPanel = new JPanel(new GridLayout(0, 1));

		gridPanel.add(regressionButton);
		gridPanel.add(dataSelection);
		gridPanel.add(regressionVisualization);
		
		JPanel flowPanel = new JPanel(new FlowLayout());

		flowPanel.add(gridPanel);
		
		jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setEnabled(false);
		
		jsp.setTopComponent(flowPanel);
		jsp.setBottomComponent(visualization3D.getParameterPanel());
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new BorderLayout());
		parameterPanel.add(jsp, BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 4000;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Regression";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("polreg");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		if (regressionFrame != null) {
			regressionFrame.dispose();
			regressionFrame = null;
		}
		this.colorSpace3D_XYZ = null;
		this.dataSelection = null;
		this.jsp = null;
		this.parameterPanel = null;
		this.regressionVisualization = null;
		this.view3D = null;
		this.visualization3D = null;
		this.visualizationNames = null;
		Visualization3D.deInitialize(visualizations);
		this.visualizations = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		colorSpace3D_XYZ = ColorSpaceManager.getColorSpace("XYZ");
		visualizationNames = new Vector();
		visualizations = new Vector();
		String names[] = VisualizationManager.getNames();

		for (int i = 0; i < names.length; i++) {
			visualizationNames.add(names[i]);
			visualizations.add(VisualizationManager.getVisualization(names[i]));
		}
		visualization3D = (Visualization3D) visualizations.elementAt(0);
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public BranchGroup[] getSceneBranch() {
		if (regressionFrame == null) {
			return null;
		}
		float xyzData[] = null;

		if (dataSelection.getSelectedIndex() == 0) {
			xyzData = regressionFrame.getXYZ_DataFromFile();
		} else if (dataSelection.getSelectedIndex() == 1) {
			xyzData = regressionFrame.getXYZ_Data();
		} else {
			xyzData = regressionFrame.getXYZ_Data2();
		}
		if (xyzData == null || xyzData.length < 5 * 3) {
			return null;
		}
		ColorSpace3D colorSpaces[] = new ColorSpace3D[2];

		colorSpaces[0] = colorSpace3D_XYZ;
		colorSpaces[1] = view3D.getColorSpaceManager().getColorSpace3D();
		float pointCoord[][] = new float[2][];

		pointCoord[0] = xyzData;
		pointCoord[1] = colorSpaces[1].fromCIEXYZ(xyzData);
		byte color[] = colorSpaces[0].toSRGB(xyzData);

		return visualization3D.createGamutObject(xyzData.length / 3, pointCoord, color,
				colorSpaces, view3D);
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
