/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.media.j3d.*;
import javax.swing.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.iccreader.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.parser.Parser2;
import icc3dpackage.utils.*;
import icc3dpackage.view3d.*;
import icc3dpackage.visualizations.*;
import icc3dpackage.visualizations.generic.*;


/**
 * This class is used to represent the surface of
 * a Gamut based on data from a file.
 */
public class SurfaceGamutModule extends Module {
	
	/** The color data read from file. */
	private float colorData[][][][];
	
	/** The number of triangles used to represent this gamut. */
	private int triangleCount;
	
	/** Determines whether a gamut has been read from file. */
	private boolean gamutLoaded;
	
	/** A Lab color space. */
	private ColorSpace3D labColorSpace3D;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The different visualization choices. */
	private JComboBox chooseVisualization;
	
	/** The View3D parent. */
	private View3D view3D;
	
	/**
	 * Default constructor.
	 */
	public SurfaceGamutModule() {}
	
	/**
	 * Creates n number of BranchGroups to represent the Gamut in 3D.
	 *
	 * @return array of BranchGroup
	 */
	public BranchGroup[] getSceneBranch() {
		if (!gamutLoaded) {
			return new BranchGroup[] {new BranchGroup()};
		}
		BranchGroup branchGroup[] = new BranchGroup[1];

		branchGroup[0] = new BranchGroup();
		TriangleArray ta = new TriangleArray(triangleCount * 3,
				GeometryArray.COORDINATES | GeometryArray.COLOR_3);
		ColorSpace3D colorSpace3D = view3D.getColorSpaceManager().getColorSpace3D();
		int index = 0;

		for (int i = 0; i < colorData.length; i++) {
			int n1 = colorData[i].length, n2 = colorData[i][0].length;

			for (int j = 0; j < n1 - 1; j++) {
				for (int k = 0; k < n2 - 1; k++) {
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j][k], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j][k]));
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j][k + 1], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j][k + 1]));
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j + 1][k], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j + 1][k]));
					
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j + 1][k + 1], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j + 1][k + 1]));
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j + 1][k], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j + 1][k]));
					ta.setCoordinate(index,
							ColorSpaceConverter.convert(colorData[i][j][k + 1], labColorSpace3D,
							colorSpace3D));
					ta.setColor(index++, labColorSpace3D.toSRGB(colorData[i][j][k + 1]));
				}
			}
		}
				
		Appearance appearance = new Appearance();

		if (chooseVisualization.getSelectedIndex() == 0) {
			appearance.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_LINE,
					PolygonAttributes.CULL_NONE, 0.0f));
		} else if (chooseVisualization.getSelectedIndex() == 1) {
			appearance.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_FILL,
					PolygonAttributes.CULL_NONE, 0.0f));
		} else if (chooseVisualization.getSelectedIndex() == 2) {
			appearance.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_FILL,
					PolygonAttributes.CULL_NONE, 0.0f));
			Shape3D shape3D2 = new Shape3D();

			shape3D2.setGeometry(ta);
			Appearance app2 = new Appearance();

			app2.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_LINE,
					PolygonAttributes.CULL_NONE, 0.0f));
			app2.setColoringAttributes(new ColoringAttributes(0.0f, 0.0f, 0.0f,
					ColoringAttributes.SHADE_FLAT));
			app2.setRenderingAttributes(new RenderingAttributes(true, true, 1.0f,
					RenderingAttributes.LESS_OR_EQUAL));
			app2.getRenderingAttributes().setIgnoreVertexColors(true);
			shape3D2.setAppearance(app2);
			
			Transform3D t3d = new Transform3D();

			t3d.setTranslation(new Vector3d(-0.5, -0.5, -0.5));
			TransformGroup translateTransform = new TransformGroup(t3d);

			translateTransform.addChild(shape3D2);
			t3d = new Transform3D();
			t3d.setScale(1.01);
			TransformGroup scaleTransform = new TransformGroup(t3d);

			scaleTransform.addChild(translateTransform);
			t3d = new Transform3D();
			t3d.setTranslation(new Vector3d(0.5, 0.5, 0.5));
			TransformGroup reverseTranslation = new TransformGroup(t3d);

			reverseTranslation.addChild(scaleTransform);
			branchGroup[0].addChild(reverseTranslation);
			
		}
		Shape3D shape3D = new Shape3D();

		shape3D.setGeometry(ta);
		shape3D.setAppearance(appearance);
		branchGroup[0].addChild(shape3D);
		return branchGroup;
	}
	
	/**
	 * Creates the parameter panel of this GamutView.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new BorderLayout());
		JPanel upperPanel = new JPanel(new FlowLayout());

		chooseVisualization = new JComboBox(new String[] {"Wireframe", "Solid",
			"Solid+Wireframe"});
		JPanel panel1 = new JPanel(new BorderLayout());
		final JLabel fileName = new JLabel("No gamut loaded");
		JButton openFile = new JButton("Load gamut");

		openFile.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser chooser = new JFileChooser(".");

				if (chooser.showOpenDialog(ICC3D.getDialogParent()) != JFileChooser.CANCEL_OPTION) {
					try {
						fileName.setText(chooser.getSelectedFile().getName());
						Parser2 parser = new Parser2(chooser.getSelectedFile().getAbsolutePath());
						float lab[] = parser.getLabData();
						int rgb[] = parser.getRGBData();
						int numR, numG, numB;

						for (numB = 1; numB < (rgb.length / 3); numB++) {
							if (rgb[numB * 3 + 1] != rgb[(numB - 1) * 3 + 1]) {
								break;
							}
						}
						for (numG = 1; numG < (rgb.length / 3); numG++) {
							if (rgb[numG * 3] != rgb[(numG - 1) * 3]) {
								break;
							}
						}
						numG /= numB;
						numR = (rgb.length / 3) / (numG * numB);
						int rVariations[] = new int[numR], gVariations[] = new int[numG],
								bVariations[] = new int[numB];
						int rIndex = 0, gIndex = 0, bIndex = 0;

						rVariations[rIndex++] = rgb[0];
						gVariations[gIndex++] = rgb[1];
						bVariations[bIndex++] = rgb[2];
						int y = 1;

						while (rIndex < numR) {
							if (rgb[y * 3] != rgb[(y - 1) * 3]) {
								rVariations[rIndex++] = rgb[y * 3];
							}
							y++;
						}
						y = 1;
						while (gIndex < numG) {
							if (rgb[y * 3 + 1] != rgb[(y - 1) * 3 + 1]) {
								gVariations[gIndex++] = rgb[y * 3 + 1];
							}
							y++;
						}
						y = 1;
						while (bIndex < numB) {
							if (rgb[y * 3 + 2] != rgb[(y - 1) * 3 + 2]) {
								bVariations[bIndex++] = rgb[y * 3 + 2];
							}
							y++;
						}
						colorData = new float[6][][][];
						triangleCount = 0;
						populateSide(colorData, 0, 0, rVariations[0], lab, rgb, numG, numB);
						populateSide(colorData, 1, 0, rVariations[numR - 1], lab, rgb, numG, numB);
						populateSide(colorData, 2, 1, gVariations[0], lab, rgb, numR, numB);
						populateSide(colorData, 3, 1, gVariations[numG - 1], lab, rgb, numR, numB);
						populateSide(colorData, 4, 2, bVariations[0], lab, rgb, numR, numG);
						populateSide(colorData, 5, 2, bVariations[numB - 1], lab, rgb, numR, numG);
						gamutLoaded = true;
					} catch (Exception e) {
						fileName.setText("No gamut loaded");
						colorData = null;
						gamutLoaded = false;
					}
				}
			}
		});
		panel1.add(openFile, BorderLayout.NORTH);
		fileName.setHorizontalAlignment(JLabel.CENTER);
		fileName.setBorder(BorderFactory.createEtchedBorder());
		fileName.setPreferredSize(new Dimension(150, 25));
		JPanel panel2 = new JPanel(new BorderLayout());

		panel1.add(panel2, BorderLayout.SOUTH);
		panel2.add(fileName, BorderLayout.NORTH);
		
		JPanel panel10 = new JPanel(new BorderLayout());

		panel2.add(panel10, BorderLayout.SOUTH);
		JButton nothingButton = new JButton("");
		JButton nothingButton2 = new JButton("Save SM");

		nothingButton.setEnabled(false);
		nothingButton.setPreferredSize(new Dimension((int) nothingButton.getPreferredSize().getWidth(),
				(int) nothingButton2.getPreferredSize().getHeight()));
		panel10.add(nothingButton, BorderLayout.NORTH);
		
		JLabel fgVis = new JLabel("Gamut visualization");

		fgVis.setHorizontalAlignment(JLabel.CENTER);
		fgVis.setBorder(BorderFactory.createEtchedBorder());
		fgVis.setPreferredSize(new Dimension(150, 25));
		JPanel panel3 = new JPanel(new BorderLayout());

		panel10.add(panel3, BorderLayout.SOUTH);
		panel3.add(fgVis, BorderLayout.NORTH);
		JPanel panel4 = new JPanel(new BorderLayout());

		panel3.add(panel4, BorderLayout.SOUTH);
		panel4.add(chooseVisualization, BorderLayout.NORTH);
		upperPanel.add(panel1);
		JSplitPane jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);

		jsp.setBorder(BorderFactory.createEmptyBorder());
		jsp.setTopComponent(upperPanel);
		jsp.setBottomComponent(new JPanel());
		parameterPanel.add(jsp, BorderLayout.CENTER);
	}
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Finds all points on a given plane(x,y,z) that satisifies that either 
	 * R, G or B are fixed.
	 * For example: Find for plane x all points that lies on this plane which 
	 * has a value of R = a, G=?, B=?.
	 * The parameters cannot be understood by anyone but the individual who
	 * wrote the functions,
	 * but rest assured that this is a very effective function.
	 * @param matrix
	 * @param matrixIndex
	 * @param fixedColumn
	 * @param fixedValue
	 * @param labData
	 * @param rgbData
	 * @param num1
	 * @param num2
	 */
	private void populateSide(float matrix[][][][], int matrixIndex, int fixedColumn, int fixedValue, float labData[], int rgbData[], int num1, int num2) {
		triangleCount += 2 * (num1 - 1) * (num2 - 1);
		matrix[matrixIndex] = new float[num1][num2][3];
		int index = 0, i1 = 0, i2 = 0;

		for (int i = 0; i < labData.length / 3; i++) {
			if (rgbData[i * 3 + fixedColumn] == fixedValue) {
				matrix[matrixIndex][i1][i2][0] = 0.5f + labData[i * 3 + 2] / 256.0f;
				matrix[matrixIndex][i1][i2][1] = labData[i * 3] / 100.0f;
				matrix[matrixIndex][i1][i2][2] = 0.5f + labData[i * 3 + 1] / 256.0f;
				i2++;
				if (i2 >= num2) {
					i1++;
					i2 = 0;
				}
			}
		}
	}
	
	/**
	 * Returns the gamut point data.
	 * @return A float array containing Lab data.
	 */
	public float[] getData() {
		return new float[] {};
	}
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public int getPriority() {
		return 100;
	}
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public String getName() {
		return "Surface device gamut";
	}
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public boolean existsAsDefault() {
		return false;
	}
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public boolean userAdd() {
		return true;
	}
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public boolean userRemove() {
		return true;
	}
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public Icon getIcon() {
		return IconManager.getIcon("gamut_vis");
	}
	
	/**
	 * Deinitializes the module.
	 */
	public void deInitialize() {
		this.chooseVisualization = null;
		this.colorData = null;
		this.labColorSpace3D = null;
		this.parameterPanel = null;
		this.view3D = null;
	}
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public boolean initialize(View3D view3D) {
		labColorSpace3D = ColorSpaceManager.getColorSpace("Lab");
		this.view3D = view3D;
		return true;
	}
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public boolean readyForUpdate() {
		return true;
	}
}
