/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules.capture;


import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.*;
import javax.media.*;
import javax.media.control.*;
import javax.media.datasink.*;
import javax.media.format.VideoFormat;
import javax.media.format.RGBFormat;
import javax.media.protocol.*;


/**
 * Creates an source stream/buffer for the movie encoder
 */
public class ImageSourceStream implements PullBufferStream {
	
	/** A list of images waiting to be processed */
	private LinkedList imageQueue;
	
	/** The image width */
	private int width;
	
	/** The image height */
	private int height;
	
	/** The video format */
	private VideoFormat format;
	
	/** The content descriptor */
	private ContentDescriptor contentDescriptor;
	
	/** Determines whether the movie maker is finished */
	private boolean ended;
	
	/** Determines whether there is more data available */
	private boolean moreData;
	
	/** The sequence number of the current frame */
	private int seqNo;
	
	/** The image data */
	private int data[];
	
	/** The max data length */
	private int maxDataLength;
	
	/** The frame rate */
	private float frameRate;
	
	/** The maximum image list size */
	private int maxListSize;

	/**
	 *	Constructor that initializes the object	
	 */
	public ImageSourceStream() {}
	
	/**
	 * Prepares the stream by setting the size and framerate
	 * @param width the width
	 * @param height the height
	 * @param frameRate the framerate
	 */
	public synchronized void prepare(int width, int height, float frameRate) {
		this.width = width;
		this.height = height;
		this.frameRate = frameRate;
		ended = false;
		moreData = true;
		seqNo = 0;
		maxListSize = 4;
		maxDataLength = width * height * 3;
		imageQueue = new LinkedList();
		data = new int[maxDataLength];
		format = new RGBFormat(new Dimension(width, height), maxDataLength,
				Format.intArray, frameRate, 32, 0xFF0000, 0xFF00, 0xFF, 1, width,
				VideoFormat.FALSE, Format.NOT_SPECIFIED);
		contentDescriptor = new ContentDescriptor(ContentDescriptor.RAW);
	}

	/**
	 * Returns if there is more data to put in the queue.
	 * @return true if there is more data and the queue is empty, false if not or the queue is not empty
	 */
	public synchronized boolean willReadBlock() {
		if (imageQueue.size() == 0 && moreData) {
			return true;
		}
		return false;
	}

	/**
	 * Extracts data from the queue and puts it into the buffer. If there is noe more data it ends the stream of data (naturally)
	 * @param buffer the buffer you put data into
	 * @throws IOException
	 */
	public synchronized void read(Buffer buffer) throws IOException {
		while (moreData && imageQueue.size() == 0) {
			try {
				this.wait();
			} catch (InterruptedException ie) {}
		}
			
		if (!moreData && imageQueue.size() == 0) {
			buffer.setEOM(true);
			buffer.setOffset(0);
			buffer.setLength(0);
			ended = true;
			return;
		}

		BufferedImage bi = (BufferedImage) imageQueue.removeFirst();

		this.notifyAll();
		
		Object outdata = buffer.getData();

		if (outdata == null || !(outdata.getClass() == Format.intArray)
				|| ((int[]) outdata).length < maxDataLength) {
			outdata = new int[maxDataLength];
			buffer.setData(outdata);
		}
		buffer.setFormat(format);
		buffer.setTimeStamp((long) (seqNo * (1000 / frameRate) * 1000000));
		bi.getRGB(0, 0, width, height, (int[]) outdata, 0, width);
		buffer.setSequenceNumber(seqNo);
		buffer.setLength(maxDataLength);
		buffer.setFlags(Buffer.FLAG_KEY_FRAME);
		buffer.setHeader(null);
		seqNo++;
	}

	/**
	 * Returns the format
	 * @return Format the format
	 */
	public synchronized Format getFormat() {
		return format;
	}

	/**
	 * Returns the content descriptor
	 * @return ContentDescriptor the content descriptor
	 */
	public synchronized ContentDescriptor getContentDescriptor() {
		return contentDescriptor;
	}
	
	/**
	 * Returns the content length
	 * @return long the content length
	 */
	public synchronized long getContentLength() {
		return LENGTH_UNKNOWN;
	}

	/**
	 * Returns true if it has reached the end of the stream
	 * @return boolean true if it is at the end, false if not
	 */
	public synchronized boolean endOfStream() {
		return ended;
	}

	/**
	 * Returns the controls
	 * @return Object[] the controls, in this case always a new Object[0]
	 */
	public synchronized Object[] getControls() {
		return new Object[0];
	}

	/**
	 * Returns the control given the type
	 * @return Object the control
	 */
	public synchronized Object getControl(String controlType) {
		try {
			Class  cls = Class.forName(controlType);
			Object cs[] = getControls();

			for (int i = 0; i < cs.length; i++) {
				if (cls.isInstance(cs[i])) {
					return cs[i];
				}
			}
			return null;
		} catch (Exception e) {
			return null;
		}
	}
	
	/**
	 * Sets the signal no more data (end of input data) and notifies all threads
	 */
	public synchronized void signalNoMoreData() {
		moreData = false;
		this.notifyAll();
	}
	
	/**
	 * Adds an image to the end of the queue
	 * @param bi the new buffered image
	 */
	public synchronized void addImage(BufferedImage bi) {
		while (imageQueue.size() >= maxListSize) {
			try {
				this.wait();
			} catch (InterruptedException ie) {}
		}
		imageQueue.addLast(bi);
		this.notifyAll();
	}
}

