/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules.capture;


import java.awt.image.*;
import javax.media.jai.*;
import javax.media.j3d.*;
import javax.vecmath.*;


/**
 * Movie object that includes information about frames, rotations and transforms that make up a sequence.
 */
public class MovieInterpolator {
	
	/** The number of frames */
	private int nFrames;
	
	/** A float defining the angle of rotation */
	private float xRot, yRot, zRot;
	
	/** A float defining a translation */
	private float xTrans, yTrans, zTrans;

	/**
	 * Constructor that initializes the object
	 */
	public MovieInterpolator() {}	

	/**
	 * Function that performes the transform and rotation this object is set to do on an transform group
	 * @param canvas Canvas3D object to do the movement on
	 * @param offscreenBuffer The ImageComponent2D that is associated with the off screen canvas.
	 * @param tiledImage The tiled image that should be used for drawing
	 * @param movieMaker MovieMaker object that shall make the movie
	 * @param tfg TransformGroup that you use to make the movement
	 */
	public void runMe(Canvas3D canvas, ImageComponent2D offscreenBuffer, TiledImage tiledImage, MovieMaker movieMaker, TransformGroup tfg) {
		Transform3D t3d = new Transform3D();
		Transform3D t3d2 = new Transform3D();

		tfg.getTransform(t3d);	
		
		float xFor = xRot / nFrames;
		float yFor = yRot / nFrames;
		float zFor = zRot / nFrames;
		
		float xTraFor = xTrans / nFrames;
		float yTraFor = yTrans / nFrames;
		float zTraFor = zTrans / nFrames;
		
		Vector3d v3dStart = new Vector3d();
		double fStart[] = new double[3];

		t3d.get(v3dStart);
		v3dStart.get(fStart);
		
		for (int i = 0; i < nFrames; i++) {
			t3d2.rotX(xFor);
			t3d.mul(t3d2);
			t3d2.rotY(yFor);
			t3d.mul(t3d2);
			t3d2.rotZ(zFor);
			t3d.mul(t3d2);
			fStart[0] += xTraFor;
			fStart[1] += yTraFor;
			fStart[2] += zTraFor;
			t3d.setTranslation(new Vector3d(fStart));
			tfg.setTransform(t3d);
			try {
				Thread.currentThread().yield();
				Thread.currentThread().sleep(50);
			} catch (InterruptedException ie) {}
			canvas.renderOffScreenBuffer();
			canvas.waitForOffScreenRendering();
			RenderedImage ri = offscreenBuffer.getRenderedImage();

			tiledImage.setData(ri.getData());
			movieMaker.addImage(tiledImage.getAsBufferedImage());
		}	
	}
	
	/**
	 * Sets the number of frames to use on this sequence
	 * @param frames number of frames
	 */
	public void setFrames(int frames) {
		nFrames = frames;
	}
		
	/**
	 * Sets the rotation around the x-axis
	 * @param xrot x-roatition in radians
	 */
	public void setXRot(float xrot) {
		xRot = (float) Math.toRadians(xrot);
	}	
	
	/**
	 * Sets the rotation around the y-axis
	 * @param yrot y-rotation in radians
	 */
	public void setYRot(float yrot) {
		yRot = (float) Math.toRadians(yrot);	
	}
		
	/**
	 * Sets the rotation around the y-axis
	 * @param zrot z-rotation in radians
	 */
	public void setZRot(float zrot) {
		zRot = (float) Math.toRadians(zrot);
	}	
	
	/**
	 * Sets the translation on the x-axis
	 * @param xtrans x-translation value
	 */
	public void setXTrans(float xtrans) {
		xTrans = xtrans;
	}	

	/**
	 * Sets the translation on the y-axis	
	 * @param ytrans y-translation value
	 */
	public void setYTrans(float ytrans) {
		yTrans = ytrans;	
	}
	
	/**
	 * Sets the translation on the z-axis	
	 * @param ztrans z-translation value
	 */	
	public void setZTrans(float ztrans) {
		zTrans = ztrans;	
	}	
}
