/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules.capture;


import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.media.*;
import javax.media.control.*;
import javax.media.datasink.*;
import javax.media.format.*;
import javax.media.protocol.*;
import javax.swing.*;


/**
 * Class that makes a movie out of the given data from the Image3D animation
 */
public class MovieMaker implements ControllerListener, DataSinkListener, Runnable {
	
	/** The image data source */
	private ImageDataSource ids;
	
	/** The data processor */
	private Processor p;
	
	/** The data sink */
	private DataSink dsink;
	
	/** Determines whether the movie maker is busy converting and compressing images */
	private boolean busy;
	
	/** The media locator describing the destination data sink */
	private MediaLocator oml;
	
	/**
	 * Constructor that initializes the object 
	 */
	public MovieMaker() {
		busy = false;
		ids = new ImageDataSource();
		try {
			p = Manager.createProcessor(ids);
		} catch (Exception e) {
			System.err.println("Cannot create a processor from the data source.");
		}
	}
	
	/**
	 * Prepares the stream by setting the size and framerate
	 * @param width the width
	 * @param height the height
	 * @param frameRate the framerate
	 */
	public synchronized void prepare(int width, int height, float frameRate) {
		while (busy) {
			try {
				this.wait();
			} catch (InterruptedException ie) {}
		}
		ids.prepare(width, height, frameRate);
		try {
			p = Manager.createProcessor(ids);
		} catch (Exception e) {
			System.err.println("Cannot create a processor from the data source.");
		}
		p.addControllerListener(this);
		p.configure();
		if (!waitForState(p, p.Configured)) {
			System.err.println("Failed to configure the processor.");
			busy = false;
			this.notifyAll();
		}

	}
	
	/**
	 * Main thread function that starts other prosess in the creation prosess, never call this function !
	 */
	public synchronized void run() {
		System.err.println("start processing...");

		try {
			System.err.println("Starting processor...");
			p.start();
			System.err.println("Starting data sink...");
			dsink.start();
		} catch (IOException e) {
			System.err.println("IO error during processing");
			busy = false;
			this.notifyAll();
			return;
		}
		System.err.println("Waiting for file to be written");
		waitForFileDone();

		System.err.println("Closing data sink");
		try {
			dsink.close();
		} catch (Exception e) {}
		System.err.println("Removing controller listener");
		p.removeControllerListener(this);

		System.err.println("...done processing.");
		busy = false;
		this.notifyAll();
	}
	
	/**
	 * Starts making the movie.
	 */
	public synchronized boolean doIt() {
		while (busy) {
			try {
				this.wait();
			} catch (InterruptedException ie) {}
		}
		busy = true;
		// p.setContentDescriptor(new FileTypeDescriptor(FileTypeDescriptor.MSVIDEO));
		/*
		 TrackControl tcs[] = p.getTrackControls();
		 Format f[] = tcs[0].getSupportedFormats();
		 if (f == null || f.length <= 0) {
		 System.err.println("The mux does not support the input format: " + tcs[0].getFormat());
		 busy = false;
		 this.notifyAll();
		 return false;
		 }

		 System.err.println("Setting the track format to: " + f[0]);
		 tcs[0].setFormat(f[0]);
		 System.out.println("Framerate: "+((VideoFormat)tcs[0].getFormat()).getFrameRate());
		 */
		p.realize();
		if (!waitForState(p, p.Realized)) {
			p.deallocate();
			System.err.println("Failed to realize the processor.");
			busy = false;
			this.notifyAll();
			return false;
		}

		if ((dsink = createDataSink(oml)) == null) {
			System.err.println("Failed to create a DataSink for the given output MediaLocator: "
					+ oml);
			busy = false;
			this.notifyAll();
			return false;
		}

		dsink.addDataSinkListener(this);
		fileDone = false;

		// System.out.println("Starting process thread");
		(new Thread(this)).start();
		// System.out.println("Returning");
		return true;
	}

	/**
	 * Creates a data sink from an given MediaLocator and returns it if it went well, if not it returns null
	 * @param outML the MediaLocator
	 * @return The created data sink
	 */
	DataSink createDataSink(MediaLocator outML) {
		DataSource ds;

		if ((ds = p.getDataOutput()) == null) {
			System.err.println("Something is really wrong: the processor does not have an output DataSource");
			return null;
		}

		DataSink dsink;

		try {
			System.err.println("- create DataSink for: " + outML);
			dsink = Manager.createDataSink(ds, outML);
			dsink.open();
		} catch (Exception e) {
			System.err.println("Cannot create the DataSink: " + e);
			return null;
		}
		return dsink;
	}

	/** Makes it possible to wait for a prosess untill its done */
	Object waitSync = new Object();

	/** True if stateTransition is ok, false if not */
	boolean stateTransitionOK = true;

	/**
	 *	Puts the prosess on hold untill it reaches the wanted state
	 * @param p the prosess
	 * @param state the wanted state
	 */
	boolean waitForState(Processor p, int state) {
		synchronized (waitSync) {
			try {
				while (p.getState() < state && stateTransitionOK) {
					waitSync.wait();
				}
			} catch (Exception e) {}
		}
		return stateTransitionOK;
	}

	/**
	 * Updates the controller given the controller event
	 * if evt is one of the following : ConfigureCompleteEvent, RealizeCompleteEvent or refetchCompleteEvent
	 * It waits and sets stateTransitionOk to true
	 * or if evt is ResourceUnavailableEvent, it sets stateTransitionOk to false and notifies all threads
	 * or if evt is EndOfMediaEvent it stops and closes up (leaves, and puts a gelly fish in his ear...arthur, where are you ???)
	 * @param evt type of controller event
	 */
	public void controllerUpdate(ControllerEvent evt) {
		if (evt instanceof ConfigureCompleteEvent
				|| evt instanceof RealizeCompleteEvent
				|| evt instanceof PrefetchCompleteEvent) {
			synchronized (waitSync) {
				stateTransitionOK = true;
				waitSync.notifyAll();
			}
		} else if (evt instanceof ResourceUnavailableEvent) {
			synchronized (waitSync) {
				stateTransitionOK = false;
				waitSync.notifyAll();
			}
		} else if (evt instanceof EndOfMediaEvent) {
			evt.getSourceController().stop();
			evt.getSourceController().close();
		}
	}

	/** Makes it possible to wait for the file untill its done */
	public Object waitFileSync = new Object();

	/** True if it is done, false if not */
	public boolean fileDone = false;

	/** True if it went well, false if not */
	public boolean fileSuccess = true;

	/**
	 * Sets the thread on hold untill the movie is done 
	 */
	boolean waitForFileDone() {
		synchronized (waitFileSync) {
			try {
				while (!fileDone) {
					waitFileSync.wait();
				}
			} catch (Exception e) {}
		}
		return fileSuccess;
	}

	/**
	 * Updates the data sink, given the event
	 * @param evt either EndOfStremEvent (the end) or DataSinkErrorEvent (something went wrong)
	 */
	public void dataSinkUpdate(DataSinkEvent evt) {
		if (evt instanceof EndOfStreamEvent) {
			synchronized (waitFileSync) {
				fileDone = true;
				waitFileSync.notifyAll();
			}
		} else if (evt instanceof DataSinkErrorEvent) {
			synchronized (waitFileSync) {
				fileDone = true;
				fileSuccess = false;
				waitFileSync.notifyAll();
			}
		}
	}

	/**
	 * Sets the filename, and returns if it went well
	 * @param fname the filename
	 * @return boolean true if it went well, false if not
	 */
	public boolean setFileName(String fname) {
		if ((oml = createMediaLocator(fname)) == null) { 
			return false;
		}
		return true;	
	}

	/**
	 * Still here for test purposes.
	 * @param args Never mind.
	 */
	public static void main(String args[]) {
		int width = 1024;
		int height = 768;
		float frameRate = 1.0f;
		String outputURL[] = {"out.avi", "out2.avi"};	
		// MediaLocator oml[] = new MediaLocator[2];
		/* if ((oml[0] = createMediaLocator(outputURL[0])) == null) {
		 System.err.println("Cannot build media locator from: " + outputURL[0]);
		 //System.exit(0);
		 }
		 if ((oml[1] = createMediaLocator(outputURL[1])) == null) {
		 System.err.println("Cannot build media locator from: " + outputURL[1]);
		 //System.exit(0);
		 }
		 */
		MovieMaker imageToMovie = new MovieMaker();

		imageToMovie.prepare(width, height, frameRate);
		
		// System.out.println("Possible content descriptors");
		ContentDescriptor cdd[] = imageToMovie.getContentDescriptors();
		// for (int i=0; i<cdd.length; i++) {
		// System.out.println(cdd[i]);
		// }
		
		// System.out.println("Possible formats for this content descriptor");
		Format ff[] = imageToMovie.getFormats();

		// for (int i=0; i<ff.length; i++) {
		// System.out.println(ff[i]);
		// }
		imageToMovie.doIt();
		// System.out.println("Back in main thread");
		// System.out.println("Threads: "+Thread.currentThread().activeCount());
		try {
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));

			imageToMovie.signalNoMoreData();
		} catch (Exception e) {
			System.out.println("Failed....");
		}
		// System.out.println("Threads: "+Thread.currentThread().activeCount());
		imageToMovie.prepare(width, height, frameRate);

		imageToMovie.doIt();
		// System.out.println("Back in main thread");
		// System.out.println("Threads: "+Thread.currentThread().activeCount());
		try {
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));
			Thread.currentThread().sleep(200);
			imageToMovie.addImage(new Robot().createScreenCapture(new Rectangle(Toolkit.getDefaultToolkit().getScreenSize())));

			imageToMovie.signalNoMoreData();
		} catch (Exception e) {
			System.out.println("Failed....");
		}
		// System.out.println("Threads: "+Thread.currentThread().activeCount());
		
		// System.out.println("Main thread finished");
	}

	/**
	 * Creates a new medialocator and returns it (given the url)
	 * @return MediaLocator the new media locator, if it fails to create it, the function returns null 
	 */
	static MediaLocator createMediaLocator(String url) {
		MediaLocator ml;

		if (url.indexOf(":") > 0 && (ml = new MediaLocator(url)) != null) {
			return ml;
		}
		if (url.startsWith(File.separator)) {
			if ((ml = new MediaLocator("file:" + url)) != null) {
				return ml;
			}
		} else {
			String file = "file:" + System.getProperty("user.dir") + File.separator + url;

			if ((ml = new MediaLocator(file)) != null) {
				return ml;
			}
		}
		return null;
	}
    
	/**
	 * Returns the content descriptors which are compatible on this system and given on video format
	 * @return ContentDescriptor[] the content descriptors 
	 */

	public ContentDescriptor[] getContentDescriptors() {
		ContentDescriptor cd[] = p.getSupportedContentDescriptors();
		int num = 0;

		for (int i = 0; i < cd.length; i++) {
			if (cd[i].getContentType().startsWith("video")) {
				num++;
			}
		}
		ContentDescriptor res[] = new ContentDescriptor[num];

		for (int i = 0; i < cd.length; i++) {
			if (cd[i].getContentType().startsWith("video")) {
				res[--num] = cd[i];
			}
		}
		return res;
	}
	
	/**
	 * Returns the video formats that are available on this system
	 * @return Format[] the video formats
	 */
	public Format[] getFormats() {
		TrackControl tcs[] = p.getTrackControls();
		Format f[] = tcs[0].getSupportedFormats();

		return f;
	}
	
	/**
	 * Adds an image to the image data source
	 * @param bi buffered image that is added
	 */
	public void addImage(BufferedImage bi) {
		ids.addImage(bi);
	}

	/**
	 * Sets the signal no more data(end of input)
	 */
	public void signalNoMoreData() {
		ids.signalNoMoreData();
	}
	
	/**
	 * Sets the movie format
	 * @param format the movie format
	 */
	public void setFormat(Format format) {
		TrackControl tcs[] = p.getTrackControls();

		tcs[0].setFormat(format);
	}
		
	/**
	 * Sets the content descriptors
	 * @param description the content descriptor
	 */
	public void setContentDescriptor(ContentDescriptor description) {
		p.setContentDescriptor(description);
	}
}

