/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules.generic;


import java.awt.Component;
import javax.swing.*;
import javax.media.j3d.BranchGroup;

import icc3dpackage.*;
import icc3dpackage.managers.*;
import icc3dpackage.view3d.*;


/**
 * A module represents a plugin that can be added to a 3D scene. The module 
 * has its own parameter panel, which is displayed in a JTabbedPane, and may
 * add one or more branch groups to the 3D scene graph.
 */
public abstract class Module extends Plugin {
	
	/**
	 * Initializes the module.
	 * @param view3D The 3D view where this module is used.
	 */
	public abstract boolean initialize(View3D view3D);
	
	/**
	 * Deinitializes the module.
	 */
	public abstract void deInitialize();
	
	/**
	 * Returns one or more branch groups which are added to the scene graph.
	 * @return An array of branch groups.
	 */
	public abstract BranchGroup[] getSceneBranch();
	
	/**
	 * Returns the icon that represents this module type.
	 * @return The icon.
	 */
	public abstract Icon getIcon();
	
	/**
	 * Returns the parameter panel.
	 * @return The parameter panel.
	 */
	public abstract Component getParameterPanel();
	
	/**
	 * Determines if the user should be able to add modules of this type.
	 * @return True if the user should be able to add modules of this type.
	 */
	public abstract boolean userAdd();
	
	/**
	 * Determines if the user should be able to remove modules of this type.
	 * @return True if the user should be able to remove modules of this type.
	 */
	public abstract boolean userRemove();
	
	/**
	 * Determines if one instance of this class should be added to all 3D
	 * scenes as default.
	 * @return True if one instance of this class should be added to all 3D
	 * scenes as default.
	 */
	public abstract boolean existsAsDefault();
	
	/**
	 * Returns the name of this module type.
	 * @return The name of the module type.
	 */
	public abstract String getName();
	
	/**
	 * Returns the priority of this module.
	 * @return The priority of this module.
	 */
	public abstract int getPriority();
	
	/**
	 * Checks whether an update of the 3D scene is possible at this time.
	 * @return True if this module is ready for a 3D update, or false if
	 * for instance an error in the parameter panel prevents an update at
	 * this time.
	 */
	public abstract boolean readyForUpdate();
}
