/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.modules.regression;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


/**
 * A class capable of performing polynomial regression.
 */
public class PolynomialRegression {
	
	/** Allow constants */
	private boolean allowConstants;
	
	/** Degree */
	private int regDegree;
	
	/** Variables */
	private int regVariables;
	
	/** Samples */
	private int regSamples;
	
	/** Equations */
	private int regEquations;
	
	/** lu decomposition */
	private double lu[][];
	
	/** Permutations */
	private int permutation[];
	
	private double m[][];
	
	private double a[];
	private double y[];
	
	private double x[][];
	private double f[];
	
	/**
	 * Creates a clone of another PolynomialRegression.
	 * @param cloneMe The original.
	 */
	public PolynomialRegression(PolynomialRegression cloneMe) {
		regDegree = cloneMe.regDegree;
		regVariables = cloneMe.regVariables;
		regSamples = cloneMe.regSamples;
		regEquations = cloneMe.regEquations;
		allowConstants = cloneMe.allowConstants;
		lu = new double[cloneMe.lu.length][];
		for (int i = 0; i < lu.length; i++) {
			lu[i] = new double[cloneMe.lu[i].length];
			System.arraycopy(cloneMe.lu[i], 0, lu[i], 0, lu[i].length);
		}
		permutation = new int[cloneMe.permutation.length];
		System.arraycopy(cloneMe.permutation, 0, permutation, 0, permutation.length);
		m = new double[cloneMe.m.length][];
		for (int i = 0; i < m.length; i++) {
			m[i] = new double[cloneMe.m[i].length];
			System.arraycopy(cloneMe.m[i], 0, m[i], 0, m[i].length);
		}
		x = new double[cloneMe.x.length][];
		for (int i = 0; i < x.length; i++) {
			x[i] = new double[cloneMe.x[i].length];
			System.arraycopy(cloneMe.x[i], 0, x[i], 0, x[i].length);
		}
		a = new double[cloneMe.a.length];
		System.arraycopy(cloneMe.a, 0, a, 0, a.length);
		y = new double[cloneMe.y.length];
		System.arraycopy(cloneMe.y, 0, y, 0, y.length);
		f = new double[cloneMe.f.length];
		System.arraycopy(cloneMe.f, 0, f, 0, f.length);
		
	}
	
	/**
	 * Creates a PolynomialRegression.
	 * @param degree The polynomial degree.
	 * @param variables The number of variables.
	 * @param samples The number of samples.
	 * @param allowConst Allow constants in equation.
	 */
	public PolynomialRegression(int degree, int variables, int samples, boolean allowConst) {
		allowConstants = allowConst;
		regDegree = degree;
		regVariables = variables;
		regSamples = samples;
		if (allowConstants) {
			regEquations = regVariables * regDegree + 1;
		} else {
			regEquations = regVariables * regDegree;
		}
		m = new double[regEquations][regEquations];
		a = new double[regEquations];
		y = new double[regEquations];
		x = new double[regSamples][regVariables];
		f = new double[regSamples];
	}
	
	/**
	 * Sum samples.
	 * @param power The power.
	 * @param variable The variable.
	 */
	private double sumFX(int power, int variable) {
		double sum = 0;

		for (int i = 0; i < regSamples; i++) {
			sum += f[i] * Math.pow(x[i][variable], power);
		}
		return sum;
	}
	
	/**
	 * Populate the M and Y matrix.
	 * @param fOut The output.
	 * @param vIn The input variables.
	 */
	public void populateMY(double fOut[], double vIn[][]) {
		for (int i = 0; i < regSamples; i++) {
			f[i] = fOut[i];
			for (int j = 0; j < regVariables; j++) {
				x[i][j] = vIn[i][j];
			}
		}
		
		int lowestDegree = 1;

		if (allowConstants) {
			lowestDegree = 0;
		}
		
		for (int i = lowestDegree; i <= regDegree; i++) {
			for (int j = 0; j < regVariables; j++) {
				if (i == 0 && j != 0) {
					continue;
				}
				int row = (i - 1) * regVariables + j;

				if (allowConstants) {
					row = 1 + (i - 1) * regVariables + j;
				}
				if (row < 0) {
					row = 0;
				}
				y[row] = sumFX(i, j);
				
				for (int k = lowestDegree; k <= regDegree; k++) {
					for (int l = 0; l < regVariables; l++) {
						if (k == 0 && k != 0) {
							continue;
						}
						int column = (k - 1) * regVariables + l;

						if (allowConstants) {
							column = 1 + (k - 1) * regVariables + l;
						}
						if (column < 0) {
							column = 0;
						}
						m[row][column] = sumXX(i, j, k, l);
					}
				}
			}
		}
	}
	
	/**
	 * Display plot of data.
	 */
	public void showPlot1() {
		if (regVariables == 1) {
			PolynomialRegression copy = new PolynomialRegression(this);
			PlotFrame pf = new PlotFrame(copy);

			pf.show();
		}
	}
	
	/**
	 * Returns if constants are allowed.
	 * @return If constants are allowed.
	 */
	public boolean getAllowConstants() {
		return allowConstants;
	}
	
	/**
	 * Returns the degree.
	 * @return The degree.
	 */
	public int getDegree() {
		return regDegree;
	}
	
	/**
	 * Solves the equation.
	 * @return The solution.
	 */
	public double[] getA() throws Exception {
		a = solve();
		return a;
	}
	
	/**
	 * Returns the prviously found equation.
	 * @return The prviously found equation.
	 */
	public double[] getOldA() {
		return a;
	}
	
	/**
	 * Returns the sample values.
	 * @return The sample values.
	 */
	public double[] getF() {
		return f;
	}
	
	/**
	 * Returns the M matrix.
	 * @return The M matrix.
	 */
	public double[][] getM() {
		return m;
	}
	
	/**
	 * Returns the number of samples.
	 * @return The number of samples.
	 */
	public int getSAMPLES() {
		return regSamples;
	}
	
	/**
	 * Returns the input sample values.
	 * @return The input sample values.
	 */
	public double[][] getX() {
		return x;
	}
	
	/**
	 * Returns the Y column.
	 * @return The Y column.
	 */
	public double[] getY() {
		return y;
	}
	
	/**
	 * Sum of two variable lists.
	 * @param power1 The power of the first variable.
	 * @param variable1 The first variable.
	 * @param power2 The power of the second variable.
	 * @param variable2 The second variable.
	 * @return The sum.
	 */
	private double sumXX(int power1, int variable1, int power2, int variable2) {
		double sum = 0;

		for (int i = 0; i < regSamples; i++) {
			sum += Math.pow(x[i][variable1], power1) * Math.pow(x[i][variable2], power2);
		}
		return sum;
	}
	
	/**
	 * Solves the equations.
	 * @return The solution.
	 */
	private double[] solve() {
		decompose();
		double yy[] = forwardSubstitution(y);
		double xx[] = backSubstitution(yy);

		return xx;
	}
	
	/**
	 * Decomposes the equations.
	 */
	private void decompose() {
		lu = new double[regEquations][regEquations];
		for (int i = 0; i < regEquations; i++) {
			System.arraycopy(m[i], 0, lu[i], 0, regEquations);
		}
		permutation = new int[regEquations];
		double scales[] = new double[regEquations];
		
		for (int r = 0; r < regEquations; ++r) {
			permutation[r] = r;
			double largest = 0;

			for (int c = 0; c < regEquations; ++c) {
				double value = Math.abs(lu[r][c]);

				if (largest < value) {
					largest = value;
				}
			}
			if (largest != 0) {
				scales[r] = 1 / largest;
			} else {
				throw new RuntimeException("Row is zero");
			}
		}
		forwardElimination(scales);
		if (lu[permutation[regEquations - 1]][regEquations - 1] == 0) {
			throw new RuntimeException("Singular matrix");
		}
	}
	
	/**
	 * Returns the calculated value given a certain input.
	 * @return The output.
	 */
	public double getValueAt(double inVar[]) {
		if (allowConstants) {
			double sum = a[0];

			for (int i = 1; i <= regDegree; i++) {
				for (int j = 0; j < regVariables; j++) {
					sum += Math.pow(inVar[j], i) * a[1 + (i - 1) * regVariables + j];
				}
			}
			return sum;
		} else {
			double sum = 0;

			for (int i = 1; i <= regDegree; i++) {
				for (int j = 0; j < regVariables; j++) {
					sum += Math.pow(inVar[j], i) * a[(i - 1) * regVariables + j];
				}
			}
			return sum;
		}
	}
	
	/**
	 * Performs forward eliminiation.
	 * @param scales The scales.
	 */
	private void forwardElimination(double scales[]) {
		for (int rPivot = 0; rPivot < regEquations - 1; ++rPivot) {
			double largestScaledElmt = 0;
			int rLargest = 0;

			for (int r = rPivot; r < regEquations; ++r) {
				int pr = permutation[r];
				double absElmt = Math.abs(lu[pr][rPivot]);
				double scaledElmt = absElmt * scales[pr];

				if (largestScaledElmt < scaledElmt) {
					largestScaledElmt = scaledElmt;
					rLargest = r;
				}
			}

			if (largestScaledElmt == 0) {
				throw new RuntimeException("Singular matrix");
			}

			if (rLargest != rPivot) {
				int temp = permutation[rPivot];

				permutation[rPivot] = permutation[rLargest];
				permutation[rLargest] = temp;
			}

			int prPivot = permutation[rPivot];
			double pivotElmt = lu[prPivot][rPivot];

			for (int r = rPivot + 1; r < regEquations; ++r) {

				int   pr = permutation[r];
				double multiple = lu[pr][rPivot] / pivotElmt;

				lu[pr][rPivot] = multiple;

				if (multiple != 0) {
					for (int c = rPivot + 1; c < regEquations; ++c) {
						double elmt = lu[pr][c];

						elmt -= multiple * lu[prPivot][c];
						lu[pr][c] = elmt;
					}
				}
			}
		}
	}
	
	/**
	 * Performs forward substitution.
	 * @param b An array.
	 */
	private double[] forwardSubstitution(double b[]) {
		double y[] = new double[regEquations];

		for (int r = 0; r < regEquations; ++r) {
			int   pr = permutation[r];
			double dot = 0;

			for (int c = 0; c < r; ++c) {
				dot += lu[pr][c] * y[c];
			}
			y[r] = b[pr] - dot;
		}
		return y;
	}
	
	/**
	 * Performs back substitution.
	 * @param y An array.
	 */
	private double[] backSubstitution(double y[]) {
		double x[] = new double[regEquations];

		for (int r = regEquations - 1; r >= 0; --r) {
			int   pr = permutation[r];
			double dot = 0;

			for (int c = r + 1; c < regEquations; ++c) {
				dot += lu[pr][c] * x[c];
			}
			x[r] = (y[r] - dot) / lu[pr][r];
		}
		return x;
	}  
}


/**
 * Displays a plot of a polynomial regression.
 */
class PlotFrame extends JFrame {
	
	/** The plot panel */
	private PlotPanel plotPanel;
	
	/** The plot controls */
	private PlotControl plotControl;
	
	/** The regression */
	private PolynomialRegression pr;
	
	/**
	 * Creates a plot frame.
	 * @param pReg The regression that should be displayed.
	 */
	public PlotFrame(PolynomialRegression pReg) {
		super("Plot of polynomial regression function");
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		pr = pReg;
		plotPanel = new PlotPanel(this);
		plotControl = new PlotControl(this, plotPanel);
		Container c = getContentPane();

		c.setLayout(new BorderLayout());
		c.add(plotPanel, BorderLayout.CENTER);
		c.add(plotControl, BorderLayout.SOUTH);
		pack();
	}
	
	/**
	 * Returns the polynomial regression.
	 * @return The polynomial regression.
	 */
	public PolynomialRegression getPR() {
		return pr;
	}
}


/**
 * A plot panel.
 */
class PlotPanel extends JPanel {
	
	/** Plot limit */
	private double lowX, lowY, highX, highY;
	
	/** The plot frame */
	private PlotFrame plotFrame;
	
	/**
	 * Creates a plot panel.
	 * @param pFrame The paren plot frame.
	 */
	public PlotPanel(PlotFrame pFrame) {
		plotFrame = pFrame;
		
		double x[][] = plotFrame.getPR().getX();
		double f[] = plotFrame.getPR().getF();
		int regSamples = plotFrame.getPR().getSAMPLES();

		lowX = x[0][0];
		highX = x[0][0];
		lowY = f[0];
		highY = f[0];
		for (int i = 1; i < regSamples; i++) {
			if (f[i] > highY) {
				highY = f[i];
			}
			if (f[i] < lowY) {
				lowY = f[i];
			}
			if (x[i][0] > highX) {
				highX = x[i][0];
			}
			if (x[i][0] < lowX) {
				lowX = x[i][0];
			}
		}
		lowX -= (highX - lowX) / 5.0;
		lowY -= (highY - lowY) / 5.0;
		highX += (highX - lowX) / 5.0;
		highY += (highY - lowY) / 5.0;
		
		if (lowX > 0) {
			lowX = -highX / 10.0;
		}
		if (lowY > 0) {
			lowY = -highY / 10.0;
		}
		if (highX < 0) {
			highX = -lowX / 10.0;
		}
		if (highY < 0) {
			highY = -lowY / 10.0;
		}
	}
	
	/**
	 * Paints the plot panel.
	 * @param g The Graphics object.
	 */
	public void paint(Graphics g) {
		g.setColor(Color.white);
		g.fillRect(0, 0, (int) getSize().getWidth(), (int) getSize().getWidth());
		g.setColor(Color.black);
		plotLine(lowX, 0, highX, 0, g);
		plotLine(0, lowY, 0, highY, g);
		
		int divisions = 1000;

		g.setColor(Color.green);
		for (int i = 0; i < divisions; i++) {
			double x1 = lowX + ((double) i / (double) divisions) * (highX - lowX);
			double x2 = lowX + ((double) (i + 1) / (double) divisions) * (highX - lowX);
			double y1 = plotFrame.getPR().getValueAt(new double[] {x1});
			double y2 = plotFrame.getPR().getValueAt(new double[] {x2});

			plotLine(x1, y1, x2, y2, g);
		}
		
		int regSamples = plotFrame.getPR().getSAMPLES();
		double x[][] = plotFrame.getPR().getX();
		double f[] = plotFrame.getPR().getF();
		
		for (int i = 0; i < regSamples; i++) {
			g.setColor(Color.blue);
			plotPoint(x[i][0], f[i], g);
			g.setColor(Color.red);
			plotText("(" + x[i][0] + " , " + f[i] + ")", x[i][0], f[i], g);
		}					
	}
	
	/**
	 * Returns the lowest x value displayed.
	 * @return The lowest x value displayed.
	 */
	public double getLowX() {
		return lowX;
	}
	
	/**
	 * Returns the lowest y value displayed.
	 * @return The lowest y value displayed.
	 */
	public double getLowY() {
		return lowY;
	}
	
	/**
	 * Returns the highest x value displayed.
	 * @return The highest x value displayed.
	 */
	public double getHighX() {
		return highX;
	}
	
	/**
	 * Returns the highest y value displayed.
	 * @return The highest y value displayed.
	 */
	public double getHighY() {
		return highY;
	}
	
	/**
	 * Set the lowest x value displayed.
	 * @param x The lowest x value displayed.
	 */
	public void setLowX(double x) {
		lowX = x;
	}
	
	/**
	 * Set the lowest y value displayed.
	 * @param y The lowest y value displayed.
	 */
	public void setLowY(double y) {
		lowY = y;
	}
	
	/**
	 * Set the highest x value displayed.
	 * @param x The highest x value displayed.
	 */
	public void setHighX(double x) {
		highX = x;
	}
	
	/**
	 * Set the highest y value displayed.
	 * @param y The highest y value displayed.
	 */
	public void setHighY(double y) {
		highY = y;
	}
	
	/**
	 * Plot a point.
	 * @param x The x coordinate.
	 * @param y The y coordinate.
	 * @param g The Graphics object used.
	 */
	private void plotPoint(double x, double y, Graphics g) {
		g.fillOval(toLocalX(x), toLocalY(y), 5, 5);
	}
	
	/**
	 * Plot a string.
	 * @param s The string.
	 * @param x The x coordinate.
	 * @param y The y coordinate.
	 * @param g The Graphics object used.
	 */
	private void plotText(String s, double x, double y, Graphics g) {
		g.drawString(s, 5 + toLocalX(x), toLocalY(y) - 5);
	}
	
	/**
	 * Plot a line.
	 * @param x1 The first x coordinate.
	 * @param y1 The first y coordinate.
	 * @param x2 The last x coordinate.
	 * @param y2 The last y coordinate.
	 * @param g The Graphics object used.
	 */
	private void plotLine(double x1, double y1, double x2, double y2, Graphics g) {
		g.drawLine(toLocalX(x1), toLocalY(y1), toLocalX(x2), toLocalY(y2));
	}
	
	/**
	 * Converts logical x coordinates to physical x coordinates.
	 * @param x A logical x coordinate.
	 * @return The physical x coordinate.
	 */
	private int toLocalX(double x) {
		Dimension d = getSize();

		return (int) (d.getWidth() * (x - lowX) / (highX - lowX));
	}
	
	/**
	 * Converts logical y coordinates to physical y coordinates.
	 * @param y A logical y coordinate.
	 * @return The physical y coordinate.
	 */
	private int toLocalY(double y) {
		Dimension d = getSize();

		return (int) d.getHeight()
				- (int) (d.getHeight() * (y - lowY) / (highY - lowY));
	}
	
	/**
	 * Returns the preferred dimension.
	 * @return The preferred dimension.
	 */
	public Dimension getPreferredSize() {
		return new Dimension(500, 300);
	}
}


/**
 * Displays plot controls.
 */
class PlotControl extends JPanel {
	
	/** The plot frame */
	private PlotFrame plotFrame;
	
	/** The plot panel */
	private PlotPanel plotPanel;
	
	/** Display control points */
	private JCheckBox showControlPoints;
	
	/** Display text control points */
	private JCheckBox showTextControlPoints;
	
	/** Increase the low x value */
	private JButton increaseLowX;
	
	/** Increase the low y value */
	private JButton increaseLowY;
	
	/** Increase the high x value */
	private JButton increaseHighX;
	
	/** Increase the high y value */
	private JButton increaseHighY;
	
	/** Decrease the low x value */
	private JButton decreaseLowX;
	
	/** Decrease the low y value */
	private JButton decreaseLowY;
	
	/** Decrease the high x value */
	private JButton decreaseHighX;
	
	/** Decrease the high y value */
	private JButton decreaseHighY;
	
	/**
	 * Creates a PlotControl.
	 * @param pFrame The plot frame.
	 * @param pPanel The plot panel.
	 */
	public PlotControl(PlotFrame pFrame, PlotPanel pPanel) {
		super(new GridLayout(1, 0));
		plotFrame = pFrame;
		plotPanel = pPanel;
		showControlPoints = new JCheckBox("Show control points");
		showTextControlPoints = new JCheckBox("Show point labels");
		increaseLowX = new JButton("low x +");
		increaseLowX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setLowX(plotPanel.getLowX()
						+ (plotPanel.getHighX() - plotPanel.getLowX()) * 0.1);
				plotPanel.repaint();
			}
		});
		increaseHighX = new JButton("high x +");
		increaseHighX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setHighX(plotPanel.getHighX()
						+ (plotPanel.getHighX() - plotPanel.getLowX()) * 0.1);
				plotPanel.repaint();
			}
		});
		increaseLowY = new JButton("low y +");
		increaseLowY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setLowY(plotPanel.getLowY()
						+ (plotPanel.getHighY() - plotPanel.getLowY()) * 0.1);
				plotPanel.repaint();
			}
		});
		increaseHighY = new JButton("high y +");
		increaseHighY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setHighY(plotPanel.getHighY()
						+ (plotPanel.getHighY() - plotPanel.getLowY()) * 0.1);
				plotPanel.repaint();
			}
		});
		decreaseHighY = new JButton("high y -");
		decreaseHighY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setHighY(plotPanel.getHighY()
						- (plotPanel.getHighY() - plotPanel.getLowY()) * 0.1);
				plotPanel.repaint();
			}
		});
		decreaseHighX = new JButton("high x -");
		decreaseHighX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setHighX(plotPanel.getHighX()
						- (plotPanel.getHighX() - plotPanel.getLowX()) * 0.1);
				plotPanel.repaint();
			}
		});
		decreaseLowY = new JButton("low y -");
		decreaseLowY.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setLowY(plotPanel.getLowY()
						- (plotPanel.getHighY() - plotPanel.getLowY()) * 0.1);
				plotPanel.repaint();
			}
		});
		decreaseLowX = new JButton("low x -");
		decreaseLowX.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				plotPanel.setLowX(plotPanel.getLowX()
						- (plotPanel.getHighX() - plotPanel.getLowX()) * 0.1);
				plotPanel.repaint();
			}
		});
		JPanel decIncPanel = new JPanel(new GridLayout(2, 0));

		decIncPanel.add(increaseLowX);
		decIncPanel.add(increaseHighX);
		decIncPanel.add(increaseLowY);
		decIncPanel.add(increaseHighY);
		decIncPanel.add(decreaseLowX);
		decIncPanel.add(decreaseHighX);
		decIncPanel.add(decreaseLowY);
		decIncPanel.add(decreaseHighY);
		
		JPanel showPanel = new JPanel(new GridLayout(1, 0));

		showPanel.add(showControlPoints);
		showPanel.add(showTextControlPoints);
		
		JPanel p = new JPanel(new BorderLayout());

		p.add(decIncPanel, BorderLayout.EAST);
		p.add(showPanel, BorderLayout.WEST);
		add(p);
	}
	
	/**
	 * Returns the preferred dimension.
	 * @return The preferred dimension.
	 */
	public Dimension getPreferredSize() {
		return new Dimension(500, 100); 
	}
}
