/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.pickable;


import javax.media.j3d.*;


/**
 * A class used to keep track of the pixels in
 * a 2D-image that are represented as an object in 3D.
 */
public class BasicObject3D implements PickableObject {
	
	/**
	 * The indices of the pixels associated with this BasicObject3D.
	 */
	private int colorIndex[];
	
	/**
	 * The position of this BasicObject3D.
	 */
	private float position[];
	
	/**
	 * The color of this BasicObject3D.
	 */
	private byte color[];
	
	/**
	 * The number of pixels added to this BasicObject3D.
	 */
	private int added;
	
	/**
	 * The PointArray index which identifies the vertex belonging to this BasicObject3D.
	 */
	private int index;
	
	/**
	 * The PointArray which contains the vertex of this BasicObject3D.
	 */
	private PointArray pointArray;
	
	/**
	 * Constructs a BasicObject3D.
	 * @param addCount The number of pixels associated with this object.
	 * @param pos The 3D position of this object.
	 * @param col The color of this object.
	 * @param index The PointArray index which identifies the vertex belonging to
	 * this BasicObject3D.
	 * @param pointArray The PointArray which contains the vertex of this BasicObject3D.
	 */
	public BasicObject3D(int addCount, float pos[], byte col[], int index, PointArray pointArray) {
		this.pointArray = pointArray;
		this.index = index;
		colorIndex = new int[addCount];
		position = pos;
		color = col;
		added = 0;
	}
	
	/**
	 * Creates a BasicObject3D with the given position.
	 * @param pos The position. 
	 */
	public BasicObject3D(float pos[]) {
		position = pos;
	}
	
	/**
	 * Adds an image color to the object's list of associated image colors.
	 * @param addColor The color to be added.
	 */
	public void addColor(int addColor) {
		colorIndex[added++] = addColor;
	}
	
	/**
	 * Returns the colors that belong to the object.
	 * @return The indices of the colors that belong to the object.
	 */
	public int[] getColorIndices() {
		return colorIndex;
	}
	
	/**
	 * Returns the position of the object.
	 * @return The object position.
	 */
	public float[] getPosition() {
		return new float[] {position[0], position[1], position[2]};
	}
	
	/**
	 * Returns the color of this object.
	 * @return The color of this object.
	 */
	public byte[] getColor() {
		return new byte[] {color[0], color[1], color[2]};
	}
	
	/**
	 * Sets the color of the object.
	 * @param col The color.
	 */
	public void setColor(byte col[]) {
		color = new byte[] {col[0], col[1], col[2]};
		pointArray.setColor(index, color);
	}
	
	/**
	 * Sets the position of the object.
	 * @param pos The position.
	 */
	public void setPosition(float pos[]) {
		position = new float[] {pos[0], pos[1], pos[2]};
		pointArray.setCoordinate(index, position);
	}
	
	/**
	 * Updates the object position.
	 */
	public void updatePosition() {
		pointArray.setCoordinate(index, position);
	}
	
	/**
	 * Updates the object color.
	 */
	public void updateColor() {
		pointArray.setColor(index, color);
	}
	
	/**
	 * Returns the object position by reference.
	 * @return The object position.
	 */
	public float[] getPositionReference() {
		return position;
	}
	
	/**
	 * Returns the object color by reference.
	 * @return The object color.
	 */
	public byte[] getColorReference() {
		return color;
	}
	
	/**
	 * Sets the color index.
	 * @param i The index.
	 */
	public void setColorIndex(int i[]) {
		this.colorIndex = i;
	}
}
