/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.pickable;


import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.image2d.Image2D;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.triangulation.*;
import icc3dpackage.visualizations.appearance.*;


/**
 * A MovingObject is inserted into the scene graph to represent the visualization
 * of an image.
 */
public class MovingObject extends Shape3D {
	
	/**
	 * The starting index.
	 */
	private int start;
	
	/**
	 * The objects that the user can move.
	 */
	private PickableObject obj[];
	
	/**
	 * The tetrahedron representation of the object.
	 */
	private Tetrahedron tetra[];
	
	/**
	 * The associated image.
	 */
	private Image2D image2D;
	
	/**
	 * The color space.
	 */
	private ColorSpace3D colorSpace;
	
	/**
	 * The colors.
	 */
	private float colors[];
	
	/**
	 * The center of the object.
	 */
	private float center[];
	
	/**
	 * The geometry.
	 */
	private GeometryArray geometryArray[];
	
	
	/**
	 * Creates a MovingObject.
	 */
	public MovingObject() {
		super();
		setAppearance(StandardAppearance.getInvisibleAppearance());
	}
	
	/**
	 * Returns the dynamic objects.
	 * @return The objects that the user can move.
	 */
	public PickableObject[] getObjects() {
		return obj;
	}
	
	/**
	 * Returns the starting index.
	 * @return The starting index.
	 */
	public int getStartingIndex() {
		return start;
	}
	
	/**
	 * Sets the object list.
	 * @param obj An array of objects.
	 */
	public void setObjects(PickableObject obj[]) {
		this.obj = obj;
	}
	
	
	/**
	 * Sets the starting index.
	 */
	public void setStart(int s) {
		start = s;
	}
	
	/**
	 * Returns the associated image.
	 * @return the associated image.
	 */
	public Image2D getImage2D() {
		return image2D;
	}
	
	/**
	 * Sets the image.
	 * @param image2D The image.
	 */
	public void setImage2D(Image2D image2D) {
		this.image2D = image2D;
	}
	
	/**
	 * Returns the color space.
	 * @return The color space.
	 */
	public ColorSpace3D getColorSpace() {
		return colorSpace;
	}
	
	/**
	 * Sets the color space.
	 * @param colorSpace The color space.
	 */
	public void setColorSpace(ColorSpace3D colorSpace) {
		this.colorSpace = colorSpace;
	}
	
	/**
	 * Sets the colors.
	 * @param colors The colors.
	 */
	public void setColors(float colors[]) {
		this.colors = colors;
	}
	
	/**
	 * Returns the colors.
	 * @return The colors.
	 */
	public float[] getColors() {
		return colors;
	}
	
	/**
	 * Returns the geometry.
	 * @return The geometry arrays.
	 */
	public GeometryArray[] getGeometryArrays() {
		return geometryArray;
	}
	
	/**
	 * Sets the geometry.
	 * @param ga The geometry arrays.
	 */
	public void setGeometryArrays(GeometryArray ga[]) {
		geometryArray = ga;
	}
	
	/**
	 * Returns the object center.
	 * @return The object center.
	 */
	public float[] getCenter() {
		if (center == null) {
			return null;
		}
		return new float[] {center[0], center[1], center[2]};
	}
	
	/**
	 * Sets the object center.
	 * @param c The center.
	 */
	public void setCenter(float c[]) {
		center = new float[] {c[0], c[1], c[2]};
	}
	
	/**
	 * Returns a reference to the object center.
	 * @return The object center.
	 */
	public float[] getCenterReference() {
		return center;
	}
	
	/**
	 * Returns the tetrahedrons.
	 * @return The tetrahedrons.
	 */
	public Tetrahedron[] getTetrahedrons() {
		return tetra;
	}
	
	/**
	 * Sets the tetrahedrons.
	 * @param tetrahedron The tetrahedrons.
	 */
	public void setTetrahedrons(Tetrahedron tetrahedron[]) {
		tetra = tetrahedron;
	}
}
