/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.pickable;


import java.util.Vector;
import javax.media.j3d.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.triangulation.*;


/**
 * A class used to keep track of the pixels in
 * a 2D-image that are represented as a segment object in 3D.
 */
public class SegmentObject3D implements PickableObject, PickableSegmentObject, PickableTetrahedronObject {
	
	/** The indices of the pixels associated with this SegmentObject3D. */
	private int colorIndex[];
	
	/** The position of this SegmentObject3D. */
	private float position[];
	
	/** The color of this SegmentObject3D. */
	private byte color[];
	
	/** The number of pixels added to this SegmentObject3D. */
	private int added;
	
	/**
	 * The array indices which identify the GeometryArrays that contain information 
	 * pertaining to this object.
	 */
	private int arrayIndex[];
	
	/**
	 * The coordinate indices which identify the individual vertices in the 
	 * GeometryArrays that contain information pertaining to this object.
	 */
	private int coordinateIndex[];
	
	/** The GeometryArrays that contain information pertaining to this object. */
	private GeometryArray geometryArray[];
	
	/** The center of the object. */
	private float center[];
	
	/** The tetrahedra that are associated with this SegmentObject3D. */
	private Tetrahedron tetrahedronNeighbor[];
	
	/** The tetrahedron list. */
	private Vector tetras;
	
	/**
	 * Constructs a SegmentObject3D.
	 * @param center The center of the object.
	 * @param pos The 3D position of this object.
	 * @param col The color of this object.
	 * @param arrayIndex The array indices which identify the GeometryArrays that contain information pertaining to this object.
	 * @param coordinateIndex The coordinate indices which identify the individual vertices in the GeometryArrays that contain information pertaining to this object.
	 * @param geometryArray The GeometryArrays that contain information pertaining to this object.
	 */
	public SegmentObject3D(float center[], int addCount, float pos[], byte col[], int arrayIndex[], int coordinateIndex[], GeometryArray geometryArray[]) {
		this.geometryArray = geometryArray;
		this.arrayIndex = arrayIndex;
		this.coordinateIndex = coordinateIndex;
		this.center = center;
		this.tetras = new Vector();
		colorIndex = new int[addCount];
		position = pos;
		color = col;
		added = 0;
	}
	
	/**
	 * Adds a image color to the object's list of associated image colors.
	 * @param addColor The color to be added.
	 */
	public void addColor(int addColor) {
		colorIndex[added++] = addColor;
	}
	
	/**
	 * Returns the colors that belong to the object.
	 * @return The indices of the colors that belong to the object.
	 */
	public int[] getColorIndices() {
		return colorIndex;
	}
	
	/**
	 * Returns the position of the object.
	 * @return The object position.
	 */
	public float[] getPosition() {
		return new float[] {position[0], position[1], position[2]};
	}
	
	/**
	 * Updates the object position.
	 */
	public void updatePosition() {
		for (int i = 0; i < arrayIndex.length; i++) {
			geometryArray[arrayIndex[i]].setCoordinate(coordinateIndex[i], position);
		}
	}
	
	/**
	 * Updates the object color.
	 */
	public void updateColor() {
		for (int i = 0; i < arrayIndex.length; i++) {
			geometryArray[arrayIndex[i]].setColor(coordinateIndex[i], color);
		}	
	}
	
	/**
	 * Returns the object position by reference.
	 * @return The object position.
	 */
	public float[] getPositionReference() {
		return position;
	}
	
	/**
	 * Returns the object color by reference.
	 * @return The object color.
	 */
	public byte[] getColorReference() {
		return color;
	}
	
	/**
	 * Returns the color of this object.
	 * @return The color of this object.
	 */
	public byte[] getColor() {
		return new byte[] {color[0], color[1], color[2]};
	}
	
	/**
	 * Sets the color of the object.
	 * @param col The color.
	 */
	public void setColor(byte col[]) {
		color = new byte[] {col[0], col[1], col[2]};
		for (int i = 0; i < arrayIndex.length; i++) {
			geometryArray[arrayIndex[i]].setColor(coordinateIndex[i], color);
		}
	}
	
	/**
	 * Sets the position of the object.
	 * @param pos The position.
	 */
	public void setPosition(float pos[]) {
		position = new float[] {pos[0], pos[1], pos[2]};
		for (int i = 0; i < arrayIndex.length; i++) {
			geometryArray[arrayIndex[i]].setCoordinate(coordinateIndex[i], position);
		}
	}	
	
	/**
	 * Returns the center of the whole object.
	 * @return An array of three floats representing the position of the center.
	 */
	public float[] getCenter() {
		return center;
	}
	
	/**
	 * Adds a tetrahedron to this object's list of associated tetrahedra.
	 * @param t The tetrahedron.
	 */
	public void addTetrahedron(Tetrahedron t) {
		tetras.add(t);
	}
	
	/**
	 * Allocates the space necessary to store the tetrahedron neighbor list.
	 */
	public void finalizeTetrahedrons() {
		tetrahedronNeighbor = new Tetrahedron[tetras.size()];
		for (int i = 0; i < tetrahedronNeighbor.length; i++) {
			tetrahedronNeighbor[i] = (Tetrahedron) tetras.elementAt(i);
		}
		tetras = null;
	}
	
	/**
	 * Returns the tetrahedra associated with this object.
	 * @return The tetrahedra associated with this object.
	 */
	public Tetrahedron[] getTetrahedra() {
		return tetrahedronNeighbor;
	}
}
