/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.quantization;


import java.util.Comparator;


/**
 * This class implements a tree based heap.
 */
public class CustomHeap {
	
	/**
	 * The Comparator that implements an ordering of the heap elements.
	 */
	private Comparator comparator;
	
	/**
	 * The tree root.
	 */
	private HeapItem root;
	
	/**
	 * The number of elements on the heap.
	 */
	private int size;
	
	/**
	 * Constructs a CustomHeap with the specified Comparator.
	 */
	public CustomHeap(Comparator c) {
		this.comparator = c;
	}

	/**
	 * Returns the number of elements on the heap.
	 * @return The current size of the heap.
	 */
	public int size() {
		return size;
	}
	
	/**
	 * Retrieves the element with the specified key from the heap.
	 * @param key The element key.
	 * @return The element, or null if there is no element on the heap with
	 * the correct key.
	 */
	public Object get(Object key) {
		HeapItem p = getHeapItem(key);

		return (p == null ? null : p.value);
	}

	/**
	 * Returns the key of the first element on the heap.
	 * (The element with the largest key)
	 * @return The largest key.
	 */
	private Object firstKey() {
		return key(firstHeapItem());
	}

	/**
	 * Returns the key of the last element on the heap.
	 * (The element with the smallest key)
	 * @return The smallest key.
	 */
	private Object lastKey() {
		return key(lastHeapItem());
	}
	
	/**
	 * Returns the HeapItem with the specified key.
	 * @param key The key.
	 * @return The HeapItem.
	 */
	private HeapItem getHeapItem(Object key) {
		HeapItem item = root;

		while (item != null) {
			int cmp = compare(key, item.key);

			if (cmp == 0) {
				return item;
			} else if (cmp < 0) {
				item = item.left;
			} else {
				item = item.right;
			}
		}
		return null;
	}
	
	/**
	 * Returns the key part of a HeapItem.
	 * @param item The HeapItem.
	 * @return The key.
	 */
	private static Object key(HeapItem item) {
		if (item == null) {
			return null;
		}
		return item.key;
	}
	
	/**
	 * Places a key/value pair on the heap.
	 * @param key The key.
	 * @param value The value.
	 */
	public void push(Object key, Object value) {
		HeapItem t = root;

		if (t == null) {
			size++;
			root = new HeapItem(key, value, null);
			return;
		}

		while (true) {
			int cmp = compare(key, t.key);

			if (cmp <= 0) {
				if (t.left != null) {
					t = t.left;
				} else {
					size++;
					t.left = new HeapItem(key, value, t);
					return;
				}
			} else {
				if (t.right != null) {
					t = t.right;
				} else {
					size++;
					t.right = new HeapItem(key, value, t);
					return;
				}
			}
		}
	}
	
	/**
	 * Removes and returns the element with the largest key
	 * on the heap.
	 * @return The element with the largest key.
	 */
	public Object pop() {
		return remove(lastKey());
	}

	/**
	 * Removes the element with the specified key from the heap.
	 * @param key The element key.
	 * @return The elemnt.
	 */
	public Object remove(Object key) {
		HeapItem p = getHeapItem(key);

		if (p == null) {
			return null;
		}
		Object oldValue = p.value;

		deleteHeapItem(p);
		return oldValue;
	}

	/**
	 * Removes all items from the heap.
	 */
	public void clear() {
		size = 0;
		root = null;
	}

	/**
	 * Compares the two keys.
	 * @param k1 The first key.
	 * @param k2 The second key.
	 * @return <0, 0 or >0
	 */
	private int compare(Object k1, Object k2) {
		return comparator.compare(k1, k2);
	}

	/**
	 * Returns the HeapItem with the largest key currently on the heap.
	 */
	private HeapItem firstHeapItem() {
		HeapItem p = root;

		if (p != null) {
			while (p.left != null) {
				p = p.left;
			}
		}
		return p;
	}

	/**
	 * Returns the HeapItem with the smallest key currently on the heap.
	 */
	private HeapItem lastHeapItem() {
		HeapItem p = root;

		if (p != null) {
			while (p.right != null) {
				p = p.right;
			}
		}
		return p;
	}

	/**
	 * Returns the successor of the specified HeapItem.
	 */
	private HeapItem successor(HeapItem t) {
		if (t == null) {
			return null;
		} else if (t.right != null) {
			HeapItem p = t.right;

			while (p.left != null) {
				p = p.left;
			}
			return p;
		} else {
			HeapItem p = t.parent;
			HeapItem ch = t;

			while (p != null && ch == p.right) {
				ch = p;
				p = p.parent;
			}
			return p;
		}
	}

	/**
	 * Deletes the HeapItem.
	 * @param p The HeapItem.
	 */
	private void deleteHeapItem(HeapItem p) {
		size--;

		if (p.left != null && p.right != null) {
			HeapItem s = successor(p);

			p.key = s.key;
			p.value = s.value;
			p = s;
		}

		HeapItem replacement = (p.left != null ? p.left : p.right);

		if (replacement != null) {
			replacement.parent = p.parent;
			if (p.parent == null) {
				root = replacement;
			} else if (p == p.parent.left) {
				p.parent.left = replacement;
			} else {
				p.parent.right = replacement;
			}

			p.left = p.right = p.parent = null;
		} else if (p.parent == null) {
			root = null;
		} else {
			if (p.parent != null) {
				if (p == p.parent.left) {
					p.parent.left = null;
				} else if (p == p.parent.right) {
					p.parent.right = null;
				}
				p.parent = null;
			}
		}
	}
}


/**
 * A HeapItem consists of a key and a value. The tree based approach
 * means that a HeapItem may have 0-1 parents, and 0-2 children.
 */
class HeapItem {
	Object key;
	Object value;
	HeapItem left;
	HeapItem right;
	HeapItem parent;

	/**
	 * Constructs a HeapItem.
	 * @param k The key.
	 * @param val The value.
	 * @param par The parent.
	 */
	HeapItem(Object k, Object val, HeapItem par) {
		key = k;
		value = val;
		parent = par;
	}
}
