/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.quantization;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;


/**
 * Quantization of an image.
 */
public class Quantization {
	
	/** Perform quantization when you open an image */
	public static final int IMAGE_QUANTIZATION = 11;
	
	/** Perform quantization in order to simplify 3D calculations */
	public static final int SIMPLIFY_QUANTIZATION = 12;
	
	/** Uniform quantization */
	public static final int UNIFORM = 0;
	
	/** Octree quantization */
	public static final int OCTREE = 1;
	
	/** Binary tree quantization */
	public static final int BINARY_TREE = 2;
	
	/** Infinite number of subdivisions (no quantization) */
	public static int INFINITY = Integer.MAX_VALUE;
	
	/** The quantization paramater panel */
	private JPanel parameterPanel;
	
	/** The type of quantization */
	private JComboBox typeBox;
	
	/** The number of subdivisions */
	private JComboBox detailCombo;
	
	/** The quantization priority parameters */
	private JComboBox priorityBox;
	
	/** The number of output colors */
	private JComboBox colorTarget;
	
	/** The preferred placement of points that represent a volume in space */
	private JComboBox pointPlacement;
	
	/** The panel that contains the components specific to the current quantization mode */
	private JPanel modeChildPanel;
	
	/** The panel that contains the components specific to tree based quantization modes */
	private JPanel treeModePanel;
	
	/** The panel that contains the components specific to uniform quantization modes */
	private JPanel uniformModePanel;
	
	/** The currently selected quantization mode */
	private int optionMode;
	
	/** Constructs a quantization object.
	 * @param q The quantization mode (IMAGE_QUANTIZATION or SIMPLIFY_QUANTIZATION)
	 */
	public Quantization(int q) {
		if (q == Quantization.IMAGE_QUANTIZATION
				|| q == Quantization.SIMPLIFY_QUANTIZATION) {
			optionMode = q;
		} else {
			throw new RuntimeException("Unknown quantization type");
		}
	}
	
	/**
	 * Constructs a quantization object.
	 */
	public Quantization() {
		optionMode = Quantization.SIMPLIFY_QUANTIZATION;
	}
	
	/**
	 * Returns the quantization parameter panel.
	 * @return The quantization parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the quantization parameter panel.
	 */
	private void createParameterPanel() {
		
		String detailChoices[] = new String[129];

		for (int i = 0; i < 128; i++) {
			detailChoices[i] = Integer.toString((i + 1));
		}
		detailChoices[128] = new String("INF");
		detailCombo = new JComboBox(detailChoices);
		detailCombo.setSelectedIndex(19);
		pointPlacement = new JComboBox(new String[] {"Box center", "Box extreme",
			"Box adjusted"});
		// pointPlacement.setSelectedIndex(2);
		typeBox = new JComboBox(new String[] {"Uniform", "Octree", "Binary tree"});
		typeBox.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED && modeChildPanel != null) {
					modeChildPanel.removeAll();
					if (typeBox.getSelectedIndex() == 0) {
						modeChildPanel.add(uniformModePanel, BorderLayout.CENTER);
						uniformModePanel.invalidate();
					} else {
						modeChildPanel.add(treeModePanel, BorderLayout.CENTER);
						treeModePanel.invalidate();
					}
					parameterPanel.doLayout();
					parameterPanel.validate();
					parameterPanel.repaint();
				}
			}
		});
		if (optionMode == Quantization.IMAGE_QUANTIZATION) {
			typeBox.setSelectedIndex(2);
		}
		priorityBox = new JComboBox(new String[] {"Type 1", "Type 2", "Type 3",
			"Type 4", "Type 5"});
		if (optionMode == Quantization.IMAGE_QUANTIZATION) {
			priorityBox.setSelectedIndex(4);
		}
		colorTarget = new JComboBox(new String[] {"16", "256", "512", "1024", "2048",
			"4096"});
		colorTarget.setSelectedIndex(1);
		JPanel commonPanel = new JPanel(new GridLayout(0, 2));

		commonPanel.add(new JLabel("Mode"));
		commonPanel.add(typeBox);
		modeChildPanel = new JPanel(new BorderLayout());
		uniformModePanel = new JPanel(new GridLayout(0, 2));
		// uniformModePanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Uniform mode"));
		uniformModePanel.add(new JLabel("Detail"));
		uniformModePanel.add(detailCombo);
		uniformModePanel.add(new JLabel("Point position"));
		uniformModePanel.add(pointPlacement);
		treeModePanel = new JPanel(new GridLayout(0, 2));
		// treeModePanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Tree mode"));
		treeModePanel.add(new JLabel("Quantized colors"));
		treeModePanel.add(colorTarget);
		treeModePanel.add(new JLabel("Color priority"));
		treeModePanel.add(priorityBox);
		
		if (typeBox.getSelectedIndex() == 0) {
			modeChildPanel.add(uniformModePanel, BorderLayout.CENTER);
		} else {
			modeChildPanel.add(treeModePanel, BorderLayout.CENTER);
		}
		
		parameterPanel = new JPanel();
		parameterPanel.setLayout(new BoxLayout(parameterPanel, BoxLayout.Y_AXIS));
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
				"Quantization"));
		parameterPanel.add(commonPanel);
		parameterPanel.add(modeChildPanel);
		// parameterPanel.add(treePanel);
	}
	
	/**
	 * Performs quantization of a number of points.
	 * @param qo The object that contains the original point data.
	 * @return An object that contains the qunatized point data.
	 */
	public QuantizedObject quantize(QuantizedObject qo) {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		int mode = typeBox.getSelectedIndex();

		if (mode == Quantization.OCTREE) {
			return quantizeOctree(qo);
		} 
		if (mode == Quantization.BINARY_TREE) {
			return quantizeBinarytree(qo); 
		}
		return quantizeSimple(qo);
	}
	
	/**
	 * Performs uniform quantization of a number of points.
	 * @param qo The object that contains the original point data.
	 * @return An object that contains the qunatized point data.
	 */
	public QuantizedObject quantizeSimple(QuantizedObject qo) {
		// System.out.println("Simple");
		int detail;

		if (detailCombo.getSelectedItem().toString().equals("INF")) {
			detail = INFINITY;
		} else {
			detail = Integer.parseInt(detailCombo.getSelectedItem().toString());
		}

		if (detail == Quantization.INFINITY) {
			return qo;
		}
		int nResultingColors = 0;
			
		int detailX = detail;
		int detailY = detail;
		int detailZ = detail;
		int i, j, k;
		int x, y, z, q;
		
		float lowX = qo.points[0];
		float lowY = qo.points[1];
		float lowZ = qo.points[2];
		
		float highX = qo.points[0];
		float highY = qo.points[1];
		float highZ = qo.points[2];
		
		int numInputColors = qo.points.length / 3;

		for (i = 1; i < numInputColors; i++) {
			q = i * 3;
			if (qo.points[q] > highX) {
				highX = qo.points[q];
			} else if (qo.points[q] < lowX) {
				lowX = qo.points[q];
			}
			
			if (qo.points[q + 1] > highY) {
				highY = qo.points[q + 1];
			} else if (qo.points[q + 1] < lowY) {
				lowY = qo.points[q + 1];
			}
			
			if (qo.points[q + 2] > highZ) {
				highZ = qo.points[q + 2];
			} else if (qo.points[q + 2] < lowZ) {
				lowZ = qo.points[q + 2];
			}
		}
		
		float xDiff = highX - lowX;

		if (xDiff <= 0.0f) {
			detailX = 1;
			lowX -= 0.1f;
			highX += 0.1f;
			xDiff = highX - lowX;
		} else {
			lowX -= xDiff * 0.005f;
			highX += xDiff * 0.005f;
			xDiff = highX - lowX;
		}
		float yDiff = highY - lowY;

		if (yDiff <= 0.0f) {
			detailY = 1;
			lowY -= 0.1f;
			highY += 0.1f;
			yDiff = highY - lowY;
		} else {
			lowY -= yDiff * 0.005f;
			highY += yDiff * 0.005f;
			yDiff = highY - lowY;
		}
		float zDiff = highZ - lowZ;

		if (zDiff <= 0.0f) {
			detailZ = 1;
			lowZ -= 0.1f;
			highZ += 0.1f;
			zDiff = highZ - lowZ;
		} else {
			lowZ -= zDiff * 0.005f;
			highZ += zDiff * 0.005f;
			zDiff = highZ - lowZ;
		}
		
		int pPosition = pointPlacement.getSelectedIndex();
		
		float xBoxSize = xDiff / detailX;
		float yBoxSize = yDiff / detailY;
		float zBoxSize = zDiff / detailZ;
		
		float adjustmentX = xBoxSize / 2.0f;
		float adjustmentY = yBoxSize / 2.0f;
		float adjustmentZ = zBoxSize / 2.0f;
		
		float centerX = (lowX + highX) / 2.0f;
		float centerY = (lowY + highY) / 2.0f;
		float centerZ = (lowZ + highZ) / 2.0f;
		
		int addCount[][][] = new int[detailX][detailY][detailZ];
		int cCount[][][] = new int[detailX][detailY][detailZ];
		int poIndex[][][] = new int[detailX][detailY][detailZ];
		int detIndex[] = new int[numInputColors * 3];
		
		for (i = 0; i < numInputColors; i++) {
			q = i * 3;
			detIndex[q] = (int) ((qo.points[q] - lowX) / xBoxSize);
			detIndex[q + 1] = (int) ((qo.points[q + 1] - lowY) / yBoxSize);
			detIndex[q + 2] = (int) ((qo.points[q + 2] - lowZ) / zBoxSize);
			
			if (addCount[detIndex[q]][detIndex[q + 1]][detIndex[q + 2]] == 0) {
				nResultingColors++;
			}
			addCount[detIndex[q]][detIndex[q + 1]][detIndex[q + 2]]++;
		}
		
		float pointResult[] = new float[nResultingColors * 3];
		int pointToColorCount[] = new int[nResultingColors];
		int pointToColorIndicesAdded[] = new int[nResultingColors];
		int pointToColorIndex[][] = new int[nResultingColors][];
		int pointToPixelCount[] = new int[nResultingColors];
		
		float centerDetailX = detailX / 2.0f;
		float centerDetailY = detailY / 2.0f;
		float centerDetailZ = detailZ / 2.0f;
		float maxDistX = centerDetailX;
		float maxDistY = centerDetailY;
		float maxDistZ = centerDetailZ;
		
		boolean he2 = false;
		
		int newIndex = 0;

		for (i = 0; i < detailX; i++) {
			for (j = 0; j < detailY; j++) {
				for (k = 0; k < detailZ; k++) {
					if (addCount[i][j][k] > 0) {
						if (pPosition == 1) {
							if (i * 2 < (detailX - 1)) {
								adjustmentX = -xBoxSize / 2.0f;
							} else if (i * 2 > (detailX - 1)) {
								adjustmentX = xBoxSize / 2.0f;
							} else {
								adjustmentX = 0;
							}
							if (j * 2 < (detailY - 1)) {
								adjustmentY = -yBoxSize / 2.0f;
							} else if (j * 2 > (detailY - 1)) {
								adjustmentY = yBoxSize / 2.0f;
							} else {
								adjustmentY = 0;
							}
							if (k * 2 < (detailZ - 1)) {
								adjustmentZ = -zBoxSize / 2.0f;
							} else if (k * 2 > (detailZ - 1)) {
								adjustmentZ = zBoxSize / 2.0f;
							} else {
								adjustmentZ = 0;
							}
						} else if (pPosition == 2) {
							adjustmentX = xBoxSize * ((i - centerDetailX) / maxDistX) / 2.0f;
							adjustmentY = yBoxSize * ((j - centerDetailY) / maxDistY) / 2.0f;
							adjustmentZ = zBoxSize * ((k - centerDetailZ) / maxDistZ) / 2.0f;

						} else {
							adjustmentX = 0;
							adjustmentY = 0;
							adjustmentZ = 0;
						}
						pointResult[newIndex * 3] = lowX + (xBoxSize * ((float) i + 0.5f))
								+ adjustmentX;
						pointResult[newIndex * 3 + 1] = lowY + (yBoxSize * ((float) j + 0.5f))
								+ adjustmentY;
						pointResult[newIndex * 3 + 2] = lowZ + (zBoxSize * ((float) k + 0.5f))
								+ adjustmentZ;
						pointToColorCount[newIndex] = addCount[i][j][k];
						pointToColorIndex[newIndex] = new int[pointToColorCount[newIndex]];
						poIndex[i][j][k] = newIndex;
						newIndex++;
					}
				}
			}
		}
		
		int index;

		for (i = 0; i < numInputColors; i++) {
			q = i * 3;
			index = poIndex[detIndex[q]][detIndex[q + 1]][detIndex[q + 2]];
			pointToColorIndex[index][pointToColorIndicesAdded[index]] = i;
			pointToColorIndicesAdded[index]++;
			if (qo.pointToPixelCount != null) {
				pointToPixelCount[index] += qo.pointToPixelCount[i];
			} else {
				pointToPixelCount[index]++;
			}
		}
		QuantizedObject result = new QuantizedObject();

		result.points = pointResult;
		result.pointToColorCount = pointToColorCount;
		result.pointToColorIndex = pointToColorIndex;
		result.pointToPixelCount = pointToPixelCount;
		result.numPoints = nResultingColors;
		return result;
	}
	
	/**
	 * Performs octree quantization of a number of points.
	 * @param qo The object that contains the original point data.
	 * @return An object that contains the qunatized point data.
	 */
	public QuantizedObject quantizeOctree(QuantizedObject qo) {
		int colorsWanted = Integer.parseInt(colorTarget.getSelectedItem().toString());
		
		// float lowX, lowY, lowZ, highX, highY, highZ, diffX, diffY, diffZ;
		
		float lowX = qo.points[0];
		float lowY = qo.points[1];
		float lowZ = qo.points[2];
		
		float highX = qo.points[0];
		float highY = qo.points[1];
		float highZ = qo.points[2];
		
		int numInputColors = qo.points.length / 3;
		int q;

		for (int i = 1; i < numInputColors; i++) {
			q = i * 3;
			if (qo.points[q] > highX) {
				highX = qo.points[q];
			} else if (qo.points[q] < lowX) {
				lowX = qo.points[q];
			}
			
			if (qo.points[q + 1] > highY) {
				highY = qo.points[q + 1];
			} else if (qo.points[q + 1] < lowY) {
				lowY = qo.points[q + 1];
			}
			
			if (qo.points[q + 2] > highZ) {
				highZ = qo.points[q + 2];
			} else if (qo.points[q + 2] < lowZ) {
				lowZ = qo.points[q + 2];
			}
		}
		
		float xDiff = highX - lowX;

		if (xDiff <= 0.0f) {
			lowX -= 0.1f;
			highX += 0.1f;
			xDiff = highX - lowX;
		} else {
			lowX -= xDiff * 0.005f;
			highX += xDiff * 0.005f;
			xDiff = highX - lowX;
		}
		float yDiff = highY - lowY;

		if (yDiff <= 0.0f) {
			lowY -= 0.1f;
			highY += 0.1f;
			yDiff = highY - lowY;
		} else {
			lowY -= yDiff * 0.005f;
			highY += yDiff * 0.005f;
			yDiff = highY - lowY;
		}
		float zDiff = highZ - lowZ;

		if (zDiff <= 0.0f) {
			lowZ -= 0.1f;
			highZ += 0.1f;
			zDiff = highZ - lowZ;
		} else {
			lowZ -= zDiff * 0.005f;
			highZ += zDiff * 0.005f;
			zDiff = highZ - lowZ;
		}
		
		CustomHeap heap = null;

		if (priorityBox.getSelectedIndex() == 0) {
			heap = new CustomHeap(new NodeComparator1());
		} else if (priorityBox.getSelectedIndex() == 1) {
			heap = new CustomHeap(new NodeComparator2(qo.pointToPixelCount));
		} else if (priorityBox.getSelectedIndex() == 2) {
			heap = new CustomHeap(new NodeComparator3(qo.pointToPixelCount, qo.points));
		} else if (priorityBox.getSelectedIndex() == 3) {
			heap = new CustomHeap(new NodeComparator4(qo.pointToPixelCount, qo.points));
		} else {
			heap = new CustomHeap(new NodeComparator5(qo.pointToPixelCount, qo.points));
		}
		ColorNode root = new ColorNode();

		root.minX = lowX;
		root.minY = lowY;
		root.minZ = lowZ;
		root.maxX = highX;
		root.maxY = highY;
		root.maxZ = highZ;
		root.centerX = (highX + lowX) / 2.0f;
		root.centerY = (highY + lowY) / 2.0f;
		root.centerZ = (highZ + lowZ) / 2.0f;
		root.colorIndex = new int[qo.numPoints];
		for (int i = 0; i < root.colorIndex.length; i++) {
			root.colorsRegistered++;
			root.colorIndex[i] = i;
		}
		heap.push(root, root);
		
		float colorData[] = qo.points; 
		int tempIndices[][] = new int[8][qo.numPoints];
		int currentColor, belongsInChild;
		
		// System.out.println("Colors wanted = "+colorsWanted);
		// int ccccc=0;
		while (heap.size() < colorsWanted) {
			ColorNode current = (ColorNode) heap.pop();

			// heap.remove(current);
			// System.out.println("Colors in current: "+current.colorsRegistered);
			// System.out.println("["+current.minX+" - "+current.centerX+" - "+current.maxX+"]");
			// System.out.println("["+current.minY+" - "+current.centerY+" - "+current.maxY+"]");
			// System.out.println("["+current.minZ+" - "+current.centerZ+" - "+current.maxZ+"]");
			// System.out.println("first color: "+colorData[
			if (current.colorsRegistered <= 1) {
				heap.push(current, current);
				break;
			}
			int colorsInChild[] = new int[8];

			for (int i = 0; i < current.colorsRegistered; i++) {
				currentColor = current.colorIndex[i];
				if (colorData[currentColor * 3] < current.centerX) {
					if (colorData[currentColor * 3 + 1] < current.centerY) {
						if (colorData[currentColor * 3 + 2] < current.centerZ) {
							belongsInChild = 0;
						} else {
							belongsInChild = 1;
						}
					} else {
						if (colorData[currentColor * 3 + 2] < current.centerZ) {
							belongsInChild = 2;
						} else {
							belongsInChild = 3;
						}
					}
				} else {
					if (colorData[currentColor * 3 + 1] < current.centerY) {
						if (colorData[currentColor * 3 + 2] < current.centerZ) {
							belongsInChild = 4;
						} else {
							belongsInChild = 5;
						}
					} else {
						if (colorData[currentColor * 3 + 2] < current.centerZ) {
							belongsInChild = 6;
						} else {
							belongsInChild = 7;
						}
					}
				}
				tempIndices[belongsInChild][colorsInChild[belongsInChild]++] = currentColor;
			}
			
			int children = 0;

			for (int j = 0; j < 8; j++) {
				// System.out.println(" - "+colorsInChild[j]);
				if (colorsInChild[j] > 0) {
					children++;
				}
			}
			if (heap.size() + children > colorsWanted) {
				heap.push(current, current);
				break;
			}
			
			if (colorsInChild[0] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[0];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[0], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.centerX;
				newNode.maxY = current.centerY;
				newNode.maxZ = current.centerZ;
				newNode.minX = current.minX;
				newNode.minY = current.minY;
				newNode.minZ = current.minZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[1] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[1];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[1], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.centerX;
				newNode.maxY = current.centerY;
				newNode.maxZ = current.maxZ;
				newNode.minX = current.minX;
				newNode.minY = current.minY;
				newNode.minZ = current.centerZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[2] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[2];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[2], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.centerX;
				newNode.maxY = current.maxY;
				newNode.maxZ = current.centerZ;
				newNode.minX = current.minX;
				newNode.minY = current.centerY;
				newNode.minZ = current.minZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[3] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[3];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[3], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.centerX;
				newNode.maxY = current.maxY;
				newNode.maxZ = current.maxZ;
				newNode.minX = current.minX;
				newNode.minY = current.centerY;
				newNode.minZ = current.centerZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[4] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[4];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[4], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.maxX;
				newNode.maxY = current.centerY;
				newNode.maxZ = current.centerZ;
				newNode.minX = current.centerX;
				newNode.minY = current.minY;
				newNode.minZ = current.minZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[5] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[5];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[5], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.maxX;
				newNode.maxY = current.centerY;
				newNode.maxZ = current.maxZ;
				newNode.minX = current.centerX;
				newNode.minY = current.minY;
				newNode.minZ = current.centerZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[6] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[6];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[6], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.maxX;
				newNode.maxY = current.maxY;
				newNode.maxZ = current.centerZ;
				newNode.minX = current.centerX;
				newNode.minY = current.centerY;
				newNode.minZ = current.minZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
			if (colorsInChild[7] > 0) {
				ColorNode newNode = new ColorNode();

				newNode.colorsRegistered = colorsInChild[7];
				newNode.colorIndex = new int[newNode.colorsRegistered];
				System.arraycopy(tempIndices[7], 0, newNode.colorIndex, 0,
						newNode.colorIndex.length);
				newNode.maxX = current.maxX;
				newNode.maxY = current.maxY;
				newNode.maxZ = current.maxZ;
				newNode.minX = current.centerX;
				newNode.minY = current.centerY;
				newNode.minZ = current.centerZ;
				newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
				newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
				newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
				heap.push(newNode, newNode);
			}
		
			/* if (ccccc++>1000)
			 break;
			 */
		}
		int resultingPoints = heap.size();
		QuantizedObject result = new QuantizedObject();

		result.numPoints = resultingPoints;
		result.points = new float[resultingPoints * 3];
		result.pointToColorCount = new int[resultingPoints];
		result.pointToColorIndex = new int[resultingPoints][];
		result.pointToPixelCount = new int[resultingPoints];
		float sumX, sumY, sumZ;
		int curPoint = 0;

		while (heap.size() > 0) {
			ColorNode current = (ColorNode) heap.pop();

			// heap.remove(current);
			sumX = 0;
			sumY = 0;
			sumZ = 0;
			for (int i = 0; i < current.colorsRegistered; i++) {
				result.pointToPixelCount[curPoint] += qo.pointToPixelCount[current.colorIndex[i]];
				sumX += colorData[current.colorIndex[i] * 3]
						* qo.pointToPixelCount[current.colorIndex[i]];
				sumY += colorData[current.colorIndex[i] * 3 + 1]
						* qo.pointToPixelCount[current.colorIndex[i]];
				sumZ += colorData[current.colorIndex[i] * 3 + 2]
						* qo.pointToPixelCount[current.colorIndex[i]];
			}
			result.points[curPoint * 3] = sumX / result.pointToPixelCount[curPoint];
			result.points[curPoint * 3 + 1] = sumY / result.pointToPixelCount[curPoint];
			result.points[curPoint * 3 + 2] = sumZ / result.pointToPixelCount[curPoint];

			/*
			 result.points[curPoint*3] = current.centerX;
			 result.points[curPoint*3+1] = current.centerY;
			 result.points[curPoint*3+2] = current.centerZ;
			 */
			result.pointToColorCount[curPoint] = current.colorsRegistered;
			result.pointToColorIndex[curPoint] = current.colorIndex;
			curPoint++;
		}
		return result;
	}
	
	/**
	 * Performs binary tree quantization of a number of points.
	 * @param qo The object that contains the original point data.
	 * @return An object that contains the qunatized point data.
	 */
	public QuantizedObject quantizeBinarytree(QuantizedObject qo) {
		int colorsWanted = Integer.parseInt(colorTarget.getSelectedItem().toString());
		
		// float lowX, lowY, lowZ, highX, highY, highZ, diffX, diffY, diffZ;
		
		float lowX = qo.points[0];
		float lowY = qo.points[1];
		float lowZ = qo.points[2];
		
		float highX = qo.points[0];
		float highY = qo.points[1];
		float highZ = qo.points[2];
		
		int numInputColors = qo.points.length / 3;
		int q;

		for (int i = 1; i < numInputColors; i++) {
			q = i * 3;
			if (qo.points[q] > highX) {
				highX = qo.points[q];
			} else if (qo.points[q] < lowX) {
				lowX = qo.points[q];
			}
			
			if (qo.points[q + 1] > highY) {
				highY = qo.points[q + 1];
			} else if (qo.points[q + 1] < lowY) {
				lowY = qo.points[q + 1];
			}
			
			if (qo.points[q + 2] > highZ) {
				highZ = qo.points[q + 2];
			} else if (qo.points[q + 2] < lowZ) {
				lowZ = qo.points[q + 2];
			}
		}
		
		float xDiff = highX - lowX;

		if (xDiff <= 0.0f) {
			lowX -= 0.1f;
			highX += 0.1f;
			xDiff = highX - lowX;
		} else {
			lowX -= xDiff * 0.005f;
			highX += xDiff * 0.005f;
			xDiff = highX - lowX;
		}
		float yDiff = highY - lowY;

		if (yDiff <= 0.0f) {
			lowY -= 0.1f;
			highY += 0.1f;
			yDiff = highY - lowY;
		} else {
			lowY -= yDiff * 0.005f;
			highY += yDiff * 0.005f;
			yDiff = highY - lowY;
		}
		float zDiff = highZ - lowZ;

		if (zDiff <= 0.0f) {
			lowZ -= 0.1f;
			highZ += 0.1f;
			zDiff = highZ - lowZ;
		} else {
			lowZ -= zDiff * 0.005f;
			highZ += zDiff * 0.005f;
			zDiff = highZ - lowZ;
		}
		
		CustomHeap heap = null;

		if (priorityBox.getSelectedIndex() == 0) {
			heap = new CustomHeap(new NodeComparator1());
		} else if (priorityBox.getSelectedIndex() == 1) {
			heap = new CustomHeap(new NodeComparator2(qo.pointToPixelCount));
		} else if (priorityBox.getSelectedIndex() == 2) {
			heap = new CustomHeap(new NodeComparator3(qo.pointToPixelCount, qo.points));
		} else if (priorityBox.getSelectedIndex() == 3) {
			heap = new CustomHeap(new NodeComparator4(qo.pointToPixelCount, qo.points));
		} else {
			heap = new CustomHeap(new NodeComparator5(qo.pointToPixelCount, qo.points));
		}
		ColorNode root = new ColorNode();

		root.minX = lowX;
		root.minY = lowY;
		root.minZ = lowZ;
		root.maxX = highX;
		root.maxY = highY;
		root.maxZ = highZ;
		root.centerX = (highX + lowX) / 2.0f;
		root.centerY = (highY + lowY) / 2.0f;
		root.centerZ = (highZ + lowZ) / 2.0f;
		root.colorIndex = new int[qo.numPoints];
		for (int i = 0; i < root.colorIndex.length; i++) {
			root.colorsRegistered++;
			root.colorIndex[i] = i;
		}
		heap.push(root, root);
		
		float colorData[] = qo.points; 
		int tempIndices[][] = new int[2][qo.numPoints];
		int currentColor, belongsInChild;
		
		// System.out.println("Colors wanted = "+colorsWanted);
		// int ccccc=0;
		while (heap.size() < colorsWanted) {
			ColorNode current = (ColorNode) heap.pop();

			// heap.remove(current);
			// System.out.println("Colors in current: "+current.colorsRegistered);
			// System.out.println("["+current.minX+" - "+current.centerX+" - "+current.maxX+"]");
			// System.out.println("["+current.minY+" - "+current.centerY+" - "+current.maxY+"]");
			// System.out.println("["+current.minZ+" - "+current.centerZ+" - "+current.maxZ+"]");
			// System.out.println("first color: "+colorData[
			if (current.colorsRegistered <= 1) {
				heap.push(current, current);
				break;
			}
			int colorsInChild[] = new int[2];
			
			int l = current.colorIndex[0];

			lowX = colorData[l * 3];
			lowY = colorData[l * 3 + 1];
			lowZ = colorData[l * 3 + 2];
			highX = colorData[l * 3];
			highY = colorData[l * 3 + 1];
			highZ = colorData[l * 3 + 2];
			for (int i = 1; i < current.colorsRegistered; i++) {
				currentColor = current.colorIndex[i];
				if (colorData[currentColor * 3] < lowX) {
					lowX = colorData[currentColor * 3];
				} else if (colorData[currentColor * 3] > highX) {
					highX = colorData[currentColor * 3];
				}
				
				if (colorData[currentColor * 3 + 1] < lowY) {
					lowY = colorData[currentColor * 3 + 1];
				} else if (colorData[currentColor * 3 + 1] > highY) {
					highY = colorData[currentColor * 3 + 1];
				}
				
				if (colorData[currentColor * 3 + 2] < lowZ) {
					lowZ = colorData[currentColor * 3 + 2];
				} else if (colorData[currentColor * 3 + 2] > highZ) {
					highZ = colorData[currentColor * 3 + 2];
				}
			}
			xDiff = highX - lowX;
			yDiff = highY - lowY;
			zDiff = highZ - lowZ;
			if (xDiff > yDiff && xDiff > zDiff) {
				for (int i = 0; i < current.colorsRegistered; i++) {
					currentColor = current.colorIndex[i];
					if (colorData[currentColor * 3] < current.centerX) {
						belongsInChild = 0;
					} else {
						belongsInChild = 1;
					}
					tempIndices[belongsInChild][colorsInChild[belongsInChild]++] = currentColor;
				}
				
				if (colorsInChild[0] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[0];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[0], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.centerX;
					newNode.maxY = current.maxY;
					newNode.maxZ = current.maxZ;
					newNode.minX = current.minX;
					newNode.minY = current.minY;
					newNode.minZ = current.minZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}
				if (colorsInChild[1] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[1];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[1], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.maxX;
					newNode.maxY = current.maxY;
					newNode.maxZ = current.maxZ;
					newNode.minX = current.centerX;
					newNode.minY = current.minY;
					newNode.minZ = current.minZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}
				
			} else if (yDiff > zDiff) {
				for (int i = 0; i < current.colorsRegistered; i++) {
					currentColor = current.colorIndex[i];
					if (colorData[currentColor * 3 + 1] < current.centerY) {
						belongsInChild = 0;
					} else {
						belongsInChild = 1;
					}
					tempIndices[belongsInChild][colorsInChild[belongsInChild]++] = currentColor;
				}
				
				if (colorsInChild[0] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[0];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[0], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.maxX;
					newNode.maxY = current.centerY;
					newNode.maxZ = current.maxZ;
					newNode.minX = current.minX;
					newNode.minY = current.minY;
					newNode.minZ = current.minZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}
				if (colorsInChild[1] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[1];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[1], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.maxX;
					newNode.maxY = current.maxY;
					newNode.maxZ = current.maxZ;
					newNode.minX = current.minX;
					newNode.minY = current.centerY;
					newNode.minZ = current.minZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}
				
			} else {
				for (int i = 0; i < current.colorsRegistered; i++) {
					currentColor = current.colorIndex[i];
					if (colorData[currentColor * 3 + 2] < current.centerZ) {
						belongsInChild = 0;
					} else {
						belongsInChild = 1;
					}
					tempIndices[belongsInChild][colorsInChild[belongsInChild]++] = currentColor;
				}
				
				if (colorsInChild[0] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[0];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[0], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.maxX;
					newNode.maxY = current.maxY;
					newNode.maxZ = current.centerZ;
					newNode.minX = current.minX;
					newNode.minY = current.minY;
					newNode.minZ = current.minZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}
				if (colorsInChild[1] > 0) {
					ColorNode newNode = new ColorNode();

					newNode.colorsRegistered = colorsInChild[1];
					newNode.colorIndex = new int[newNode.colorsRegistered];
					System.arraycopy(tempIndices[1], 0, newNode.colorIndex, 0,
							newNode.colorIndex.length);
					newNode.maxX = current.maxX;
					newNode.maxY = current.maxY;
					newNode.maxZ = current.maxZ;
					newNode.minX = current.minX;
					newNode.minY = current.minY;
					newNode.minZ = current.centerZ;
					newNode.centerX = (newNode.minX + newNode.maxX) / 2.0f;
					newNode.centerY = (newNode.minY + newNode.maxY) / 2.0f;
					newNode.centerZ = (newNode.minZ + newNode.maxZ) / 2.0f;
					heap.push(newNode, newNode);
				}	
			}
		}
		int resultingPoints = heap.size();
		QuantizedObject result = new QuantizedObject();

		result.numPoints = resultingPoints;
		result.points = new float[resultingPoints * 3];
		result.pointToColorCount = new int[resultingPoints];
		result.pointToColorIndex = new int[resultingPoints][];
		result.pointToPixelCount = new int[resultingPoints];
		float sumX, sumY, sumZ;
		int curPoint = 0;

		while (heap.size() > 0) {
			ColorNode current = (ColorNode) heap.pop();

			// heap.remove(current);
			sumX = 0;
			sumY = 0;
			sumZ = 0;
			for (int i = 0; i < current.colorsRegistered; i++) {
				result.pointToPixelCount[curPoint] += qo.pointToPixelCount[current.colorIndex[i]];
				sumX += colorData[current.colorIndex[i] * 3]
						* qo.pointToPixelCount[current.colorIndex[i]];
				sumY += colorData[current.colorIndex[i] * 3 + 1]
						* qo.pointToPixelCount[current.colorIndex[i]];
				sumZ += colorData[current.colorIndex[i] * 3 + 2]
						* qo.pointToPixelCount[current.colorIndex[i]];
			}
			result.points[curPoint * 3] = sumX / result.pointToPixelCount[curPoint];
			result.points[curPoint * 3 + 1] = sumY / result.pointToPixelCount[curPoint];
			result.points[curPoint * 3 + 2] = sumZ / result.pointToPixelCount[curPoint];

			/*
			 result.points[curPoint*3] = current.centerX;
			 result.points[curPoint*3+1] = current.centerY;
			 result.points[curPoint*3+2] = current.centerZ;
			 */
			result.pointToColorCount[curPoint] = current.colorsRegistered;
			result.pointToColorIndex[curPoint] = current.colorIndex;
			curPoint++;
		}
		return result;
	}
}


/**
 * A node that represents a box in in 3D space and the colors that it contains.
 */
class ColorNode {
	
	/** The number of colors in this node */
	public int colorsRegistered;
	
	/** The x coordinate of the node center */
	public float centerX;
	
	/** The y coordinate of the node center */
	public float centerY;
	
	/** The z coordinate of the node center */
	public float centerZ;
	
	/** The minimum x coordinate */
	public float minX;
	
	/** The minimum y coordinate */
	public float minY;
	
	/** The minimum z coordinate */
	public float minZ;
	
	/** The maximum x coordinate */
	public float maxX;
	
	/** The maximum y coordinate */
	public float maxY;
	
	/** The maximum z coordinate */
	public float maxZ;
	
	/** The indices of the original colors that this node represents */
	public int colorIndex[];
	
	/** Used to compare the priorities of two nodes */
	public float comparatorData;
}


/**
 * A comparator that prioritizes nodes that contain the most colors.
 */
class NodeComparator1 implements Comparator {
	
	/**
	 * Compares two color nodes.
	 * @param a Node a.
	 * @param b Node b.
	 * @return 0 if the two nodes have the same priority,
	 * 1 if node a has got a higher priority than b,
	 * or -1 if node b has got a higher priority than a.
	 */
	public int compare(Object a, Object b) {
		if (((ColorNode) a).colorsRegistered > ((ColorNode) b).colorsRegistered) {
			return 1;
		}
		if (((ColorNode) a).colorsRegistered < ((ColorNode) b).colorsRegistered) {
			return -1;
		}
		return 0;
	}
}


/**
 * A comparator that prioritizes nodes that represent the most image pixels.
 */
class NodeComparator2 implements Comparator {
	
	/** The pixel counts for each of the original colors */
	private int numPix[];
	
	/**
	 * Constructs a node comparator.
	 * @param pix The pixel counts for each of the original colors.
	 */
	public NodeComparator2(int pix[]) {
		numPix = pix;
	}
	
	/**
	 * Compares two color nodes.
	 * @param a Node a.
	 * @param b Node b.
	 * @return 0 if the two nodes have the same priority,
	 * 1 if node a has got a higher priority than b,
	 * or -1 if node b has got a higher priority than a.
	 */
	public int compare(Object a, Object b) {
		ColorNode nodea = ((ColorNode) a);
		ColorNode nodeb = ((ColorNode) b);

		// System.out.println("A "+a+", B "+b);
		if (nodea.colorsRegistered == 1 && nodeb.colorsRegistered > 1) {
			return -1;
		}
		if (nodea.colorsRegistered > 1 && nodeb.colorsRegistered == 1) {
			return 1;
		}
		int pxa = 0, pxb = 0;

		for (int i = 0; i < nodea.colorsRegistered; i++) {
			pxa += numPix[nodea.colorIndex[i]];
		}
		
		for (int i = 0; i < nodeb.colorsRegistered; i++) {
			pxb += numPix[nodeb.colorIndex[i]];
		}
		
		if (pxa < pxb) {
			return -1;
		}
		if (pxb < pxa) {
			return 1;
		}
		
		return 0;
	}
}


/**
 * A comparator that prioritizes nodes with a high color error sum.
 */
class NodeComparator3 implements Comparator {
	
	/** The pixel counts for each of the original colors */
	private int numPix[];
	
	/** The color positions for each of the original colors */
	private float colorData[];
	
	/**
	 * Constructs a node comparator.
	 * @param pix The pixel counts for each of the original colors.
	 * @param colors The color positions for each of the original colors.
	 */
	public NodeComparator3(int pix[], float colors[]) {
		numPix = pix;
		colorData = colors;
	}
	
	/**
	 * Compares two color nodes.
	 * @param a Node a.
	 * @param b Node b.
	 * @return 0 if the two nodes have the same priority,
	 * 1 if node a has got a higher priority than b,
	 * or -1 if node b has got a higher priority than a.
	 */
	public int compare(Object a, Object b) {
		ColorNode nodea = ((ColorNode) a);
		ColorNode nodeb = ((ColorNode) b);
		
		// System.out.println("A "+a+", B "+b);
		if (nodea.colorsRegistered == 1 && nodeb.colorsRegistered > 1) {
			return -1;
		}
		if (nodea.colorsRegistered > 1 && nodeb.colorsRegistered == 1) {
			return 1;
		}
		
		if (nodea.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				totalDistance += numPix[k]
						* Math.sqrt((colorData[k * 3] - pointX) * (colorData[k * 3] - pointX)
								+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
								+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
			}
			nodea.comparatorData = totalDistance;
		}
		if (nodeb.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				totalDistance += numPix[k]
						* Math.sqrt((colorData[k * 3] - pointX) * (colorData[k * 3] - pointX)
								+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
								+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
			}
			nodeb.comparatorData = totalDistance;
		}
		
		if (nodea.comparatorData < nodeb.comparatorData) {
			return -1;
		}
		if (nodeb.comparatorData < nodea.comparatorData) {
			return 1;
		}
		
		return 0;
	}
}


/**
 * A comparator that prioritizes nodes with a high color error sum, but
 * assigns a higher weight to the color difference over the pixel count
 * compared to NodeComparator3.
 */
class NodeComparator4 implements Comparator {
	
	/** The pixel counts for each of the original colors */
	private int numPix[];
	
	/** The color positions for each of the original colors */
	private float colorData[];
	
	/**
	 * Constructs a node comparator.
	 * @param pix The pixel counts for each of the original colors.
	 * @param colors The color positions for each of the original colors.
	 */
	public NodeComparator4(int pix[], float colors[]) {
		numPix = pix;
		colorData = colors;
	}
	
	/**
	 * Compares two color nodes.
	 * @param a Node a.
	 * @param b Node b.
	 * @return 0 if the two nodes have the same priority,
	 * 1 if node a has got a higher priority than b,
	 * or -1 if node b has got a higher priority than a.
	 */
	public int compare(Object a, Object b) {
		ColorNode nodea = ((ColorNode) a);
		ColorNode nodeb = ((ColorNode) b);
		
		// System.out.println("A "+a+", B "+b);
		if (nodea.colorsRegistered == 1 && nodeb.colorsRegistered > 1) {
			return -1;
		}
		if (nodea.colorsRegistered > 1 && nodeb.colorsRegistered == 1) {
			return 1;
		}
		
		if (nodea.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				totalDistance += (float) Math.pow(numPix[k], 0.33)
						* Math.sqrt((colorData[k * 3] - pointX) * (colorData[k * 3] - pointX)
								+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
								+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
			}
			nodea.comparatorData = totalDistance;
		}
		if (nodeb.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				totalDistance += (float) Math.pow(numPix[k], 0.33)
						* Math.sqrt((colorData[k * 3] - pointX) * (colorData[k * 3] - pointX)
								+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
								+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
			}
			nodeb.comparatorData = totalDistance;
		}
		
		if (nodea.comparatorData < nodeb.comparatorData) {
			return -1;
		}
		if (nodeb.comparatorData < nodea.comparatorData) {
			return 1;
		}
		
		return 0;
	}
}


/**
 * A comparator that prioritizes nodes with a high color error sum, but
 * assigns a higher weight to the color difference over the pixel count
 * compared to NodeComparator3.
 */
class NodeComparator5 implements Comparator {
	
	/** The pixel counts for each of the original colors */
	private int numPix[];
	
	/** The color positions for each of the original colors */
	private float colorData[];
	
	/**
	 * Constructs a node comparator.
	 * @param pix The pixel counts for each of the original colors.
	 * @param colors The color positions for each of the original colors.
	 */
	public NodeComparator5(int pix[], float colors[]) {
		numPix = pix;
		colorData = colors;
	}
	
	/**
	 * Compares two color nodes.
	 * @param a Node a.
	 * @param b Node b.
	 * @return 0 if the two nodes have the same priority,
	 * 1 if node a has got a higher priority than b,
	 * or -1 if node b has got a higher priority than a.
	 */
	public int compare(Object a, Object b) {
		ColorNode nodea = ((ColorNode) a);
		ColorNode nodeb = ((ColorNode) b);
		
		// System.out.println("A "+a+", B "+b);
		if (nodea.colorsRegistered == 1 && nodeb.colorsRegistered > 1) {
			return -1;
		}
		if (nodea.colorsRegistered > 1 && nodeb.colorsRegistered == 1) {
			return 1;
		}
		
		if (nodea.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;
			float fTemp;

			for (int i = 0; i < nodea.colorsRegistered; i++) {
				k = nodea.colorIndex[i];
				fTemp = (float) Math.sqrt((colorData[k * 3] - pointX)
								* (colorData[k * 3] - pointX)
						+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
						+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
				if (fTemp > totalDistance) {
					totalDistance = fTemp;
				}
			}
			nodea.comparatorData = totalDistance;
		}
		if (nodeb.comparatorData == 0.0f) {
			float sumX = 0;
			float sumY = 0;
			float sumZ = 0;
			float pointX, pointY, pointZ;
			int total = 0;
			int k;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				total += numPix[k];
				sumX += colorData[k * 3] * numPix[k];
				sumY += colorData[k * 3 + 1] * numPix[k];
				sumZ += colorData[k * 3 + 2] * numPix[k];
			}
			pointX = sumX / total;
			pointY = sumY / total;
			pointZ = sumZ / total;
			float totalDistance = 0.0f;
			float fTemp;

			for (int i = 0; i < nodeb.colorsRegistered; i++) {
				k = nodeb.colorIndex[i];
				fTemp = (float) Math.sqrt((colorData[k * 3] - pointX)
								* (colorData[k * 3] - pointX)
						+ (colorData[k * 3 + 1] - pointY) * (colorData[k * 3 + 1] - pointY)
						+ (colorData[k * 3 + 2] - pointZ) * (colorData[k * 3 + 2] - pointZ));
				if (fTemp > totalDistance) {
					totalDistance = fTemp;
				}
			}
			nodeb.comparatorData = totalDistance;
		}
		
		if (nodea.comparatorData < nodeb.comparatorData) {
			return -1;
		}
		if (nodeb.comparatorData < nodea.comparatorData) {
			return 1;
		}
		
		return 0;
	}
}
