/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader;


import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A generic TIFF tag implementation. 
 */
public class GenericTag implements Comparable {
	
	/** A tag of this type contains byte data. */
	public static final int TYPE_BYTE = 1;
	
	/** A tag of this type contains ASCII data. */
	public static final int TYPE_ASCII = 2;
	
	/** A tag of this type contains short integer data. */
	public static final int TYPE_SHORT = 3;
	
	/** A tag of this type contains long integer data. */
	public static final int TYPE_LONG = 4;
	
	/** A tag of this type contains rational number data. */
	public static final int TYPE_RATIONAL = 5;
	
	/** A tag of this type contains signed byte data. */
	public static final int TYPE_SBYTE = 6;
	
	/** A tag of this type contains undefined data. */
	public static final int TYPE_UNDEFINED = 7;
	
	/** A tag of this type contains signed short integer data. */
	public static final int TYPE_SSHORT = 8;
	
	/** A tag of this type contains signed long integer data. */
	public static final int TYPE_SLONG = 9;
	
	/** A tag of this type contains signed rational number data. */
	public static final int TYPE_SRATIONAL = 10;
	
	/** A tag of this type contains floating point data. */
	public static final int TYPE_FLOAT = 11;
	
	/** A tag of this type contains double floating point data. */
	public static final int TYPE_DOUBLE = 12;
	
	/** A tag of this type contains unknown data. */
	public static final int TYPE_ICC3D_UNKNOWN = -1;
	
	/** A tag that defines a new subfile. */
	public static final int TAG_NEW_SUBFILE_TYPE = 254;
	
	/** A tag that defines the type of a subfile. */
	public static final int TAG_SUBFILE_TYPE = 255;
	
	/** A tag that defines the width of an image. */
	public static final int TAG_IMAGE_WIDTH = 256;
	
	/** A tag that defines the height of an image. */
	public static final int TAG_IMAGE_LENGTH = 257;
	
	/** A tag that defines the bits per sample. */
	public static final int TAG_BITS_PER_SAMPLE = 258;
	
	/** A tag that defines the image compression used. */
	public static final int TAG_COMPRESSION = 259;
	
	/** A tag that determines how image data should be interpreted. */
	public static final int TAG_PHOTOMETRIC_INTERPRETATION = 262;
	
	/** A TIFF file tag. */
	public static final int TAG_THRESHHOLDING = 263;
	
	/** A TIFF file tag. */
	public static final int TAG_CELL_WIDTH = 264;
	
	/** A TIFF file tag. */
	public static final int TAG_CELL_LENGTH = 265;
	
	/** A TIFF file tag. */
	public static final int TAG_FILL_ORDER = 266;
	
	/** A TIFF file tag. */
	public static final int TAG_DOCUMENT_NAME = 269;
	
	/** A TIFF file tag. */
	public static final int TAG_IMAGE_DESCRIPTION = 270;
	
	/** A TIFF file tag. */
	public static final int TAG_MAKE = 271;
	
	/** A TIFF file tag. */
	public static final int TAG_MODEL = 272;
	
	/** A TIFF file tag. */
	public static final int TAG_STRIP_OFFSETS = 273;
	
	/** A TIFF file tag. */
	public static final int TAG_ORIENTATION = 274;
	
	/** A TIFF file tag. */
	public static final int TAG_SAMPLES_PER_PIXEL = 277;
	
	/** A TIFF file tag. */
	public static final int TAG_ROWS_PER_STRIP = 278;
	
	/** A TIFF file tag. */
	public static final int TAG_STRIP_BYTE_COUNTS = 279;
	
	/** A TIFF file tag. */
	public static final int TAG_MIN_SAMPLE_VALUE = 280;
	
	/** A TIFF file tag. */
	public static final int TAG_MAX_SAMPLE_VALUE = 281;
	
	/** A TIFF file tag. */
	public static final int TAG_X_RESOLUTION = 282;
	
	/** A TIFF file tag. */
	public static final int TAG_Y_RESOLUTION = 283;
	
	/** A TIFF file tag. */
	public static final int TAG_PLANAR_CONFIGURATION = 284;
	
	/** A TIFF file tag. */
	public static final int TAG_PAGE_NAME = 285;
	
	/** A TIFF file tag. */
	public static final int TAG_X_POSITION = 286;
	
	/** A TIFF file tag. */
	public static final int TAG_Y_POSITION = 287;
	
	/** A TIFF file tag. */
	public static final int TAG_T4_OPTIONS = 292;
	
	/** A TIFF file tag. */
	public static final int TAG_T6_OPTIONS = 293;
	
	/** A TIFF file tag. */
	public static final int TAG_RESOLUTION_UNIT = 296;
	
	/** A TIFF file tag. */
	public static final int TAG_PAGE_NUMBER = 297;
	
	/** A TIFF file tag. */
	public static final int TAG_TRANSFER_FUNCTION = 301;
	
	/** A TIFF file tag. */
	public static final int TAG_SOFTWARE = 305;
	
	/** A TIFF file tag. */
	public static final int TAG_DATETIME = 306;
	
	/** A TIFF file tag. */
	public static final int TAG_ARTIST = 315;
	
	/** A TIFF file tag. */
	public static final int TAG_HOST_COMPUTER = 316;
	
	/** A TIFF file tag. */
	public static final int TAG_PREDICTOR = 317;
	
	/** A TIFF file tag. */
	public static final int TAG_WHITE_POINT = 318;
	
	/** A TIFF file tag. */
	public static final int TAG_PRIMARY_CHROMATICITIES = 319;
	
	/** A TIFF file tag. */
	public static final int TAG_COLOR_MAP = 320;
	
	/** A TIFF file tag. */
	public static final int TAG_HALFTONE_HINTS = 321;
	
	/** A TIFF file tag. */
	public static final int TAG_TILE_WIDTH = 322;
	
	/** A TIFF file tag. */
	public static final int TAG_TILE_LENGTH = 323;
	
	/** A TIFF file tag. */
	public static final int TAG_TILE_OFFSETS = 324;
	
	/** A TIFF file tag. */
	public static final int TAG_TILE_BYTE_COUNTS = 325;
	
	/** A TIFF file tag. */
	public static final int TAG_INK_SET = 332;
	
	/** A TIFF file tag. */
	public static final int TAG_INK_NAMES = 333;
	
	/** A TIFF file tag. */
	public static final int TAG_NUMBER_OF_INKS = 334;
	
	/** A TIFF file tag. */
	public static final int TAG_DOT_RANGE = 336;
	
	/** A TIFF file tag. */
	public static final int TAG_TARGET_PRINTER = 337;
	
	/** A TIFF file tag. */
	public static final int TAG_EXTRA_SAMPLES = 338;
	
	/** A TIFF file tag. */
	public static final int TAG_SAMPLE_FORMAT = 339;
	
	/** A TIFF file tag. */
	public static final int TAG_S_MIN_SAMPLE_VALUE = 340;
	
	/** A TIFF file tag. */
	public static final int TAG_S_MAX_SAMPLE_VALUE = 341;
	
	/** A TIFF file tag. */
	public static final int TAG_TRANSFER_RANGE = 342;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_PROC = 512;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_INTERCHANGE_FORMAT = 513;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_INTERCHANGE_FORMAT_LENGTH = 514;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_RESTART_INTERVAL = 515;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_LOSSLESS_PREDICTORS = 517;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_POINT_TRANSFORMS = 518;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_Q_TABLES = 519;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_DCT_TABLES = 520;
	
	/** A TIFF file tag. */
	public static final int TAG_JPEG_AC_TABLES = 521;
	
	/** A TIFF file tag. */
	public static final int TAG_YCC_COEFFICIENTS = 529;
	
	/** A TIFF file tag. */
	public static final int TAG_YCC_SUBSAMPLING = 530;
	
	/** A TIFF file tag. */
	public static final int TAG_YCC_POSITIONING = 531;
	
	/** A TIFF file tag. */
	public static final int TAG_REFERENCE_BLACK_WHITE = 532;
	
	/** A TIFF file tag. */
	public static final int TAG_COPYRIGHT = 33432;
	
	/** A TIFF file tag. */
	public static final int TAG_ICC_PROFILE = 34675;
	
	/**
	 * Returns the TIFF tag of the given type, or null if no such tag is found.
	 *
	 * @param gt The available tags.
	 * @param tagid The id of the requested tag.
	 * @return The requested TIFF tag, or null if no tag is found.
	 */
	public static GenericTag getTag(GenericTag gt[], int tagid) {
		for (int i = 0; i < gt.length; i++) {
			if (gt[i].id == tagid) {
				return gt[i];
			}
		}
		return null;
	}
	
	/**
	 * Create a tag that contains short integer data, with the given id.
	 * @param d The integer data.
	 * @param tagid The id of the tag.
	 * @return The new TIFF tag.
	 */
	public static GenericTag createShortTag(int d[], int tagid) {
		return new GenericTag(d, tagid, GenericTag.TYPE_SHORT, d.length);
	}
	
	/**
	 * Create a tag that contains long integer data, with the given id.
	 * @param d The integer data.
	 * @param tagid The id of the tag.
	 * @return The new TIFF tag.
	 */
	public static GenericTag createLongTag(long d[], int tagid) {
		return new GenericTag(d, tagid, GenericTag.TYPE_LONG, d.length);
	}
	
	/**
	 * Create a tag that contains text data, with the given id.
	 * @param s The text.
	 * @param tagid The id of the tag.
	 * @return The new TIFF tag.
	 */
	public static GenericTag createTextTag(String s, int tagid) {
		byte b[] = s.getBytes();
		byte b2[] = new byte[b.length + 1];

		System.arraycopy(b, 0, b2, 0, b.length);
		return new GenericTag(b2, tagid, GenericTag.TYPE_ASCII, b2.length);
	}
	
	/**
	 * Create a tag that contains rational number data, with the given id.
	 * @param d The floating point data.
	 * @param tagid The id of the tag.
	 * @return The new TIFF tag.
	 */
	public static GenericTag createRationalTag(float d[], int tagid) {
		return new GenericTag(d, tagid, GenericTag.TYPE_RATIONAL, d.length);
	}
	
	/**
	 * Returns the size of one primitive of the given type.
	 * @param datatype The TIFF tag data type.
	 * @return The size of the data type.
	 */
	public static int sizeOfData(int datatype) {
		switch (datatype) {
		case GenericTag.TYPE_UNDEFINED:
		case GenericTag.TYPE_ASCII:
		case GenericTag.TYPE_BYTE:
		case GenericTag.TYPE_SBYTE:
			return 1;

		case GenericTag.TYPE_SHORT:
		case GenericTag.TYPE_SSHORT:
			return 2;

		case GenericTag.TYPE_DOUBLE:
		case GenericTag.TYPE_RATIONAL:
		case GenericTag.TYPE_SRATIONAL:
			return 8;

		case GenericTag.TYPE_LONG:
		case GenericTag.TYPE_SLONG:
		case GenericTag.TYPE_FLOAT:
			return 4;

		default:
			return 16;
		}
	}
	
	/** The TIFF tag data */
	protected Object data;
	
	/** The size of the data that is stored outside the entry. */
	protected long bytesOfDataOutsideEntry;
	
	/** The TIFF tag id */
	protected int id;
	
	/** The TIFF tag data type */
	protected int type;
	
	/** The TIFF tag data primitive count */
	protected int count;
	
	/**
	 * Constructs a copy of the TIFF tag.
	 * @param gt The TIFF tag to be copied.
	 */
	public GenericTag(GenericTag gt) {
		bytesOfDataOutsideEntry = gt.bytesOfDataOutsideEntry;
		data = gt.data;
		id = gt.id;
		type = gt.type;
		count = gt.count;
	}
	
	/**
	 * Constructs a TIFF tag.
	 * @param d The data.
	 * @param tagid The tag id.
	 * @param datatype The tag data type.
	 * @param datacount The number of data primitives.
	 */
	public GenericTag(Object d, int tagid, int datatype, int datacount) {
		data = d;
		id = tagid;
		type = datatype;
		count = datacount;
		long l = count * sizeOfData(datatype);

		if (l > 4) {
			bytesOfDataOutsideEntry = l;
		} else {
			bytesOfDataOutsideEntry = 0;
		}
	}
	
	/**
	 * Creates a TIFF tag from a file at a certain offset.
	 * @param rafw The TIFF file.
	 * @param offset The offset where the tag starts.
	 */
	public GenericTag(RandomAccessFileWrapper rafw, long offset) {
		rafw.setOffset(offset);
		id = rafw.getInt();
		type = rafw.getInt();
		count = (int) rafw.getLong();
		long ll = count * sizeOfData(type);

		if (ll > 4) {
			bytesOfDataOutsideEntry = ll;
		} else {
			bytesOfDataOutsideEntry = 0;
		}
		switch (type) {
		case TYPE_ASCII:
		case TYPE_SBYTE:
		case TYPE_UNDEFINED:
			byte b[] = new byte[count];

			if (count > 4) {
				rafw.setOffset(rafw.getLong());
			}
			rafw.getBytes(b);
			data = b;
			break;
				
		case TYPE_SHORT:
			int i[] = new int[count];

			if (count > 2) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				i[j] = rafw.getInt();
			}
			data = i;
			break;
				
		case TYPE_LONG:
			long l[] = new long[count];

			if (count > 1) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				l[j] = rafw.getLong();
			}
			data = l;
			break;
				
		case TYPE_BYTE:
			char c[] = new char[count];

			if (count > 4) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				c[j] = rafw.getChar();
			}
			data = c;
			break;
			
		case TYPE_SSHORT:
			char c2[] = new char[count];

			if (count > 2) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				c2[j] = (char) rafw.getInt();
			}
			data = c2;
			break;
				
		case TYPE_SLONG:
			int i2[] = new int[count];

			if (count > 1) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				i2[j] = (int) rafw.getLong();
			}
			data = i2;
			break;
				
		case TYPE_RATIONAL:
			float f[] = new float[count];

			rafw.setOffset(rafw.getLong());
			for (int j = 0; j < count; j++) {
				long l1 = rafw.getLong();
				long l2 = rafw.getLong();

				f[j] = (float) l1 / (float) l2;
			}
			data = f;
			break;
				
		case TYPE_SRATIONAL:
			float f3[] = new float[count];

			rafw.setOffset(rafw.getLong());

			for (int j = 0; j < count; j++) {
				int i11 = (int) rafw.getLong();
				int i22 = (int) rafw.getLong();

				f3[j] = (float) i11 / (float) i22;
			}
			data = f3;
			break;
			
		case TYPE_FLOAT:
			float f2[] = new float[count];

			if (count > 1) {
				rafw.setOffset(rafw.getLong());
			}
			for (int j = 0; j < count; j++) {
				f2[j] = rafw.getFloat();
			}
			data = f2;
			break;
				
		case TYPE_DOUBLE:
			double d[] = new double[count];

			rafw.setOffset(rafw.getLong());
			for (int j = 0; j < count; j++) {
				d[j] = rafw.getDouble();
			}
			data = d;
			break;
				
		default:
			type = TYPE_ICC3D_UNKNOWN;
			break;
		}
		rafw.setOffset(offset + 12);
	}
	
	/**
	 * Writes the tag to a file.
	 * @param rafw The TIFF file.
	 * @param entryOffset The offset of the entry.
	 * @param dataOffset The offset of the data.
	 * @return The next data offset. 
	 */
	public long writeTag(RandomAccessFileWrapper rafw, long entryOffset, long dataOffset) {
		dataOffset = ensureWordBoundary(dataOffset);
		rafw.setOffset(entryOffset);
		rafw.put((char) id);
		rafw.put((char) type);
		rafw.put(count);
		switch (type) {
		case TYPE_ASCII:
		case TYPE_UNDEFINED:
		case TYPE_SBYTE:
			if (count > 4) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count;
			}
			for (int i = 0; i < count; i++) {
				rafw.put(((byte[]) data)[i]);
			}
			break;
			
		case TYPE_BYTE:
			if (count > 4) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((byte) (((char[]) data)[i]));
			}
			break;
			
		case TYPE_SHORT:
			if (count > 2) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count * 2;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((char) (((int[]) data)[i]));
			}
			break;
				
		case TYPE_LONG:
			if (count > 1) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count * 4;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((int) (((long[]) data)[i]));
			}
			break;
			
		case TYPE_SSHORT:
			if (count > 2) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count * 2;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((char) (((char[]) data)[i]));
			}
			break;
				
		case TYPE_SLONG:
			if (count > 1) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count * 4;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((int) (((int[]) data)[i]));
			}
			break;
			
		case TYPE_FLOAT:
			if (count > 1) {
				rafw.put((int) dataOffset);
				rafw.setOffset(dataOffset);
				dataOffset += count * 4;
			}
			for (int i = 0; i < count; i++) {
				rafw.put((float) (((float[]) data)[i]));
			}
			break;
				
		case TYPE_DOUBLE:
			rafw.put((int) dataOffset);
			rafw.setOffset(dataOffset);
			dataOffset += count * 8;

			for (int i = 0; i < count; i++) {
				rafw.put((double) (((double[]) data)[i]));
			}
			break;
			
		case TYPE_SRATIONAL:
		case TYPE_RATIONAL:
			rafw.put((int) dataOffset);
			rafw.setOffset(dataOffset);
			dataOffset += count * 8;

			for (int i = 0; i < count; i++) {
				float f = ((float[]) data)[i];
				// System.out.println("Rational "+f);
				long l1 = (long) (f * 1000);
				long l2 = 1000;

				// System.out.println("l1:"+l1+", l2:"+l2);
				rafw.put((int) l1);
				rafw.put((int) l2);
			}
			break;
					
		default:
			throw new RuntimeException("Trying to write an unknown TIFF tag type");
		}
		return dataOffset;
	}
	
	/**
	 * Returns ASCII/byte data as a String.
	 * @return A string representation of the data.
	 */
	public String getString() {
		if (data instanceof byte[] && ((byte[]) data).length > 1) {
			byte b[] = (byte[]) data;
			byte b2[] = new byte[b.length - 1];

			for (int i = 0; i < b.length - 1; i++) {
				if (b[i] == 0) {
					b2[i] = ' ';
				} else {
					b2[i] = b[i];
				}
			}
			return new String(b2);
		}
		return "";
	}
	
	/**
	 * Returns the data as an array of long integers.
	 * @return An array of long integers.
	 */
	public long[] getLong() {
		long l[] = new long[count];

		if (data instanceof byte[]) {
			byte b[] = (byte[]) data;

			for (int i = 0; i < count; i++) {
				l[i] = b[i];
			}
		}
		if (data instanceof char[]) {
			char b[] = (char[]) data;

			for (int i = 0; i < count; i++) {
				l[i] = b[i];
			}
		}
		if (data instanceof int[]) {
			int b[] = (int[]) data;

			for (int i = 0; i < count; i++) {
				l[i] = b[i];
			}
		}
		if (data instanceof long[]) {
			long b[] = (long[]) data;

			for (int i = 0; i < count; i++) {
				l[i] = b[i];
			}
		}
		return l;
	}
	
	/**
	 * Returns the data as an array of floats.
	 * @return An array of floats.
	 */
	public float[] getFloat() {
		float f[] = new float[count];

		if (data instanceof float[]) {
			float b[] = (float[]) data;

			for (int i = 0; i < count; i++) {
				f[i] = b[i];
			}
		}
		if (data instanceof double[]) {
			double b[] = (double[]) data;

			for (int i = 0; i < count; i++) {
				f[i] = (float) b[i];
			}
		}
		return f;
	}
	
	/**
	 * Returns the data as an array of doubles.
	 * @return An array of double precision floating point numbers.
	 */
	public double[] getDouble() {
		double f[] = new double[count];

		if (data instanceof float[]) {
			float b[] = (float[]) data;

			for (int i = 0; i < count; i++) {
				f[i] = b[i];
			}
		}
		if (data instanceof double[]) {
			double b[] = (double[]) data;

			for (int i = 0; i < count; i++) {
				f[i] = b[i];
			}
		}
		return f;
	}
	
	/**
	 * Compares the ids of the tags.
	 * @return 0, 1, -1 ...
	 */
	public int compareTo(Object o) {
		GenericTag gt = (GenericTag) o;

		if (gt.id > id) {
			return -1;
		}
		if (gt.id < id) {
			return 1;
		}
		return 0;
	}
	
	/**
	 * Prints debug output to System.out
	 */
	public void printInfo() {
		System.out.println("Tag " + getTagName(this.id));
		System.out.println(" Data type " + getTypeName(type));
		System.out.println(" Data count " + count);
		if (count < 1) {
			return;
		}
		byte b[], b2[];
		char c[];
		int ints[];
		long l[];
		float f[];
		double d[];

		switch (type) {
		case GenericTag.TYPE_ASCII:
			b = (byte[]) data;
			if (b.length > 1 && b[b.length - 1] == 0) {
				b2 = new byte[b.length - 1];
				System.arraycopy(b, 0, b2, 0, b.length - 1);
			} else {
				b2 = new byte[b.length];
				System.arraycopy(b, 0, b2, 0, b.length);
			}
			for (int i = 0; i < b2.length; i++) {
				if (b2[i] == 0) {
					b2[i] = ' ';
				}
			}
			String s = new String(b2);

			System.out.println(" [ " + s + " ]");
			break;

		case GenericTag.TYPE_SBYTE:
			b = (byte[]) data;
			System.out.print(" [ ");
			System.out.print(b[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(b[i]);
			}
			System.out.print(" ]\n");
			break;

		case GenericTag.TYPE_SLONG:
		case GenericTag.TYPE_SHORT:
			ints = (int[]) data;
			System.out.print(" [ ");
			System.out.print(ints[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(ints[i]);
			}
			System.out.print(" ]\n");
			break;

		case GenericTag.TYPE_BYTE:
		case GenericTag.TYPE_SSHORT:
			c = (char[]) data;
			System.out.print(" [ ");
			System.out.print(c[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(c[i]);
			}
			System.out.print(" ]\n");
			break;

		case GenericTag.TYPE_LONG:
			l = (long[]) data;
			System.out.print(" [ ");
			System.out.print(l[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(l[i]);
			}
			System.out.print(" ]\n");
			break;

		case GenericTag.TYPE_FLOAT:
		case GenericTag.TYPE_RATIONAL:
		case GenericTag.TYPE_SRATIONAL:
			f = (float[]) data;
			System.out.print(" [ ");
			System.out.print(f[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(f[i]);
			}
			System.out.print(" ]\n");
			break;

		case GenericTag.TYPE_DOUBLE:
			d = (double[]) data;
			System.out.print(" [ ");
			System.out.print(d[0]);
			for (int i = 1; i < count; i++) {
				System.out.print(", ");
				System.out.print(d[i]);
			}
			System.out.print(" ]\n");
			break;

		default:
			break;
		}
	}
	
	/**
	 * Returns a string describing the data type.
	 * @param type The data type.
	 * @return A string.
	 */
	public static String getTypeName(int type) {
		switch (type) {
		case GenericTag.TYPE_ASCII:
			return "ASCII";

		case GenericTag.TYPE_BYTE:
			return "Unsigned byte";

		case GenericTag.TYPE_DOUBLE:
			return "Double";

		case GenericTag.TYPE_FLOAT:
			return "Float";

		case GenericTag.TYPE_LONG:
			return "Unsigned long";

		case GenericTag.TYPE_RATIONAL:
			return "Unsigned rational";

		case GenericTag.TYPE_SBYTE:
			return "Signed byte";

		case GenericTag.TYPE_SHORT:
			return "Unsigned short";

		case GenericTag.TYPE_SLONG:
			return "Signed long";

		case GenericTag.TYPE_SRATIONAL:
			return "Signed rational";

		case GenericTag.TYPE_SSHORT:
			return "Signed short";

		case GenericTag.TYPE_UNDEFINED:
			return "Undefined";

		default:
			return "Unknown (" + type + ")";
		}
	}
	
	/**
	 * Returns a string describing the tag type.
	 * @param tag The tag type.
	 * @return A description of the tag type.
	 */
	public static String getTagName(int tag) {
		switch (tag) {
		case GenericTag.TAG_PHOTOMETRIC_INTERPRETATION:
			return "PhotometricInterpretation";

		case GenericTag.TAG_COMPRESSION:
			return "Compression";

		case GenericTag.TAG_IMAGE_LENGTH:
			return "ImageLength";

		case GenericTag.TAG_IMAGE_WIDTH:
			return "ImageWidth";

		case GenericTag.TAG_RESOLUTION_UNIT:
			return "ResolutionUnit";

		case GenericTag.TAG_X_RESOLUTION:
			return "XResolution";

		case GenericTag.TAG_Y_RESOLUTION:
			return "YResolution";

		case GenericTag.TAG_ROWS_PER_STRIP:
			return "Rows per strip";

		case GenericTag.TAG_STRIP_OFFSETS:
			return "StripOffsets";

		case GenericTag.TAG_STRIP_BYTE_COUNTS:
			return "StripByteCounts";

		case GenericTag.TAG_BITS_PER_SAMPLE:
			return "BitsPerSample";

		case GenericTag.TAG_COLOR_MAP:
			return "ColorMap";

		case GenericTag.TAG_SAMPLES_PER_PIXEL:
			return "SamplesPerPixel";

		case GenericTag.TAG_ARTIST:
			return "Artist";

		case GenericTag.TAG_CELL_LENGTH:
			return "CellLength";

		case GenericTag.TAG_CELL_WIDTH:
			return "CellWidth";

		case GenericTag.TAG_COPYRIGHT:
			return "Copyright";

		case GenericTag.TAG_DATETIME:
			return "DateTime";

		case GenericTag.TAG_EXTRA_SAMPLES:
			return "ExtraSamples";

		case GenericTag.TAG_FILL_ORDER:
			return "FillOrder";

		case GenericTag.TAG_HOST_COMPUTER:
			return "HostComputer";

		case GenericTag.TAG_IMAGE_DESCRIPTION:
			return "ImageDescription";

		case GenericTag.TAG_MAKE:
			return "Make";

		case GenericTag.TAG_MAX_SAMPLE_VALUE:
			return "MaxSampleValue";

		case GenericTag.TAG_MIN_SAMPLE_VALUE:
			return "MinSampleValue";

		case GenericTag.TAG_MODEL:
			return "Model";

		case GenericTag.TAG_NEW_SUBFILE_TYPE:
			return "NewSubfileType";

		case GenericTag.TAG_ORIENTATION:
			return "Orientation";

		case GenericTag.TAG_PLANAR_CONFIGURATION:
			return "PlanarConfiguration";

		case GenericTag.TAG_SOFTWARE:
			return "Software";

		case GenericTag.TAG_SUBFILE_TYPE:
			return "SubfileType";

		case GenericTag.TAG_THRESHHOLDING:
			return "Threshholding";

		case GenericTag.TAG_T4_OPTIONS:
			return "T4Options";

		case GenericTag.TAG_T6_OPTIONS:
			return "T6Options";

		case GenericTag.TAG_DOCUMENT_NAME:
			return "DocumentName";

		case GenericTag.TAG_PAGE_NAME:
			return "PageName";

		case GenericTag.TAG_PAGE_NUMBER:
			return "PageNumber";

		case GenericTag.TAG_X_POSITION:
			return "XPosition";

		case GenericTag.TAG_Y_POSITION:
			return "YPosition";

		case GenericTag.TAG_PREDICTOR:
			return "Predictor";

		case GenericTag.TAG_TILE_WIDTH:
			return "TileWidth";

		case GenericTag.TAG_TILE_LENGTH:
			return "TileLength";

		case GenericTag.TAG_TILE_OFFSETS:
			return "TileOffsets";

		case GenericTag.TAG_TILE_BYTE_COUNTS:
			return "TileByteCounts";

		case GenericTag.TAG_INK_SET:
			return "InkSet";

		case GenericTag.TAG_NUMBER_OF_INKS:
			return "NumberOfInks";

		case GenericTag.TAG_INK_NAMES:
			return "InkNames";

		case GenericTag.TAG_DOT_RANGE:
			return "DotRange";

		case GenericTag.TAG_TARGET_PRINTER:
			return "TargetPrinter";

		case GenericTag.TAG_HALFTONE_HINTS:
			return "HalftoneHints";

		case GenericTag.TAG_SAMPLE_FORMAT:
			return "SampleFormat";

		case GenericTag.TAG_S_MIN_SAMPLE_VALUE:
			return "SMinSampleValue";

		case GenericTag.TAG_S_MAX_SAMPLE_VALUE:
			return "SMaxSampleValue";

		case GenericTag.TAG_WHITE_POINT:
			return "WhitePoint";

		case GenericTag.TAG_PRIMARY_CHROMATICITIES:
			return "PrimaryChromaticities";

		case GenericTag.TAG_TRANSFER_FUNCTION:
			return "TransferFunction";

		case GenericTag.TAG_TRANSFER_RANGE:
			return "TransferRange";

		case GenericTag.TAG_REFERENCE_BLACK_WHITE:
			return "ReferenceBlackWhite";

		case GenericTag.TAG_YCC_COEFFICIENTS:
			return "YCCCoefficients";

		case GenericTag.TAG_YCC_SUBSAMPLING:
			return "YCCSubSampling";

		case GenericTag.TAG_YCC_POSITIONING:
			return "YCC_Positioning";

		case GenericTag.TAG_JPEG_PROC:
			return "JPEGProc";

		case GenericTag.TAG_JPEG_INTERCHANGE_FORMAT:
			return "JPEGInterchangeFormat";

		case GenericTag.TAG_JPEG_INTERCHANGE_FORMAT_LENGTH:
			return "JPEGInterchangeFormatLength";

		case GenericTag.TAG_JPEG_RESTART_INTERVAL:
			return "JPEGRestartInterval";

		case GenericTag.TAG_JPEG_LOSSLESS_PREDICTORS:
			return "JPEGLosslessPredictors";

		case GenericTag.TAG_JPEG_POINT_TRANSFORMS:
			return "JPEGPointTransforms";

		case GenericTag.TAG_JPEG_Q_TABLES:
			return "JPEGQTables";

		case GenericTag.TAG_JPEG_DCT_TABLES:
			return "JPEGDCTTables";

		case GenericTag.TAG_JPEG_AC_TABLES:
			return "JPEGACTables";

		default:
			return "Unknown (" + tag + ")";
		}		
	}
	
	/**
	 * Ensure that data offsets start at a word boundary.
	 * @param l The original offset.
	 * @return The new offset, possibly increased to ensure that it is at a
	 * word boundary.
	 */
	public static final long ensureWordBoundary(long l) {
		if (l != 2 * ((long) (l / 2))) {
			l++;
		}
		return l;
	}
}
