/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader;


import java.io.ByteArrayOutputStream;
import java.awt.color.ICC_Profile;

import icc3dpackage.tiffreader.tags.*;
import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A class representing a TIFF IFD.
 */
public class IFD {
	
	/** The number of directory entries. */
	private int numDirEntries;
	
	/** The IFD entries. */
	private IFD_Entry entry[];
	
	/** The next IFD in the linked list. */
	private IFD next;
	
	/** The bits per sample tag. */
	private BitsPerSampleTag bitsPerSampleTag;
	
	/** The cell length tag. */
	private CellLengthTag cellLengthTag;
	
	/** The cell width tag. */
	private CellWidthTag cellWidthTag;
	
	/** The color map tag. */
	private ColorMapTag colorMapTag;
	
	/** The compression tag. */
	private CompressionTag compressionTag;
	
	/** The copyright tag. */
	private CopyrightTag copyrightTag;
	
	/** The date time tag. */
	private DateTimeTag dateTimeTag;
	
	/** The extra samples tag. */
	private ExtraSamplesTag extraSamplesTag;
	
	/** The image length tag. */
	private ImageLengthTag imageLengthTag;
	
	/** The image width tag. */
	private ImageWidthTag imageWidthTag;
	
	/** The photometric interpretation tag. */
	private PhotometricInterpretationTag photometricInterpretationTag;
	
	/** The resolution unit tag. */
	private ResolutionUnitTag resolutionUnitTag;
	
	/** The rows per strip tag. */
	private RowsPerStripTag rowsPerStripTag;
	
	/** The sample per pixel tag. */
	private SamplesPerPixelTag samplesPerPixelTag;
	
	/** The strip byte counts tag. */
	private StripByteCountsTag stripByteCountsTag;
	
	/** The strip offsets tag. */
	private StripOffsetsTag stripOffsetsTag;
	
	/** The threshholding tag. */
	private ThreshholdingTag threshholdingTag;
	
	/** The x resolution tag. */
	private XResolutionTag xResolutionTag;
	
	/** The y resolution tag. */
	private YResolutionTag yResolutionTag;
	
	/**
	 * Creates a TIFF IFD.
	 * @param rafw The TIFF file.
	 * @param offset The IFD offset.
	 */
	public IFD(RandomAccessFileWrapper rafw, long offset) {
		if (offset < 8) {
			throw new RuntimeException("IFD offset ");
		}
		rafw.setOffset(offset);
		numDirEntries = rafw.getInt();
		entry = new IFD_Entry[numDirEntries];
		GenericTag gt[] = new GenericTag[numDirEntries];

		for (int i = 0; i < numDirEntries; i++) {
			entry[i] = new IFD_Entry(rafw, offset + 2 + i * 12);
			gt[i] = new GenericTag(rafw, offset + 2 + i * 12);
			// gt[i].printInfo();
		}
		rafw.setOffset(offset + 2 + entry.length * 12);
		long nextOffset = rafw.getLong();

		if (nextOffset != 0) {
			next = new IFD(rafw, nextOffset);
		}
	}
	
	/**
	 * Returns the number of IFDs in a TIFF file.
	 * @return The number of IFDs in a TIFF file.
	 */
	public int getNumIfds() {
		if (next == null) {
			return 1;
		}
		return (next.getNumIfds() + 1);
	}
	
	/**
	 * Copy IFDs to array.
	 * @param ifd The IFD array.
	 * @param index The next array index to fill.
	 */
	public void populate(IFD ifd[], int index) {
		ifd[index] = this;
		if (next != null) {
			next.populate(ifd, index + 1);
		}
	}
	
	/**
	 * Returns true if the image contains alpha information.
	 * @return True if the image contains alpha information.
	 */
	public boolean hasAlpha() {
		if (photometricInterpretationTag.getType() == PhotometricInterpretationTag.RGB
				&& (int) samplesPerPixelTag.getSamplesPerPixel() == 4) {
			return true;
		}
		return false;
	}
	
	/**
	 * Displays information about the IFD.
	 */
	public void printChildren() {
		if (bitsPerSampleTag != null) {
			bitsPerSampleTag.printInfo();
		}
		if (cellLengthTag != null) {
			cellLengthTag.printInfo();
		}
		if (cellWidthTag != null) {
			cellWidthTag.printInfo();
		}
		if (cellWidthTag != null) {
			colorMapTag.printInfo();
		}
		if (compressionTag != null) {
			compressionTag.printInfo();
		}
		if (copyrightTag != null) {
			copyrightTag.printInfo();
		}
		if (dateTimeTag != null) {
			dateTimeTag.printInfo();
		}
		if (extraSamplesTag != null) {
			extraSamplesTag.printInfo();
		}
		if (imageLengthTag != null) {
			imageLengthTag.printInfo();
		}
		if (imageWidthTag != null) {
			imageWidthTag.printInfo();
		}
		if (photometricInterpretationTag != null) {
			photometricInterpretationTag.printInfo();
		}
		if (resolutionUnitTag != null) {
			resolutionUnitTag.printInfo();
		}
		if (rowsPerStripTag != null) {
			rowsPerStripTag.printInfo();
		}
		if (samplesPerPixelTag != null) {
			samplesPerPixelTag.printInfo();
		}
		if (stripByteCountsTag != null) {
			stripByteCountsTag.printInfo();
		}
		if (stripOffsetsTag != null) {
			stripOffsetsTag.printInfo();
		}
		if (threshholdingTag != null) {
			threshholdingTag.printInfo();
		}
		if (xResolutionTag != null) {
			xResolutionTag.printInfo();
		}
		if (yResolutionTag != null) {
			yResolutionTag.printInfo();
		}
	}
	
	/**
	 * Print information about the image.
	 */
	public void print() {
		System.out.println("Is a grayscale image: " + isGrayscaleImage());
		System.out.println("Is a bilevel image: " + isBilevelImage());
		System.out.println("Is a palette-color image: " + isPaletteColorImage());
		System.out.println("Number of entries: " + numDirEntries);
		for (int i = 0; i < entry.length; i++) {
			System.out.println("Entry " + i);
			entry[i].print();
		}
	}
	
	/**
	 * Returns any ICC profile contained in the IFD.
	 * @param rafw The TIFF file.
	 * @return The ICC profile.
	 */
	public ICC_Profile getICC_Profile(RandomAccessFileWrapper rafw) {
		for (int i = 0; i < entry.length; i++) {
			if (entry[i].getTag() == 34675) {
				return entry[i].getICC_Profile(rafw);
			}
		}
		return null;
	}
	
	/**
	 * Returns true if the image is a palettized color image.
	 * @return true if the image is a palettized color image, otherwise false.
	 */
	public boolean isPaletteColorImage() {
		boolean imgWidth = false;
		boolean imgLength = false;
		boolean bitsPerSample = false;
		boolean compression = false;
		boolean photmInterp = false;
		boolean stripOffsets = false;
		boolean rowsPerStrip = false;
		boolean stripByteCounts = false;
		boolean xRes = false;
		boolean yRes = false;
		boolean resUnit = false;
		boolean colorMap = false;
		
		for (int i = 0; i < entry.length; i++) {
			switch (entry[i].getTag()) {
			case ImageWidthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgWidth = true;
				}
				break;

			case ImageLengthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgLength = true;
				}
				break;

			case BitsPerSampleTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 4 || entry[i].getValueOrOffset() == 8)) {
					bitsPerSample = true;
				}
				break;

			case CompressionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1
								|| entry[i].getValueOrOffset() == 32773)) {
					compression = true;
				}
				break;

			case PhotometricInterpretationTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& entry[i].getValueOrOffset() == 3) {
					photmInterp = true;
				}
				break;

			case StripOffsetsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripOffsets = true;
				}
				break;

			case RowsPerStripTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					rowsPerStrip = true;
				}
				break;

			case StripByteCountsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripByteCounts = true;
				}
				break;

			case XResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					xRes = true;
				}
				break;

			case YResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					yRes = true;
				}
				break;

			case ResolutionUnitTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1 || entry[i].getValueOrOffset() == 2
								|| entry[i].getValueOrOffset() == 3)) {
					resUnit = true;
				}
				break;

			case ColorMapTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT) {
					colorMap = true;
				}
				break;

			default:
				break;
			}
		}
		
		if (imgWidth && imgLength && compression && photmInterp && stripOffsets
				&& rowsPerStrip && stripByteCounts && xRes && yRes && resUnit
				&& bitsPerSample && colorMap) {
			return true;
		}
		return false;
	}
	
	/**
	 * Returns true if the image is a grayscale image.
	 * @return true if the image is a grayscale image, otherwise false.
	 */
	public boolean isGrayscaleImage() {
		if (isPaletteColorImage()) {
			return false;
		}
		boolean imgWidth = false;
		boolean imgLength = false;
		boolean bitsPerSample = false;
		boolean compression = false;
		boolean photmInterp = false;
		boolean stripOffsets = false;
		boolean rowsPerStrip = false;
		boolean stripByteCounts = false;
		boolean xRes = false;
		boolean yRes = false;
		boolean resUnit = false;
		
		for (int i = 0; i < entry.length; i++) {
			switch (entry[i].getTag()) {
			case ImageWidthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgWidth = true;
				}
				break;

			case ImageLengthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgLength = true;
				}
				break;

			case BitsPerSampleTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 4 || entry[i].getValueOrOffset() == 8)) {
					bitsPerSample = true;
				}
				break;

			case CompressionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1
								|| entry[i].getValueOrOffset() == 32773)) {
					compression = true;
				}
				break;

			case PhotometricInterpretationTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 0 || entry[i].getValueOrOffset() == 1)) {
					photmInterp = true;
				}
				break;

			case StripOffsetsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripOffsets = true;
				}
				break;

			case RowsPerStripTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					rowsPerStrip = true;
				}
				break;

			case StripByteCountsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripByteCounts = true;
				}
				break;

			case XResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					xRes = true;
				}
				break;

			case YResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					yRes = true;
				}
				break;

			case ResolutionUnitTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1 || entry[i].getValueOrOffset() == 2
								|| entry[i].getValueOrOffset() == 3)) {
					resUnit = true;
				}
				break;

			default:
				break;
			}
		}
		
		if (imgWidth && imgLength && compression && photmInterp && stripOffsets
				&& rowsPerStrip && stripByteCounts && xRes && yRes && resUnit
				&& bitsPerSample) {
			return true;
		}
		return false;
	}
	
	/**
	 * Returns true if the image is a black and white image.
	 * @return true if the image is a black and white image, otherwise false.
	 */
	public boolean isBilevelImage() {
		if (isGrayscaleImage() || isPaletteColorImage()) {
			return false;
		}
		boolean imgWidth = false;
		boolean imgLength = false;
		boolean compression = false;
		boolean photmInterp = false;
		boolean stripOffsets = false;
		boolean rowsPerStrip = false;
		boolean stripByteCounts = false;
		boolean xRes = false;
		boolean yRes = false;
		boolean resUnit = false;
		
		for (int i = 0; i < entry.length; i++) {
			switch (entry[i].getTag()) {
			case ImageWidthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgWidth = true;
				}
				break;

			case ImageLengthTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					imgLength = true;
				}
				break;

			case CompressionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1 || entry[i].getValueOrOffset() == 2
								|| entry[i].getValueOrOffset() == 32773)) {
					compression = true;
				}
				break;

			case PhotometricInterpretationTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 0 || entry[i].getValueOrOffset() == 1)) {
					photmInterp = true;
				}
				break;

			case StripOffsetsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripOffsets = true;
				}
				break;

			case RowsPerStripTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					rowsPerStrip = true;
				}
				break;

			case StripByteCountsTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						|| entry[i].getType() == IFD_Entry.TYPE_LONG) {
					stripByteCounts = true;
				}
				break;

			case XResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					xRes = true;
				}
				break;

			case YResolutionTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_RATIONAL) {
					yRes = true;
				}
				break;

			case ResolutionUnitTag.TAG_ID:
				if (entry[i].getType() == IFD_Entry.TYPE_SHORT
						&& (entry[i].getValueOrOffset() == 1 || entry[i].getValueOrOffset() == 2
								|| entry[i].getValueOrOffset() == 3)) {
					resUnit = true;
				}
				break;

			default:
				break;
			}
		}
		
		if (imgWidth && imgLength && compression && photmInterp && stripOffsets
				&& rowsPerStrip && stripByteCounts && xRes && yRes && resUnit) {
			return true;
		}
		return false;
	}
	
	/**
	 * Creates the neceessary TIFF tags.
	 * @param rafw The TIFF file.
	 */
	public void makeTags(RandomAccessFileWrapper rafw) {
		// System.out.println("Making tags");
		bitsPerSampleTag = new BitsPerSampleTag();
		
		for (int i = 0; i < entry.length; i++) {
			switch (entry[i].getTag()) {
			case BitsPerSampleTag.TAG_ID:
				bitsPerSampleTag = new BitsPerSampleTag(entry[i], rafw);
				break;

			case ImageWidthTag.TAG_ID:
				imageWidthTag = new ImageWidthTag(entry[i], rafw);
				break;

			case ImageLengthTag.TAG_ID:
				imageLengthTag = new ImageLengthTag(entry[i], rafw);
				break;

			case CompressionTag.TAG_ID:
				compressionTag = new CompressionTag(entry[i], rafw);
				break;

			case PhotometricInterpretationTag.TAG_ID:
				photometricInterpretationTag = new PhotometricInterpretationTag(entry[i],
						rafw);
				break;

			case StripOffsetsTag.TAG_ID:
				stripOffsetsTag = new StripOffsetsTag(entry[i], rafw);
				break;

			case RowsPerStripTag.TAG_ID:
				rowsPerStripTag = new RowsPerStripTag(entry[i], rafw);
				break;

			case SamplesPerPixelTag.TAG_ID:
				samplesPerPixelTag = new SamplesPerPixelTag(entry[i], rafw);
				break;

			case StripByteCountsTag.TAG_ID:
				stripByteCountsTag = new StripByteCountsTag(entry[i], rafw);
				break;

			case ThreshholdingTag.TAG_ID:
				threshholdingTag = new ThreshholdingTag(entry[i], rafw);
				break;

			case XResolutionTag.TAG_ID:
				xResolutionTag = new XResolutionTag(entry[i], rafw);
				break;

			case YResolutionTag.TAG_ID:
				yResolutionTag = new YResolutionTag(entry[i], rafw);
				break;

			case ResolutionUnitTag.TAG_ID:
				resolutionUnitTag = new ResolutionUnitTag(entry[i], rafw);
				break;

			default:
				break;
			}
		}
	}
}