/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader;


import java.io.FileOutputStream;
import java.awt.color.ICC_Profile;

import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A class representing a TIFF IFD entry.
 */
public class IFD_Entry {
	
	/** TIFF tag data of type BYTE. */
	public final static int TYPE_BYTE = 1;
	
	/** TIFF tag data of type ASCII. */
	public final static int TYPE_ASCII = 2;
	
	/** TIFF tag data of type SHORT. */
	public final static int TYPE_SHORT = 3;
	
	/** TIFF tag data of type LONG. */
	public final static int TYPE_LONG = 4;
	
	/** TIFF tag data of type RATIONAL. */
	public final static int TYPE_RATIONAL = 5;
	
	/** TIFF tag data of type SIGNED BYTE. */
	public final static int TYPE_SIGNED_BYTE = 6;
	
	/** TIFF tag data of type UNDEFINED. */
	public final static int TYPE_UNDEFINED = 7;
	
	/** TIFF tag data of type SIGNED SHORT. */
	public final static int TYPE_SIGNED_SHORT = 8;
	
	/** TIFF tag data of type SIGNED LONG. */
	public final static int TYPE_SIGNED_LONG = 9;
	
	/** TIFF tag data of type SIGNED RATIONAL. */
	public final static int TYPE_SIGNED_RATIONAL = 10;
	
	/** TIFF tag data of type FLOAT. */
	public final static int TYPE_FLOAT = 11;
	
	/** TIFF tag data of type DOUBLE. */
	public final static int TYPE_DOUBLE = 12;	
	
	/** Tag id. */
	private int tag;
	
	/** Tag type. */
	private int type;
	
	/** Value/count. */
	private long valueCount;
	
	/** Value/offset. */
	private long valueOffset;
	
	/** First value. */
	private int value1;
	
	/** Second value. */
	private int value2;
	
	/**
	 * Creates an IFD_Entry.
	 * @param rafw The file.
	 * @param offset The entry offset.
	 */
	public IFD_Entry(RandomAccessFileWrapper rafw, long offset) {
		rafw.setOffset(offset);
		tag = rafw.getInt();
		type = rafw.getInt();
		valueCount = rafw.getLong();
		long off = rafw.getOffset();

		valueOffset = rafw.getLong();
		rafw.setOffset(off);
		value1 = rafw.getInt();
		value2 = rafw.getInt();
	}
	
	/**
	 * Print information about the entry.
	 */
	public void print() {
		System.out.println("Tag: " + tag);
		System.out.println("Type: " + type);
		System.out.println("Count: " + valueCount);
		System.out.println("Offset/Value: " + valueOffset);
	}
	
	/**
	 * Returns the tag id.
	 * @return The tag id.
	 */
	public int getTag() {
		return tag;
	}
	
	/**
	 * Returns the tag type.
	 * @return The tag type.
	 */
	public int getType() {
		return type;
	}
	
	/**
	 * Returns the value count.
	 * @return The value count.
	 */
	public long getCount() {
		return valueCount;
	}
	
	/**
	 * Returns the value of offset.
	 * @return The value of offset.
	 */
	public long getValueOrOffset() {
		return valueOffset;
	}
	
	/**
	 * Returns the first value.
	 * @return the first value.
	 */
	public int getValue1() {
		return value1;
	}
	
	/**
	 * Returns the second value.
	 * @return the second value.
	 */
	public int getValue2() {
		return value2;
	}
	
	/**
	 * Fetches the ICC profile represented by this IFD entry.
	 * @return The ICC profile represented by this IFD entry.
	 */
	public ICC_Profile getICC_Profile(RandomAccessFileWrapper rafw) {
		ICC_IFD_Entry iccEntry = new ICC_IFD_Entry(rafw, valueOffset, valueCount);

		return iccEntry.getICC_Profile();
	}
	
	/**
	 * Converts a tag data type to a description.
	 * @param type The data type.
	 * @return The description of the data type.
	 */
	public static String getType(int type) {
		switch (type) {
		case 1:
			return "Byte";

		case 2:
			return "ASCII";

		case 3:
			return "Short";

		case 4:
			return "Long";

		case 5:
			return "Rational";

		case 6:
			return "Signed byte";

		case 7:
			return "Undefined";

		case 8:
			return "Signed Short";

		case 9:
			return "Signed long";

		case 10:
			return "Signed Rational";

		case 11:
			return "Float";

		case 12:
			return "Double";

		default:
			return "Unknown";
		}
	}
	
	/**
	 * Fetches data of type long from this entry.
	 * @param rafw The file.
	 * @return An array of long integer data.
	 */
	public long[] getLongData(RandomAccessFileWrapper rafw) {
		if (getType() != IFD_Entry.TYPE_LONG) {
			throw new RuntimeException("Trying to access long data in TIFF tag that doesn't contain long data");
		}
		int count = (int) getCount();
		long result[] = new long[count];

		if (count == 1) {
			result[0] = getValueOrOffset();
		} else {
			rafw.setOffset(getValueOrOffset());
			for (int i = 0; i < count; i++) {
				result[i] = rafw.getLong();
			}
		}
		return result;
	}
	
	/**
	 * Fetches data of type short from this entry.
	 * @param rafw The file.
	 * @return An array of short integer data.
	 */
	public int[] getShortData(RandomAccessFileWrapper rafw) {
		if (getType() != IFD_Entry.TYPE_SHORT) {
			throw new RuntimeException("Trying to access short data in TIFF tag that doesn't contain short data");
		}
		int count = (int) getCount();
		int result[] = new int[count];

		if (count < 3) {
			result[0] = getValue1();
			if (count == 2) {
				result[1] = getValue2();
			}
		} else {
			rafw.setOffset(getValueOrOffset());
			for (int i = 0; i < count; i++) {
				result[i] = rafw.getInt();
			}
		}
		return result;
	}
}

/**
 * This entry represents an embedded ICC profile.
 */
class ICC_IFD_Entry {
	
	/** The file. */
	private RandomAccessFileWrapper rafw;
	
	/** The data offset. */
	private long offset;
	
	/** The data length. */
	private long length;
	
	/**
	 * Constructor.
	 * @param rafw The file.
	 * @param offset The data offset.
	 * @param length The data length.
	 */
	public ICC_IFD_Entry(RandomAccessFileWrapper rafw, long offset, long length) {
		this.rafw = rafw;
		this.offset = offset;
		this.length = length;
	}
	
	/**
	 * Retrieves a ICC profile.
	 * @return The ICC profile.
	 */
	public ICC_Profile getICC_Profile() {
		try {
			rafw.setOffset(offset);
			byte b[] = new byte[(int) length];

			rafw.getBytes(b);
			ICC_Profile profile = ICC_Profile.getInstance(b);

			return profile;
		} catch (Exception e) {
			throw new RuntimeException("Unable to extract ICC profile from TIFF file");
		} 
	}
	/*
	public void export(String filename) {
		try {
			rafw.setOffset(offset);
			FileOutputStream fos = new FileOutputStream(filename);
			byte b[] = new byte[(int) length];

			rafw.getBytes(b);
			fos.write(b);
			fos.flush();
			fos.close();
		} catch (Exception e) {
			throw new RuntimeException("Unable to export ICC profile from TIFF file");
		}
	}
	*/
}

