/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader;


import java.io.*;
import java.util.Vector;
import java.awt.color.ICC_Profile;

import icc3dpackage.iccreader.*;
import icc3dpackage.tiffreader.tags.*;
import icc3dpackage.tiffreader.tags.generic.TIFF_Tag;
import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * A class capable of reading TIFF files.
 */
public class TIFF_Reader {
	
	/**
	 * The TIFF file.
	 */
	private RandomAccessFileWrapper rafw;
	
	/**
	 * The first IFD offset in the file.
	 */
	private long firstOffset;
	
	/**
	 * The first IFD in the file.
	 */
	private IFD firstIFD;
	
	/**
	 * The IFDs in the file.
	 */
	private IFD ifd[];
	
	/**
	 * Reads the specified TIFF file.
	 * @param filename The TIFF file name.
	 */
	public TIFF_Reader(String filename) {
		RandomAccessFile randomAccessFile;

		try {
			randomAccessFile = new RandomAccessFile(filename, "r");
		} catch (FileNotFoundException fnfe) {
			throw new RuntimeException("File not found!");
		}
		rafw = new RandomAccessFileWrapper(randomAccessFile, false);
		byte b[] = new byte[2];

		rafw.getBytes(b);
		boolean endianForm = false;

		if (b[0] == Byte.parseByte("49", 16) && b[1] == Byte.parseByte("49", 16)) {
			endianForm = true;
		} else if (b[0] == Byte.parseByte("4D", 16)
				&& b[1] == Byte.parseByte("4D", 16)) {
			endianForm = false;
		} else {
			throw new RuntimeException(filename + " does not seem to be a TIFF file");
		}
		rafw.setEndian(endianForm);
		
		int magic = rafw.getInt();

		if (magic != 42) {
			throw new RuntimeException(filename + " does not seem to be a TIFF file");
		}
		firstOffset = rafw.getLong();
		ensureWordBoundary(firstOffset);
		firstIFD = new IFD(rafw, firstOffset);
		int numIfds = firstIFD.getNumIfds();

		ifd = new IFD[numIfds];
		firstIFD.populate(ifd, 0);
		makeTags();
	}
	
	/**
	 * Returns true if the image contains an alpha channel, otherwise false.
	 * @return True if the image contains an alpha channel, otherwise false.
	 */
	public boolean hasAlpha() {
		return ifd[0].hasAlpha();
	}
	
	/**
	 * Retrieves ICC profiles from the file.
	 * @return An array of ICC profiles.
	 */
	public ICC_Profile[] getICC_Profiles() {
		Vector profiles = new Vector();

		for (int i = 0; i < ifd.length; i++) {
			ICC_Profile profile = ifd[i].getICC_Profile(rafw);

			if (profile != null) {
				profiles.add(profile);
			}
		}
		if (profiles.size() < 1) {
			return new ICC_Profile[] {};
		}
		ICC_Profile ret[] = new ICC_Profile[profiles.size()];

		for (int i = 0; i < ret.length; i++) {
			ret[i] = (ICC_Profile) profiles.elementAt(i);
		}
		return ret;
	}
	
	/**
	 * Print information about the IFDs in the file.
	 */
	public void print() {
		for (int i = 0; i < ifd.length; i++) {
			System.out.println("IFD " + i);
			ifd[i].print();
		}
	}
	
	/**
	 * Prints information about the file contents.
	 */
	public void printChildren() {
		for (int i = 0; i < ifd.length; i++) {
			ifd[i].printChildren();
		}
	}
	
	/**
	 * Makes the TIFF tags.
	 */
	private void makeTags() {
		for (int i = 0; i < ifd.length; i++) {
			ifd[i].makeTags(rafw);
		}
	}
	
	/**
	 * Ensure that data starts at a word boundary.
	 */
	private void ensureWordBoundary(long l) {
		if (l != 2 * ((long) (l / 2))) {
			throw new RuntimeException("Offset not on a word boundary");
		}
	}
}
