/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader;


import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.RandomAccessFile;
import java.util.*;
import javax.swing.*;

import icc3dpackage.*;
import icc3dpackage.image2d.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ColorSpaceManager;
import icc3dpackage.utils.*;


/**
 * A class capable of writing TIFF files.
 */
public class TIFF_Writer {
	
	/** RGB image data. */
	public static final int MODE_RGB = 1;
	
	/** Lab image data. */
	public static final int MODE_LAB = 2;
	
	/** CMYK image data. */
	public static final int MODE_CMYK = 3;
	
	/** 8 bits per pixel component. */
	public static final int BPP_8 = 8;
	
	/** 16 bits per pixel component. */
	public static final int BPP_16 = 16;
	
	/**
	 * Saves a TIFF image.
	 * @param image The image.
	 */
	public static void saveImage(ImageCanvas2D image) {
		final JComboBox color = new JComboBox(new String[] {"Lab", "sRGB"});
		final JComboBox bpp = new JComboBox(new String[] {"24", "48"});
		final JButton file = new JButton("Click here to select a file");

		file.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				JFileChooser chooser = new JFileChooser(".");
				int choice = chooser.showSaveDialog(ICC3D.getDialogParent());

				if (choice == JFileChooser.APPROVE_OPTION
						&& chooser.getSelectedFile() != null) {
					file.setText(chooser.getSelectedFile().getAbsolutePath());
				}
			}
		});
		JPanel dialogPanel = new JPanel(new GridLayout(0, 2));

		dialogPanel.add(new JLabel("File"));
		dialogPanel.add(file);
		dialogPanel.add(new JLabel("Color space"));
		dialogPanel.add(color);
		dialogPanel.add(new JLabel("Bits per pixel"));
		dialogPanel.add(bpp);
		
		final JDialog dialog = new JDialog((JFrame) null, "Save TIFF", true);
		JButton okButton = new JButton("OK");
		JButton cancelButton = new JButton("Cancel");

		cancelButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				dialog.dispose();
			}
		});
		final ImageCanvas2D i2d = image;

		okButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				if (file.getText().equals("Click here to select a file")) {
					SystemManager.displayError(
						"Error",
						"Please specify a file name");
					return;
				}
				try {
					File f = new File(file.getText());

					f.delete();
					TIFF_Writer tw = new TIFF_Writer(i2d, file.getText(),
							bpp.getSelectedIndex(), (String) color.getSelectedItem());

					dialog.dispose();
				} catch (Exception e) {
					dialog.dispose();
					SystemManager.displayError(
						"Error",
						"Failed to write to file");
				}
			}
		});
		
		JPanel buttonPanel = new JPanel();
		JPanel buttonGrid = new JPanel(new GridLayout(1, 0));

		buttonGrid.add(okButton);
		buttonGrid.add(new JPanel() {
			public Dimension getPreferredSize() {
				return new Dimension(100, 0);
			}
		});
		buttonGrid.add(cancelButton);
		buttonPanel.add(buttonGrid);
		
		Container c = dialog.getContentPane();

		c.setLayout(new BorderLayout());
		c.add(dialogPanel, BorderLayout.CENTER);
		c.add(buttonPanel, BorderLayout.SOUTH);
		dialog.setResizable(false);
		dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		dialog.pack();
		dialog.setLocation(Toolkit.getDefaultToolkit().getScreenSize().width / 2
				- (dialog.getWidth() / 2),
				Toolkit.getDefaultToolkit().getScreenSize().height / 2
				- (dialog.getHeight() / 2));
		dialog.show();
		
	}
	
	/**
	 * Writes the image to a file.
	 * @param image The image.
	 * @param file The filename.
	 * @param mode The image mode.
	 * @param color The color space.
	 */
	public TIFF_Writer(ImageCanvas2D image, String file, int mode, String color) {
		int bitsPerComponent = 8;

		if (mode == 1) {
			bitsPerComponent = 16;
		}
		int interpretation = InterpretationInfo.RGB;

		if (color.equals("Lab")) {
			interpretation = InterpretationInfo.CIELab;
		}
		int height = image.getImageHeight();
		int width = image.getImageWidth();
		
		GenericTag tags[] = createTags(image, bitsPerComponent, interpretation, width,
				height);

		Arrays.sort(tags);
		long dataSize = 0;

		for (int i = 0; i < tags.length; i++) {
			dataSize += tags[i].bytesOfDataOutsideEntry;
			tags[i].printInfo();
			// System.out.println("----------------------------------------");
		}
		boolean isLittleEndian = false;
		RandomAccessFile raf = null;

		try {
			raf = new RandomAccessFile(file, "rw");
		} catch (Exception e) {
			throw new RuntimeException("Unable to open file for output");
		}
		RandomAccessFileWrapper rafw = new RandomAccessFileWrapper(raf, isLittleEndian);

		if (isLittleEndian) {
			rafw.put((char) 0x4949);
		} else {
			rafw.put((char) 0x4d4d);
		}
		rafw.put((char) 42);
		long off = rafw.getOffset() + 4;

		off = GenericTag.ensureWordBoundary(off);
		rafw.put((int) off);
		rafw.put((char) tags.length);
		long dataOffset = rafw.getOffset() + 12 * tags.length + 4;
		long firstEntryOffset = rafw.getOffset();
		long imageDataOffset = dataOffset + dataSize;
		// System.out.println("Datasize "+dataSize);
		GenericTag offsetTag = GenericTag.getTag(tags, GenericTag.TAG_STRIP_OFFSETS);

		((long[]) offsetTag.data)[0] = imageDataOffset;
		for (int i = 0; i < tags.length; i++) {
			dataOffset = tags[i].writeTag(rafw, firstEntryOffset + 12 * i, dataOffset);
		}
		// System.out.println("Offset after writing the tags "+rafw.getOffset());
		rafw.setOffset(firstEntryOffset + 12 * tags.length);
		rafw.put((int) 0);
		// System.out.println("Offset after writing the IFD "+rafw.getOffset());
		rafw.setOffset(imageDataOffset);
		
		float data[] = null;
		char cData[] = null;
		byte bData[] = null;
		int temp;
		int pixelToColorIndex[] = image.getPixelToColorIndex();

		if (interpretation == InterpretationInfo.RGB) {
			float col[] = ColorSpaceConverter.convert(image.getXYZ_Colors(),
					ColorSpaceManager.getColorSpace("XYZ"),
					ColorSpaceManager.getColorSpace("sRGB"));

			data = new float[width * height * 3];
			for (int i = 0; i < width * height; i++) {
				data[i * 3] = col[3 * pixelToColorIndex[i]];
				data[i * 3 + 1] = col[1 + 3 * pixelToColorIndex[i]];
				data[i * 3 + 2] = col[2 + 3 * pixelToColorIndex[i]];
			}
			if (bitsPerComponent == 8) {
				bData = new byte[data.length];
				for (int i = 0; i < data.length; i++) {
					temp = (int) (data[i] * 255.99);
					if (temp < 0) {
						temp = 0;
					} else if (temp > 255) {
						temp = 255;
					}
					bData[i] = (byte) temp;
				}
			} else {
				cData = new char[data.length];
				for (int i = 0; i < data.length; i++) {
					temp = (int) (data[i] * 65535.99);
					if (temp < 0) {
						temp = 0;
					} else if (temp > 65535) {
						temp = 65535;
					}
					cData[i] = (char) temp;
				}
			}
		} else {
			float col[] = ColorSpaceConverter.convert(image.getXYZ_Colors(),
					ColorSpaceManager.getColorSpace("XYZ"),
					ColorSpaceManager.getColorSpace("Lab"));

			data = new float[width * height * 3];
			for (int i = 0; i < width * height; i++) {
				data[i * 3] = col[3 * pixelToColorIndex[i]];
				data[i * 3 + 1] = col[1 + 3 * pixelToColorIndex[i]];
				data[i * 3 + 2] = col[2 + 3 * pixelToColorIndex[i]];
			}
			if (bitsPerComponent == 8) {
				bData = new byte[data.length];
				for (int i = 0; i < height * width; i++) {
					temp = (int) (data[i * 3 + 1] * 255.99);
					if (temp < 0) {
						temp = 0;
					} else if (temp > 255) {
						temp = 255;
					}
					bData[i * 3] = (byte) temp;
					temp = (int) ((data[i * 3] - 0.5f) * 256);
					if (temp < -128) {
						temp = -128;
					} else if (temp > 127) {
						temp = 127;
					}
					bData[i * 3 + 2] = (byte) temp;
					temp = (int) ((data[i * 3 + 2] - 0.5f) * 256);
					if (temp < -128) {
						temp = -128;
					} else if (temp > 127) {
						temp = 127;
					}
					bData[i * 3 + 1] = (byte) temp;
				}
			} else {
				cData = new char[data.length];
				for (int i = 0; i < height * width; i++) {
					temp = (int) (data[i * 3 + 1] * 65535.99f);
					if (temp < 0) {
						temp = 0;
					} else if (temp > 65535) {
						temp = 65535;
					}
					cData[i * 3] = (char) temp;
					temp = (int) ((data[i * 3] - 0.5f) * 65536);
					if (temp < -32768) {
						temp = -32768;
					} else if (temp > 32767) {
						temp = 32767;
					}
					cData[i * 3 + 2] = (char) temp;
					temp = (int) ((data[i * 3 + 2] - 0.5f) * 65536);
					if (temp < -32768) {
						temp = -32768;
					} else if (temp > 32767) {
						temp = 32767;
					}
					cData[i * 3 + 1] = (char) temp;
				}
			}
		}
		for (int i = 0; i < height; i++) {
			for (int j = 0; j < width; j++) {
				if (bitsPerComponent == 8) {
					rafw.put(bData[3 * (width * i + j)]);
					rafw.put(bData[3 * (width * i + j) + 1]);
					rafw.put(bData[3 * (width * i + j) + 2]);
				} else {
					rafw.put(cData[3 * (width * i + j)]);
					rafw.put(cData[3 * (width * i + j) + 1]);
					rafw.put(cData[3 * (width * i + j) + 2]);
				}
			}
		}
		try {
			raf.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Create the necessary TIFF tags.
	 * @param image The image.
	 * @param bitsPerComponent The bits per component,
	 * @param interpretation The data interpretation type.
	 * @param width The image width.
	 * @param height The image height.
	 * @return The TIFF tags.
	 */
	private GenericTag[] createTags(ImageCanvas2D image, int bitsPerComponent, int interpretation, int width, int height) {
		long bytesOfData = height * width * 3 * 2;
		long dataOffset = 500;
		
		Vector v = new Vector();
		GenericTag softwareTag = GenericTag.createTextTag("ICC3D",
				GenericTag.TAG_SOFTWARE);
		GenericTag bitsPerSampleTag = GenericTag.createShortTag(new int[] {bitsPerComponent,
			bitsPerComponent, bitsPerComponent},
				GenericTag.TAG_BITS_PER_SAMPLE);
		GenericTag compressionTag = GenericTag.createShortTag(new int[] {1},
				GenericTag.TAG_COMPRESSION);
		GenericTag imageLengthTag = GenericTag.createShortTag(new int[] {height},
				GenericTag.TAG_IMAGE_LENGTH);
		GenericTag imageWidthTag = GenericTag.createShortTag(new int[] {width},
				GenericTag.TAG_IMAGE_WIDTH);
		GenericTag photometricInterpretationTag = GenericTag.createShortTag(new int[] {interpretation},
				GenericTag.TAG_PHOTOMETRIC_INTERPRETATION);
		GenericTag rowsPerStripTag = GenericTag.createShortTag(new int[] {height},
				GenericTag.TAG_ROWS_PER_STRIP);
		GenericTag stripOffsetsTag = GenericTag.createLongTag(new long[] {dataOffset},
				GenericTag.TAG_STRIP_OFFSETS);
		GenericTag stripByteCountsTag = GenericTag.createLongTag(new long[] {bytesOfData},
				GenericTag.TAG_STRIP_BYTE_COUNTS);
		GenericTag samplesPerPixelTag = GenericTag.createShortTag(new int[] {3},
				GenericTag.TAG_SAMPLES_PER_PIXEL);
		GenericTag resolutionUnitTag = GenericTag.createShortTag(new int[] {2},
				GenericTag.TAG_RESOLUTION_UNIT);
		GenericTag xResolutionTag = GenericTag.createRationalTag(new float[] {72.0f},
				GenericTag.TAG_X_RESOLUTION);
		GenericTag yResolutionTag = GenericTag.createRationalTag(new float[] {72.0f},
				GenericTag.TAG_Y_RESOLUTION);

		v.add(softwareTag);
		v.add(bitsPerSampleTag);
		v.add(compressionTag);
		v.add(imageLengthTag);
		v.add(imageWidthTag);
		v.add(photometricInterpretationTag);
		v.add(rowsPerStripTag);
		v.add(stripOffsetsTag);
		v.add(stripByteCountsTag);
		v.add(samplesPerPixelTag);
		v.add(resolutionUnitTag);
		v.add(xResolutionTag);
		v.add(yResolutionTag);
		
		GenericTag tag[] = new GenericTag[v.size()];

		for (int i = 0; i < v.size(); i++) {
			tag[i] = (GenericTag) v.elementAt(i);
		}
		return tag;
		
		// --- black/white ---
		// PhotometricInterpretation
		// Compression
		// ImageLength
		// ImageWidth
		// ResolutionUnit
		// XResolution
		// YResolution
		// RowsPerStrip
		// StripOffsets
		// StripByteCounts;
		
		// --- grayscale ---
		// BitsPerSample
		
		// --- RGB ---
		// SamplesPerPixel	
	}
}


/**
 * Defines TIFF data interpretation types.
 */
class InterpretationInfo {
	
	/** Black and white image, white is 0 */
	public static final int WHITE_IS_ZERO = 0;
	
	/** Black and white image, black is 0 */
	public static final int BLACK_IS_ZERO = 1;
	
	/** RGB image */
	public static final int RGB = 2;
	
	/** RGB image with palette */
	public static final int RGB_PALETTE = 3;
	
	/** Image is a transparency mask */
	public static final int TRANSPARENCY_MASK = 4;
	
	/** CMYK image */
	public static final int CMYK = 5;
	
	/** YCbCr image */
	public static final int YCbCr = 6;
	
	/** CIELAB image */
	public static final int CIELab = 8;
	
	/**
	 * Returns a description of the interpretation type.
	 * @param type The interpretation type.
	 * @return A string.
	 */
	public static String getDescription(int type) {
		switch (type) {
		case InterpretationInfo.BLACK_IS_ZERO:
			return "Grayscale inverted";

		case InterpretationInfo.CIELab:
			return "CIELab";

		case InterpretationInfo.CMYK:
			return "CMYK";

		case InterpretationInfo.RGB:
			return "RGB";

		case InterpretationInfo.RGB_PALETTE:
			return "Indexed RGB";

		case InterpretationInfo.TRANSPARENCY_MASK:
			return "Transparency mask";

		case InterpretationInfo.WHITE_IS_ZERO:
			return "Grayscale";

		case InterpretationInfo.YCbCr:
			return "YCbCr";

		default:
			return "Unknown";
		}
	}
}