/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader.tags;


import icc3dpackage.tiffreader.tags.generic.TIFF_Tag;
import icc3dpackage.tiffreader.IFD_Entry;
import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * This TIFF tag contains information about the number of bits per sample.
 */
public class BitsPerSampleTag extends TIFF_Tag {
	
	/** The tag description. */
	public final static String TAG_DESCRIPTION = "Bits per sample";
	
	/** The tag id. */
	public final static int TAG_ID = 258;
	
	/** The number of samples. */
	private int length;
	
	/** The bits per samples. */
	private int bits[];
	
	/**
	 * Returns the tag id.
	 * @return The tag id.
	 */
	public static int getTagID() {
		return TAG_ID;
	}
	
	/**
	 * Returns a description of the tag.
	 * @return A String that describes the tag.
	 */
	public static String getTagDescription() {
		return TAG_DESCRIPTION;
	}
	
	/**
	 * Creates the tag.
	 * @param entry The parent directory.
	 * @param rafw The TIFF file.
	 */
	public BitsPerSampleTag(IFD_Entry entry, RandomAccessFileWrapper rafw) {
		if (entry.getType() != IFD_Entry.TYPE_SHORT) {
			throw new RuntimeException("TIFF bits per sample tag doesn't contain short unsigned integers");
		}
		length = (int) entry.getCount();
		bits = new int[length];
		rafw.setOffset(entry.getValueOrOffset());
		for (int i = 0; i < length; i++) {
			bits[i] = rafw.getInt();
		}
	}
	
	/**
	 * Default contructor.
	 */
	public BitsPerSampleTag() {
		length = 1;
		bits = new int[1];
		bits[0] = 1;
	}
	
	/**
	 * The bits per sample.
	 * @param sample The requested sample.
	 * @return The number of bits used to describe the sample.
	 */
	public int getBitPerSample(int sample) {
		if (sample >= bits.length) {
			return 1;
		}
		return bits[sample];
	}
	
	/**
	 * Prints information about the tag.
	 */
	public void printInfo() {
		System.out.println("Bits per sample");
		for (int i = 0; i < length; i++) {
			System.out.println("\t" + bits[i]);
		}
	}
}
