/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.tiffreader.tags;


import icc3dpackage.tiffreader.tags.generic.TIFF_Tag;
import icc3dpackage.tiffreader.IFD_Entry;
import icc3dpackage.utils.RandomAccessFileWrapper;


/**
 * This TIFF tag contains information about the reolution unit.
 */
public class ResolutionUnitTag extends TIFF_Tag {
	
	/** The tag description. */
	public final static String TAG_DESCRIPTION = "Resolution unit";
	
	/** The tag id. */
	public final static int TAG_ID = 296;
	
	/** No resolution unit specified. */
	public final static int NO_UNIT = 1;
	
	/** Inches. */
	public final static int INCH = 2;
	
	/** Cenitmetres. */
	public final static int CENTIMETER = 3;
	
	/** Default unit. */
	public final static int DEFAULT_UNIT = 2;
	
	/** The resolution unit. */
	private int resUnit;
	
	/**
	 * Returns the tag id.
	 * @return The tag id.
	 */
	public static int getTagID() {
		return TAG_ID;
	}
	
	/**
	 * Returns a description of the tag.
	 * @return A String that describes the tag.
	 */
	public static String getTagDescription() {
		return TAG_DESCRIPTION;
	}
	
	/**
	 * Creates the tag.
	 * @param entry The parent directory.
	 * @param rafw The TIFF file.
	 */
	public ResolutionUnitTag(IFD_Entry entry, RandomAccessFileWrapper rafw) {
		if (entry.getType() != IFD_Entry.TYPE_SHORT) {
			throw new RuntimeException("TIFF resolution unit tag doesn't contain short unsigned integers");
		}
		if (entry.getCount() != 1) {
			throw new RuntimeException("TIFF resolution unit tag doesn't contain the correct number of values");
		}
		resUnit = entry.getValue1();
	}
	
	/**
	 * Returns the type of resolution unit.
	 * @param r The type.
	 * @return A string that describes the unit type.
	 */
	public static String getResolutionUnit(int r) {
		if (r == NO_UNIT) {
			return "No unit specified";
		}
		if (r == INCH) {
			return "Inches";
		}
		if (r == CENTIMETER) {
			return "Centimetres";
		}
		return "Unknown";
	}
	
	/**
	 * Prints information about the tag.
	 */
	public void printInfo() {
		System.out.println("Resolution unit: " + getResolutionUnit(resUnit));
	}
}
