/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


import javax.vecmath.*;
import javax.swing.JOptionPane;

import icc3dpackage.triangulation.*;


/**
 * A class that represents an n dimensional Facet
 * in n dimensions. (n=3 or n=4)
 */
public class Facet {
	
	/** The facet is discarded if it no longer is a part of the convex hull. */
	boolean discarded;
	
	/** The sequence number when this facet was last visited. */
	int lastVisit;
	
	/** Integer used to connect a facet to data stored elsewhere. */
	public int data;
	
	/** The vertices of this facet. */
	public Vertex[] corner;
  
	/** The triangles of this Facet. */
	public AlphaTriangle triangle[];
  
	/** The neighboring facets. */
	public Facet neighbor[];
  
	/** The vertices that are associated with this facet. */
	OptimizedVector associatedVertices;
	
	/** The normal and distance that define this Facet. */
	public double distance, normalX, normalY, normalZ, normalW;
	
	/** The radius of smallest circumsphere (r^2). */
	public double radiusOfSmallestCircumsphere;
	
	/** Sequence number. */
	public static int seq;
	
	/**
	 * Temporary storage.
	 */
	static double d10, d11, d12, d13, d20, d21, d22, len;
	
	/**
	 * Constructs a Facet from the vertices.
	 * @param a Vertex a.
	 * @param b Vertex b.
	 * @param c Vertex c.
	 */
	public Facet(Vertex a, Vertex b, Vertex c) {
		// id = global_id++;
		corner = new Vertex[3];
		neighbor = new Facet[3];
		corner[0] = a;
		corner[1] = b;
		corner[2] = c;
		d10 = b.coordinates[0] - a.coordinates[0];
		d11 = b.coordinates[1] - a.coordinates[1];
		d12 = b.coordinates[2] - a.coordinates[2];
		d20 = c.coordinates[0] - a.coordinates[0];
		d21 = c.coordinates[1] - a.coordinates[1];
		d22 = c.coordinates[2] - a.coordinates[2];
		normalX = d11 * d22 - d12 * d21;
		normalY = d12 * d20 - d10 * d22;
		normalZ = d10 * d21 - d11 * d20;
		len = Math.sqrt(normalX * normalX + normalY * normalY + normalZ * normalZ);
		normalX /= len;
		normalY /= len;
		normalZ /= len;
		distance = normalX * a.coordinates[0] + normalY * a.coordinates[1]
				+ normalZ * a.coordinates[2];
		associatedVertices = new OptimizedVector(10);
	}
	
	/**
	 * Constructs a Facet from the vertices.
	 * @param a Vertex a.
	 * @param b Vertex b.
	 * @param c Vertex c.
	 * @param dummy seperates this constructor from others.
	 */
	public Facet(Vertex a, Vertex b, Vertex c, boolean dummy) {
		// id = global_id++;
		corner = new Vertex[3];
		corner[0] = a;
		corner[1] = b;
		corner[2] = c;
	}
	
	/**
	 * Constructs a Facet from the vertices.
	 * @param a Vertex a.
	 * @param b Vertex b.
	 * @param c Vertex c.
	 * @param d Vertex d.
	 */
	public Facet(Vertex a, Vertex b, Vertex c, Vertex d) {
		// id = global_id++;
		corner = new Vertex[4];
		triangle = new AlphaTriangle[4];
		neighbor = new Facet[4];
		corner[0] = a;
		corner[1] = b;
		corner[2] = c;
		corner[3] = d;
		associatedVertices = new OptimizedVector(10);
		
		double x1, x2, x3, x4;
		double y1, y2, y3, y4;
		double z1, z2, z3, z4;
		double w1, w2, w3, w4;

		x1 = a.coordinates[0];
		x2 = b.coordinates[0];
		x3 = c.coordinates[0];
		x4 = d.coordinates[0];
		y1 = a.coordinates[1];
		y2 = b.coordinates[1];
		y3 = c.coordinates[1];
		y4 = d.coordinates[1];
		z1 = a.coordinates[2];
		z2 = b.coordinates[2];
		z3 = c.coordinates[2];
		z4 = d.coordinates[2];
		w1 = a.coordinates[3];
		w2 = b.coordinates[3];
		w3 = c.coordinates[3];
		w4 = d.coordinates[3];
		
		double det = x1 * y2 * z3 * w4 - x1 * y2 * w3 * z4 - x1 * y3 * z2 * w4
				+ x1 * y3 * w2 * z4 + x1 * y4 * z2 * w3 - x1 * y4 * w2 * z3
				- x2 * y1 * z3 * w4 + x2 * y1 * w3 * z4 + x2 * y3 * z1 * w4
				- x2 * y3 * w1 * z4 - x2 * y4 * z1 * w3 + x2 * y4 * w1 * z3
				+ x3 * y1 * z2 * w4 - x3 * y1 * w2 * z4 - x3 * y2 * z1 * w4
				+ x3 * y2 * w1 * z4 + x3 * y4 * z1 * w2 - x3 * y4 * w1 * z2
				- x4 * y1 * z2 * w3 + x4 * y1 * w2 * z3 + x4 * y2 * z1 * w3
				- x4 * y2 * w1 * z3 - x4 * y3 * z1 * w2 + x4 * y3 * w1 * z2;

		double denoter = z4 * x3 * y2 * w1 - y3 * z4 * x2 * w1 + z4 * y1 * x2 * w3
				+ x4 * z1 * y2 * w3 + y4 * z3 * x2 * w1 + x3 * y4 * z1 * w2
				- z4 * x3 * y1 * w2 + y1 * x4 * z3 * w2 - x2 * y1 * z3 * w4
				+ z2 * x3 * y1 * w4 - z2 * x3 * y4 * w1 - z2 * y1 * x4 * w3
				- y2 * x3 * z1 * w4 - x2 * z1 * y4 * w3 - y3 * x4 * z1 * w2
				- y2 * z3 * x4 * w1 + y3 * x2 * z1 * w4 + y3 * z2 * x4 * w1
				+ x1 * y3 * z4 * w2 - x1 * y3 * z2 * w4 + x1 * z2 * y4 * w3
				- x1 * z4 * y2 * w3 - x1 * y4 * z3 * w2 + x1 * y2 * z3 * w4;

		d10 = -(-z4 * y1 * w3 + y1 * z3 * w4 + y3 * z2 * w4 - y3 * z2 * w1
				- y3 * z1 * w4 + y3 * z1 * w2 - y3 * z4 * w2 - y1 * z3 * w2 + z2 * y1 * w3
				- z2 * y4 * w3 + z4 * y2 * w3 - z1 * y2 * w3 - y4 * z3 * w1 + y4 * z3 * w2
				- y2 * z3 * w4 + y2 * z3 * w1 + z1 * y4 * w3 + y3 * z4 * w1 - z4 * y2 * w1
				- y4 * z1 * w2 + z4 * y1 * w2 - z2 * y1 * w4 + z2 * y4 * w1 + y2 * z1 * w4)
				/ (denoter);
		d11 = (-x1 * z4 * w3 - x1 * z3 * w2 + x1 * z3 * w4 + x1 * z2 * w3
				- x1 * z2 * w4 + x1 * z4 * w2 - z3 * x4 * w1 + z2 * x3 * w4 - z2 * x3 * w1
				- x3 * z1 * w4 + z3 * x2 * w1 + x3 * z1 * w2 + x4 * z3 * w2 - z4 * x3 * w2
				+ x4 * z1 * w3 - z2 * x4 * w3 + z4 * x2 * w3 - x2 * z3 * w4 + z4 * x3 * w1
				- x2 * z1 * w3 - z4 * x2 * w1 - x4 * z1 * w2 + x2 * z1 * w4 + z2 * x4 * w1)
				/ (denoter);
		d12 = -(x1 * y3 * w4 - x1 * y3 * w2 + x1 * y4 * w2 + x1 * y2 * w3
				- x1 * y2 * w4 - y4 * x1 * w3 + y3 * x4 * w2 - y3 * x2 * w4 - x4 * y3 * w1
				+ y3 * x2 * w1 - y4 * x2 * w1 - x3 * y2 * w1 - y1 * x4 * w2 + y1 * x4 * w3
				- y1 * x2 * w3 + y2 * x3 * w4 + y2 * x4 * w1 + x2 * y1 * w4 + x3 * y1 * w2
				+ x3 * y4 * w1 - x3 * y4 * w2 - x4 * y2 * w3 - x3 * y1 * w4 + x2 * y4 * w3)
				/ (denoter);
		d13 = (y1 * x4 * z3 - y4 * x1 * z3 - x4 * y3 * z1 + x1 * y3 * z4
				- x3 * y1 * z4 - x1 * y3 * z2 + x1 * y2 * z3 - x2 * z1 * y4 - y2 * x3 * z1
				- z2 * y1 * x4 + x4 * z1 * y2 + y4 * z3 * x2 + z4 * y2 * x3 - z2 * x3 * y4
				- y2 * z3 * x4 + z4 * x2 * y1 + y3 * x2 * z1 + x1 * z2 * y4 - x1 * z4 * y2
				- y3 * z4 * x2 + y3 * z2 * x4 - x2 * y1 * z3 + x3 * y4 * z1 + x3 * y1 * z2)
				/ (denoter);
		len = Math.sqrt(d10 * d10 + d11 * d11 + d12 * d12 + d13 * d13);
		normalX = d10 / len;
		normalY = d11 / len;
		normalZ = d12 / len;
		normalW = d13 / len;
		if (det < 0) {
			normalX = -normalX;
			normalY = -normalY;
			normalZ = -normalZ;
			normalW = -normalW;
		}
		distance = normalX * x1 + normalY * y1 + normalZ * z1 + normalW * w1;
	}
	
	/**
	 * Constructs a Facet from the vertices.
	 * @param a Vertex a.
	 * @param b Vertex b.
	 * @param c Vertex c.
	 * @param d Vertex d.
	 * @param isntOutside A vertex that is inside the convex hull. (or at least not outside)
	 */
	public Facet(Vertex a, Vertex b, Vertex c, Vertex d, Vertex isntOutside) {
		// id = global_id++;
		corner = new Vertex[4];
		triangle = new AlphaTriangle[4];
		neighbor = new Facet[4];
		corner[0] = a;
		corner[1] = b;
		corner[2] = c;
		corner[3] = d;
		associatedVertices = new OptimizedVector(10);
		
		double x1, x2, x3, x4;
		double y1, y2, y3, y4;
		double z1, z2, z3, z4;
		double w1, w2, w3, w4;

		x1 = a.coordinates[0];
		x2 = b.coordinates[0];
		x3 = c.coordinates[0];
		x4 = d.coordinates[0];
		y1 = a.coordinates[1];
		y2 = b.coordinates[1];
		y3 = c.coordinates[1];
		y4 = d.coordinates[1];
		z1 = a.coordinates[2];
		z2 = b.coordinates[2];
		z3 = c.coordinates[2];
		z4 = d.coordinates[2];
		w1 = a.coordinates[3];
		w2 = b.coordinates[3];
		w3 = c.coordinates[3];
		w4 = d.coordinates[3];
				
		double denoter = z4 * x3 * y2 * w1 - y3 * z4 * x2 * w1 + z4 * y1 * x2 * w3
				+ x4 * z1 * y2 * w3 + y4 * z3 * x2 * w1 + x3 * y4 * z1 * w2
				- z4 * x3 * y1 * w2 + y1 * x4 * z3 * w2 - x2 * y1 * z3 * w4
				+ z2 * x3 * y1 * w4 - z2 * x3 * y4 * w1 - z2 * y1 * x4 * w3
				- y2 * x3 * z1 * w4 - x2 * z1 * y4 * w3 - y3 * x4 * z1 * w2
				- y2 * z3 * x4 * w1 + y3 * x2 * z1 * w4 + y3 * z2 * x4 * w1
				+ x1 * y3 * z4 * w2 - x1 * y3 * z2 * w4 + x1 * z2 * y4 * w3
				- x1 * z4 * y2 * w3 - x1 * y4 * z3 * w2 + x1 * y2 * z3 * w4;

		d10 = -(-z4 * y1 * w3 + y1 * z3 * w4 + y3 * z2 * w4 - y3 * z2 * w1
				- y3 * z1 * w4 + y3 * z1 * w2 - y3 * z4 * w2 - y1 * z3 * w2 + z2 * y1 * w3
				- z2 * y4 * w3 + z4 * y2 * w3 - z1 * y2 * w3 - y4 * z3 * w1 + y4 * z3 * w2
				- y2 * z3 * w4 + y2 * z3 * w1 + z1 * y4 * w3 + y3 * z4 * w1 - z4 * y2 * w1
				- y4 * z1 * w2 + z4 * y1 * w2 - z2 * y1 * w4 + z2 * y4 * w1 + y2 * z1 * w4)
				/ (denoter);
		d11 = (-x1 * z4 * w3 - x1 * z3 * w2 + x1 * z3 * w4 + x1 * z2 * w3
				- x1 * z2 * w4 + x1 * z4 * w2 - z3 * x4 * w1 + z2 * x3 * w4 - z2 * x3 * w1
				- x3 * z1 * w4 + z3 * x2 * w1 + x3 * z1 * w2 + x4 * z3 * w2 - z4 * x3 * w2
				+ x4 * z1 * w3 - z2 * x4 * w3 + z4 * x2 * w3 - x2 * z3 * w4 + z4 * x3 * w1
				- x2 * z1 * w3 - z4 * x2 * w1 - x4 * z1 * w2 + x2 * z1 * w4 + z2 * x4 * w1)
				/ (denoter);
		d12 = -(x1 * y3 * w4 - x1 * y3 * w2 + x1 * y4 * w2 + x1 * y2 * w3
				- x1 * y2 * w4 - y4 * x1 * w3 + y3 * x4 * w2 - y3 * x2 * w4 - x4 * y3 * w1
				+ y3 * x2 * w1 - y4 * x2 * w1 - x3 * y2 * w1 - y1 * x4 * w2 + y1 * x4 * w3
				- y1 * x2 * w3 + y2 * x3 * w4 + y2 * x4 * w1 + x2 * y1 * w4 + x3 * y1 * w2
				+ x3 * y4 * w1 - x3 * y4 * w2 - x4 * y2 * w3 - x3 * y1 * w4 + x2 * y4 * w3)
				/ (denoter);
		d13 = (y1 * x4 * z3 - y4 * x1 * z3 - x4 * y3 * z1 + x1 * y3 * z4
				- x3 * y1 * z4 - x1 * y3 * z2 + x1 * y2 * z3 - x2 * z1 * y4 - y2 * x3 * z1
				- z2 * y1 * x4 + x4 * z1 * y2 + y4 * z3 * x2 + z4 * y2 * x3 - z2 * x3 * y4
				- y2 * z3 * x4 + z4 * x2 * y1 + y3 * x2 * z1 + x1 * z2 * y4 - x1 * z4 * y2
				- y3 * z4 * x2 + y3 * z2 * x4 - x2 * y1 * z3 + x3 * y4 * z1 + x3 * y1 * z2)
				/ (denoter);
		len = Math.sqrt(d10 * d10 + d11 * d11 + d12 * d12 + d13 * d13);
		normalX = d10 / len;
		normalY = d11 / len;
		normalZ = d12 / len;
		normalW = d13 / len;
		distance = normalX * x1 + normalY * y1 + normalZ * z1 + normalW * w1;
		
		if (normalX * isntOutside.coordinates[0]
						+ normalY * isntOutside.coordinates[1]
						+ normalZ * isntOutside.coordinates[2]
						+ normalW * isntOutside.coordinates[3]
				> distance) {
			normalX = -normalX;
			normalY = -normalY;
			normalZ = -normalZ;
			normalW = -normalW;
			distance = -distance;
			Vertex temp = corner[1];

			corner[1] = corner[0];
			corner[0] = temp;
		}
	}

	/**
	 * Try to add a vertex to this facet. Succeeds if vertex is outside Facet.
	 * @param p The vertex.
	 * @return True if success, otherwise false.
	 */
	public boolean add(Vertex p) {
		if (normalX * p.coordinates[0] + normalY * p.coordinates[1]
						+ normalZ * p.coordinates[2]
				> distance) {
			associatedVertices.addElement(p);
			return true;
		}
		return false;
	}
	
	/**
	 * Try to add a vertex to this facet. Succeeds if vertex is outside Facet.
	 * @param p The vertex.
	 * @return True if success, otherwise false.
	 */
	public boolean add4D(Vertex p) {
		if (normalX * p.coordinates[0] + normalY * p.coordinates[1]
				+ normalZ * p.coordinates[2] + normalW * p.coordinates[3]
				> distance) {
			associatedVertices.addElement(p);
			return true;
		}
		return false;
	}

	/**
	 * Find the vertex furthest away from the facet.
	 * @return The vertex furthest away away from the facet.
	 */
	public Vertex extreme() {
		Vertex res = null, p;
		double maxd = Double.NEGATIVE_INFINITY;
		double d;

		for (int i = 0; i < associatedVertices.elementCount; i++) {
			p = (Vertex) associatedVertices.elementData[i];
			d = normalX * p.coordinates[0] + normalY * p.coordinates[1]
					+ normalZ * p.coordinates[2];
			if (d > maxd) {
				res = (Vertex) associatedVertices.elementData[i];
				maxd = d;
			}
		}
		return res;
	}
	
	/**
	 * Find the vertex furthest away from the facet.
	 * @return The vertex furthest away away from the facet.
	 */
	public Vertex extreme4D() {
		Vertex res = null, p;
		double maxd = Double.NEGATIVE_INFINITY;
		double d;

		for (int i = 0; i < associatedVertices.elementCount; i++) {
			p = (Vertex) associatedVertices.elementData[i];
			d = normalX * p.coordinates[0] + normalY * p.coordinates[1]
					+ normalZ * p.coordinates[2] + normalW * p.coordinates[3];
			if (d > maxd) {
				res = (Vertex) associatedVertices.elementData[i];
				maxd = d;
			}
		}
		return res;
	}

	/**
	 * Decide whether the vertex is outside the plane of the facet.
	 * @param x The vertex.
	 * @return True if the vertex is outside the plane, otherwise false.
	 */
	public boolean outside(Vertex x) {
		return (normalX * x.coordinates[0] + normalY * x.coordinates[1]
						+ normalZ * x.coordinates[2]
				> distance);
	}
	
	/**
	 * Decide whether the vertex is outside the plane of the facet.
	 * @param x The vertex.
	 * @return True if the vertex is outside the plane, otherwise false.
	 */
	public boolean outside4D(Vertex x) {
		return (normalX * x.coordinates[0] + normalY * x.coordinates[1]
				+ normalZ * x.coordinates[2] + normalW * x.coordinates[3]
				> distance);
	}
	
	/**
	 * Returns whether the facet is discarded.
	 * @return True if the vertex is discarded, otherwise false.
	 */
	public boolean isDiscarded() {
		return discarded;
	}
	
	/**
	 * Sets the discarded property of this Facet to true;
	 */
	public void discard() {
		discarded = true;
	}
	
	/**
	 * Returns the vertices of this facet.
	 * @return The vertices.
	 */
	public Vertex[] getVertices() {
		return corner;
	}
	
	/**
	 * Calculate the radius of the smallest circumsphere.
	 */
	public void calculateRadiusOfSmallestCircumsphere() {
		double pi1, pi2, pi3, pi4;
		double pj1, pj2, pj3, pj4;
		double pk1, pk2, pk3, pk4;
		double pu1, pu2, pu3, pu4;

		pi1 = corner[0].coordinates[0];
		pi2 = corner[0].coordinates[1];
		pi3 = corner[0].coordinates[2];
		pi4 = corner[0].coordinates[3];
		pj1 = corner[1].coordinates[0];
		pj2 = corner[1].coordinates[1];
		pj3 = corner[1].coordinates[2];
		pj4 = corner[1].coordinates[3];
		pk1 = corner[2].coordinates[0];
		pk2 = corner[2].coordinates[1];
		pk3 = corner[2].coordinates[2];
		pk4 = corner[2].coordinates[3];
		pu1 = corner[3].coordinates[0];
		pu2 = corner[3].coordinates[1];
		pu3 = corner[3].coordinates[2];
		pu4 = corner[3].coordinates[3];
		
		double a1 = pi2 * pj3 * pk4 - pi2 * pj3 * pu4 - pi2 * pk3 * pj4
				+ pi2 * pk3 * pu4 + pi2 * pu3 * pj4 - pi2 * pu3 * pk4 - pj2 * pi3 * pk4
				+ pj2 * pi3 * pu4 + pj2 * pk3 * pi4 - pj2 * pk3 * pu4 - pj2 * pu3 * pi4
				+ pj2 * pu3 * pk4 + pk2 * pi3 * pj4 - pk2 * pi3 * pu4 - pk2 * pj3 * pi4
				+ pk2 * pj3 * pu4 + pk2 * pu3 * pi4 - pk2 * pu3 * pj4 - pu2 * pi3 * pj4
				+ pu2 * pi3 * pk4 + pu2 * pj3 * pi4 - pu2 * pj3 * pk4 - pu2 * pk3 * pi4
				+ pu2 * pk3 * pj4;
		double a2 = pi1 * pj3 * pk4 - pi1 * pj3 * pu4 - pi1 * pk3 * pj4
				+ pi1 * pk3 * pu4 + pi1 * pu3 * pj4 - pi1 * pu3 * pk4 - pj1 * pi3 * pk4
				+ pj1 * pi3 * pu4 + pj1 * pk3 * pi4 - pj1 * pk3 * pu4 - pj1 * pu3 * pi4
				+ pj1 * pu3 * pk4 + pk1 * pi3 * pj4 - pk1 * pi3 * pu4 - pk1 * pj3 * pi4
				+ pk1 * pj3 * pu4 + pk1 * pu3 * pi4 - pk1 * pu3 * pj4 - pu1 * pi3 * pj4
				+ pu1 * pi3 * pk4 + pu1 * pj3 * pi4 - pu1 * pj3 * pk4 - pu1 * pk3 * pi4
				+ pu1 * pk3 * pj4;
		double a3 = pi1 * pj2 * pk4 - pi1 * pj2 * pu4 - pi1 * pk2 * pj4
				+ pi1 * pk2 * pu4 + pi1 * pu2 * pj4 - pi1 * pu2 * pk4 - pj1 * pi2 * pk4
				+ pj1 * pi2 * pu4 + pj1 * pk2 * pi4 - pj1 * pk2 * pu4 - pj1 * pu2 * pi4
				+ pj1 * pu2 * pk4 + pk1 * pi2 * pj4 - pk1 * pi2 * pu4 - pk1 * pj2 * pi4
				+ pk1 * pj2 * pu4 + pk1 * pu2 * pi4 - pk1 * pu2 * pj4 - pu1 * pi2 * pj4
				+ pu1 * pi2 * pk4 + pu1 * pj2 * pi4 - pu1 * pj2 * pk4 - pu1 * pk2 * pi4
				+ pu1 * pk2 * pj4;
		double a4 = pi1 * pj2 * pk3 - pi1 * pj2 * pu3 - pi1 * pk2 * pj3
				+ pi1 * pk2 * pu3 + pi1 * pu2 * pj3 - pi1 * pu2 * pk3 - pj1 * pi2 * pk3
				+ pj1 * pi2 * pu3 + pj1 * pk2 * pi3 - pj1 * pk2 * pu3 - pj1 * pu2 * pi3
				+ pj1 * pu2 * pk3 + pk1 * pi2 * pj3 - pk1 * pi2 * pu3 - pk1 * pj2 * pi3
				+ pk1 * pj2 * pu3 + pk1 * pu2 * pi3 - pk1 * pu2 * pj3 - pu1 * pi2 * pj3
				+ pu1 * pi2 * pk3 + pu1 * pj2 * pi3 - pu1 * pj2 * pk3 - pu1 * pk2 * pi3
				+ pu1 * pk2 * pj3;
		double a5 = pi1 * pj2 * pk3 * pu4 - pi1 * pj2 * pu3 * pk4
				- pi1 * pk2 * pj3 * pu4 + pi1 * pk2 * pu3 * pj4 + pi1 * pu2 * pj3 * pk4
				- pi1 * pu2 * pk3 * pj4 - pj1 * pi2 * pk3 * pu4 + pj1 * pi2 * pu3 * pk4
				+ pj1 * pk2 * pi3 * pu4 - pj1 * pk2 * pu3 * pi4 - pj1 * pu2 * pi3 * pk4
				+ pj1 * pu2 * pk3 * pi4 + pk1 * pi2 * pj3 * pu4 - pk1 * pi2 * pu3 * pj4
				- pk1 * pj2 * pi3 * pu4 + pk1 * pj2 * pu3 * pi4 + pk1 * pu2 * pi3 * pj4
				- pk1 * pu2 * pj3 * pi4 - pu1 * pi2 * pj3 * pk4 + pu1 * pi2 * pk3 * pj4
				+ pu1 * pj2 * pi3 * pk4 - pu1 * pj2 * pk3 * pi4 - pu1 * pk2 * pi3 * pj4
				+ pu1 * pk2 * pj3 * pi4;

		// not really radius, radius^2;
		radiusOfSmallestCircumsphere = (a1 * a1 + a2 * a2 + a3 * a3 + 4 * a4 * a5)
				/ (4 * a4 * a4);
	}
	
	/**
	 * Locate the 4D Facet (3D tetrahedron) that contains the 3D vertex.
	 * @return The Facet that contains the vertex, or null if the vertex is outside the Facets.
	 */
	public Facet getClosestTetrahedron(Vertex v, int sequence) {
		if (lastVisit == sequence) {
			return this;
		}
		lastVisit = sequence;
		
		float px = (float) v.coordinates[0];
		float py = (float) v.coordinates[1];
		float pz = (float) v.coordinates[2];
		
		float c0c0 = (float) corner[0].coordinates[0];
		float c0c1 = (float) corner[0].coordinates[1];
		float c0c2 = (float) corner[0].coordinates[2];
		
		float c1c0 = (float) corner[1].coordinates[0];
		float c1c1 = (float) corner[1].coordinates[1];
		float c1c2 = (float) corner[1].coordinates[2];
		
		float c2c0 = (float) corner[2].coordinates[0];
		float c2c1 = (float) corner[2].coordinates[1];
		float c2c2 = (float) corner[2].coordinates[2];
		
		float c3c0 = (float) corner[3].coordinates[0];
		float c3c1 = (float) corner[3].coordinates[1];
		float c3c2 = (float) corner[3].coordinates[2];
		
		Plane plane = new Plane();

		plane.set(c0c0, c0c1, c0c2, c1c0, c1c1, c1c2, c2c0, c2c1, c2c2);
		
		if (plane.classifyPoint2(c3c0, c3c1, c3c2) != plane.classifyPoint2(px, py, pz)) {
			if (neighbor[0] == null) {
				return this;
			}
			return neighbor[0].getClosestTetrahedron(v, sequence);
		}
		plane.set(c1c0, c1c1, c1c2, c2c0, c2c1, c2c2, c3c0, c3c1, c3c2);
		
		if (plane.classifyPoint(c0c0, c0c1, c0c2) != plane.classifyPoint2(px, py, pz)) {
			if (neighbor[1] == null) {
				return this;
			}
			return neighbor[1].getClosestTetrahedron(v, sequence);
		}
		plane.set(c2c0, c2c1, c2c2, c3c0, c3c1, c3c2, c0c0, c0c1, c0c2);
								
		if (plane.classifyPoint(c1c0, c1c1, c1c2) != plane.classifyPoint2(px, py, pz)) {
			if (neighbor[2] == null) {
				return this;
			}
			return neighbor[2].getClosestTetrahedron(v, sequence);
		}
		plane.set(c3c0, c3c1, c3c2, c0c0, c0c1, c0c2, c1c0, c1c1, c1c2);
								
		if (plane.classifyPoint(c2c0, c2c1, c2c2) != plane.classifyPoint2(px, py, pz)) {
			if (neighbor[3] == null) {
				return this;
			}
			return neighbor[3].getClosestTetrahedron(v, sequence);
		}
		
		return this;
	}
	
	public boolean contains(Vertex v) {
		boolean isInside = true;
		
		float px = (float) v.coordinates[0];
		float py = (float) v.coordinates[1];
		float pz = (float) v.coordinates[2];
		
		float c0c0 = (float) corner[0].coordinates[0];
		float c0c1 = (float) corner[0].coordinates[1];
		float c0c2 = (float) corner[0].coordinates[2];
		
		float c1c0 = (float) corner[1].coordinates[0];
		float c1c1 = (float) corner[1].coordinates[1];
		float c1c2 = (float) corner[1].coordinates[2];
		
		float c2c0 = (float) corner[2].coordinates[0];
		float c2c1 = (float) corner[2].coordinates[1];
		float c2c2 = (float) corner[2].coordinates[2];
		
		float c3c0 = (float) corner[3].coordinates[0];
		float c3c1 = (float) corner[3].coordinates[1];
		float c3c2 = (float) corner[3].coordinates[2];
		
		Plane plane = new Plane();

		plane.set(c0c0, c0c1, c0c2, c1c0, c1c1, c1c2, c2c0, c2c1, c2c2);
		
		if (plane.classifyPoint2(c3c0, c3c1, c3c2) != plane.classifyPoint2(px, py, pz)) {
			isInside = false;
		}
		plane.set(c1c0, c1c1, c1c2, c2c0, c2c1, c2c2, c3c0, c3c1, c3c2);
		
		if (plane.classifyPoint(c0c0, c0c1, c0c2) != plane.classifyPoint2(px, py, pz)) {
			isInside = false;
		}
		plane.set(c2c0, c2c1, c2c2, c3c0, c3c1, c3c2, c0c0, c0c1, c0c2);
								
		if (plane.classifyPoint(c1c0, c1c1, c1c2) != plane.classifyPoint2(px, py, pz)) {
			isInside = false;
		}
		plane.set(c3c0, c3c1, c3c2, c0c0, c0c1, c0c2, c1c0, c1c1, c1c2);
								
		if (plane.classifyPoint(c2c0, c2c1, c2c2) != plane.classifyPoint2(px, py, pz)) {
			isInside = false;
		}
		return isInside;
	}
	
	public boolean vertexInTriangle(Vertex v) {
		final float delta = 1.0e-5f;
		float u0, u1, u2;
		float v0, v1, v2;
		float a, b;
		float max;
		int i, j;

		max = (float) Math.abs(normalX);
		i = 1;
		j = 2;
		a = (float) Math.abs(normalY);
		if (a > max) {
			max = a;
			i = 0;
		}
		a = (float) Math.abs(normalZ);
		if (a > max) {
			max = a;
			i = 0;
			j = 1;
		}
		u0 = (float) (v.coordinates[i] - corner[0].coordinates[i]);
		v0 = (float) (v.coordinates[j] - corner[0].coordinates[j]);
		u1 = (float) (corner[1].coordinates[i] - corner[0].coordinates[i]);
		v1 = (float) (corner[1].coordinates[j] - corner[0].coordinates[j]);
		u2 = (float) (corner[2].coordinates[i] - corner[0].coordinates[i]);
		v2 = (float) (corner[2].coordinates[j] - corner[0].coordinates[j]);
	
		if (u1 > -delta && u1 < delta) {
			b = u0 / u2;
			if (0.0f <= b && b <= 1.0f) {
				a = (v0 - b * v2) / v1;
				if ((a >= 0.0f) && ((a + b) <= 1.0f)) {
					return true;
				}
			}
		} else {
			b = (v0 * u1 - u0 * v1) / (v2 * u1 - u2 * v1);
			if (0.0f <= b && b <= 1.0f) {
				a = (u0 - b * u2) / u1;
				if ((a >= 0.0f) && ((a + b) <= 1.0f)) {
					return true;
				}
			}
		}
		return false;
	}
}
