/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


import java.util.ArrayList;
import javax.vecmath.Vector3f;


/**
 * This class is used to represent 3D points on the gamut boundary in a certain
 * hue plane by their 2D chroma and lightness coordinates.
 */
public class GamutBoundaryPoint {
	
	/**
	 * A list of unused gamut boundary points.
	 */
	protected static final ArrayList gbps = new ArrayList();
	
	/**
	 * The lightness of the gamut boundary point.
	 */
	public float y;
	
	/**
	 * The chroma of the gamut boundary point.
	 */
	public float chroma;
	
	/**
	 * Caclulates the distance along the line a-b where the ray p+r hits.
	 * @param a0 First coordinate of line start point.
	 * @param a1 Second coordinate of line start point.
	 * @param b0 First coordinate of line end point.
	 * @param b1 Second coordinate of line end point.
	 * @param p0 First coordinate of ray start point.
	 * @param p1 Second coordinate of ray start point.
	 * @param r0 First coordinate of ray direction.
	 * @param r1 Second coordinate of ray direction.
	 * @return The distance t along the line a-b where the ray p+u[b-a] hits. If t<0 or t>1, no intersection was detected.
	 */
	public static final float distanseAlongLine(float a0, float a1, float b0, float b1, float p0, float p1, float r0, float r1) {

		/* float u = (-b0*a1+b1*a0-b1*p0+a1*p0+b0*p1-a0*p1)/(-r1*b0+r1*a0+r0*b1-r0*a1)*/
		float denom = -r1 * b0 + r1 * a0 + r0 * b1 - r0 * a1;

		if (denom == 0.0f) {
			return -1;
		}
		return (r1 * a0 - r1 * p0 - r0 * a1 + r0 * p1) / (denom);
		
	}
	
	/**
	 * Transforms the 2D point into a full 3D vector, given the hue plane and
	 * color space scale.
	 * @param hue The hue.
	 * @param scale The color space scale.
	 * @return The vector representing the position in 3D.
	 */
	public Vector3f getVector(float hue, float scale[]) {
		Vector3f v3f = new Vector3f();

		v3f.y = y / scale[1];
		v3f.x = 0.5f + ((float) Math.sin(hue) * chroma) / scale[0];
		v3f.z = 0.5f + ((float) Math.cos(hue) * chroma) / scale[2];
		return v3f;
	}
	
	/**
	 * Returns the cusp lightness, given points that define the intersection
	 * between the gamut and a hue plane.
	 * @param al A list of points on the surface that define the gamut in a hue
	 * plane.
	 * @return The lightness of the cusp.
	 */
	public static final float getCusp(ArrayList al) {
		float largestChroma = 0;
		float lightness = 50;

		for (int i = 0; i < al.size(); i++) {
			GamutBoundaryPoint gbp = (GamutBoundaryPoint) al.get(i);

			if (gbp.chroma > largestChroma) {
				largestChroma = gbp.chroma;
				lightness = gbp.y;
			}
		}
		return lightness;
	}
	
	/**
	 * Returns the cusp point, given points that define the intersection
	 * between the gamut and a hue plane.
	 * @param al A list of points on the surface that define the gamut in a hue
	 * plane.
	 * @return The cusp point.
	 */
	public static final GamutBoundaryPoint getCuspPoint(ArrayList al) {
		float largestChroma = 0;
		int index = 0;
		
		for (int i = 0; i < al.size(); i++) {
			GamutBoundaryPoint gbp = (GamutBoundaryPoint) al.get(i);

			if (gbp.chroma > largestChroma) {
				largestChroma = gbp.chroma;
				index = i;
			}
		}
		return (GamutBoundaryPoint) al.get(index);
	}
	
	/**
	 * Returns the point on the line a-b that is closest to p.
	 * @param a The line start point.
	 * @param b The line end point.
	 * @param p A point.
	 * @return The point on the line a-b that is closest to p.
	 */
	public static final GamutBoundaryPoint getClosest(GamutBoundaryPoint a, GamutBoundaryPoint b, GamutBoundaryPoint p) {
		float cx = p.chroma - a.chroma;
		float cy = p.y - a.y;
		float vx = b.chroma - a.chroma;
		float vy = b.y - a.y;
		float d = (float) Math.sqrt(vx * vx + vy * vy);

		vx /= d;
		vy /= d;
		float t = vx * cx + vy * cy;

		if (t <= 0) {
			return a;
		}
		if (t >= d) {
			return b;
		}
		GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

		gbp.chroma = a.chroma + t * vx;
		gbp.y = a.y + t * vy;
		return gbp;
	}
	
	/**
	 * Returns a new gamut boundary point.
	 * @return A new gamut boundary point.
	 */
	public static final GamutBoundaryPoint getGBP() {
		int last = gbps.size() - 1;

		if (last >= 0) {
			return (GamutBoundaryPoint) gbps.remove(last);
		}
		return new GamutBoundaryPoint();
	}
	
	/**
	 * Frees the gamut boundary point.
	 * @param gbp The gamut boundary point.
	 */
	public static final void release(GamutBoundaryPoint gbp) {
		gbps.add(gbp);
	}
	
	/**
	 * Frees the gamut boundary points.
	 * @param al A list of gamut boundary points.
	 */
	public static final void release(ArrayList al) {
		gbps.addAll(al);
	}
}
