/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


import javax.vecmath.*;


/**
 * A class used to represent a plane in 3D.
 */
public class Plane {
	
	/**
	 * A small floating point value used in internal calculations.
	 */
	private static final float EPSILON = 0.00001f;
	
	/**
	 * The point is on the inside of the plane.
	 */
	public static final int PLANE_INSIDE = -1;
	
	/**
	 * The point lies on the plane.
	 */
	public static final int IN_PLANE = 0;
	
	/**
	 * The point is on the outside of the plane.
	 */
	public static final int PLANE_OUTSIDE = 1;
	
	/**
	 * The plane equation, consisting of a normal vector and a distance..
	 */
	protected float normalX, normalY, normalZ, distance;
	
	/**
	 * Creates a plane with an unspecified plane equation.
	 */
	public Plane() {}
	
	/**
	 * Creates a plane from the 3 input points.
	 * @param a Point 1.
	 * @param b Point 2.
	 * @param c Point 3.
	 */
	public Plane(Vector3f a, Vector3f b, Vector3f c) {
		set(a, b, c);
	}
	
	/**
	 * Calculates the plane equation from the 3 input points.
	 * @param a Point 1.
	 * @param b Point 2.
	 * @param c Point 3.
	 */
	public void set(Tuple3f a, Tuple3f b, Tuple3f c) {
		float v1x = b.x - a.x;
		float v1y = b.y - a.y;
		float v1z = b.z - a.z;
		
		float v2x = c.x - a.x;
		float v2y = c.y - a.y;
		float v2z = c.z - a.z;

		normalX = v1y * v2z - v1z * v2y;
		normalY = v1z * v2x - v1x * v2z;
		normalZ = v1x * v2y - v1y * v2x;
		
		float len = (float) Math.sqrt(normalX * normalX + normalY * normalY
				+ normalZ * normalZ);

		if (len == 0) {
			normalX = 1.0f;
			normalY = 0.0f;
			normalZ = 0.0f;
		} else {
			normalX /= len;
			normalY /= len;
			normalZ /= len;
		}
		distance = normalX * a.x + normalY * a.y + normalZ * a.z;
	}
	
	/**
	 * Creates a copy of a plane.
	 */
	public Plane(Plane p) {
		normalX = p.normalX;
		normalY = p.normalY;
		normalZ = p.normalZ;
		distance = p.distance;
	}
	
	/**
	 * Creates a plane from the given plane equation.
	 */
	public Plane(float nX, float nY, float nZ, float d) {
		normalX = nX;
		normalY = nY;
		normalZ = nZ;
		distance = d;
	}
	
	/**
	 * Creates a plane from the 3 input points.
	 * @param p1x The x coordinate of the 1st point.
	 * @param p1y The y coordinate of the 1st point.
	 * @param p1z The z coordinate of the 1st point.
	 *
	 * @param p2x The x coordinate of the 2nd point.
	 * @param p2y The y coordinate of the 2nd point.
	 * @param p2z The z coordinate of the 2nd point.
	 *
	 * @param p3x The x coordinate of the 3rd point.
	 * @param p3y The y coordinate of the 3rd point.
	 * @param p3z The z coordinate of the 3rd point.
	 **/
	public Plane(float p1x, float p1y, float p1z, float p2x, float p2y, float p2z, float p3x, float p3y, float p3z) {
		set(p1x, p1y, p1z, p2x, p2y, p2z, p3x, p3y, p3z);
	}
	
	/**
	 * Calculates the plane equation from the 3 input points.
	 * @param p1x The x coordinate of the 1st point.
	 * @param p1y The y coordinate of the 1st point.
	 * @param p1z The z coordinate of the 1st point.
	 *
	 * @param p2x The x coordinate of the 2nd point.
	 * @param p2y The y coordinate of the 2nd point.
	 * @param p2z The z coordinate of the 2nd point.
	 *
	 * @param p3x The x coordinate of the 3rd point.
	 * @param p3y The y coordinate of the 3rd point.
	 * @param p3z The z coordinate of the 3rd point.
	 */
	public final void set(float p1x, float p1y, float p1z, float p2x, float p2y, float p2z, float p3x, float p3y, float p3z) {
		float v1x = p2x - p1x;
		float v1y = p2y - p1y;
		float v1z = p2z - p1z;
		
		float v2x = p3x - p1x;
		float v2y = p3y - p1y;
		float v2z = p3z - p1z;

		normalX = v1y * v2z - v1z * v2y;
		normalY = v1z * v2x - v1x * v2z;
		normalZ = v1x * v2y - v1y * v2x;
		
		float len = (float) Math.sqrt(normalX * normalX + normalY * normalY
				+ normalZ * normalZ);

		if (len == 0) {
			normalX = 1.0f;
			normalY = 0.0f;
			normalZ = 0.0f;
		} else {
			normalX /= len;
			normalY /= len;
			normalZ /= len;
		}
		distance = normalX * p1x + normalY * p1y + normalZ * p1z;
	}
	
	/**
	 * Classifies a point with respect to this plane.
	 * @param pointX The x coordinate of the point.
	 * @param pointY The y coordinate of the point.
	 * @param pointZ The z coordinate of the point.
	 * @return PLANE_INSIDE - if the point lies on the inside of the plane.
	 * PLANE_OUSIDE - if the point lies on the outside of the plane.
	 * IN_PLANE - if the point lies on the plane.
	 */
	public final int classifyPoint(float pointX, float pointY, float pointZ) {
		float f = normalX * pointX + normalY * pointY + normalZ * pointZ;

		if (f > distance + 0.0001f) {
			return Plane.PLANE_OUTSIDE;
		}
		if (f < distance - 0.0001f) {
			return Plane.PLANE_INSIDE;
		}
		return Plane.IN_PLANE;
	}
	
	/**
	 * Classifies a point with respect to this plane. No floating point error
	 * correction.
	 * @param pointX The x coordinate of the point.
	 * @param pointY The y coordinate of the point.
	 * @param pointZ The z coordinate of the point.
	 * @return PLANE_INSIDE - if the point lies on the inside of the plane.
	 * PLANE_OUSIDE - if the point lies on the outside of the plane.
	 * IN_PLANE - if the point lies on the plane.
	 */
	public final int classifyPoint2(float pointX, float pointY, float pointZ) {
		float f = normalX * pointX + normalY * pointY + normalZ * pointZ;

		if (f > distance) {
			return Plane.PLANE_OUTSIDE;
		}
		if (f < distance) {
			return Plane.PLANE_INSIDE;
		}
		return Plane.IN_PLANE;
	}
	
	/**
	 * Returns a point projected down to the plane.
	 * @param px The x coordinate of the point.
	 * @param py The y coordinate of the point.
	 * @param pz The z coordinate of the point.
	 * @return The projected point.
	 */
	public Point3f projectPointToPlane(float px, float py, float pz) {
		int side = classifyPoint(px, py, pz);
		
		if (side != Plane.IN_PLANE) {
			float dist = normalX * px + normalY * py + normalZ * pz - distance;
			
			Point3f retPoint = new Point3f(px, py, pz);

			retPoint.x -= dist * normalX;
			retPoint.y -= dist * normalY;
			retPoint.z -= dist * normalZ;
			return retPoint;
		}
		
		return new Point3f(px, py, pz);
	}
	
	/**
	 * Calculates the intersection point between a line and the plane.
	 * The line is defined by its two end points.
	 * @param start The 1st line point.
	 * @param end The 2nd line point.
	 * @return The intersection point.
	 */
	public Vector3f intersectionPoint(Tuple3f start, Tuple3f end) {
		Vector3f ray = new Vector3f();

		ray.sub(end, start);
		float t = -(normalX * start.x + normalY * start.y + normalZ * start.z
						- distance)
				/ (normalX * ray.x + normalY * ray.y + normalZ * ray.z);

		if (t > 1) {
			t = 1;
		} else if (t < 0) {
			t = 0;
		}
		ray.scale(t);
		ray.add(start);
		return ray;
	}
	
	/**
	 * Calculates the intersection point between a line and the plane.
	 * The line is given by its two end points, and it is assumed that
	 * the points are on different sides of the plane.
	 * The intersection point is calculated by applying linear interpolation
	 * between the two end points.
	 *
	 * float w1 = intersectionTimeAsWeight(p1x, p1y, p1z, p2x, p2y, p2z);
	 * float w2 = 1- w1;
	 * float intersectionX = p1x*w1 + p2x*w2;
	 * float intersectionY = p1y*w1 + p2y*w2;
	 * float intersectionZ = p1z*w1 + p2z*w2;
	 *
	 * @param p1x The x coordinate of the 1st point.
	 * @param p1y The y coordinate of the 1st point.
	 * @param p1z The z coordinate of the 1st point.
	 * @param p2x The x coordinate of the 2nd point.
	 * @param p2y The y coordinate of the 2nd point.
	 * @param p2z The z coordinate of the 2nd point.
	 * @return The weight of the 1st point.
	 */
	public final float intersectionTimeAsWeight(float p1x, float p1y, float p1z, float p2x, float p2y, float p2z) {
		float dist1 = p1x * normalX + p1y * normalY + p1z * normalZ - distance;
		float dist2 = p2x * normalX + p2y * normalY + p2z * normalZ - distance;

		if (dist1 < 0) {
			dist1 = -dist1;
		}
		if (dist2 < 0) {
			dist2 = -dist2;
		}
		if (dist1 == 0) {
			if (dist2 == 0) {
				return 0.5f;
			}
			return 0;
		}
		return (dist2 / (dist1 + dist2));
	}
	
	/**
	 * Caclulates the distance from a point to the plane.
	 * @param x The x coordinate of the point.
	 * @param y The y coordinate of the point.
	 * @param z The z coordinate of the point.
	 * @return The distance from the point to the plane.
	 */
	public final float distanceTo(float x, float y, float z) {
		return x * normalX + y * normalY + z * normalZ - distance;
	}
	
	/**
	 * Increases or decreased the distance element in the plane equation.
	 */
	public void incDistance() {
		float epsilon = EPSILON;
		float oldDistance = distance;

		while (distance == oldDistance) {
			if (Math.random() < 0.5) {
				distance += epsilon;
			} else {
				distance -= epsilon;
			}
			epsilon *= 2;
		}
	}
}
