/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


/**
 * A class that is used to construct the 3D convex
 * hull of a given set of input points. Employs
 * the QuickHull algorithm.
 */
public class QuickHull {

	/**
	 * Default constructor for QuickHull
	 */
	public QuickHull() {}

	/**
	 * Locates the extreme points and moves them to the start of the array.
	 * @param vertices The vertex array.
	 */
	private void locateMaxAndMin(Vertex[] vertices) {
		Vertex temp;

		for (int i = 0; i < vertices.length; i++) {
			if (vertices[i].coordinates[0] > vertices[0].coordinates[0]) {
				temp = vertices[0];
				vertices[0] = vertices[i];
				vertices[i] = temp;
			}
			if (vertices[i].coordinates[0] < vertices[1].coordinates[0]) {
				temp = vertices[1];
				vertices[1] = vertices[i];
				vertices[i] = temp;
			}
		}
	}

	/**
	 * Recursively visit all visible facets. Discard vertices as necessary, and build a horizon consisting of
	 * all surrounding ridges.
	 * @param face The facet that should be visited.
	 * @param discardedVertices The vertices that are discarded.
	 * @param horizon The horizon ridges.
	 * @param vertex The vertex being processed.
	 * @param sequence The sequence number.
	 */
	private void visit(Facet face, OptimizedVector discardedVertices, OptimizedVector horizon, Vertex vertex, int sequence) {
		discardedVertices.append(face.associatedVertices);
		Facet neighbor;

		for (int i = 0; i < 3; i++) {
			neighbor = face.neighbor[i];
			if (!neighbor.discarded) {
				if (neighbor.lastVisit < sequence) {
					neighbor.lastVisit = sequence;
					if (neighbor.outside(vertex)) {
						neighbor.discarded = true;
						visit(neighbor, discardedVertices, horizon, vertex, sequence);
					} else {
						Edge edge = new Edge(face.corner[i], face.corner[(i + 1) % 3], neighbor);

						horizon.addElement(edge);
					}
				} else {
					Edge edge = new Edge(face.corner[i], face.corner[(i + 1) % 3], neighbor);

					horizon.addElement(edge);
				}
			}
		}
	}

	/**
	 * Builds the 3D convex hull.
	 * @param vertices The vertices.
	 * @return The facets.
	 */
	public OptimizedVector build3D(Vertex[] vertices) {
		Facet face1, face2;
		OptimizedVector faces = new OptimizedVector(1000);

		locateMaxAndMin(vertices);
		HyperPlane hyperPlane = new HyperPlane(vertices[0], vertices[1]);
		
		Vertex temp;
		double dist1 = hyperPlane.normal.dot(vertices[2]), dist2;

		for (int i = 3; i < vertices.length; i++) {
			if ((dist2 = hyperPlane.normal.dot(vertices[i])) > dist1) {
				temp = vertices[2];
				vertices[2] = vertices[i];
				vertices[i] = temp;
				dist1 = dist2;
			}
		}
		faces.addElement(face1 = new Facet(vertices[0], vertices[1], vertices[2]));
		faces.addElement(face2 = new Facet(vertices[0], vertices[2], vertices[1]));
		
		for (int i = 0; i < 3; i++) {
			face1.neighbor[i] = face2;
			face2.neighbor[i] = face1;
		}
	
		for (int i = 3; i < vertices.length; i++) {
			if (!face1.add(vertices[i])) {
				face2.add(vertices[i]);
			}
		}
    
		OptimizedVector discardedVertices = new OptimizedVector(100);
		
		int oldFaces;
		OptimizedVector horizon = new OptimizedVector(20);
		
		for (int i = 0; i < faces.elementCount; i++) {
			oldFaces = faces.elementCount;
			discardedVertices.elementCount = 0;
			Facet selected = (Facet) faces.elementAt(i);

			if (selected.discarded || selected.associatedVertices.elementCount == 0) {
				continue;
			}
			Vertex vertex = selected.extreme();

			horizon.elementCount = 0;
			selected.discarded = true;
			selected.lastVisit = ++Facet.seq;
			visit(selected, discardedVertices, horizon, vertex, Facet.seq);
			for (int iii = 0; iii < horizon.elementCount; iii++) {
				Edge e = (Edge) horizon.elementData[iii];
				Facet newFacet = new Facet(e.start, e.end, vertex);

				newFacet.neighbor[0] = e.outside;
				e.inside = newFacet;
				for (int co = 0; co < 3; co++) {
					if (e.outside.corner[co] == e.end) {
						e.outside.neighbor[co] = newFacet;
						break;
					}
				}
				faces.addElement(newFacet);
			}
			
			for (int iii = 0; iii < horizon.elementCount; iii++) {
				Edge e = (Edge) horizon.elementData[iii];
				Facet current = e.inside;
				Facet up;
				Edge upper;

				for (int jjj = 0;; jjj++) {
					upper = ((Edge) horizon.elementData[jjj]);
					if (upper.start == e.end) {
						up = upper.inside;
						break;
					}
				}
				current.neighbor[1] = up;
				up.neighbor[2] = current;
			}
			
			Vertex tmpVertex;
			
			for (int currentPoint = 0; currentPoint < discardedVertices.elementCount; currentPoint++) {
				((Vertex) discardedVertices.elementData[currentPoint]).isAssigned = false;
			}
			
			Facet tmpFace;
			
			for (int fc = oldFaces - 1; fc < faces.elementCount; fc++) {
				tmpFace = ((Facet) faces.elementData[fc]);
				if (tmpFace.discarded) {
					continue;
				}
				for (int currentPoint = 0; currentPoint < discardedVertices.elementCount; currentPoint++) {
					tmpVertex = ((Vertex) discardedVertices.elementData[currentPoint]);
					if (!tmpVertex.isAssigned && !(vertex == tmpVertex)) {
						tmpVertex.isAssigned = tmpFace.add(tmpVertex);
					}
				}	
			}
		}
		return faces;
	}
}
