/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


/**
 * A class that is used to construct the 4D convex
 * hull of a given set of input points. Employs
 * the QuickHull algorithm.
 */
public class QuickHull4D {
	
	/**
	 * Records the time used on this stage.
	 */
	public static long stage1, stage2, stage3, stage4, stage5, stage6;
	
	/**
	 * Default constructor for QuickHull4D
	 */
	public QuickHull4D() {}

	/**
	 * Locate the extreme points and move them to the front of the array.
	 * @param vertices The array of vertices.
	 */
	private void locateMaxAndMin(Vertex[] vertices) {
		Vertex temp;

		for (int i = 0; i < vertices.length; i++) {
			if (vertices[i].coordinates[0] > vertices[0].coordinates[0]) {
				temp = vertices[0];
				vertices[0] = vertices[i];
				vertices[i] = temp;
			}
			if (vertices[i].coordinates[0] < vertices[1].coordinates[0]) {
				temp = vertices[1];
				vertices[1] = vertices[i];
				vertices[i] = temp;
			}
			if (vertices[i].coordinates[1] > vertices[2].coordinates[1]) {
				temp = vertices[2];
				vertices[2] = vertices[i];
				vertices[i] = temp;
			}
			if (vertices[i].coordinates[1] < vertices[3].coordinates[1]) {
				temp = vertices[3];
				vertices[3] = vertices[i];
				vertices[i] = temp;
			}
		}
	}

	/**
	 * Recursively visit all visible facets. Discard vertices as necessary, and build a horizon consisting of
	 * all surrounding ridges.
	 * @param face The facet that should be visited.
	 * @param discardedVertices The vertices that are discarded.
	 * @param horizon The horizon ridges.
	 * @param vertex The vertex being processed.
	 * @param sequence The sequence number.
	 */
	private void visit(Facet face, OptimizedVector discardedVertices, OptimizedVector horizon, Vertex vertex, int sequence) {
		discardedVertices.append(face.associatedVertices);
		Facet neighbor;

		for (int i = 0; i < 4; i++) {
			neighbor = face.neighbor[i];
			if (!neighbor.discarded) {
				if (neighbor.lastVisit < sequence) {
					neighbor.lastVisit = sequence;
					if (neighbor.outside4D(vertex)) {
						neighbor.discarded = true;
						visit(neighbor, discardedVertices, horizon, vertex, sequence);
					} else {
						Edge edge = new Edge(face.corner[i], face.corner[(i + 1) % 4],
								face.corner[(i + 2) % 4], neighbor);

						horizon.addElement(edge);
					}
				} else {
					Edge edge = new Edge(face.corner[i], face.corner[(i + 1) % 4],
							face.corner[(i + 2) % 4], neighbor);

					horizon.addElement(edge);
				}
			}
		}
	}

	/**
	 * Builds the 4D convex hull.
	 * @param vertices The vertices.
	 * @return The facets.
	 */
	public OptimizedVector build4D(Vertex[] vertices) {
		Facet face1, face2;
		OptimizedVector faces = new OptimizedVector(1000);

		locateMaxAndMin(vertices);
		faces.addElement(face1 = new Facet(vertices[0], vertices[1], vertices[2],
				vertices[3]));
		faces.addElement(face2 = new Facet(vertices[0], vertices[2], vertices[1],
				vertices[3]));
		
		Vertex middle = new Vertex((vertices[0].coordinates[0]
						+ vertices[1].coordinates[0] + vertices[2].coordinates[0]
						+ vertices[3].coordinates[0])
				/ 4.0,
				(vertices[0].coordinates[1] + vertices[1].coordinates[1]
				+ vertices[2].coordinates[1] + vertices[3].coordinates[1])
				/ 4.0,
				(vertices[0].coordinates[2] + vertices[1].coordinates[2]
				+ vertices[2].coordinates[2] + vertices[3].coordinates[2])
				/ 4.0,
				(vertices[0].coordinates[3] + vertices[1].coordinates[3]
				+ vertices[2].coordinates[3] + vertices[3].coordinates[3])
				/ 4.0);
				
		for (int i = 0; i < 4; i++) {
			face1.neighbor[i] = face2;
			face2.neighbor[i] = face1;
		}
	
		for (int i = 4; i < vertices.length; i++) {
			if (!face1.add4D(vertices[i])) {
				face2.add4D(vertices[i]);
			}
		}
    
		OptimizedVector discardedVertices = new OptimizedVector(100);
		
		int oldFaces;
		OptimizedVector horizon = new OptimizedVector(20);
		
		for (int i = 0; i < faces.elementCount; i++) {
			oldFaces = faces.elementCount;
			discardedVertices.elementCount = 0;
			Facet selected = (Facet) faces.elementAt(i);

			if (selected.discarded || selected.associatedVertices.elementCount == 0) {
				continue;
			}
			Vertex vertex = selected.extreme4D();

			horizon.elementCount = 0;
			selected.discarded = true;
			selected.lastVisit = ++Facet.seq;
			visit(selected, discardedVertices, horizon, vertex, Facet.seq);
			for (int iii = 0; iii < horizon.elementCount; iii++) {
				Edge e = (Edge) horizon.elementData[iii];
				Facet newFacet = new Facet(e.start, e.middle, e.end, vertex, middle);

				newFacet.neighbor[0] = e.outside;
				e.inside = newFacet;
				for (int co = 0; co < 4; co++) {
					if ((e.outside.corner[co] == e.start
									&& ((e.outside.corner[(co + 1) % 4] == e.end
													&& e.outside.corner[(co + 2) % 4] == e.middle)
											|| (e.outside.corner[(co + 1) % 4] == e.middle
													&& e.outside.corner[(co + 2) % 4] == e.end)))
							|| (e.outside.corner[co] == e.middle
									&& ((e.outside.corner[(co + 1) % 4] == e.start
													&& e.outside.corner[(co + 2) % 4] == e.end)
											|| (e.outside.corner[(co + 1) % 4] == e.end
													&& e.outside.corner[(co + 2) % 4] == e.start)))
							|| (e.outside.corner[co] == e.end
									&& ((e.outside.corner[(co + 1) % 4] == e.middle
													&& e.outside.corner[(co + 2) % 4] == e.start)
											|| (e.outside.corner[(co + 1) % 4] == e.start
													&& e.outside.corner[(co + 2) % 4] == e.middle)))) {
						e.outside.neighbor[co] = newFacet;
						break;
					}
				}
				faces.addElement(newFacet);
			}
			
			for (int iii = oldFaces; iii < faces.elementCount; iii++) {
				Facet current = (Facet) faces.elementData[iii];
				Facet testFace;

				for (int f = 1; f < 4; f++) {
					for (int y = iii + 1; y < faces.elementCount; y++) {
						if (current.neighbor[f] != null) {
							break;
						}
						testFace = (Facet) faces.elementData[y];
						for (int u = 1; u < 4; u++) {
							if (
									((testFace.corner[u] == current.corner[f]
											&& ((testFace.corner[(u + 1) % 4] == current.corner[(f + 2) % 4]
															&& testFace.corner[(u + 2) % 4] == current.corner[(f + 1) % 4])
													|| (testFace.corner[(u + 1) % 4] == current.corner[(f + 1) % 4]
															&& testFace.corner[(u + 2) % 4] == current.corner[(f + 2) % 4])))
									|| (testFace.corner[u] == current.corner[(f + 1) % 4]
											&& ((testFace.corner[(u + 1) % 4] == current.corner[f]
															&& testFace.corner[(u + 2) % 4] == current.corner[(f + 2) % 4])
													|| (testFace.corner[(u + 1) % 4] == current.corner[(f + 2) % 4]
															&& testFace.corner[(u + 2) % 4] == current.corner[f])))
									|| (testFace.corner[u] == current.corner[(f + 2) % 4]
											&& ((testFace.corner[(u + 1) % 4] == current.corner[(f + 1) % 4]
															&& testFace.corner[(u + 2) % 4] == current.corner[f])
													|| (testFace.corner[(u + 1) % 4] == current.corner[f]
															&& testFace.corner[(u + 2) % 4] == current.corner[(f + 1) % 4])))
									)) {
								
								current.neighbor[f] = testFace;
								testFace.neighbor[u] = current;
								
								break;
							}
						}
					}
				}
			}
			Vertex tmpVertex;
			Facet tmpFace;
			
			for (int currentPoint = 0; currentPoint < discardedVertices.elementCount; currentPoint++) {
				tmpVertex = (Vertex) discardedVertices.elementData[currentPoint];
				if (tmpVertex == vertex) {
					continue;
				}
				tmpVertex.isAssigned = false;
				for (int fc = oldFaces - 1; fc < faces.elementCount; fc++) {
					tmpFace = ((Facet) faces.elementData[fc]);
					if (tmpFace.discarded) {
						continue;
					}
					if (tmpVertex.isAssigned = tmpFace.add4D(tmpVertex)) {
						break;
					}
				}	
			}
		}
		return faces;
	}
}
