/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


import java.util.*;
import javax.vecmath.*;

import icc3dpackage.pickable.*;


/**
 * A class used to represent a tetrahedron in 3D.
 */
public class Tetrahedron {
	
	/** The planes that define the interior of this tetrahedron. */
	public Plane plane[];
	
	/** The tetrahedra that surround this ttetrahedron. */
	private Tetrahedron neighbor[];
	
	/** The PickableObjects that define the corners of this tetrahedron. */
	private PickableObject cornerObj[];
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos0_0;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos0_1;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos0_2;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos1_0;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos1_1;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos1_2;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos2_0;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos2_1;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos2_2;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos3_0;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos3_1;
	
	/** A coordinate of one of the original tetrahedron corner vertices. */
	public float cornerpos3_2;
	
	/** Signals that this tetrahedron has been searched. */
	private boolean searched;
	
	/** The number of image indices already associated with this tetrahedron. */
	private int indicesUsed;
	
	/** The positions. */
	private float position[];
	
	/** The image indices associated with this tetrahedron. */
	private int colorIndices[];
	
	/** The total number of image indices. */
	private int totalIndices;
	
	/** The weights of the image indices associated with this tetrahedron. */
	private float weight[][];
	
	/** The volume of this tetrahedron. */
	private float myVolume;
	
	/** Temporary vector used to speed up calculations. */
	private static final Vector3f cVector = new Vector3f();
	
	/**
	 * Constructs a Tetrahedron from four PickableObjects.
	 * @param p0 PickableObject 1.
	 * @param p1 PickableObject 2.
	 * @param p2 PickableObject 3.
	 * @param p3 PickableObject 4.
	 */
	public Tetrahedron(PickableObject p0, PickableObject p1, PickableObject p2, PickableObject p3) {
		// id = last_id++;
		cornerObj = new PickableObject[4];
		plane = new Plane[4];
		neighbor = new Tetrahedron[4];
		cornerObj[0] = p0;
		float f[] = p0.getPosition();

		this.cornerpos0_0 = f[0];
		this.cornerpos0_1 = f[1];
		this.cornerpos0_2 = f[2];
		cornerObj[1] = p1;
		f = p1.getPosition();
		this.cornerpos1_0 = f[0];
		this.cornerpos1_1 = f[1];
		this.cornerpos1_2 = f[2];
		cornerObj[2] = p2;
		f = p2.getPosition();
		this.cornerpos2_0 = f[0];
		this.cornerpos2_1 = f[1];
		this.cornerpos2_2 = f[2];
		cornerObj[3] = p3;
		f = p3.getPosition();
		this.cornerpos3_0 = f[0];
		this.cornerpos3_1 = f[1];
		this.cornerpos3_2 = f[2];
		
		myVolume = volume(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos1_0,
				cornerpos1_1, cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2,
				cornerpos3_0, cornerpos3_1, cornerpos3_2);
		plane[0] = new Plane(new Vector3f(p0.getPosition()),
				new Vector3f(p1.getPosition()), new Vector3f(p2.getPosition()));
		plane[1] = new Plane(new Vector3f(p0.getPosition()),
				new Vector3f(p2.getPosition()), new Vector3f(p3.getPosition()));
		plane[2] = new Plane(new Vector3f(p0.getPosition()),
				new Vector3f(p3.getPosition()), new Vector3f(p1.getPosition()));
		plane[3] = new Plane(new Vector3f(p1.getPosition()),
				new Vector3f(p3.getPosition()), new Vector3f(p2.getPosition()));
	}
	
	/**
	 * Synchronizes the plane equations and corner positions with the current
	 * positions of the corner objects.
	 */
	public void updatePositionFromCorners() {
		
		float f[] = cornerObj[0].getPositionReference();

		this.cornerpos0_0 = f[0];
		this.cornerpos0_1 = f[1];
		this.cornerpos0_2 = f[2];

		f = cornerObj[1].getPositionReference();
		this.cornerpos1_0 = f[0];
		this.cornerpos1_1 = f[1];
		this.cornerpos1_2 = f[2];
		f = cornerObj[2].getPositionReference();
		this.cornerpos2_0 = f[0];
		this.cornerpos2_1 = f[1];
		this.cornerpos2_2 = f[2];
		f = cornerObj[3].getPositionReference();
		this.cornerpos3_0 = f[0];
		this.cornerpos3_1 = f[1];
		this.cornerpos3_2 = f[2];
		
		myVolume = volume(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos1_0,
				cornerpos1_1, cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2,
				cornerpos3_0, cornerpos3_1, cornerpos3_2);
		plane[0].set(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos1_0,
				cornerpos1_1, cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
		plane[1].set(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos2_0,
				cornerpos2_1, cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
		plane[2].set(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos3_0,
				cornerpos3_1, cornerpos3_2, cornerpos1_0, cornerpos1_1, cornerpos1_2);
		plane[3].set(cornerpos1_0, cornerpos1_1, cornerpos1_2, cornerpos3_0,
				cornerpos3_1, cornerpos3_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
	}
	
	/**
	 * Sets the tetrahedron neighbors.
	 * @param i The index of the neighbor.
	 * @param tetra The tetrahedron.
	 */
	public void setNeighbor(int i, Tetrahedron tetra) {
		neighbor[i] = tetra;
	}
	
	/**
	 * Returns the tetrahedron neighbor.
	 * @param i The index of the neighbor.
	 * @return The tetrahedron.
	 */
	public final Tetrahedron getNeighbor(int i) {
		return neighbor[i];
	}
	
	/**
	 * Returns the segment tetrahedron that contains the point. Allows points to be 
	 * outside the segment.
	 * @param pointx The point x coordinate.
	 * @param pointy The point y coordinate.
	 * @param pointz The point z coordinate.
	 * @return The tetrahedron that contains/is closest to the point.
	 */
	public Tetrahedron findSegmentTetrahedron(float pointx, float pointy, float pointz) {
		// System.out.println("Tetra :"+id);
		for (int i = 0; i < 4; i++) {
			if (plane[i].classifyPoint(pointx, pointy, pointz) == Plane.PLANE_OUTSIDE
					&& neighbor[i] != null) {
				return neighbor[i].findSegmentTetrahedron(pointx, pointy, pointz);
			}
		}
		// System.out.println("Inside "+id);
		return this;
	}
	
	/**
	 * Returns the segment tetrahedron that contains the point. Allows points to be 
	 * outside the segment.
	 * @param from The last tetrahedron.
	 * @param pointx The point x coordinate.
	 * @param pointy The point y coordinate.
	 * @param pointz The point z coordinate.
	 * @return The tetrahedron that contains/is closest to the point.
	 */
	public Tetrahedron findSegmentTetrahedron(Tetrahedron from, float pointx, float pointy, float pointz) {
		// System.out.println("Tetra :"+id);
		for (int i = 0; i < 3; i++) {
			if (neighbor[i] != from
					&& plane[i].classifyPoint(pointx, pointy, pointz) == Plane.PLANE_OUTSIDE) {
				return neighbor[i].findSegmentTetrahedron(this, pointx, pointy, pointz);
			}
		}
		// System.out.println("Inside "+id);
		return this;
	}
	
	/**
	 * Returns the segment tetrahedron that contains the point. Allows points to be 
	 * outside the segment. Absolute precision assumed.
	 * @param from The last tetrahedron.
	 * @param pointx The point x coordinate.
	 * @param pointy The point y coordinate.
	 * @param pointz The point z coordinate.
	 * @return The tetrahedron that contains/is closest to the point.
	 */
	public Tetrahedron findSegmentTetrahedron2(Tetrahedron from, float pointx, float pointy, float pointz) {
		// System.out.println("Tetra :"+id);
		for (int i = 0; i < 3; i++) {
			if (neighbor[i] != from
					&& plane[i].classifyPoint2(pointx, pointy, pointz) == Plane.PLANE_OUTSIDE) {
				return neighbor[i].findSegmentTetrahedron(this, pointx, pointy, pointz);
			}
		}
		// System.out.println("Inside "+id);
		return this;
	}
	
	/**
	 * Returns true if the point is inside the tetrahedron sector.
	 * @return True if the point is inside the tetrahedron sector.
	 */	
	public boolean isDefinitelyInSegment(float x, float y, float z) {
		for (int i = 0; i < 3; i++) {
			if (plane[i].classifyPoint2(x, y, z) == Plane.PLANE_OUTSIDE) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Checks whether the tetrahedron is legal.
	 * @return True if the tetrahedron is legal, otherwise false.
	 */
	public boolean test(boolean fix) {
		
		// for (int k=0; k<4; k++) {
		// int ttt = plane[k].classifyPoint(this.cornerPos[3-k]);
		// if (ttt==Plane.IN_PLANE)
		// System.out.println("In plane");
		// if (ttt==Plane.PLANE_INSIDE)
		// System.out.println("plane inside");
		// if (ttt==Plane.PLANE_OUTSIDE)
		// System.out.println("plane outside");
		// }
			
		
		if (plane[3].classifyPoint2(cornerpos0_0, cornerpos0_1, cornerpos0_2)
				!= Plane.PLANE_INSIDE) {
			System.out.println("Tetrahedron test failed "/* +id*/);
			if (fix) {
				if (neighbor[0].isDefinitelyInSegment(cornerpos3_0, cornerpos3_1,
						cornerpos3_2)) {
					float epsilon = 0.00001f;

					// Vector3f negNormal = new Vector3f();
					while (plane[0].classifyPoint2(cornerpos3_0, cornerpos3_1, cornerpos3_2)
							!= Plane.PLANE_INSIDE) {
						cornerpos3_0 -= plane[0].normalX * epsilon;
						cornerpos3_1 -= plane[0].normalY * epsilon;
						cornerpos3_2 -= plane[0].normalZ * epsilon;
						epsilon *= 2.0f;
					}
					float f[] = cornerObj[3].getPositionReference();

					f[0] = cornerpos3_0;
					f[1] = cornerpos3_1;
					f[2] = cornerpos3_2;

					boolean success = true;

					try {
						PickableTetrahedronObject pto = (PickableTetrahedronObject) cornerObj[3];
						Tetrahedron t[] = pto.getTetrahedra();

						if (t != null) {
							for (int i = 0; i < t.length; i++) {
								t[i].updatePositionFromCorners();
								success &= t[i].test(false);
							}
						}
					} catch (Exception e) {}
					System.out.println("Corner 3 in neighbor 0");
					success &= test(false);
					System.out.println("Tetrahedron fix success=" + success);
					return success;
				} else if (neighbor[1].isDefinitelyInSegment(cornerpos1_0, cornerpos1_1,
						cornerpos1_2)) {
					float epsilon = 0.00001f;

					// Vector3f negNormal = new Vector3f();
					while (plane[1].classifyPoint2(cornerpos1_0, cornerpos1_1, cornerpos1_2)
							!= Plane.PLANE_INSIDE) {
						// negNormal.set(-plane[1].normalX, -plane[1].normalY, -plane[1].normalZ);
						// negNormal.scale(epsilon);
						// cornerPos[1].add(negNormal);
						cornerpos1_0 -= plane[1].normalX * epsilon;
						cornerpos1_1 -= plane[1].normalY * epsilon;
						cornerpos1_2 -= plane[1].normalZ * epsilon;
						epsilon *= 2.0f;
					}
					float f[] = cornerObj[1].getPositionReference();

					f[0] = cornerpos1_0;
					f[1] = cornerpos1_1;
					f[2] = cornerpos1_2;
					
					boolean success = true;

					try {
						PickableTetrahedronObject pto = (PickableTetrahedronObject) cornerObj[1];
						Tetrahedron t[] = pto.getTetrahedra();

						if (t != null) {
							for (int i = 0; i < t.length; i++) {
								t[i].updatePositionFromCorners();
								success &= t[i].test(false);
							}
						}
					} catch (Exception e) {}
					System.out.println("Corner 1 in neighbor 1");
					success &= test(false);
					System.out.println("Tetrahedron fix success=" + success);
					return success;
				} else if (neighbor[2].isDefinitelyInSegment(cornerpos2_0, cornerpos2_1,
						cornerpos2_2)) {
					float epsilon = 0.00001f;

					// Vector3f negNormal = new Vector3f();
					while (plane[2].classifyPoint2(cornerpos2_0, cornerpos2_1, cornerpos2_2)
							!= Plane.PLANE_INSIDE) {
						// negNormal.set(-plane[2].normalX, -plane[2].normalY, -plane[2].normalZ);
						// negNormal.scale(epsilon);
						// cornerPos[2].add(negNormal);
						cornerpos2_0 -= plane[2].normalX * epsilon;
						cornerpos2_1 -= plane[2].normalY * epsilon;
						cornerpos2_2 -= plane[2].normalZ * epsilon;
						epsilon *= 2.0f;
					}
					float f[] = cornerObj[2].getPositionReference();

					f[0] = cornerpos2_0;
					f[1] = cornerpos2_1;
					f[2] = cornerpos2_2;
					// cornerPos[2].get(f);
					// cornerObj[2].setPosition(f);
					
					boolean success = true;

					try {
						PickableTetrahedronObject pto = (PickableTetrahedronObject) cornerObj[2];
						Tetrahedron t[] = pto.getTetrahedra();

						if (t != null) {
							for (int i = 0; i < t.length; i++) {
								t[i].updatePositionFromCorners();
								success &= t[i].test(false);
							}
						}
					} catch (Exception e) {}
					System.out.println("Corner 2 in neighbor 2");
					success &= test(false);
					System.out.println("Tetrahedron fix success=" + success);
					return success;
				}
			}
			
			return false;
		}
		return true;
		
		// float f[][] = new float[4][3];
		// float midPoint[] = new float[3];
		// for (int i=0; i<4; i++) {
		// cornerPos[i].get(f[i]);
		// for (int j=0; j<3; j++) {
		// midPoint[j]+=(f[i][j]/4);;
		// }
		// }
		// Point3f testPoint = new Point3f(midPoint);
		// if (!isInside(testPoint)) {
		//
		// System.out.println("Center "+this.cornerpos0_0+", "+this.cornerpos0_1+", "+this.cornerpos0_2);
		// System.out.println("Vertex1 "+this.cornerpos1_0+", "+this.cornerpos1_1+", "+this.cornerpos1_2);
		// System.out.println("Vertex2 "+this.cornerpos2_0+", "+this.cornerpos2_1+", "+this.cornerpos2_2);
		// System.out.println("Vertex3 "+this.cornerpos3_0+", "+this.cornerpos3_1+", "+this.cornerpos3_2);
		// return false;
		// }
		// return true;
		//
	}
	
	/**
	 * Returns true if the point is in this segment.
	 * @param pointx The point x coordinate.
	 * @param pointy The point y coordinate.
	 * @param pointz The point z coordinate.
	 * @return True if the point is inside the segment, otherwise false.
	 */
	public boolean isInSegment(float pointx, float pointy, float pointz) {
		for (int i = 0; i < 3; i++) {
			if (plane[i].classifyPoint(pointx, pointy, pointz) == Plane.PLANE_OUTSIDE) {
				return false;
			}
		}
		
		return true;
	}
	
	/**
	 * Returns true if the point is outside the segment plane.
	 * @param pointx The point x coordinate.
	 * @param pointy The point y coordinate.
	 * @param pointz The point z coordinate.
	 * @return True if the point is outside the segment plane, otherwise false.
	 */
	public final boolean isOutsideSegmentPlane(float pointx, float pointy, float pointz) {
		if (plane[3].classifyPoint(pointx, pointy, pointz) == Plane.PLANE_OUTSIDE) {
			return true;
		}
		return false;
	}
	
	/**
	 * Disconnects the tetrahedron from the neighbors. Not yet implemented.
	 */
	public void disconnect() {}
	
	/**
	 * Increased the color count.
	 */
	public void addColor() {
		totalIndices++;
	}
	
	/**
	 * Allocates space for the image indices.
	 */
	public void finalizeColorIndices() {
		colorIndices = new int[totalIndices];
		weight = new float[totalIndices][4];
		position = new float[totalIndices * 3];
	}
	
	/**
	 * Adds image colors.
	 * @param addColor The color.
	 * @param point The position of the indices.
	 */
	public void addColor(int addColor, float point[]) {
		colorIndices[indicesUsed] = addColor;
		float w1, w2, w3, w4;

		w1 = volume(cornerpos1_0, cornerpos1_1, cornerpos1_2, cornerpos2_0,
				cornerpos2_1, cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2,
				point[0], point[1], point[2])
				/ myVolume;
		w2 = volume(cornerpos0_0, cornerpos0_1, cornerpos0_2, cornerpos2_0,
				cornerpos2_1, cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2,
				point[0], point[1], point[2])
				/ myVolume;
		w3 = volume(cornerpos1_0, cornerpos1_1, cornerpos1_2, cornerpos0_0,
				cornerpos0_1, cornerpos0_2, cornerpos3_0, cornerpos3_1, cornerpos3_2,
				point[0], point[1], point[2])
				/ myVolume;
		w4 = volume(cornerpos1_0, cornerpos1_1, cornerpos1_2, cornerpos2_0,
				cornerpos2_1, cornerpos2_2, cornerpos0_0, cornerpos0_1, cornerpos0_2,
				point[0], point[1], point[2])
				/ myVolume;
		
		weight[indicesUsed][0] = w1;
		weight[indicesUsed][1] = w2;
		weight[indicesUsed][2] = w3;
		weight[indicesUsed][3] = w4;
		position[indicesUsed * 3] = point[0];
		position[indicesUsed * 3 + 1] = point[1];
		position[indicesUsed * 3 + 2] = point[2];
		indicesUsed++;
	}
	
	/**
	 * Returns the positions.
	 * @return The positions.
	 */
	public float[] getPositions() {
		return position;
	}
	
	/**
	 * Computes the volume of a tetrahedron.
	 * 
	 */
	public final float volume(float x1, float y1, float z1,
			float x2, float y2, float z2,
			float x3, float y3, float z3,
			float x4, float y4, float z4) {
		
		float vol = (1.0f / 6.0f)
				* (x1 * y2 * (z3 - z4) + x1 * y3 * (z4 - z2) + x1 * y4 * (z2 - z3)
						+ x2 * y1 * (z4 - z3) + x2 * y3 * (z1 - z4) + x2 * y4 * (z3 - z1)
						+ x3 * y1 * (z2 - z4) + x3 * y2 * (z4 - z1) + x3 * y4 * (z1 - z2)
						+ x4 * y1 * (z3 - z2) + x4 * y2 * (z1 - z3) + x4 * y3 * (z2 - z1));

		if (vol < 0.0f) {
			vol = -vol;
		}
		return vol;
	}
	
	/**
	 * Returns the color indices.
	 * @return The color indices.
	 */
	public int[] getColorIndices() {
		return colorIndices;
	}
	
	/**
	 * Returns the weights of the image indices.
	 * @return The weights of the image indices.
	 */
	public float[][] getWeights() {
		return weight;
	}
	
	/**
	 * Returns the corners.
	 * @return The corners.
	 */
	public float[][] getCorners() {
		float corny[][] = new float[4][];

		for (int i = 0; i < 4; i++) {
			corny[i] = this.cornerObj[i].getPosition();
		}
		return corny;
	}
	
	/**
	 * Returns the initial corners.
	 * @return The initial corners.
	 */
	public float[][] getStartCorners() {
		return new float[][] { { cornerpos0_0, cornerpos0_1, cornerpos0_2 },
			{ cornerpos1_0, cornerpos1_1, cornerpos1_2 },
			{ cornerpos2_0, cornerpos2_1, cornerpos2_2 },
			{ cornerpos3_0, cornerpos3_1, cornerpos3_2 }};
	}
	
	/**
	 * Finds the intersection with a plane.
	 * @param p The plane.
	 * @param al The list of points on the surface that define the intersection.
	 * @param from The last tetrahedron visited.
	 * @param start The first tetrahedron visited.
	 * @param scale The color space scale.
	 * @param posDiffX Used to decide positive/negative chroma values.
	 * @param posDiffZ Used to decide positive/negative chroma values.
	 * @param hue The hue that defines the plane.
	 * @return 4 If plane doesn't intersect the start tetrahedron. 2 if there was a
	 * floating point precision error. ...
	 */
	public int findIntersection(
			Plane p,
			ArrayList al,
			Tetrahedron from,
			Tetrahedron start,
			float scale[],
			boolean posDiffX,
			boolean posDiffZ,
			float hue) {
		if (start == this) {
			return 0;
		}
			
		int numOnInside = 0;
		
		int k = p.classifyPoint2(cornerpos1_0, cornerpos1_1, cornerpos1_2);
		int l = p.classifyPoint2(cornerpos2_0, cornerpos2_1, cornerpos2_2);
		int m = p.classifyPoint2(cornerpos3_0, cornerpos3_1, cornerpos3_2);
		
		if (k == Plane.IN_PLANE || l == Plane.IN_PLANE || m == Plane.IN_PLANE) {
			return 1;
		}
		if (k == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		if (l == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		if (m == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		
		if (numOnInside == 0 || numOnInside == 3) {
			if (from == null) {
				return 4;
			}
			return 2;
		}
		
		if (from == null) {
			if (k != l) {
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;

				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = (v3y) * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}
				al.add(gbp);
				return neighbor[0].findIntersection(p, al, this, this, scale, posDiffX,
						posDiffZ, hue);
				
			} else {
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos3_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}

				al.add(gbp);
				return neighbor[2].findIntersection(p, al, this, this, scale, posDiffX,
						posDiffZ, hue);
			}
		}
		
		if (this.neighbor[0] == from) {
			if (l != m) {
				
				float t = p.intersectionTimeAsWeight(cornerpos2_0, cornerpos2_1,
						cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos2_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos2_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos2_2 + oneMinusT * cornerpos3_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}
				al.add(gbp);
				return neighbor[1].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos3_0, cornerpos3_1,
						cornerpos3_2, cornerpos1_0, cornerpos1_1, cornerpos1_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos3_0 + oneMinusT * cornerpos1_0;
				float v3y = t * cornerpos3_1 + oneMinusT * cornerpos1_1;
				float v3z = t * cornerpos3_2 + oneMinusT * cornerpos1_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}

				al.add(gbp);
				return neighbor[2].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
			}
		}
		if (this.neighbor[1] == from) {
			if (m != k) {
				
				float t = p.intersectionTimeAsWeight(cornerpos3_0, cornerpos3_1,
						cornerpos3_2, cornerpos1_0, cornerpos1_1, cornerpos1_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos3_0 + oneMinusT * cornerpos1_0;
				float v3y = t * cornerpos3_1 + oneMinusT * cornerpos1_1;
				float v3z = t * cornerpos3_2 + oneMinusT * cornerpos1_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}

				al.add(gbp);
				return neighbor[2].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}
				al.add(gbp);
				return neighbor[0].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
			}
		}
		if (this.neighbor[2] == from) {
			if (k != l) {
				
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}

				al.add(gbp);
				return neighbor[0].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos2_0, cornerpos2_1,
						cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos2_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos2_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos2_2 + oneMinusT * cornerpos3_2;
				
				GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();

				gbp.y = v3y * scale[1];
				float xDiff = (v3x - 0.5f) * scale[0];
				float zDiff = (v3z - 0.5f) * scale[2];

				gbp.chroma = (float) Math.sqrt(xDiff * xDiff + zDiff * zDiff);
				if (Math.abs(v3x - 0.5f) > Math.abs(v3z - 0.5f)) {
					if ((v3x > 0.5f) != posDiffX) {
						gbp.chroma = -gbp.chroma;
					}
				} else {
					if ((v3z > 0.5f) != posDiffZ) {
						gbp.chroma = -gbp.chroma;
					}
				}
				al.add(gbp);
				return neighbor[1].findIntersection(p, al, this, start, scale, posDiffX,
						posDiffZ, hue);
			}
		}
		return 3;	
	}

	/**
	 * Finds the intersection with a plane.
	 * @param p The plane.
	 * @param al The list of points on the surface that define the intersection.
	 * @param from The last tetrahedron visited.
	 * @param start The first tetrahedron visited.
	 * @return 4 If plane doesn't intersect the start tetrahedron. 2 if there was a
	 * floating point precision error. ...
	 */
	public int findIntersection(Plane p, ArrayList al, Tetrahedron from, Tetrahedron start) {
		if (start == this) {
			return 0;
		}
			
		int numOnInside = 0;
		
		int k = p.classifyPoint2(cornerpos1_0, cornerpos1_1, cornerpos1_2);
		int l = p.classifyPoint2(cornerpos2_0, cornerpos2_1, cornerpos2_2);
		int m = p.classifyPoint2(cornerpos3_0, cornerpos3_1, cornerpos3_2);
		
		if (k == Plane.IN_PLANE || l == Plane.IN_PLANE || m == Plane.IN_PLANE) {
			return 1;
		}
		if (k == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		if (l == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		if (m == Plane.PLANE_INSIDE) {
			numOnInside++;
		}
		
		if (numOnInside == 0 || numOnInside == 3) {
			if (from == null) {
				return 4;
			}
			return 2;
		}
		
		if (from == null) {
			if (k != l) {
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[0].findIntersection(p, al, this, this);
				
			} else {
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos3_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[2].findIntersection(p, al, this, this);
			}
		}
		
		if (this.neighbor[0] == from) {
			if (l != m) {
				
				float t = p.intersectionTimeAsWeight(cornerpos2_0, cornerpos2_1,
						cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos2_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos2_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos2_2 + oneMinusT * cornerpos3_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[1].findIntersection(p, al, this, start);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos3_0, cornerpos3_1,
						cornerpos3_2, cornerpos1_0, cornerpos1_1, cornerpos1_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos3_0 + oneMinusT * cornerpos1_0;
				float v3y = t * cornerpos3_1 + oneMinusT * cornerpos1_1;
				float v3z = t * cornerpos3_2 + oneMinusT * cornerpos1_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[2].findIntersection(p, al, this, start);
			}
		}
		if (this.neighbor[1] == from) {
			if (m != k) {
				
				float t = p.intersectionTimeAsWeight(cornerpos3_0, cornerpos3_1,
						cornerpos3_2, cornerpos1_0, cornerpos1_1, cornerpos1_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos3_0 + oneMinusT * cornerpos1_0;
				float v3y = t * cornerpos3_1 + oneMinusT * cornerpos1_1;
				float v3z = t * cornerpos3_2 + oneMinusT * cornerpos1_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[2].findIntersection(p, al, this, start);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[0].findIntersection(p, al, this, start);
			}
		}
		if (this.neighbor[2] == from) {
			if (k != l) {
				
				float t = p.intersectionTimeAsWeight(cornerpos1_0, cornerpos1_1,
						cornerpos1_2, cornerpos2_0, cornerpos2_1, cornerpos2_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos1_0 + oneMinusT * cornerpos2_0;
				float v3y = t * cornerpos1_1 + oneMinusT * cornerpos2_1;
				float v3z = t * cornerpos1_2 + oneMinusT * cornerpos2_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[0].findIntersection(p, al, this, start);
				
			} else {
				
				float t = p.intersectionTimeAsWeight(cornerpos2_0, cornerpos2_1,
						cornerpos2_2, cornerpos3_0, cornerpos3_1, cornerpos3_2);
				float oneMinusT = 1 - t;
				float v3x = t * cornerpos2_0 + oneMinusT * cornerpos3_0;
				float v3y = t * cornerpos2_1 + oneMinusT * cornerpos3_1;
				float v3z = t * cornerpos2_2 + oneMinusT * cornerpos3_2;

				al.add(new Vector3f(v3x, v3y, v3z));
				return neighbor[1].findIntersection(p, al, this, start);
			}
		}
		return 3;	
	}

	/**
	 * Find the closest point on a line to a point.
	 * @param lineStart The line start point.
	 * @param lineEnd The line end point.
	 * @param point The point.
	 * @return The closest point on the line.
	 */
	public static final Vector3f closestPointOnLine(Vector3f lineStart, Vector3f lineEnd, Vector3f point) {
		cVector.sub(point, lineStart);
		
		Vector3f v = new Vector3f();

		v.sub(lineEnd, lineStart);
		float d = v.length();

		v.x /= d;
		v.y /= d;
		v.z /= d;
		
		float t = v.dot(cVector);

		if (t < 0) {
			return lineStart;
		}
		if (t > d) {
			return lineEnd;
		}
		v.scale(t);

		v.add(lineStart);
		return v;
	}
	
	/**
	 * Find the gamut boundary.
	 * @param p The intersecting plane.
	 * @param tetra The gamut tetrahedrons.
	 * @param startSearch The start tetrahedron.
	 * @param al The list of boundary points.
	 * @return True if successfull.
	 */
	public static boolean findBoundary(Plane p, Tetrahedron tetra[], Tetrahedron startSearch, ArrayList al) {
		int res;

		do {
			res = startSearch.findIntersection(p, al, null, null);
			if (res == 0) {
				return true;
			}
			p.incDistance();
			al.clear();
		} while (res != 4);
		
		int c = 0;

		while (res != 0 && c++ < 20) {
			for (int i = 0; i < tetra.length; i++) {
				res = tetra[i].findIntersection(p, al, null, null);
				if (res != 4) {
					break;
				}
			}
			if (res != 0) {
				p.incDistance();
				al.clear();
			}
		}
		return (res == 0);
	}
	
	/**
	 * Find the gamut boundary.
	 * @param p The plane.
	 * @param tetra The gamut tetrahedrons.
	 * @param startSearch The start tetrahedron.
	 * @param al The list of points on the surface that define the intersection.
	 * @param scale The color space scale.
	 * @param posDiffX Used to decide positive/negative chroma values.
	 * @param posDiffZ Used to decide positive/negative chroma values.
	 * @param hue The hue that defines the plane.
	 * @return True if successfull.
	 */
	public static boolean findBoundary(Plane p, Tetrahedron tetra[], Tetrahedron startSearch, ArrayList al, float scale[], boolean posDiffX, boolean posDiffZ, float hue) {
		int res;

		do {
			res = startSearch.findIntersection(p, al, null, null, scale, posDiffX,
					posDiffZ, hue);
			if (res == 0) {
				return true;
			}
			p.incDistance();
			al.clear();
		} while (res != 4);
		
		int c = 0;

		while (res != 0 && c++ < 20) {
			for (int i = 0; i < tetra.length; i++) {
				res = tetra[i].findIntersection(p, al, null, null, scale, posDiffX, posDiffZ,
						hue);
				if (res != 4) {
					break;
				}
			}
			if (res != 0) {
				p.incDistance();
				al.clear();
			}
		}
		return (res == 0);
	}
	
	/**
	 * Find the closest intersection point.
	 * @param al The gamut boundary points.
	 * @param pointx The chroma coordinate of the start point.
	 * @param pointy The lightness coordinate of the start point.
	 * @param rayx The x coordinate of the ray vector.
	 * @param rayy The y coordinate of the ray vector.
	 * @return The point on the gamut boundary where the ray intersects the surface.
	 */
	public static final GamutBoundaryPoint findClosestIntersection(ArrayList al, float pointx, float pointy, float rayx, float rayy) {
		GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();
		float closestIntersection = Float.MAX_VALUE;
		// float epsilon = 0.000001f;
		// int nr = 0;
		// do {
		int num = al.size();

		for (int i = 0; i < num; i++) {
			GamutBoundaryPoint a = (GamutBoundaryPoint) al.get(i);
			GamutBoundaryPoint b = (GamutBoundaryPoint) al.get((i + 1) % num);
			float t = GamutBoundaryPoint.distanseAlongLine(a.chroma, a.y, b.chroma, b.y,
					pointx, pointy, rayx, rayy);

			if (t >= -0.0001f && t <= 1 + 0.0001f) {
				float chroma = a.chroma + t * (b.chroma - a.chroma);

				if (chroma >= 0) {
					float y = a.y + t * (b.y - a.y);
					float dist = (chroma - pointx) * (chroma - pointx)
							+ (y - pointy) * (y - pointy);

					if (dist < closestIntersection) {
						closestIntersection = dist;
						gbp.chroma = chroma;
						gbp.y = y;
					}
				}
			}
		}
		// epsilon *=2.0f;
		// if (nr++>10) {
		/*
		 */
		// return null;
		// break;
		// }
		// } while (closestIntersection==Float.MAX_VALUE);
		// System.out.println("N"+nr);
		if (closestIntersection == Float.MAX_VALUE) {

			/*
			 System.out.println("");
			 System.out.println("Num "+num);
			 //System.out.println("Epsilon "+epsilon);
			 System.out.println("Point: "+pointx+","+pointy);
			 System.out.println("Ray: "+rayx+","+rayy);
			 for (int i=0; i<num; i++) {
			 GamutBoundaryPoint a = (GamutBoundaryPoint)al.get(i);
			 System.out.println("gbp: "+a.chroma+","+a.y);
			 }
			 */
			GamutBoundaryPoint.release(gbp);
			return null;
		}
		return gbp;
	}
	
	/**
	 * Find the extreme intersection point.
	 * @param al The gamut boundary points.
	 * @param pointx The chroma coordinate of the start point.
	 * @param pointy The lightness coordinate of the start point.
	 * @param rayx The x coordinate of the ray vector.
	 * @param rayy The y coordinate of the ray vector.
	 * @return The point on the gamut boundary where the ray intersects the
	 * surface, that has the largest chroma coordinate.
	 */
	public static final GamutBoundaryPoint findExtremeIntersection(ArrayList al, float pointx, float pointy, float rayx, float rayy) {
		GamutBoundaryPoint gbp = GamutBoundaryPoint.getGBP();
		float closestIntersection = -1000.0f;
		boolean found = false;
		int num = al.size();

		for (int i = 0; i < num; i++) {
			GamutBoundaryPoint a = (GamutBoundaryPoint) al.get(i);
			GamutBoundaryPoint b = (GamutBoundaryPoint) al.get((i + 1) % num);
			float t = GamutBoundaryPoint.distanseAlongLine(a.chroma, a.y, b.chroma, b.y,
					pointx, pointy, rayx, rayy);

			if (t >= -0.0001f && t <= 1 + 0.0001f) {
				float chroma = a.chroma + t * (b.chroma - a.chroma);

				if (chroma >= closestIntersection) {
					closestIntersection = chroma;
					found = true;
					float y = a.y + t * (b.y - a.y);

					gbp.chroma = chroma;
					gbp.y = y;
				}
			}
		}
		if (!found) {
			for (int i = 0; i < num; i++) {
				GamutBoundaryPoint a = (GamutBoundaryPoint) al.get(i);
				GamutBoundaryPoint b = (GamutBoundaryPoint) al.get((i + 1) % num);
				float t = GamutBoundaryPoint.distanseAlongLine(a.chroma, a.y, b.chroma, b.y,
						pointx, pointy, rayx, rayy);
				float chroma = a.chroma + t * (b.chroma - a.chroma);
				float y = a.y + t * (b.y - a.y);

				if (t < 0) {
					float distSquared = (a.y - y) * (a.y - y)
							+ (a.chroma - chroma) * (a.chroma - chroma);

					if (distSquared < 1) {
						if (chroma >= closestIntersection) {
							closestIntersection = chroma;
							found = true;
							gbp.chroma = chroma;
							gbp.y = y;
						}
					}
				} else {
					float distSquared = (b.y - y) * (b.y - y)
							+ (b.chroma - chroma) * (b.chroma - chroma);

					if (distSquared < 1) {
						if (chroma >= closestIntersection) {
							closestIntersection = chroma;
							found = true;
							gbp.chroma = chroma;
							gbp.y = y;
						}
					}
				}
			}
			if (!found) {
				System.out.println("");
				System.out.println("Num " + num);
				// System.out.println("Epsilon "+epsilon);
				System.out.println("Point: " + pointx + "," + pointy);
				System.out.println("Ray: " + rayx + "," + rayy);
				for (int i = 0; i < num; i++) {
					GamutBoundaryPoint a = (GamutBoundaryPoint) al.get(i);

					System.out.println("gbp: " + a.chroma + "," + a.y);
				}
				GamutBoundaryPoint.release(gbp);
				return null;
			}
		}
		return gbp;
	}
}


/**
 * A class used as an argument for a forced tetrahedron search.
 */
class TetrahedronSearchArgument {
	
	/**
	 * The distance to the nearest tetrahedron.
	 */
	private float nearestDistance;
	
	/**
	 * The nearest tetrahedron.
	 */
	private Tetrahedron nearestTetrahedron;
	
	/**
	 * Constructs a TetrahedronSearchArgument with an initial distance.
	 * @param dist The distance.
	 */
	public TetrahedronSearchArgument(float dist) {
		nearestDistance = dist;
	}
	
	/**
	 * Sets the distance.
	 * @param f The distance.
	 */
	public void setDistance(float f) {
		nearestDistance = f;
	}
	
	/**
	 * Returns the distance.
	 * @return The distance.
	 */
	public float getDistance() {
		return nearestDistance;
	}
	
	/**
	 * Sets the nearest tetrahedron.
	 * @param t The tetrahedron.
	 */
	public void setTetrahedron(Tetrahedron t) {
		nearestTetrahedron = t;
	}
	
	/**
	 * Returns the nearest tetrahedron.
	 * @return The tetrahedron.
	 */
	public Tetrahedron getTetrahedron() {
		return nearestTetrahedron;
	}
}
