/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


import javax.vecmath.*;


/**
 * A class used to represent a triangle in 3D.
 */
public class Triangle {

	/**
	 * The vertices of this triangle.
	 */
	private Vector3f vertex[];
	
	/**
	 * Constructs a Triangle from three points.
	 * @param v1 Point 1.
	 * @param v2 Point 2.
	 * @param v3 Point 3.
	 */
	public Triangle(Vector3f v1, Vector3f v2, Vector3f v3) {
		vertex = new Vector3f[3];
		vertex[0] = v1;
		vertex[1] = v2;
		vertex[2] = v3;
	}
	
	/**
	 * Finds the closest point on the triangle.
	 * @param pointP The point.
	 * @return The closest point on the triangle.
	 */
	public Point3f closestPointOnTriangle(Point3f pointP) {
		Vector3f point = new Vector3f(pointP);
		
		Vector3f rab = closestPointOnLine(point, vertex[0], vertex[2]);
		Vector3f rbc = closestPointOnLine(point, vertex[2], vertex[1]);
		Vector3f rca = closestPointOnLine(point, vertex[1], vertex[0]);
		Vector3f tmp = new Vector3f();

		tmp.sub(point, rab);
		double dab = tmp.length();

		tmp.sub(point, rbc);
		double dbc = tmp.length();

		tmp.sub(point, rca);
		double dca = tmp.length();

		double min = dab;
		Vector3f result = rab;

		if (dbc < min) {
			min = dbc;
			result = rbc;
		}
		if (dca < min) {
			result = rca;
		}
		return(new Point3f(result));
	}
	
	/**
	 * Finds the closest point on a line.
	 * @param point The point.
	 * @param lineStart The start point of the line.
	 * @param lineEnd The end point of the line.
	 * @return The closest point on the line.
	 */
	private Vector3f closestPointOnLine(Vector3f point, Vector3f lineStart, Vector3f lineEnd) {
		Vector3f result;
		Vector3f c, v, p = new Vector3f(point);

		c = new Vector3f();
		c.sub(p, lineStart);
		v = new Vector3f();
		v.sub(lineEnd, lineStart);
		float d = v.length();

		v.normalize();
		float t = v.dot(c);

		if (t < 0.0f) {
			return (lineStart);
		}
		if (t > d) {
			return (lineEnd);
		}
		v.scale(t);
		result = new Vector3f();
		result.add(lineStart, v);
		return (result);
	}
	
	/**
	 * Checks whether the point is in the triangle.
	 * @param pointP The point.
	 * @return True if the point lies inside the triangle, otherwise false.
	 */
	boolean pointInTriangle(Point3f pointP) {
		Vector3f point = new Vector3f(pointP);
		Vector3f v1 = new Vector3f(), v2 = new Vector3f(), v3 = new Vector3f();

		v1.sub(point, vertex[0]);
		v2.sub(point, vertex[2]);
		v3.sub(point, vertex[1]);
		v1.normalize();
		v2.normalize();
		v3.normalize();
		float totalangle = (float) Math.acos(v1.dot(v2));

		totalangle += (float) Math.acos(v2.dot(v3));
		totalangle += (float) Math.acos(v3.dot(v1));
		if ((float) Math.abs(totalangle - 2 * Math.PI) <= 0.005f) {
			return true;
		}
		return false;
	}
}
