/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.triangulation;


/**
 * A class that is used to represent a vertex in 3D.
 */
public class Vertex implements Comparable {
	
	/**
	 * Used to store information about a vertex.
	 */
	public int data;
	
	/**
	 * The vertex is assigned to a facet. Used to construct the convex hull.
	 */
	boolean isAssigned;
	
	/**
	 * The dimension of the vertex.
	 */
	int dimension;
	
	/**
	 * The vertex coordinates.
	 */
	public double coordinates[];
  
	/**
	 * Constructs a vertex in 3D.
	 */
	public Vertex() {
		this.coordinates = new double[3];
		dimension = 3;
	}
  
	/**
	 * Constructs a 2D vertex.
	 * @param x The x coordinate.
	 * @param y The x coordinate.
	 */
	public Vertex(double x, double y) {
		this.coordinates = new double[2];
		dimension = 2;
		this.coordinates[0] = x;
		this.coordinates[1] = y;
	}
  
	/**
	 * Retrieves the vertex coordinates.
	 * @return A double array of vertex coordinates.
	 */
	public double[] get() {
		return coordinates;
	}

	/**
	 * Constructs a 3D vertex.
	 * @param x The x coordinate.
	 * @param y The y coordinate.
	 * @param z The z coordinate.
	 */
	public Vertex(double x, double y, double z) {
		this.coordinates = new double[3];
		dimension = 3;
		this.coordinates[0] = x;
		this.coordinates[1] = y;
		this.coordinates[2] = z;
	}
  
	/**
	 * Sets the vertex coordinates.
	 * @param x The x coordinate.
	 * @param y The y coordinate.
	 * @param z The z coordinate.
	 */
	public void set(double x, double y, double z) {
		coordinates[0] = x;
		coordinates[1] = y;
		coordinates[2] = z;
	}
  
	/**
	 * Constructs a 4D vertex.
	 * @param x The x coordinate.
	 * @param y The y coordinate.
	 * @param z The z coordinate.
	 * @param w The w coordinate.
	 */
	public Vertex(double x, double y, double z, double w) {
		this.coordinates = new double[4];
		dimension = 4;
		this.coordinates[0] = x;
		this.coordinates[1] = y;
		this.coordinates[2] = z;
		this.coordinates[3] = w;
	}

	/**
	 * Adds a vertex by adding all coordinates.
	 * @param x The vertex to add.
	 * @return The resulting vertex.
	 */
	public Vertex add(Vertex x) {
		Vertex a = new Vertex();

		for (int i = 0; i < dimension; i++) {
			a.coordinates[i] = coordinates[i] + x.coordinates[i];
		}
		return a;
	}

	/**
	 * Subtracts a vertex by subtracting all coordinates.
	 * @param x The vertex to subtract.
	 * @return The resulting vertex.
	 */
	public Vertex subtract(Vertex x) {
		Vertex a = new Vertex();

		for (int i = 0; i < dimension; i++) {
			a.coordinates[i] = coordinates[i] - x.coordinates[i];
		}
		return a;
	}

	/**
	 * Scales the vertex by multiplying all coordinates with a scalar.
	 * @param x The scalar.
	 * @return The resulting vertex.
	 */
	public Vertex scale(double x) {
		Vertex a = new Vertex();

		for (int i = 0; i < dimension; i++) {
			a.coordinates[i] = x * coordinates[i];
		}
		return a;
	}

	/**
	 * Scales the vertex by multiplying all coordinates with scalars.
	 * @param x The x scalar.
	 * @param y The y scalar.
	 * @param z The z scalar.
	 * @return The resulting vertex.
	 */
	public Vertex scale(double x, double y, double z) {
		return new Vertex(x * coordinates[0], y * coordinates[1], z * coordinates[2]);
	}
	
	/**
	 * Scales the vertex by multiplying all coordinates with scalars.
	 * @param x The x scalar.
	 * @param y The y scalar.
	 * @param z The z scalar.
	 * @param w The z scalar.
	 * @return The resulting vertex.
	 */
	public Vertex scale(double x, double y, double z, double w) {
		return new Vertex(x * coordinates[0], y * coordinates[1], z * coordinates[2],
				w * coordinates[3]);
	}

	/**
	 * Computes the dot product.
	 * @param x The other vertex.
	 * @return The result.
	 */
	public double dot(Vertex x) {
		double d = 0;

		for (int i = 0; i < dimension; i++) {
			d += coordinates[i] * x.coordinates[i];
		}
		return d;
	}
	
	/**
	 * Normalizes the vertex
	 * @return The resulting vertex.
	 */
	public Vertex normalize() {
		return scale(1 / length());
	}

	/**
	 * Returns the length of this vertex/vector.
	 * @return The length.
	 */
	public double length() {
		return Math.sqrt(dot(this));
	}
	
	/**
	 * Computes the cross product.
	 * @param x The other vertex.
	 * @return The result.
	 */
	public Vertex cross(Vertex x) {
		return new Vertex(coordinates[1] * x.coordinates[2]
				- x.coordinates[1] * coordinates[2],
				coordinates[2] * x.coordinates[0] - x.coordinates[2] * coordinates[0],
				coordinates[0] * x.coordinates[1] - x.coordinates[0] * coordinates[1]);
	}
	
	/**
	 * Adds a dimension to this vertex by putting the new coordinate on the paraboloid 
	 * h = a*a+b*b+c*c+...
	 */
	public void addDimension() {
		double d = 0;

		for (int i = 0; i < dimension; i++) {
			d += (coordinates[i] * coordinates[i]);
		}
		addDimension(d);
	}
	
	/**
	 * Adds a dimension to this vertex by adding the specified coordinate.
	 * @param d The new coordinate.
	 */
	public void addDimension(double d) {
		double old[] = coordinates;

		coordinates = new double[dimension + 1];
		System.arraycopy(old, 0, coordinates, 0, old.length);
		coordinates[dimension] = d;
		dimension++;
	}
	
	/**
	 * Compares the vertex to another vertex.
	 * @param o The other vertex.
	 * @return 0, -1, +1 depending on the coordinates of the vertices.
	 */
	public int compareTo(Object o) {
		Vertex v = (Vertex) o;

		if (v.coordinates[0] > coordinates[0]) {
			return -1;
		}
		if (v.coordinates[0] < coordinates[0]) {
			return 1;
		}
		double diff = 0;

		for (int i = 1; i < dimension; i++) {
			diff += coordinates[i] * coordinates[i] - v.coordinates[i] * v.coordinates[i];
		}
		
		if (diff < 0.0) {
			return 1;
		}
		if (diff > 0.0) {
			return -1;
		}
		return 0;
	}
	
	/**
	 * Returns a String representation of this vertex.
	 * @return A string representing this vertex.
	 */
	public String toString() {
		String s = new String();

		for (int i = 0; i < dimension - 1; i++) {
			s += coordinates[i] + " ";
		}
		s += coordinates[dimension - 1];
		return s;
	}
}
