/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.utils;


/**
 * This abstract class is used to convert between byte arrays and integers of
 * different types.
 */
public abstract class ByteConverter {

	/**
	 * Converts a byte array to a type long.
	 *
	 * @param b Byte array to convert
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return Byte array as type long
	 */
	public static long byteArrayToLong(byte b[], boolean isLittleEndian) {
		long l = 0;

		if (isLittleEndian) {
			l = b[3] & 0xff;
			l = l << 8;
			l += b[2] & 0xff;
			l = l << 8;
			l += b[1] & 0xff;
			l = l << 8;
			l += b[0] & 0xff;

		} else {
			l = b[0] & 0xff;
			l = l << 8;
			l += b[1] & 0xff;
			l = l << 8;
			l += b[2] & 0xff;
			l = l << 8;
			l += b[3] & 0xff;
		}
		return l;
	}
	
	/**
	 * Converts a long integer to a byte array.
	 *
	 * @param l A long integer
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return The long integer as a byte array
	 */
	public static byte[] toByteArray(long l, boolean isLittleEndian) {
		byte b[] = new byte[8];

		if (isLittleEndian) {
			b[0] = (byte) (l & 0xff);
			l = l >> 8;
			b[1] = (byte) (l & 0xff);
			l = l >> 8;
			b[2] = (byte) (l & 0xff);
			l = l >> 8;
			b[3] = (byte) (l & 0xff);
			l = l >> 8;
			b[4] = (byte) (l & 0xff);
			l = l >> 8;
			b[5] = (byte) (l & 0xff);
			l = l >> 8;
			b[6] = (byte) (l & 0xff);
			l = l >> 8;
			b[7] = (byte) (l & 0xff);
		} else {
			b[7] = (byte) (l & 0xff);
			l = l >> 8;
			b[6] = (byte) (l & 0xff);
			l = l >> 8;
			b[5] = (byte) (l & 0xff);
			l = l >> 8;
			b[4] = (byte) (l & 0xff);
			l = l >> 8;
			b[3] = (byte) (l & 0xff);
			l = l >> 8;
			b[2] = (byte) (l & 0xff);
			l = l >> 8;
			b[1] = (byte) (l & 0xff);
			l = l >> 8;
			b[0] = (byte) (l & 0xff);
		}
		return b;
	}
	
	/**
	 * Converts a byte array to a type char.
	 *
	 * @param b byte array to convert
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return byte array as type char
	 */
	public static char byteArrayToChar(byte b[], boolean isLittleEndian) {
		char c = (char) (b[0] & 0xff);

		return c;
	}
	
	/**
	 * Convert an integer to a byte array
	 * @param i The integer
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return The integer as a byte array
	 */
	public static byte[] toByteArray(int i, boolean isLittleEndian) {
		byte b[] = new byte[4];

		if (isLittleEndian) {
			b[0] = (byte) (i & 0xff);
			i = i >> 8;
			b[1] = (byte) (i & 0xff);
			i = i >> 8;
			b[2] = (byte) (i & 0xff);
			i = i >> 8;
			b[3] = (byte) (i & 0xff);
		} else {
			b[3] = (byte) (i & 0xff);
			i = i >> 8;
			b[2] = (byte) (i & 0xff);
			i = i >> 8;
			b[1] = (byte) (i & 0xff);
			i = i >> 8;
			b[0] = (byte) (i & 0xff);
		}
		return b;
	}
	
	/**
	 * Converts a char to a byte array
	 * @param c The char
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return The char as a byte array
	 */
	public static byte[] toByteArray(char c, boolean isLittleEndian) {
		byte b[] = new byte[2];

		if (isLittleEndian) {
			b[0] = (byte) (c & 0xff);
			c = (char) (c >> 8);
			b[1] = (byte) (c & 0xff);
		} else {
			b[1] = (byte) (c & 0xff);
			c = (char) (c >> 8);
			b[0] = (byte) (c & 0xff);
		}
		return b;
	}
	
	/**
	 * Converts a bytearray to a type int.
	 *
	 * @param b Byte array to convert
	 * @param isLittleEndian Is the integer represented by little endian bytes
	 * @return Byte array as an integer
	 */	
	public static int byteArrayToInt(byte b[], boolean isLittleEndian) {
		int i = 0;

		if (isLittleEndian) {
			i = b[1] & 0xff;
			i = i << 8;
			i += b[0] & 0xff;
		} else {
			i = b[0] & 0xff;
			i = i << 8;
			i += b[1] & 0xff;
		}
		return i;
	}
}
