/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.utils;


import java.util.Arrays;

import icc3dpackage.colorspaces.*;


/**
 * This class provides two different approaches to converting between
 * absolute and relative colorimetric XYZ colors.
 */
public class Colorimetry {
	
	/**
	 * Converts absolute XYZ to relative XYZ. Basic white point scaling.
	 * @param xyz The color(s) that should be converted.
	 * @param relativeWhiteX Relative white X component.
	 * @param relativeWhiteY Relative white Y component.
	 * @param relativeWhiteZ Relative white Z component.
	 */
	public static void toRelativeXYZ(float xyz[], float relativeWhiteX, float relativeWhiteY, float relativeWhiteZ) {
		int num = xyz.length / 3;
		int q;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			xyz[q] = xyz[q] * XYZ_ColorSpace3D.whiteX / relativeWhiteX;
			xyz[q + 1] = xyz[q + 1] * XYZ_ColorSpace3D.whiteY / relativeWhiteY;
			xyz[q + 2] = xyz[q + 2] * XYZ_ColorSpace3D.whiteZ / relativeWhiteZ;
		}
	}
	
	/**
	 * Converts relative XYZ to absolute XYZ. Basic white point scaling.
	 * @param xyz The color(s) that should be converted.
	 * @param relativeWhiteX Relative white X component.
	 * @param relativeWhiteY Relative white Y component.
	 * @param relativeWhiteZ Relative white Z component.
	 */
	public static void toAbsoluteXYZ(float xyz[], float relativeWhiteX, float relativeWhiteY, float relativeWhiteZ) {
		int num = xyz.length / 3;
		int q;

		for (int i = 0; i < num; i++) {
			q = i * 3;
			xyz[q] = xyz[q] * relativeWhiteX / XYZ_ColorSpace3D.whiteX;
			xyz[q + 1] = xyz[q + 1] * relativeWhiteY / XYZ_ColorSpace3D.whiteY;
			xyz[q + 2] = xyz[q + 2] * relativeWhiteZ / XYZ_ColorSpace3D.whiteZ;
		}
	}
	
	/**
	 * Converts absolute XYZ to relative XYZ. Based on CAT02.
	 * @param xyz The color(s) that should be converted.
	 * @param relativeWhiteX Relative white X component.
	 * @param relativeWhiteY Relative white Y component.
	 * @param relativeWhiteZ Relative white Z component.
	 */
	public static void toRelativeXYZ_CAT02(float xyz[], float relativeWhiteX, float relativeWhiteY, float relativeWhiteZ) {
		int num = xyz.length / 3;
		int q;
		CAT02_ColorSpace3D catSpace = new CAT02_ColorSpace3D();
		float cat02[] = catSpace.fromCIEXYZ(xyz);
		float relativeWhiteXYZ[] = new float[] {relativeWhiteX, relativeWhiteY,
			relativeWhiteZ};
		float absoluteWhiteXYZ[] = new float[] {XYZ_ColorSpace3D.whiteX,
			XYZ_ColorSpace3D.whiteY, XYZ_ColorSpace3D.whiteZ};
		float relativeWhiteCAT[] = catSpace.fromCIEXYZ(relativeWhiteXYZ);
		float absoluteWhiteCAT[] = catSpace.fromCIEXYZ(absoluteWhiteXYZ);

		for (int i = 0; i < num; i++) {
			q = i * 3;
			cat02[q] = cat02[q] * absoluteWhiteCAT[0] / relativeWhiteCAT[0];
			cat02[q + 1] = cat02[q + 1] * absoluteWhiteCAT[1] / relativeWhiteCAT[1];
			cat02[q + 2] = cat02[q + 2] * absoluteWhiteCAT[2] / relativeWhiteCAT[2];
		}
		boolean dirty[] = new boolean[num];

		Arrays.fill(dirty, true);
		catSpace.toCIEXYZ(cat02, xyz, dirty);
	}
	
	/**
	 * Converts relative XYZ to absolute XYZ. Based on CAT02.
	 * @param xyz The color(s) that should be converted.
	 * @param relativeWhiteX Relative white X component.
	 * @param relativeWhiteY Relative white Y component.
	 * @param relativeWhiteZ Relative white Z component.
	 */
	public static void toAbsoluteXYZ_CAT02(float xyz[], float relativeWhiteX, float relativeWhiteY, float relativeWhiteZ) {
		int num = xyz.length / 3;
		int q;
		CAT02_ColorSpace3D catSpace = new CAT02_ColorSpace3D();
		float cat02[] = catSpace.fromCIEXYZ(xyz);
		float relativeWhiteXYZ[] = new float[] {relativeWhiteX, relativeWhiteY,
			relativeWhiteZ};
		float absoluteWhiteXYZ[] = new float[] {XYZ_ColorSpace3D.whiteX,
			XYZ_ColorSpace3D.whiteY, XYZ_ColorSpace3D.whiteZ};
		float relativeWhiteCAT[] = catSpace.fromCIEXYZ(relativeWhiteXYZ);
		float absoluteWhiteCAT[] = catSpace.fromCIEXYZ(absoluteWhiteXYZ);

		for (int i = 0; i < num; i++) {
			q = i * 3;
			cat02[q] = cat02[q] * relativeWhiteCAT[0] / absoluteWhiteCAT[0];
			cat02[q + 1] = cat02[q + 1] * relativeWhiteCAT[1] / absoluteWhiteCAT[1];
			cat02[q + 2] = cat02[q + 2] * relativeWhiteCAT[2] / absoluteWhiteCAT[2];
		}
		boolean dirty[] = new boolean[num];

		Arrays.fill(dirty, true);
		catSpace.toCIEXYZ(cat02, xyz, dirty);
	}
}
