/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.utils;


import java.io.*;


/**
 * This class is used as a wrapper to provide additional functions used when
 * accessing a RandomAccessFile with a specific byte order.
 */
public class RandomAccessFileWrapper {
	
	/** The wrapped RandomAccessFile */
	private RandomAccessFile raf;
	
	/** Indicates whether the file uses little endian byte order */
	private boolean isLittleEndian;

	/**
	 * Default constructor.
	 * @param raf RandomAccessFile to wrap
	 * @param isLittleEndian Indicates whether the file uses little endian byte order
	 */
	public RandomAccessFileWrapper(RandomAccessFile raf, boolean isLittleEndian) {
		this.raf = raf;
		this.isLittleEndian = isLittleEndian;
	} 
	
	/**
	 * Writes a float to the file
	 * @param f The float
	 */
	public void put(float f) {
		int i = Float.floatToIntBits(f);

		put(i);
	}
	
	/**
	 * Writes a double precision float to the file
	 * @param d The double
	 */
	public void put(double d) {
		long l = Double.doubleToLongBits(d);

		put(l);
	}
	
	/**
	 * Writes a long integer to the file
	 * @param l The long integer
	 */
	public void put(long l) {
		byte b[] = ByteConverter.toByteArray(l, isLittleEndian);

		try {
			raf.write(b);
		} catch (Exception e) {
			throw new RuntimeException("Error writing file");
		}
	}
	
	/**
	 * Writes a byte to the file
	 * @param l The byte
	 */
	public void put(byte l) {
		try {
			raf.writeByte(l);
		} catch (Exception e) {
			throw new RuntimeException("Error writing file");
		}
	}
	
	/**
	 * Writes an integer to the file
	 * @param l The integer
	 */
	public void put(int l) {
		byte b[] = ByteConverter.toByteArray(l, isLittleEndian);

		try {
			raf.write(b);
		} catch (Exception e) {
			throw new RuntimeException("Error writing file");
		}
	}
	
	/**
	 * Writes a char to the file
	 * @param l The char
	 */
	public void put(char l) {
		byte b[] = ByteConverter.toByteArray(l, isLittleEndian);

		try {
			raf.write(b);
		} catch (Exception e) {
			throw new RuntimeException("Error writing file");
		}
	}
	
	/**
	 * Reads four (4) bytes from the file and returns it as a long.
	 *
	 * @return long value
	 */
	public long getLong() {
		byte fourBytes[] = new byte[4];

		try {
			raf.readFully(fourBytes);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
		return ByteConverter.byteArrayToLong(fourBytes, isLittleEndian);
	}
	
	/**
	 * Reads four (4) bytes from the file and returns it as a float.
	 *
	 * @return long value
	 */
	public float getFloat() {
		long l = getLong();

		return Float.intBitsToFloat((int) l);
	}
	
	/**
	 * Reads four (8) bytes from the file and returns it as a double.
	 * @return double value
	 */
	public double getDouble() {
		byte b[] = new byte[8];

		try {
			raf.readFully(b);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
		
		long l = 0;

		if (isLittleEndian) {
			l = b[7] & 0xff;
			l = l << 8;
			l |= b[6] & 0xff;
			l = l << 8;
			l |= b[5] & 0xff;
			l = l << 8;
			l |= b[4] & 0xff;
			l = l << 8;
			l |= b[3] & 0xff;
			l = l << 8;
			l |= b[2] & 0xff;
			l = l << 8;
			l |= b[1] & 0xff;
			l = l << 8;
			l |= b[0] & 0xff;

		} else {
			l = b[0] & 0xff;
			l = l << 8;
			l |= b[1] & 0xff;
			l = l << 8;
			l |= b[2] & 0xff;
			l = l << 8;
			l |= b[3] & 0xff;
			l = l << 8;
			l |= b[4] & 0xff;
			l = l << 8;
			l |= b[5] & 0xff;
			l = l << 8;
			l |= b[6] & 0xff;
			l = l << 8;
			l |= b[7] & 0xff;
		}
		return Double.longBitsToDouble(l);
	}
	
	/**
	 * Reads one (1) byte from the file and returns it as a char.
	 *
	 * @return long value
	 */
	public char getChar() {
		byte oneByte[] = new byte[1];

		try {
			raf.readFully(oneByte);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
		return ByteConverter.byteArrayToChar(oneByte, isLittleEndian);
	}
	
	/**
	 * Reads eight (8) bytes from the file and returns it as a fraction.
	 * @return float value
	 */
	public float getRational() {
		long l1 = getLong();
		long l2 = getLong();

		return (float) l1 / (float) l2;
	}
	
	/**
	 * Reads two (2) bytes from the file and returns it as an int.
	 * @return int value
	 */	
	public int getInt() {
		byte twoBytes[] = new byte[2];

		try {
			raf.readFully(twoBytes);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
		return ByteConverter.byteArrayToInt(twoBytes, isLittleEndian);
	}
	
	/**
	 * Positions the seek pointer at offset position in the file.
	 * @param offset position to place the seek pointer
	 */	
	public void setOffset(long offset) {
		try {
			raf.seek(offset);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
	}
	
	/**
	 * Returns the current file offset.
	 * @return The offset.
	 */
	public long getOffset() {
		long off = 0;

		try {
			off = raf.getFilePointer();
		} catch (Exception e) {
			off = 0;
		}
		return off;
	}
	
	/**
	 * Reads a number of bytes from the file.
	 * @param b The array of bytes that should be filled.
	 */	
	public void getBytes(byte b[]) {
		try {
			raf.readFully(b);
		} catch (Exception e) {
			throw new RuntimeException("Error reading file");
		}
	}
	
	/**
	 * Returns the length of the file.
	 * @return The length of the file.
	 */
	public long getLength() {
		try {
			return raf.length();
		} catch (IOException ioe) {
			return 0;
		}
	}
	
	/**
	 * Sets the byte order used by the file.
	 * @param isLittleEndian Indicates whether the file uses little endian byte order
	 */
	public void setEndian(boolean isLittleEndian) {
		this.isLittleEndian = isLittleEndian;
	}
}
