/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.view3d;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import com.sun.j3d.utils.picking.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.pickable.*;


/**
 * A mouse behavior that gives the user the ability to move around, and select
 * and move objects in the 3D scene.
 */
public class MouseMoveBehavior extends MouseBehavior {
	
	/**
	 * No action.
	 */
	public static final int MOUSE_NONE = 0;
	
	/**
	 * Translation of world coordinates.
	 */
	public static final int MOUSE_TRANSLATE = 1;
	
	/**
	 * Zoom in/out.
	 */
	public static final int MOUSE_ZOOM = 2;
	
	/**
	 * Rotate world coordinates.
	 */
	public static final int MOUSE_ROTATE = 3;
	
	/**
	 * An object is selected, and should be manipulated.
	 */
	public static final int MOUSE_OBJECT_SELECTED = 4;
	
	private int type;
	private double rotateAngleX;
	private double rotateAngleY;
	private double rotateFactorX = 0.02;
	private double rotateFactorY = 0.02;
	private double translateFactorX = 0.002;
	private double translateFactorY = 0.002;
	private double zoomFactor = 0.02;
	private Vector3d translateTranslation = new Vector3d();
	private Vector3d zoomTranslation = new Vector3d();
	private MouseBehaviorCallback callback = null;
	
	private View3D view3D;
	private int picked;
	private MovingObject mo;
	private TransformGroup viewTransform, scaleTransform, translateTransform;
	private byte old[] = null, notOld[] = new byte[3];
	private float pointInMotion[], oldPosition[], newPosition[];
	
	private final float PICK_TOLERANCE = 7.0f;
	protected PickCanvas pickCanvas;
	
	private float transformedForward[];
	private float transformedUp[];
	private float transformedRight[];
	
	/**
	 * Creates a mouse behavior.
	 * @param view3D The View3D that contains the canvas and scene.
	 * @param canvas The canvas used to display the 3D scene.
	 * @param scene The BranchGroup that contains the scene. 
	 */
	public MouseMoveBehavior(View3D view3D, Canvas3D canvas, BranchGroup scene) {
		super(0);
		this.view3D = view3D;
		pickCanvas = new PickCanvas(canvas, scene);
		pickCanvas.setTolerance(PICK_TOLERANCE);
		pickCanvas.setMode(PickCanvas.GEOMETRY_INTERSECT_INFO);
		transformedForward = new float[3];
		transformedUp = new float[3];
		transformedRight = new float[3];
	}
	
	/**
	 * Initializes the mouse behavior.
	 */
	public void initialize() {
		super.initialize();
		if ((flags & INVERT_INPUT) == INVERT_INPUT) {
			invert = true;
			translateFactorX *= -1;
			translateFactorY *= -1;
			rotateFactorX *= -1;
			rotateFactorY *= -1;
			zoomFactor *= -1;
		}
	}
	
	/**
	 * Returns the zoom factor.
	 * @return The zoom factor.
	 */
	public double getZoomFactor() {
		return zoomFactor;
	}
	
	/**
	 * Sets the zoom factor.
	 * @param factor The zoom factor.
	 */
	public void setZoomFactor(double factor) {
		zoomFactor = factor;
	}
	
	/**
	 * Returns the X translation factor.
	 * @return The X translation factor.
	 */
	public double getTranslateFactorX() {
		return translateFactorX;
	}
	
	/**
	 * Returns the Y translation factor.
	 * @return The Y translation factor.
	 */
	public double getTranslateFactorY() {
		return translateFactorY;
	}
	
	/**
	 * Returns the X rotation factor.
	 * @return The X rotation factor.
	 */
	public double getRotateFactorX() {
		return rotateFactorX;
	}
	
	/**
	 * Returns the Y rotation factor.
	 * @return The Y rotation factor.
	 */
	public double getRotateFactorY() {
		return rotateFactorY;
	}
	
	/**
	 * Sets the translation factor.
	 * @param factor The translation factor.
	 */
	public void setTranslateFactor(double factor) {
		translateFactorX = translateFactorY = factor;
	}
	
	/**
	 * Sets the translation factor.
	 * @param xFactor The X translation factor.
	 * @param yFactor The Y translation factor.
	 */
	public void setTranslateFactor(double xFactor, double yFactor) {
		translateFactorX = xFactor;
		translateFactorY = yFactor;    
	}
	
	/**
	 * Sets the rotation factor.
	 * @param factor The rotation factor.
	 */
	public void setRotateFactor(double factor) {
		rotateFactorX = rotateFactorY = factor;
	}
	
	/**
	 * Sets the rotation factor.
	 * @param xFactor The X rotation factor.
	 * @param yFactor The Y rotation factor.
	 */
	public void setRotateFactor(double xFactor, double yFactor) {
		rotateFactorX = xFactor;
		rotateFactorY = yFactor;    
	}
	
	/**
	 * Processes mouse events.
	 * @param criteria The wake up criteria.
	 */
	public void processStimulus(Enumeration criteria) {
		WakeupCriterion wakeup;
		AWTEvent[] events;
		MouseEvent evt;
		
		while (criteria.hasMoreElements()) {
			wakeup = (WakeupCriterion) criteria.nextElement();
			if (wakeup instanceof WakeupOnAWTEvent) {
				events = ((WakeupOnAWTEvent) wakeup).getAWTEvent();
				if (events.length > 0) {
					evt = (MouseEvent) events[events.length - 1];
					doProcess(evt);
				}
			} else if (wakeup instanceof WakeupOnBehaviorPost) {
				while (true) {
					synchronized (mouseq) {
						if (mouseq.isEmpty()) {
							break;
						}
						evt = (MouseEvent) mouseq.remove(0);
						while ((evt.getID() == MouseEvent.MOUSE_DRAGGED) && !mouseq.isEmpty()
								&& (((MouseEvent) mouseq.get(0)).getID() == MouseEvent.MOUSE_DRAGGED)) {
							evt = (MouseEvent) mouseq.remove(0);
						}
					}
					doProcess(evt);
				}
			}
		}
		wakeupOn(mouseCriterion);
	}
	
	/**
	 * Processes the mouse event.
	 * @param evt The mouse event.
	 */
	void doProcess(MouseEvent evt) {
		int id = evt.getID();
		boolean shift = evt.isShiftDown();
		boolean alt = evt.isAltDown();
		boolean ctrl = evt.isControlDown();
		boolean meta = evt.isMetaDown();

		x = evt.getX();
		y = evt.getY();
		if (id == MouseEvent.MOUSE_DRAGGED) {
			int dx = x - x_last;
			int dy = y - y_last;

			if (type == MouseMoveBehavior.MOUSE_TRANSLATE && !reset) {
				if (((Math.abs(dy) < 50) && (Math.abs(dx) < 50))) {
					transformGroup.getTransform(currXform);
					translateTranslation.x = dx * translateFactorX;
					translateTranslation.y = -dy * translateFactorY;
					transformX.set(translateTranslation);
			    
					if (invert) {
						currXform.mul(currXform, transformX);
					} else {
						currXform.mul(transformX, currXform);
					}
					transformGroup.setTransform(currXform);
					// transformChanged(currXform);
					if (callback != null) {
						callback.transformChanged(MouseBehaviorCallback.TRANSLATE, currXform);
					}
				}
			} else if (type == MouseMoveBehavior.MOUSE_ROTATE) {
				rotateAngleX = dy * rotateFactorY;
				rotateAngleY = dx * rotateFactorX;
				transformX.rotX(rotateAngleX);
				transformY.rotY(rotateAngleY);
				transformGroup.getTransform(currXform);
		    
				Matrix4d mat = new Matrix4d();

				currXform.get(mat);
				currXform.setTranslation(new Vector3d(0.0, 0.0, 0.0));
				if (invert) {
					currXform.mul(currXform, transformX);
					currXform.mul(currXform, transformY);
				} else {
					currXform.mul(transformX, currXform);
					currXform.mul(transformY, currXform);
				}
	    
				Vector3d translation = new Vector3d(mat.m03, mat.m13, mat.m23);

				currXform.setTranslation(translation);
				transformGroup.setTransform(currXform);
				// transformChanged( currXform );
	    
				if (callback != null) {
					callback.transformChanged(MouseBehaviorCallback.ROTATE, currXform);
				}
			} else if (type == MouseMoveBehavior.MOUSE_ZOOM) {
				transformGroup.getTransform(currXform);
				zoomTranslation.z = dy * zoomFactor;
				transformX.set(zoomTranslation);
	    
				if (invert) {
					currXform.mul(currXform, transformX);
				} else {
					currXform.mul(transformX, currXform);
				}
				transformGroup.setTransform(currXform);
				// transformChanged( currXform );
	    
				if (callback != null) {
					callback.transformChanged(MouseBehaviorCallback.ZOOM, currXform);
				}
			} else if (type == MouseMoveBehavior.MOUSE_OBJECT_SELECTED) {
						
				float diffF[] = new float[3];
				float diffR[] = new float[3];

				if (!shift) {
					for (int i = 0; i < 3; i++) {
						diffF[i] = transformedUp[i] * -0.1f * (float) dy / 100.0f;
					}
				} else {
					for (int i = 0; i < 3; i++) {
						diffF[i] = transformedForward[i] * -0.1f * (float) dy / 100.0f;
					}
				}
				for (int i = 0; i < 3; i++) {
					diffR[i] = transformedRight[i] * -0.1f * (float) dx / 100.0f;
				}

				for (int i = 0; i < 3; i++) {
					pointInMotion[i] += diffF[i];
					pointInMotion[i] += diffR[i];
				}
			
				if (old == null) {
					old = mo.getObjects()[picked].getColor();
					oldPosition = mo.getObjects()[picked].getPosition();
				}
				pointInMotion = view3D.getConstraintsManager().getConstraint().getRealEndPoint(oldPosition,
						pointInMotion, mo.getObjects());
				view3D.getInteractionManager().getInteraction().update3D(picked, mo,
						mo.getObjects()[picked].getPosition(), pointInMotion);
				newPosition = mo.getObjects()[picked].getPosition();
				pointInMotion = mo.getObjects()[picked].getPosition();
				notOld = mo.getObjects()[picked].getColor();
				ColorSpace3D cs3d = mo.getColorSpace();
				float presColor[] = cs3d.toPresentation(newPosition);

				view3D.setStatusLeft("Position " + cs3d.getColorLabels()[0] + ":"
						+ presColor[0] + " " + cs3d.getColorLabels()[1] + ":" + presColor[1] + " "
						+ cs3d.getColorLabels()[2] + ":" + presColor[2]);
				if (view3D.getInteractionManager().getContinuousUpdate()) {
					updateImage();
				}
			}
				
		} else if (id == MouseEvent.MOUSE_PRESSED) {
			if (shift && !alt && !meta) {
				type = MouseMoveBehavior.MOUSE_TRANSLATE;
			} else if (!shift && !alt && !meta) {
				type = MouseMoveBehavior.MOUSE_ROTATE;
			} else if (alt && !meta) {
				type = MouseMoveBehavior.MOUSE_ZOOM;
			} else if (!alt && meta) {
				pickCanvas.setShapeLocation(x, y);
				Point3d eyePos = pickCanvas.getStartPosition();
				PickResult pickResult = pickCanvas.pickClosest();

				if (pickResult != null) {
					type = MouseMoveBehavior.MOUSE_OBJECT_SELECTED;
					// --------------------------------------------------------------------------
					Point3f destForwardPoint = new Point3f(new float[] {0.0f, 0.0f, -1.0f});
					Point3f destRightPoint = new Point3f(new float[] {-1.0f, 0.0f, 0.0f});
					Point3f destUpPoint = new Point3f(new float[] {0.0f, 1.0f, 0.0f});
					Point3f startPoint = new Point3f(new float[] {0.0f, 0.0f, 0.0f});
					Point3f forwardMotion = new Point3f();
					Point3f rightMotion = new Point3f();
					Point3f upMotion = new Point3f();
								
					Transform3D t3d = new Transform3D();

					viewTransform.getTransform(t3d);
					t3d.setTranslation(new Vector3d(0.0, 0.0, 0.0));
					t3d.setScale(1.0);
					t3d.invert();
					t3d.transform(destForwardPoint);
					t3d.transform(startPoint);
					t3d.transform(destRightPoint);
					t3d.transform(destUpPoint);
					translateTransform.getTransform(t3d);
					t3d.setTranslation(new Vector3d(0.0, 0.0, 0.0));
					t3d.setScale(1.0);
					t3d.invert();
					t3d.transform(destForwardPoint);
					t3d.transform(startPoint);
					t3d.transform(destRightPoint);
					t3d.transform(destUpPoint);
					scaleTransform.getTransform(t3d);
					t3d.setTranslation(new Vector3d(0.0, 0.0, 0.0));
					t3d.setScale(1.0);
					t3d.invert();
					t3d.transform(destForwardPoint);
					t3d.transform(startPoint);
					t3d.transform(destRightPoint);
					t3d.transform(destUpPoint);
					forwardMotion.sub(destForwardPoint, startPoint);
					rightMotion.sub(destRightPoint, startPoint);
					upMotion.sub(destUpPoint, startPoint);
					forwardMotion.get(transformedForward);
					upMotion.get(transformedUp);
					rightMotion.get(transformedRight);
					// --------------------------------------------------------------------------
					PickIntersection pickIntersection = pickResult.getClosestIntersection(eyePos);
					MovingObject mo = (MovingObject) pickResult.getObject();

					if (pickIntersection.getGeometryArray() instanceof PointArray) {
						int[] indices = pickIntersection.getPrimitiveVertexIndices();
						PointArray pa = (PointArray) pickIntersection.getGeometryArray();

						setSomethingPicked(mo.getStartingIndex() + indices[0], mo);
					} else {
						setSomethingPicked(mo.getStartingIndex(), mo);
					}
				} else {
					type = MouseMoveBehavior.MOUSE_NONE;
				}
			} else {
				type = MouseMoveBehavior.MOUSE_NONE;
			}
		} else if (id == MouseEvent.MOUSE_RELEASED) {
			type = MouseMoveBehavior.MOUSE_NONE;
			if (old != null) {
				updateImage();
			}
			view3D.setStatusLeft("ICC3D " + ICC3D.getVersion() + " ready");
		}
		x_last = x;
		y_last = y;
		reset = false;
	}
	
	/**
	 * Selects one of the moveable objects.
	 * @param i The index of the selected object.
	 * @param mo The MovingObject that is selected.
	 */
	public void setSomethingPicked(int i, MovingObject mo) {
		picked = i;
		this.mo = mo;
		pointInMotion = mo.getObjects()[picked].getPosition();
		ColorSpace3D cs3d = mo.getColorSpace();
		float presColor[] = cs3d.toPresentation(pointInMotion);

		view3D.setStatusLeft("Position " + cs3d.getColorLabels()[0] + ":"
				+ presColor[0] + " " + cs3d.getColorLabels()[1] + ":" + presColor[1] + " "
				+ cs3d.getColorLabels()[2] + ":" + presColor[2]);
	}
	
	/**
	 * Updates the 2D image after changes has been made to the 3D scene.
	 */
	private void updateImage() {
		view3D.getInteractionManager().getInteraction().update2D(picked, oldPosition,
				newPosition, old, notOld, mo);
		old = null;
	}
	
	/**
	 * Installs a MouseBehaviorCallback.
	 * @param callback The MouseBehaviorCallback.
	 */
	public void setupCallback(MouseBehaviorCallback callback) {
		this.callback = callback;
	}
	
	/**
	 * Updates the view transform.
	 * @param tg The new view transform.
	 */
	public void setViewTransform(TransformGroup tg) {
		viewTransform = tg;
	}
	
	/**
	 * Updates the scale transform.
	 * @param tg The new scale transform.
	 */
	public void setScaleTransform(TransformGroup tg) {
		scaleTransform = tg;
	}
	
	/**
	 * Updates the translation transform.
	 * @param tg The translation transform.
	 */
	public void setTranslateTransform(TransformGroup tg) {
		translateTransform = tg;
	}
}
