/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.view3d;


import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.vecmath.*;
import javax.media.j3d.*;

import icc3dpackage.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.*;
import icc3dpackage.modules.generic.Module;
import icc3dpackage.visualizations.appearance.*;


/**
 * Creates and maintains the frame and behaviors for a 3D view.
 */
public class View3D extends JFrame implements MouseBehaviorCallback {
	
	/** The main application object. */
	private ICC3D icc3d;

	/** The Java3D universe. */
	private SimpleUniverse universe;
	
	/** The manager that controls the modules in this 3d view. */
	private ModuleManager moduleManager;
	
	/** The 3D scene canvas. */
	private Canvas3D canvas;
	
	/** The screen buffer. */
	private Canvas3D screenBuffer;
	
	/** An ImageComponent2D used for off screen rendering. */
	private ImageComponent2D offscreenBuffer;
	
	/** The root of the scene graph. */
	private BranchGroup objRoot;
	
	/** The bounding sphere surrounding the active Java3D universe. */
	private BoundingSphere bounds;
	
	/** The canvas background. */
	private Background background;
	
	/** The scale size. */
	private double scaleSize;
	
	/** The object scale transform. */
	private TransformGroup objScale;
	
	/** The object translation transform. */
	private TransformGroup objTrans;
	
	/** The color space manager. */
	private ColorSpaceManager colorSpaceManager;
	
	/** The constraint manager. */
	private ConstraintsManager constraintsManager;
	
	/** The mapping manager. */
	private MappingManager mappingManager;
	
	/** The interaction manager. */
	private InteractionManager interactionManager;
	
	/** The panel used to control View3D parameters. */
	private JPanel adjustmentPanel;
	
	/** The tabbed pane used for GUI purposes. */
	private JTabbedPane tabs;
	
	/** The status bar. */
	private Status3D status3D;
	
	/** The image gamuts. */
	private ArrayList imageGamuts;
	
	/** The device gamuts. */
	private ArrayList deviceGamuts;
	
	/**
	 * Constructs a View3D object.
	 * @param icc3D The ICC3D application object.
	 * @param id The id that identifies the View3D.
	 */
	public View3D(ICC3D icc3D, int id) {
		this.icc3d = icc3D;
		imageGamuts = new ArrayList();
		deviceGamuts = new ArrayList();
		moduleManager = new ModuleManager(this);
		initOK();
		scaleSize = -100.0;
		setTitle("3D View " + id);
		setIconImage(ICC3D.image);
		Container c = getContentPane();
		c.setLayout(new BorderLayout(5, 5));

		setUpDefault();
		colorSpaceManager.signalMovement();
		initFrame();
		this.setLocation(300, 310);
		this.addWindowListener(new WindowAdapter() {
			public void windowActivated(WindowEvent e) {
				icc3d.getControlArea().setControl(adjustmentPanel);
				icc3d.getControlArea().setTitle("Control panel - "+ View3D.this.getTitle());
				if (!icc3d.getControlArea().isVisible()) {
					icc3d.getControlArea().setVisible(true);
				}
			}
		});
	}

	/**
	 * Initializes the frame. Sets the closing function and size.
	 */
	private void initFrame() {
		this.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				kill();
			}
		});
		this.setSize(getPreferredSize());
		this.setVisible(true);	
	}
	
	/**
	 * Returns the preferred dimension.
	 * @return Dimension the preferred dimension.
	 */
	public Dimension getPreferredSize() {
		return new Dimension(800, 600);	
	}

	/**
	 * Creates the control panel.
	 */
	public void createAdjustmentPanel() {
		adjustmentPanel = new JPanel();
		adjustmentPanel.setLayout(new BorderLayout());
		tabs = new JTabbedPane(JTabbedPane.LEFT);
		
		Module mods[] = moduleManager.getModules();

		for (int i = 0; i < mods.length; i++) {
			addModule(mods[i]);
		}

		JPanel updatePanel = new JPanel(new FlowLayout());

		updatePanel.setBorder(BorderFactory.createEmptyBorder());
		JButton btnUpdate = new JButton("Update", IconManager.getIcon("update"));

		btnUpdate.setHorizontalAlignment(JButton.LEFT);
		btnUpdate.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				update();
			}
		});
		updatePanel.add(btnUpdate);
		JPanel modulePanel = new JPanel(new FlowLayout());

		modulePanel.add(moduleManager.getParameterPanel());
		JPanel topPanel = new JPanel(new BorderLayout());

		topPanel.add(updatePanel, BorderLayout.NORTH);
		topPanel.add(modulePanel, BorderLayout.SOUTH);
		adjustmentPanel.add(topPanel, BorderLayout.NORTH);
		adjustmentPanel.add(tabs, BorderLayout.CENTER);
		
	}

	/**
	 * Updates the 3D view.
	 */
	public void update3D() {
		imageGamuts.clear();
		deviceGamuts.clear();
		universe.getCanvas().stopRenderer();
		universe.getCanvas().getGraphics().drawString("Performing update. Please stand by...",
				20, 30);
		universe.getCanvas().getGraphics().drawRect(10, 10, 275, 30);
		universe.getLocale().removeBranchGraph(objRoot);
		objRoot = null;
		SystemManager.majorClean();
		universe.addBranchGraph(createBG());
		colorSpaceManager.signalMovement();
		universe.getCanvas().startRenderer();
		if (mappingManager != null)
			mappingManager.notifyGamutsChanged();
	}

	/**
	 * Sets the default configuration for the 3d view, like the size 
	 * and distance to the object.
	 */
	private void setUpDefault() {
		bounds = new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 5.0);
		GraphicsConfiguration config = null;// SimpleUniverse.getPreferredConfiguration();

		if (config == null) {
			GraphicsDevice[] gs = GraphicsEnvironment.getLocalGraphicsEnvironment().getScreenDevices();
			Vector v = new Vector();

			for (int j = 0; j < gs.length; j++) { 
				GraphicsDevice gd = gs[j];
				GraphicsConfiguration[] gc = gd.getConfigurations();

				for (int i = 0; i < gc.length; i++) {
					v.add(gc[i]);
				}
			}
			GraphicsConfiguration gc[] = new GraphicsConfiguration[v.size()];

			for (int i = 0; i < v.size(); i++) {
				gc[i] = (GraphicsConfiguration) v.get(i);
			}
			System.out.println("GCs: " + gc.length);
			GraphicsConfigTemplate3D template = new GraphicsConfigTemplate3D();

			template.setDoubleBuffer(GraphicsConfigTemplate.PREFERRED);
			config = null;// template.getBestConfiguration(gc);
			if (config == null) {
				for (int k = 32; k >= 0; k--) {
					template.setDepthSize(k);
					config = template.getBestConfiguration(gc);
					if (config != null) {
						StandardAppearance.setDepth(k);
						System.out.println("Depth buffer: " + k + " bits");
						break;
					}
				}
			}
		}
		if (config == null) {
			SystemManager.displayError(
				"Error",
				"Unable to aquire a valid graphics configuration.\nMatrox users "
					+ "should try to enable 32 bit z-buffer.");
		}
		canvas = new Canvas3D(config);
		universe = new SimpleUniverse(canvas);
		screenBuffer = new Canvas3D(config, true);
		screenBuffer.getScreen3D().setPhysicalScreenHeight(canvas.getScreen3D().getPhysicalScreenHeight());
		screenBuffer.getScreen3D().setPhysicalScreenWidth(canvas.getScreen3D().getPhysicalScreenWidth());
		screenBuffer.getScreen3D().setSize(canvas.getScreen3D().getSize());
		universe.getViewer().getView().addCanvas3D(screenBuffer);
		universe.addBranchGraph(createBG());
		universe.getViewer().getView().setDepthBufferFreezeTransparent(false);
		universe.getViewer().getView().setFrontClipDistance(0.01);
		universe.getViewer().getView().setBackClipDistance(20.0);
		universe.getViewingPlatform().setNominalViewingTransform();
		
		// -----TRANSPARENCY
		// u.getViewer().getView().setDepthBufferFreezeTransparent(true);
		// u.getViewer().getView().setTransparencySortingPolicy(
		// View.TRANSPARENCY_SORT_GEOMETRY);
		// -----TRANSPARENCY
		
		Transform3D t3d = new Transform3D();
		// canvas.setViewTransform();
		
		JPanel view3DPanel = new JPanel();

		view3DPanel.setLayout(new BorderLayout());
		view3DPanel.setBorder(BorderFactory.createBevelBorder(BevelBorder.LOWERED));
		view3DPanel.add(canvas, BorderLayout.CENTER);
		view3DPanel.setVisible(true);
		Container c = getContentPane();
		c.add(view3DPanel, BorderLayout.CENTER);
		status3D = new Status3D();
		status3D.setRight("Memory usage: " + SystemManager.getUsedMemory() + " "
				+ (java.text.DecimalFormat.getPercentInstance().format((double) SystemManager.getUsedMemoryInBytes()
						/ (double) SystemManager.getTotalMemory())));
		c.add(status3D, BorderLayout.SOUTH);
		createAdjustmentPanel();
	}
	
	/**
	 * Creates the root branchgroup for the 3D object and returns it.
	 * @return BranchGroup the root branchgroup.
	 */
	public BranchGroup createBG() {
		Module mods[] = moduleManager.getModules();
		ProgressBar bar = new ProgressBar(1 + mods.length, "Updating 3D view",
				"Creating 3D scene structure");

		if (background == null) {
			background = new Background(new Color3f(0.5f, 0.5f, 0.5f));
			background.setCapability(Background.ALLOW_COLOR_WRITE);
			background.setCapability(Background.ALLOW_COLOR_READ);
			background.setApplicationBounds(bounds);
		} else {
			Color3f color3f = new Color3f();

			background.getColor(color3f);
			background = new Background(color3f);
			background.setCapability(Background.ALLOW_COLOR_WRITE);
			background.setCapability(Background.ALLOW_COLOR_READ);
			background.setApplicationBounds(bounds);
		}
		objRoot = new BranchGroup();
		objRoot.setCapability(BranchGroup.ALLOW_DETACH);
		objRoot.addChild(background);
		
		// DirectionalLight dl = new DirectionalLight(true, new Color3f(0.0f, 0.0f, 1.0f), new Vector3f(-1.0f, 0.0f, 0.0f));
		// dl.setInfluencingBounds(new BoundingSphere(new Point3d(0.0, 0.0, 0.0), 1000.0));
		// objRoot.addChild(dl);
		
		
		Transform3D colorScaleTransform = colorSpaceManager.getColorSpace3D().getTransform();
		TransformGroup colorScale = new TransformGroup(colorScaleTransform);
		Transform3D scale = new Transform3D();

		if (objScale != null) {
			objScale.getTransform(scale);
		}
		objScale = new TransformGroup(scale);
		objScale.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		objScale.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		objScale.setCapability(TransformGroup.ENABLE_PICK_REPORTING);
				
		Transform3D trans = new Transform3D();

		trans.setTranslation(new Vector3f(-0.5f, -0.5f, -0.5f));
		objTrans = new TransformGroup(trans);
		objTrans.setCapability(TransformGroup.ENABLE_PICK_REPORTING);
		objScale.addChild(objTrans);
		objTrans.addChild(colorScale);
		
		BranchGroup pickableBranch = new BranchGroup();

		for (int i = 0; i < mods.length; i++) {
			bar.inc("Processing module " + (i + 1));
			BranchGroup tmp[] = mods[i].getSceneBranch();

			if (tmp != null && tmp.length > 0) {
				if (tmp[0] != null) {
					pickableBranch.addChild(tmp[0]);
				}
				for (int c = 1; c < tmp.length; c++) {
					if (tmp[c] != null) {
						colorScale.addChild(tmp[c]);
					}
				}
			}
		}
		TriangleArray ta = new TriangleArray(3,
				GeometryArray.COORDINATES | GeometryArray.COLOR_3);
		float colors[] = new float[]{0, 0, 0, 0, 1, 0, 0, 0, 1};
		ta.setCoordinates(0, colors);
		ta.setColors(0, colorSpaceManager.getColorSpace3D().toSRGB(colors));
		/*Shape3D shape3D = new Shape3D(ta,
				StandardAppearance.getFilledTriangleApperance());
		colorScale.addChild(shape3D);
		*/
		//if (!initialized) {
			Color3f color3f = new Color3f();
			background.getColor(color3f);
			
			Appearance app = StandardAppearance.getLineAppearance(color3f,
					true, false);
			app.getRenderingAttributes().setAlphaTestFunction(RenderingAttributes.NEVER);
			Shape3D shape3D2 = new Shape3D(ta, app);
			colorScale.addChild(shape3D2);
		//}
		
				
		colorScale.addChild(pickableBranch);
		MouseMoveBehavior mouseBehavior = new MouseMoveBehavior(this, canvas,
				pickableBranch);

		mouseBehavior.setTransformGroup(objScale);
		mouseBehavior.setupCallback(this);
		mouseBehavior.setSchedulingBounds(bounds);
		mouseBehavior.setViewTransform(universe.getViewingPlatform().getViewPlatformTransform());
		mouseBehavior.setTranslateTransform(objTrans);
		mouseBehavior.setScaleTransform(objScale);
		objScale.addChild(mouseBehavior);
		objRoot.addChild(objScale);
		objTrans.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		objScale.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		objRoot.compile();
		bar.inc("Done");
		bar.setVisible(false);
		bar.dispose();
		return objRoot;
	}
	
	/**
	 * Returns the current canvas 3d.
	 * @return Canvas3D the current canvas 3d object.
	 */
	public Canvas3D getCanvas() {
		return canvas;
	}
	
	/**
	 * Cleans up the when you exit.
	 */
	public void kill() {
		moduleManager.deInitialize();
		icc3d.getSessionManager().remove3D_View(this);
		icc3d.getControlArea().invalidateControl(adjustmentPanel);
		icc3d.getControlArea().hide();
		universe.removeAllLocales();
		this.adjustmentPanel = null;
		this.background = null;
		this.bounds = null;
		this.canvas = null;
		this.colorSpaceManager = null;
		this.constraintsManager = null;
		this.icc3d = null;
		this.interactionManager = null;
		this.moduleManager = null;
		this.objRoot = null;
		this.objScale = null;
		this.objTrans = null;
		this.offscreenBuffer = null;
		this.screenBuffer = null;
		this.status3D = null;
		this.tabs = null;
		this.universe = null;
		this.dispose();
		this.mappingManager = null;
		SystemManager.majorClean();
	}
	
	/**
	 * Enables zoom in the orthographic view mode
	 * @param type The type of mouse behavior event (zoom, translate, rotate)
	 * @param transform The current transformgroup to which you want to make 
	 * the zoom (changes the viewers position) .
	 */
	public void transformChanged(int type, Transform3D transform) {
		colorSpaceManager.signalMovement();
		if (type == MouseBehaviorCallback.ZOOM) {
			scaleSize = 0.15;
			Vector3d v3d = new Vector3d();

			transform.get(v3d);
			universe.getViewer().getView().setScreenScale(scaleSize
					/ (0.1 + v3d.length()));
		}
	}
	
	/**
	 * Animates the object from one position to an pre-defined end position. This is done by
	 * linear interpolation between to matrixes, and gives quite nifty results :-) .
	 * @param destination the pre-defined end position.
	 */
	public void animate(Transform3D destination, String type) {
		
		Transform3D source = new Transform3D();
		int steps = 50;

		objScale.getTransform(source);
		Matrix4d m4dSource = new Matrix4d();
		Matrix3d m3dSource = new Matrix3d();
		Vector3d v3dSourceTranslation = new Vector3d();
		Vector3d v3dSourceScale = new Vector3d();

		source.get(m4dSource);
		source.get(m3dSource, v3dSourceTranslation);
		source.getScale(v3dSourceScale);
		
		Matrix4d m4dDestination = new Matrix4d();
		Matrix3d m3dDestination = new Matrix3d();
		Vector3d v3dDestinationTranslation = new Vector3d();
		Vector3d v3dDestinationScale = new Vector3d();

		destination.get(m4dDestination);
		destination.get(m3dDestination, v3dDestinationTranslation);
		destination.getScale(v3dDestinationScale);
		
		Matrix4d m4dChange = new Matrix4d();
		Matrix3d m3dChange = new Matrix3d();
		Vector3d v3dChangeTranslation = new Vector3d();
		Vector3d v3dChangeScale = new Vector3d();

		m4dChange.sub(m4dDestination, m4dSource);
		m3dChange.sub(m3dDestination, m3dSource);
		v3dChangeTranslation.sub(v3dDestinationTranslation, v3dSourceTranslation);
		v3dChangeScale.sub(v3dDestinationScale, v3dSourceScale);
		m4dChange.mul(1.0 / (double) steps);
		Matrix4d current = new Matrix4d(m4dSource);
		
		for (int i = 0; i < steps; i++) {
			current.add(m4dChange);
			source.set(current);
			objScale.setTransform(source);
			if (!type.equals("Perspective")) {
				transformChanged(MouseBehaviorCallback.ZOOM, source);
			}
			colorSpaceManager.getColorSpace3D().signalMovement();
			try {
				Thread.currentThread().sleep(10);
			} catch (InterruptedException ie) {}
		}							
	}
	
	/**
	 * Returns the current interaction manager.
	 * @return The current interaction manager.
	 */
	public InteractionManager getInteractionManager() {
		return interactionManager;
	}
	
	/**
	 * Sets the current interaction manager.
	 * @param intMan The interaction manager.
	 */
	public void setInteractionManager(InteractionManager intMan) {
		interactionManager = intMan;
	}
	
	/**
	 * Returns the current constraints manager.
	 * @return The current constraints manager.
	 */
	public ConstraintsManager getConstraintsManager() {
		return constraintsManager;
	}
	
	/**
	 * Sets the current constraints manager.
	 * @param conMan The constraints manager.
	 */
	public void setConstraintsManager(ConstraintsManager conMan) {
		constraintsManager = conMan;
	}
	
	/**
	 * Returns the current color space manager.
	 * @return the current color space manager.
	 */
	public ColorSpaceManager getColorSpaceManager() {
		return colorSpaceManager;
	}
	
	/**
	 * Sets the current color space manager.
	 * @param colMan The color space manager.
	 */
	public void setColorSpaceManager(ColorSpaceManager colMan) {
		colorSpaceManager = colMan;
	}
	
	/**
	 * Sets the current gamut mapping manager.
	 * @param man The gamut mapping manager.
	 */
	public void setMappingManager(MappingManager man) {
		mappingManager = man;
	}
	
	/**
	 * Returns the object scale transform group.
	 * @return The object scale transform group.
	 */
	public TransformGroup getObjectScale() {
		return objScale;
	}
	
	/**
	 * Sets the left status text.
	 * @param s The status text.
	 */
	public void setStatusLeft(String s) {
		status3D.setLeft(s);
	}
	
	/**
	 * Sets the center status text.
	 * @param s The status text.
	 */
	public void setStatusCenter(String s) {
		status3D.setMiddle(s);
	}
	
	/**
	 * Sets the right status text.
	 * @param s The status text.
	 */
	public void setStatusRight(String s) {
		status3D.setRight(s);
	}
	
	/**
	 * Sets the status color.
	 * @param c The color.
	 */
	public void setStatusColor(Color c) {
		status3D.setColor(c);
	}
	
	/**
	 * Changes to a different view type.
	 * @param type The new view type.
	 */
	public void changeView(String type) {
		if (type.equals("Perspective")) {
			universe.getViewer().getView().setProjectionPolicy(View.PERSPECTIVE_PROJECTION);
			universe.getViewer().getView().setScreenScalePolicy(View.SCALE_SCREEN_SIZE);
		} else {
			universe.getViewer().getView().setProjectionPolicy(View.PARALLEL_PROJECTION);
			universe.getViewer().getView().setScreenScalePolicy(View.SCALE_EXPLICIT);
			Transform3D trans = new Transform3D();

			objScale.getTransform(trans);
			transformChanged(MouseBehaviorCallback.ZOOM, trans);
		}
	}
	
	/**
	 * Returns the current background.
	 * @return The current background.
	 */
	public Background background() {
		return background;
	}
	
	/**
	 * Adds a module to this view.
	 * @param mod The module that should be added to this view.
	 */
	public void addModule(Module mod) {
		if (!moduleManager.exists(mod)) {
			moduleManager.add(mod);
		} else {
			tabs.addTab("", mod.getIcon(), mod.getParameterPanel(), mod.getName());
		}
	}
	
	/**
	 * Check module initialization.
	 */
	private void initOK() {
		if (colorSpaceManager == null) {
			SystemManager.displayError(
				"Error",
				"Failed to initialize color space module");
			System.exit(1);
		}
		if (constraintsManager == null) {
			SystemManager.displayError(
				"Error",
				"Failed to initialize constraint module");
			System.exit(1);
		}
		if (interactionManager == null) {
			SystemManager.displayError(
				"Error",
				"Failed to initialize interaction module");
			System.exit(1);
		}
		if (mappingManager == null) {
			SystemManager.displayError(
				"Warning",
				"Failed to initialize gamut mapping module");
		}
	}
	
	/**
	 * Returns a canvas for off screen rendering.
	 * @return A canvas for off screen rendering.
	 */
	public Canvas3D getOffScreenCanvas() {
		return this.screenBuffer;
	}
	
	/**
	 * Returns the main application object.
	 * @return The main application object.
	 */
	public ICC3D getICC3D() {
		return icc3d;
	}
	
	/**
	 * Returns the module manager that controls this view.
	 * @return The module manager that controls this view.
	 */
	public ModuleManager getModuleManager() {
		return moduleManager;
	}
	
	/**
	 * Update view.
	 */
	public void update() {
		StandardAppearance.updateDepthOffsetFromConfig();
		boolean ok = true;
		Module mods[] = moduleManager.getModules();

		for (int i = 0; i < mods.length; i++) {
			if (mods[i].readyForUpdate() == false) {
				ok = false;
			}
		}
		if (ok) {
			try {
				update3D();
			} catch (Throwable t) {
				t.printStackTrace();
				if (t instanceof OutOfMemoryError) {
					SystemManager.displayError(
						"Error",
						"Out of memory");
				} else {
					SystemManager.displayError(
						"Error",
						"An error occurred while updating the 3D view");
				}
			}
			SystemManager.majorClean();
			status3D.setRight("Memory usage: " + SystemManager.getUsedMemory() + " "
					+ (java.text.DecimalFormat.getPercentInstance().format((double) SystemManager.getUsedMemoryInBytes()
							/ (double) SystemManager.getTotalMemory())));
		}
	}
	
	/**
	 * Returns a list of all available image gamuts.
	 * @return A list of all available image gamuts.
	 */
	public ArrayList getImageGamuts() {
		return imageGamuts;
	}
	
	/**
	 * Returns a list of all available device gamuts.
	 * @return A list of all available device gamuts.
	 */
	public ArrayList getDeviceGamuts() {
		return deviceGamuts;
	}
	
	/**
	 * Returns the tabbed pane.
	 * @return The tabbed pane.
	 */
	public JTabbedPane getTabs() {
		return tabs;
	}
}
