/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import javax.media.j3d.*;
import javax.vecmath.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.text.NumberFormat;
import java.io.*;
import java.text.DecimalFormat;
import javax.swing.event.*;
import java.util.BitSet;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.quantization.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.*;


/**
 * AlphaShapesVisualization3D is a visualization which displays
 * the input points as a solid object in 3D, specified by a
 * certain alpha value.
 */
public class AlphaShapesVisualization3D extends Visualization3D {
	
	/** The alpha shape parameter panel */
	private JPanel parameterPanel;
	
	/** The branch group where the alpha shape child is added */
	private BranchGroup parent;
	
	/** The branch group that contains the alpha shape */
	private BranchGroup child;
	
	/** The alpha values that determine a unique alpha shape */
	private double alphaValues[];
	
	/** The alpha value slider */
	private JSlider alphaSlider;
	
	/** The current tetrahedron traversal sequence number */
	private int currentSequence = Integer.MIN_VALUE;
	
	/** The number of 4D facets (4D facet = 3D tetrahedron) */
	private int tetraCount;
	
	/** The facets that make up the 4D convex hull */
	private Facet tetras[];
	
	/** The tetrahedron volumes */
	private double volumes[];
	
	/** The scaling factors of the color space */
	private float scaleUsed[];
	
	/** The color space where the alpha shape is computed */
	private ColorSpace3D colorSpace3D[];
	
	/** The quantization method used to limit the number of vertices */
	private Quantization quantization;
	
	/** The volume of the convex hull */
	private double convVolume;
	
	/** The volume of the alpha shape */
	private double alphaVolume;
	
	/** The volume of the convex hull */
	private JLabel convVolumeLabel;
	
	/** The volume of the alpha shape */
	private JLabel alphaVolumeLabel;
	
	/** The number of vertices */
	private JLabel tetraVertexCountLabel;
		
	/**
	 * Default constructor for an AlphaShapesVisualization3D.
	 */
	public AlphaShapesVisualization3D() {}

	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0= outside).
	 */
	public BitSet getVoxel(int detail) {
		BitSet voxel = new BitSet(detail * detail * detail);

		if (tetras == null) {
			return voxel;
		}
		double suggestedAlpha = (alphaValues[alphaSlider.getValue()]
						+ alphaValues[alphaSlider.getValue() + 1])
				/ 2.0;
		double adjust = 0.5 / detail;
		double dDetail = detail;
		Vertex checkVertex = new Vertex(0, 0, 0);
		Facet startFacet = tetras[0];
		int lastX = 0;

		for (int x = 0; x < detail; x++) {
			if (x % 2 == 0) {
				for (int y = 0; y < detail; y++) {
					if (lastX == 0) {
						lastX = 1;
						for (int z = 0; z < detail; z++) {
							checkVertex.set((adjust + x / dDetail) * scaleUsed[0],
									(adjust + y / dDetail) * scaleUsed[1],
									(adjust + z / dDetail) * scaleUsed[2]);
							startFacet = startFacet.getClosestTetrahedron(checkVertex, Facet.seq++);
							if (startFacet.contains(checkVertex)
									&& startFacet.radiusOfSmallestCircumsphere < suggestedAlpha) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					} else {
						lastX = 0;
						for (int z = detail - 1; z >= 0; z--) {
							checkVertex.set((adjust + x / dDetail) * scaleUsed[0],
									(adjust + y / dDetail) * scaleUsed[1],
									(adjust + z / dDetail) * scaleUsed[2]);
							startFacet = startFacet.getClosestTetrahedron(checkVertex, Facet.seq++);
							if (startFacet.contains(checkVertex)
									&& startFacet.radiusOfSmallestCircumsphere < suggestedAlpha) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					}
				}
			} else {
				for (int y = detail - 1; y >= 0; y--) {
					if (lastX == 0) {
						lastX = 1;
						for (int z = 0; z < detail; z++) {
							checkVertex.set((adjust + x / dDetail) * scaleUsed[0],
									(adjust + y / dDetail) * scaleUsed[1],
									(adjust + z / dDetail) * scaleUsed[2]);
							startFacet = startFacet.getClosestTetrahedron(checkVertex, Facet.seq++);
							if (startFacet.contains(checkVertex)
									&& startFacet.radiusOfSmallestCircumsphere < suggestedAlpha) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					} else {
						lastX = 0;
						for (int z = detail - 1; z >= 0; z--) {
							checkVertex.set((adjust + x / dDetail) * scaleUsed[0],
									(adjust + y / dDetail) * scaleUsed[1],
									(adjust + z / dDetail) * scaleUsed[2]);
							startFacet = startFacet.getClosestTetrahedron(checkVertex, Facet.seq++);
							if (startFacet.contains(checkVertex)
									&& startFacet.radiusOfSmallestCircumsphere < suggestedAlpha) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					}
				}
			}
		}
		
		return voxel;
	}

	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */	
	public boolean checkParams() {
		return true;
	}

	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces. This particular implementation computes alpha shapes by
	 * first applying a 4D QuickHull algorithm, and then classifying all triangles
	 * by computing the numbers that decide whether a given triangle should be displayed.
	 * The user then has the possibility to specify an alpha value, and the resulting
	 * alpha shape is displayed in the 3D view.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		scaleUsed = colorSpace3D[1].getScale();
		// String detailString = detailBox.getSelectedItem().toString();
		QuantizedObject qo = new QuantizedObject();

		qo.numPoints = numPoints;
		qo.points = pointCoord[1];
		qo.pointToPixelCount = new int[numPoints];
		qo = quantization.quantize(qo);
		int oIndex = qo.numPoints;

		if (qo.points != pointCoord[1]) {
			numPoints = qo.numPoints;
			pointCoord = new float[2][];
			pointCoord[1] = qo.points;
			pointColor = colorSpace3D[1].toSRGB(pointCoord[1]);
		}
		SystemManager.majorClean();
		this.colorSpace3D = colorSpace3D;
		int choice = JOptionPane.YES_OPTION;
			
		if (oIndex < 8) {
			return new BranchGroup[] {null, null};
		}
		if (oIndex > 10000) {
			choice = SystemManager.confirmYesNo(
				"Continue with operation?",
				"Are you sure you want to compute the Delaunay triangulation of " + oIndex
					+ " points ?");
		}
		if (choice == JOptionPane.CANCEL_OPTION || choice == JOptionPane.NO_OPTION) {
			return new BranchGroup[] {new BranchGroup(), new BranchGroup()};
		}
		QuickHull4D quickHull = new QuickHull4D();
		Vertex vertices[] = new Vertex[numPoints];

		for (int i = 0; i < numPoints; i++) {
			vertices[i] = new Vertex(pointCoord[1][i * 3] * scaleUsed[0]
					+ Math.random() / 1000000.0,
					pointCoord[1][i * 3 + 1] * scaleUsed[1] + Math.random() / 1000000.0,
					pointCoord[1][i * 3 + 2] * scaleUsed[2] + Math.random() / 1000000.0);
			vertices[i].addDimension();
		}

		tetraVertexCountLabel.setText(Integer.toString(numPoints));
			
		long start = System.currentTimeMillis();
		OptimizedVector list = quickHull.build4D(vertices);
		long stop = System.currentTimeMillis();

		tetraCount = 0;
		int accumulatedTetras = 0;

		for (int i = 0; i < list.elementCount; i++) {
			Facet tetra = (Facet) list.elementAt(i);

			if (!tetra.isDiscarded() && tetra.normalW < 0.0) {
				tetraCount++;
				tetra.calculateRadiusOfSmallestCircumsphere();
			}
		}
			
		int triangleCount = 0;
		double tempAlphaValues[] = new double[tetraCount * 8];
		int tempAlphaUsed = 0;

		tempAlphaValues[tempAlphaUsed++] = Double.MAX_VALUE;
		tempAlphaValues[tempAlphaUsed++] = Double.MIN_VALUE;
		tetras = new Facet[tetraCount];
		for (int i = 0; i < list.elementCount; i++) {
			Facet tetra = (Facet) list.elementAt(i);

			if (!tetra.isDiscarded() && tetra.normalW < 0.0) {
				tetras[accumulatedTetras++] = tetra;
				for (int j = 0; j < 4; j++) {
					if (tetra.neighbor[j].isDiscarded() || tetra.neighbor[j].normalW >= 0.0) {
						tetra.neighbor[j] = null;
						tetra.triangle[j] = new AlphaTriangle(tetra.radiusOfSmallestCircumsphere,
								Double.MAX_VALUE);
						tempAlphaValues[tempAlphaUsed++] = tetra.radiusOfSmallestCircumsphere;
					} else if (tetra.triangle[j] == null) {
						Facet temp = tetra.neighbor[j];

						tetra.triangle[j] = new AlphaTriangle(Math.min(tetra.radiusOfSmallestCircumsphere,
								temp.radiusOfSmallestCircumsphere),
								Math.max(tetra.radiusOfSmallestCircumsphere,
								temp.radiusOfSmallestCircumsphere));
						tempAlphaValues[tempAlphaUsed++] = tetra.triangle[j].minU;
						tempAlphaValues[tempAlphaUsed++] = tetra.triangle[j].maxU;
						triangleCount++;
						for (int k = 0; k < 4; k++) {
							if (temp.neighbor[k] == tetra) {
								temp.triangle[k] = tetra.triangle[j];
								break;
							}
						}
					}
				}
			}
		}
		convVolume = 0;
		// float scale[] = colorSpace3D[1].getScale();
		double m[][] = new double[4][4];

		volumes = new double[tetraCount];
		for (int i = 0; i < tetraCount; i++) {
			volumes[i] = determinantVolume(tetras[i].corner, m);
			convVolume += volumes[i];
		}
		convVolumeLabel.setText(Integer.toString((int) Math.round(convVolume)));
		list = null;
		SystemManager.majorClean();
		alphaValues = new double[tempAlphaUsed];
		System.arraycopy(tempAlphaValues, 0, alphaValues, 0, tempAlphaUsed);
		java.util.Arrays.sort(alphaValues);
		int len = alphaValues.length;
		int forwardStep = 1;
		int kop;

		for (kop = 0; kop < len - forwardStep; kop++) {
			if (alphaValues[kop] == alphaValues[kop + forwardStep]) {
				forwardStep++;
				kop--;
			} else {
				alphaValues[kop + 1] = alphaValues[kop + forwardStep];
			}
		}
		len = kop;
		if (alphaSlider.getValue() > len - 2 || alphaSlider.getValue() < 11) {
			alphaSlider.setMaximum(len - 2);
			double largestAlphaDiff = 0;
			int suggestedValue = len - 2;
			
			for (int i = (3 * (len - 2) / 4); i < len - 4; i++) {
				double tmp = (alphaValues[i + 1] - alphaValues[i])
						/ (alphaValues[i] * Math.sqrt(alphaValues[i]));

				if (tmp > largestAlphaDiff) {
					largestAlphaDiff = tmp;
					suggestedValue = i;
				}
			}
			alphaSlider.setValue(suggestedValue);
		}
		alphaSlider.setMaximum(len - 2);
		parent = new BranchGroup();
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		rebuildChild();
		parent.addChild(child);
		BranchGroup branchGroup[] = new BranchGroup[2];

		branchGroup[0] = new BranchGroup();
		branchGroup[1] = parent;
		return branchGroup;
	}

	/**
	 * Returns the volume of the tetrahedron
	 */
	private double determinantVolume(Vertex[] v, double m[][]) {
		for (int i = 0; i < 4; i++) {
			m[0][i] = v[i].coordinates[0];
			m[1][i] = v[i].coordinates[1];
			m[2][i] = v[i].coordinates[2];
			m[3][i] = 1;
		}
		return Math.abs((
										m[0][0]
								* (m[1][1] * (m[2][2] * m[3][3] - m[2][3] * m[3][2])
										- m[1][2] * (m[2][1] * m[3][3] - m[2][3] * m[3][1])
										+ m[1][3] * (m[2][1] * m[3][2] - m[2][2] * m[3][1]))
						- m[0][1]
								* (m[1][0] * (m[2][2] * m[3][3] - m[2][3] * m[3][2])
										- m[1][2] * (m[2][0] * m[3][3] - m[2][3] * m[3][0])
										+ m[1][3] * (m[2][0] * m[3][2] - m[2][2] * m[3][0]))
						+ m[0][2]
								* (m[1][0] * (m[2][1] * m[3][3] - m[2][3] * m[3][1])
										- m[1][1] * (m[2][0] * m[3][3] - m[2][3] * m[3][0])
										+ m[1][3] * (m[2][0] * m[3][1] - m[2][1] * m[3][0]))
						- m[0][3]
								* (m[1][0] * (m[2][1] * m[3][2] - m[2][2] * m[3][1])
										- m[1][1] * (m[2][0] * m[3][2] - m[2][2] * m[3][0])
										+ m[1][2] * (m[2][0] * m[3][1] - m[2][1] * m[3][0]))
						)
				/ 6.0);
	}

	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization. (Alpha Shapes)
	 */	
	public String getName() {
		return "Alpha Shapes";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 5;
	}

	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel for this visualization.
	 */
	private void createParameterPanel() {
		quantization = new Quantization(Quantization.SIMPLIFY_QUANTIZATION);
		
		JPanel innerPanel = new JPanel(new BorderLayout());

		innerPanel.add(new JLabel("Alpha: "), BorderLayout.NORTH);
		alphaSlider = new JSlider(JSlider.HORIZONTAL, 0, 10, 0);
		innerPanel.add(alphaSlider, BorderLayout.SOUTH);

		JPanel volumesPanel = new JPanel();

		volumesPanel.setLayout(new BoxLayout(volumesPanel, BoxLayout.Y_AXIS));
		convVolumeLabel = new JLabel(Integer.toString((int) Math.round(convVolume)));
		JPanel tmpPanel = new JPanel(new BorderLayout());

		tmpPanel.add(new JLabel("Conv. hull vol.: "), BorderLayout.WEST);
		tmpPanel.add(convVolumeLabel, BorderLayout.EAST);
		volumesPanel.add(tmpPanel);
		tmpPanel = new JPanel(new BorderLayout());
		alphaVolumeLabel = new JLabel(Integer.toString((int) Math.round(alphaVolume)));
		tmpPanel.add(new JLabel("Alpha shape vol.: "), BorderLayout.WEST);
		tmpPanel.add(alphaVolumeLabel, BorderLayout.EAST);
		volumesPanel.add(tmpPanel);
		tmpPanel = new JPanel(new BorderLayout());
		tetraVertexCountLabel = new JLabel("0");
		tmpPanel.add(new JLabel("No. of vertices:"), BorderLayout.WEST);
		tmpPanel.add(tetraVertexCountLabel, BorderLayout.EAST);
		volumesPanel.add(tmpPanel);

		alphaSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				if (parent != null && !alphaSlider.getValueIsAdjusting()) {
					child.detach();
					rebuildChild();
					parent.addChild(child);
				}
			}
		});
		parameterPanel = new JPanel(new BorderLayout());
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Alpha Shapes"));
		
		parameterPanel.add(quantization.getParameterPanel(), BorderLayout.NORTH);
		JPanel anotherPanel = new JPanel(new BorderLayout());
		JPanel yetAnotherPanel = new JPanel(new BorderLayout());

		yetAnotherPanel.add(volumesPanel, BorderLayout.NORTH);
		yetAnotherPanel.add(innerPanel, BorderLayout.SOUTH);
		anotherPanel.add(yetAnotherPanel, BorderLayout.NORTH);
		anotherPanel.add(new JPanel(), BorderLayout.CENTER);
		parameterPanel.add(anotherPanel, BorderLayout.CENTER);
	}
	
	/**
	 * Rebuilds the child branch group with a new alpha value.
	 */
	private void rebuildChild() {
		System.out.println("Rebuilding");
		child = new BranchGroup();
		child.setCapability(BranchGroup.ALLOW_DETACH);
		double suggestedAlpha = (alphaValues[alphaSlider.getValue()]
						+ alphaValues[alphaSlider.getValue() + 1])
				/ 2.0;

		currentSequence++;
		
		int numDrawnTriangles = 0;

		alphaVolume = 0;
		for (int i = 0; i < tetraCount; i++) {
			Facet tetra = tetras[i];

			if (tetra.radiusOfSmallestCircumsphere <= suggestedAlpha) {
				alphaVolume += volumes[i];
			}
			for (int j = 0; j < 4; j++) {
				if (tetra.triangle[j].lastSeen != currentSequence) {
					tetra.triangle[j].lastSeen = currentSequence;
					if (tetra.triangle[j].minU <= suggestedAlpha 
							&& tetra.triangle[j].maxU >= suggestedAlpha) {
						numDrawnTriangles++;
					}
					// if (tetra.triangle[j].minU <= suggestedAlpha)
					// numDrawnTriangles++;
				}
			}
		}
		alphaVolumeLabel.setText(Integer.toString((int) Math.round(alphaVolume)));

		if (numDrawnTriangles < 1) {
			return;
		}

		OptimizedVector faceList = new OptimizedVector(numDrawnTriangles);
		
		Facet tmp;

		currentSequence++;
		
		for (int i = 0; i < tetraCount; i++) {
			Facet tetra = tetras[i];

			for (int j = 0; j < 4; j++) {
				if (tetra.triangle[j].lastSeen != currentSequence) {
					tetra.triangle[j].lastSeen = currentSequence;
					if (tetra.triangle[j].minU <= suggestedAlpha 
							&& tetra.triangle[j].maxU >= suggestedAlpha) {
						faceList.addElement(new Facet(tetra.corner[j], tetra.corner[(j + 1) % 4],
								tetra.corner[(j + 2) % 4]));
					}
					// if (tetra.triangle[j].minU<=suggestedAlpha)
					// faceList.addElement(new Facet(tetra.corner[j], tetra.corner[(j+1)%4], 
					// tetra.corner[(j+2)%4]));
				}
			}
		}

		int count = faceList.size();
		OptimizedVector list = faceList;
		TriangleArray ta = new TriangleArray(count * 3,
				GeometryArray.COORDINATES | GeometryArray.COLOR_3);
		int index = 0;
		Facet face;

		for (int i = 0; i < list.size(); i++) {
			face = (Facet) list.elementAt(i);
			if (!face.isDiscarded()) {
				ta.setCoordinate(index,
						new double[] {
					face.getVertices()[0].coordinates[0] / scaleUsed[0],
					face.getVertices()[0].coordinates[1] / scaleUsed[1],
					face.getVertices()[0].coordinates[2] / scaleUsed[2]});
				ta.setColor(index++,
						colorSpace3D[1].toSRGB(new float[] {(float) face.getVertices()[0].coordinates[0]
							/ scaleUsed[0],
					(float) face.getVertices()[0].coordinates[1] / scaleUsed[1],
					(float) face.getVertices()[0].coordinates[2] / scaleUsed[2]}));
				ta.setCoordinate(index,
						new double[] {
					face.getVertices()[1].coordinates[0] / scaleUsed[0],
					face.getVertices()[1].coordinates[1] / scaleUsed[1],
					face.getVertices()[1].coordinates[2] / scaleUsed[2]});
				ta.setColor(index++,
						colorSpace3D[1].toSRGB(new float[] {(float) face.getVertices()[1].coordinates[0]
							/ scaleUsed[0],
					(float) face.getVertices()[1].coordinates[1] / scaleUsed[1],
					(float) face.getVertices()[1].coordinates[2] / scaleUsed[2]}));
				ta.setCoordinate(index,
						new double[] {
					face.getVertices()[2].coordinates[0] / scaleUsed[0],
					face.getVertices()[2].coordinates[1] / scaleUsed[1],
					face.getVertices()[2].coordinates[2] / scaleUsed[2]});
				ta.setColor(index++,
						colorSpace3D[1].toSRGB(new float[] {(float) face.getVertices()[2].coordinates[0]
							/ scaleUsed[0],
					(float) face.getVertices()[2].coordinates[1] / scaleUsed[1],
					(float) face.getVertices()[2].coordinates[2] / scaleUsed[2]}));
			}
		}
			
		Shape3D shape3D = new Shape3D(ta,
				StandardAppearance.getFilledTriangleApperance());
		Shape3D shape3D2 = new Shape3D(ta,
				StandardAppearance.getLineAppearance(new Color3f(0.0f, 0.0f, 0.0f), true,
				false));

		child.addChild(shape3D);
		child.addChild(shape3D2);

	}	
			
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {
		tetras = null;
		parent = null;
		child = null;
		alphaValues = null;
	}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {
		this.alphaSlider = null;
		this.alphaValues = null;
		this.alphaVolumeLabel = null;
		this.child = null;
		this.colorSpace3D = null;
		this.convVolumeLabel = null;
		this.quantization = null;
		this.parameterPanel = null;
		this.parent = null;
		this.tetras = null;
		this.tetraVertexCountLabel = null;
	}
}
