/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import java.awt.*;
import java.awt.event.*;
import java.text.NumberFormat;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;
import javax.media.j3d.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.vecmath.*;

import icc3dpackage.modules.*;
import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.utils.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.*;


/**
 * ConvexHullVisualization3D is a visualization which displays
 * the input points as a solid object in 3D, calculated by
 * finding the convex hull of the input points in 3D.
 */
public class ConvexHullVisualization3D extends Visualization3D {
	
	/** The color space that was used when the convex hull was constructed. */
	private ColorSpace3D colorSpaceUsed;
	
	/** The gamma parameter used when applying the radial expansion. */
	private float gamma = 0.2f;
	
	/** The radius that the radial expansion converges towards. */
	private float maxRadius = 1.0f;
	
	/** The center of the convex hull. */
	private float middle[] = new float[] {0.5f, 0.5f, 0.5f};
	
	/** The convex hull parameter panel. */
	private JPanel parameterPanel;
	
	/** The scaling factors of the color space. */
	private float scaleUsed[];
	
	/** Determines whether the extents along the color axis are the same. */
	private JCheckBox normalizeAspect;
	
	/** Determines whether the radial expansion (blowing algorithm) is used. */
	private JCheckBox radialExpansion;
	
	/** The gamma parameter slider. */
	private JSlider gammaSlider;
	
	/** The gamma parameter label. */
	private JLabel gammaDisplay;
	
	/** The volume label. */
	private JLabel convexVolume;
	
	/** Determines the appearance of the convex hull. */
	private TriangleAppearance triangleAppearance;
	
	/** The tetrahedron structure. */
	private Tetrahedron tetra[];
	
	/** Determines whether display options are included in the parameter panel */
	private boolean noDisplayParameters;
	
	/**
	 * Default constructor for a ConvexHullVisualization3D.
	 */
	public ConvexHullVisualization3D() {}
	
	/**
	 * Constructs a ConvexHullVisualization3D without display paramaters.
	 */
	public ConvexHullVisualization3D(boolean noDispParam) {
		noDisplayParameters = noDispParam;
	}
	
	/**
	 * @return The tetrahedron structure, or null if no such structure has been
	 * built.
	 */
	public Tetrahedron[] getTetras() {
		return tetra;
	}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0=outside).
	 */
	public BitSet getVoxel(int detail) {
		return ConvexHullVisualization3D.getVoxel(detail, tetra);
	}
	
	/**
	 * Computes and returns a voxel representation of a gamut.
	 *
	 * @param detail The detail level.
	 * @param tetraDescriptor The tetrahedron structure that defines a gamut.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0=outside).
	 */
	public static BitSet getVoxel(int detail, Tetrahedron tetraDescriptor[]) {
		BitSet voxel = new BitSet(detail * detail * detail);

		if (tetraDescriptor == null) {
			return voxel;
		}

		float adjust = 0.5f / detail;
		float dDetail = detail;
		float posX, posY, posZ;
		Tetrahedron startFacet = tetraDescriptor[0];
		int lastX = 0;

		for (int x = 0; x < detail; x++) {
			if (x % 2 == 0) {
				for (int y = 0; y < detail; y++) {
					if (lastX == 0) {
						lastX = 1;
						for (int z = 0; z < detail; z++) {
							posX = (adjust + x / dDetail);
							posY = (adjust + y / dDetail);
							posZ = (adjust + z / dDetail);
							startFacet = startFacet.findSegmentTetrahedron2(null, posX, posY, posZ);
							if (startFacet.plane[3].classifyPoint2(posX, posY, posZ)
									!= Plane.PLANE_OUTSIDE) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					} else {
						lastX = 0;
						for (int z = detail - 1; z >= 0; z--) {
							posX = (adjust + x / dDetail);
							posY = (adjust + y / dDetail);
							posZ = (adjust + z / dDetail);
							startFacet = startFacet.findSegmentTetrahedron2(null, posX, posY, posZ);
							if (startFacet.plane[3].classifyPoint2(posX, posY, posZ)
									!= Plane.PLANE_OUTSIDE) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					}
				}
			} else {
				for (int y = detail - 1; y >= 0; y--) {
					if (lastX == 0) {
						lastX = 1;
						for (int z = 0; z < detail; z++) {
							posX = (adjust + x / dDetail);
							posY = (adjust + y / dDetail);
							posZ = (adjust + z / dDetail);
							startFacet = startFacet.findSegmentTetrahedron2(null, posX, posY, posZ);
							if (startFacet.plane[3].classifyPoint2(posX, posY, posZ)
									!= Plane.PLANE_OUTSIDE) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					} else {
						lastX = 0;
						for (int z = detail - 1; z >= 0; z--) {
							posX = (adjust + x / dDetail);
							posY = (adjust + y / dDetail);
							posZ = (adjust + z / dDetail);
							startFacet = startFacet.findSegmentTetrahedron2(null, posX, posY, posZ);
							if (startFacet.plane[3].classifyPoint2(posX, posY, posZ)
									!= Plane.PLANE_OUTSIDE) {
								voxel.set(x * detail * detail + y * detail + z);
							}
						}
					}
				}
			}
		}
		
		return voxel;
	}
	
	/**
	 * Calculates the radius of a color point.
	 * @param p Ana array of color coordinates.
	 * @param i The number of the point.
	 * @return The radius of the color point.
	 */
	private float radius(float p[], int i) {
		return (float) Math.pow((p[i * 3] - middle[0]) * (p[i * 3] - middle[0])
				+ (p[i * 3 + 1] - middle[1]) * (p[i * 3 + 1] - middle[1])
				+ (p[i * 3 + 2] - middle[2]) * (p[i * 3 + 2] - middle[2]),
				0.5);
	}
	
	/**
	 * Calculates a transformed radius of a color point by applying a gamma
	 * function to the original radius.
	 * @param r The original radius.
	 * @return The transfomed radius.
	 */
	private float gammaRadius(float r) {
		return maxRadius * (float) Math.pow(r / maxRadius, gamma);
	}
	
	/**
	 * Calculates the original radius of a color point, using the previously
	 * applied gamma parameter and the transformed radius.
	 * @param r The transformed radius.
	 * @return The original radius.
	 */
	private float normalRadius(float r) {
		return maxRadius * (float) Math.pow((r / maxRadius), 1 / gamma);
	}

	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */	
	public boolean checkParams() {
		return true;
	}

	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces. This particular implementation computes a convex hull by
	 * first applying one of several fast 3D convex hull algorithms.
	 * The resulting triangles are then included in the branch group.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		if (numPoints < 5) {
			return null;
		}

		/*
		 vertices[numPoints+0] = new Vertex(middle[0], middle[1]+0.01f*scaleUsed[1], middle[2]);
		 vertices[numPoints+1] = new Vertex(middle[0]+0.01f*scaleUsed[0], middle[1]-0.01f*scaleUsed[1], middle[2]-0.01f*scaleUsed[2]);
		 vertices[numPoints+2] = new Vertex(middle[0]-0.01f*scaleUsed[0], middle[1]-0.01f*scaleUsed[1], middle[2]-0.01f*scaleUsed[2]);
		 vertices[numPoints+3] = new Vertex(middle[0], middle[1]-0.01f*scaleUsed[1], middle[2]+0.01f*scaleUsed[2]);
		 */
		tetra = null;
		QuickHull quickHull = new QuickHull();
		Vertex vertices[] = new Vertex[numPoints/* +4*/];
		int mode = 0;// useColorSpace.getSelectedIndex();
		float posInSpace[] = null;

		if (mode == 2) {
			posInSpace = ColorSpaceConverter.convert(pointCoord[0], colorSpace3D[0],
					ColorSpaceManager.getColorSpace("XYZ"));
			scaleUsed = new float[] {1.0f, 1.0f, 1.0f};
			
		} else if (mode == 1) {
			scaleUsed = colorSpace3D[0].getScale();
			posInSpace = new float[pointCoord[0].length];
			System.arraycopy(pointCoord[0], 0, posInSpace, 0, posInSpace.length);
		} else {
			scaleUsed = colorSpace3D[1].getScale();
			posInSpace = new float[pointCoord[1].length];
			System.arraycopy(pointCoord[1], 0, posInSpace, 0, posInSpace.length);
		}
		int cCount = posInSpace.length / 3;

		for (int i = 0; i < cCount; i++) {
			posInSpace[i * 3] *= scaleUsed[0];
			posInSpace[i * 3 + 1] *= scaleUsed[1];
			posInSpace[i * 3 + 2] *= scaleUsed[2];
		}
		
		float m1 = 0, m2 = 0, m3 = 0;
		
		float lowX = posInSpace[0];
		float lowY = posInSpace[1];
		float lowZ = posInSpace[2];
		float highX = posInSpace[0];
		float highY = posInSpace[1];
		float highZ = posInSpace[2];
		
		for (int i = 0; i < numPoints; i++) {
			if (posInSpace[i * 3] < lowX) {
				lowX = posInSpace[i * 3];
			} else if (posInSpace[i * 3] > highX) {
				highX = posInSpace[i * 3];
			}
			if (posInSpace[i * 3 + 1] < lowY) {
				lowY = posInSpace[i * 3 + 1];
			} else if (posInSpace[i * 3 + 1] > highY) {
				highY = posInSpace[i * 3 + 1];
			}
			if (posInSpace[i * 3 + 2] < lowZ) {
				lowZ = posInSpace[i * 3 + 2];
			} else if (posInSpace[i * 3 + 2] > highZ) {
				highZ = posInSpace[i * 3 + 2];
			}
			m1 += posInSpace[i * 3];
			m2 += posInSpace[i * 3 + 1];
			m3 += posInSpace[i * 3 + 2];
		}

		middle[0] = 0.5f * scaleUsed[0];
		middle[1] = 0.5f * scaleUsed[1];
		middle[2] = 0.5f * scaleUsed[2];
		
		if (lowX >= highX) {
			lowX = highX - 0.01f;
			highX += 0.01f;
		}
		if (lowY >= highY) {
			lowY = highY - 0.01f;
			highY += 0.01f;
		}
		if (lowZ >= highZ) {
			lowZ = highZ - 0.01f;
			highZ += 0.01f;
		}
		float xDiff = highX - lowX;
		float yDiff = highY - lowY;
		float zDiff = highZ - lowZ;
		
		boolean useNormalization = normalizeAspect.isSelected();

		if (useNormalization) {
			float oldPos[] = posInSpace;

			posInSpace = new float[oldPos.length];
			for (int i = 0; i < numPoints; i++) {
				posInSpace[i * 3] = middle[0] + (oldPos[i * 3] - middle[0]) / xDiff;
				posInSpace[i * 3 + 1] = middle[1] + (oldPos[i * 3 + 1] - middle[1]) / yDiff;
				posInSpace[i * 3 + 2] = middle[2] + (oldPos[i * 3 + 2] - middle[2]) / zDiff;
			}
		}
		
		boolean useBlow = radialExpansion.isSelected();

		if (useBlow) {
			float oldPos[] = posInSpace;

			posInSpace = new float[oldPos.length];
			System.arraycopy(oldPos, 0, posInSpace, 0, oldPos.length);
			for (int i = 0; i < numPoints; i++) {
				float r = radius(oldPos, i);

				if (r > 0) {
					float newR = gammaRadius(r);

					posInSpace[i * 3] = middle[0] + (newR / r) * (oldPos[i * 3] - middle[0]);
					posInSpace[i * 3 + 1] = middle[1]
							+ (newR / r) * (oldPos[i * 3 + 1] - middle[1]);
					posInSpace[i * 3 + 2] = middle[2]
							+ (newR / r) * (oldPos[i * 3 + 2] - middle[2]);
				}
			}	
		}
		
		for (int i = 0; i < numPoints; i++) {
			vertices[i] = new Vertex(posInSpace[i * 3] + Math.random() / 10000.0f,
					posInSpace[i * 3 + 1] + Math.random() / 10000.0f,
					posInSpace[i * 3 + 2] + Math.random() / 10000.0f);
		}

		/*
		 vertices[numPoints+0] = new Vertex(middle[0], 			middle[1]+0.1f, 		middle[2]);
		 vertices[numPoints+1] = new Vertex(middle[0]+0.1f, 	middle[1]-0.1f, 		middle[2]-0.1f);
		 vertices[numPoints+2] = new Vertex(middle[0]-0.1f, 	middle[1]-0.1f, 		middle[2]-0.1f);
		 vertices[numPoints+3] = new Vertex(middle[0], 			middle[1]-0.1f, 		middle[2]+0.1f);
		 */
		
		OptimizedVector list = quickHull.build3D(vertices);
		int count = 0;

		for (int i = 0; i < list.size(); i++) {
			if (list.elementAt(i) instanceof Facet) {
				if (!((Facet) list.elementAt(i)).isDiscarded()) {
					count++;
				}
			}
		}
		
		BranchGroup branchGroup[] = new BranchGroup[2];

		branchGroup[0] = new BranchGroup();
		branchGroup[1] = new BranchGroup();

		int index = 0;
		Facet face;
		Facet triFacet[] = new Facet[count];
		int triCount = 0;
		float tmpFloat[] = new float[3];
		ColorSpace3D tmpSpace = null;

		if (mode == 2) {
			tmpSpace = ColorSpaceManager.getColorSpace("XYZ");
		} else if (mode == 1) {
			tmpSpace = colorSpace3D[0];
		} else {
			tmpSpace = colorSpace3D[1];
		}
		colorSpaceUsed = tmpSpace;
		// output = colorSpace3D[1];
		
		ArrayList usedVertices = new ArrayList();
		ArrayList usedFaces = new ArrayList();
		Vertex facePoint[];

		for (int i = 0; i < list.size(); i++) {
			face = (Facet) list.elementAt(i);
			if (!face.isDiscarded()) {
				usedFaces.add(face);
				face.data = usedFaces.size();
				facePoint = face.getVertices();
				for (int j = 0; j < 3; j++) {
					if (facePoint[j].data == 0) {
						usedVertices.add(facePoint[j]);
						facePoint[j].data = usedVertices.size();
						// scale coordinates
						
						float tmpFloat2[] = null;

						tmpFloat[0] = (float) facePoint[j].coordinates[0];
						tmpFloat[1] = (float) facePoint[j].coordinates[1];
						tmpFloat[2] = (float) facePoint[j].coordinates[2];
						if (useBlow) {
							float r = radius(tmpFloat, 0);

							if (r > 0) {
								float newR = normalRadius(r);

								tmpFloat[0] = middle[0] + (newR / r) * (tmpFloat[0] - middle[0]);
								tmpFloat[1] = middle[1] + (newR / r) * (tmpFloat[1] - middle[1]);
								tmpFloat[2] = middle[2] + (newR / r) * (tmpFloat[2] - middle[2]);
							}
						}
						
						if (useNormalization) {
							tmpFloat[0] = middle[0] + xDiff * (tmpFloat[0] - middle[0]);
							tmpFloat[1] = middle[1] + yDiff * (tmpFloat[1] - middle[1]);
							tmpFloat[2] = middle[2] + zDiff * (tmpFloat[2] - middle[2]);
						}
						tmpFloat[0] /= scaleUsed[0];
						tmpFloat[1] /= scaleUsed[1];
						tmpFloat[2] /= scaleUsed[2];
						tmpFloat2 = ColorSpaceConverter.convert(tmpFloat, tmpSpace,
								colorSpace3D[1]);
						facePoint[j].coordinates[0] = tmpFloat2[0];
						facePoint[j].coordinates[1] = tmpFloat2[1];
						facePoint[j].coordinates[2] = tmpFloat2[2];
					}
				}
			}
		}
		
		for (int i = 0; i < 3; i++) {
			middle[i] /= scaleUsed[i];
		}
		tetra = new Tetrahedron[usedFaces.size()];
		return ImprovedSegmentMaxima.buildStructure(usedVertices, usedFaces, middle,
				colorSpace3D, tetra, scaleUsed, convexVolume, numPoints, pointCoord,
				numEqual, imageManager, view3D, triangleAppearance);
	}
	
	/**
	 * Computes the scaled volume of the tetrahedron.
	 * @param t The tetrahedron
	 * @param scale The color space scale.
	 * @param m A temporary matrix.
	 * @return The scaled tetrahedron volume.
	 */
	public static float volume(Tetrahedron t, float scale[], float m[][]) {
		m[0][0] = t.cornerpos0_0 * scale[0];
		m[1][0] = t.cornerpos0_1 * scale[1];
		m[2][0] = t.cornerpos0_2 * scale[2];
		m[3][0] = 1;
		m[0][1] = t.cornerpos1_0 * scale[0];
		m[1][1] = t.cornerpos1_1 * scale[1];
		m[2][1] = t.cornerpos1_2 * scale[2];
		m[3][1] = 1;
		m[0][2] = t.cornerpos2_0 * scale[0];
		m[1][2] = t.cornerpos2_1 * scale[1];
		m[2][2] = t.cornerpos2_2 * scale[2];
		m[3][2] = 1;
		m[0][3] = t.cornerpos3_0 * scale[0];
		m[1][3] = t.cornerpos3_1 * scale[1];
		m[2][3] = t.cornerpos3_2 * scale[2];
		m[3][3] = 1;

		/*
		 for (int i = 0; i < 4; i++) {
		 m[0][i] = v[i].coordinates[0];
		 m[1][i] = v[i].coordinates[1];
		 m[2][i] = v[i].coordinates[2];
		 m[3][i] = 1;
		 }
		 */
		return (float) Math.abs((
										m[0][0]
								* (m[1][1] * (m[2][2] * m[3][3] - m[2][3] * m[3][2])
										- m[1][2] * (m[2][1] * m[3][3] - m[2][3] * m[3][1])
										+ m[1][3] * (m[2][1] * m[3][2] - m[2][2] * m[3][1]))
						- m[0][1]
								* (m[1][0] * (m[2][2] * m[3][3] - m[2][3] * m[3][2])
										- m[1][2] * (m[2][0] * m[3][3] - m[2][3] * m[3][0])
										+ m[1][3] * (m[2][0] * m[3][2] - m[2][2] * m[3][0]))
						+ m[0][2]
								* (m[1][0] * (m[2][1] * m[3][3] - m[2][3] * m[3][1])
										- m[1][1] * (m[2][0] * m[3][3] - m[2][3] * m[3][0])
										+ m[1][3] * (m[2][0] * m[3][1] - m[2][1] * m[3][0]))
						- m[0][3]
								* (m[1][0] * (m[2][1] * m[3][2] - m[2][2] * m[3][1])
										- m[1][1] * (m[2][0] * m[3][2] - m[2][2] * m[3][0])
										+ m[1][2] * (m[2][0] * m[3][1] - m[2][1] * m[3][0]))
						)
				/ 6.0);
	}

	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization. (Convex Hull)
	 */	
	public String getName() {
		return "Convex Hull";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 4;
	}

	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel for this visualization.
	 */
	private void createParameterPanel() {
		normalizeAspect = new JCheckBox();
		radialExpansion = new JCheckBox();
		JPanel twoPanel = new JPanel(new GridLayout(0, 2));

		twoPanel.add(new JLabel("Normalize aspect"));
		twoPanel.add(normalizeAspect);
		gammaDisplay = new JLabel("0.2");
		gammaDisplay.setHorizontalAlignment(JLabel.RIGHT);
		gammaSlider = new JSlider(JSlider.HORIZONTAL, 1, 1000, 200);
		gammaSlider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent ce) {
				gamma = gammaSlider.getValue() / 1000.0f;
				String gText = Float.toString(gamma);

				while (gText.length() < 5) {
					gText = gText + "0";
				}
				if (gText.length() > 5) {
					gText = gText.substring(0, 5);
				}
				gammaDisplay.setText(gText);
			}
		});
		gammaSlider.getChangeListeners()[0].stateChanged(null);
		parameterPanel = new JPanel();
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Convex Hull"));
		parameterPanel.setLayout(new BorderLayout());
		// chooseDisplayMode = new JComboBox(new String[]{"Wireframe", "Solid", "Solid+Wireframe"});
		// useColorSpace = new JComboBox(new String[]{"Output", "Input", "CIEXYZ"});
		JPanel pPanel = new JPanel(new GridLayout(0, 1));

		pPanel.setLayout(new BoxLayout(pPanel, BoxLayout.Y_AXIS));
		pPanel.add(twoPanel);
		twoPanel = new JPanel(new GridLayout(0, 2));
		twoPanel.add(new JLabel("Radial expansion"));
		twoPanel.add(radialExpansion);
		pPanel.add(twoPanel);
		pPanel.add(gammaSlider);
		JPanel gammaPanel = new JPanel(new GridLayout(1, 2));

		gammaPanel.add(new JLabel("Gamma: "));
		gammaPanel.add(gammaDisplay);
		pPanel.add(gammaPanel);
		triangleAppearance = new TriangleAppearance();
		// pPanel.add(useColorSpace);
		// JPanel dispPanel = new JPanel(new GridLayout(1, 2));
		// dispPanel.add(new JLabel("Display"));
		// dispPanel.add(chooseDisplayMode);
		// pPanel.add(dispPanel);
		convexVolume = new JLabel("0");
		convexVolume.setHorizontalAlignment(JLabel.RIGHT);
		JPanel volPanel = new JPanel(new GridLayout(1, 2));

		volPanel.add(new JLabel("Volume"));
		volPanel.add(convexVolume);
		pPanel.add(volPanel);
		if (!noDisplayParameters) {
			pPanel.add(triangleAppearance);
		}
		parameterPanel.add(pPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
		
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {
		this.triangleAppearance = null;
		this.colorSpaceUsed = null;
		this.gammaDisplay = null;
		this.gammaSlider = null;
		this.middle = null;
		this.parameterPanel = null;
		this.radialExpansion = null;
		this.normalizeAspect = null;
	}
}
