/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.media.j3d.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.*;


/**
 * CrossSectionVisualization3D is a visualization which displays a
 * cross-section of a gamut.
 */
public class CrossSectionVisualization3D extends Visualization3D {
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The convex hull parameter panel. */
	private Component convexPanel;
	
	/** The convex hull visualization. */
	private ConvexHullVisualization3D convex;
	
	/** The parent branch group. */
	private BranchGroup parent;
	
	/** The child branch group. (Removed and changed as necessary). */
	private BranchGroup child;
	
	/** The slider value. */
	private JLabel value;
	
	/** The slider that determines the cross-section. */
	private JSlider slider;
	
	/** The type of cross-section. */
	private JComboBox type;
	
	/** The current color space. */
	private ColorSpace3D curSpace;
	
	/** The visualization mode (triangles, outline ...). */
	private JComboBox visualizationMode;
	
	/** Use a fixed solid color. */
	private JCheckBox fixedSolidColor;
	
	/** The fixed solid color. */
	private JButton solidColor;
	
	/** Use a fixed outline color. */
	private JCheckBox fixedOutlineColor;
	
	/** The fixed outline color. */
	private JButton outlineColor;
	
	/**
	 * Default constructor for a CrossSectionVisualization3D.
	 */
	public CrossSectionVisualization3D() {
		convex = new ConvexHullVisualization3D(true);
		convexPanel = convex.getParameterPanel();
	}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0= outside).
	 */
	public BitSet getVoxel(int detail) {
		if (convex == null) {
			BitSet voxel = new BitSet(detail * detail * detail);

			return voxel;
		} else {
			return convex.getVoxel(detail);
		}
	}
	
	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces. This particular implementation returns an empty scene graph.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */ 
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		convex.createSceneObject(numPoints, pointCoord, numEqual, pointColor,
				pixelIndex, colorSpace3D, imageManager, view3D);
		curSpace = colorSpace3D[1];
		parent = new BranchGroup();
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		parent.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		child = getChild();
		child.setCapability(BranchGroup.ALLOW_DETACH);
		parent.addChild(child);
		BranchGroup ret[] = new BranchGroup[] {new BranchGroup(), parent};

		return ret;
	}
	
	/**
	 * Traverses the gamut boundary, and constructs a triangle fan of the resulting points.
	 * @return The BranchGroup that visualizes the cross-section of the gamut.
	 */
	private BranchGroup getChild() {
		if (convex == null || curSpace == null) {
			return new BranchGroup();
		}
		Tetrahedron tetra[] = convex.getTetras();

		if (tetra == null) {
			return new BranchGroup();
		}
		
		Vector3f p1 = new Vector3f();
		Vector3f p2 = new Vector3f();
		Vector3f p3 = new Vector3f();
		Vector3f center = new Vector3f();
		
		if (type.getSelectedIndex() == 0) {
			float param = (float) slider.getValue() / (float) slider.getMaximum();

			p1.x = 0.0f;
			p1.y = param;
			p1.z = 0.0f;
			p2.x = 0.0f;
			p2.y = param;
			p2.z = 1.0f;
			p3.x = 1.0f;
			p3.y = param;
			p3.z = 0.0f;
			center.x = 0.5f;
			center.y = param;
			center.z = 0.5f;
		} else {
			double param = 2 * Math.PI * (double) slider.getValue()
					/ (double) slider.getMaximum();

			p1.x = 0.5f;
			p1.y = 0.4f;
			p1.z = 0.5f;
			p2.x = 0.5f;
			p2.y = 0.6f;
			p2.z = 0.5f;
			p3.x = 0.5f + (float) Math.cos(param);
			p3.y = 0.5f;
			p3.z = 0.5f + (float) Math.sin(param);
			center.x = 0.5f;
			center.y = 0.5f;
			center.z = 0.5f;
		}
		
		ArrayList al = new ArrayList();
		Plane p = new Plane(p1, p2, p3);
		Tetrahedron startSearch = tetra[0];

		if (Tetrahedron.findBoundary(p, tetra, startSearch, al)) {
			if (al.size() > 2) {
				BranchGroup newChild = new BranchGroup();
				if ((visualizationMode.getSelectedIndex() == 0) || (visualizationMode.getSelectedIndex() == 2)) {
					float colors[] = new float[(al.size() + 2) * 3];
					colors[0] = center.x;
					colors[1] = center.y;
					colors[2] = center.z;
					for (int i = 0; i <= al.size(); i++) {
						Vector3f v3f = (Vector3f) al.get(i % al.size());
						colors[(i + 1) * 3] = v3f.x;
						colors[(i + 1) * 3 + 1] = v3f.y;
						colors[(i + 1) * 3 + 2] = v3f.z;
					}
					byte sRGB[] = curSpace.toSRGB(colors);
					TriangleFanArray tfa = new TriangleFanArray(al.size() + 2,
							GeometryArray.COLOR_3 | GeometryArray.COORDINATES,
							new int[] {al.size() + 2});
	
					tfa.setCoordinates(0, colors);
					tfa.setColors(0, sRGB);
					
					newChild.addChild(new Shape3D(tfa,
						StandardAppearance.getFilledTriangleApperance(!fixedSolidColor.isSelected(),
						new Color3f(solidColor.getBackground()),
						false,
						0.0f)));
				}
				if ((visualizationMode.getSelectedIndex() == 1) || (visualizationMode.getSelectedIndex() == 2)) {
					
					int[] stripVertexCounts = new int[]{al.size()+1};
					int vertexCount = al.size() + 1;
					int vertexFormat = GeometryArray.COLOR_3 | GeometryArray.COORDINATES;
					float colors[] = new float[(al.size() + 1) * 3];
					for (int i = 0; i <= al.size(); i++) {
						Vector3f v3f = (Vector3f) al.get(i % al.size());
						colors[i * 3] = v3f.x;
						colors[i * 3 + 1] = v3f.y;
						colors[i * 3 + 2] = v3f.z;
					}
					byte sRGB[] = curSpace.toSRGB(colors);
					
					LineStripArray lsa = new LineStripArray(
						vertexCount,
						vertexFormat,
						stripVertexCounts);
					lsa.setCoordinates(0, colors);
					lsa.setColors(0, sRGB);
					
					newChild.addChild(new Shape3D(lsa,
						StandardAppearance.getLineAppearance(new Color3f(outlineColor.getBackground()),
						true,
						!fixedOutlineColor.isSelected())));
				}
				return newChild;
			}
		}
		return new BranchGroup();
	}

	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization.
	 */	
	public String getName() {
		return "Cross-section";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 6;
	}
	
	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Rebuilds the child object, using the new visualization parameters.
	 */
	public void rebuildChild() {
		if (parent != null && child != null) {
			child.detach();
			child = getChild();
			child.setCapability(BranchGroup.ALLOW_DETACH);
			parent.addChild(child);
		}
	}
	
	/**
	 * Creates the parameter panel for this visualization.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		JPanel separatePanel = new JPanel(new BorderLayout());

		separatePanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Cross-section"));
		slider = new JSlider(JSlider.HORIZONTAL, 0, 10000, 5000);
		type = new JComboBox(new String[] {"L plane", "Hue plane"});
		type.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				rebuildChild();
			}
		});
		value = new JLabel(""
				+ (float) slider.getValue() / (float) slider.getMaximum());
		slider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				value.setText("" + (float) slider.getValue() / (float) slider.getMaximum());
				rebuildChild();
			}
		});
		JPanel typePanel = new JPanel(new GridLayout(0, 2));

		typePanel.add(new JLabel("Cross-section"));
		typePanel.add(type);
		
		visualizationMode = new JComboBox(new String[]{"Solid", "Outline", "Solid+Outline"});
		visualizationMode.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent e) {
				rebuildChild();
			}
		});
		
		typePanel.add(new JLabel("Mode"));
		typePanel.add(visualizationMode);
		
		fixedSolidColor = new JCheckBox();
		fixedSolidColor.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				rebuildChild();
			}
		});
		solidColor = new JButton();
		solidColor.setBackground(Color.red);
		solidColor.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Color retColor = JColorChooser.showDialog(ICC3D.getDialogParent(),
						"Choose a color for the solid body", solidColor.getBackground());

				if (retColor != null) {
					solidColor.setBackground(retColor);
					rebuildChild();
				}
			}
		});
		JPanel solidColorPanel = new JPanel(new BorderLayout());
		solidColorPanel.add(fixedSolidColor, BorderLayout.WEST);
		solidColorPanel.add(solidColor, BorderLayout.CENTER);
		typePanel.add(new JLabel("Solid color"));
		typePanel.add(solidColorPanel);
		
		fixedOutlineColor = new JCheckBox();
		fixedOutlineColor.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				rebuildChild();
			}
		});
		outlineColor = new JButton();
		outlineColor.setBackground(Color.black);
		outlineColor.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Color retColor = JColorChooser.showDialog(ICC3D.getDialogParent(),
						"Choose a color for the outline", outlineColor.getBackground());
				if (retColor != null) {
					outlineColor.setBackground(retColor);
					rebuildChild();
				}
			}
		});
		JPanel outlineColorPanel = new JPanel(new BorderLayout());
		outlineColorPanel.add(fixedOutlineColor, BorderLayout.WEST);
		outlineColorPanel.add(outlineColor, BorderLayout.CENTER);
		typePanel.add(new JLabel("Outline color"));
		typePanel.add(outlineColorPanel);
		
		JPanel innerPanel = new JPanel();
		innerPanel.setLayout(new BoxLayout(innerPanel, BoxLayout.Y_AXIS));
		innerPanel.add(slider);
		innerPanel.add(value);
		innerPanel.add(typePanel);
		
		JPanel adjustmentPanel = new JPanel(new BorderLayout());
		adjustmentPanel.add(innerPanel, BorderLayout.NORTH);
		adjustmentPanel.add(new JPanel(), BorderLayout.CENTER);
		separatePanel.add(adjustmentPanel, BorderLayout.NORTH);
		separatePanel.add(convexPanel, BorderLayout.CENTER);
		parameterPanel.add(separatePanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}

	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */	
	public boolean checkParams() {
		return true;
	}
	
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {
		this.parameterPanel = null;
		this.convexPanel = null;
		this.convex.deInitialize();
		this.convex = null;
		this.parent = null;
		this.child = null;
		this.value = null;
		this.slider = null;
		this.type = null;
		this.curSpace = null;
	}
}
