/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.lang.Math;
import java.awt.*;
import java.awt.image.*;
import java.awt.image.Raster;
import com.sun.media.jai.codecimpl.*;
import java.awt.*;
import javax.swing.JOptionPane;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;
import icc3dpackage.quantization.*;
import icc3dpackage.utils.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.*;


/**
 * DetailedVisualization3D is a visualization which displays
 * the input points as a simplified selection of points in 3D.
 * The points are computed by dividing the color space into cubes of
 * a certain size, and then deciding whether there exists point(s)
 * inside each cube.
 */
public class DetailedVisualization3D extends Visualization3D {
	
	/** The point size. */
	private float pointSize;
	
	/** The quantization of the input points. */
	private Quantization quantization;
	
	/** The number of points/objects. */
	private int nObjects;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The label that contains the point count. **/
	private JLabel edtObjectCount;
	
	/** The visualization mode (points or spheres). */
	private JComboBox cmbMode;
	
	/** The choice of quantization space. */
	private JComboBox detChoice;
	
	/** The point appearance. */
	private Appearance app;
	
	/** The available point size choices. */
	private JComboBox cmbPointSize;
	
	/**
	 * Default constructor for a DetailedVisualization3D.
	 */
	public DetailedVisualization3D() {}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0= outside).
	 */
	public BitSet getVoxel(int detail) {
		BitSet voxel = new BitSet(detail * detail * detail);
		return voxel;
	}
	
	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces. This particular implementation computes a quantized version of
	 * the input, and then displays the result either as simple points or spheres with
	 * different radi.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		BasicObject3D objects[];
		QuantizedObject qo = new QuantizedObject();

		// qo.detailRequested = Integer.parseInt(edtDetail.getText());
		qo.numPoints = numPoints;
		int detMode = detChoice.getSelectedIndex();

		qo.points = pointCoord[detMode];
		qo.pointToPixelCount = numEqual;
		QuantizedObject quantizedData = quantization.quantize(qo);

		nObjects = quantizedData.numPoints;
		
		Shape3D shape3D = null;
		BranchGroup branchGroup[] = null;

		if (numEqual == null) {
			shape3D = new Shape3D();
			branchGroup = new BranchGroup[2];
			branchGroup[0] = null;
			branchGroup[1] = new BranchGroup();
		} else {
			shape3D = new MovingObject();
			branchGroup = new BranchGroup[1];
			branchGroup[0] = new BranchGroup();
		}
		
		if (cmbMode.getSelectedIndex() == 0) {
			objects = new BasicObject3D[nObjects];
			PointArray pointArray = new PointArray(nObjects,
					GeometryArray.COORDINATES | GeometryArray.COLOR_3);

			if (numEqual != null) {
				((MovingObject) shape3D).setStart(0);
				((MovingObject) shape3D).setGeometryArrays(new GeometryArray[] {pointArray});
				((MovingObject) shape3D).setObjects(objects);
				((MovingObject) shape3D).setColors(pointCoord[1]);
				((MovingObject) shape3D).setImage2D(imageManager.getImage2D());
				((MovingObject) shape3D).setColorSpace(colorSpace3D[1]);
				view3D.getImageGamuts().add(shape3D);
			}
			
			for (int i = 0; i < nObjects; i++) {
				int q = i * 3;
				float pos[] = new float[] {quantizedData.points[q],
					quantizedData.points[q + 1], quantizedData.points[q + 2]};

				if (detMode == 0) {
					pos = ColorSpaceConverter.convert(pos, colorSpace3D[0], colorSpace3D[1]);
				}
				byte col[] = colorSpace3D[1].toSRGB(pos);

				objects[i] = new BasicObject3D(quantizedData.pointToColorCount[i], pos, col,
						i, pointArray);
				objects[i].setColorIndex(quantizedData.pointToColorIndex[i]);
			}
			StandardVisualization3D.setCapability(pointArray);
			shape3D.setCapability(Shape3D.ALLOW_GEOMETRY_READ);

			SystemManager.majorClean();
			Point3f vertexData[] = new Point3f[nObjects];
			Color3b colorData[] = new Color3b[nObjects];

			for (int i = 0; i < nObjects; i++) {
				vertexData[i] = new Point3f(objects[i].getPosition());
				colorData[i] = new Color3b(objects[i].getColor());
			}
			pointArray.setCoordinates(0, vertexData);
			pointArray.setColors(0, colorData);
			app = StandardAppearance.getChangeablePointAppearance(pointSize);
			shape3D.setAppearance(app);
			shape3D.setGeometry(pointArray);
			if (numEqual == null) {
				branchGroup[1].addChild(shape3D);
			} else {
				branchGroup[0].addChild(shape3D);
			}
		} else {
			Sphere sphere[] = new Sphere[nObjects];
			TransformGroup transformGroup[] = new TransformGroup[nObjects];
			Transform3D transform[] = new Transform3D[nObjects];
			Appearance appear[] = new Appearance[nObjects];
			
			TransformObject sphereObjs[] = new TransformObject[nObjects];
			
			for (int i = 0; i < nObjects; i++) {
				int q = i * 3;
				float pos[] = new float[] {quantizedData.points[q],
					quantizedData.points[q + 1], quantizedData.points[q + 2]};

				if (detMode == 0) {
					pos = ColorSpaceConverter.convert(pos, colorSpace3D[0], colorSpace3D[1]);
				}
				byte col[] = colorSpace3D[1].toSRGB(pos);
				float fCol[] = new float[3];

				for (int iii = 0; iii < 3; iii++) {
					fCol[iii] = (col[iii] & 0xff) / 255.0f;
				}
				Appearance appearance = new Appearance();

				appear[i] = appearance;
				appearance.setColoringAttributes(new ColoringAttributes(fCol[0], fCol[1],
						fCol[2], ColoringAttributes.SHADE_GOURAUD));
				Transform3D t3d = colorSpace3D[1].getTransform();

				t3d.invert();
				t3d.setTranslation(new Vector3f(pos));
				transform[i] = t3d;
				transformGroup[i] = new TransformGroup(t3d);
				sphere[i] = new Sphere(getRadius(quantizedData.pointToPixelCount[i]));
				sphere[i].setAppearance(appearance);
				if (numEqual == null) {
					transformGroup[i].addChild(sphere[i]);
				} else {
					MovingObject mo = new MovingObject();

					if (i == 0) {
						view3D.getImageGamuts().add(mo);
					}
					mo.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
					appearance.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_READ);
					appearance.getColoringAttributes().setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
					mo.setAppearance(appearance);
					for (Enumeration enum = sphere[i].getShape().getAllGeometries(); enum.hasMoreElements();) {
						try {
							GeometryArray garr = (GeometryArray) enum.nextElement();

							if (!garr.isLive() && !garr.isCompiled()) {
								StandardVisualization3D.setCapability(garr);
							}
							mo.addGeometry(garr);
						} catch (Exception e) {
							e.printStackTrace();
						}
						
					}
					mo.setStart(i);
					mo.setImage2D(imageManager.getImage2D());
					mo.setColorSpace(colorSpace3D[1]);
					mo.setObjects(sphereObjs);
					mo.setColors(pointCoord[1]);
					transformGroup[i].setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
					transformGroup[i].setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
					sphereObjs[i] = new TransformObject(quantizedData.pointToColorCount[i], pos,
							col, appearance, transformGroup[i]);
					sphereObjs[i].setColorIndex(quantizedData.pointToColorIndex[i]);
					transformGroup[i].addChild(mo);
				}
			}
			SystemManager.majorClean();		
			
			if (numEqual == null) {
				for (int iO = 0; iO < nObjects; iO++) {
					branchGroup[1].addChild(transformGroup[iO]);
				}
			} else {
				for (int iO = 0; iO < nObjects; iO++) {
					branchGroup[0].addChild(transformGroup[iO]);
				}
			}
		}
		updateParameterPanel();
		return branchGroup;
	}

	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization. (Quantized)
	 */	
	public String getName() {
		return "Quantized";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 1;
	}
	
	/**
	 * Creates the parameter panel for this visualization.
	 */	
	private void createParameterPanel() {
		quantization = new Quantization();

		edtObjectCount = new JLabel();
		edtObjectCount.setHorizontalAlignment(JLabel.RIGHT);
		edtObjectCount.setBorder(BorderFactory.createEtchedBorder());
		edtObjectCount.setText(Integer.toString(nObjects));
		
		/* edtDetail = new JTextField();
		 edtDetail.setHorizontalAlignment(JTextField.RIGHT);
		 edtDetail.setText("25");
		 */
		String dets[] = {"Input", "Output"};

		detChoice = new JComboBox(dets);
		detChoice.setSelectedIndex(1);
		
		String cmbAlternatives[] = new String[] {"Points", "Spheres"};

		cmbMode = new JComboBox(cmbAlternatives);
		pointSize = 10.0f;
		cmbPointSize = new JComboBox(new String[] {"1", "2", "3", "4", "5", "6", "8",
			"10", "12", "15", "20"});
		cmbPointSize.setSelectedIndex(7);
		cmbPointSize.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				pointSize = Integer.parseInt(cmbPointSize.getSelectedItem().toString());
				if (app != null) {
					app.setPointAttributes(new PointAttributes(pointSize, true));
				}
			}
		});		
		
		JPanel panel1 = new JPanel(new GridLayout(0, 2));

		panel1.add(new JLabel("Object count: "));
		panel1.add(edtObjectCount);
		panel1.add(new JLabel("Mode:"));
		panel1.add(cmbMode);
		panel1.add(new JLabel("Quantize: "));
		panel1.add(detChoice);
		panel1.add(new JLabel("Point size: "));
		panel1.add(cmbPointSize);
		
		JPanel panel2 = new JPanel(new BorderLayout());

		panel2.add(panel1, BorderLayout.NORTH);
		panel2.add(quantization.getParameterPanel(), BorderLayout.SOUTH);

		parameterPanel = new JPanel(new BorderLayout());
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Quantized"));
		parameterPanel.add(panel2, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}

	/**
	 * Updates the parameter panel so that it displays the correct number of objects.
	 */	
	public void updateParameterPanel() {
		// edtDetail.setText(Integer.toString(detail));	
		edtObjectCount.setText(Integer.toString(nObjects));
	}

	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}

	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */	
	public boolean checkParams() {
		return true;
	}
	
	/**
	 * Computes a radius for a sphere representing a certain number of pixels.
	 * @param numPixels The number of pixels.
	 * @return The radius of the sphere.
	 */
	private float getRadius(int numPixels) {
		double retValue;

		retValue = Math.pow(numPixels, 1.0 / 3.0) / 400.0;
		if (retValue > 0.15) {
			retValue = 0.15;
		}
		return (float) (retValue);
	}
	
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {
	}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {
		this.app = null;
		this.cmbMode = null;
		this.cmbPointSize = null;
		this.detChoice = null;
		this.quantization = null;
		this.edtObjectCount = null;
		this.parameterPanel = null;
	}
}
