/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import icc3dpackage.*;
import icc3dpackage.colorspaces.generic.ColorSpace3D;
import icc3dpackage.managers.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.pickable.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.Visualization3D;


/**
 * A segment maxima based visualization.
 */
public class ImprovedSegmentMaxima extends Visualization3D {
	
	/** The tetrahedrons. */
	private Tetrahedron tetra[];
	
	/** The volume label. */
	private JLabel convexVolume;
	
	/** The parameter panel. */
	private JPanel parameterPanel;
	
	/** The triangle appearance panel. */
	private TriangleAppearance triangleAppearance;
	
	/** The detail level. */
	private JComboBox chooseDetail;
	
	/** The smooth level. */
	private JComboBox chooseSmooth;
	
	/**
	 * Default constructor for ImprovedSegmentMaxima.
	 */
	public ImprovedSegmentMaxima() {}
	
	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		if (numPoints < 5) {
			return null;
		}
		// parameters
		/*
		 float newPointCoord[] = new float[numPoints*3];
		 for (int i=0; i<numPoints*3; i++) {
		 newPointCoord[i] = pointCoord[1][i] + 0.00001f*(float)(Math.random()-0.5);
		 }
		 pointCoord[1] = newPointCoord;
		 */
		
		int detail = 5 + chooseDetail.getSelectedIndex();
		
		// scale to color space
		float center[] = {0.5f, 0.5f, 0.5f};
		float scale[] = colorSpace3D[1].getScale();

		center[0] *= scale[0];
		center[1] *= scale[1];
		center[2] *= scale[2];
		
		float scaledCoord[] = new float[numPoints * 3];

		for (int i = 0; i < numPoints; i++) {
			scaledCoord[i * 3 + 0] = pointCoord[1][i * 3 + 0] * scale[0];
			scaledCoord[i * 3 + 1] = pointCoord[1][i * 3 + 1] * scale[1];
			scaledCoord[i * 3 + 2] = pointCoord[1][i * 3 + 2] * scale[2];
		}
		
		Vertex v[][] = new Vertex[detail][detail];
		float maxRadius[][] = new float[detail][detail];

		for (int i = 0; i < detail; i++) {
			for (int j = 0; j < detail; j++) {
				maxRadius[i][j] = -1.0f;
				v[i][j] = new Vertex();
				v[i][j].data = 1 + i * detail + j;
			}
		}
		
		// find maxima points
		for (int i = 0; i < numPoints; i++) {
			float fAlpha = alpha(scaledCoord, i * 3, center);
			float fTheta = theta(scaledCoord, i * 3, center);
			float fRadius = radius(scaledCoord, i * 3, center);
			int alpha = (int) (detail * fAlpha / (2 * Math.PI));
			int theta = (int) (detail * fTheta / (Math.PI));

			if (alpha >= detail) {
				alpha = detail - 1;
			} else if (alpha < 0) {
				alpha = 0;
			}
			if (theta >= detail) {
				theta = detail - 1;
			} else if (theta < 0) {
				theta = 0;
			}
			if (maxRadius[theta][alpha] < fRadius) {
				maxRadius[theta][alpha] = fRadius;
				v[theta][alpha].coordinates[0] = scaledCoord[i * 3 + 0];
				v[theta][alpha].coordinates[1] = scaledCoord[i * 3 + 1];
				v[theta][alpha].coordinates[2] = scaledCoord[i * 3 + 2];
			}
		}
		
		// interpolate missing data
		
		boolean noneInterp = true;
		
		float xyzz[] = new float[3];

		do {
			noneInterp = true;
			for (int i = 0; i < detail; i++) {
				for (int j = 0; j < detail; j++) {
					if (maxRadius[i][j] < 3.0f) {
						noneInterp = false;
						float neighborsTotal = 0;

						if (i < detail - 1 && maxRadius[i + 1][j] > 0) {
							neighborsTotal += maxRadius[i + 1][j];
						}
						if (i > 0 && maxRadius[i - 1][j] > 0) {
							neighborsTotal += maxRadius[i - 1][j];
						}
						if (maxRadius[i][(j + 1) % detail] > 0) {
							neighborsTotal += maxRadius[i][(j + 1) % detail];
						}
						if (maxRadius[i][(j - 1 + detail) % detail] > 0) {
							neighborsTotal += maxRadius[i][(j - 1 + detail) % detail];
						}
						if (neighborsTotal > 0) {
							float alphaDirection = (float) (2 * Math.PI
									- 2 * Math.PI * (j + 0.5f) / (detail));
							float thetaDirection = (float) (Math.PI * (i + 0.5f) / (detail));
							
							float newRadius = neighborsTotal / 4.0f;

							if (newRadius < 3.0f) {
								newRadius = 3.0f;
							}
							float newChroma = newRadius * (float) Math.sin(thetaDirection);

							maxRadius[i][j] = newRadius;
							// float x, y, z;
							xyzz[0] = center[0] + (float) (Math.sin(alphaDirection) * newChroma);
							xyzz[1] = center[1] - (float) (Math.cos(thetaDirection) * newRadius);
							xyzz[2] = center[2] + (float) (Math.cos(alphaDirection) * newChroma);
							v[i][j].coordinates[0] = xyzz[0];
							v[i][j].coordinates[1] = xyzz[1];
							v[i][j].coordinates[2] = xyzz[2];
							// System.err.println("Vertex "+v[i][j].data+" interpolated");
							// System.out.println("Center alpha: "+alphaDirection+", theta: "+thetaDirection);
							// System.out.println("Vertex alpha: "+alpha(xyzz, 0, center)+", theta: "+theta(xyzz, 0, center));
							// System.out.println("Segment: "+i+","+j);
							float fAlpha = alpha(xyzz, 0, center);
							float fTheta = theta(xyzz, 0, center);
							float fRadius = radius(xyzz, 0, center);
							int iAlpha = (int) (detail * fAlpha / (2 * Math.PI));
							int iTheta = (int) (detail * fTheta / (Math.PI));
							// System.out.println("Vertex segment: "+iTheta+","+iAlpha);
							// System.out.println("Radius: "+fRadius);
						}
					}
				}
			}
		} while (!noneInterp);
		
		// smooth
		
		int smooth = chooseSmooth.getSelectedIndex();
		float smoothRadius[][] = new float[detail][detail];

		for (int s = 0; s < smooth; s++) {
			for (int i = 0; i < detail; i++) {
				for (int j = 0; j < detail; j++) {
					float myR = maxRadius[i][j];
					float oldR = myR;
					float lR = maxRadius[i][(j - 1 + detail) % detail];
					float rR = maxRadius[i][(j + 1) % detail];
					float uR = 0.0f;
					float loR = 0.0f;

					if (i > 0) {
						loR = maxRadius[i - 1][j];
					} else {
						loR = maxRadius[i][(j + detail / 2) % 2];
					}
					if (i < detail - 1) {
						uR = maxRadius[i + 1][j];
					} else {
						uR = maxRadius[i][(j + detail / 2) % 2];
					}
					myR = /* myR/2.0f+*/(lR + rR + uR + loR) / 4.0f/*/ 2.0f*/;
					if (myR < oldR) {
						myR = oldR;
					}
					smoothRadius[i][j] = myR;
					v[i][j].coordinates[0] = center[0]
							+ myR * (v[i][j].coordinates[0] - center[0]) / oldR;
					v[i][j].coordinates[1] = center[1]
							+ myR * (v[i][j].coordinates[1] - center[1]) / oldR;
					v[i][j].coordinates[2] = center[2]
							+ myR * (v[i][j].coordinates[2] - center[2]) / oldR;
				}
			}
			float tmpR[][] = maxRadius;

			maxRadius = smoothRadius;
			smoothRadius = tmpR;
		}

		// unscale ...
		for (int i = 0; i < detail; i++) {
			for (int j = 0; j < detail; j++) {
				v[i][j].coordinates[0] /= scale[0];
				v[i][j].coordinates[1] /= scale[1];
				v[i][j].coordinates[2] /= scale[2];
			}
		}
		center[0] /= scale[0];
		center[1] /= scale[1];
		center[2] /= scale[2];
		
		// make facets;
		int lastFacet = 0;
		// body
		Vertex cV = new Vertex(center[0], center[1], center[2]);
		SMRect smRect[][] = new SMRect[detail - 1][detail];

		for (int i = 0; i < detail - 1; i++) {
			for (int j = 0; j < detail; j++) {
				smRect[i][j] = new SMRect();
			}
		}
		
		for (int i = 0; i < detail - 1; i++) {
			for (int j = 0; j < detail; j++) {
				int rep = 0;

				do {
					smRect[i][j].createFrom(cV, v[i + 1][j], v[i + 1][(j + 1) % detail],
							v[i + 0][j], v[i + 0][(j + 1) % detail]);
																		
					if (smRect[i][j].possibleProblem) {
						for (int o = 0; o < 2; o++) {
							for (int p = 0; p < 2; p++) {
								int iIndex = i + o;
								int jIndex = (j + p) % detail;
								float alphaDirection = (float) (2 * Math.PI
										- 2 * Math.PI * (jIndex + 0.5f) / (detail));
								float thetaDirection = (float) (Math.PI * (iIndex + 0.5f) / (detail));
								float newRadius = maxRadius[iIndex][jIndex];
								float newChroma = newRadius * (float) Math.sin(thetaDirection);
								float x, y, z;

								x = center[0] * scale[0]
										+ (float) (Math.sin(alphaDirection) * newChroma);
								y = center[1] * scale[1]
										- (float) (Math.cos(thetaDirection) * newRadius);
								z = center[2] * scale[2]
										+ (float) (Math.cos(alphaDirection) * newChroma);
								x /= scale[0];
								y /= scale[1];
								z /= scale[2];
								
								float xDir = x - (float) v[iIndex][jIndex].coordinates[0];
								float yDir = y - (float) v[iIndex][jIndex].coordinates[1];
								float zDir = z - (float) v[iIndex][jIndex].coordinates[2];
								float len = (float) Math.sqrt(xDir * xDir + yDir * yDir + zDir * zDir);

								if (len == 0) {
									float coord[] = new float[] {(float) v[iIndex][jIndex].coordinates[0],
										(float) v[iIndex][jIndex].coordinates[1],
										(float) v[iIndex][jIndex].coordinates[2]};
									// System.out.println("Alpha: "+alpha(coord, 0, new float[]{0.5f, 0.5f, 0.5f}));
									// System.out.println("Center alpha: "+alphaDirection);
									// System.out.println("Theta: "+theta(coord, 0, new float[]{0.5f, 0.5f, 0.5f}));
									// System.out.println("Center theta: "+thetaDirection);
								}
								if (len != 0) {
									float move = 0.001f * (rep + 1);

									if (rep == 10 || move > len) {
										v[iIndex][jIndex].coordinates[0] = x;
										v[iIndex][jIndex].coordinates[1] = y;
										v[iIndex][jIndex].coordinates[2] = z;
									} else {
										v[iIndex][jIndex].coordinates[0] += move * xDir / len;
										v[iIndex][jIndex].coordinates[1] += move * yDir / len;
										v[iIndex][jIndex].coordinates[2] += move * zDir / len;
									}
								}
							}
						}
					}
				} while (smRect[i][j].possibleProblem && ++rep < 11);
				if (smRect[i][j].possibleProblem) {// System.out.println("Problem remains");
					// System.out.println("Rep is "+rep);
				}
				smRect[i][j].left.data = ++lastFacet;
				smRect[i][j].right.data = ++lastFacet;
				
			}
		}
		
		for (int i = 0; i < detail - 1; i++) {
			for (int j = 0; j < detail; j++) {
				smRect[i][(j + detail - 1) % detail].setRightNeighbor(smRect[i][j].left);
				smRect[i][(j + detail + 1) % detail].setLeftNeighbor(smRect[i][j].right);
				int below = i - 1;
				int above = i + 1;

				if (smRect[i][j].leftIsUpper) {
					if (above < detail - 1) {
						smRect[above][j].setLowerNeighbor(smRect[i][j].left);
					}
					if (below >= 0) {
						smRect[below][j].setUpperNeighbor(smRect[i][j].right);
					}
				} else {
					if (above < detail - 1) {
						smRect[above][j].setLowerNeighbor(smRect[i][j].right);
					}
					if (below >= 0) {
						smRect[below][j].setUpperNeighbor(smRect[i][j].left);
					}
				}
			}
		}
		
		// top
		
		float prevCoord[] = new float[3 * detail];

		// System.out.println("Coords: ");
		for (int i = 0; i < detail; i++) {
			// System.out.println("Vertex "+v[detail-1][i].data+": ");
			prevCoord[i * 3 + 0] = (float) v[detail - 1][i].coordinates[0];
			prevCoord[i * 3 + 1] = (float) v[detail - 1][i].coordinates[1];
			prevCoord[i * 3 + 2] = (float) v[detail - 1][i].coordinates[2];
			v[detail - 1][i].coordinates[0] = prevCoord[i * 3 + 0];
			v[detail - 1][i].coordinates[1] = prevCoord[i * 3 + 1];
			v[detail - 1][i].coordinates[2] = prevCoord[i * 3 + 2];
			float x = prevCoord[i * 3 + 0];
			float y = prevCoord[i * 3 + 1];
			float z = prevCoord[i * 3 + 2];
			
			// System.out.println("Original: "+x+", "+y+", "+z);
			float extX = center[0] + (x - center[0]) * 0.5f / (y - center[1]);
			float extY = center[1] + (y - center[1]) * 0.5f / (y - center[1]);
			float extZ = center[2] + (z - center[2]) * 0.5f / (y - center[1]);

			// System.out.println("New: "+extX+", "+extY+", "+extZ);
			// y==0 ???
			v[detail - 1][i].coordinates[0] = extX;
			v[detail - 1][i].coordinates[1] = extY;
			v[detail - 1][i].coordinates[2] = extZ;
			
		}
		
		Vertex vList[];
		int numV;
		Plane plane = new Plane();
		OptimizedVector faces;
		
		vList = new Vertex[detail];
		faces = new OptimizedVector();
		for (int i = 0; i < detail; i++) {
			vList[i] = v[detail - 1][i];
		}
		numV = detail;
		float abovePointX, abovePointY, abovePointZ;

		while (numV >= 3) {
			int earV = -1;

			for (int i = 0; i < numV; i++) {
				abovePointX = (float) vList[i].coordinates[0];
				abovePointY = (float) vList[i].coordinates[1] + 0.5f;
				abovePointZ = (float) vList[i].coordinates[2];
				plane.set((float) vList[(i + detail - 1) % detail].coordinates[0],
						(float) vList[(i + detail - 1) % detail].coordinates[1],
						(float) vList[(i + detail - 1) % detail].coordinates[2], abovePointX,
						abovePointY, abovePointZ, (float) vList[i].coordinates[0],
						(float) vList[i].coordinates[1], (float) vList[i].coordinates[2]);
				if (plane.classifyPoint2((float) vList[(i + 1) % detail].coordinates[0],
						(float) vList[(i + 1) % detail].coordinates[1],
						(float) vList[(i + 1) % detail].coordinates[2])
						== Plane.PLANE_INSIDE) {
					//
					Facet f = new Facet(vList[i], vList[(i + numV - 1) % numV],
							vList[(i + 1) % numV]);
					boolean verticesInTriangle = false;

					for (int k = 0; k < numV; k++) {
						if (k == i || k == (i + 1) % numV || k == (i + numV - 1) % numV) {
							continue;
						}
						if (f.vertexInTriangle(vList[k])) {
							// System.err.println("Vertex "+vList[k].data+" in triangle");
							verticesInTriangle = true;
							break;
						} // else {
						// System.err.println("Vertex "+vList[k].data+" not in triangle");
						// }
						
					}
					if (!verticesInTriangle) {
						earV = i;
						break;
					}
				}
			}
			if (earV == -1) {
				SystemManager.displayError("Error", "Unable to triangulate top polygon");
				// System.err.println("No 'ear' vertex found by triangulation algorithm");
				// System.err.println("Vertices left: "+numV);
				return null;
			}
			// add face
			Facet f = new Facet(vList[earV], vList[(earV + numV - 1) % numV],
					vList[(earV + 1) % numV]);

			// System.err.println("New facet from "+f.corner[0].data+", "+f.corner[1].data+", "+f.corner[2].data);
			f.data = ++lastFacet;
			faces.addElement(f);
			for (int i = earV + 1; i < numV; i++) {
				vList[i - 1] = vList[i];
			}
			numV--;
		}
		
		for (int i = 0; i < detail; i++) {
			v[detail - 1][i].coordinates[0] = prevCoord[i * 3 + 0];
			v[detail - 1][i].coordinates[1] = prevCoord[i * 3 + 1];
			v[detail - 1][i].coordinates[2] = prevCoord[i * 3 + 2];			
		}
		
		// internal neighbor structure
		for (int i = 0; i < faces.size(); i++) {
			Facet f = (Facet) faces.elementAt(i);
			Vertex va = f.corner[0];
			Vertex vb = f.corner[1];
			Vertex vc = f.corner[2];

			for (int j = 0; j < faces.size(); j++) {
				if (i == j) {
					continue;
				}
				Facet f2 = (Facet) faces.elementAt(j);

				for (int k = 0; k < 3; k++) {
					Vertex v1 = f2.corner[k];
					Vertex v2 = f2.corner[(k + 1) % 3];

					if ((v1 == va || v1 == vb || v1 == vc)
							&& (v2 == va || v2 == vb || v2 == vc)) {
						f2.neighbor[k] = f;
					}
				}
			}
		}
		
		// external neighbor structure
		
		for (int i = 0; i < faces.size(); i++) {
			Facet f = (Facet) faces.elementAt(i);

			for (int j = 0; j < 3; j++) {
				if (f.neighbor[j] == null) {
					int v0 = f.corner[j].data;
					int v1 = f.corner[(j + 1) % 3].data;

					v0 = (v0 - 1) % detail;
					v1 = (v1 - 1) % detail;
					// System.err.println("-v0: "+v0);
					// System.err.println("-v1: "+v1);
					int low = (v0 < v1) ? v0 : v1;
					int high = (v0 > v1) ? v0 : v1;

					if (low != high - 1) {
						low = high;
					}
					smRect[detail - 2][low].setUpperNeighbor(f);
					Facet upperBelow = smRect[detail - 2][low].right;

					if (smRect[detail - 2][low].leftIsUpper) {
						upperBelow = smRect[detail - 2][low].left;
					}
					f.neighbor[j] = upperBelow;
				}
			}
		}
		
		// bottom
		
		OptimizedVector faces1 = faces;
		
		for (int i = 0; i < detail; i++) {
			// System.out.println("Vertex "+v[0][i].data+": ");
			prevCoord[i * 3 + 0] = (float) v[0][i].coordinates[0];
			prevCoord[i * 3 + 1] = (float) v[0][i].coordinates[1];
			prevCoord[i * 3 + 2] = (float) v[0][i].coordinates[2];
			v[0][i].coordinates[0] = prevCoord[i * 3 + 0];
			v[0][i].coordinates[1] = prevCoord[i * 3 + 1];
			v[0][i].coordinates[2] = prevCoord[i * 3 + 2];
			float x = prevCoord[i * 3 + 0];
			float y = prevCoord[i * 3 + 1];
			float z = prevCoord[i * 3 + 2];
			
			// System.out.println("Original: "+x+", "+y+", "+z);
			float extX = center[0] + (x - center[0]) * 0.5f / (center[1] - y);
			float extY = center[1] + (y - center[1]) * 0.5f / (center[1] - y);
			float extZ = center[2] + (z - center[2]) * 0.5f / (center[1] - y);

			// System.out.println("New: "+extX+", "+extY+", "+extZ);
			// y==0 ???
			v[0][i].coordinates[0] = extX;
			v[0][i].coordinates[1] = extY;
			v[0][i].coordinates[2] = extZ;
			
		}
		
		vList = new Vertex[detail];
		faces = new OptimizedVector();
		for (int i = 0; i < detail; i++) {
			vList[i] = v[0][i];
		}
		numV = detail;
		while (numV >= 3) {
			int earV = -1;

			for (int i = 0; i < numV; i++) {
				abovePointX = (float) vList[i].coordinates[0];
				abovePointY = (float) vList[i].coordinates[1] + 0.5f;
				abovePointZ = (float) vList[i].coordinates[2];
				plane.set((float) vList[(i + detail - 1) % detail].coordinates[0],
						(float) vList[(i + detail - 1) % detail].coordinates[1],
						(float) vList[(i + detail - 1) % detail].coordinates[2], abovePointX,
						abovePointY, abovePointZ, (float) vList[i].coordinates[0],
						(float) vList[i].coordinates[1], (float) vList[i].coordinates[2]);
				if (plane.classifyPoint2((float) vList[(i + 1) % detail].coordinates[0],
						(float) vList[(i + 1) % detail].coordinates[1],
						(float) vList[(i + 1) % detail].coordinates[2])
						== Plane.PLANE_INSIDE) {
					//
					
					Facet f = new Facet(vList[i], vList[(i + 1) % numV],
							vList[(i + numV - 1) % numV]);
					boolean verticesInTriangle = false;

					for (int k = 0; k < numV; k++) {
						if (k == i || k == (i + 1) % numV || k == (i + numV - 1) % numV) {
							continue;
						}
						if (f.vertexInTriangle(vList[k])) {
							// System.err.println("Vertex "+vList[k].data+" in triangle");
							verticesInTriangle = true;
							break;
						} // else {
						// System.err.println("Vertex "+vList[k].data+" not in triangle");
						// }
						
					}
					if (!verticesInTriangle) {
						earV = i;
						break;
					}
				}
			}
			if (earV == -1) {
				// System.err.println("No 'ear' vertex found by triangulation algorithm");
				// System.err.println("Vertices left: "+numV);
				SystemManager.displayError("Error", "Unable to triangulate bottom polygon");
				return null;
			}
			// add face
			Facet f = new Facet(vList[earV], vList[(earV + 1) % numV],
					vList[(earV + numV - 1) % numV]);

			// System.err.println("New facet from "+f.corner[0].data+", "+f.corner[1].data+", "+f.corner[2].data);
			f.data = ++lastFacet;
			faces.addElement(f);
			for (int i = earV + 1; i < numV; i++) {
				vList[i - 1] = vList[i];
			}
			numV--;
		}
		
		for (int i = 0; i < detail; i++) {
			v[0][i].coordinates[0] = prevCoord[i * 3 + 0];
			v[0][i].coordinates[1] = prevCoord[i * 3 + 1];
			v[0][i].coordinates[2] = prevCoord[i * 3 + 2];			
		}
		
		// internal neighbor structure
		for (int i = 0; i < faces.size(); i++) {
			Facet f = (Facet) faces.elementAt(i);
			Vertex va = f.corner[0];
			Vertex vb = f.corner[1];
			Vertex vc = f.corner[2];

			for (int j = 0; j < faces.size(); j++) {
				if (i == j) {
					continue;
				}
				Facet f2 = (Facet) faces.elementAt(j);

				for (int k = 0; k < 3; k++) {
					Vertex v1 = f2.corner[k];
					Vertex v2 = f2.corner[(k + 1) % 3];

					if ((v1 == va || v1 == vb || v1 == vc)
							&& (v2 == va || v2 == vb || v2 == vc)) {
						f2.neighbor[k] = f;
					}
				}
			}
		}
		
		// external neighbor structure
		for (int i = 0; i < faces.size(); i++) {
			Facet f = (Facet) faces.elementAt(i);

			// System.err.println("Facet "+f.data);
			for (int j = 0; j < 3; j++) {
				if (f.neighbor[j] == null) {
					int v0 = f.corner[j].data;
					int v1 = f.corner[(j + 1) % 3].data;

					v0 = (v0 - 1) % detail;
					v1 = (v1 - 1) % detail;
					// System.err.println("-v0: "+v0);
					// System.err.println("-v1: "+v1);
					int low = (v0 < v1) ? v0 : v1;
					int high = (v0 > v1) ? v0 : v1;

					if (low != high - 1) {
						low = high;
					}
					smRect[0][low].setLowerNeighbor(f);
					Facet lowerAbove = smRect[0][low].left;

					if (smRect[0][low].leftIsUpper) {
						lowerAbove = smRect[0][low].right;
					}
					f.neighbor[j] = lowerAbove;
				}
			}
		}
		
		ArrayList usedVertices = new ArrayList();
		ArrayList usedFaces = new ArrayList();
		
		for (int i = 0; i < detail; i++) {
			for (int j = 0; j < detail; j++) {
				usedVertices.add(v[i][j]);
				if (i != detail - 1) {
					usedFaces.add(smRect[i][j].left);
					usedFaces.add(smRect[i][j].right);
				}
			}
		}
		
		int lowIndex = 100000000;
		int highIndex = -1;

		for (int i = 0; i < faces1.size(); i++) {
			Facet f = (Facet) faces1.elementAt(i);

			if (f.isDiscarded()) {
				continue;
			}
			if (lowIndex > f.data) {
				lowIndex = f.data;
			}
			if (highIndex < f.data) {
				highIndex = f.data;
			}
			usedFaces.add(f);
		}
		// System.out.println("Top faces between "+lowIndex+" and "+highIndex);
		lowIndex = 100000000;
		highIndex = -1;
		for (int i = 0; i < faces.size(); i++) {
			Facet f = (Facet) faces.elementAt(i);

			if (f.isDiscarded()) {
				continue;
			}
			if (lowIndex > f.data) {
				lowIndex = f.data;
			}
			if (highIndex < f.data) {
				highIndex = f.data;
			}
			usedFaces.add(f);
		}
		// System.out.println("Bottom faces between "+lowIndex+" and "+highIndex);
		
		for (int i = 0; i < usedFaces.size(); i++) {
			Facet f = (Facet) usedFaces.get(i);

			for (int j = 0; j < 3; j++) {
				if (f.neighbor[j] == null || f.neighbor[j].isDiscarded()) {
					if (f.neighbor[j] == null) {
						System.err.println("Face " + f.data + " neighbor " + j + " is null");
					} else {
						System.err.println("Face " + f.data + " neighbor " + j + " is discarded");
					}
					System.err.println("Edge between vertex " + f.corner[j].data + " and "
							+ f.corner[(j + 1) % 3].data);
					continue;
				}
				Facet n = f.neighbor[j];
				int foundInList = 0;

				for (int k = 0; k < 3; k++) {
					if (n.neighbor[k] == f) {
						foundInList++;
					}
				}
				if (foundInList != 1) {
					System.err.println("Facet " + f.data + " has neighbor " + n.data
							+ ", but the neighbor has facet " + foundInList + " times in its list");
				}
			}
		}
		tetra = new Tetrahedron[usedFaces.size()];
		return buildStructure(usedVertices, usedFaces, center, colorSpace3D, tetra,
				scale, convexVolume, numPoints, pointCoord, numEqual, imageManager, view3D,
				triangleAppearance);
	}
	
	/**
	 * Builds the 3D structure.
	 * @param usedVertices A list of vertices.
	 * @param usedFaces A list of faces/triangles.
	 * @param center The color space center.
	 * @param colorSpace3D The input and output color spaces.
	 * @param tetra The tetrahedrons.
	 * @param scale The color space scale.
	 * @param volumeLabel The object volume label.
	 * @param numPoints The number of points.
	 * @param pointCoord Thye vertex coordinates.
	 * @param numEqual The number of equal pixels.
	 * @param imageManager The corresponding image manager.
	 * @param view3D The View3D where the 3D branch are to be added.
	 * @param triangleAppearance The panel that defines the appearance of the object.
	 */
	public static BranchGroup[] buildStructure(
			ArrayList usedVertices,
			ArrayList usedFaces,
			float center[],
			ColorSpace3D colorSpace3D[],
			Tetrahedron tetra[],
			float scale[],
			JLabel volumeLabel,
			int numPoints,
			float pointCoord[][],
			int numEqual[],
			ImageModule imageManager,
			View3D view3D,
			TriangleAppearance triangleAppearance) {
		Facet face = null;
		BranchGroup branchGroup[] = new BranchGroup[2];

		branchGroup[0] = new BranchGroup();
		branchGroup[1] = new BranchGroup();
		
		IndexedTriangleArray ita = new IndexedTriangleArray(usedVertices.size(),
				GeometryArray.COORDINATES | GeometryArray.COLOR_3, usedFaces.size() * 3);
		
		PointArray pa = new PointArray(usedVertices.size(),
				GeometryArray.COORDINATES | GeometryArray.COLOR_3);
		GeometryArray ga[] = new GeometryArray[2];

		ga[0] = ita;
		ga[1] = pa;
		
		SegmentObject3D objects[] = new SegmentObject3D[usedVertices.size()];
		
		BasicObject3D centerObject = new BasicObject3D(0,
				new float[] {center[0], center[1], center[2]}, new byte[] {127, 127, 127}, 0,
				null);
		
		Vertex tmpVertex;

		for (int i = 0; i < objects.length; i++) {
			tmpVertex = (Vertex) usedVertices.get(i);
			objects[i] = new SegmentObject3D(centerObject.getPositionReference(), 0,
					new float[] {(float) tmpVertex.coordinates[0],
				(float) tmpVertex.coordinates[1], (float) tmpVertex.coordinates[2]},
					colorSpace3D[1].toSRGB(new float[] {(float) tmpVertex.coordinates[0],
				(float) tmpVertex.coordinates[1], (float) tmpVertex.coordinates[2]}),
					new int[] {0, 1}, new int[] {i, i}, ga);
			ita.setCoordinate(i, objects[i].getPositionReference());
			pa.setCoordinate(i, objects[i].getPositionReference());
			ita.setColor(i, objects[i].getColorReference());
		}
		for (int i = 0; i < usedFaces.size(); i++) {
			face = (Facet) usedFaces.get(i);
			for (int j = 0; j < 3; j++) {
				ita.setCoordinateIndex(i * 3 + j, face.corner[j].data - 1);
				ita.setColorIndex(i * 3 + j, face.corner[j].data - 1);
			}
		}
		// make all tetrahedrons and connect them and their vertices together
		
		int a, b, c;

		for (int i = 0; i < tetra.length; i++) {
			a = ((Facet) (usedFaces.get(i))).corner[0].data - 1;
			c = ((Facet) (usedFaces.get(i))).corner[1].data - 1;
			b = ((Facet) (usedFaces.get(i))).corner[2].data - 1;
			tetra[i] = new Tetrahedron(centerObject, objects[a], objects[b], objects[c]);
			objects[a].addTetrahedron(tetra[i]);
			objects[b].addTetrahedron(tetra[i]);
			objects[c].addTetrahedron(tetra[i]);
		}
		for (int i = 0; i < objects.length; i++) {
			objects[i].finalizeTetrahedrons();
		}
		for (int i = 0; i < tetra.length; i++) {
			tetra[i].setNeighbor(0,
					tetra[((Facet) (usedFaces.get(i))).neighbor[2].data - 1]);
			tetra[i].setNeighbor(1,
					tetra[((Facet) (usedFaces.get(i))).neighbor[1].data - 1]);
			tetra[i].setNeighbor(2,
					tetra[((Facet) (usedFaces.get(i))).neighbor[0].data - 1]);
		}
		
		boolean success = true;
		
		if (true) {
			for (int i = 0; i < tetra.length; i++) {
				if (tetra[i].getNeighbor(0) == tetra[i].getNeighbor(1) 
						|| tetra[i].getNeighbor(0) == tetra[i].getNeighbor(2) 
						|| tetra[i].getNeighbor(1) == tetra[i].getNeighbor(2)) {
					System.out.println("Tetra " + i + " has the same neighbor more than"
							+ " once in its neighbor list");
				}
				for (int j = 0; j < 3; j++) {
					if (tetra[i].getNeighbor(j) == null) {
						System.out.println("Tetra " + i + " is missing neighbor " + j);
					} else {
						Tetrahedron nbor = tetra[i].getNeighbor(j);
						boolean good = false;

						for (int k = 0; k < 3; k++) {
							if (nbor.getNeighbor(k) == tetra[i]) {
								good = true;
							}
						}
						if (!good) {
							System.out.println("Tetra " + i + "has neighbor " + j + ", not good");
						}
					}
				}
			}
				
			for (int i = 0; i < tetra.length; i++) {
				if (tetra[i] != null && !tetra[i].test(true)) {
					success = false;
					System.err.println("Problem with tetra " + i);
				}
			}
			if (!success) {
				success = true;
				for (int i = 0; i < tetra.length; i++) {
					if (tetra[i] != null && !tetra[i].test(false)) {
						success = false;
					}
				}
			}
		}
			
		if (!success) {
			SystemManager.displayError("Error", "Error building tetrahedron structure");
		}
		
		float volume = 0f;
		float m[][] = new float[4][4];

		for (int i = 0; i < tetra.length; i++) {
			volume += ConvexHullVisualization3D.volume(tetra[i], scale, m);
		}
		volumeLabel.setText(Float.toString(volume));
		Tetrahedron tetraPointers[] = new Tetrahedron[numPoints];
		Tetrahedron last = tetra[0];
		Tetrahedron aTet;

		for (int p = 0; p < numPoints; p++) {
			try {
				aTet = last.findSegmentTetrahedron(null, pointCoord[1][p * 3],
						pointCoord[1][p * 3 + 1], pointCoord[1][p * 3 + 2]);
				tetraPointers[p] = aTet;
				last = aTet;
				aTet.addColor();
			} catch (StackOverflowError e) {
				e.printStackTrace();
				System.out.println("Alternative method");
				for (int i = 0; i < tetra.length; i++) {
					if (tetra[i].isInSegment(pointCoord[1][p * 3], pointCoord[1][p * 3 + 1],
							pointCoord[1][p * 3 + 2])) {
						tetraPointers[p] = tetra[i];
						break;
					}
				}
				System.out.println("Unable to find a suitable tetrahedron");
			}
		}
		for (int i = 0; i < tetra.length; i++) {
			tetra[i].finalizeColorIndices();
		}
		
		for (int p = 0; p < numPoints; p++) {
			if (tetraPointers[p] != null) {
				tetraPointers[p].addColor(p,
						new float[] {pointCoord[1][p * 3], pointCoord[1][p * 3 + 1],
					pointCoord[1][p * 3 + 2]});	
			}
		}
		
		if (numEqual != null) {
			MovingObject mo = new MovingObject();

			mo.setColors(pointCoord[1]);
			mo.setColorSpace(colorSpace3D[1]);
			mo.setImage2D(imageManager.getImage2D());
			mo.setObjects(objects);
			mo.setTetrahedrons(tetra);
			mo.setStart(0);
			mo.setGeometryArrays(ga);
			StandardVisualization3D.setCapability(pa);
			StandardVisualization3D.setCapability(ita);
			mo.setGeometry(pa);
			mo.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
			branchGroup[0].addChild(mo);
			view3D.getImageGamuts().add(mo);
		} else {
			view3D.getDeviceGamuts().add(tetra);
		}
		
		Shape3D shape3D[] = triangleAppearance.getShapes(ita);

		for (int i = 0; i < shape3D.length; i++) {
			branchGroup[1].addChild(shape3D[i]);
		}
		return branchGroup;
	}
	
	/**
	 * Computes the radius of a given point.
	 * @param lab An array of points (x, y, z, x, y, z).
	 * @param i The index of the first float representing the point.
	 * @param middle The center point.
	 * @return The radius of the point.
	 */
	public static final float radius(float lab[], int i, float middle[]) {
		float retRadius = (float) Math.sqrt((lab[i] - middle[0])
						* (lab[i] - middle[0])
				+ (lab[i + 1] - middle[1]) * (lab[i + 1] - middle[1])
				+ (lab[i + 2] - middle[2]) * (lab[i + 2] - middle[2]));

		return retRadius;
	}
	
	/**
	 * Computes the alpha angle for a given point.
	 * @param lab An array of points (x, y, z, x, y, z).
	 * @param i The index of the first float representing the point.
	 * @param middle The center point.
	 * @return The alpha angle of the point.
	 */
	public static final float alpha(float lab[], int i, float middle[]) {
		
		float a = lab[i + 2] - middle[2];
		float b = lab[i] - middle[0];
		
		if (a == 0.0f) {
			a = 0.00000000000001f;
		}
		
		float res = (float) Math.atan(b / a);

		if (b < 0.0f && a > 0.0f) {
			res += (float) (2 * Math.PI);
		} else if (a < 0.0f) {
			res += (float) Math.PI;
		}
		res = (float) (2 * Math.PI) - res;
		if (res < 0) {
			res = 0;
		}
		if (res >= (float) (Math.PI * 2 - 0.000001f)) {
			res = (float) Math.PI * 2 - 0.000001f;
		}
		return res;
	}
	
	/**
	 * Computes the hue of a color.
	 * @param x The x coordinate.
	 * @param z The z coordinate.
	 * @return The hue of the color.
	 */
	public static final float hue(float x, float z) {
		float a = z - 0.5f;
		float b = x - 0.5f;
		
		if (a == 0.0f) {
			a = 0.00000000000001f;
		}
		
		float res = (float) Math.atan(b / a);

		if (b < 0.0f && a > 0.0f) {
			res += (float) (2 * Math.PI);
		} else if (a < 0.0f) {
			res += (float) Math.PI;
		}
		if (res >= (float) (Math.PI * 2 - 0.000001f)) {
			res = (float) Math.PI * 2 - 0.000001f;
		}
		return res;
	}
	
	/**
	 * Computes the theta angle for a given point.
	 * @param lab An array of points (x, y, z, x, y, z).
	 * @param i The index of the first float representing the point.
	 * @param middle The center point.
	 * @return The theta angle of the point.
	 */
	public static final float theta(float lab[], int i, float middle[]) {
		float a = lab[i + 2] - middle[2];
		float b = lab[i] - middle[0];
		float l = lab[i + 1] - middle[1];
		Vector3f vec3f = new Vector3f(a, b, l);
		float p = vec3f.length();
		
		if (p == 0.0f) {
			p = 0.00000000000001f;
		}
		
		float res = (float) (Math.PI / 2.0 + Math.asin(l / p));

		if (res >= (float) (Math.PI - 0.000001f)) {
			res = (float) Math.PI - 0.000001f;
		}
		return res;
	}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0=outside).
	 */
	public BitSet getVoxel(int detail) {
		return ConvexHullVisualization3D.getVoxel(detail, tetra);
	}
	
	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization.
	 */
	public String getName() {
		return "Segment Maxima";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 3;
	}
	
	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}
	
	/**
	 * Creates the parameter panel.
	 **/
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		JPanel upperGridPanel = new JPanel();

		upperGridPanel.setLayout(new BoxLayout(upperGridPanel, BoxLayout.Y_AXIS));
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Segment Maxima"));
		
		String detLevels[] = new String[46];

		for (int i = 0; i < detLevels.length; i++) {
			detLevels[i] = Integer.toString(i + 5);
		}
		chooseDetail = new JComboBox(detLevels);
		chooseSmooth = new JComboBox(new String[] {"0", "1", "2", "3", "4", "5", "6",
			"7", "8", "10"});
		
		JPanel pPanel = new JPanel(new GridLayout(0, 1));
		
		JPanel dispPanel = new JPanel(new GridLayout(1, 2));

		dispPanel.add(new JLabel("Detail"));
		dispPanel.add(chooseDetail);
		pPanel.add(dispPanel);
		dispPanel = new JPanel(new GridLayout(1, 2));
		dispPanel.add(new JLabel("Smooth"));
		dispPanel.add(chooseSmooth);
		// pPanel.add(dispPanel);
		convexVolume = new JLabel("0");
		convexVolume.setHorizontalAlignment(JLabel.RIGHT);
		JPanel volPanel = new JPanel(new GridLayout(1, 2));

		volPanel.add(new JLabel("Volume"));
		volPanel.add(convexVolume);
		pPanel.add(volPanel);
		upperGridPanel.add(pPanel);
		triangleAppearance = new TriangleAppearance();
		upperGridPanel.add(triangleAppearance);
		parameterPanel.add(upperGridPanel, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}
	
	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */
	public boolean checkParams() {
		return true;
	}
	
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {}
		
}


/**
 * A segment maxima rectangle defined by four sm points.
 */
class SMRect {
	
	/**
	 * The left facet.
	 */
	Facet left;
	
	/**
	 * The right facet.
	 */
	Facet right;
	
	/**
	 * Certain constellations represent a problem when dividing the rectangle into triangles.
	 */
	boolean possibleProblem;
	
	/**
	 * True if the left facet is the upper neigbor.
	 */
	boolean leftIsUpper;
	
	/**
	 * Creates a rectangle from 4 vertices.
	 * NOTICE: Parameters are mislabeled ... TODO
	 * @param center The object center.
	 * @param upperLeft The upper left vertex.
	 * @param upperRight The upper right facet.
	 * @param lowerLeft The lower left facet.
	 * @param lowerRight The lower right facet.
	 */
	void createFrom(Vertex center, Vertex upperLeft, Vertex upperRight, Vertex lowerLeft, Vertex lowerRight) {
		possibleProblem = false;
		Plane crossPlane = new Plane((float) center.coordinates[0],
				(float) center.coordinates[1], (float) center.coordinates[2],
				(float) lowerRight.coordinates[0], (float) lowerRight.coordinates[1],
				(float) lowerRight.coordinates[2], (float) upperLeft.coordinates[0],
				(float) upperLeft.coordinates[1], (float) upperLeft.coordinates[2]);
																	
		Plane crossPlane2 = new Plane((float) center.coordinates[0],
				(float) center.coordinates[1], (float) center.coordinates[2],
				(float) upperRight.coordinates[0], (float) upperRight.coordinates[1],
				(float) upperRight.coordinates[2], (float) lowerLeft.coordinates[0],
				(float) lowerLeft.coordinates[1], (float) lowerLeft.coordinates[2]);
																	
		int ur = crossPlane.classifyPoint((float) upperRight.coordinates[0],
				(float) upperRight.coordinates[1], (float) upperRight.coordinates[2]);
		int ll = crossPlane.classifyPoint((float) lowerLeft.coordinates[0],
				(float) lowerLeft.coordinates[1], (float) lowerLeft.coordinates[2]);
		
		int lr = crossPlane2.classifyPoint((float) lowerRight.coordinates[0],
				(float) lowerRight.coordinates[1], (float) lowerRight.coordinates[2]);
		int ul = crossPlane2.classifyPoint((float) upperLeft.coordinates[0],
				(float) upperLeft.coordinates[1], (float) upperLeft.coordinates[2]);
		boolean leftCanBeUpper = (ul == Plane.PLANE_OUTSIDE
				&& lr == Plane.PLANE_INSIDE);
		boolean rightCanBeUpper = (ur == Plane.PLANE_OUTSIDE
				&& ll == Plane.PLANE_INSIDE);

		if (leftCanBeUpper && !rightCanBeUpper) {
			leftIsUpper = true;
		} else if (!leftCanBeUpper && rightCanBeUpper) {
			leftIsUpper = false;
		} else if (leftCanBeUpper && rightCanBeUpper) {
			Facet l1 = new Facet(lowerLeft, upperLeft, lowerRight);

			if (l1.outside(upperRight)) {
				leftIsUpper = true;
			} else {
				leftIsUpper = false;
			}
		} else {
			// System.err.println("Possible problem");
			// System.err.println("Vertices: "+upperLeft.data+", "+upperRight.data+", "+lowerLeft.data+", "+lowerRight.data);
			possibleProblem = true;
			Facet l1 = new Facet(lowerLeft, upperLeft, lowerRight);
			Facet r1 = new Facet(upperRight, lowerRight, upperLeft);

			if (l1.outside(center) || r1.outside(center)) {
				leftIsUpper = true;
			} else {
				leftIsUpper = false;
			}
		}
		if (leftIsUpper == false) {
			left = new Facet(lowerLeft, upperLeft, lowerRight);
			right = new Facet(upperRight, lowerRight, upperLeft);
			left.neighbor[1] = right;
			right.neighbor[1] = left;
		} else {
			left = new Facet(lowerLeft, upperLeft, upperRight);
			right = new Facet(upperRight, lowerRight, lowerLeft);
			left.neighbor[2] = right;
			right.neighbor[2] = left;
		}
		if (left.outside(center)) {
			System.err.println("Center outside error 2");
		}
		if (right.outside(center)) {
			System.err.println("Center outside error 4");
		}
	}
	
	/**
	 * Sets the lower facet neighbor,
	 * @param f The lower neighbor.
	 */
	void setLowerNeighbor(Facet f) {
		if (leftIsUpper && right != null) {
			if (right.neighbor[1] != null && right.neighbor[1] != f) {
				System.err.println("Overwriting previous lower neighbor");
			}
			right.neighbor[1] = f;
		} else if (!leftIsUpper && left != null) {
			left.neighbor[2] = f;
			if (left.neighbor[2] != null && left.neighbor[2] != f) {
				System.err.println("Overwriting previous lower neighbor");
			}
		}
	}
	
	/**
	 * Sets the upper facet neighbor,
	 * @param f The upper neighbor.
	 */
	void setUpperNeighbor(Facet f) {
		if (leftIsUpper && left != null) {
			if (left.neighbor[1] != null && left.neighbor[1] != f) {
				System.err.println("Overwriting previous upper neighbor");
			}
			left.neighbor[1] = f;
		} else if (!leftIsUpper && right != null) {
			if (right.neighbor[2] != null && right.neighbor[2] != f) {
				System.err.println("Overwriting previous upper neighbor");
			}
			right.neighbor[2] = f;
		}
	}
	
	/**
	 * Sets the left facet neighbor,
	 * @param f The left neighbor.
	 */
	void setLeftNeighbor(Facet f) {
		if (left != null) {
			if (left.neighbor[0] != null && left.neighbor[0] != f) {
				System.err.println("Overwriting previous left neighbor");
			}
			left.neighbor[0] = f;
		}
	}
	
	/**
	 * Sets the right facet neighbor,
	 * @param f The right neighbor.
	 */
	void setRightNeighbor(Facet f) {
		if (right != null) {
			if (right.neighbor[0] != null && right.neighbor[0] != f) {
				System.err.println("Overwriting previous right neighbor");
			}
			right.neighbor[0] = f;
		}
	}
}
