/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import java.awt.*;
import javax.swing.JPanel;
import javax.media.j3d.*;
import java.util.*;

import icc3dpackage.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.triangulation.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.generic.Visualization3D;


/**
 * MountainRangeVisualization3D displays the object as a height map.
 */
public class MountainRangeVisualization3D extends Visualization3D {

	/**
	 * Constructs a MountainRangeVisualization3D object.
	 */
	public MountainRangeVisualization3D() {}
	
	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		float lowX = pointCoord[1][0], lowZ = pointCoord[1][2], highX = pointCoord[1][0], highZ = pointCoord[1][2];

		for (int i = 0; i < numPoints; i++) {
			if (pointCoord[1][i * 3] < lowX) {
				lowX = pointCoord[1][i * 3];
			} else if (pointCoord[1][i * 3] > highX) {
				highX = pointCoord[1][i * 3];
			}
			if (pointCoord[1][i * 3 + 2] < lowZ) {
				lowZ = pointCoord[1][i * 3 + 2];
			} else if (pointCoord[1][i * 3 + 2] > highZ) {
				highZ = pointCoord[1][i * 3 + 2];
			}
		}
		
		if (lowX >= highX || lowZ >= highZ) {
			return new BranchGroup[] {null, null};
		}
		
		float diffX = highX - lowX;
		float diffZ = highZ - lowZ;
		int detail = 30;
		int numOutputPoints = 0;
		float grid[][][] = new float[detail][detail][];
		
		float xBoxSize = diffX / detail;
		float zBoxSize = diffZ / detail;
		
		int xIndex, zIndex;

		for (int i = 0; i < numPoints; i++) {
			xIndex = (int) ((pointCoord[1][i * 3] - lowX) / xBoxSize);
			zIndex = (int) ((pointCoord[1][i * 3 + 2] - lowZ) / zBoxSize);
			if (xIndex < 0) {
				xIndex = 0;
			}
			if (xIndex > detail - 1) {
				xIndex = detail - 1;
			}
			if (zIndex < 0) {
				zIndex = 0;
			}
			if (zIndex > detail - 1) {
				zIndex = detail - 1;
			}
			
			if (grid[xIndex][zIndex] == null) {
				grid[xIndex][zIndex] = new float[] {pointCoord[1][i * 3],
					pointCoord[1][i * 3 + 1], pointCoord[1][i * 3 + 2]};
				numOutputPoints++;
			} else if (grid[xIndex][zIndex][1] < pointCoord[1][i * 3 + 1]) {
				grid[xIndex][zIndex] = new float[] {pointCoord[1][i * 3],
					pointCoord[1][i * 3 + 1], pointCoord[1][i * 3 + 2]};
			}
		}
		
		if (numOutputPoints < 5) {
			return new BranchGroup[] {null, null};
		}
		
		Vertex v[] = new Vertex[numOutputPoints];
		int arrayIndex = 0;

		for (int i = 0; i < detail; i++) {
			for (int j = 0; j < detail; j++) {
				if (grid[i][j] != null) {
					v[arrayIndex] = new Vertex(grid[i][j][0] + Math.random() / 10000.0f,
							grid[i][j][2] + Math.random() / 10000.0f);
					v[arrayIndex].data = i * detail + j;
					v[arrayIndex].addDimension();
					arrayIndex++;
				}
			}
		}
		
		QuickHull qh = new QuickHull();
		OptimizedVector ov = qh.build3D(v);
		OptimizedVector ov2 = new OptimizedVector(ov.elementCount / 2);

		for (int i = 0; i < ov.elementCount; i++) {
			Facet face = (Facet) ov.elementAt(i);

			if (face.normalZ < 0 && !face.isDiscarded()) {
				ov2.addElement(face);
			}
		}
		ov = ov2;
		TriangleArray ta = new TriangleArray(3 * ov.elementCount,
				GeometryArray.COLOR_3 | GeometryArray.COORDINATES);

		for (int i = 0; i < ov.elementCount; i++) {
			Facet face = (Facet) ov.elementAt(i);
			Vertex v1 = face.corner[0];
			// float pos[] = new float[]{(float)v1.coordinates[0], (float)v1.coordinates[1], 0};
			float pos[] = grid[v1.data / detail][v1.data % detail];
			byte col[] = colorSpace3D[1].toSRGB(pos);

			ta.setColor(i * 3, col);
			ta.setCoordinate(i * 3, pos);
			v1 = face.corner[1];
			// pos = new float[]{(float)v1.coordinates[0], (float)v1.coordinates[1], 0};
			pos = grid[v1.data / detail][v1.data % detail];
			col = colorSpace3D[1].toSRGB(pos);
			ta.setColor(i * 3 + 1, col);
			ta.setCoordinate(i * 3 + 1, pos);
			v1 = face.corner[2];
			// pos = new float[]{(float)v1.coordinates[0], (float)v1.coordinates[1], 0};
			pos = grid[v1.data / detail][v1.data % detail];
			col = colorSpace3D[1].toSRGB(pos);
			ta.setColor(i * 3 + 2, col);
			ta.setCoordinate(i * 3 + 2, pos);
		}
		PolygonAttributes pa = new PolygonAttributes();

		// pa.setPolygonMode(PolygonAttributes.POLYGON_LINE);
		pa.setCullFace(PolygonAttributes.CULL_NONE);
		Appearance app = new Appearance();

		app.setPolygonAttributes(pa);
		Shape3D shape3D = new Shape3D();

		shape3D.setGeometry(ta);
		shape3D.setAppearance(app);
		BranchGroup branchGroup[] = new BranchGroup[] {null, new BranchGroup()};

		branchGroup[1].addChild(shape3D);
		return branchGroup;
	}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0= outside).
	 */
	public BitSet getVoxel(int detail) {
		BitSet bs = new BitSet(detail * detail * detail);

		return bs;
	}
	
	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization.
	 */
	public String getName() {
		return "Height map";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 10;
	}
	
	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */
	public Component getParameterPanel() {
		return new JPanel();
	}
	
	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */
	public boolean checkParams() {
		return true;
	}
	
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {}
} 

