/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations;


import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import java.lang.Math;
import java.awt.*;
import java.awt.image.*;
import java.awt.image.Raster;
import com.sun.media.jai.codecimpl.*;
import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.BitSet;

import icc3dpackage.*;
import icc3dpackage.modules.ImageModule;
import icc3dpackage.colorspaces.generic.*;
import icc3dpackage.pickable.*;
import icc3dpackage.view3d.View3D;
import icc3dpackage.visualizations.appearance.*;
import icc3dpackage.visualizations.generic.*;


/**
 * StandardVisualization3D is a visualization which displays a point array
 * representing the input points in 3D.
 */
public class StandardVisualization3D extends Visualization3D {
	
	/**
	 * The point size.
	 */
	private float pointSize;
	
	/**
	 * The parameter panel.
	 */
	private JPanel parameterPanel = null;
	
	/**
	 * The number of objects/points.
	 */
	private int nObjects;
	
	/**
	 * A label that contains the number of points/objects.
	 */
	private JLabel edtObjectCount;
	
	/**
	 * The point appearance.
	 */
	private Appearance app;
	
	/**
	 * The available point size choices.
	 */
	private JComboBox cmbPointSize;
	
	/**
	 * Default constructor for a StandardVisualizatiuon3D.
	 */
	public StandardVisualization3D() {}
	
	/**
	 * Sets the necessary capability bits on the given GeometryArray.
	 * @param pa The GeometryArray.
	 */
	public static void setCapability(GeometryArray pa) {
		pa.setCapability(GeometryArray.ALLOW_COUNT_READ);
		pa.setCapability(GeometryArray.ALLOW_FORMAT_READ);
		pa.setCapability(GeometryArray.ALLOW_COORDINATE_READ);
		pa.setCapability(GeometryArray.ALLOW_COORDINATE_WRITE);
		pa.setCapability(GeometryArray.ALLOW_COLOR_READ);
		pa.setCapability(GeometryArray.ALLOW_COLOR_WRITE);	
	}
	
	/**
	 * Computes and returns a voxel representation of the gamut.
	 *
	 * @param detail The detail level.
	 * @return A BitSet voxel representaion of the gamut. (1=inside, 0= outside).
	 */
	public BitSet getVoxel(int detail) {
		BitSet voxel = new BitSet(detail * detail * detail);

		return voxel;
	}
	
	/**
	 * Computes and returns a scene graph given a set of points and input/output
	 * color spaces. This particular implementation returns a scene graph 
	 * consisting of the original input points, displayed as a point array.
	 *
	 * @param numPoints The number of input points.
	 * @param pointCoord The coordinates of the points, in both the input and output color space.
	 *	@param numEqual The number of pixels that share the given vertex coordinates in 3D.
	 * @param pointColor The color of the points.
	 * @param pixelIndex For each point, gives the indices of the image pixels that belong to the point.
	 * @param colorSpace3D The input and output color space.
	 * @return The branch groups that represent the visualization. BranchGroup[0] is the pickable point array,
	 * while BranchGroup[1] is the non-selectable part of the visualization. Both branch groups
	 * may be null.        
	 */
	public BranchGroup[] createSceneObject(int numPoints, float pointCoord[][], int numEqual[], byte pointColor[], int pixelIndex[][], ColorSpace3D colorSpace3D[], ImageModule imageManager, View3D view3D) {
		int i;
		PointArray pointArray = new PointArray(numPoints,
				GeometryArray.COORDINATES | GeometryArray.COLOR_3);

		nObjects = numPoints;
		MovingObject shape3D = new MovingObject();

		if (numEqual != null) {
			view3D.getImageGamuts().add(shape3D);
		}
		
		Point3f vertexData[] = new Point3f[numPoints];
		Color3b colorData[] = new Color3b[numPoints];

		for (i = 0; i < numPoints; i++) {
			vertexData[i] = new Point3f(pointCoord[1][3 * i], pointCoord[1][3 * i + 1],
					pointCoord[1][3 * i + 2]);
			colorData[i] = new Color3b(pointColor[3 * i], pointColor[3 * i + 1],
					pointColor[3 * i + 2]);
		}
		pointArray.setCoordinates(0, vertexData);
		pointArray.setColors(0, colorData);
		app = StandardAppearance.getChangeablePointAppearance(pointSize);
		shape3D.setAppearance(app);
		shape3D.setGeometry(pointArray);
		
		BranchGroup branchGroup[];

		if (numEqual == null) {
			branchGroup = new BranchGroup[2];
			branchGroup[1] = new BranchGroup();
			branchGroup[1].addChild(shape3D);
		} else {
			BasicObject3D objects[] = new BasicObject3D[nObjects];

			for (i = 0; i < nObjects; i++) {
				objects[i] = new BasicObject3D(1,
						new float[] {pointCoord[1][i * 3], pointCoord[1][i * 3 + 1],
					pointCoord[1][i * 3 + 2]},
						new byte[] {pointColor[i * 3], pointColor[i * 3 + 1],
					pointColor[i * 3 + 2]}, i, pointArray);
				objects[i].addColor(i);
			}
			setCapability(pointArray);
			shape3D.setCapability(Shape3D.ALLOW_GEOMETRY_READ);
			shape3D.setStart(0);
			shape3D.setGeometryArrays(new GeometryArray[] {pointArray});
			shape3D.setObjects(objects);
			shape3D.setColors(pointCoord[1]);
			shape3D.setImage2D(imageManager.getImage2D());
			shape3D.setColorSpace(colorSpace3D[1]);
			branchGroup = new BranchGroup[1];
			branchGroup[0] = new BranchGroup();
			branchGroup[0].addChild(shape3D);
		}
		updateParameterPanel();
		return branchGroup;
	}

	/**
	 * Returns the name (short description) of the visualization.
	 * @return The name of the visualization. (Standard)
	 */	
	public String getName() {
		return "Standard";
	}
	
	/**
	 * The priority of the visualization.
	 * @return An int representing the priority of the visualization (lower => higher priority).
	 */
	public int getPriority() {
		return 2;
	}
	
	/**
	 * Creates the parameter panel for this visualization.
	 */
	private void createParameterPanel() {
		parameterPanel = new JPanel(new BorderLayout());
		parameterPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createRaisedBevelBorder(),
				"Standard"));
		JPanel panel1 = new JPanel(new GridLayout(0, 2));

		edtObjectCount = new JLabel();
		edtObjectCount.setHorizontalAlignment(JLabel.RIGHT);
		edtObjectCount.setBorder(BorderFactory.createEtchedBorder());
		edtObjectCount.setText(Integer.toString(nObjects));
		cmbPointSize = new JComboBox(new String[] {"1", "2", "3", "4", "5", "6", "8",
			"10", "12", "15", "20"});
		pointSize = 10.0f;
		cmbPointSize.setSelectedIndex(7);
		cmbPointSize.addItemListener(new ItemListener() {
			public void itemStateChanged(ItemEvent ie) {
				pointSize = Integer.parseInt(cmbPointSize.getSelectedItem().toString());
				if (app != null) {
					app.setPointAttributes(new PointAttributes(pointSize, true));
				}
			}
		});
		panel1.add(new JLabel("Object count: "));
		panel1.add(edtObjectCount);
		panel1.add(new JLabel("Point size: "));
		panel1.add(cmbPointSize);
		parameterPanel.add(panel1, BorderLayout.NORTH);
		parameterPanel.add(new JPanel(), BorderLayout.CENTER);
	}	
	
	/**
	 * Updates the parameter panel so that it displays the correct number of objects.
	 */
	public void updateParameterPanel() {
		edtObjectCount.setText(Integer.toString(nObjects));
	}

	/**
	 * Returns the panel that allows the user to modify visualization parameters.
	 * @return The parameter panel.
	 */	
	public Component getParameterPanel() {
		if (parameterPanel == null) {
			createParameterPanel();
		}
		return parameterPanel;
	}

	/**
	 * Checks that the parameters are within normal range.
	 * @return True if everyting checks out, false if there are errors.
	 */	
	public boolean checkParams() {
		return true;
	}
	
	/**
	 * Releases any memory that is no longer needed.
	 */
	public void clearMemory() {}
	
	/**
	 * Deinitializes the visualization, and frees any memory that is no longer needed.
	 */
	public void deInitialize() {
		this.app = null;
		this.cmbPointSize = null;
		this.edtObjectCount = null;
		this.parameterPanel = null;
	}
}
