/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations.appearance;


import java.io.*;
import javax.media.j3d.*;
import javax.vecmath.*;


/**
 * StandardAppearance provides methods for creating standard appearance
 * objects for Java3D geometry.
 */
public abstract class StandardAppearance {
	
	/** The depth buffer bit depth */
	private static int depthSize;
	
	/** Depth buffer offset. */
	private static float offsetAt16, offsetAt24, offsetAt32;
	
	/** Update depth buffer offsets from configuration file. */
	public static void updateDepthOffsetFromConfig() {
		try {
			BufferedReader br = new BufferedReader(new FileReader("config.ini"));
			offsetAt16 = Float.parseFloat(br.readLine());
			offsetAt24 = Float.parseFloat(br.readLine());
			offsetAt32 = Float.parseFloat(br.readLine());
			// javax.swing.JOptionPane.showMessageDialog(null, "16: " + offsetAt16+", 24: " + offsetAt24+", 32: " + offsetAt32);
		} catch (Exception e) {
			offsetAt16 = -5.0f;
			offsetAt24 = -150.0f;
			offsetAt32 = -1536.0f;
		}
	}
	static {
		updateDepthOffsetFromConfig();
	}
	
	/**
	 * Sets the depth buffer bit depth.
	 * @param d The number of bits in the depth buffer.
	 */
	public static void setDepth(int d) {
		depthSize = d;
	}
	
	/**
	 * Returns a line appearance.
	 * @return An appearance that displays objects as lines/wireframe.
	 */
	public static Appearance getLineAppearance() {
		return getLineAppearance(new Color3f(1.0f, 1.0f, 1.0f), false, true);
	}
	
	/**
	 * Returns a line appearance.
	 * @param color The line color.
	 * @return An appearance that displays objects as lines/wireframe.
	 */
	public static Appearance getLineAppearance(Color3f color) {
		if (color == null) {
			color = new Color3f(1.0f, 1.0f, 1.0f);
		}
		return getLineAppearance(color, false, true);
	}
	
	/**
	 * Returns a line appearance.
	 * @param color The line color.
	 * @param polygonOffset The polygon depth offset.
	 * @param useVertexColors Specify the color of each vertex.
	 * @return An appearance that displays objects as lines/wireframe.
	 */
	public static Appearance getLineAppearance(Color3f color, boolean polygonOffset, boolean useVertexColors) {		
		Appearance appearance = new Appearance();
		
		LineAttributes lineAttributes = new LineAttributes(
				0.5f,															// line width
				LineAttributes.PATTERN_SOLID,			// line pattern
				true															// line antialiasing
				);
		ColoringAttributes coloringAttributes = new ColoringAttributes(
				color.x,																// red
				color.y,																// green
				color.z,																// blue
				ColoringAttributes.SHADE_GOURAUD				// shade model
				);
		/*
		PointAttributes pointAttributes = new PointAttributes(
				1.0f,															// point size
				true															// point antialiasing
				);
				*/
		
		PolygonAttributes polygonAttributes = new PolygonAttributes(
				PolygonAttributes.POLYGON_LINE,		// polygon fill mode
				PolygonAttributes.CULL_NONE,			// polygon cull mode
				0.0f															// polygon offset
				);

		if (polygonOffset) {
			float d = 0.0f;
			if (depthSize == 32) {
				d = offsetAt32;
			} else if (depthSize == 24) {
				d = offsetAt24;
			} else if (depthSize == 16) {
				d = offsetAt16;
			} else if (depthSize<16) {
				d = offsetAt16 * (float) depthSize / 16.0f;
			} else if (depthSize < 24) {
				float w1 = (24.0f - depthSize)/8.0f;
				float w2 = (depthSize - 16.0f)/8.0f;
				d = offsetAt16 * w1 + offsetAt24 * w2;
			} else {
				float w1 = (32.0f - depthSize)/8.0f;
				float w2 = (depthSize - 24.0f)/8.0f;
				d = offsetAt24 * w1 + offsetAt32 * w2;
			}
			polygonAttributes.setPolygonOffset(d);
			polygonAttributes.setPolygonOffsetFactor(0.0f);
		}
		
		/*
		Material material = new Material(
				color,														// ambient
				new Color3f(0.0f, 0.0f, 0.0f),		// emissive
				color,														// diffuse
				new Color3f(1.0f, 1.0f, 1.0f),		// specular
				64.0f															// shininess
				);
		*/
		
		RenderingAttributes renderingAttributes = new RenderingAttributes(
				true, 														// boolean depthBufferEnable, 
				true,															// boolean depthBufferWriteEnable, 
				0.0f,															// float alphaTestValue,
				RenderingAttributes.ALWAYS,				// int alphaTestFunction,
				true,															// boolean visible,
				!useVertexColors,									// boolean ignoreVertexColors, 
				false,														// boolean rasterOpEnable,
				RenderingAttributes.ROP_COPY			// int rasterOp
				);

		appearance.setLineAttributes(lineAttributes);
		appearance.setColoringAttributes(coloringAttributes);
		//appearance.setPointAttributes(pointAttributes);
		appearance.setPolygonAttributes(polygonAttributes);
		appearance.setRenderingAttributes(renderingAttributes);
		//appearance.setMaterial(material);
		return appearance;
	}
	
	/**
	 * Returns a point appearance. The point size can be changed.
	 * @param pointSize The point size.
	 * @return An appearance that displays objects as points.
	 */
	public static Appearance getChangeablePointAppearance(float pointSize) {
		Appearance appearance = new Appearance();
		
		LineAttributes lineAttributes = new LineAttributes(
				0.5f,															// line width
				LineAttributes.PATTERN_SOLID,			// line pattern
				true															// line antialiasing
				);
		ColoringAttributes coloringAttributes = new ColoringAttributes(
				1.0f,															// red
				1.0f,															// green
				1.0f,															// blue
				ColoringAttributes.SHADE_FLAT			// shade model
				);
		
		PointAttributes pointAttributes = new PointAttributes(
				pointSize,												// point size
				true															// point antialiasing
				);
		
		PolygonAttributes polygonAttributes = new PolygonAttributes(
				PolygonAttributes.POLYGON_POINT,	// polygon fill mode
				PolygonAttributes.CULL_NONE,			// polygon cull mode
				0.0f															// polygon offset
				);
		/*
		Material material = new Material(
				new Color3f(0.0f, 0.0f, 0.0f),		// ambient
				new Color3f(0.0f, 0.0f, 0.0f),		// emissive
				new Color3f(0.0f, 0.0f, 0.0f),		// diffuse
				new Color3f(1.0f, 1.0f, 1.0f),		// specular
				64.0f															// shininess
				);
		*/
		
		appearance.setLineAttributes(lineAttributes);
		appearance.setColoringAttributes(coloringAttributes);
		appearance.setPointAttributes(pointAttributes);
		appearance.setPolygonAttributes(polygonAttributes);
		//appearance.setMaterial(material);
		appearance.setCapability(Appearance.ALLOW_POINT_ATTRIBUTES_WRITE);
		return appearance;
	}
	
	/**
	 * Returns a point appearance.
	 * @param pointSize The point size.
	 * @return An appearance that displays objects as points.
	 */
	public static Appearance getPointAppearance(float pointSize) {
		Appearance appearance = new Appearance();
		
		LineAttributes lineAttributes = new LineAttributes(
				0.5f,														// line width
				LineAttributes.PATTERN_SOLID,		// line pattern
				true														// line antialiasing
				);
		ColoringAttributes coloringAttributes = new ColoringAttributes(
				1.0f,													// red
				1.0f,													// green
				1.0f,													// blue
				ColoringAttributes.SHADE_FLAT	// shade model
				);
		
		PointAttributes pointAttributes = new PointAttributes(
				pointSize,	// point size
				true				// point antialiasing
				);
		
		PolygonAttributes polygonAttributes = new PolygonAttributes(
				PolygonAttributes.POLYGON_POINT,	// polygon fill mode
				PolygonAttributes.CULL_NONE,			// polygon cull mode
				0.0f															// polygon offset
				);
		
		/*
		Material material = new Material(
				new Color3f(0.0f, 0.0f, 0.0f),		// ambient
				new Color3f(0.0f, 0.0f, 0.0f),		// emissive
				new Color3f(0.0f, 0.0f, 0.0f),		// diffuse
				new Color3f(1.0f, 1.0f, 1.0f),		// specular
				64.0f															// shininess
				);
		*/
		appearance.setLineAttributes(lineAttributes);
		appearance.setColoringAttributes(coloringAttributes);
		appearance.setPointAttributes(pointAttributes);
		appearance.setPolygonAttributes(polygonAttributes);
		//appearance.setMaterial(material);
		
		return appearance;
	}
	
	/**
	 * Returns a filled triangle appearance.
	 * @return An appearance that displays objects as filled triangles.
	 */
	public static Appearance getFilledTriangleApperance() {
		Appearance appearance = new Appearance();
		
		LineAttributes lineAttributes = new LineAttributes(
				0.5f,															// line width
				LineAttributes.PATTERN_SOLID,			// line pattern
				true															// line antialiasing
				);
		ColoringAttributes coloringAttributes = new ColoringAttributes(
				1.0f,															// red
				1.0f,															// green
				1.0f,															// blue
				ColoringAttributes.SHADE_GOURAUD	// shade model
				);
		
		PointAttributes pointAttributes = new PointAttributes(
				1.0f,															// point size
				true															// point antialiasing
				);
		
		PolygonAttributes polygonAttributes = new PolygonAttributes(
				PolygonAttributes.POLYGON_FILL,		// polygon fill mode
				PolygonAttributes.CULL_NONE,			// polygon cull mode
				0.0f															// polygon offset
				);
		
		/*
		Material material = new Material(
				new Color3f(0.0f, 0.0f, 0.0f),		// ambient
				new Color3f(0.0f, 0.0f, 0.0f),		// emissive
				new Color3f(0.0f, 0.0f, 0.0f),		// diffuse
				new Color3f(1.0f, 1.0f, 1.0f),		// specular
				64.0f															// shininess
				);
		*/
		
		appearance.setLineAttributes(lineAttributes);
		appearance.setColoringAttributes(coloringAttributes);
		appearance.setPointAttributes(pointAttributes);
		appearance.setPolygonAttributes(polygonAttributes);
		//appearance.setMaterial(material);
		return appearance;
	}
	
	/**
	 * Returns a filled triangle appearance.
	 * @param useVertexColors Specify the color of each vertex.
	 * @param color The triangle color.
	 * @param useTransparency Apply transparency to the object.
	 * @param transparency The level of transparency (0.0-1.0).
	 * @return An appearance that displays objects as filled triangles.
	 */
	public static Appearance getFilledTriangleApperance(boolean useVertexColors, Color3f color, boolean useTransparency, float transparency) {
		Appearance appearance = new Appearance();
		
		LineAttributes lineAttributes = new LineAttributes(
				0.5f,															// line width
				LineAttributes.PATTERN_SOLID,			// line pattern
				true															// line antialiasing
				);
		ColoringAttributes coloringAttributes = new ColoringAttributes(
				color.x,															// red
				color.y,															// green
				color.z,															// blue
				ColoringAttributes.SHADE_GOURAUD			// shade model
				);
		
		PointAttributes pointAttributes = new PointAttributes(
				1.0f,															// point size
				true															// point antialiasing
				);
		
		PolygonAttributes polygonAttributes = new PolygonAttributes(
				PolygonAttributes.POLYGON_FILL,		// polygon fill mode
				PolygonAttributes.CULL_NONE,			// polygon cull mode
				0.0f															// polygon offset
				);
		
		RenderingAttributes renderingAttributes = new RenderingAttributes(
				true, 														// boolean depthBufferEnable, 
				true,															// boolean depthBufferWriteEnable, 
				0.0f,															// float alphaTestValue,
				RenderingAttributes.ALWAYS,				// int alphaTestFunction,
				true,															// boolean visible,
				!useVertexColors,									// boolean ignoreVertexColors, 
				false,														// boolean rasterOpEnable,
				RenderingAttributes.ROP_COPY			// int rasterOp
				);
		
		TransparencyAttributes transparencyAttributes = new TransparencyAttributes(
				TransparencyAttributes.SCREEN_DOOR,		// transparency type
				transparency);
		
		/*
		Material material = new Material(
				color,														// ambient
				new Color3f(0.0f, 0.0f, 0.0f),		// emissive
				color,														// diffuse
				new Color3f(1.0f, 1.0f, 1.0f),		// specular
				64.0f															// shininess
				);
		*/
		
		appearance.setLineAttributes(lineAttributes);
		appearance.setColoringAttributes(coloringAttributes);
		appearance.setPointAttributes(pointAttributes);
		appearance.setPolygonAttributes(polygonAttributes);
		appearance.setRenderingAttributes(renderingAttributes);
		
		if (useTransparency) {
			appearance.setTransparencyAttributes(transparencyAttributes);
		}
		//appearance.setMaterial(material);
		return appearance;
	}
	
	/**
	 * Returns an appearance that makes an object invisible.
	 * @return An appearance that makes an object invisible.
	 */
	public static Appearance getInvisibleAppearance() {
		Appearance appearance = new Appearance();
		RenderingAttributes renderingAttributes = new RenderingAttributes(
				true,
				true,
				0.0f,
				RenderingAttributes.NEVER);

		appearance.setRenderingAttributes(renderingAttributes);
		return appearance;
	}
}
