/*
 ICC3D - Interactive Color Correction in 3 Dimensions.
 Copyright (C) 2002-2003 Gjvik University College.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


package icc3dpackage.visualizations.appearance;


import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import icc3dpackage.*;


/**
 * TriangleAppearance determines the wireframe/solid appearance of 3D triangles.
 */
public class TriangleAppearance extends JPanel {
	
	/**
	 * Apply transparency to the object.
	 */
	private JCheckBox useTransparency;
	
	/**
	 * The level of transparency.
	 */
	private JComboBox transparencyLevel;
	
	/**
	 * The triangle display mode.
	 */
	private JComboBox triangleMode;
	
	/**
	 * Apply a constant color to the solid triangles.
	 */
	private JCheckBox fixedSolidColor;
	
	/**
	 * Choose the solid triangle color.
	 */
	private JButton		solidColor;
	
	/**
	 * Apply a constant color to the wireframe.
	 */
	private JCheckBox fixedWireframeColor;
	
	/**
	 * Choose the wireframe color.
	 */
	private JButton		wireframeColor;
	
	/**
	 * Constructs a TriangleAppearance object.
	 */
	public TriangleAppearance() {
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		triangleMode = new JComboBox(new String[] {"Solid+Wireframe", "Solid",
			"Wireframe", "Points"});
		
		JPanel trianglePanel = new JPanel(new GridLayout(0, 2));

		trianglePanel.setBorder(BorderFactory.createEtchedBorder());
		trianglePanel.add(new JLabel("Display"));
		trianglePanel.add(triangleMode);
		
		fixedSolidColor = new JCheckBox();
		solidColor = new JButton();
		solidColor.setBackground(Color.red);
		solidColor.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Color retColor = JColorChooser.showDialog(ICC3D.getDialogParent(),
						"Choose a color for the solid body", solidColor.getBackground());

				if (retColor != null) {
					solidColor.setBackground(retColor);
				}
			}
		});
		JPanel solidColorPanel = new JPanel(new BorderLayout());
		solidColorPanel.add(fixedSolidColor, BorderLayout.WEST);
		solidColorPanel.add(solidColor, BorderLayout.CENTER);
		
		trianglePanel.add(new JLabel("Solid color"));
		trianglePanel.add(solidColorPanel);

		fixedWireframeColor = new JCheckBox();
		fixedWireframeColor.setSelected(true);
		wireframeColor = new JButton();
		wireframeColor.setBackground(Color.black);
		wireframeColor.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				Color retColor = JColorChooser.showDialog(ICC3D.getDialogParent(),
						"Choose a color for the wireframe", wireframeColor.getBackground());

				if (retColor != null) {
					wireframeColor.setBackground(retColor);
				}
			}
		});
		JPanel wireframeColorPanel = new JPanel(new BorderLayout());

		wireframeColorPanel.add(fixedWireframeColor, BorderLayout.WEST);
		wireframeColorPanel.add(wireframeColor, BorderLayout.CENTER);
		trianglePanel.add(new JLabel("Wireframe color"));
		trianglePanel.add(wireframeColorPanel);
		useTransparency = new JCheckBox();
		useTransparency.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				transparencyLevel.setEnabled(transparencyLevel.isEnabled() ? false : true);
			}
		});
		
		String levels[] = {"0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.7", "0.8",
			"0.9"};

		transparencyLevel = new JComboBox(levels);
		transparencyLevel.setEnabled(false);
		trianglePanel.add(new JLabel("Transparency"));
		trianglePanel.add(useTransparency);
		trianglePanel.add(new JLabel("Level"));
		trianglePanel.add(transparencyLevel);
		add(trianglePanel);
	}
	
	/**
	 * Returns an array of shapes with the given geometry, and appeance as specified by
	 * the current parameter panel.
	 * @param triangles The triangle geometry.
	 * @return An array of Shape3D objects.
	 */
	public Shape3D[] getShapes(Geometry triangles) {
		Shape3D result[] = null;

		switch (triangleMode.getSelectedIndex()) {
		case 0:
			result = new Shape3D[2];
			result[0] = getSolid(triangles);
			result[1] = getWireframe(triangles, true);
			break;

		case 1:
			result = new Shape3D[1];
			result[0] = getSolid(triangles);
			break;

		case 2:
			result = new Shape3D[1];
			result[0] = getWireframe(triangles, false);
			break;

		case 3:
			result = new Shape3D[1];
			result[0] = getPoint(triangles);
			break;

		default:
			result = new Shape3D[0];
			break;
		}
		return result;
	}
	
	/**
	 * Returns a shape that contains filled triangles.
	 * @param geometry The triangle geometry.
	 * @return A shape that contains the geometry.
	 */
	private Shape3D getSolid(Geometry geometry) {
		return new Shape3D(geometry,
				StandardAppearance.getFilledTriangleApperance(!fixedSolidColor.isSelected(),
				new Color3f(solidColor.getBackground()), useTransparency.isSelected(),
				Float.parseFloat(transparencyLevel.getSelectedItem().toString())));
	}
	
	/**
	 * Returns a shape that displays the geometry as points.
	 * @param geometry The geometry.
	 * @return A point shape.
	 */
	private Shape3D getPoint(Geometry geometry) {
		return new Shape3D(geometry, StandardAppearance.getPointAppearance(10.0f));
	}
	
	/**
	 * Returns a shape that displays the geometry as a wireframe.
	 * @param geometry The geometry.
	 * @param offset Apply depth offset to the wireframe to display it in fron of
	 * the filled triangles.
	 * @return A wireframe shape.
	 */
	private Shape3D getWireframe(Geometry geometry, boolean offset) {
		float fOffset = 0.0f;

		return new Shape3D(geometry,
				StandardAppearance.getLineAppearance(new Color3f(wireframeColor.getBackground()),
				offset, !fixedWireframeColor.isSelected()));
	}
}
